/*
  This software is copyright Kong Inc. and its licensors.
  Use of the software is subject to the agreement between your organization
  and Kong Inc. If there is no such agreement, use is governed by and
  subject to the terms of the Kong Master Software License Agreement found
  at https://konghq.com/enterprisesoftwarelicense/.
*/

(() => {
  // node_modules/monaco-editor/esm/vs/base/common/errors.js
  var ErrorHandler = class {
    constructor() {
      this.listeners = [];
      this.unexpectedErrorHandler = function(e) {
        setTimeout(() => {
          if (e.stack) {
            throw new Error(e.message + "\n\n" + e.stack);
          }
          throw e;
        }, 0);
      };
    }
    emit(e) {
      this.listeners.forEach((listener) => {
        listener(e);
      });
    }
    onUnexpectedError(e) {
      this.unexpectedErrorHandler(e);
      this.emit(e);
    }
    onUnexpectedExternalError(e) {
      this.unexpectedErrorHandler(e);
    }
  };
  var errorHandler = new ErrorHandler();
  function onUnexpectedError(e) {
    if (!isPromiseCanceledError(e)) {
      errorHandler.onUnexpectedError(e);
    }
    return void 0;
  }
  function transformErrorForSerialization(error) {
    if (error instanceof Error) {
      let { name, message } = error;
      const stack = error.stacktrace || error.stack;
      return {
        $isError: true,
        name,
        message,
        stack
      };
    }
    return error;
  }
  var canceledName = "Canceled";
  function isPromiseCanceledError(error) {
    return error instanceof Error && error.name === canceledName && error.message === canceledName;
  }

  // node_modules/monaco-editor/esm/vs/base/common/iterator.js
  var Iterable;
  (function(Iterable2) {
    function is(thing) {
      return thing && typeof thing === "object" && typeof thing[Symbol.iterator] === "function";
    }
    Iterable2.is = is;
    const _empty3 = Object.freeze([]);
    function empty() {
      return _empty3;
    }
    Iterable2.empty = empty;
    function* single(element) {
      yield element;
    }
    Iterable2.single = single;
    function from(iterable) {
      return iterable || _empty3;
    }
    Iterable2.from = from;
    function first(iterable) {
      return iterable[Symbol.iterator]().next().value;
    }
    Iterable2.first = first;
    function some(iterable, predicate) {
      for (const element of iterable) {
        if (predicate(element)) {
          return true;
        }
      }
      return false;
    }
    Iterable2.some = some;
    function* filter(iterable, predicate) {
      for (const element of iterable) {
        if (predicate(element)) {
          yield element;
        }
      }
    }
    Iterable2.filter = filter;
    function* map(iterable, fn) {
      for (const element of iterable) {
        yield fn(element);
      }
    }
    Iterable2.map = map;
    function* concat(...iterables) {
      for (const iterable of iterables) {
        for (const element of iterable) {
          yield element;
        }
      }
    }
    Iterable2.concat = concat;
    function consume(iterable, atMost = Number.POSITIVE_INFINITY) {
      const consumed = [];
      if (atMost === 0) {
        return [consumed, iterable];
      }
      const iterator = iterable[Symbol.iterator]();
      for (let i = 0; i < atMost; i++) {
        const next = iterator.next();
        if (next.done) {
          return [consumed, Iterable2.empty()];
        }
        consumed.push(next.value);
      }
      return [consumed, { [Symbol.iterator]() {
        return iterator;
      } }];
    }
    Iterable2.consume = consume;
  })(Iterable || (Iterable = {}));

  // node_modules/monaco-editor/esm/vs/base/common/lifecycle.js
  var TRACK_DISPOSABLES = false;
  var __is_disposable_tracked__ = "__is_disposable_tracked__";
  function markTracked(x) {
    if (!TRACK_DISPOSABLES) {
      return;
    }
    if (x && x !== Disposable.None) {
      try {
        x[__is_disposable_tracked__] = true;
      } catch (_a3) {
      }
    }
  }
  function trackDisposable(x) {
    if (!TRACK_DISPOSABLES) {
      return x;
    }
    const stack = new Error("Potentially leaked disposable").stack;
    setTimeout(() => {
      if (!x[__is_disposable_tracked__]) {
        console.log(stack);
      }
    }, 3e3);
    return x;
  }
  var MultiDisposeError = class extends Error {
    constructor(errors) {
      super(`Encounter errors while disposing of store. Errors: [${errors.join(", ")}]`);
      this.errors = errors;
    }
  };
  function dispose(arg) {
    if (Iterable.is(arg)) {
      let errors = [];
      for (const d of arg) {
        if (d) {
          markTracked(d);
          try {
            d.dispose();
          } catch (e) {
            errors.push(e);
          }
        }
      }
      if (errors.length === 1) {
        throw errors[0];
      } else if (errors.length > 1) {
        throw new MultiDisposeError(errors);
      }
      return Array.isArray(arg) ? [] : arg;
    } else if (arg) {
      markTracked(arg);
      arg.dispose();
      return arg;
    }
  }
  function combinedDisposable(...disposables) {
    disposables.forEach(markTracked);
    return trackDisposable({ dispose: () => dispose(disposables) });
  }
  var DisposableStore = class {
    constructor() {
      this._toDispose = /* @__PURE__ */ new Set();
      this._isDisposed = false;
    }
    dispose() {
      if (this._isDisposed) {
        return;
      }
      markTracked(this);
      this._isDisposed = true;
      this.clear();
    }
    clear() {
      try {
        dispose(this._toDispose.values());
      } finally {
        this._toDispose.clear();
      }
    }
    add(t) {
      if (!t) {
        return t;
      }
      if (t === this) {
        throw new Error("Cannot register a disposable on itself!");
      }
      markTracked(t);
      if (this._isDisposed) {
        if (!DisposableStore.DISABLE_DISPOSED_WARNING) {
          console.warn(new Error("Trying to add a disposable to a DisposableStore that has already been disposed of. The added object will be leaked!").stack);
        }
      } else {
        this._toDispose.add(t);
      }
      return t;
    }
  };
  DisposableStore.DISABLE_DISPOSED_WARNING = false;
  var Disposable = class {
    constructor() {
      this._store = new DisposableStore();
      trackDisposable(this);
    }
    dispose() {
      markTracked(this);
      this._store.dispose();
    }
    _register(t) {
      if (t === this) {
        throw new Error("Cannot register a disposable on itself!");
      }
      return this._store.add(t);
    }
  };
  Disposable.None = Object.freeze({ dispose() {
  } });

  // node_modules/monaco-editor/esm/vs/base/common/platform.js
  var LANGUAGE_DEFAULT = "en";
  var _isWindows = false;
  var _isMacintosh = false;
  var _isLinux = false;
  var _isNative = false;
  var _isWeb = false;
  var _isIOS = false;
  var _locale = void 0;
  var _language = LANGUAGE_DEFAULT;
  var _translationsConfigFile = void 0;
  var _userAgent = void 0;
  var isElectronRenderer = typeof process !== "undefined" && typeof process.versions !== "undefined" && typeof process.versions.electron !== "undefined" && process.type === "renderer";
  if (typeof navigator === "object" && !isElectronRenderer) {
    _userAgent = navigator.userAgent;
    _isWindows = _userAgent.indexOf("Windows") >= 0;
    _isMacintosh = _userAgent.indexOf("Macintosh") >= 0;
    _isIOS = (_userAgent.indexOf("Macintosh") >= 0 || _userAgent.indexOf("iPad") >= 0 || _userAgent.indexOf("iPhone") >= 0) && !!navigator.maxTouchPoints && navigator.maxTouchPoints > 0;
    _isLinux = _userAgent.indexOf("Linux") >= 0;
    _isWeb = true;
    _locale = navigator.language;
    _language = _locale;
  } else if (typeof process === "object") {
    _isWindows = process.platform === "win32";
    _isMacintosh = process.platform === "darwin";
    _isLinux = process.platform === "linux";
    _locale = LANGUAGE_DEFAULT;
    _language = LANGUAGE_DEFAULT;
    const rawNlsConfig = process.env["VSCODE_NLS_CONFIG"];
    if (rawNlsConfig) {
      try {
        const nlsConfig = JSON.parse(rawNlsConfig);
        const resolved = nlsConfig.availableLanguages["*"];
        _locale = nlsConfig.locale;
        _language = resolved ? resolved : LANGUAGE_DEFAULT;
        _translationsConfigFile = nlsConfig._translationsConfigFile;
      } catch (e) {
      }
    }
    _isNative = true;
  }
  var _platform = 0;
  if (_isMacintosh) {
    _platform = 1;
  } else if (_isWindows) {
    _platform = 3;
  } else if (_isLinux) {
    _platform = 2;
  }
  var isWindows = _isWindows;
  var isMacintosh = _isMacintosh;
  var _globals = typeof self === "object" ? self : typeof global === "object" ? global : {};
  var globals = _globals;
  var setImmediate = function defineSetImmediate() {
    if (globals.setImmediate) {
      return globals.setImmediate.bind(globals);
    }
    if (typeof globals.postMessage === "function" && !globals.importScripts) {
      let pending = [];
      globals.addEventListener("message", (e) => {
        if (e.data && e.data.vscodeSetImmediateId) {
          for (let i = 0, len = pending.length; i < len; i++) {
            const candidate = pending[i];
            if (candidate.id === e.data.vscodeSetImmediateId) {
              pending.splice(i, 1);
              candidate.callback();
              return;
            }
          }
        }
      });
      let lastId = 0;
      return (callback) => {
        const myId = ++lastId;
        pending.push({
          id: myId,
          callback
        });
        globals.postMessage({ vscodeSetImmediateId: myId }, "*");
      };
    }
    if (typeof process !== "undefined" && typeof process.nextTick === "function") {
      return process.nextTick.bind(process);
    }
    const _promise = Promise.resolve();
    return (callback) => _promise.then(callback);
  }();

  // node_modules/monaco-editor/esm/vs/base/common/types.js
  function getAllPropertyNames(obj) {
    let res = [];
    let proto = Object.getPrototypeOf(obj);
    while (Object.prototype !== proto) {
      res = res.concat(Object.getOwnPropertyNames(proto));
      proto = Object.getPrototypeOf(proto);
    }
    return res;
  }
  function getAllMethodNames(obj) {
    const methods = [];
    for (const prop of getAllPropertyNames(obj)) {
      if (typeof obj[prop] === "function") {
        methods.push(prop);
      }
    }
    return methods;
  }
  function createProxyObject(methodNames, invoke) {
    const createProxyMethod = (method) => {
      return function() {
        const args = Array.prototype.slice.call(arguments, 0);
        return invoke(method, args);
      };
    };
    let result = {};
    for (const methodName of methodNames) {
      result[methodName] = createProxyMethod(methodName);
    }
    return result;
  }

  // node_modules/monaco-editor/esm/vs/base/common/worker/simpleWorker.js
  var INITIALIZE = "$initialize";
  var SimpleWorkerProtocol = class {
    constructor(handler) {
      this._workerId = -1;
      this._handler = handler;
      this._lastSentReq = 0;
      this._pendingReplies = /* @__PURE__ */ Object.create(null);
    }
    setWorkerId(workerId) {
      this._workerId = workerId;
    }
    sendMessage(method, args) {
      let req = String(++this._lastSentReq);
      return new Promise((resolve2, reject) => {
        this._pendingReplies[req] = {
          resolve: resolve2,
          reject
        };
        this._send({
          vsWorker: this._workerId,
          req,
          method,
          args
        });
      });
    }
    handleMessage(message) {
      if (!message || !message.vsWorker) {
        return;
      }
      if (this._workerId !== -1 && message.vsWorker !== this._workerId) {
        return;
      }
      this._handleMessage(message);
    }
    _handleMessage(msg) {
      if (msg.seq) {
        let replyMessage = msg;
        if (!this._pendingReplies[replyMessage.seq]) {
          console.warn("Got reply to unknown seq");
          return;
        }
        let reply = this._pendingReplies[replyMessage.seq];
        delete this._pendingReplies[replyMessage.seq];
        if (replyMessage.err) {
          let err = replyMessage.err;
          if (replyMessage.err.$isError) {
            err = new Error();
            err.name = replyMessage.err.name;
            err.message = replyMessage.err.message;
            err.stack = replyMessage.err.stack;
          }
          reply.reject(err);
          return;
        }
        reply.resolve(replyMessage.res);
        return;
      }
      let requestMessage = msg;
      let req = requestMessage.req;
      let result = this._handler.handleMessage(requestMessage.method, requestMessage.args);
      result.then((r) => {
        this._send({
          vsWorker: this._workerId,
          seq: req,
          res: r,
          err: void 0
        });
      }, (e) => {
        if (e.detail instanceof Error) {
          e.detail = transformErrorForSerialization(e.detail);
        }
        this._send({
          vsWorker: this._workerId,
          seq: req,
          res: void 0,
          err: transformErrorForSerialization(e)
        });
      });
    }
    _send(msg) {
      let transfer = [];
      if (msg.req) {
        const m = msg;
        for (let i = 0; i < m.args.length; i++) {
          if (m.args[i] instanceof ArrayBuffer) {
            transfer.push(m.args[i]);
          }
        }
      } else {
        const m = msg;
        if (m.res instanceof ArrayBuffer) {
          transfer.push(m.res);
        }
      }
      this._handler.sendMessage(msg, transfer);
    }
  };
  var SimpleWorkerServer = class {
    constructor(postMessage, requestHandlerFactory) {
      this._requestHandlerFactory = requestHandlerFactory;
      this._requestHandler = null;
      this._protocol = new SimpleWorkerProtocol({
        sendMessage: (msg, transfer) => {
          postMessage(msg, transfer);
        },
        handleMessage: (method, args) => this._handleMessage(method, args)
      });
    }
    onmessage(msg) {
      this._protocol.handleMessage(msg);
    }
    _handleMessage(method, args) {
      if (method === INITIALIZE) {
        return this.initialize(args[0], args[1], args[2], args[3]);
      }
      if (!this._requestHandler || typeof this._requestHandler[method] !== "function") {
        return Promise.reject(new Error("Missing requestHandler or method: " + method));
      }
      try {
        return Promise.resolve(this._requestHandler[method].apply(this._requestHandler, args));
      } catch (e) {
        return Promise.reject(e);
      }
    }
    initialize(workerId, loaderConfig, moduleId, hostMethods) {
      this._protocol.setWorkerId(workerId);
      const proxyMethodRequest = (method, args) => {
        return this._protocol.sendMessage(method, args);
      };
      const hostProxy = createProxyObject(hostMethods, proxyMethodRequest);
      if (this._requestHandlerFactory) {
        this._requestHandler = this._requestHandlerFactory(hostProxy);
        return Promise.resolve(getAllMethodNames(this._requestHandler));
      }
      if (loaderConfig) {
        if (typeof loaderConfig.baseUrl !== "undefined") {
          delete loaderConfig["baseUrl"];
        }
        if (typeof loaderConfig.paths !== "undefined") {
          if (typeof loaderConfig.paths.vs !== "undefined") {
            delete loaderConfig.paths["vs"];
          }
        }
        loaderConfig.catchError = true;
        self.require.config(loaderConfig);
      }
      return new Promise((resolve2, reject) => {
        self.require([moduleId], (module) => {
          this._requestHandler = module.create(hostProxy);
          if (!this._requestHandler) {
            reject(new Error(`No RequestHandler!`));
            return;
          }
          resolve2(getAllMethodNames(this._requestHandler));
        }, reject);
      });
    }
  };

  // node_modules/monaco-editor/esm/vs/base/common/arrays.js
  function mergeSort(data, compare) {
    _sort(data, compare, 0, data.length - 1, []);
    return data;
  }
  function _merge(a2, compare, lo, mid, hi, aux) {
    let leftIdx = lo, rightIdx = mid + 1;
    for (let i = lo; i <= hi; i++) {
      aux[i] = a2[i];
    }
    for (let i = lo; i <= hi; i++) {
      if (leftIdx > mid) {
        a2[i] = aux[rightIdx++];
      } else if (rightIdx > hi) {
        a2[i] = aux[leftIdx++];
      } else if (compare(aux[rightIdx], aux[leftIdx]) < 0) {
        a2[i] = aux[rightIdx++];
      } else {
        a2[i] = aux[leftIdx++];
      }
    }
  }
  function _sort(a2, compare, lo, hi, aux) {
    if (hi <= lo) {
      return;
    }
    const mid = lo + (hi - lo) / 2 | 0;
    _sort(a2, compare, lo, mid, aux);
    _sort(a2, compare, mid + 1, hi, aux);
    if (compare(a2[mid], a2[mid + 1]) <= 0) {
      return;
    }
    _merge(a2, compare, lo, mid, hi, aux);
  }

  // node_modules/monaco-editor/esm/vs/base/common/diff/diffChange.js
  var DiffChange = class {
    constructor(originalStart, originalLength, modifiedStart, modifiedLength) {
      this.originalStart = originalStart;
      this.originalLength = originalLength;
      this.modifiedStart = modifiedStart;
      this.modifiedLength = modifiedLength;
    }
    getOriginalEnd() {
      return this.originalStart + this.originalLength;
    }
    getModifiedEnd() {
      return this.modifiedStart + this.modifiedLength;
    }
  };

  // node_modules/monaco-editor/esm/vs/base/common/strings.js
  function firstNonWhitespaceIndex(str) {
    for (let i = 0, len = str.length; i < len; i++) {
      const chCode = str.charCodeAt(i);
      if (chCode !== 32 && chCode !== 9) {
        return i;
      }
    }
    return -1;
  }
  function lastNonWhitespaceIndex(str, startIndex = str.length - 1) {
    for (let i = startIndex; i >= 0; i--) {
      const chCode = str.charCodeAt(i);
      if (chCode !== 32 && chCode !== 9) {
        return i;
      }
    }
    return -1;
  }
  function isHighSurrogate(charCode) {
    return 55296 <= charCode && charCode <= 56319;
  }
  function isLowSurrogate(charCode) {
    return 56320 <= charCode && charCode <= 57343;
  }
  function computeCodePoint(highSurrogate, lowSurrogate) {
    return (highSurrogate - 55296 << 10) + (lowSurrogate - 56320) + 65536;
  }
  var UTF8_BOM_CHARACTER = String.fromCharCode(65279);
  var GraphemeBreakTree = class {
    constructor() {
      this._data = getGraphemeBreakRawData();
    }
    static getInstance() {
      if (!GraphemeBreakTree._INSTANCE) {
        GraphemeBreakTree._INSTANCE = new GraphemeBreakTree();
      }
      return GraphemeBreakTree._INSTANCE;
    }
    getGraphemeBreakType(codePoint) {
      if (codePoint < 32) {
        if (codePoint === 10) {
          return 3;
        }
        if (codePoint === 13) {
          return 2;
        }
        return 4;
      }
      if (codePoint < 127) {
        return 0;
      }
      const data = this._data;
      const nodeCount = data.length / 3;
      let nodeIndex = 1;
      while (nodeIndex <= nodeCount) {
        if (codePoint < data[3 * nodeIndex]) {
          nodeIndex = 2 * nodeIndex;
        } else if (codePoint > data[3 * nodeIndex + 1]) {
          nodeIndex = 2 * nodeIndex + 1;
        } else {
          return data[3 * nodeIndex + 2];
        }
      }
      return 0;
    }
  };
  GraphemeBreakTree._INSTANCE = null;
  function getGraphemeBreakRawData() {
    return JSON.parse("[0,0,0,51592,51592,11,44424,44424,11,72251,72254,5,7150,7150,7,48008,48008,11,55176,55176,11,128420,128420,14,3276,3277,5,9979,9980,14,46216,46216,11,49800,49800,11,53384,53384,11,70726,70726,5,122915,122916,5,129320,129327,14,2558,2558,5,5906,5908,5,9762,9763,14,43360,43388,8,45320,45320,11,47112,47112,11,48904,48904,11,50696,50696,11,52488,52488,11,54280,54280,11,70082,70083,1,71350,71350,7,73111,73111,5,127892,127893,14,128726,128727,14,129473,129474,14,2027,2035,5,2901,2902,5,3784,3789,5,6754,6754,5,8418,8420,5,9877,9877,14,11088,11088,14,44008,44008,5,44872,44872,11,45768,45768,11,46664,46664,11,47560,47560,11,48456,48456,11,49352,49352,11,50248,50248,11,51144,51144,11,52040,52040,11,52936,52936,11,53832,53832,11,54728,54728,11,69811,69814,5,70459,70460,5,71096,71099,7,71998,71998,5,72874,72880,5,119149,119149,7,127374,127374,14,128335,128335,14,128482,128482,14,128765,128767,14,129399,129400,14,129680,129685,14,1476,1477,5,2377,2380,7,2759,2760,5,3137,3140,7,3458,3459,7,4153,4154,5,6432,6434,5,6978,6978,5,7675,7679,5,9723,9726,14,9823,9823,14,9919,9923,14,10035,10036,14,42736,42737,5,43596,43596,5,44200,44200,11,44648,44648,11,45096,45096,11,45544,45544,11,45992,45992,11,46440,46440,11,46888,46888,11,47336,47336,11,47784,47784,11,48232,48232,11,48680,48680,11,49128,49128,11,49576,49576,11,50024,50024,11,50472,50472,11,50920,50920,11,51368,51368,11,51816,51816,11,52264,52264,11,52712,52712,11,53160,53160,11,53608,53608,11,54056,54056,11,54504,54504,11,54952,54952,11,68108,68111,5,69933,69940,5,70197,70197,7,70498,70499,7,70845,70845,5,71229,71229,5,71727,71735,5,72154,72155,5,72344,72345,5,73023,73029,5,94095,94098,5,121403,121452,5,126981,127182,14,127538,127546,14,127990,127990,14,128391,128391,14,128445,128449,14,128500,128505,14,128752,128752,14,129160,129167,14,129356,129356,14,129432,129442,14,129648,129651,14,129751,131069,14,173,173,4,1757,1757,1,2274,2274,1,2494,2494,5,2641,2641,5,2876,2876,5,3014,3016,7,3262,3262,7,3393,3396,5,3570,3571,7,3968,3972,5,4228,4228,7,6086,6086,5,6679,6680,5,6912,6915,5,7080,7081,5,7380,7392,5,8252,8252,14,9096,9096,14,9748,9749,14,9784,9786,14,9833,9850,14,9890,9894,14,9938,9938,14,9999,9999,14,10085,10087,14,12349,12349,14,43136,43137,7,43454,43456,7,43755,43755,7,44088,44088,11,44312,44312,11,44536,44536,11,44760,44760,11,44984,44984,11,45208,45208,11,45432,45432,11,45656,45656,11,45880,45880,11,46104,46104,11,46328,46328,11,46552,46552,11,46776,46776,11,47000,47000,11,47224,47224,11,47448,47448,11,47672,47672,11,47896,47896,11,48120,48120,11,48344,48344,11,48568,48568,11,48792,48792,11,49016,49016,11,49240,49240,11,49464,49464,11,49688,49688,11,49912,49912,11,50136,50136,11,50360,50360,11,50584,50584,11,50808,50808,11,51032,51032,11,51256,51256,11,51480,51480,11,51704,51704,11,51928,51928,11,52152,52152,11,52376,52376,11,52600,52600,11,52824,52824,11,53048,53048,11,53272,53272,11,53496,53496,11,53720,53720,11,53944,53944,11,54168,54168,11,54392,54392,11,54616,54616,11,54840,54840,11,55064,55064,11,65438,65439,5,69633,69633,5,69837,69837,1,70018,70018,7,70188,70190,7,70368,70370,7,70465,70468,7,70712,70719,5,70835,70840,5,70850,70851,5,71132,71133,5,71340,71340,7,71458,71461,5,71985,71989,7,72002,72002,7,72193,72202,5,72281,72283,5,72766,72766,7,72885,72886,5,73104,73105,5,92912,92916,5,113824,113827,4,119173,119179,5,121505,121519,5,125136,125142,5,127279,127279,14,127489,127490,14,127570,127743,14,127900,127901,14,128254,128254,14,128369,128370,14,128400,128400,14,128425,128432,14,128468,128475,14,128489,128494,14,128715,128720,14,128745,128745,14,128759,128760,14,129004,129023,14,129296,129304,14,129340,129342,14,129388,129392,14,129404,129407,14,129454,129455,14,129485,129487,14,129659,129663,14,129719,129727,14,917536,917631,5,13,13,2,1160,1161,5,1564,1564,4,1807,1807,1,2085,2087,5,2363,2363,7,2402,2403,5,2507,2508,7,2622,2624,7,2691,2691,7,2786,2787,5,2881,2884,5,3006,3006,5,3072,3072,5,3170,3171,5,3267,3268,7,3330,3331,7,3406,3406,1,3538,3540,5,3655,3662,5,3897,3897,5,4038,4038,5,4184,4185,5,4352,4447,8,6068,6069,5,6155,6157,5,6448,6449,7,6742,6742,5,6783,6783,5,6966,6970,5,7042,7042,7,7143,7143,7,7212,7219,5,7412,7412,5,8206,8207,4,8294,8303,4,8596,8601,14,9410,9410,14,9742,9742,14,9757,9757,14,9770,9770,14,9794,9794,14,9828,9828,14,9855,9855,14,9882,9882,14,9900,9903,14,9929,9933,14,9963,9967,14,9987,9988,14,10006,10006,14,10062,10062,14,10175,10175,14,11744,11775,5,42607,42607,5,43043,43044,7,43263,43263,5,43444,43445,7,43569,43570,5,43698,43700,5,43766,43766,5,44032,44032,11,44144,44144,11,44256,44256,11,44368,44368,11,44480,44480,11,44592,44592,11,44704,44704,11,44816,44816,11,44928,44928,11,45040,45040,11,45152,45152,11,45264,45264,11,45376,45376,11,45488,45488,11,45600,45600,11,45712,45712,11,45824,45824,11,45936,45936,11,46048,46048,11,46160,46160,11,46272,46272,11,46384,46384,11,46496,46496,11,46608,46608,11,46720,46720,11,46832,46832,11,46944,46944,11,47056,47056,11,47168,47168,11,47280,47280,11,47392,47392,11,47504,47504,11,47616,47616,11,47728,47728,11,47840,47840,11,47952,47952,11,48064,48064,11,48176,48176,11,48288,48288,11,48400,48400,11,48512,48512,11,48624,48624,11,48736,48736,11,48848,48848,11,48960,48960,11,49072,49072,11,49184,49184,11,49296,49296,11,49408,49408,11,49520,49520,11,49632,49632,11,49744,49744,11,49856,49856,11,49968,49968,11,50080,50080,11,50192,50192,11,50304,50304,11,50416,50416,11,50528,50528,11,50640,50640,11,50752,50752,11,50864,50864,11,50976,50976,11,51088,51088,11,51200,51200,11,51312,51312,11,51424,51424,11,51536,51536,11,51648,51648,11,51760,51760,11,51872,51872,11,51984,51984,11,52096,52096,11,52208,52208,11,52320,52320,11,52432,52432,11,52544,52544,11,52656,52656,11,52768,52768,11,52880,52880,11,52992,52992,11,53104,53104,11,53216,53216,11,53328,53328,11,53440,53440,11,53552,53552,11,53664,53664,11,53776,53776,11,53888,53888,11,54000,54000,11,54112,54112,11,54224,54224,11,54336,54336,11,54448,54448,11,54560,54560,11,54672,54672,11,54784,54784,11,54896,54896,11,55008,55008,11,55120,55120,11,64286,64286,5,66272,66272,5,68900,68903,5,69762,69762,7,69817,69818,5,69927,69931,5,70003,70003,5,70070,70078,5,70094,70094,7,70194,70195,7,70206,70206,5,70400,70401,5,70463,70463,7,70475,70477,7,70512,70516,5,70722,70724,5,70832,70832,5,70842,70842,5,70847,70848,5,71088,71089,7,71102,71102,7,71219,71226,5,71231,71232,5,71342,71343,7,71453,71455,5,71463,71467,5,71737,71738,5,71995,71996,5,72000,72000,7,72145,72147,7,72160,72160,5,72249,72249,7,72273,72278,5,72330,72342,5,72752,72758,5,72850,72871,5,72882,72883,5,73018,73018,5,73031,73031,5,73109,73109,5,73461,73462,7,94031,94031,5,94192,94193,7,119142,119142,7,119155,119162,4,119362,119364,5,121476,121476,5,122888,122904,5,123184,123190,5,126976,126979,14,127184,127231,14,127344,127345,14,127405,127461,14,127514,127514,14,127561,127567,14,127778,127779,14,127896,127896,14,127985,127986,14,127995,127999,5,128326,128328,14,128360,128366,14,128378,128378,14,128394,128397,14,128405,128406,14,128422,128423,14,128435,128443,14,128453,128464,14,128479,128480,14,128484,128487,14,128496,128498,14,128640,128709,14,128723,128724,14,128736,128741,14,128747,128748,14,128755,128755,14,128762,128762,14,128981,128991,14,129096,129103,14,129292,129292,14,129311,129311,14,129329,129330,14,129344,129349,14,129360,129374,14,129394,129394,14,129402,129402,14,129413,129425,14,129445,129450,14,129466,129471,14,129483,129483,14,129511,129535,14,129653,129655,14,129667,129670,14,129705,129711,14,129731,129743,14,917505,917505,4,917760,917999,5,10,10,3,127,159,4,768,879,5,1471,1471,5,1536,1541,1,1648,1648,5,1767,1768,5,1840,1866,5,2070,2073,5,2137,2139,5,2307,2307,7,2366,2368,7,2382,2383,7,2434,2435,7,2497,2500,5,2519,2519,5,2563,2563,7,2631,2632,5,2677,2677,5,2750,2752,7,2763,2764,7,2817,2817,5,2879,2879,5,2891,2892,7,2914,2915,5,3008,3008,5,3021,3021,5,3076,3076,5,3146,3149,5,3202,3203,7,3264,3265,7,3271,3272,7,3298,3299,5,3390,3390,5,3402,3404,7,3426,3427,5,3535,3535,5,3544,3550,7,3635,3635,7,3763,3763,7,3893,3893,5,3953,3966,5,3981,3991,5,4145,4145,7,4157,4158,5,4209,4212,5,4237,4237,5,4520,4607,10,5970,5971,5,6071,6077,5,6089,6099,5,6277,6278,5,6439,6440,5,6451,6456,7,6683,6683,5,6744,6750,5,6765,6770,7,6846,6846,5,6964,6964,5,6972,6972,5,7019,7027,5,7074,7077,5,7083,7085,5,7146,7148,7,7154,7155,7,7222,7223,5,7394,7400,5,7416,7417,5,8204,8204,5,8233,8233,4,8288,8292,4,8413,8416,5,8482,8482,14,8986,8987,14,9193,9203,14,9654,9654,14,9733,9733,14,9745,9745,14,9752,9752,14,9760,9760,14,9766,9766,14,9774,9775,14,9792,9792,14,9800,9811,14,9825,9826,14,9831,9831,14,9852,9853,14,9872,9873,14,9880,9880,14,9885,9887,14,9896,9897,14,9906,9916,14,9926,9927,14,9936,9936,14,9941,9960,14,9974,9974,14,9982,9985,14,9992,9997,14,10002,10002,14,10017,10017,14,10055,10055,14,10071,10071,14,10145,10145,14,11013,11015,14,11503,11505,5,12334,12335,5,12951,12951,14,42612,42621,5,43014,43014,5,43047,43047,7,43204,43205,5,43335,43345,5,43395,43395,7,43450,43451,7,43561,43566,5,43573,43574,5,43644,43644,5,43710,43711,5,43758,43759,7,44005,44005,5,44012,44012,7,44060,44060,11,44116,44116,11,44172,44172,11,44228,44228,11,44284,44284,11,44340,44340,11,44396,44396,11,44452,44452,11,44508,44508,11,44564,44564,11,44620,44620,11,44676,44676,11,44732,44732,11,44788,44788,11,44844,44844,11,44900,44900,11,44956,44956,11,45012,45012,11,45068,45068,11,45124,45124,11,45180,45180,11,45236,45236,11,45292,45292,11,45348,45348,11,45404,45404,11,45460,45460,11,45516,45516,11,45572,45572,11,45628,45628,11,45684,45684,11,45740,45740,11,45796,45796,11,45852,45852,11,45908,45908,11,45964,45964,11,46020,46020,11,46076,46076,11,46132,46132,11,46188,46188,11,46244,46244,11,46300,46300,11,46356,46356,11,46412,46412,11,46468,46468,11,46524,46524,11,46580,46580,11,46636,46636,11,46692,46692,11,46748,46748,11,46804,46804,11,46860,46860,11,46916,46916,11,46972,46972,11,47028,47028,11,47084,47084,11,47140,47140,11,47196,47196,11,47252,47252,11,47308,47308,11,47364,47364,11,47420,47420,11,47476,47476,11,47532,47532,11,47588,47588,11,47644,47644,11,47700,47700,11,47756,47756,11,47812,47812,11,47868,47868,11,47924,47924,11,47980,47980,11,48036,48036,11,48092,48092,11,48148,48148,11,48204,48204,11,48260,48260,11,48316,48316,11,48372,48372,11,48428,48428,11,48484,48484,11,48540,48540,11,48596,48596,11,48652,48652,11,48708,48708,11,48764,48764,11,48820,48820,11,48876,48876,11,48932,48932,11,48988,48988,11,49044,49044,11,49100,49100,11,49156,49156,11,49212,49212,11,49268,49268,11,49324,49324,11,49380,49380,11,49436,49436,11,49492,49492,11,49548,49548,11,49604,49604,11,49660,49660,11,49716,49716,11,49772,49772,11,49828,49828,11,49884,49884,11,49940,49940,11,49996,49996,11,50052,50052,11,50108,50108,11,50164,50164,11,50220,50220,11,50276,50276,11,50332,50332,11,50388,50388,11,50444,50444,11,50500,50500,11,50556,50556,11,50612,50612,11,50668,50668,11,50724,50724,11,50780,50780,11,50836,50836,11,50892,50892,11,50948,50948,11,51004,51004,11,51060,51060,11,51116,51116,11,51172,51172,11,51228,51228,11,51284,51284,11,51340,51340,11,51396,51396,11,51452,51452,11,51508,51508,11,51564,51564,11,51620,51620,11,51676,51676,11,51732,51732,11,51788,51788,11,51844,51844,11,51900,51900,11,51956,51956,11,52012,52012,11,52068,52068,11,52124,52124,11,52180,52180,11,52236,52236,11,52292,52292,11,52348,52348,11,52404,52404,11,52460,52460,11,52516,52516,11,52572,52572,11,52628,52628,11,52684,52684,11,52740,52740,11,52796,52796,11,52852,52852,11,52908,52908,11,52964,52964,11,53020,53020,11,53076,53076,11,53132,53132,11,53188,53188,11,53244,53244,11,53300,53300,11,53356,53356,11,53412,53412,11,53468,53468,11,53524,53524,11,53580,53580,11,53636,53636,11,53692,53692,11,53748,53748,11,53804,53804,11,53860,53860,11,53916,53916,11,53972,53972,11,54028,54028,11,54084,54084,11,54140,54140,11,54196,54196,11,54252,54252,11,54308,54308,11,54364,54364,11,54420,54420,11,54476,54476,11,54532,54532,11,54588,54588,11,54644,54644,11,54700,54700,11,54756,54756,11,54812,54812,11,54868,54868,11,54924,54924,11,54980,54980,11,55036,55036,11,55092,55092,11,55148,55148,11,55216,55238,9,65056,65071,5,65529,65531,4,68097,68099,5,68159,68159,5,69446,69456,5,69688,69702,5,69808,69810,7,69815,69816,7,69821,69821,1,69888,69890,5,69932,69932,7,69957,69958,7,70016,70017,5,70067,70069,7,70079,70080,7,70089,70092,5,70095,70095,5,70191,70193,5,70196,70196,5,70198,70199,5,70367,70367,5,70371,70378,5,70402,70403,7,70462,70462,5,70464,70464,5,70471,70472,7,70487,70487,5,70502,70508,5,70709,70711,7,70720,70721,7,70725,70725,7,70750,70750,5,70833,70834,7,70841,70841,7,70843,70844,7,70846,70846,7,70849,70849,7,71087,71087,5,71090,71093,5,71100,71101,5,71103,71104,5,71216,71218,7,71227,71228,7,71230,71230,7,71339,71339,5,71341,71341,5,71344,71349,5,71351,71351,5,71456,71457,7,71462,71462,7,71724,71726,7,71736,71736,7,71984,71984,5,71991,71992,7,71997,71997,7,71999,71999,1,72001,72001,1,72003,72003,5,72148,72151,5,72156,72159,7,72164,72164,7,72243,72248,5,72250,72250,1,72263,72263,5,72279,72280,7,72324,72329,1,72343,72343,7,72751,72751,7,72760,72765,5,72767,72767,5,72873,72873,7,72881,72881,7,72884,72884,7,73009,73014,5,73020,73021,5,73030,73030,1,73098,73102,7,73107,73108,7,73110,73110,7,73459,73460,5,78896,78904,4,92976,92982,5,94033,94087,7,94180,94180,5,113821,113822,5,119141,119141,5,119143,119145,5,119150,119154,5,119163,119170,5,119210,119213,5,121344,121398,5,121461,121461,5,121499,121503,5,122880,122886,5,122907,122913,5,122918,122922,5,123628,123631,5,125252,125258,5,126980,126980,14,127183,127183,14,127245,127247,14,127340,127343,14,127358,127359,14,127377,127386,14,127462,127487,6,127491,127503,14,127535,127535,14,127548,127551,14,127568,127569,14,127744,127777,14,127780,127891,14,127894,127895,14,127897,127899,14,127902,127984,14,127987,127989,14,127991,127994,14,128000,128253,14,128255,128317,14,128329,128334,14,128336,128359,14,128367,128368,14,128371,128377,14,128379,128390,14,128392,128393,14,128398,128399,14,128401,128404,14,128407,128419,14,128421,128421,14,128424,128424,14,128433,128434,14,128444,128444,14,128450,128452,14,128465,128467,14,128476,128478,14,128481,128481,14,128483,128483,14,128488,128488,14,128495,128495,14,128499,128499,14,128506,128591,14,128710,128714,14,128721,128722,14,128725,128725,14,128728,128735,14,128742,128744,14,128746,128746,14,128749,128751,14,128753,128754,14,128756,128758,14,128761,128761,14,128763,128764,14,128884,128895,14,128992,129003,14,129036,129039,14,129114,129119,14,129198,129279,14,129293,129295,14,129305,129310,14,129312,129319,14,129328,129328,14,129331,129338,14,129343,129343,14,129351,129355,14,129357,129359,14,129375,129387,14,129393,129393,14,129395,129398,14,129401,129401,14,129403,129403,14,129408,129412,14,129426,129431,14,129443,129444,14,129451,129453,14,129456,129465,14,129472,129472,14,129475,129482,14,129484,129484,14,129488,129510,14,129536,129647,14,129652,129652,14,129656,129658,14,129664,129666,14,129671,129679,14,129686,129704,14,129712,129718,14,129728,129730,14,129744,129750,14,917504,917504,4,917506,917535,4,917632,917759,4,918000,921599,4,0,9,4,11,12,4,14,31,4,169,169,14,174,174,14,1155,1159,5,1425,1469,5,1473,1474,5,1479,1479,5,1552,1562,5,1611,1631,5,1750,1756,5,1759,1764,5,1770,1773,5,1809,1809,5,1958,1968,5,2045,2045,5,2075,2083,5,2089,2093,5,2259,2273,5,2275,2306,5,2362,2362,5,2364,2364,5,2369,2376,5,2381,2381,5,2385,2391,5,2433,2433,5,2492,2492,5,2495,2496,7,2503,2504,7,2509,2509,5,2530,2531,5,2561,2562,5,2620,2620,5,2625,2626,5,2635,2637,5,2672,2673,5,2689,2690,5,2748,2748,5,2753,2757,5,2761,2761,7,2765,2765,5,2810,2815,5,2818,2819,7,2878,2878,5,2880,2880,7,2887,2888,7,2893,2893,5,2903,2903,5,2946,2946,5,3007,3007,7,3009,3010,7,3018,3020,7,3031,3031,5,3073,3075,7,3134,3136,5,3142,3144,5,3157,3158,5,3201,3201,5,3260,3260,5,3263,3263,5,3266,3266,5,3270,3270,5,3274,3275,7,3285,3286,5,3328,3329,5,3387,3388,5,3391,3392,7,3398,3400,7,3405,3405,5,3415,3415,5,3457,3457,5,3530,3530,5,3536,3537,7,3542,3542,5,3551,3551,5,3633,3633,5,3636,3642,5,3761,3761,5,3764,3772,5,3864,3865,5,3895,3895,5,3902,3903,7,3967,3967,7,3974,3975,5,3993,4028,5,4141,4144,5,4146,4151,5,4155,4156,7,4182,4183,7,4190,4192,5,4226,4226,5,4229,4230,5,4253,4253,5,4448,4519,9,4957,4959,5,5938,5940,5,6002,6003,5,6070,6070,7,6078,6085,7,6087,6088,7,6109,6109,5,6158,6158,4,6313,6313,5,6435,6438,7,6441,6443,7,6450,6450,5,6457,6459,5,6681,6682,7,6741,6741,7,6743,6743,7,6752,6752,5,6757,6764,5,6771,6780,5,6832,6845,5,6847,6848,5,6916,6916,7,6965,6965,5,6971,6971,7,6973,6977,7,6979,6980,7,7040,7041,5,7073,7073,7,7078,7079,7,7082,7082,7,7142,7142,5,7144,7145,5,7149,7149,5,7151,7153,5,7204,7211,7,7220,7221,7,7376,7378,5,7393,7393,7,7405,7405,5,7415,7415,7,7616,7673,5,8203,8203,4,8205,8205,13,8232,8232,4,8234,8238,4,8265,8265,14,8293,8293,4,8400,8412,5,8417,8417,5,8421,8432,5,8505,8505,14,8617,8618,14,9000,9000,14,9167,9167,14,9208,9210,14,9642,9643,14,9664,9664,14,9728,9732,14,9735,9741,14,9743,9744,14,9746,9746,14,9750,9751,14,9753,9756,14,9758,9759,14,9761,9761,14,9764,9765,14,9767,9769,14,9771,9773,14,9776,9783,14,9787,9791,14,9793,9793,14,9795,9799,14,9812,9822,14,9824,9824,14,9827,9827,14,9829,9830,14,9832,9832,14,9851,9851,14,9854,9854,14,9856,9861,14,9874,9876,14,9878,9879,14,9881,9881,14,9883,9884,14,9888,9889,14,9895,9895,14,9898,9899,14,9904,9905,14,9917,9918,14,9924,9925,14,9928,9928,14,9934,9935,14,9937,9937,14,9939,9940,14,9961,9962,14,9968,9973,14,9975,9978,14,9981,9981,14,9986,9986,14,9989,9989,14,9998,9998,14,10000,10001,14,10004,10004,14,10013,10013,14,10024,10024,14,10052,10052,14,10060,10060,14,10067,10069,14,10083,10084,14,10133,10135,14,10160,10160,14,10548,10549,14,11035,11036,14,11093,11093,14,11647,11647,5,12330,12333,5,12336,12336,14,12441,12442,5,12953,12953,14,42608,42610,5,42654,42655,5,43010,43010,5,43019,43019,5,43045,43046,5,43052,43052,5,43188,43203,7,43232,43249,5,43302,43309,5,43346,43347,7,43392,43394,5,43443,43443,5,43446,43449,5,43452,43453,5,43493,43493,5,43567,43568,7,43571,43572,7,43587,43587,5,43597,43597,7,43696,43696,5,43703,43704,5,43713,43713,5,43756,43757,5,43765,43765,7,44003,44004,7,44006,44007,7,44009,44010,7,44013,44013,5,44033,44059,12,44061,44087,12,44089,44115,12,44117,44143,12,44145,44171,12,44173,44199,12,44201,44227,12,44229,44255,12,44257,44283,12,44285,44311,12,44313,44339,12,44341,44367,12,44369,44395,12,44397,44423,12,44425,44451,12,44453,44479,12,44481,44507,12,44509,44535,12,44537,44563,12,44565,44591,12,44593,44619,12,44621,44647,12,44649,44675,12,44677,44703,12,44705,44731,12,44733,44759,12,44761,44787,12,44789,44815,12,44817,44843,12,44845,44871,12,44873,44899,12,44901,44927,12,44929,44955,12,44957,44983,12,44985,45011,12,45013,45039,12,45041,45067,12,45069,45095,12,45097,45123,12,45125,45151,12,45153,45179,12,45181,45207,12,45209,45235,12,45237,45263,12,45265,45291,12,45293,45319,12,45321,45347,12,45349,45375,12,45377,45403,12,45405,45431,12,45433,45459,12,45461,45487,12,45489,45515,12,45517,45543,12,45545,45571,12,45573,45599,12,45601,45627,12,45629,45655,12,45657,45683,12,45685,45711,12,45713,45739,12,45741,45767,12,45769,45795,12,45797,45823,12,45825,45851,12,45853,45879,12,45881,45907,12,45909,45935,12,45937,45963,12,45965,45991,12,45993,46019,12,46021,46047,12,46049,46075,12,46077,46103,12,46105,46131,12,46133,46159,12,46161,46187,12,46189,46215,12,46217,46243,12,46245,46271,12,46273,46299,12,46301,46327,12,46329,46355,12,46357,46383,12,46385,46411,12,46413,46439,12,46441,46467,12,46469,46495,12,46497,46523,12,46525,46551,12,46553,46579,12,46581,46607,12,46609,46635,12,46637,46663,12,46665,46691,12,46693,46719,12,46721,46747,12,46749,46775,12,46777,46803,12,46805,46831,12,46833,46859,12,46861,46887,12,46889,46915,12,46917,46943,12,46945,46971,12,46973,46999,12,47001,47027,12,47029,47055,12,47057,47083,12,47085,47111,12,47113,47139,12,47141,47167,12,47169,47195,12,47197,47223,12,47225,47251,12,47253,47279,12,47281,47307,12,47309,47335,12,47337,47363,12,47365,47391,12,47393,47419,12,47421,47447,12,47449,47475,12,47477,47503,12,47505,47531,12,47533,47559,12,47561,47587,12,47589,47615,12,47617,47643,12,47645,47671,12,47673,47699,12,47701,47727,12,47729,47755,12,47757,47783,12,47785,47811,12,47813,47839,12,47841,47867,12,47869,47895,12,47897,47923,12,47925,47951,12,47953,47979,12,47981,48007,12,48009,48035,12,48037,48063,12,48065,48091,12,48093,48119,12,48121,48147,12,48149,48175,12,48177,48203,12,48205,48231,12,48233,48259,12,48261,48287,12,48289,48315,12,48317,48343,12,48345,48371,12,48373,48399,12,48401,48427,12,48429,48455,12,48457,48483,12,48485,48511,12,48513,48539,12,48541,48567,12,48569,48595,12,48597,48623,12,48625,48651,12,48653,48679,12,48681,48707,12,48709,48735,12,48737,48763,12,48765,48791,12,48793,48819,12,48821,48847,12,48849,48875,12,48877,48903,12,48905,48931,12,48933,48959,12,48961,48987,12,48989,49015,12,49017,49043,12,49045,49071,12,49073,49099,12,49101,49127,12,49129,49155,12,49157,49183,12,49185,49211,12,49213,49239,12,49241,49267,12,49269,49295,12,49297,49323,12,49325,49351,12,49353,49379,12,49381,49407,12,49409,49435,12,49437,49463,12,49465,49491,12,49493,49519,12,49521,49547,12,49549,49575,12,49577,49603,12,49605,49631,12,49633,49659,12,49661,49687,12,49689,49715,12,49717,49743,12,49745,49771,12,49773,49799,12,49801,49827,12,49829,49855,12,49857,49883,12,49885,49911,12,49913,49939,12,49941,49967,12,49969,49995,12,49997,50023,12,50025,50051,12,50053,50079,12,50081,50107,12,50109,50135,12,50137,50163,12,50165,50191,12,50193,50219,12,50221,50247,12,50249,50275,12,50277,50303,12,50305,50331,12,50333,50359,12,50361,50387,12,50389,50415,12,50417,50443,12,50445,50471,12,50473,50499,12,50501,50527,12,50529,50555,12,50557,50583,12,50585,50611,12,50613,50639,12,50641,50667,12,50669,50695,12,50697,50723,12,50725,50751,12,50753,50779,12,50781,50807,12,50809,50835,12,50837,50863,12,50865,50891,12,50893,50919,12,50921,50947,12,50949,50975,12,50977,51003,12,51005,51031,12,51033,51059,12,51061,51087,12,51089,51115,12,51117,51143,12,51145,51171,12,51173,51199,12,51201,51227,12,51229,51255,12,51257,51283,12,51285,51311,12,51313,51339,12,51341,51367,12,51369,51395,12,51397,51423,12,51425,51451,12,51453,51479,12,51481,51507,12,51509,51535,12,51537,51563,12,51565,51591,12,51593,51619,12,51621,51647,12,51649,51675,12,51677,51703,12,51705,51731,12,51733,51759,12,51761,51787,12,51789,51815,12,51817,51843,12,51845,51871,12,51873,51899,12,51901,51927,12,51929,51955,12,51957,51983,12,51985,52011,12,52013,52039,12,52041,52067,12,52069,52095,12,52097,52123,12,52125,52151,12,52153,52179,12,52181,52207,12,52209,52235,12,52237,52263,12,52265,52291,12,52293,52319,12,52321,52347,12,52349,52375,12,52377,52403,12,52405,52431,12,52433,52459,12,52461,52487,12,52489,52515,12,52517,52543,12,52545,52571,12,52573,52599,12,52601,52627,12,52629,52655,12,52657,52683,12,52685,52711,12,52713,52739,12,52741,52767,12,52769,52795,12,52797,52823,12,52825,52851,12,52853,52879,12,52881,52907,12,52909,52935,12,52937,52963,12,52965,52991,12,52993,53019,12,53021,53047,12,53049,53075,12,53077,53103,12,53105,53131,12,53133,53159,12,53161,53187,12,53189,53215,12,53217,53243,12,53245,53271,12,53273,53299,12,53301,53327,12,53329,53355,12,53357,53383,12,53385,53411,12,53413,53439,12,53441,53467,12,53469,53495,12,53497,53523,12,53525,53551,12,53553,53579,12,53581,53607,12,53609,53635,12,53637,53663,12,53665,53691,12,53693,53719,12,53721,53747,12,53749,53775,12,53777,53803,12,53805,53831,12,53833,53859,12,53861,53887,12,53889,53915,12,53917,53943,12,53945,53971,12,53973,53999,12,54001,54027,12,54029,54055,12,54057,54083,12,54085,54111,12,54113,54139,12,54141,54167,12,54169,54195,12,54197,54223,12,54225,54251,12,54253,54279,12,54281,54307,12,54309,54335,12,54337,54363,12,54365,54391,12,54393,54419,12,54421,54447,12,54449,54475,12,54477,54503,12,54505,54531,12,54533,54559,12,54561,54587,12,54589,54615,12,54617,54643,12,54645,54671,12,54673,54699,12,54701,54727,12,54729,54755,12,54757,54783,12,54785,54811,12,54813,54839,12,54841,54867,12,54869,54895,12,54897,54923,12,54925,54951,12,54953,54979,12,54981,55007,12,55009,55035,12,55037,55063,12,55065,55091,12,55093,55119,12,55121,55147,12,55149,55175,12,55177,55203,12,55243,55291,10,65024,65039,5,65279,65279,4,65520,65528,4,66045,66045,5,66422,66426,5,68101,68102,5,68152,68154,5,68325,68326,5,69291,69292,5,69632,69632,7,69634,69634,7,69759,69761,5]");
  }

  // node_modules/monaco-editor/esm/vs/base/common/hash.js
  function numberHash(val, initialHashVal) {
    return (initialHashVal << 5) - initialHashVal + val | 0;
  }
  function stringHash(s, hashVal) {
    hashVal = numberHash(149417, hashVal);
    for (let i = 0, length = s.length; i < length; i++) {
      hashVal = numberHash(s.charCodeAt(i), hashVal);
    }
    return hashVal;
  }
  function leftRotate(value, bits, totalBits = 32) {
    const delta = totalBits - bits;
    const mask = ~((1 << delta) - 1);
    return (value << bits | (mask & value) >>> delta) >>> 0;
  }
  function fill(dest, index = 0, count = dest.byteLength, value = 0) {
    for (let i = 0; i < count; i++) {
      dest[index + i] = value;
    }
  }
  function leftPad(value, length, char = "0") {
    while (value.length < length) {
      value = char + value;
    }
    return value;
  }
  function toHexString(value, bitsize = 32) {
    return leftPad((value >>> 0).toString(16), bitsize / 4);
  }
  var StringSHA1 = class {
    constructor() {
      this._h0 = 1732584193;
      this._h1 = 4023233417;
      this._h2 = 2562383102;
      this._h3 = 271733878;
      this._h4 = 3285377520;
      this._buff = new Uint8Array(64 + 3);
      this._buffDV = new DataView(this._buff.buffer);
      this._buffLen = 0;
      this._totalLen = 0;
      this._leftoverHighSurrogate = 0;
      this._finished = false;
    }
    update(str) {
      const strLen = str.length;
      if (strLen === 0) {
        return;
      }
      const buff = this._buff;
      let buffLen = this._buffLen;
      let leftoverHighSurrogate = this._leftoverHighSurrogate;
      let charCode;
      let offset;
      if (leftoverHighSurrogate !== 0) {
        charCode = leftoverHighSurrogate;
        offset = -1;
        leftoverHighSurrogate = 0;
      } else {
        charCode = str.charCodeAt(0);
        offset = 0;
      }
      while (true) {
        let codePoint = charCode;
        if (isHighSurrogate(charCode)) {
          if (offset + 1 < strLen) {
            const nextCharCode = str.charCodeAt(offset + 1);
            if (isLowSurrogate(nextCharCode)) {
              offset++;
              codePoint = computeCodePoint(charCode, nextCharCode);
            } else {
              codePoint = 65533;
            }
          } else {
            leftoverHighSurrogate = charCode;
            break;
          }
        } else if (isLowSurrogate(charCode)) {
          codePoint = 65533;
        }
        buffLen = this._push(buff, buffLen, codePoint);
        offset++;
        if (offset < strLen) {
          charCode = str.charCodeAt(offset);
        } else {
          break;
        }
      }
      this._buffLen = buffLen;
      this._leftoverHighSurrogate = leftoverHighSurrogate;
    }
    _push(buff, buffLen, codePoint) {
      if (codePoint < 128) {
        buff[buffLen++] = codePoint;
      } else if (codePoint < 2048) {
        buff[buffLen++] = 192 | (codePoint & 1984) >>> 6;
        buff[buffLen++] = 128 | (codePoint & 63) >>> 0;
      } else if (codePoint < 65536) {
        buff[buffLen++] = 224 | (codePoint & 61440) >>> 12;
        buff[buffLen++] = 128 | (codePoint & 4032) >>> 6;
        buff[buffLen++] = 128 | (codePoint & 63) >>> 0;
      } else {
        buff[buffLen++] = 240 | (codePoint & 1835008) >>> 18;
        buff[buffLen++] = 128 | (codePoint & 258048) >>> 12;
        buff[buffLen++] = 128 | (codePoint & 4032) >>> 6;
        buff[buffLen++] = 128 | (codePoint & 63) >>> 0;
      }
      if (buffLen >= 64) {
        this._step();
        buffLen -= 64;
        this._totalLen += 64;
        buff[0] = buff[64 + 0];
        buff[1] = buff[64 + 1];
        buff[2] = buff[64 + 2];
      }
      return buffLen;
    }
    digest() {
      if (!this._finished) {
        this._finished = true;
        if (this._leftoverHighSurrogate) {
          this._leftoverHighSurrogate = 0;
          this._buffLen = this._push(this._buff, this._buffLen, 65533);
        }
        this._totalLen += this._buffLen;
        this._wrapUp();
      }
      return toHexString(this._h0) + toHexString(this._h1) + toHexString(this._h2) + toHexString(this._h3) + toHexString(this._h4);
    }
    _wrapUp() {
      this._buff[this._buffLen++] = 128;
      fill(this._buff, this._buffLen);
      if (this._buffLen > 56) {
        this._step();
        fill(this._buff);
      }
      const ml = 8 * this._totalLen;
      this._buffDV.setUint32(56, Math.floor(ml / 4294967296), false);
      this._buffDV.setUint32(60, ml % 4294967296, false);
      this._step();
    }
    _step() {
      const bigBlock32 = StringSHA1._bigBlock32;
      const data = this._buffDV;
      for (let j = 0; j < 64; j += 4) {
        bigBlock32.setUint32(j, data.getUint32(j, false), false);
      }
      for (let j = 64; j < 320; j += 4) {
        bigBlock32.setUint32(j, leftRotate(bigBlock32.getUint32(j - 12, false) ^ bigBlock32.getUint32(j - 32, false) ^ bigBlock32.getUint32(j - 56, false) ^ bigBlock32.getUint32(j - 64, false), 1), false);
      }
      let a2 = this._h0;
      let b = this._h1;
      let c = this._h2;
      let d = this._h3;
      let e = this._h4;
      let f2, k;
      let temp;
      for (let j = 0; j < 80; j++) {
        if (j < 20) {
          f2 = b & c | ~b & d;
          k = 1518500249;
        } else if (j < 40) {
          f2 = b ^ c ^ d;
          k = 1859775393;
        } else if (j < 60) {
          f2 = b & c | b & d | c & d;
          k = 2400959708;
        } else {
          f2 = b ^ c ^ d;
          k = 3395469782;
        }
        temp = leftRotate(a2, 5) + f2 + e + k + bigBlock32.getUint32(j * 4, false) & 4294967295;
        e = d;
        d = c;
        c = leftRotate(b, 30);
        b = a2;
        a2 = temp;
      }
      this._h0 = this._h0 + a2 & 4294967295;
      this._h1 = this._h1 + b & 4294967295;
      this._h2 = this._h2 + c & 4294967295;
      this._h3 = this._h3 + d & 4294967295;
      this._h4 = this._h4 + e & 4294967295;
    }
  };
  StringSHA1._bigBlock32 = new DataView(new ArrayBuffer(320));

  // node_modules/monaco-editor/esm/vs/base/common/diff/diff.js
  var StringDiffSequence = class {
    constructor(source) {
      this.source = source;
    }
    getElements() {
      const source = this.source;
      const characters = new Int32Array(source.length);
      for (let i = 0, len = source.length; i < len; i++) {
        characters[i] = source.charCodeAt(i);
      }
      return characters;
    }
  };
  function stringDiff(original, modified, pretty) {
    return new LcsDiff(new StringDiffSequence(original), new StringDiffSequence(modified)).ComputeDiff(pretty).changes;
  }
  var Debug = class {
    static Assert(condition, message) {
      if (!condition) {
        throw new Error(message);
      }
    }
  };
  var MyArray = class {
    static Copy(sourceArray, sourceIndex, destinationArray, destinationIndex, length) {
      for (let i = 0; i < length; i++) {
        destinationArray[destinationIndex + i] = sourceArray[sourceIndex + i];
      }
    }
    static Copy2(sourceArray, sourceIndex, destinationArray, destinationIndex, length) {
      for (let i = 0; i < length; i++) {
        destinationArray[destinationIndex + i] = sourceArray[sourceIndex + i];
      }
    }
  };
  var DiffChangeHelper = class {
    constructor() {
      this.m_changes = [];
      this.m_originalStart = 1073741824;
      this.m_modifiedStart = 1073741824;
      this.m_originalCount = 0;
      this.m_modifiedCount = 0;
    }
    MarkNextChange() {
      if (this.m_originalCount > 0 || this.m_modifiedCount > 0) {
        this.m_changes.push(new DiffChange(this.m_originalStart, this.m_originalCount, this.m_modifiedStart, this.m_modifiedCount));
      }
      this.m_originalCount = 0;
      this.m_modifiedCount = 0;
      this.m_originalStart = 1073741824;
      this.m_modifiedStart = 1073741824;
    }
    AddOriginalElement(originalIndex, modifiedIndex) {
      this.m_originalStart = Math.min(this.m_originalStart, originalIndex);
      this.m_modifiedStart = Math.min(this.m_modifiedStart, modifiedIndex);
      this.m_originalCount++;
    }
    AddModifiedElement(originalIndex, modifiedIndex) {
      this.m_originalStart = Math.min(this.m_originalStart, originalIndex);
      this.m_modifiedStart = Math.min(this.m_modifiedStart, modifiedIndex);
      this.m_modifiedCount++;
    }
    getChanges() {
      if (this.m_originalCount > 0 || this.m_modifiedCount > 0) {
        this.MarkNextChange();
      }
      return this.m_changes;
    }
    getReverseChanges() {
      if (this.m_originalCount > 0 || this.m_modifiedCount > 0) {
        this.MarkNextChange();
      }
      this.m_changes.reverse();
      return this.m_changes;
    }
  };
  var LcsDiff = class {
    constructor(originalSequence, modifiedSequence, continueProcessingPredicate = null) {
      this.ContinueProcessingPredicate = continueProcessingPredicate;
      const [originalStringElements, originalElementsOrHash, originalHasStrings] = LcsDiff._getElements(originalSequence);
      const [modifiedStringElements, modifiedElementsOrHash, modifiedHasStrings] = LcsDiff._getElements(modifiedSequence);
      this._hasStrings = originalHasStrings && modifiedHasStrings;
      this._originalStringElements = originalStringElements;
      this._originalElementsOrHash = originalElementsOrHash;
      this._modifiedStringElements = modifiedStringElements;
      this._modifiedElementsOrHash = modifiedElementsOrHash;
      this.m_forwardHistory = [];
      this.m_reverseHistory = [];
    }
    static _isStringArray(arr) {
      return arr.length > 0 && typeof arr[0] === "string";
    }
    static _getElements(sequence) {
      const elements = sequence.getElements();
      if (LcsDiff._isStringArray(elements)) {
        const hashes = new Int32Array(elements.length);
        for (let i = 0, len = elements.length; i < len; i++) {
          hashes[i] = stringHash(elements[i], 0);
        }
        return [elements, hashes, true];
      }
      if (elements instanceof Int32Array) {
        return [[], elements, false];
      }
      return [[], new Int32Array(elements), false];
    }
    ElementsAreEqual(originalIndex, newIndex) {
      if (this._originalElementsOrHash[originalIndex] !== this._modifiedElementsOrHash[newIndex]) {
        return false;
      }
      return this._hasStrings ? this._originalStringElements[originalIndex] === this._modifiedStringElements[newIndex] : true;
    }
    OriginalElementsAreEqual(index1, index2) {
      if (this._originalElementsOrHash[index1] !== this._originalElementsOrHash[index2]) {
        return false;
      }
      return this._hasStrings ? this._originalStringElements[index1] === this._originalStringElements[index2] : true;
    }
    ModifiedElementsAreEqual(index1, index2) {
      if (this._modifiedElementsOrHash[index1] !== this._modifiedElementsOrHash[index2]) {
        return false;
      }
      return this._hasStrings ? this._modifiedStringElements[index1] === this._modifiedStringElements[index2] : true;
    }
    ComputeDiff(pretty) {
      return this._ComputeDiff(0, this._originalElementsOrHash.length - 1, 0, this._modifiedElementsOrHash.length - 1, pretty);
    }
    _ComputeDiff(originalStart, originalEnd, modifiedStart, modifiedEnd, pretty) {
      const quitEarlyArr = [false];
      let changes = this.ComputeDiffRecursive(originalStart, originalEnd, modifiedStart, modifiedEnd, quitEarlyArr);
      if (pretty) {
        changes = this.PrettifyChanges(changes);
      }
      return {
        quitEarly: quitEarlyArr[0],
        changes
      };
    }
    ComputeDiffRecursive(originalStart, originalEnd, modifiedStart, modifiedEnd, quitEarlyArr) {
      quitEarlyArr[0] = false;
      while (originalStart <= originalEnd && modifiedStart <= modifiedEnd && this.ElementsAreEqual(originalStart, modifiedStart)) {
        originalStart++;
        modifiedStart++;
      }
      while (originalEnd >= originalStart && modifiedEnd >= modifiedStart && this.ElementsAreEqual(originalEnd, modifiedEnd)) {
        originalEnd--;
        modifiedEnd--;
      }
      if (originalStart > originalEnd || modifiedStart > modifiedEnd) {
        let changes;
        if (modifiedStart <= modifiedEnd) {
          Debug.Assert(originalStart === originalEnd + 1, "originalStart should only be one more than originalEnd");
          changes = [
            new DiffChange(originalStart, 0, modifiedStart, modifiedEnd - modifiedStart + 1)
          ];
        } else if (originalStart <= originalEnd) {
          Debug.Assert(modifiedStart === modifiedEnd + 1, "modifiedStart should only be one more than modifiedEnd");
          changes = [
            new DiffChange(originalStart, originalEnd - originalStart + 1, modifiedStart, 0)
          ];
        } else {
          Debug.Assert(originalStart === originalEnd + 1, "originalStart should only be one more than originalEnd");
          Debug.Assert(modifiedStart === modifiedEnd + 1, "modifiedStart should only be one more than modifiedEnd");
          changes = [];
        }
        return changes;
      }
      const midOriginalArr = [0];
      const midModifiedArr = [0];
      const result = this.ComputeRecursionPoint(originalStart, originalEnd, modifiedStart, modifiedEnd, midOriginalArr, midModifiedArr, quitEarlyArr);
      const midOriginal = midOriginalArr[0];
      const midModified = midModifiedArr[0];
      if (result !== null) {
        return result;
      } else if (!quitEarlyArr[0]) {
        const leftChanges = this.ComputeDiffRecursive(originalStart, midOriginal, modifiedStart, midModified, quitEarlyArr);
        let rightChanges = [];
        if (!quitEarlyArr[0]) {
          rightChanges = this.ComputeDiffRecursive(midOriginal + 1, originalEnd, midModified + 1, modifiedEnd, quitEarlyArr);
        } else {
          rightChanges = [
            new DiffChange(midOriginal + 1, originalEnd - (midOriginal + 1) + 1, midModified + 1, modifiedEnd - (midModified + 1) + 1)
          ];
        }
        return this.ConcatenateChanges(leftChanges, rightChanges);
      }
      return [
        new DiffChange(originalStart, originalEnd - originalStart + 1, modifiedStart, modifiedEnd - modifiedStart + 1)
      ];
    }
    WALKTRACE(diagonalForwardBase, diagonalForwardStart, diagonalForwardEnd, diagonalForwardOffset, diagonalReverseBase, diagonalReverseStart, diagonalReverseEnd, diagonalReverseOffset, forwardPoints, reversePoints, originalIndex, originalEnd, midOriginalArr, modifiedIndex, modifiedEnd, midModifiedArr, deltaIsEven, quitEarlyArr) {
      let forwardChanges = null;
      let reverseChanges = null;
      let changeHelper = new DiffChangeHelper();
      let diagonalMin = diagonalForwardStart;
      let diagonalMax = diagonalForwardEnd;
      let diagonalRelative = midOriginalArr[0] - midModifiedArr[0] - diagonalForwardOffset;
      let lastOriginalIndex = -1073741824;
      let historyIndex = this.m_forwardHistory.length - 1;
      do {
        const diagonal = diagonalRelative + diagonalForwardBase;
        if (diagonal === diagonalMin || diagonal < diagonalMax && forwardPoints[diagonal - 1] < forwardPoints[diagonal + 1]) {
          originalIndex = forwardPoints[diagonal + 1];
          modifiedIndex = originalIndex - diagonalRelative - diagonalForwardOffset;
          if (originalIndex < lastOriginalIndex) {
            changeHelper.MarkNextChange();
          }
          lastOriginalIndex = originalIndex;
          changeHelper.AddModifiedElement(originalIndex + 1, modifiedIndex);
          diagonalRelative = diagonal + 1 - diagonalForwardBase;
        } else {
          originalIndex = forwardPoints[diagonal - 1] + 1;
          modifiedIndex = originalIndex - diagonalRelative - diagonalForwardOffset;
          if (originalIndex < lastOriginalIndex) {
            changeHelper.MarkNextChange();
          }
          lastOriginalIndex = originalIndex - 1;
          changeHelper.AddOriginalElement(originalIndex, modifiedIndex + 1);
          diagonalRelative = diagonal - 1 - diagonalForwardBase;
        }
        if (historyIndex >= 0) {
          forwardPoints = this.m_forwardHistory[historyIndex];
          diagonalForwardBase = forwardPoints[0];
          diagonalMin = 1;
          diagonalMax = forwardPoints.length - 1;
        }
      } while (--historyIndex >= -1);
      forwardChanges = changeHelper.getReverseChanges();
      if (quitEarlyArr[0]) {
        let originalStartPoint = midOriginalArr[0] + 1;
        let modifiedStartPoint = midModifiedArr[0] + 1;
        if (forwardChanges !== null && forwardChanges.length > 0) {
          const lastForwardChange = forwardChanges[forwardChanges.length - 1];
          originalStartPoint = Math.max(originalStartPoint, lastForwardChange.getOriginalEnd());
          modifiedStartPoint = Math.max(modifiedStartPoint, lastForwardChange.getModifiedEnd());
        }
        reverseChanges = [
          new DiffChange(originalStartPoint, originalEnd - originalStartPoint + 1, modifiedStartPoint, modifiedEnd - modifiedStartPoint + 1)
        ];
      } else {
        changeHelper = new DiffChangeHelper();
        diagonalMin = diagonalReverseStart;
        diagonalMax = diagonalReverseEnd;
        diagonalRelative = midOriginalArr[0] - midModifiedArr[0] - diagonalReverseOffset;
        lastOriginalIndex = 1073741824;
        historyIndex = deltaIsEven ? this.m_reverseHistory.length - 1 : this.m_reverseHistory.length - 2;
        do {
          const diagonal = diagonalRelative + diagonalReverseBase;
          if (diagonal === diagonalMin || diagonal < diagonalMax && reversePoints[diagonal - 1] >= reversePoints[diagonal + 1]) {
            originalIndex = reversePoints[diagonal + 1] - 1;
            modifiedIndex = originalIndex - diagonalRelative - diagonalReverseOffset;
            if (originalIndex > lastOriginalIndex) {
              changeHelper.MarkNextChange();
            }
            lastOriginalIndex = originalIndex + 1;
            changeHelper.AddOriginalElement(originalIndex + 1, modifiedIndex + 1);
            diagonalRelative = diagonal + 1 - diagonalReverseBase;
          } else {
            originalIndex = reversePoints[diagonal - 1];
            modifiedIndex = originalIndex - diagonalRelative - diagonalReverseOffset;
            if (originalIndex > lastOriginalIndex) {
              changeHelper.MarkNextChange();
            }
            lastOriginalIndex = originalIndex;
            changeHelper.AddModifiedElement(originalIndex + 1, modifiedIndex + 1);
            diagonalRelative = diagonal - 1 - diagonalReverseBase;
          }
          if (historyIndex >= 0) {
            reversePoints = this.m_reverseHistory[historyIndex];
            diagonalReverseBase = reversePoints[0];
            diagonalMin = 1;
            diagonalMax = reversePoints.length - 1;
          }
        } while (--historyIndex >= -1);
        reverseChanges = changeHelper.getChanges();
      }
      return this.ConcatenateChanges(forwardChanges, reverseChanges);
    }
    ComputeRecursionPoint(originalStart, originalEnd, modifiedStart, modifiedEnd, midOriginalArr, midModifiedArr, quitEarlyArr) {
      let originalIndex = 0, modifiedIndex = 0;
      let diagonalForwardStart = 0, diagonalForwardEnd = 0;
      let diagonalReverseStart = 0, diagonalReverseEnd = 0;
      originalStart--;
      modifiedStart--;
      midOriginalArr[0] = 0;
      midModifiedArr[0] = 0;
      this.m_forwardHistory = [];
      this.m_reverseHistory = [];
      const maxDifferences = originalEnd - originalStart + (modifiedEnd - modifiedStart);
      const numDiagonals = maxDifferences + 1;
      const forwardPoints = new Int32Array(numDiagonals);
      const reversePoints = new Int32Array(numDiagonals);
      const diagonalForwardBase = modifiedEnd - modifiedStart;
      const diagonalReverseBase = originalEnd - originalStart;
      const diagonalForwardOffset = originalStart - modifiedStart;
      const diagonalReverseOffset = originalEnd - modifiedEnd;
      const delta = diagonalReverseBase - diagonalForwardBase;
      const deltaIsEven = delta % 2 === 0;
      forwardPoints[diagonalForwardBase] = originalStart;
      reversePoints[diagonalReverseBase] = originalEnd;
      quitEarlyArr[0] = false;
      for (let numDifferences = 1; numDifferences <= maxDifferences / 2 + 1; numDifferences++) {
        let furthestOriginalIndex = 0;
        let furthestModifiedIndex = 0;
        diagonalForwardStart = this.ClipDiagonalBound(diagonalForwardBase - numDifferences, numDifferences, diagonalForwardBase, numDiagonals);
        diagonalForwardEnd = this.ClipDiagonalBound(diagonalForwardBase + numDifferences, numDifferences, diagonalForwardBase, numDiagonals);
        for (let diagonal = diagonalForwardStart; diagonal <= diagonalForwardEnd; diagonal += 2) {
          if (diagonal === diagonalForwardStart || diagonal < diagonalForwardEnd && forwardPoints[diagonal - 1] < forwardPoints[diagonal + 1]) {
            originalIndex = forwardPoints[diagonal + 1];
          } else {
            originalIndex = forwardPoints[diagonal - 1] + 1;
          }
          modifiedIndex = originalIndex - (diagonal - diagonalForwardBase) - diagonalForwardOffset;
          const tempOriginalIndex = originalIndex;
          while (originalIndex < originalEnd && modifiedIndex < modifiedEnd && this.ElementsAreEqual(originalIndex + 1, modifiedIndex + 1)) {
            originalIndex++;
            modifiedIndex++;
          }
          forwardPoints[diagonal] = originalIndex;
          if (originalIndex + modifiedIndex > furthestOriginalIndex + furthestModifiedIndex) {
            furthestOriginalIndex = originalIndex;
            furthestModifiedIndex = modifiedIndex;
          }
          if (!deltaIsEven && Math.abs(diagonal - diagonalReverseBase) <= numDifferences - 1) {
            if (originalIndex >= reversePoints[diagonal]) {
              midOriginalArr[0] = originalIndex;
              midModifiedArr[0] = modifiedIndex;
              if (tempOriginalIndex <= reversePoints[diagonal] && 1447 > 0 && numDifferences <= 1447 + 1) {
                return this.WALKTRACE(diagonalForwardBase, diagonalForwardStart, diagonalForwardEnd, diagonalForwardOffset, diagonalReverseBase, diagonalReverseStart, diagonalReverseEnd, diagonalReverseOffset, forwardPoints, reversePoints, originalIndex, originalEnd, midOriginalArr, modifiedIndex, modifiedEnd, midModifiedArr, deltaIsEven, quitEarlyArr);
              } else {
                return null;
              }
            }
          }
        }
        const matchLengthOfLongest = (furthestOriginalIndex - originalStart + (furthestModifiedIndex - modifiedStart) - numDifferences) / 2;
        if (this.ContinueProcessingPredicate !== null && !this.ContinueProcessingPredicate(furthestOriginalIndex, matchLengthOfLongest)) {
          quitEarlyArr[0] = true;
          midOriginalArr[0] = furthestOriginalIndex;
          midModifiedArr[0] = furthestModifiedIndex;
          if (matchLengthOfLongest > 0 && 1447 > 0 && numDifferences <= 1447 + 1) {
            return this.WALKTRACE(diagonalForwardBase, diagonalForwardStart, diagonalForwardEnd, diagonalForwardOffset, diagonalReverseBase, diagonalReverseStart, diagonalReverseEnd, diagonalReverseOffset, forwardPoints, reversePoints, originalIndex, originalEnd, midOriginalArr, modifiedIndex, modifiedEnd, midModifiedArr, deltaIsEven, quitEarlyArr);
          } else {
            originalStart++;
            modifiedStart++;
            return [
              new DiffChange(originalStart, originalEnd - originalStart + 1, modifiedStart, modifiedEnd - modifiedStart + 1)
            ];
          }
        }
        diagonalReverseStart = this.ClipDiagonalBound(diagonalReverseBase - numDifferences, numDifferences, diagonalReverseBase, numDiagonals);
        diagonalReverseEnd = this.ClipDiagonalBound(diagonalReverseBase + numDifferences, numDifferences, diagonalReverseBase, numDiagonals);
        for (let diagonal = diagonalReverseStart; diagonal <= diagonalReverseEnd; diagonal += 2) {
          if (diagonal === diagonalReverseStart || diagonal < diagonalReverseEnd && reversePoints[diagonal - 1] >= reversePoints[diagonal + 1]) {
            originalIndex = reversePoints[diagonal + 1] - 1;
          } else {
            originalIndex = reversePoints[diagonal - 1];
          }
          modifiedIndex = originalIndex - (diagonal - diagonalReverseBase) - diagonalReverseOffset;
          const tempOriginalIndex = originalIndex;
          while (originalIndex > originalStart && modifiedIndex > modifiedStart && this.ElementsAreEqual(originalIndex, modifiedIndex)) {
            originalIndex--;
            modifiedIndex--;
          }
          reversePoints[diagonal] = originalIndex;
          if (deltaIsEven && Math.abs(diagonal - diagonalForwardBase) <= numDifferences) {
            if (originalIndex <= forwardPoints[diagonal]) {
              midOriginalArr[0] = originalIndex;
              midModifiedArr[0] = modifiedIndex;
              if (tempOriginalIndex >= forwardPoints[diagonal] && 1447 > 0 && numDifferences <= 1447 + 1) {
                return this.WALKTRACE(diagonalForwardBase, diagonalForwardStart, diagonalForwardEnd, diagonalForwardOffset, diagonalReverseBase, diagonalReverseStart, diagonalReverseEnd, diagonalReverseOffset, forwardPoints, reversePoints, originalIndex, originalEnd, midOriginalArr, modifiedIndex, modifiedEnd, midModifiedArr, deltaIsEven, quitEarlyArr);
              } else {
                return null;
              }
            }
          }
        }
        if (numDifferences <= 1447) {
          let temp = new Int32Array(diagonalForwardEnd - diagonalForwardStart + 2);
          temp[0] = diagonalForwardBase - diagonalForwardStart + 1;
          MyArray.Copy2(forwardPoints, diagonalForwardStart, temp, 1, diagonalForwardEnd - diagonalForwardStart + 1);
          this.m_forwardHistory.push(temp);
          temp = new Int32Array(diagonalReverseEnd - diagonalReverseStart + 2);
          temp[0] = diagonalReverseBase - diagonalReverseStart + 1;
          MyArray.Copy2(reversePoints, diagonalReverseStart, temp, 1, diagonalReverseEnd - diagonalReverseStart + 1);
          this.m_reverseHistory.push(temp);
        }
      }
      return this.WALKTRACE(diagonalForwardBase, diagonalForwardStart, diagonalForwardEnd, diagonalForwardOffset, diagonalReverseBase, diagonalReverseStart, diagonalReverseEnd, diagonalReverseOffset, forwardPoints, reversePoints, originalIndex, originalEnd, midOriginalArr, modifiedIndex, modifiedEnd, midModifiedArr, deltaIsEven, quitEarlyArr);
    }
    PrettifyChanges(changes) {
      for (let i = 0; i < changes.length; i++) {
        const change = changes[i];
        const originalStop = i < changes.length - 1 ? changes[i + 1].originalStart : this._originalElementsOrHash.length;
        const modifiedStop = i < changes.length - 1 ? changes[i + 1].modifiedStart : this._modifiedElementsOrHash.length;
        const checkOriginal = change.originalLength > 0;
        const checkModified = change.modifiedLength > 0;
        while (change.originalStart + change.originalLength < originalStop && change.modifiedStart + change.modifiedLength < modifiedStop && (!checkOriginal || this.OriginalElementsAreEqual(change.originalStart, change.originalStart + change.originalLength)) && (!checkModified || this.ModifiedElementsAreEqual(change.modifiedStart, change.modifiedStart + change.modifiedLength))) {
          change.originalStart++;
          change.modifiedStart++;
        }
        let mergedChangeArr = [null];
        if (i < changes.length - 1 && this.ChangesOverlap(changes[i], changes[i + 1], mergedChangeArr)) {
          changes[i] = mergedChangeArr[0];
          changes.splice(i + 1, 1);
          i--;
          continue;
        }
      }
      for (let i = changes.length - 1; i >= 0; i--) {
        const change = changes[i];
        let originalStop = 0;
        let modifiedStop = 0;
        if (i > 0) {
          const prevChange = changes[i - 1];
          if (prevChange.originalLength > 0) {
            originalStop = prevChange.originalStart + prevChange.originalLength;
          }
          if (prevChange.modifiedLength > 0) {
            modifiedStop = prevChange.modifiedStart + prevChange.modifiedLength;
          }
        }
        const checkOriginal = change.originalLength > 0;
        const checkModified = change.modifiedLength > 0;
        let bestDelta = 0;
        let bestScore = this._boundaryScore(change.originalStart, change.originalLength, change.modifiedStart, change.modifiedLength);
        for (let delta = 1; ; delta++) {
          const originalStart = change.originalStart - delta;
          const modifiedStart = change.modifiedStart - delta;
          if (originalStart < originalStop || modifiedStart < modifiedStop) {
            break;
          }
          if (checkOriginal && !this.OriginalElementsAreEqual(originalStart, originalStart + change.originalLength)) {
            break;
          }
          if (checkModified && !this.ModifiedElementsAreEqual(modifiedStart, modifiedStart + change.modifiedLength)) {
            break;
          }
          const score = this._boundaryScore(originalStart, change.originalLength, modifiedStart, change.modifiedLength);
          if (score > bestScore) {
            bestScore = score;
            bestDelta = delta;
          }
        }
        change.originalStart -= bestDelta;
        change.modifiedStart -= bestDelta;
      }
      return changes;
    }
    _OriginalIsBoundary(index) {
      if (index <= 0 || index >= this._originalElementsOrHash.length - 1) {
        return true;
      }
      return this._hasStrings && /^\s*$/.test(this._originalStringElements[index]);
    }
    _OriginalRegionIsBoundary(originalStart, originalLength) {
      if (this._OriginalIsBoundary(originalStart) || this._OriginalIsBoundary(originalStart - 1)) {
        return true;
      }
      if (originalLength > 0) {
        const originalEnd = originalStart + originalLength;
        if (this._OriginalIsBoundary(originalEnd - 1) || this._OriginalIsBoundary(originalEnd)) {
          return true;
        }
      }
      return false;
    }
    _ModifiedIsBoundary(index) {
      if (index <= 0 || index >= this._modifiedElementsOrHash.length - 1) {
        return true;
      }
      return this._hasStrings && /^\s*$/.test(this._modifiedStringElements[index]);
    }
    _ModifiedRegionIsBoundary(modifiedStart, modifiedLength) {
      if (this._ModifiedIsBoundary(modifiedStart) || this._ModifiedIsBoundary(modifiedStart - 1)) {
        return true;
      }
      if (modifiedLength > 0) {
        const modifiedEnd = modifiedStart + modifiedLength;
        if (this._ModifiedIsBoundary(modifiedEnd - 1) || this._ModifiedIsBoundary(modifiedEnd)) {
          return true;
        }
      }
      return false;
    }
    _boundaryScore(originalStart, originalLength, modifiedStart, modifiedLength) {
      const originalScore = this._OriginalRegionIsBoundary(originalStart, originalLength) ? 1 : 0;
      const modifiedScore = this._ModifiedRegionIsBoundary(modifiedStart, modifiedLength) ? 1 : 0;
      return originalScore + modifiedScore;
    }
    ConcatenateChanges(left, right) {
      let mergedChangeArr = [];
      if (left.length === 0 || right.length === 0) {
        return right.length > 0 ? right : left;
      } else if (this.ChangesOverlap(left[left.length - 1], right[0], mergedChangeArr)) {
        const result = new Array(left.length + right.length - 1);
        MyArray.Copy(left, 0, result, 0, left.length - 1);
        result[left.length - 1] = mergedChangeArr[0];
        MyArray.Copy(right, 1, result, left.length, right.length - 1);
        return result;
      } else {
        const result = new Array(left.length + right.length);
        MyArray.Copy(left, 0, result, 0, left.length);
        MyArray.Copy(right, 0, result, left.length, right.length);
        return result;
      }
    }
    ChangesOverlap(left, right, mergedChangeArr) {
      Debug.Assert(left.originalStart <= right.originalStart, "Left change is not less than or equal to right change");
      Debug.Assert(left.modifiedStart <= right.modifiedStart, "Left change is not less than or equal to right change");
      if (left.originalStart + left.originalLength >= right.originalStart || left.modifiedStart + left.modifiedLength >= right.modifiedStart) {
        const originalStart = left.originalStart;
        let originalLength = left.originalLength;
        const modifiedStart = left.modifiedStart;
        let modifiedLength = left.modifiedLength;
        if (left.originalStart + left.originalLength >= right.originalStart) {
          originalLength = right.originalStart + right.originalLength - left.originalStart;
        }
        if (left.modifiedStart + left.modifiedLength >= right.modifiedStart) {
          modifiedLength = right.modifiedStart + right.modifiedLength - left.modifiedStart;
        }
        mergedChangeArr[0] = new DiffChange(originalStart, originalLength, modifiedStart, modifiedLength);
        return true;
      } else {
        mergedChangeArr[0] = null;
        return false;
      }
    }
    ClipDiagonalBound(diagonal, numDifferences, diagonalBaseIndex, numDiagonals) {
      if (diagonal >= 0 && diagonal < numDiagonals) {
        return diagonal;
      }
      const diagonalsBelow = diagonalBaseIndex;
      const diagonalsAbove = numDiagonals - diagonalBaseIndex - 1;
      const diffEven = numDifferences % 2 === 0;
      if (diagonal < 0) {
        const lowerBoundEven = diagonalsBelow % 2 === 0;
        return diffEven === lowerBoundEven ? 0 : 1;
      } else {
        const upperBoundEven = diagonalsAbove % 2 === 0;
        return diffEven === upperBoundEven ? numDiagonals - 1 : numDiagonals - 2;
      }
    }
  };

  // node_modules/monaco-editor/esm/vs/base/common/process.js
  var safeProcess = typeof process === "undefined" ? {
    cwd() {
      return "/";
    },
    env: /* @__PURE__ */ Object.create(null),
    get platform() {
      return isWindows ? "win32" : isMacintosh ? "darwin" : "linux";
    },
    nextTick(callback) {
      return setImmediate(callback);
    }
  } : process;
  var cwd = safeProcess.cwd;
  var env = safeProcess.env;
  var platform = safeProcess.platform;

  // node_modules/monaco-editor/esm/vs/base/common/path.js
  var CHAR_UPPERCASE_A = 65;
  var CHAR_LOWERCASE_A = 97;
  var CHAR_UPPERCASE_Z = 90;
  var CHAR_LOWERCASE_Z = 122;
  var CHAR_DOT = 46;
  var CHAR_FORWARD_SLASH = 47;
  var CHAR_BACKWARD_SLASH = 92;
  var CHAR_COLON = 58;
  var CHAR_QUESTION_MARK = 63;
  var ErrorInvalidArgType = class extends Error {
    constructor(name, expected, actual) {
      let determiner;
      if (typeof expected === "string" && expected.indexOf("not ") === 0) {
        determiner = "must not be";
        expected = expected.replace(/^not /, "");
      } else {
        determiner = "must be";
      }
      const type = name.indexOf(".") !== -1 ? "property" : "argument";
      let msg = `The "${name}" ${type} ${determiner} of type ${expected}`;
      msg += `. Received type ${typeof actual}`;
      super(msg);
      this.code = "ERR_INVALID_ARG_TYPE";
    }
  };
  function validateString(value, name) {
    if (typeof value !== "string") {
      throw new ErrorInvalidArgType(name, "string", value);
    }
  }
  function isPathSeparator(code) {
    return code === CHAR_FORWARD_SLASH || code === CHAR_BACKWARD_SLASH;
  }
  function isPosixPathSeparator(code) {
    return code === CHAR_FORWARD_SLASH;
  }
  function isWindowsDeviceRoot(code) {
    return code >= CHAR_UPPERCASE_A && code <= CHAR_UPPERCASE_Z || code >= CHAR_LOWERCASE_A && code <= CHAR_LOWERCASE_Z;
  }
  function normalizeString(path, allowAboveRoot, separator, isPathSeparator2) {
    let res = "";
    let lastSegmentLength = 0;
    let lastSlash = -1;
    let dots = 0;
    let code = 0;
    for (let i = 0; i <= path.length; ++i) {
      if (i < path.length) {
        code = path.charCodeAt(i);
      } else if (isPathSeparator2(code)) {
        break;
      } else {
        code = CHAR_FORWARD_SLASH;
      }
      if (isPathSeparator2(code)) {
        if (lastSlash === i - 1 || dots === 1) {
        } else if (dots === 2) {
          if (res.length < 2 || lastSegmentLength !== 2 || res.charCodeAt(res.length - 1) !== CHAR_DOT || res.charCodeAt(res.length - 2) !== CHAR_DOT) {
            if (res.length > 2) {
              const lastSlashIndex = res.lastIndexOf(separator);
              if (lastSlashIndex === -1) {
                res = "";
                lastSegmentLength = 0;
              } else {
                res = res.slice(0, lastSlashIndex);
                lastSegmentLength = res.length - 1 - res.lastIndexOf(separator);
              }
              lastSlash = i;
              dots = 0;
              continue;
            } else if (res.length !== 0) {
              res = "";
              lastSegmentLength = 0;
              lastSlash = i;
              dots = 0;
              continue;
            }
          }
          if (allowAboveRoot) {
            res += res.length > 0 ? `${separator}..` : "..";
            lastSegmentLength = 2;
          }
        } else {
          if (res.length > 0) {
            res += `${separator}${path.slice(lastSlash + 1, i)}`;
          } else {
            res = path.slice(lastSlash + 1, i);
          }
          lastSegmentLength = i - lastSlash - 1;
        }
        lastSlash = i;
        dots = 0;
      } else if (code === CHAR_DOT && dots !== -1) {
        ++dots;
      } else {
        dots = -1;
      }
    }
    return res;
  }
  function _format(sep2, pathObject) {
    if (pathObject === null || typeof pathObject !== "object") {
      throw new ErrorInvalidArgType("pathObject", "Object", pathObject);
    }
    const dir = pathObject.dir || pathObject.root;
    const base = pathObject.base || `${pathObject.name || ""}${pathObject.ext || ""}`;
    if (!dir) {
      return base;
    }
    return dir === pathObject.root ? `${dir}${base}` : `${dir}${sep2}${base}`;
  }
  var win32 = {
    resolve(...pathSegments) {
      let resolvedDevice = "";
      let resolvedTail = "";
      let resolvedAbsolute = false;
      for (let i = pathSegments.length - 1; i >= -1; i--) {
        let path;
        if (i >= 0) {
          path = pathSegments[i];
          validateString(path, "path");
          if (path.length === 0) {
            continue;
          }
        } else if (resolvedDevice.length === 0) {
          path = cwd();
        } else {
          path = env[`=${resolvedDevice}`] || cwd();
          if (path === void 0 || path.slice(0, 2).toLowerCase() !== resolvedDevice.toLowerCase() && path.charCodeAt(2) === CHAR_BACKWARD_SLASH) {
            path = `${resolvedDevice}\\`;
          }
        }
        const len = path.length;
        let rootEnd = 0;
        let device = "";
        let isAbsolute = false;
        const code = path.charCodeAt(0);
        if (len === 1) {
          if (isPathSeparator(code)) {
            rootEnd = 1;
            isAbsolute = true;
          }
        } else if (isPathSeparator(code)) {
          isAbsolute = true;
          if (isPathSeparator(path.charCodeAt(1))) {
            let j = 2;
            let last = j;
            while (j < len && !isPathSeparator(path.charCodeAt(j))) {
              j++;
            }
            if (j < len && j !== last) {
              const firstPart = path.slice(last, j);
              last = j;
              while (j < len && isPathSeparator(path.charCodeAt(j))) {
                j++;
              }
              if (j < len && j !== last) {
                last = j;
                while (j < len && !isPathSeparator(path.charCodeAt(j))) {
                  j++;
                }
                if (j === len || j !== last) {
                  device = `\\\\${firstPart}\\${path.slice(last, j)}`;
                  rootEnd = j;
                }
              }
            }
          } else {
            rootEnd = 1;
          }
        } else if (isWindowsDeviceRoot(code) && path.charCodeAt(1) === CHAR_COLON) {
          device = path.slice(0, 2);
          rootEnd = 2;
          if (len > 2 && isPathSeparator(path.charCodeAt(2))) {
            isAbsolute = true;
            rootEnd = 3;
          }
        }
        if (device.length > 0) {
          if (resolvedDevice.length > 0) {
            if (device.toLowerCase() !== resolvedDevice.toLowerCase()) {
              continue;
            }
          } else {
            resolvedDevice = device;
          }
        }
        if (resolvedAbsolute) {
          if (resolvedDevice.length > 0) {
            break;
          }
        } else {
          resolvedTail = `${path.slice(rootEnd)}\\${resolvedTail}`;
          resolvedAbsolute = isAbsolute;
          if (isAbsolute && resolvedDevice.length > 0) {
            break;
          }
        }
      }
      resolvedTail = normalizeString(resolvedTail, !resolvedAbsolute, "\\", isPathSeparator);
      return resolvedAbsolute ? `${resolvedDevice}\\${resolvedTail}` : `${resolvedDevice}${resolvedTail}` || ".";
    },
    normalize(path) {
      validateString(path, "path");
      const len = path.length;
      if (len === 0) {
        return ".";
      }
      let rootEnd = 0;
      let device;
      let isAbsolute = false;
      const code = path.charCodeAt(0);
      if (len === 1) {
        return isPosixPathSeparator(code) ? "\\" : path;
      }
      if (isPathSeparator(code)) {
        isAbsolute = true;
        if (isPathSeparator(path.charCodeAt(1))) {
          let j = 2;
          let last = j;
          while (j < len && !isPathSeparator(path.charCodeAt(j))) {
            j++;
          }
          if (j < len && j !== last) {
            const firstPart = path.slice(last, j);
            last = j;
            while (j < len && isPathSeparator(path.charCodeAt(j))) {
              j++;
            }
            if (j < len && j !== last) {
              last = j;
              while (j < len && !isPathSeparator(path.charCodeAt(j))) {
                j++;
              }
              if (j === len) {
                return `\\\\${firstPart}\\${path.slice(last)}\\`;
              }
              if (j !== last) {
                device = `\\\\${firstPart}\\${path.slice(last, j)}`;
                rootEnd = j;
              }
            }
          }
        } else {
          rootEnd = 1;
        }
      } else if (isWindowsDeviceRoot(code) && path.charCodeAt(1) === CHAR_COLON) {
        device = path.slice(0, 2);
        rootEnd = 2;
        if (len > 2 && isPathSeparator(path.charCodeAt(2))) {
          isAbsolute = true;
          rootEnd = 3;
        }
      }
      let tail = rootEnd < len ? normalizeString(path.slice(rootEnd), !isAbsolute, "\\", isPathSeparator) : "";
      if (tail.length === 0 && !isAbsolute) {
        tail = ".";
      }
      if (tail.length > 0 && isPathSeparator(path.charCodeAt(len - 1))) {
        tail += "\\";
      }
      if (device === void 0) {
        return isAbsolute ? `\\${tail}` : tail;
      }
      return isAbsolute ? `${device}\\${tail}` : `${device}${tail}`;
    },
    isAbsolute(path) {
      validateString(path, "path");
      const len = path.length;
      if (len === 0) {
        return false;
      }
      const code = path.charCodeAt(0);
      return isPathSeparator(code) || len > 2 && isWindowsDeviceRoot(code) && path.charCodeAt(1) === CHAR_COLON && isPathSeparator(path.charCodeAt(2));
    },
    join(...paths) {
      if (paths.length === 0) {
        return ".";
      }
      let joined;
      let firstPart;
      for (let i = 0; i < paths.length; ++i) {
        const arg = paths[i];
        validateString(arg, "path");
        if (arg.length > 0) {
          if (joined === void 0) {
            joined = firstPart = arg;
          } else {
            joined += `\\${arg}`;
          }
        }
      }
      if (joined === void 0) {
        return ".";
      }
      let needsReplace = true;
      let slashCount = 0;
      if (typeof firstPart === "string" && isPathSeparator(firstPart.charCodeAt(0))) {
        ++slashCount;
        const firstLen = firstPart.length;
        if (firstLen > 1 && isPathSeparator(firstPart.charCodeAt(1))) {
          ++slashCount;
          if (firstLen > 2) {
            if (isPathSeparator(firstPart.charCodeAt(2))) {
              ++slashCount;
            } else {
              needsReplace = false;
            }
          }
        }
      }
      if (needsReplace) {
        while (slashCount < joined.length && isPathSeparator(joined.charCodeAt(slashCount))) {
          slashCount++;
        }
        if (slashCount >= 2) {
          joined = `\\${joined.slice(slashCount)}`;
        }
      }
      return win32.normalize(joined);
    },
    relative(from, to) {
      validateString(from, "from");
      validateString(to, "to");
      if (from === to) {
        return "";
      }
      const fromOrig = win32.resolve(from);
      const toOrig = win32.resolve(to);
      if (fromOrig === toOrig) {
        return "";
      }
      from = fromOrig.toLowerCase();
      to = toOrig.toLowerCase();
      if (from === to) {
        return "";
      }
      let fromStart = 0;
      while (fromStart < from.length && from.charCodeAt(fromStart) === CHAR_BACKWARD_SLASH) {
        fromStart++;
      }
      let fromEnd = from.length;
      while (fromEnd - 1 > fromStart && from.charCodeAt(fromEnd - 1) === CHAR_BACKWARD_SLASH) {
        fromEnd--;
      }
      const fromLen = fromEnd - fromStart;
      let toStart = 0;
      while (toStart < to.length && to.charCodeAt(toStart) === CHAR_BACKWARD_SLASH) {
        toStart++;
      }
      let toEnd = to.length;
      while (toEnd - 1 > toStart && to.charCodeAt(toEnd - 1) === CHAR_BACKWARD_SLASH) {
        toEnd--;
      }
      const toLen = toEnd - toStart;
      const length = fromLen < toLen ? fromLen : toLen;
      let lastCommonSep = -1;
      let i = 0;
      for (; i < length; i++) {
        const fromCode = from.charCodeAt(fromStart + i);
        if (fromCode !== to.charCodeAt(toStart + i)) {
          break;
        } else if (fromCode === CHAR_BACKWARD_SLASH) {
          lastCommonSep = i;
        }
      }
      if (i !== length) {
        if (lastCommonSep === -1) {
          return toOrig;
        }
      } else {
        if (toLen > length) {
          if (to.charCodeAt(toStart + i) === CHAR_BACKWARD_SLASH) {
            return toOrig.slice(toStart + i + 1);
          }
          if (i === 2) {
            return toOrig.slice(toStart + i);
          }
        }
        if (fromLen > length) {
          if (from.charCodeAt(fromStart + i) === CHAR_BACKWARD_SLASH) {
            lastCommonSep = i;
          } else if (i === 2) {
            lastCommonSep = 3;
          }
        }
        if (lastCommonSep === -1) {
          lastCommonSep = 0;
        }
      }
      let out = "";
      for (i = fromStart + lastCommonSep + 1; i <= fromEnd; ++i) {
        if (i === fromEnd || from.charCodeAt(i) === CHAR_BACKWARD_SLASH) {
          out += out.length === 0 ? ".." : "\\..";
        }
      }
      toStart += lastCommonSep;
      if (out.length > 0) {
        return `${out}${toOrig.slice(toStart, toEnd)}`;
      }
      if (toOrig.charCodeAt(toStart) === CHAR_BACKWARD_SLASH) {
        ++toStart;
      }
      return toOrig.slice(toStart, toEnd);
    },
    toNamespacedPath(path) {
      if (typeof path !== "string") {
        return path;
      }
      if (path.length === 0) {
        return "";
      }
      const resolvedPath = win32.resolve(path);
      if (resolvedPath.length <= 2) {
        return path;
      }
      if (resolvedPath.charCodeAt(0) === CHAR_BACKWARD_SLASH) {
        if (resolvedPath.charCodeAt(1) === CHAR_BACKWARD_SLASH) {
          const code = resolvedPath.charCodeAt(2);
          if (code !== CHAR_QUESTION_MARK && code !== CHAR_DOT) {
            return `\\\\?\\UNC\\${resolvedPath.slice(2)}`;
          }
        }
      } else if (isWindowsDeviceRoot(resolvedPath.charCodeAt(0)) && resolvedPath.charCodeAt(1) === CHAR_COLON && resolvedPath.charCodeAt(2) === CHAR_BACKWARD_SLASH) {
        return `\\\\?\\${resolvedPath}`;
      }
      return path;
    },
    dirname(path) {
      validateString(path, "path");
      const len = path.length;
      if (len === 0) {
        return ".";
      }
      let rootEnd = -1;
      let offset = 0;
      const code = path.charCodeAt(0);
      if (len === 1) {
        return isPathSeparator(code) ? path : ".";
      }
      if (isPathSeparator(code)) {
        rootEnd = offset = 1;
        if (isPathSeparator(path.charCodeAt(1))) {
          let j = 2;
          let last = j;
          while (j < len && !isPathSeparator(path.charCodeAt(j))) {
            j++;
          }
          if (j < len && j !== last) {
            last = j;
            while (j < len && isPathSeparator(path.charCodeAt(j))) {
              j++;
            }
            if (j < len && j !== last) {
              last = j;
              while (j < len && !isPathSeparator(path.charCodeAt(j))) {
                j++;
              }
              if (j === len) {
                return path;
              }
              if (j !== last) {
                rootEnd = offset = j + 1;
              }
            }
          }
        }
      } else if (isWindowsDeviceRoot(code) && path.charCodeAt(1) === CHAR_COLON) {
        rootEnd = len > 2 && isPathSeparator(path.charCodeAt(2)) ? 3 : 2;
        offset = rootEnd;
      }
      let end = -1;
      let matchedSlash = true;
      for (let i = len - 1; i >= offset; --i) {
        if (isPathSeparator(path.charCodeAt(i))) {
          if (!matchedSlash) {
            end = i;
            break;
          }
        } else {
          matchedSlash = false;
        }
      }
      if (end === -1) {
        if (rootEnd === -1) {
          return ".";
        }
        end = rootEnd;
      }
      return path.slice(0, end);
    },
    basename(path, ext) {
      if (ext !== void 0) {
        validateString(ext, "ext");
      }
      validateString(path, "path");
      let start = 0;
      let end = -1;
      let matchedSlash = true;
      let i;
      if (path.length >= 2 && isWindowsDeviceRoot(path.charCodeAt(0)) && path.charCodeAt(1) === CHAR_COLON) {
        start = 2;
      }
      if (ext !== void 0 && ext.length > 0 && ext.length <= path.length) {
        if (ext === path) {
          return "";
        }
        let extIdx = ext.length - 1;
        let firstNonSlashEnd = -1;
        for (i = path.length - 1; i >= start; --i) {
          const code = path.charCodeAt(i);
          if (isPathSeparator(code)) {
            if (!matchedSlash) {
              start = i + 1;
              break;
            }
          } else {
            if (firstNonSlashEnd === -1) {
              matchedSlash = false;
              firstNonSlashEnd = i + 1;
            }
            if (extIdx >= 0) {
              if (code === ext.charCodeAt(extIdx)) {
                if (--extIdx === -1) {
                  end = i;
                }
              } else {
                extIdx = -1;
                end = firstNonSlashEnd;
              }
            }
          }
        }
        if (start === end) {
          end = firstNonSlashEnd;
        } else if (end === -1) {
          end = path.length;
        }
        return path.slice(start, end);
      }
      for (i = path.length - 1; i >= start; --i) {
        if (isPathSeparator(path.charCodeAt(i))) {
          if (!matchedSlash) {
            start = i + 1;
            break;
          }
        } else if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
      }
      if (end === -1) {
        return "";
      }
      return path.slice(start, end);
    },
    extname(path) {
      validateString(path, "path");
      let start = 0;
      let startDot = -1;
      let startPart = 0;
      let end = -1;
      let matchedSlash = true;
      let preDotState = 0;
      if (path.length >= 2 && path.charCodeAt(1) === CHAR_COLON && isWindowsDeviceRoot(path.charCodeAt(0))) {
        start = startPart = 2;
      }
      for (let i = path.length - 1; i >= start; --i) {
        const code = path.charCodeAt(i);
        if (isPathSeparator(code)) {
          if (!matchedSlash) {
            startPart = i + 1;
            break;
          }
          continue;
        }
        if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
        if (code === CHAR_DOT) {
          if (startDot === -1) {
            startDot = i;
          } else if (preDotState !== 1) {
            preDotState = 1;
          }
        } else if (startDot !== -1) {
          preDotState = -1;
        }
      }
      if (startDot === -1 || end === -1 || preDotState === 0 || preDotState === 1 && startDot === end - 1 && startDot === startPart + 1) {
        return "";
      }
      return path.slice(startDot, end);
    },
    format: _format.bind(null, "\\"),
    parse(path) {
      validateString(path, "path");
      const ret = { root: "", dir: "", base: "", ext: "", name: "" };
      if (path.length === 0) {
        return ret;
      }
      const len = path.length;
      let rootEnd = 0;
      let code = path.charCodeAt(0);
      if (len === 1) {
        if (isPathSeparator(code)) {
          ret.root = ret.dir = path;
          return ret;
        }
        ret.base = ret.name = path;
        return ret;
      }
      if (isPathSeparator(code)) {
        rootEnd = 1;
        if (isPathSeparator(path.charCodeAt(1))) {
          let j = 2;
          let last = j;
          while (j < len && !isPathSeparator(path.charCodeAt(j))) {
            j++;
          }
          if (j < len && j !== last) {
            last = j;
            while (j < len && isPathSeparator(path.charCodeAt(j))) {
              j++;
            }
            if (j < len && j !== last) {
              last = j;
              while (j < len && !isPathSeparator(path.charCodeAt(j))) {
                j++;
              }
              if (j === len) {
                rootEnd = j;
              } else if (j !== last) {
                rootEnd = j + 1;
              }
            }
          }
        }
      } else if (isWindowsDeviceRoot(code) && path.charCodeAt(1) === CHAR_COLON) {
        if (len <= 2) {
          ret.root = ret.dir = path;
          return ret;
        }
        rootEnd = 2;
        if (isPathSeparator(path.charCodeAt(2))) {
          if (len === 3) {
            ret.root = ret.dir = path;
            return ret;
          }
          rootEnd = 3;
        }
      }
      if (rootEnd > 0) {
        ret.root = path.slice(0, rootEnd);
      }
      let startDot = -1;
      let startPart = rootEnd;
      let end = -1;
      let matchedSlash = true;
      let i = path.length - 1;
      let preDotState = 0;
      for (; i >= rootEnd; --i) {
        code = path.charCodeAt(i);
        if (isPathSeparator(code)) {
          if (!matchedSlash) {
            startPart = i + 1;
            break;
          }
          continue;
        }
        if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
        if (code === CHAR_DOT) {
          if (startDot === -1) {
            startDot = i;
          } else if (preDotState !== 1) {
            preDotState = 1;
          }
        } else if (startDot !== -1) {
          preDotState = -1;
        }
      }
      if (end !== -1) {
        if (startDot === -1 || preDotState === 0 || preDotState === 1 && startDot === end - 1 && startDot === startPart + 1) {
          ret.base = ret.name = path.slice(startPart, end);
        } else {
          ret.name = path.slice(startPart, startDot);
          ret.base = path.slice(startPart, end);
          ret.ext = path.slice(startDot, end);
        }
      }
      if (startPart > 0 && startPart !== rootEnd) {
        ret.dir = path.slice(0, startPart - 1);
      } else {
        ret.dir = ret.root;
      }
      return ret;
    },
    sep: "\\",
    delimiter: ";",
    win32: null,
    posix: null
  };
  var posix = {
    resolve(...pathSegments) {
      let resolvedPath = "";
      let resolvedAbsolute = false;
      for (let i = pathSegments.length - 1; i >= -1 && !resolvedAbsolute; i--) {
        const path = i >= 0 ? pathSegments[i] : cwd();
        validateString(path, "path");
        if (path.length === 0) {
          continue;
        }
        resolvedPath = `${path}/${resolvedPath}`;
        resolvedAbsolute = path.charCodeAt(0) === CHAR_FORWARD_SLASH;
      }
      resolvedPath = normalizeString(resolvedPath, !resolvedAbsolute, "/", isPosixPathSeparator);
      if (resolvedAbsolute) {
        return `/${resolvedPath}`;
      }
      return resolvedPath.length > 0 ? resolvedPath : ".";
    },
    normalize(path) {
      validateString(path, "path");
      if (path.length === 0) {
        return ".";
      }
      const isAbsolute = path.charCodeAt(0) === CHAR_FORWARD_SLASH;
      const trailingSeparator = path.charCodeAt(path.length - 1) === CHAR_FORWARD_SLASH;
      path = normalizeString(path, !isAbsolute, "/", isPosixPathSeparator);
      if (path.length === 0) {
        if (isAbsolute) {
          return "/";
        }
        return trailingSeparator ? "./" : ".";
      }
      if (trailingSeparator) {
        path += "/";
      }
      return isAbsolute ? `/${path}` : path;
    },
    isAbsolute(path) {
      validateString(path, "path");
      return path.length > 0 && path.charCodeAt(0) === CHAR_FORWARD_SLASH;
    },
    join(...paths) {
      if (paths.length === 0) {
        return ".";
      }
      let joined;
      for (let i = 0; i < paths.length; ++i) {
        const arg = paths[i];
        validateString(arg, "path");
        if (arg.length > 0) {
          if (joined === void 0) {
            joined = arg;
          } else {
            joined += `/${arg}`;
          }
        }
      }
      if (joined === void 0) {
        return ".";
      }
      return posix.normalize(joined);
    },
    relative(from, to) {
      validateString(from, "from");
      validateString(to, "to");
      if (from === to) {
        return "";
      }
      from = posix.resolve(from);
      to = posix.resolve(to);
      if (from === to) {
        return "";
      }
      const fromStart = 1;
      const fromEnd = from.length;
      const fromLen = fromEnd - fromStart;
      const toStart = 1;
      const toLen = to.length - toStart;
      const length = fromLen < toLen ? fromLen : toLen;
      let lastCommonSep = -1;
      let i = 0;
      for (; i < length; i++) {
        const fromCode = from.charCodeAt(fromStart + i);
        if (fromCode !== to.charCodeAt(toStart + i)) {
          break;
        } else if (fromCode === CHAR_FORWARD_SLASH) {
          lastCommonSep = i;
        }
      }
      if (i === length) {
        if (toLen > length) {
          if (to.charCodeAt(toStart + i) === CHAR_FORWARD_SLASH) {
            return to.slice(toStart + i + 1);
          }
          if (i === 0) {
            return to.slice(toStart + i);
          }
        } else if (fromLen > length) {
          if (from.charCodeAt(fromStart + i) === CHAR_FORWARD_SLASH) {
            lastCommonSep = i;
          } else if (i === 0) {
            lastCommonSep = 0;
          }
        }
      }
      let out = "";
      for (i = fromStart + lastCommonSep + 1; i <= fromEnd; ++i) {
        if (i === fromEnd || from.charCodeAt(i) === CHAR_FORWARD_SLASH) {
          out += out.length === 0 ? ".." : "/..";
        }
      }
      return `${out}${to.slice(toStart + lastCommonSep)}`;
    },
    toNamespacedPath(path) {
      return path;
    },
    dirname(path) {
      validateString(path, "path");
      if (path.length === 0) {
        return ".";
      }
      const hasRoot = path.charCodeAt(0) === CHAR_FORWARD_SLASH;
      let end = -1;
      let matchedSlash = true;
      for (let i = path.length - 1; i >= 1; --i) {
        if (path.charCodeAt(i) === CHAR_FORWARD_SLASH) {
          if (!matchedSlash) {
            end = i;
            break;
          }
        } else {
          matchedSlash = false;
        }
      }
      if (end === -1) {
        return hasRoot ? "/" : ".";
      }
      if (hasRoot && end === 1) {
        return "//";
      }
      return path.slice(0, end);
    },
    basename(path, ext) {
      if (ext !== void 0) {
        validateString(ext, "ext");
      }
      validateString(path, "path");
      let start = 0;
      let end = -1;
      let matchedSlash = true;
      let i;
      if (ext !== void 0 && ext.length > 0 && ext.length <= path.length) {
        if (ext === path) {
          return "";
        }
        let extIdx = ext.length - 1;
        let firstNonSlashEnd = -1;
        for (i = path.length - 1; i >= 0; --i) {
          const code = path.charCodeAt(i);
          if (code === CHAR_FORWARD_SLASH) {
            if (!matchedSlash) {
              start = i + 1;
              break;
            }
          } else {
            if (firstNonSlashEnd === -1) {
              matchedSlash = false;
              firstNonSlashEnd = i + 1;
            }
            if (extIdx >= 0) {
              if (code === ext.charCodeAt(extIdx)) {
                if (--extIdx === -1) {
                  end = i;
                }
              } else {
                extIdx = -1;
                end = firstNonSlashEnd;
              }
            }
          }
        }
        if (start === end) {
          end = firstNonSlashEnd;
        } else if (end === -1) {
          end = path.length;
        }
        return path.slice(start, end);
      }
      for (i = path.length - 1; i >= 0; --i) {
        if (path.charCodeAt(i) === CHAR_FORWARD_SLASH) {
          if (!matchedSlash) {
            start = i + 1;
            break;
          }
        } else if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
      }
      if (end === -1) {
        return "";
      }
      return path.slice(start, end);
    },
    extname(path) {
      validateString(path, "path");
      let startDot = -1;
      let startPart = 0;
      let end = -1;
      let matchedSlash = true;
      let preDotState = 0;
      for (let i = path.length - 1; i >= 0; --i) {
        const code = path.charCodeAt(i);
        if (code === CHAR_FORWARD_SLASH) {
          if (!matchedSlash) {
            startPart = i + 1;
            break;
          }
          continue;
        }
        if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
        if (code === CHAR_DOT) {
          if (startDot === -1) {
            startDot = i;
          } else if (preDotState !== 1) {
            preDotState = 1;
          }
        } else if (startDot !== -1) {
          preDotState = -1;
        }
      }
      if (startDot === -1 || end === -1 || preDotState === 0 || preDotState === 1 && startDot === end - 1 && startDot === startPart + 1) {
        return "";
      }
      return path.slice(startDot, end);
    },
    format: _format.bind(null, "/"),
    parse(path) {
      validateString(path, "path");
      const ret = { root: "", dir: "", base: "", ext: "", name: "" };
      if (path.length === 0) {
        return ret;
      }
      const isAbsolute = path.charCodeAt(0) === CHAR_FORWARD_SLASH;
      let start;
      if (isAbsolute) {
        ret.root = "/";
        start = 1;
      } else {
        start = 0;
      }
      let startDot = -1;
      let startPart = 0;
      let end = -1;
      let matchedSlash = true;
      let i = path.length - 1;
      let preDotState = 0;
      for (; i >= start; --i) {
        const code = path.charCodeAt(i);
        if (code === CHAR_FORWARD_SLASH) {
          if (!matchedSlash) {
            startPart = i + 1;
            break;
          }
          continue;
        }
        if (end === -1) {
          matchedSlash = false;
          end = i + 1;
        }
        if (code === CHAR_DOT) {
          if (startDot === -1) {
            startDot = i;
          } else if (preDotState !== 1) {
            preDotState = 1;
          }
        } else if (startDot !== -1) {
          preDotState = -1;
        }
      }
      if (end !== -1) {
        const start2 = startPart === 0 && isAbsolute ? 1 : startPart;
        if (startDot === -1 || preDotState === 0 || preDotState === 1 && startDot === end - 1 && startDot === startPart + 1) {
          ret.base = ret.name = path.slice(start2, end);
        } else {
          ret.name = path.slice(start2, startDot);
          ret.base = path.slice(start2, end);
          ret.ext = path.slice(startDot, end);
        }
      }
      if (startPart > 0) {
        ret.dir = path.slice(0, startPart - 1);
      } else if (isAbsolute) {
        ret.dir = "/";
      }
      return ret;
    },
    sep: "/",
    delimiter: ":",
    win32: null,
    posix: null
  };
  posix.win32 = win32.win32 = win32;
  posix.posix = win32.posix = posix;
  var normalize = platform === "win32" ? win32.normalize : posix.normalize;
  var resolve = platform === "win32" ? win32.resolve : posix.resolve;
  var relative = platform === "win32" ? win32.relative : posix.relative;
  var dirname = platform === "win32" ? win32.dirname : posix.dirname;
  var basename = platform === "win32" ? win32.basename : posix.basename;
  var extname = platform === "win32" ? win32.extname : posix.extname;
  var sep = platform === "win32" ? win32.sep : posix.sep;

  // node_modules/monaco-editor/esm/vs/base/common/uri.js
  var _schemePattern = /^\w[\w\d+.-]*$/;
  var _singleSlashStart = /^\//;
  var _doubleSlashStart = /^\/\//;
  function _validateUri(ret, _strict) {
    if (!ret.scheme && _strict) {
      throw new Error(`[UriError]: Scheme is missing: {scheme: "", authority: "${ret.authority}", path: "${ret.path}", query: "${ret.query}", fragment: "${ret.fragment}"}`);
    }
    if (ret.scheme && !_schemePattern.test(ret.scheme)) {
      throw new Error("[UriError]: Scheme contains illegal characters.");
    }
    if (ret.path) {
      if (ret.authority) {
        if (!_singleSlashStart.test(ret.path)) {
          throw new Error('[UriError]: If a URI contains an authority component, then the path component must either be empty or begin with a slash ("/") character');
        }
      } else {
        if (_doubleSlashStart.test(ret.path)) {
          throw new Error('[UriError]: If a URI does not contain an authority component, then the path cannot begin with two slash characters ("//")');
        }
      }
    }
  }
  function _schemeFix(scheme, _strict) {
    if (!scheme && !_strict) {
      return "file";
    }
    return scheme;
  }
  function _referenceResolution(scheme, path) {
    switch (scheme) {
      case "https":
      case "http":
      case "file":
        if (!path) {
          path = _slash;
        } else if (path[0] !== _slash) {
          path = _slash + path;
        }
        break;
    }
    return path;
  }
  var _empty = "";
  var _slash = "/";
  var _regexp = /^(([^:/?#]+?):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/;
  var URI = class {
    constructor(schemeOrData, authority, path, query, fragment, _strict = false) {
      if (typeof schemeOrData === "object") {
        this.scheme = schemeOrData.scheme || _empty;
        this.authority = schemeOrData.authority || _empty;
        this.path = schemeOrData.path || _empty;
        this.query = schemeOrData.query || _empty;
        this.fragment = schemeOrData.fragment || _empty;
      } else {
        this.scheme = _schemeFix(schemeOrData, _strict);
        this.authority = authority || _empty;
        this.path = _referenceResolution(this.scheme, path || _empty);
        this.query = query || _empty;
        this.fragment = fragment || _empty;
        _validateUri(this, _strict);
      }
    }
    static isUri(thing) {
      if (thing instanceof URI) {
        return true;
      }
      if (!thing) {
        return false;
      }
      return typeof thing.authority === "string" && typeof thing.fragment === "string" && typeof thing.path === "string" && typeof thing.query === "string" && typeof thing.scheme === "string" && typeof thing.fsPath === "function" && typeof thing.with === "function" && typeof thing.toString === "function";
    }
    get fsPath() {
      return uriToFsPath(this, false);
    }
    with(change) {
      if (!change) {
        return this;
      }
      let { scheme, authority, path, query, fragment } = change;
      if (scheme === void 0) {
        scheme = this.scheme;
      } else if (scheme === null) {
        scheme = _empty;
      }
      if (authority === void 0) {
        authority = this.authority;
      } else if (authority === null) {
        authority = _empty;
      }
      if (path === void 0) {
        path = this.path;
      } else if (path === null) {
        path = _empty;
      }
      if (query === void 0) {
        query = this.query;
      } else if (query === null) {
        query = _empty;
      }
      if (fragment === void 0) {
        fragment = this.fragment;
      } else if (fragment === null) {
        fragment = _empty;
      }
      if (scheme === this.scheme && authority === this.authority && path === this.path && query === this.query && fragment === this.fragment) {
        return this;
      }
      return new Uri(scheme, authority, path, query, fragment);
    }
    static parse(value, _strict = false) {
      const match = _regexp.exec(value);
      if (!match) {
        return new Uri(_empty, _empty, _empty, _empty, _empty);
      }
      return new Uri(match[2] || _empty, percentDecode(match[4] || _empty), percentDecode(match[5] || _empty), percentDecode(match[7] || _empty), percentDecode(match[9] || _empty), _strict);
    }
    static file(path) {
      let authority = _empty;
      if (isWindows) {
        path = path.replace(/\\/g, _slash);
      }
      if (path[0] === _slash && path[1] === _slash) {
        const idx = path.indexOf(_slash, 2);
        if (idx === -1) {
          authority = path.substring(2);
          path = _slash;
        } else {
          authority = path.substring(2, idx);
          path = path.substring(idx) || _slash;
        }
      }
      return new Uri("file", authority, path, _empty, _empty);
    }
    static from(components) {
      return new Uri(components.scheme, components.authority, components.path, components.query, components.fragment);
    }
    static joinPath(uri, ...pathFragment) {
      if (!uri.path) {
        throw new Error(`[UriError]: cannot call joinPaths on URI without path`);
      }
      let newPath;
      if (isWindows && uri.scheme === "file") {
        newPath = URI.file(win32.join(uriToFsPath(uri, true), ...pathFragment)).path;
      } else {
        newPath = posix.join(uri.path, ...pathFragment);
      }
      return uri.with({ path: newPath });
    }
    toString(skipEncoding = false) {
      return _asFormatted(this, skipEncoding);
    }
    toJSON() {
      return this;
    }
    static revive(data) {
      if (!data) {
        return data;
      } else if (data instanceof URI) {
        return data;
      } else {
        const result = new Uri(data);
        result._formatted = data.external;
        result._fsPath = data._sep === _pathSepMarker ? data.fsPath : null;
        return result;
      }
    }
  };
  var _pathSepMarker = isWindows ? 1 : void 0;
  var Uri = class extends URI {
    constructor() {
      super(...arguments);
      this._formatted = null;
      this._fsPath = null;
    }
    get fsPath() {
      if (!this._fsPath) {
        this._fsPath = uriToFsPath(this, false);
      }
      return this._fsPath;
    }
    toString(skipEncoding = false) {
      if (!skipEncoding) {
        if (!this._formatted) {
          this._formatted = _asFormatted(this, false);
        }
        return this._formatted;
      } else {
        return _asFormatted(this, true);
      }
    }
    toJSON() {
      const res = {
        $mid: 1
      };
      if (this._fsPath) {
        res.fsPath = this._fsPath;
        res._sep = _pathSepMarker;
      }
      if (this._formatted) {
        res.external = this._formatted;
      }
      if (this.path) {
        res.path = this.path;
      }
      if (this.scheme) {
        res.scheme = this.scheme;
      }
      if (this.authority) {
        res.authority = this.authority;
      }
      if (this.query) {
        res.query = this.query;
      }
      if (this.fragment) {
        res.fragment = this.fragment;
      }
      return res;
    }
  };
  var encodeTable = {
    [58]: "%3A",
    [47]: "%2F",
    [63]: "%3F",
    [35]: "%23",
    [91]: "%5B",
    [93]: "%5D",
    [64]: "%40",
    [33]: "%21",
    [36]: "%24",
    [38]: "%26",
    [39]: "%27",
    [40]: "%28",
    [41]: "%29",
    [42]: "%2A",
    [43]: "%2B",
    [44]: "%2C",
    [59]: "%3B",
    [61]: "%3D",
    [32]: "%20"
  };
  function encodeURIComponentFast(uriComponent, allowSlash) {
    let res = void 0;
    let nativeEncodePos = -1;
    for (let pos = 0; pos < uriComponent.length; pos++) {
      const code = uriComponent.charCodeAt(pos);
      if (code >= 97 && code <= 122 || code >= 65 && code <= 90 || code >= 48 && code <= 57 || code === 45 || code === 46 || code === 95 || code === 126 || allowSlash && code === 47) {
        if (nativeEncodePos !== -1) {
          res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
          nativeEncodePos = -1;
        }
        if (res !== void 0) {
          res += uriComponent.charAt(pos);
        }
      } else {
        if (res === void 0) {
          res = uriComponent.substr(0, pos);
        }
        const escaped = encodeTable[code];
        if (escaped !== void 0) {
          if (nativeEncodePos !== -1) {
            res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
            nativeEncodePos = -1;
          }
          res += escaped;
        } else if (nativeEncodePos === -1) {
          nativeEncodePos = pos;
        }
      }
    }
    if (nativeEncodePos !== -1) {
      res += encodeURIComponent(uriComponent.substring(nativeEncodePos));
    }
    return res !== void 0 ? res : uriComponent;
  }
  function encodeURIComponentMinimal(path) {
    let res = void 0;
    for (let pos = 0; pos < path.length; pos++) {
      const code = path.charCodeAt(pos);
      if (code === 35 || code === 63) {
        if (res === void 0) {
          res = path.substr(0, pos);
        }
        res += encodeTable[code];
      } else {
        if (res !== void 0) {
          res += path[pos];
        }
      }
    }
    return res !== void 0 ? res : path;
  }
  function uriToFsPath(uri, keepDriveLetterCasing) {
    let value;
    if (uri.authority && uri.path.length > 1 && uri.scheme === "file") {
      value = `//${uri.authority}${uri.path}`;
    } else if (uri.path.charCodeAt(0) === 47 && (uri.path.charCodeAt(1) >= 65 && uri.path.charCodeAt(1) <= 90 || uri.path.charCodeAt(1) >= 97 && uri.path.charCodeAt(1) <= 122) && uri.path.charCodeAt(2) === 58) {
      if (!keepDriveLetterCasing) {
        value = uri.path[1].toLowerCase() + uri.path.substr(2);
      } else {
        value = uri.path.substr(1);
      }
    } else {
      value = uri.path;
    }
    if (isWindows) {
      value = value.replace(/\//g, "\\");
    }
    return value;
  }
  function _asFormatted(uri, skipEncoding) {
    const encoder = !skipEncoding ? encodeURIComponentFast : encodeURIComponentMinimal;
    let res = "";
    let { scheme, authority, path, query, fragment } = uri;
    if (scheme) {
      res += scheme;
      res += ":";
    }
    if (authority || scheme === "file") {
      res += _slash;
      res += _slash;
    }
    if (authority) {
      let idx = authority.indexOf("@");
      if (idx !== -1) {
        const userinfo = authority.substr(0, idx);
        authority = authority.substr(idx + 1);
        idx = userinfo.indexOf(":");
        if (idx === -1) {
          res += encoder(userinfo, false);
        } else {
          res += encoder(userinfo.substr(0, idx), false);
          res += ":";
          res += encoder(userinfo.substr(idx + 1), false);
        }
        res += "@";
      }
      authority = authority.toLowerCase();
      idx = authority.indexOf(":");
      if (idx === -1) {
        res += encoder(authority, false);
      } else {
        res += encoder(authority.substr(0, idx), false);
        res += authority.substr(idx);
      }
    }
    if (path) {
      if (path.length >= 3 && path.charCodeAt(0) === 47 && path.charCodeAt(2) === 58) {
        const code = path.charCodeAt(1);
        if (code >= 65 && code <= 90) {
          path = `/${String.fromCharCode(code + 32)}:${path.substr(3)}`;
        }
      } else if (path.length >= 2 && path.charCodeAt(1) === 58) {
        const code = path.charCodeAt(0);
        if (code >= 65 && code <= 90) {
          path = `${String.fromCharCode(code + 32)}:${path.substr(2)}`;
        }
      }
      res += encoder(path, true);
    }
    if (query) {
      res += "?";
      res += encoder(query, false);
    }
    if (fragment) {
      res += "#";
      res += !skipEncoding ? encodeURIComponentFast(fragment, false) : fragment;
    }
    return res;
  }
  function decodeURIComponentGraceful(str) {
    try {
      return decodeURIComponent(str);
    } catch (_a3) {
      if (str.length > 3) {
        return str.substr(0, 3) + decodeURIComponentGraceful(str.substr(3));
      } else {
        return str;
      }
    }
  }
  var _rEncodedAsHex = /(%[0-9A-Za-z][0-9A-Za-z])+/g;
  function percentDecode(str) {
    if (!str.match(_rEncodedAsHex)) {
      return str;
    }
    return str.replace(_rEncodedAsHex, (match) => decodeURIComponentGraceful(match));
  }

  // node_modules/monaco-editor/esm/vs/editor/common/core/position.js
  var Position = class {
    constructor(lineNumber, column) {
      this.lineNumber = lineNumber;
      this.column = column;
    }
    with(newLineNumber = this.lineNumber, newColumn = this.column) {
      if (newLineNumber === this.lineNumber && newColumn === this.column) {
        return this;
      } else {
        return new Position(newLineNumber, newColumn);
      }
    }
    delta(deltaLineNumber = 0, deltaColumn = 0) {
      return this.with(this.lineNumber + deltaLineNumber, this.column + deltaColumn);
    }
    equals(other) {
      return Position.equals(this, other);
    }
    static equals(a2, b) {
      if (!a2 && !b) {
        return true;
      }
      return !!a2 && !!b && a2.lineNumber === b.lineNumber && a2.column === b.column;
    }
    isBefore(other) {
      return Position.isBefore(this, other);
    }
    static isBefore(a2, b) {
      if (a2.lineNumber < b.lineNumber) {
        return true;
      }
      if (b.lineNumber < a2.lineNumber) {
        return false;
      }
      return a2.column < b.column;
    }
    isBeforeOrEqual(other) {
      return Position.isBeforeOrEqual(this, other);
    }
    static isBeforeOrEqual(a2, b) {
      if (a2.lineNumber < b.lineNumber) {
        return true;
      }
      if (b.lineNumber < a2.lineNumber) {
        return false;
      }
      return a2.column <= b.column;
    }
    static compare(a2, b) {
      let aLineNumber = a2.lineNumber | 0;
      let bLineNumber = b.lineNumber | 0;
      if (aLineNumber === bLineNumber) {
        let aColumn = a2.column | 0;
        let bColumn = b.column | 0;
        return aColumn - bColumn;
      }
      return aLineNumber - bLineNumber;
    }
    clone() {
      return new Position(this.lineNumber, this.column);
    }
    toString() {
      return "(" + this.lineNumber + "," + this.column + ")";
    }
    static lift(pos) {
      return new Position(pos.lineNumber, pos.column);
    }
    static isIPosition(obj) {
      return obj && typeof obj.lineNumber === "number" && typeof obj.column === "number";
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/core/range.js
  var Range = class {
    constructor(startLineNumber, startColumn, endLineNumber, endColumn) {
      if (startLineNumber > endLineNumber || startLineNumber === endLineNumber && startColumn > endColumn) {
        this.startLineNumber = endLineNumber;
        this.startColumn = endColumn;
        this.endLineNumber = startLineNumber;
        this.endColumn = startColumn;
      } else {
        this.startLineNumber = startLineNumber;
        this.startColumn = startColumn;
        this.endLineNumber = endLineNumber;
        this.endColumn = endColumn;
      }
    }
    isEmpty() {
      return Range.isEmpty(this);
    }
    static isEmpty(range) {
      return range.startLineNumber === range.endLineNumber && range.startColumn === range.endColumn;
    }
    containsPosition(position) {
      return Range.containsPosition(this, position);
    }
    static containsPosition(range, position) {
      if (position.lineNumber < range.startLineNumber || position.lineNumber > range.endLineNumber) {
        return false;
      }
      if (position.lineNumber === range.startLineNumber && position.column < range.startColumn) {
        return false;
      }
      if (position.lineNumber === range.endLineNumber && position.column > range.endColumn) {
        return false;
      }
      return true;
    }
    containsRange(range) {
      return Range.containsRange(this, range);
    }
    static containsRange(range, otherRange) {
      if (otherRange.startLineNumber < range.startLineNumber || otherRange.endLineNumber < range.startLineNumber) {
        return false;
      }
      if (otherRange.startLineNumber > range.endLineNumber || otherRange.endLineNumber > range.endLineNumber) {
        return false;
      }
      if (otherRange.startLineNumber === range.startLineNumber && otherRange.startColumn < range.startColumn) {
        return false;
      }
      if (otherRange.endLineNumber === range.endLineNumber && otherRange.endColumn > range.endColumn) {
        return false;
      }
      return true;
    }
    strictContainsRange(range) {
      return Range.strictContainsRange(this, range);
    }
    static strictContainsRange(range, otherRange) {
      if (otherRange.startLineNumber < range.startLineNumber || otherRange.endLineNumber < range.startLineNumber) {
        return false;
      }
      if (otherRange.startLineNumber > range.endLineNumber || otherRange.endLineNumber > range.endLineNumber) {
        return false;
      }
      if (otherRange.startLineNumber === range.startLineNumber && otherRange.startColumn <= range.startColumn) {
        return false;
      }
      if (otherRange.endLineNumber === range.endLineNumber && otherRange.endColumn >= range.endColumn) {
        return false;
      }
      return true;
    }
    plusRange(range) {
      return Range.plusRange(this, range);
    }
    static plusRange(a2, b) {
      let startLineNumber;
      let startColumn;
      let endLineNumber;
      let endColumn;
      if (b.startLineNumber < a2.startLineNumber) {
        startLineNumber = b.startLineNumber;
        startColumn = b.startColumn;
      } else if (b.startLineNumber === a2.startLineNumber) {
        startLineNumber = b.startLineNumber;
        startColumn = Math.min(b.startColumn, a2.startColumn);
      } else {
        startLineNumber = a2.startLineNumber;
        startColumn = a2.startColumn;
      }
      if (b.endLineNumber > a2.endLineNumber) {
        endLineNumber = b.endLineNumber;
        endColumn = b.endColumn;
      } else if (b.endLineNumber === a2.endLineNumber) {
        endLineNumber = b.endLineNumber;
        endColumn = Math.max(b.endColumn, a2.endColumn);
      } else {
        endLineNumber = a2.endLineNumber;
        endColumn = a2.endColumn;
      }
      return new Range(startLineNumber, startColumn, endLineNumber, endColumn);
    }
    intersectRanges(range) {
      return Range.intersectRanges(this, range);
    }
    static intersectRanges(a2, b) {
      let resultStartLineNumber = a2.startLineNumber;
      let resultStartColumn = a2.startColumn;
      let resultEndLineNumber = a2.endLineNumber;
      let resultEndColumn = a2.endColumn;
      let otherStartLineNumber = b.startLineNumber;
      let otherStartColumn = b.startColumn;
      let otherEndLineNumber = b.endLineNumber;
      let otherEndColumn = b.endColumn;
      if (resultStartLineNumber < otherStartLineNumber) {
        resultStartLineNumber = otherStartLineNumber;
        resultStartColumn = otherStartColumn;
      } else if (resultStartLineNumber === otherStartLineNumber) {
        resultStartColumn = Math.max(resultStartColumn, otherStartColumn);
      }
      if (resultEndLineNumber > otherEndLineNumber) {
        resultEndLineNumber = otherEndLineNumber;
        resultEndColumn = otherEndColumn;
      } else if (resultEndLineNumber === otherEndLineNumber) {
        resultEndColumn = Math.min(resultEndColumn, otherEndColumn);
      }
      if (resultStartLineNumber > resultEndLineNumber) {
        return null;
      }
      if (resultStartLineNumber === resultEndLineNumber && resultStartColumn > resultEndColumn) {
        return null;
      }
      return new Range(resultStartLineNumber, resultStartColumn, resultEndLineNumber, resultEndColumn);
    }
    equalsRange(other) {
      return Range.equalsRange(this, other);
    }
    static equalsRange(a2, b) {
      return !!a2 && !!b && a2.startLineNumber === b.startLineNumber && a2.startColumn === b.startColumn && a2.endLineNumber === b.endLineNumber && a2.endColumn === b.endColumn;
    }
    getEndPosition() {
      return Range.getEndPosition(this);
    }
    static getEndPosition(range) {
      return new Position(range.endLineNumber, range.endColumn);
    }
    getStartPosition() {
      return Range.getStartPosition(this);
    }
    static getStartPosition(range) {
      return new Position(range.startLineNumber, range.startColumn);
    }
    toString() {
      return "[" + this.startLineNumber + "," + this.startColumn + " -> " + this.endLineNumber + "," + this.endColumn + "]";
    }
    setEndPosition(endLineNumber, endColumn) {
      return new Range(this.startLineNumber, this.startColumn, endLineNumber, endColumn);
    }
    setStartPosition(startLineNumber, startColumn) {
      return new Range(startLineNumber, startColumn, this.endLineNumber, this.endColumn);
    }
    collapseToStart() {
      return Range.collapseToStart(this);
    }
    static collapseToStart(range) {
      return new Range(range.startLineNumber, range.startColumn, range.startLineNumber, range.startColumn);
    }
    static fromPositions(start, end = start) {
      return new Range(start.lineNumber, start.column, end.lineNumber, end.column);
    }
    static lift(range) {
      if (!range) {
        return null;
      }
      return new Range(range.startLineNumber, range.startColumn, range.endLineNumber, range.endColumn);
    }
    static isIRange(obj) {
      return obj && typeof obj.startLineNumber === "number" && typeof obj.startColumn === "number" && typeof obj.endLineNumber === "number" && typeof obj.endColumn === "number";
    }
    static areIntersectingOrTouching(a2, b) {
      if (a2.endLineNumber < b.startLineNumber || a2.endLineNumber === b.startLineNumber && a2.endColumn < b.startColumn) {
        return false;
      }
      if (b.endLineNumber < a2.startLineNumber || b.endLineNumber === a2.startLineNumber && b.endColumn < a2.startColumn) {
        return false;
      }
      return true;
    }
    static areIntersecting(a2, b) {
      if (a2.endLineNumber < b.startLineNumber || a2.endLineNumber === b.startLineNumber && a2.endColumn <= b.startColumn) {
        return false;
      }
      if (b.endLineNumber < a2.startLineNumber || b.endLineNumber === a2.startLineNumber && b.endColumn <= a2.startColumn) {
        return false;
      }
      return true;
    }
    static compareRangesUsingStarts(a2, b) {
      if (a2 && b) {
        const aStartLineNumber = a2.startLineNumber | 0;
        const bStartLineNumber = b.startLineNumber | 0;
        if (aStartLineNumber === bStartLineNumber) {
          const aStartColumn = a2.startColumn | 0;
          const bStartColumn = b.startColumn | 0;
          if (aStartColumn === bStartColumn) {
            const aEndLineNumber = a2.endLineNumber | 0;
            const bEndLineNumber = b.endLineNumber | 0;
            if (aEndLineNumber === bEndLineNumber) {
              const aEndColumn = a2.endColumn | 0;
              const bEndColumn = b.endColumn | 0;
              return aEndColumn - bEndColumn;
            }
            return aEndLineNumber - bEndLineNumber;
          }
          return aStartColumn - bStartColumn;
        }
        return aStartLineNumber - bStartLineNumber;
      }
      const aExists = a2 ? 1 : 0;
      const bExists = b ? 1 : 0;
      return aExists - bExists;
    }
    static compareRangesUsingEnds(a2, b) {
      if (a2.endLineNumber === b.endLineNumber) {
        if (a2.endColumn === b.endColumn) {
          if (a2.startLineNumber === b.startLineNumber) {
            return a2.startColumn - b.startColumn;
          }
          return a2.startLineNumber - b.startLineNumber;
        }
        return a2.endColumn - b.endColumn;
      }
      return a2.endLineNumber - b.endLineNumber;
    }
    static spansMultipleLines(range) {
      return range.endLineNumber > range.startLineNumber;
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/diff/diffComputer.js
  var MINIMUM_MATCHING_CHARACTER_LENGTH = 3;
  function computeDiff(originalSequence, modifiedSequence, continueProcessingPredicate, pretty) {
    const diffAlgo = new LcsDiff(originalSequence, modifiedSequence, continueProcessingPredicate);
    return diffAlgo.ComputeDiff(pretty);
  }
  var LineSequence = class {
    constructor(lines) {
      const startColumns = [];
      const endColumns = [];
      for (let i = 0, length = lines.length; i < length; i++) {
        startColumns[i] = getFirstNonBlankColumn(lines[i], 1);
        endColumns[i] = getLastNonBlankColumn(lines[i], 1);
      }
      this.lines = lines;
      this._startColumns = startColumns;
      this._endColumns = endColumns;
    }
    getElements() {
      const elements = [];
      for (let i = 0, len = this.lines.length; i < len; i++) {
        elements[i] = this.lines[i].substring(this._startColumns[i] - 1, this._endColumns[i] - 1);
      }
      return elements;
    }
    getStartLineNumber(i) {
      return i + 1;
    }
    getEndLineNumber(i) {
      return i + 1;
    }
    createCharSequence(shouldIgnoreTrimWhitespace, startIndex, endIndex) {
      const charCodes = [];
      const lineNumbers = [];
      const columns = [];
      let len = 0;
      for (let index = startIndex; index <= endIndex; index++) {
        const lineContent = this.lines[index];
        const startColumn = shouldIgnoreTrimWhitespace ? this._startColumns[index] : 1;
        const endColumn = shouldIgnoreTrimWhitespace ? this._endColumns[index] : lineContent.length + 1;
        for (let col = startColumn; col < endColumn; col++) {
          charCodes[len] = lineContent.charCodeAt(col - 1);
          lineNumbers[len] = index + 1;
          columns[len] = col;
          len++;
        }
      }
      return new CharSequence(charCodes, lineNumbers, columns);
    }
  };
  var CharSequence = class {
    constructor(charCodes, lineNumbers, columns) {
      this._charCodes = charCodes;
      this._lineNumbers = lineNumbers;
      this._columns = columns;
    }
    getElements() {
      return this._charCodes;
    }
    getStartLineNumber(i) {
      return this._lineNumbers[i];
    }
    getStartColumn(i) {
      return this._columns[i];
    }
    getEndLineNumber(i) {
      return this._lineNumbers[i];
    }
    getEndColumn(i) {
      return this._columns[i] + 1;
    }
  };
  var CharChange = class {
    constructor(originalStartLineNumber, originalStartColumn, originalEndLineNumber, originalEndColumn, modifiedStartLineNumber, modifiedStartColumn, modifiedEndLineNumber, modifiedEndColumn) {
      this.originalStartLineNumber = originalStartLineNumber;
      this.originalStartColumn = originalStartColumn;
      this.originalEndLineNumber = originalEndLineNumber;
      this.originalEndColumn = originalEndColumn;
      this.modifiedStartLineNumber = modifiedStartLineNumber;
      this.modifiedStartColumn = modifiedStartColumn;
      this.modifiedEndLineNumber = modifiedEndLineNumber;
      this.modifiedEndColumn = modifiedEndColumn;
    }
    static createFromDiffChange(diffChange, originalCharSequence, modifiedCharSequence) {
      let originalStartLineNumber;
      let originalStartColumn;
      let originalEndLineNumber;
      let originalEndColumn;
      let modifiedStartLineNumber;
      let modifiedStartColumn;
      let modifiedEndLineNumber;
      let modifiedEndColumn;
      if (diffChange.originalLength === 0) {
        originalStartLineNumber = 0;
        originalStartColumn = 0;
        originalEndLineNumber = 0;
        originalEndColumn = 0;
      } else {
        originalStartLineNumber = originalCharSequence.getStartLineNumber(diffChange.originalStart);
        originalStartColumn = originalCharSequence.getStartColumn(diffChange.originalStart);
        originalEndLineNumber = originalCharSequence.getEndLineNumber(diffChange.originalStart + diffChange.originalLength - 1);
        originalEndColumn = originalCharSequence.getEndColumn(diffChange.originalStart + diffChange.originalLength - 1);
      }
      if (diffChange.modifiedLength === 0) {
        modifiedStartLineNumber = 0;
        modifiedStartColumn = 0;
        modifiedEndLineNumber = 0;
        modifiedEndColumn = 0;
      } else {
        modifiedStartLineNumber = modifiedCharSequence.getStartLineNumber(diffChange.modifiedStart);
        modifiedStartColumn = modifiedCharSequence.getStartColumn(diffChange.modifiedStart);
        modifiedEndLineNumber = modifiedCharSequence.getEndLineNumber(diffChange.modifiedStart + diffChange.modifiedLength - 1);
        modifiedEndColumn = modifiedCharSequence.getEndColumn(diffChange.modifiedStart + diffChange.modifiedLength - 1);
      }
      return new CharChange(originalStartLineNumber, originalStartColumn, originalEndLineNumber, originalEndColumn, modifiedStartLineNumber, modifiedStartColumn, modifiedEndLineNumber, modifiedEndColumn);
    }
  };
  function postProcessCharChanges(rawChanges) {
    if (rawChanges.length <= 1) {
      return rawChanges;
    }
    const result = [rawChanges[0]];
    let prevChange = result[0];
    for (let i = 1, len = rawChanges.length; i < len; i++) {
      const currChange = rawChanges[i];
      const originalMatchingLength = currChange.originalStart - (prevChange.originalStart + prevChange.originalLength);
      const modifiedMatchingLength = currChange.modifiedStart - (prevChange.modifiedStart + prevChange.modifiedLength);
      const matchingLength = Math.min(originalMatchingLength, modifiedMatchingLength);
      if (matchingLength < MINIMUM_MATCHING_CHARACTER_LENGTH) {
        prevChange.originalLength = currChange.originalStart + currChange.originalLength - prevChange.originalStart;
        prevChange.modifiedLength = currChange.modifiedStart + currChange.modifiedLength - prevChange.modifiedStart;
      } else {
        result.push(currChange);
        prevChange = currChange;
      }
    }
    return result;
  }
  var LineChange = class {
    constructor(originalStartLineNumber, originalEndLineNumber, modifiedStartLineNumber, modifiedEndLineNumber, charChanges) {
      this.originalStartLineNumber = originalStartLineNumber;
      this.originalEndLineNumber = originalEndLineNumber;
      this.modifiedStartLineNumber = modifiedStartLineNumber;
      this.modifiedEndLineNumber = modifiedEndLineNumber;
      this.charChanges = charChanges;
    }
    static createFromDiffResult(shouldIgnoreTrimWhitespace, diffChange, originalLineSequence, modifiedLineSequence, continueCharDiff, shouldComputeCharChanges, shouldPostProcessCharChanges) {
      let originalStartLineNumber;
      let originalEndLineNumber;
      let modifiedStartLineNumber;
      let modifiedEndLineNumber;
      let charChanges = void 0;
      if (diffChange.originalLength === 0) {
        originalStartLineNumber = originalLineSequence.getStartLineNumber(diffChange.originalStart) - 1;
        originalEndLineNumber = 0;
      } else {
        originalStartLineNumber = originalLineSequence.getStartLineNumber(diffChange.originalStart);
        originalEndLineNumber = originalLineSequence.getEndLineNumber(diffChange.originalStart + diffChange.originalLength - 1);
      }
      if (diffChange.modifiedLength === 0) {
        modifiedStartLineNumber = modifiedLineSequence.getStartLineNumber(diffChange.modifiedStart) - 1;
        modifiedEndLineNumber = 0;
      } else {
        modifiedStartLineNumber = modifiedLineSequence.getStartLineNumber(diffChange.modifiedStart);
        modifiedEndLineNumber = modifiedLineSequence.getEndLineNumber(diffChange.modifiedStart + diffChange.modifiedLength - 1);
      }
      if (shouldComputeCharChanges && diffChange.originalLength > 0 && diffChange.originalLength < 20 && diffChange.modifiedLength > 0 && diffChange.modifiedLength < 20 && continueCharDiff()) {
        const originalCharSequence = originalLineSequence.createCharSequence(shouldIgnoreTrimWhitespace, diffChange.originalStart, diffChange.originalStart + diffChange.originalLength - 1);
        const modifiedCharSequence = modifiedLineSequence.createCharSequence(shouldIgnoreTrimWhitespace, diffChange.modifiedStart, diffChange.modifiedStart + diffChange.modifiedLength - 1);
        let rawChanges = computeDiff(originalCharSequence, modifiedCharSequence, continueCharDiff, true).changes;
        if (shouldPostProcessCharChanges) {
          rawChanges = postProcessCharChanges(rawChanges);
        }
        charChanges = [];
        for (let i = 0, length = rawChanges.length; i < length; i++) {
          charChanges.push(CharChange.createFromDiffChange(rawChanges[i], originalCharSequence, modifiedCharSequence));
        }
      }
      return new LineChange(originalStartLineNumber, originalEndLineNumber, modifiedStartLineNumber, modifiedEndLineNumber, charChanges);
    }
  };
  var DiffComputer = class {
    constructor(originalLines, modifiedLines, opts) {
      this.shouldComputeCharChanges = opts.shouldComputeCharChanges;
      this.shouldPostProcessCharChanges = opts.shouldPostProcessCharChanges;
      this.shouldIgnoreTrimWhitespace = opts.shouldIgnoreTrimWhitespace;
      this.shouldMakePrettyDiff = opts.shouldMakePrettyDiff;
      this.originalLines = originalLines;
      this.modifiedLines = modifiedLines;
      this.original = new LineSequence(originalLines);
      this.modified = new LineSequence(modifiedLines);
      this.continueLineDiff = createContinueProcessingPredicate(opts.maxComputationTime);
      this.continueCharDiff = createContinueProcessingPredicate(opts.maxComputationTime === 0 ? 0 : Math.min(opts.maxComputationTime, 5e3));
    }
    computeDiff() {
      if (this.original.lines.length === 1 && this.original.lines[0].length === 0) {
        return {
          quitEarly: false,
          changes: [{
            originalStartLineNumber: 1,
            originalEndLineNumber: 1,
            modifiedStartLineNumber: 1,
            modifiedEndLineNumber: this.modified.lines.length,
            charChanges: [{
              modifiedEndColumn: 0,
              modifiedEndLineNumber: 0,
              modifiedStartColumn: 0,
              modifiedStartLineNumber: 0,
              originalEndColumn: 0,
              originalEndLineNumber: 0,
              originalStartColumn: 0,
              originalStartLineNumber: 0
            }]
          }]
        };
      }
      if (this.modified.lines.length === 1 && this.modified.lines[0].length === 0) {
        return {
          quitEarly: false,
          changes: [{
            originalStartLineNumber: 1,
            originalEndLineNumber: this.original.lines.length,
            modifiedStartLineNumber: 1,
            modifiedEndLineNumber: 1,
            charChanges: [{
              modifiedEndColumn: 0,
              modifiedEndLineNumber: 0,
              modifiedStartColumn: 0,
              modifiedStartLineNumber: 0,
              originalEndColumn: 0,
              originalEndLineNumber: 0,
              originalStartColumn: 0,
              originalStartLineNumber: 0
            }]
          }]
        };
      }
      const diffResult = computeDiff(this.original, this.modified, this.continueLineDiff, this.shouldMakePrettyDiff);
      const rawChanges = diffResult.changes;
      const quitEarly = diffResult.quitEarly;
      if (this.shouldIgnoreTrimWhitespace) {
        const lineChanges = [];
        for (let i = 0, length = rawChanges.length; i < length; i++) {
          lineChanges.push(LineChange.createFromDiffResult(this.shouldIgnoreTrimWhitespace, rawChanges[i], this.original, this.modified, this.continueCharDiff, this.shouldComputeCharChanges, this.shouldPostProcessCharChanges));
        }
        return {
          quitEarly,
          changes: lineChanges
        };
      }
      const result = [];
      let originalLineIndex = 0;
      let modifiedLineIndex = 0;
      for (let i = -1, len = rawChanges.length; i < len; i++) {
        const nextChange = i + 1 < len ? rawChanges[i + 1] : null;
        const originalStop = nextChange ? nextChange.originalStart : this.originalLines.length;
        const modifiedStop = nextChange ? nextChange.modifiedStart : this.modifiedLines.length;
        while (originalLineIndex < originalStop && modifiedLineIndex < modifiedStop) {
          const originalLine = this.originalLines[originalLineIndex];
          const modifiedLine = this.modifiedLines[modifiedLineIndex];
          if (originalLine !== modifiedLine) {
            {
              let originalStartColumn = getFirstNonBlankColumn(originalLine, 1);
              let modifiedStartColumn = getFirstNonBlankColumn(modifiedLine, 1);
              while (originalStartColumn > 1 && modifiedStartColumn > 1) {
                const originalChar = originalLine.charCodeAt(originalStartColumn - 2);
                const modifiedChar = modifiedLine.charCodeAt(modifiedStartColumn - 2);
                if (originalChar !== modifiedChar) {
                  break;
                }
                originalStartColumn--;
                modifiedStartColumn--;
              }
              if (originalStartColumn > 1 || modifiedStartColumn > 1) {
                this._pushTrimWhitespaceCharChange(result, originalLineIndex + 1, 1, originalStartColumn, modifiedLineIndex + 1, 1, modifiedStartColumn);
              }
            }
            {
              let originalEndColumn = getLastNonBlankColumn(originalLine, 1);
              let modifiedEndColumn = getLastNonBlankColumn(modifiedLine, 1);
              const originalMaxColumn = originalLine.length + 1;
              const modifiedMaxColumn = modifiedLine.length + 1;
              while (originalEndColumn < originalMaxColumn && modifiedEndColumn < modifiedMaxColumn) {
                const originalChar = originalLine.charCodeAt(originalEndColumn - 1);
                const modifiedChar = originalLine.charCodeAt(modifiedEndColumn - 1);
                if (originalChar !== modifiedChar) {
                  break;
                }
                originalEndColumn++;
                modifiedEndColumn++;
              }
              if (originalEndColumn < originalMaxColumn || modifiedEndColumn < modifiedMaxColumn) {
                this._pushTrimWhitespaceCharChange(result, originalLineIndex + 1, originalEndColumn, originalMaxColumn, modifiedLineIndex + 1, modifiedEndColumn, modifiedMaxColumn);
              }
            }
          }
          originalLineIndex++;
          modifiedLineIndex++;
        }
        if (nextChange) {
          result.push(LineChange.createFromDiffResult(this.shouldIgnoreTrimWhitespace, nextChange, this.original, this.modified, this.continueCharDiff, this.shouldComputeCharChanges, this.shouldPostProcessCharChanges));
          originalLineIndex += nextChange.originalLength;
          modifiedLineIndex += nextChange.modifiedLength;
        }
      }
      return {
        quitEarly,
        changes: result
      };
    }
    _pushTrimWhitespaceCharChange(result, originalLineNumber, originalStartColumn, originalEndColumn, modifiedLineNumber, modifiedStartColumn, modifiedEndColumn) {
      if (this._mergeTrimWhitespaceCharChange(result, originalLineNumber, originalStartColumn, originalEndColumn, modifiedLineNumber, modifiedStartColumn, modifiedEndColumn)) {
        return;
      }
      let charChanges = void 0;
      if (this.shouldComputeCharChanges) {
        charChanges = [new CharChange(originalLineNumber, originalStartColumn, originalLineNumber, originalEndColumn, modifiedLineNumber, modifiedStartColumn, modifiedLineNumber, modifiedEndColumn)];
      }
      result.push(new LineChange(originalLineNumber, originalLineNumber, modifiedLineNumber, modifiedLineNumber, charChanges));
    }
    _mergeTrimWhitespaceCharChange(result, originalLineNumber, originalStartColumn, originalEndColumn, modifiedLineNumber, modifiedStartColumn, modifiedEndColumn) {
      const len = result.length;
      if (len === 0) {
        return false;
      }
      const prevChange = result[len - 1];
      if (prevChange.originalEndLineNumber === 0 || prevChange.modifiedEndLineNumber === 0) {
        return false;
      }
      if (prevChange.originalEndLineNumber + 1 === originalLineNumber && prevChange.modifiedEndLineNumber + 1 === modifiedLineNumber) {
        prevChange.originalEndLineNumber = originalLineNumber;
        prevChange.modifiedEndLineNumber = modifiedLineNumber;
        if (this.shouldComputeCharChanges && prevChange.charChanges) {
          prevChange.charChanges.push(new CharChange(originalLineNumber, originalStartColumn, originalLineNumber, originalEndColumn, modifiedLineNumber, modifiedStartColumn, modifiedLineNumber, modifiedEndColumn));
        }
        return true;
      }
      return false;
    }
  };
  function getFirstNonBlankColumn(txt, defaultValue) {
    const r = firstNonWhitespaceIndex(txt);
    if (r === -1) {
      return defaultValue;
    }
    return r + 1;
  }
  function getLastNonBlankColumn(txt, defaultValue) {
    const r = lastNonWhitespaceIndex(txt);
    if (r === -1) {
      return defaultValue;
    }
    return r + 2;
  }
  function createContinueProcessingPredicate(maximumRuntime) {
    if (maximumRuntime === 0) {
      return () => true;
    }
    const startTime = Date.now();
    return () => {
      return Date.now() - startTime < maximumRuntime;
    };
  }

  // node_modules/monaco-editor/esm/vs/base/common/uint.js
  function toUint8(v) {
    if (v < 0) {
      return 0;
    }
    if (v > 255) {
      return 255;
    }
    return v | 0;
  }
  function toUint32(v) {
    if (v < 0) {
      return 0;
    }
    if (v > 4294967295) {
      return 4294967295;
    }
    return v | 0;
  }

  // node_modules/monaco-editor/esm/vs/editor/common/viewModel/prefixSumComputer.js
  var PrefixSumIndexOfResult = class {
    constructor(index, remainder) {
      this.index = index;
      this.remainder = remainder;
    }
  };
  var PrefixSumComputer = class {
    constructor(values2) {
      this.values = values2;
      this.prefixSum = new Uint32Array(values2.length);
      this.prefixSumValidIndex = new Int32Array(1);
      this.prefixSumValidIndex[0] = -1;
    }
    insertValues(insertIndex, insertValues) {
      insertIndex = toUint32(insertIndex);
      const oldValues = this.values;
      const oldPrefixSum = this.prefixSum;
      const insertValuesLen = insertValues.length;
      if (insertValuesLen === 0) {
        return false;
      }
      this.values = new Uint32Array(oldValues.length + insertValuesLen);
      this.values.set(oldValues.subarray(0, insertIndex), 0);
      this.values.set(oldValues.subarray(insertIndex), insertIndex + insertValuesLen);
      this.values.set(insertValues, insertIndex);
      if (insertIndex - 1 < this.prefixSumValidIndex[0]) {
        this.prefixSumValidIndex[0] = insertIndex - 1;
      }
      this.prefixSum = new Uint32Array(this.values.length);
      if (this.prefixSumValidIndex[0] >= 0) {
        this.prefixSum.set(oldPrefixSum.subarray(0, this.prefixSumValidIndex[0] + 1));
      }
      return true;
    }
    changeValue(index, value) {
      index = toUint32(index);
      value = toUint32(value);
      if (this.values[index] === value) {
        return false;
      }
      this.values[index] = value;
      if (index - 1 < this.prefixSumValidIndex[0]) {
        this.prefixSumValidIndex[0] = index - 1;
      }
      return true;
    }
    removeValues(startIndex, cnt) {
      startIndex = toUint32(startIndex);
      cnt = toUint32(cnt);
      const oldValues = this.values;
      const oldPrefixSum = this.prefixSum;
      if (startIndex >= oldValues.length) {
        return false;
      }
      let maxCnt = oldValues.length - startIndex;
      if (cnt >= maxCnt) {
        cnt = maxCnt;
      }
      if (cnt === 0) {
        return false;
      }
      this.values = new Uint32Array(oldValues.length - cnt);
      this.values.set(oldValues.subarray(0, startIndex), 0);
      this.values.set(oldValues.subarray(startIndex + cnt), startIndex);
      this.prefixSum = new Uint32Array(this.values.length);
      if (startIndex - 1 < this.prefixSumValidIndex[0]) {
        this.prefixSumValidIndex[0] = startIndex - 1;
      }
      if (this.prefixSumValidIndex[0] >= 0) {
        this.prefixSum.set(oldPrefixSum.subarray(0, this.prefixSumValidIndex[0] + 1));
      }
      return true;
    }
    getTotalValue() {
      if (this.values.length === 0) {
        return 0;
      }
      return this._getAccumulatedValue(this.values.length - 1);
    }
    getAccumulatedValue(index) {
      if (index < 0) {
        return 0;
      }
      index = toUint32(index);
      return this._getAccumulatedValue(index);
    }
    _getAccumulatedValue(index) {
      if (index <= this.prefixSumValidIndex[0]) {
        return this.prefixSum[index];
      }
      let startIndex = this.prefixSumValidIndex[0] + 1;
      if (startIndex === 0) {
        this.prefixSum[0] = this.values[0];
        startIndex++;
      }
      if (index >= this.values.length) {
        index = this.values.length - 1;
      }
      for (let i = startIndex; i <= index; i++) {
        this.prefixSum[i] = this.prefixSum[i - 1] + this.values[i];
      }
      this.prefixSumValidIndex[0] = Math.max(this.prefixSumValidIndex[0], index);
      return this.prefixSum[index];
    }
    getIndexOf(accumulatedValue) {
      accumulatedValue = Math.floor(accumulatedValue);
      this.getTotalValue();
      let low = 0;
      let high = this.values.length - 1;
      let mid = 0;
      let midStop = 0;
      let midStart = 0;
      while (low <= high) {
        mid = low + (high - low) / 2 | 0;
        midStop = this.prefixSum[mid];
        midStart = midStop - this.values[mid];
        if (accumulatedValue < midStart) {
          high = mid - 1;
        } else if (accumulatedValue >= midStop) {
          low = mid + 1;
        } else {
          break;
        }
      }
      return new PrefixSumIndexOfResult(mid, accumulatedValue - midStart);
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/model/mirrorTextModel.js
  var MirrorTextModel = class {
    constructor(uri, lines, eol, versionId) {
      this._uri = uri;
      this._lines = lines;
      this._eol = eol;
      this._versionId = versionId;
      this._lineStarts = null;
      this._cachedTextValue = null;
    }
    dispose() {
      this._lines.length = 0;
    }
    getText() {
      if (this._cachedTextValue === null) {
        this._cachedTextValue = this._lines.join(this._eol);
      }
      return this._cachedTextValue;
    }
    onEvents(e) {
      if (e.eol && e.eol !== this._eol) {
        this._eol = e.eol;
        this._lineStarts = null;
      }
      const changes = e.changes;
      for (const change of changes) {
        this._acceptDeleteRange(change.range);
        this._acceptInsertText(new Position(change.range.startLineNumber, change.range.startColumn), change.text);
      }
      this._versionId = e.versionId;
      this._cachedTextValue = null;
    }
    _ensureLineStarts() {
      if (!this._lineStarts) {
        const eolLength = this._eol.length;
        const linesLength = this._lines.length;
        const lineStartValues = new Uint32Array(linesLength);
        for (let i = 0; i < linesLength; i++) {
          lineStartValues[i] = this._lines[i].length + eolLength;
        }
        this._lineStarts = new PrefixSumComputer(lineStartValues);
      }
    }
    _setLineText(lineIndex, newValue) {
      this._lines[lineIndex] = newValue;
      if (this._lineStarts) {
        this._lineStarts.changeValue(lineIndex, this._lines[lineIndex].length + this._eol.length);
      }
    }
    _acceptDeleteRange(range) {
      if (range.startLineNumber === range.endLineNumber) {
        if (range.startColumn === range.endColumn) {
          return;
        }
        this._setLineText(range.startLineNumber - 1, this._lines[range.startLineNumber - 1].substring(0, range.startColumn - 1) + this._lines[range.startLineNumber - 1].substring(range.endColumn - 1));
        return;
      }
      this._setLineText(range.startLineNumber - 1, this._lines[range.startLineNumber - 1].substring(0, range.startColumn - 1) + this._lines[range.endLineNumber - 1].substring(range.endColumn - 1));
      this._lines.splice(range.startLineNumber, range.endLineNumber - range.startLineNumber);
      if (this._lineStarts) {
        this._lineStarts.removeValues(range.startLineNumber, range.endLineNumber - range.startLineNumber);
      }
    }
    _acceptInsertText(position, insertText) {
      if (insertText.length === 0) {
        return;
      }
      let insertLines = insertText.split(/\r\n|\r|\n/);
      if (insertLines.length === 1) {
        this._setLineText(position.lineNumber - 1, this._lines[position.lineNumber - 1].substring(0, position.column - 1) + insertLines[0] + this._lines[position.lineNumber - 1].substring(position.column - 1));
        return;
      }
      insertLines[insertLines.length - 1] += this._lines[position.lineNumber - 1].substring(position.column - 1);
      this._setLineText(position.lineNumber - 1, this._lines[position.lineNumber - 1].substring(0, position.column - 1) + insertLines[0]);
      let newLengths = new Uint32Array(insertLines.length - 1);
      for (let i = 1; i < insertLines.length; i++) {
        this._lines.splice(position.lineNumber + i - 1, 0, insertLines[i]);
        newLengths[i - 1] = insertLines[i].length + this._eol.length;
      }
      if (this._lineStarts) {
        this._lineStarts.insertValues(position.lineNumber, newLengths);
      }
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/model/wordHelper.js
  var USUAL_WORD_SEPARATORS = "`~!@#$%^&*()-=+[{]}\\|;:'\",.<>/?";
  function createWordRegExp(allowInWords = "") {
    let source = "(-?\\d*\\.\\d\\w*)|([^";
    for (const sep2 of USUAL_WORD_SEPARATORS) {
      if (allowInWords.indexOf(sep2) >= 0) {
        continue;
      }
      source += "\\" + sep2;
    }
    source += "\\s]+)";
    return new RegExp(source, "g");
  }
  var DEFAULT_WORD_REGEXP = createWordRegExp();
  function ensureValidWordDefinition(wordDefinition) {
    let result = DEFAULT_WORD_REGEXP;
    if (wordDefinition && wordDefinition instanceof RegExp) {
      if (!wordDefinition.global) {
        let flags = "g";
        if (wordDefinition.ignoreCase) {
          flags += "i";
        }
        if (wordDefinition.multiline) {
          flags += "m";
        }
        if (wordDefinition.unicode) {
          flags += "u";
        }
        result = new RegExp(wordDefinition.source, flags);
      } else {
        result = wordDefinition;
      }
    }
    result.lastIndex = 0;
    return result;
  }
  var _defaultConfig = {
    maxLen: 1e3,
    windowSize: 15,
    timeBudget: 150
  };
  function getWordAtText(column, wordDefinition, text, textOffset, config = _defaultConfig) {
    if (text.length > config.maxLen) {
      let start = column - config.maxLen / 2;
      if (start < 0) {
        textOffset += column;
        start = 0;
      } else {
        textOffset += start;
      }
      text = text.substring(start, column + config.maxLen / 2);
      return getWordAtText(column, wordDefinition, text, textOffset, config);
    }
    const t1 = Date.now();
    const pos = column - 1 - textOffset;
    let prevRegexIndex = -1;
    let match = null;
    for (let i = 1; ; i++) {
      if (Date.now() - t1 >= config.timeBudget) {
      }
      const regexIndex = pos - config.windowSize * i;
      wordDefinition.lastIndex = Math.max(0, regexIndex);
      const thisMatch = _findRegexMatchEnclosingPosition(wordDefinition, text, pos, prevRegexIndex);
      if (!thisMatch && match) {
        break;
      }
      match = thisMatch;
      if (regexIndex <= 0) {
        break;
      }
      prevRegexIndex = regexIndex;
    }
    if (match) {
      let result = {
        word: match[0],
        startColumn: textOffset + 1 + match.index,
        endColumn: textOffset + 1 + match.index + match[0].length
      };
      wordDefinition.lastIndex = 0;
      return result;
    }
    return null;
  }
  function _findRegexMatchEnclosingPosition(wordDefinition, text, pos, stopPos) {
    let match;
    while (match = wordDefinition.exec(text)) {
      const matchIndex = match.index || 0;
      if (matchIndex <= pos && wordDefinition.lastIndex >= pos) {
        return match;
      } else if (stopPos > 0 && matchIndex > stopPos) {
        return null;
      }
    }
    return null;
  }

  // node_modules/monaco-editor/esm/vs/editor/common/core/characterClassifier.js
  var CharacterClassifier = class {
    constructor(_defaultValue) {
      let defaultValue = toUint8(_defaultValue);
      this._defaultValue = defaultValue;
      this._asciiMap = CharacterClassifier._createAsciiMap(defaultValue);
      this._map = /* @__PURE__ */ new Map();
    }
    static _createAsciiMap(defaultValue) {
      let asciiMap = new Uint8Array(256);
      for (let i = 0; i < 256; i++) {
        asciiMap[i] = defaultValue;
      }
      return asciiMap;
    }
    set(charCode, _value) {
      let value = toUint8(_value);
      if (charCode >= 0 && charCode < 256) {
        this._asciiMap[charCode] = value;
      } else {
        this._map.set(charCode, value);
      }
    }
    get(charCode) {
      if (charCode >= 0 && charCode < 256) {
        return this._asciiMap[charCode];
      } else {
        return this._map.get(charCode) || this._defaultValue;
      }
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/modes/linkComputer.js
  var Uint8Matrix = class {
    constructor(rows, cols, defaultValue) {
      const data = new Uint8Array(rows * cols);
      for (let i = 0, len = rows * cols; i < len; i++) {
        data[i] = defaultValue;
      }
      this._data = data;
      this.rows = rows;
      this.cols = cols;
    }
    get(row, col) {
      return this._data[row * this.cols + col];
    }
    set(row, col, value) {
      this._data[row * this.cols + col] = value;
    }
  };
  var StateMachine = class {
    constructor(edges) {
      let maxCharCode = 0;
      let maxState = 0;
      for (let i = 0, len = edges.length; i < len; i++) {
        let [from, chCode, to] = edges[i];
        if (chCode > maxCharCode) {
          maxCharCode = chCode;
        }
        if (from > maxState) {
          maxState = from;
        }
        if (to > maxState) {
          maxState = to;
        }
      }
      maxCharCode++;
      maxState++;
      let states = new Uint8Matrix(maxState, maxCharCode, 0);
      for (let i = 0, len = edges.length; i < len; i++) {
        let [from, chCode, to] = edges[i];
        states.set(from, chCode, to);
      }
      this._states = states;
      this._maxCharCode = maxCharCode;
    }
    nextState(currentState, chCode) {
      if (chCode < 0 || chCode >= this._maxCharCode) {
        return 0;
      }
      return this._states.get(currentState, chCode);
    }
  };
  var _stateMachine = null;
  function getStateMachine() {
    if (_stateMachine === null) {
      _stateMachine = new StateMachine([
        [1, 104, 2],
        [1, 72, 2],
        [1, 102, 6],
        [1, 70, 6],
        [2, 116, 3],
        [2, 84, 3],
        [3, 116, 4],
        [3, 84, 4],
        [4, 112, 5],
        [4, 80, 5],
        [5, 115, 9],
        [5, 83, 9],
        [5, 58, 10],
        [6, 105, 7],
        [6, 73, 7],
        [7, 108, 8],
        [7, 76, 8],
        [8, 101, 9],
        [8, 69, 9],
        [9, 58, 10],
        [10, 47, 11],
        [11, 47, 12]
      ]);
    }
    return _stateMachine;
  }
  var _classifier = null;
  function getClassifier() {
    if (_classifier === null) {
      _classifier = new CharacterClassifier(0);
      const FORCE_TERMINATION_CHARACTERS = ` 	<>'"\u3001\u3002\uFF61\uFF64\uFF0C\uFF0E\uFF1A\uFF1B\u2018\u201C\u3008\u300A\u300C\u300E\u3010\u3014\uFF08\uFF3B\uFF5B\uFF62\uFF63\uFF5D\uFF3D\uFF09\u3015\u3011\u300F\u300D\u300B\u3009\u201D\u2019\uFF40\uFF5E\u2026`;
      for (let i = 0; i < FORCE_TERMINATION_CHARACTERS.length; i++) {
        _classifier.set(FORCE_TERMINATION_CHARACTERS.charCodeAt(i), 1);
      }
      const CANNOT_END_WITH_CHARACTERS = ".,;";
      for (let i = 0; i < CANNOT_END_WITH_CHARACTERS.length; i++) {
        _classifier.set(CANNOT_END_WITH_CHARACTERS.charCodeAt(i), 2);
      }
    }
    return _classifier;
  }
  var LinkComputer = class {
    static _createLink(classifier, line, lineNumber, linkBeginIndex, linkEndIndex) {
      let lastIncludedCharIndex = linkEndIndex - 1;
      do {
        const chCode = line.charCodeAt(lastIncludedCharIndex);
        const chClass = classifier.get(chCode);
        if (chClass !== 2) {
          break;
        }
        lastIncludedCharIndex--;
      } while (lastIncludedCharIndex > linkBeginIndex);
      if (linkBeginIndex > 0) {
        const charCodeBeforeLink = line.charCodeAt(linkBeginIndex - 1);
        const lastCharCodeInLink = line.charCodeAt(lastIncludedCharIndex);
        if (charCodeBeforeLink === 40 && lastCharCodeInLink === 41 || charCodeBeforeLink === 91 && lastCharCodeInLink === 93 || charCodeBeforeLink === 123 && lastCharCodeInLink === 125) {
          lastIncludedCharIndex--;
        }
      }
      return {
        range: {
          startLineNumber: lineNumber,
          startColumn: linkBeginIndex + 1,
          endLineNumber: lineNumber,
          endColumn: lastIncludedCharIndex + 2
        },
        url: line.substring(linkBeginIndex, lastIncludedCharIndex + 1)
      };
    }
    static computeLinks(model, stateMachine = getStateMachine()) {
      const classifier = getClassifier();
      let result = [];
      for (let i = 1, lineCount = model.getLineCount(); i <= lineCount; i++) {
        const line = model.getLineContent(i);
        const len = line.length;
        let j = 0;
        let linkBeginIndex = 0;
        let linkBeginChCode = 0;
        let state = 1;
        let hasOpenParens = false;
        let hasOpenSquareBracket = false;
        let inSquareBrackets = false;
        let hasOpenCurlyBracket = false;
        while (j < len) {
          let resetStateMachine = false;
          const chCode = line.charCodeAt(j);
          if (state === 13) {
            let chClass;
            switch (chCode) {
              case 40:
                hasOpenParens = true;
                chClass = 0;
                break;
              case 41:
                chClass = hasOpenParens ? 0 : 1;
                break;
              case 91:
                inSquareBrackets = true;
                hasOpenSquareBracket = true;
                chClass = 0;
                break;
              case 93:
                inSquareBrackets = false;
                chClass = hasOpenSquareBracket ? 0 : 1;
                break;
              case 123:
                hasOpenCurlyBracket = true;
                chClass = 0;
                break;
              case 125:
                chClass = hasOpenCurlyBracket ? 0 : 1;
                break;
              case 39:
                chClass = linkBeginChCode === 34 || linkBeginChCode === 96 ? 0 : 1;
                break;
              case 34:
                chClass = linkBeginChCode === 39 || linkBeginChCode === 96 ? 0 : 1;
                break;
              case 96:
                chClass = linkBeginChCode === 39 || linkBeginChCode === 34 ? 0 : 1;
                break;
              case 42:
                chClass = linkBeginChCode === 42 ? 1 : 0;
                break;
              case 124:
                chClass = linkBeginChCode === 124 ? 1 : 0;
                break;
              case 32:
                chClass = inSquareBrackets ? 0 : 1;
                break;
              default:
                chClass = classifier.get(chCode);
            }
            if (chClass === 1) {
              result.push(LinkComputer._createLink(classifier, line, i, linkBeginIndex, j));
              resetStateMachine = true;
            }
          } else if (state === 12) {
            let chClass;
            if (chCode === 91) {
              hasOpenSquareBracket = true;
              chClass = 0;
            } else {
              chClass = classifier.get(chCode);
            }
            if (chClass === 1) {
              resetStateMachine = true;
            } else {
              state = 13;
            }
          } else {
            state = stateMachine.nextState(state, chCode);
            if (state === 0) {
              resetStateMachine = true;
            }
          }
          if (resetStateMachine) {
            state = 1;
            hasOpenParens = false;
            hasOpenSquareBracket = false;
            hasOpenCurlyBracket = false;
            linkBeginIndex = j + 1;
            linkBeginChCode = chCode;
          }
          j++;
        }
        if (state === 13) {
          result.push(LinkComputer._createLink(classifier, line, i, linkBeginIndex, len));
        }
      }
      return result;
    }
  };
  function computeLinks(model) {
    if (!model || typeof model.getLineCount !== "function" || typeof model.getLineContent !== "function") {
      return [];
    }
    return LinkComputer.computeLinks(model);
  }

  // node_modules/monaco-editor/esm/vs/editor/common/modes/supports/inplaceReplaceSupport.js
  var BasicInplaceReplace = class {
    constructor() {
      this._defaultValueSet = [
        ["true", "false"],
        ["True", "False"],
        ["Private", "Public", "Friend", "ReadOnly", "Partial", "Protected", "WriteOnly"],
        ["public", "protected", "private"]
      ];
    }
    navigateValueSet(range1, text1, range2, text2, up) {
      if (range1 && text1) {
        let result = this.doNavigateValueSet(text1, up);
        if (result) {
          return {
            range: range1,
            value: result
          };
        }
      }
      if (range2 && text2) {
        let result = this.doNavigateValueSet(text2, up);
        if (result) {
          return {
            range: range2,
            value: result
          };
        }
      }
      return null;
    }
    doNavigateValueSet(text, up) {
      let numberResult = this.numberReplace(text, up);
      if (numberResult !== null) {
        return numberResult;
      }
      return this.textReplace(text, up);
    }
    numberReplace(value, up) {
      let precision = Math.pow(10, value.length - (value.lastIndexOf(".") + 1));
      let n1 = Number(value);
      let n2 = parseFloat(value);
      if (!isNaN(n1) && !isNaN(n2) && n1 === n2) {
        if (n1 === 0 && !up) {
          return null;
        } else {
          n1 = Math.floor(n1 * precision);
          n1 += up ? precision : -precision;
          return String(n1 / precision);
        }
      }
      return null;
    }
    textReplace(value, up) {
      return this.valueSetsReplace(this._defaultValueSet, value, up);
    }
    valueSetsReplace(valueSets, value, up) {
      let result = null;
      for (let i = 0, len = valueSets.length; result === null && i < len; i++) {
        result = this.valueSetReplace(valueSets[i], value, up);
      }
      return result;
    }
    valueSetReplace(valueSet, value, up) {
      let idx = valueSet.indexOf(value);
      if (idx >= 0) {
        idx += up ? 1 : -1;
        if (idx < 0) {
          idx = valueSet.length - 1;
        } else {
          idx %= valueSet.length;
        }
        return valueSet[idx];
      }
      return null;
    }
  };
  BasicInplaceReplace.INSTANCE = new BasicInplaceReplace();

  // node_modules/monaco-editor/esm/vs/base/common/linkedList.js
  var Node = class {
    constructor(element) {
      this.element = element;
      this.next = Node.Undefined;
      this.prev = Node.Undefined;
    }
  };
  Node.Undefined = new Node(void 0);
  var LinkedList = class {
    constructor() {
      this._first = Node.Undefined;
      this._last = Node.Undefined;
      this._size = 0;
    }
    get size() {
      return this._size;
    }
    isEmpty() {
      return this._first === Node.Undefined;
    }
    clear() {
      this._first = Node.Undefined;
      this._last = Node.Undefined;
      this._size = 0;
    }
    unshift(element) {
      return this._insert(element, false);
    }
    push(element) {
      return this._insert(element, true);
    }
    _insert(element, atTheEnd) {
      const newNode = new Node(element);
      if (this._first === Node.Undefined) {
        this._first = newNode;
        this._last = newNode;
      } else if (atTheEnd) {
        const oldLast = this._last;
        this._last = newNode;
        newNode.prev = oldLast;
        oldLast.next = newNode;
      } else {
        const oldFirst = this._first;
        this._first = newNode;
        newNode.next = oldFirst;
        oldFirst.prev = newNode;
      }
      this._size += 1;
      let didRemove = false;
      return () => {
        if (!didRemove) {
          didRemove = true;
          this._remove(newNode);
        }
      };
    }
    shift() {
      if (this._first === Node.Undefined) {
        return void 0;
      } else {
        const res = this._first.element;
        this._remove(this._first);
        return res;
      }
    }
    pop() {
      if (this._last === Node.Undefined) {
        return void 0;
      } else {
        const res = this._last.element;
        this._remove(this._last);
        return res;
      }
    }
    _remove(node) {
      if (node.prev !== Node.Undefined && node.next !== Node.Undefined) {
        const anchor = node.prev;
        anchor.next = node.next;
        node.next.prev = anchor;
      } else if (node.prev === Node.Undefined && node.next === Node.Undefined) {
        this._first = Node.Undefined;
        this._last = Node.Undefined;
      } else if (node.next === Node.Undefined) {
        this._last = this._last.prev;
        this._last.next = Node.Undefined;
      } else if (node.prev === Node.Undefined) {
        this._first = this._first.next;
        this._first.prev = Node.Undefined;
      }
      this._size -= 1;
    }
    *[Symbol.iterator]() {
      let node = this._first;
      while (node !== Node.Undefined) {
        yield node.element;
        node = node.next;
      }
    }
    toArray() {
      const result = [];
      for (let node = this._first; node !== Node.Undefined; node = node.next) {
        result.push(node.element);
      }
      return result;
    }
  };

  // node_modules/monaco-editor/esm/vs/base/common/event.js
  var Event;
  (function(Event2) {
    Event2.None = () => Disposable.None;
    function once(event) {
      return (listener, thisArgs = null, disposables) => {
        let didFire = false;
        let result;
        result = event((e) => {
          if (didFire) {
            return;
          } else if (result) {
            result.dispose();
          } else {
            didFire = true;
          }
          return listener.call(thisArgs, e);
        }, null, disposables);
        if (didFire) {
          result.dispose();
        }
        return result;
      };
    }
    Event2.once = once;
    function map(event, map2) {
      return snapshot((listener, thisArgs = null, disposables) => event((i) => listener.call(thisArgs, map2(i)), null, disposables));
    }
    Event2.map = map;
    function forEach(event, each) {
      return snapshot((listener, thisArgs = null, disposables) => event((i) => {
        each(i);
        listener.call(thisArgs, i);
      }, null, disposables));
    }
    Event2.forEach = forEach;
    function filter(event, filter2) {
      return snapshot((listener, thisArgs = null, disposables) => event((e) => filter2(e) && listener.call(thisArgs, e), null, disposables));
    }
    Event2.filter = filter;
    function signal(event) {
      return event;
    }
    Event2.signal = signal;
    function any(...events) {
      return (listener, thisArgs = null, disposables) => combinedDisposable(...events.map((event) => event((e) => listener.call(thisArgs, e), null, disposables)));
    }
    Event2.any = any;
    function reduce(event, merge, initial) {
      let output = initial;
      return map(event, (e) => {
        output = merge(output, e);
        return output;
      });
    }
    Event2.reduce = reduce;
    function snapshot(event) {
      let listener;
      const emitter = new Emitter({
        onFirstListenerAdd() {
          listener = event(emitter.fire, emitter);
        },
        onLastListenerRemove() {
          listener.dispose();
        }
      });
      return emitter.event;
    }
    Event2.snapshot = snapshot;
    function debounce(event, merge, delay = 100, leading = false, leakWarningThreshold) {
      let subscription;
      let output = void 0;
      let handle = void 0;
      let numDebouncedCalls = 0;
      const emitter = new Emitter({
        leakWarningThreshold,
        onFirstListenerAdd() {
          subscription = event((cur) => {
            numDebouncedCalls++;
            output = merge(output, cur);
            if (leading && !handle) {
              emitter.fire(output);
              output = void 0;
            }
            clearTimeout(handle);
            handle = setTimeout(() => {
              const _output = output;
              output = void 0;
              handle = void 0;
              if (!leading || numDebouncedCalls > 1) {
                emitter.fire(_output);
              }
              numDebouncedCalls = 0;
            }, delay);
          });
        },
        onLastListenerRemove() {
          subscription.dispose();
        }
      });
      return emitter.event;
    }
    Event2.debounce = debounce;
    function stopwatch(event) {
      const start = new Date().getTime();
      return map(once(event), (_) => new Date().getTime() - start);
    }
    Event2.stopwatch = stopwatch;
    function latch(event) {
      let firstCall = true;
      let cache;
      return filter(event, (value) => {
        const shouldEmit = firstCall || value !== cache;
        firstCall = false;
        cache = value;
        return shouldEmit;
      });
    }
    Event2.latch = latch;
    function buffer(event, nextTick = false, _buffer = []) {
      let buffer2 = _buffer.slice();
      let listener = event((e) => {
        if (buffer2) {
          buffer2.push(e);
        } else {
          emitter.fire(e);
        }
      });
      const flush = () => {
        if (buffer2) {
          buffer2.forEach((e) => emitter.fire(e));
        }
        buffer2 = null;
      };
      const emitter = new Emitter({
        onFirstListenerAdd() {
          if (!listener) {
            listener = event((e) => emitter.fire(e));
          }
        },
        onFirstListenerDidAdd() {
          if (buffer2) {
            if (nextTick) {
              setTimeout(flush);
            } else {
              flush();
            }
          }
        },
        onLastListenerRemove() {
          if (listener) {
            listener.dispose();
          }
          listener = null;
        }
      });
      return emitter.event;
    }
    Event2.buffer = buffer;
    class ChainableEvent {
      constructor(event) {
        this.event = event;
      }
      map(fn) {
        return new ChainableEvent(map(this.event, fn));
      }
      forEach(fn) {
        return new ChainableEvent(forEach(this.event, fn));
      }
      filter(fn) {
        return new ChainableEvent(filter(this.event, fn));
      }
      reduce(merge, initial) {
        return new ChainableEvent(reduce(this.event, merge, initial));
      }
      latch() {
        return new ChainableEvent(latch(this.event));
      }
      debounce(merge, delay = 100, leading = false, leakWarningThreshold) {
        return new ChainableEvent(debounce(this.event, merge, delay, leading, leakWarningThreshold));
      }
      on(listener, thisArgs, disposables) {
        return this.event(listener, thisArgs, disposables);
      }
      once(listener, thisArgs, disposables) {
        return once(this.event)(listener, thisArgs, disposables);
      }
    }
    function chain(event) {
      return new ChainableEvent(event);
    }
    Event2.chain = chain;
    function fromNodeEventEmitter(emitter, eventName, map2 = (id) => id) {
      const fn = (...args) => result.fire(map2(...args));
      const onFirstListenerAdd = () => emitter.on(eventName, fn);
      const onLastListenerRemove = () => emitter.removeListener(eventName, fn);
      const result = new Emitter({ onFirstListenerAdd, onLastListenerRemove });
      return result.event;
    }
    Event2.fromNodeEventEmitter = fromNodeEventEmitter;
    function fromDOMEventEmitter(emitter, eventName, map2 = (id) => id) {
      const fn = (...args) => result.fire(map2(...args));
      const onFirstListenerAdd = () => emitter.addEventListener(eventName, fn);
      const onLastListenerRemove = () => emitter.removeEventListener(eventName, fn);
      const result = new Emitter({ onFirstListenerAdd, onLastListenerRemove });
      return result.event;
    }
    Event2.fromDOMEventEmitter = fromDOMEventEmitter;
    function fromPromise(promise) {
      const emitter = new Emitter();
      let shouldEmit = false;
      promise.then(void 0, () => null).then(() => {
        if (!shouldEmit) {
          setTimeout(() => emitter.fire(void 0), 0);
        } else {
          emitter.fire(void 0);
        }
      });
      shouldEmit = true;
      return emitter.event;
    }
    Event2.fromPromise = fromPromise;
    function toPromise(event) {
      return new Promise((c) => once(event)(c));
    }
    Event2.toPromise = toPromise;
  })(Event || (Event = {}));
  var _globalLeakWarningThreshold = -1;
  var LeakageMonitor = class {
    constructor(customThreshold, name = Math.random().toString(18).slice(2, 5)) {
      this.customThreshold = customThreshold;
      this.name = name;
      this._warnCountdown = 0;
    }
    dispose() {
      if (this._stacks) {
        this._stacks.clear();
      }
    }
    check(listenerCount) {
      let threshold = _globalLeakWarningThreshold;
      if (typeof this.customThreshold === "number") {
        threshold = this.customThreshold;
      }
      if (threshold <= 0 || listenerCount < threshold) {
        return void 0;
      }
      if (!this._stacks) {
        this._stacks = /* @__PURE__ */ new Map();
      }
      const stack = new Error().stack.split("\n").slice(3).join("\n");
      const count = this._stacks.get(stack) || 0;
      this._stacks.set(stack, count + 1);
      this._warnCountdown -= 1;
      if (this._warnCountdown <= 0) {
        this._warnCountdown = threshold * 0.5;
        let topStack;
        let topCount = 0;
        for (const [stack2, count2] of this._stacks) {
          if (!topStack || topCount < count2) {
            topStack = stack2;
            topCount = count2;
          }
        }
        console.warn(`[${this.name}] potential listener LEAK detected, having ${listenerCount} listeners already. MOST frequent listener (${topCount}):`);
        console.warn(topStack);
      }
      return () => {
        const count2 = this._stacks.get(stack) || 0;
        this._stacks.set(stack, count2 - 1);
      };
    }
  };
  var Emitter = class {
    constructor(options) {
      this._disposed = false;
      this._options = options;
      this._leakageMon = _globalLeakWarningThreshold > 0 ? new LeakageMonitor(this._options && this._options.leakWarningThreshold) : void 0;
    }
    get event() {
      if (!this._event) {
        this._event = (listener, thisArgs, disposables) => {
          if (!this._listeners) {
            this._listeners = new LinkedList();
          }
          const firstListener = this._listeners.isEmpty();
          if (firstListener && this._options && this._options.onFirstListenerAdd) {
            this._options.onFirstListenerAdd(this);
          }
          const remove = this._listeners.push(!thisArgs ? listener : [listener, thisArgs]);
          if (firstListener && this._options && this._options.onFirstListenerDidAdd) {
            this._options.onFirstListenerDidAdd(this);
          }
          if (this._options && this._options.onListenerDidAdd) {
            this._options.onListenerDidAdd(this, listener, thisArgs);
          }
          let removeMonitor;
          if (this._leakageMon) {
            removeMonitor = this._leakageMon.check(this._listeners.size);
          }
          let result;
          result = {
            dispose: () => {
              if (removeMonitor) {
                removeMonitor();
              }
              result.dispose = Emitter._noop;
              if (!this._disposed) {
                remove();
                if (this._options && this._options.onLastListenerRemove) {
                  const hasListeners = this._listeners && !this._listeners.isEmpty();
                  if (!hasListeners) {
                    this._options.onLastListenerRemove(this);
                  }
                }
              }
            }
          };
          if (disposables instanceof DisposableStore) {
            disposables.add(result);
          } else if (Array.isArray(disposables)) {
            disposables.push(result);
          }
          return result;
        };
      }
      return this._event;
    }
    fire(event) {
      if (this._listeners) {
        if (!this._deliveryQueue) {
          this._deliveryQueue = new LinkedList();
        }
        for (let listener of this._listeners) {
          this._deliveryQueue.push([listener, event]);
        }
        while (this._deliveryQueue.size > 0) {
          const [listener, event2] = this._deliveryQueue.shift();
          try {
            if (typeof listener === "function") {
              listener.call(void 0, event2);
            } else {
              listener[0].call(listener[1], event2);
            }
          } catch (e) {
            onUnexpectedError(e);
          }
        }
      }
    }
    dispose() {
      if (this._listeners) {
        this._listeners.clear();
      }
      if (this._deliveryQueue) {
        this._deliveryQueue.clear();
      }
      if (this._leakageMon) {
        this._leakageMon.dispose();
      }
      this._disposed = true;
    }
  };
  Emitter._noop = function() {
  };

  // node_modules/monaco-editor/esm/vs/base/common/cancellation.js
  var shortcutEvent = Object.freeze(function(callback, context) {
    const handle = setTimeout(callback.bind(context), 0);
    return { dispose() {
      clearTimeout(handle);
    } };
  });
  var CancellationToken;
  (function(CancellationToken2) {
    function isCancellationToken(thing) {
      if (thing === CancellationToken2.None || thing === CancellationToken2.Cancelled) {
        return true;
      }
      if (thing instanceof MutableToken) {
        return true;
      }
      if (!thing || typeof thing !== "object") {
        return false;
      }
      return typeof thing.isCancellationRequested === "boolean" && typeof thing.onCancellationRequested === "function";
    }
    CancellationToken2.isCancellationToken = isCancellationToken;
    CancellationToken2.None = Object.freeze({
      isCancellationRequested: false,
      onCancellationRequested: Event.None
    });
    CancellationToken2.Cancelled = Object.freeze({
      isCancellationRequested: true,
      onCancellationRequested: shortcutEvent
    });
  })(CancellationToken || (CancellationToken = {}));
  var MutableToken = class {
    constructor() {
      this._isCancelled = false;
      this._emitter = null;
    }
    cancel() {
      if (!this._isCancelled) {
        this._isCancelled = true;
        if (this._emitter) {
          this._emitter.fire(void 0);
          this.dispose();
        }
      }
    }
    get isCancellationRequested() {
      return this._isCancelled;
    }
    get onCancellationRequested() {
      if (this._isCancelled) {
        return shortcutEvent;
      }
      if (!this._emitter) {
        this._emitter = new Emitter();
      }
      return this._emitter.event;
    }
    dispose() {
      if (this._emitter) {
        this._emitter.dispose();
        this._emitter = null;
      }
    }
  };
  var CancellationTokenSource = class {
    constructor(parent) {
      this._token = void 0;
      this._parentListener = void 0;
      this._parentListener = parent && parent.onCancellationRequested(this.cancel, this);
    }
    get token() {
      if (!this._token) {
        this._token = new MutableToken();
      }
      return this._token;
    }
    cancel() {
      if (!this._token) {
        this._token = CancellationToken.Cancelled;
      } else if (this._token instanceof MutableToken) {
        this._token.cancel();
      }
    }
    dispose(cancel = false) {
      if (cancel) {
        this.cancel();
      }
      if (this._parentListener) {
        this._parentListener.dispose();
      }
      if (!this._token) {
        this._token = CancellationToken.None;
      } else if (this._token instanceof MutableToken) {
        this._token.dispose();
      }
    }
  };

  // node_modules/monaco-editor/esm/vs/base/common/keyCodes.js
  var KeyCodeStrMap = class {
    constructor() {
      this._keyCodeToStr = [];
      this._strToKeyCode = /* @__PURE__ */ Object.create(null);
    }
    define(keyCode, str) {
      this._keyCodeToStr[keyCode] = str;
      this._strToKeyCode[str.toLowerCase()] = keyCode;
    }
    keyCodeToStr(keyCode) {
      return this._keyCodeToStr[keyCode];
    }
    strToKeyCode(str) {
      return this._strToKeyCode[str.toLowerCase()] || 0;
    }
  };
  var uiMap = new KeyCodeStrMap();
  var userSettingsUSMap = new KeyCodeStrMap();
  var userSettingsGeneralMap = new KeyCodeStrMap();
  (function() {
    function define(keyCode, uiLabel, usUserSettingsLabel = uiLabel, generalUserSettingsLabel = usUserSettingsLabel) {
      uiMap.define(keyCode, uiLabel);
      userSettingsUSMap.define(keyCode, usUserSettingsLabel);
      userSettingsGeneralMap.define(keyCode, generalUserSettingsLabel);
    }
    define(0, "unknown");
    define(1, "Backspace");
    define(2, "Tab");
    define(3, "Enter");
    define(4, "Shift");
    define(5, "Ctrl");
    define(6, "Alt");
    define(7, "PauseBreak");
    define(8, "CapsLock");
    define(9, "Escape");
    define(10, "Space");
    define(11, "PageUp");
    define(12, "PageDown");
    define(13, "End");
    define(14, "Home");
    define(15, "LeftArrow", "Left");
    define(16, "UpArrow", "Up");
    define(17, "RightArrow", "Right");
    define(18, "DownArrow", "Down");
    define(19, "Insert");
    define(20, "Delete");
    define(21, "0");
    define(22, "1");
    define(23, "2");
    define(24, "3");
    define(25, "4");
    define(26, "5");
    define(27, "6");
    define(28, "7");
    define(29, "8");
    define(30, "9");
    define(31, "A");
    define(32, "B");
    define(33, "C");
    define(34, "D");
    define(35, "E");
    define(36, "F");
    define(37, "G");
    define(38, "H");
    define(39, "I");
    define(40, "J");
    define(41, "K");
    define(42, "L");
    define(43, "M");
    define(44, "N");
    define(45, "O");
    define(46, "P");
    define(47, "Q");
    define(48, "R");
    define(49, "S");
    define(50, "T");
    define(51, "U");
    define(52, "V");
    define(53, "W");
    define(54, "X");
    define(55, "Y");
    define(56, "Z");
    define(57, "Meta");
    define(58, "ContextMenu");
    define(59, "F1");
    define(60, "F2");
    define(61, "F3");
    define(62, "F4");
    define(63, "F5");
    define(64, "F6");
    define(65, "F7");
    define(66, "F8");
    define(67, "F9");
    define(68, "F10");
    define(69, "F11");
    define(70, "F12");
    define(71, "F13");
    define(72, "F14");
    define(73, "F15");
    define(74, "F16");
    define(75, "F17");
    define(76, "F18");
    define(77, "F19");
    define(78, "NumLock");
    define(79, "ScrollLock");
    define(80, ";", ";", "OEM_1");
    define(81, "=", "=", "OEM_PLUS");
    define(82, ",", ",", "OEM_COMMA");
    define(83, "-", "-", "OEM_MINUS");
    define(84, ".", ".", "OEM_PERIOD");
    define(85, "/", "/", "OEM_2");
    define(86, "`", "`", "OEM_3");
    define(110, "ABNT_C1");
    define(111, "ABNT_C2");
    define(87, "[", "[", "OEM_4");
    define(88, "\\", "\\", "OEM_5");
    define(89, "]", "]", "OEM_6");
    define(90, "'", "'", "OEM_7");
    define(91, "OEM_8");
    define(92, "OEM_102");
    define(93, "NumPad0");
    define(94, "NumPad1");
    define(95, "NumPad2");
    define(96, "NumPad3");
    define(97, "NumPad4");
    define(98, "NumPad5");
    define(99, "NumPad6");
    define(100, "NumPad7");
    define(101, "NumPad8");
    define(102, "NumPad9");
    define(103, "NumPad_Multiply");
    define(104, "NumPad_Add");
    define(105, "NumPad_Separator");
    define(106, "NumPad_Subtract");
    define(107, "NumPad_Decimal");
    define(108, "NumPad_Divide");
  })();
  var KeyCodeUtils;
  (function(KeyCodeUtils2) {
    function toString(keyCode) {
      return uiMap.keyCodeToStr(keyCode);
    }
    KeyCodeUtils2.toString = toString;
    function fromString(key) {
      return uiMap.strToKeyCode(key);
    }
    KeyCodeUtils2.fromString = fromString;
    function toUserSettingsUS(keyCode) {
      return userSettingsUSMap.keyCodeToStr(keyCode);
    }
    KeyCodeUtils2.toUserSettingsUS = toUserSettingsUS;
    function toUserSettingsGeneral(keyCode) {
      return userSettingsGeneralMap.keyCodeToStr(keyCode);
    }
    KeyCodeUtils2.toUserSettingsGeneral = toUserSettingsGeneral;
    function fromUserSettings(key) {
      return userSettingsUSMap.strToKeyCode(key) || userSettingsGeneralMap.strToKeyCode(key);
    }
    KeyCodeUtils2.fromUserSettings = fromUserSettings;
  })(KeyCodeUtils || (KeyCodeUtils = {}));
  function KeyChord(firstPart, secondPart) {
    const chordPart = (secondPart & 65535) << 16 >>> 0;
    return (firstPart | chordPart) >>> 0;
  }

  // node_modules/monaco-editor/esm/vs/editor/common/core/selection.js
  var Selection = class extends Range {
    constructor(selectionStartLineNumber, selectionStartColumn, positionLineNumber, positionColumn) {
      super(selectionStartLineNumber, selectionStartColumn, positionLineNumber, positionColumn);
      this.selectionStartLineNumber = selectionStartLineNumber;
      this.selectionStartColumn = selectionStartColumn;
      this.positionLineNumber = positionLineNumber;
      this.positionColumn = positionColumn;
    }
    toString() {
      return "[" + this.selectionStartLineNumber + "," + this.selectionStartColumn + " -> " + this.positionLineNumber + "," + this.positionColumn + "]";
    }
    equalsSelection(other) {
      return Selection.selectionsEqual(this, other);
    }
    static selectionsEqual(a2, b) {
      return a2.selectionStartLineNumber === b.selectionStartLineNumber && a2.selectionStartColumn === b.selectionStartColumn && a2.positionLineNumber === b.positionLineNumber && a2.positionColumn === b.positionColumn;
    }
    getDirection() {
      if (this.selectionStartLineNumber === this.startLineNumber && this.selectionStartColumn === this.startColumn) {
        return 0;
      }
      return 1;
    }
    setEndPosition(endLineNumber, endColumn) {
      if (this.getDirection() === 0) {
        return new Selection(this.startLineNumber, this.startColumn, endLineNumber, endColumn);
      }
      return new Selection(endLineNumber, endColumn, this.startLineNumber, this.startColumn);
    }
    getPosition() {
      return new Position(this.positionLineNumber, this.positionColumn);
    }
    setStartPosition(startLineNumber, startColumn) {
      if (this.getDirection() === 0) {
        return new Selection(startLineNumber, startColumn, this.endLineNumber, this.endColumn);
      }
      return new Selection(this.endLineNumber, this.endColumn, startLineNumber, startColumn);
    }
    static fromPositions(start, end = start) {
      return new Selection(start.lineNumber, start.column, end.lineNumber, end.column);
    }
    static liftSelection(sel) {
      return new Selection(sel.selectionStartLineNumber, sel.selectionStartColumn, sel.positionLineNumber, sel.positionColumn);
    }
    static selectionsArrEqual(a2, b) {
      if (a2 && !b || !a2 && b) {
        return false;
      }
      if (!a2 && !b) {
        return true;
      }
      if (a2.length !== b.length) {
        return false;
      }
      for (let i = 0, len = a2.length; i < len; i++) {
        if (!this.selectionsEqual(a2[i], b[i])) {
          return false;
        }
      }
      return true;
    }
    static isISelection(obj) {
      return obj && typeof obj.selectionStartLineNumber === "number" && typeof obj.selectionStartColumn === "number" && typeof obj.positionLineNumber === "number" && typeof obj.positionColumn === "number";
    }
    static createWithDirection(startLineNumber, startColumn, endLineNumber, endColumn, direction) {
      if (direction === 0) {
        return new Selection(startLineNumber, startColumn, endLineNumber, endColumn);
      }
      return new Selection(endLineNumber, endColumn, startLineNumber, startColumn);
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/core/token.js
  var Token = class {
    constructor(offset, type, language) {
      this.offset = offset | 0;
      this.type = type;
      this.language = language;
    }
    toString() {
      return "(" + this.offset + ", " + this.type + ")";
    }
  };

  // node_modules/monaco-editor/esm/vs/editor/common/standalone/standaloneEnums.js
  var AccessibilitySupport;
  (function(AccessibilitySupport2) {
    AccessibilitySupport2[AccessibilitySupport2["Unknown"] = 0] = "Unknown";
    AccessibilitySupport2[AccessibilitySupport2["Disabled"] = 1] = "Disabled";
    AccessibilitySupport2[AccessibilitySupport2["Enabled"] = 2] = "Enabled";
  })(AccessibilitySupport || (AccessibilitySupport = {}));
  var CompletionItemInsertTextRule;
  (function(CompletionItemInsertTextRule2) {
    CompletionItemInsertTextRule2[CompletionItemInsertTextRule2["KeepWhitespace"] = 1] = "KeepWhitespace";
    CompletionItemInsertTextRule2[CompletionItemInsertTextRule2["InsertAsSnippet"] = 4] = "InsertAsSnippet";
  })(CompletionItemInsertTextRule || (CompletionItemInsertTextRule = {}));
  var CompletionItemKind;
  (function(CompletionItemKind3) {
    CompletionItemKind3[CompletionItemKind3["Method"] = 0] = "Method";
    CompletionItemKind3[CompletionItemKind3["Function"] = 1] = "Function";
    CompletionItemKind3[CompletionItemKind3["Constructor"] = 2] = "Constructor";
    CompletionItemKind3[CompletionItemKind3["Field"] = 3] = "Field";
    CompletionItemKind3[CompletionItemKind3["Variable"] = 4] = "Variable";
    CompletionItemKind3[CompletionItemKind3["Class"] = 5] = "Class";
    CompletionItemKind3[CompletionItemKind3["Struct"] = 6] = "Struct";
    CompletionItemKind3[CompletionItemKind3["Interface"] = 7] = "Interface";
    CompletionItemKind3[CompletionItemKind3["Module"] = 8] = "Module";
    CompletionItemKind3[CompletionItemKind3["Property"] = 9] = "Property";
    CompletionItemKind3[CompletionItemKind3["Event"] = 10] = "Event";
    CompletionItemKind3[CompletionItemKind3["Operator"] = 11] = "Operator";
    CompletionItemKind3[CompletionItemKind3["Unit"] = 12] = "Unit";
    CompletionItemKind3[CompletionItemKind3["Value"] = 13] = "Value";
    CompletionItemKind3[CompletionItemKind3["Constant"] = 14] = "Constant";
    CompletionItemKind3[CompletionItemKind3["Enum"] = 15] = "Enum";
    CompletionItemKind3[CompletionItemKind3["EnumMember"] = 16] = "EnumMember";
    CompletionItemKind3[CompletionItemKind3["Keyword"] = 17] = "Keyword";
    CompletionItemKind3[CompletionItemKind3["Text"] = 18] = "Text";
    CompletionItemKind3[CompletionItemKind3["Color"] = 19] = "Color";
    CompletionItemKind3[CompletionItemKind3["File"] = 20] = "File";
    CompletionItemKind3[CompletionItemKind3["Reference"] = 21] = "Reference";
    CompletionItemKind3[CompletionItemKind3["Customcolor"] = 22] = "Customcolor";
    CompletionItemKind3[CompletionItemKind3["Folder"] = 23] = "Folder";
    CompletionItemKind3[CompletionItemKind3["TypeParameter"] = 24] = "TypeParameter";
    CompletionItemKind3[CompletionItemKind3["User"] = 25] = "User";
    CompletionItemKind3[CompletionItemKind3["Issue"] = 26] = "Issue";
    CompletionItemKind3[CompletionItemKind3["Snippet"] = 27] = "Snippet";
  })(CompletionItemKind || (CompletionItemKind = {}));
  var CompletionItemTag;
  (function(CompletionItemTag3) {
    CompletionItemTag3[CompletionItemTag3["Deprecated"] = 1] = "Deprecated";
  })(CompletionItemTag || (CompletionItemTag = {}));
  var CompletionTriggerKind;
  (function(CompletionTriggerKind2) {
    CompletionTriggerKind2[CompletionTriggerKind2["Invoke"] = 0] = "Invoke";
    CompletionTriggerKind2[CompletionTriggerKind2["TriggerCharacter"] = 1] = "TriggerCharacter";
    CompletionTriggerKind2[CompletionTriggerKind2["TriggerForIncompleteCompletions"] = 2] = "TriggerForIncompleteCompletions";
  })(CompletionTriggerKind || (CompletionTriggerKind = {}));
  var ContentWidgetPositionPreference;
  (function(ContentWidgetPositionPreference2) {
    ContentWidgetPositionPreference2[ContentWidgetPositionPreference2["EXACT"] = 0] = "EXACT";
    ContentWidgetPositionPreference2[ContentWidgetPositionPreference2["ABOVE"] = 1] = "ABOVE";
    ContentWidgetPositionPreference2[ContentWidgetPositionPreference2["BELOW"] = 2] = "BELOW";
  })(ContentWidgetPositionPreference || (ContentWidgetPositionPreference = {}));
  var CursorChangeReason;
  (function(CursorChangeReason2) {
    CursorChangeReason2[CursorChangeReason2["NotSet"] = 0] = "NotSet";
    CursorChangeReason2[CursorChangeReason2["ContentFlush"] = 1] = "ContentFlush";
    CursorChangeReason2[CursorChangeReason2["RecoverFromMarkers"] = 2] = "RecoverFromMarkers";
    CursorChangeReason2[CursorChangeReason2["Explicit"] = 3] = "Explicit";
    CursorChangeReason2[CursorChangeReason2["Paste"] = 4] = "Paste";
    CursorChangeReason2[CursorChangeReason2["Undo"] = 5] = "Undo";
    CursorChangeReason2[CursorChangeReason2["Redo"] = 6] = "Redo";
  })(CursorChangeReason || (CursorChangeReason = {}));
  var DefaultEndOfLine;
  (function(DefaultEndOfLine2) {
    DefaultEndOfLine2[DefaultEndOfLine2["LF"] = 1] = "LF";
    DefaultEndOfLine2[DefaultEndOfLine2["CRLF"] = 2] = "CRLF";
  })(DefaultEndOfLine || (DefaultEndOfLine = {}));
  var DocumentHighlightKind;
  (function(DocumentHighlightKind3) {
    DocumentHighlightKind3[DocumentHighlightKind3["Text"] = 0] = "Text";
    DocumentHighlightKind3[DocumentHighlightKind3["Read"] = 1] = "Read";
    DocumentHighlightKind3[DocumentHighlightKind3["Write"] = 2] = "Write";
  })(DocumentHighlightKind || (DocumentHighlightKind = {}));
  var EditorAutoIndentStrategy;
  (function(EditorAutoIndentStrategy2) {
    EditorAutoIndentStrategy2[EditorAutoIndentStrategy2["None"] = 0] = "None";
    EditorAutoIndentStrategy2[EditorAutoIndentStrategy2["Keep"] = 1] = "Keep";
    EditorAutoIndentStrategy2[EditorAutoIndentStrategy2["Brackets"] = 2] = "Brackets";
    EditorAutoIndentStrategy2[EditorAutoIndentStrategy2["Advanced"] = 3] = "Advanced";
    EditorAutoIndentStrategy2[EditorAutoIndentStrategy2["Full"] = 4] = "Full";
  })(EditorAutoIndentStrategy || (EditorAutoIndentStrategy = {}));
  var EditorOption;
  (function(EditorOption2) {
    EditorOption2[EditorOption2["acceptSuggestionOnCommitCharacter"] = 0] = "acceptSuggestionOnCommitCharacter";
    EditorOption2[EditorOption2["acceptSuggestionOnEnter"] = 1] = "acceptSuggestionOnEnter";
    EditorOption2[EditorOption2["accessibilitySupport"] = 2] = "accessibilitySupport";
    EditorOption2[EditorOption2["accessibilityPageSize"] = 3] = "accessibilityPageSize";
    EditorOption2[EditorOption2["ariaLabel"] = 4] = "ariaLabel";
    EditorOption2[EditorOption2["autoClosingBrackets"] = 5] = "autoClosingBrackets";
    EditorOption2[EditorOption2["autoClosingOvertype"] = 6] = "autoClosingOvertype";
    EditorOption2[EditorOption2["autoClosingQuotes"] = 7] = "autoClosingQuotes";
    EditorOption2[EditorOption2["autoIndent"] = 8] = "autoIndent";
    EditorOption2[EditorOption2["automaticLayout"] = 9] = "automaticLayout";
    EditorOption2[EditorOption2["autoSurround"] = 10] = "autoSurround";
    EditorOption2[EditorOption2["codeLens"] = 11] = "codeLens";
    EditorOption2[EditorOption2["colorDecorators"] = 12] = "colorDecorators";
    EditorOption2[EditorOption2["columnSelection"] = 13] = "columnSelection";
    EditorOption2[EditorOption2["comments"] = 14] = "comments";
    EditorOption2[EditorOption2["contextmenu"] = 15] = "contextmenu";
    EditorOption2[EditorOption2["copyWithSyntaxHighlighting"] = 16] = "copyWithSyntaxHighlighting";
    EditorOption2[EditorOption2["cursorBlinking"] = 17] = "cursorBlinking";
    EditorOption2[EditorOption2["cursorSmoothCaretAnimation"] = 18] = "cursorSmoothCaretAnimation";
    EditorOption2[EditorOption2["cursorStyle"] = 19] = "cursorStyle";
    EditorOption2[EditorOption2["cursorSurroundingLines"] = 20] = "cursorSurroundingLines";
    EditorOption2[EditorOption2["cursorSurroundingLinesStyle"] = 21] = "cursorSurroundingLinesStyle";
    EditorOption2[EditorOption2["cursorWidth"] = 22] = "cursorWidth";
    EditorOption2[EditorOption2["disableLayerHinting"] = 23] = "disableLayerHinting";
    EditorOption2[EditorOption2["disableMonospaceOptimizations"] = 24] = "disableMonospaceOptimizations";
    EditorOption2[EditorOption2["dragAndDrop"] = 25] = "dragAndDrop";
    EditorOption2[EditorOption2["emptySelectionClipboard"] = 26] = "emptySelectionClipboard";
    EditorOption2[EditorOption2["extraEditorClassName"] = 27] = "extraEditorClassName";
    EditorOption2[EditorOption2["fastScrollSensitivity"] = 28] = "fastScrollSensitivity";
    EditorOption2[EditorOption2["find"] = 29] = "find";
    EditorOption2[EditorOption2["fixedOverflowWidgets"] = 30] = "fixedOverflowWidgets";
    EditorOption2[EditorOption2["folding"] = 31] = "folding";
    EditorOption2[EditorOption2["foldingStrategy"] = 32] = "foldingStrategy";
    EditorOption2[EditorOption2["foldingHighlight"] = 33] = "foldingHighlight";
    EditorOption2[EditorOption2["unfoldOnClickAfterEndOfLine"] = 34] = "unfoldOnClickAfterEndOfLine";
    EditorOption2[EditorOption2["fontFamily"] = 35] = "fontFamily";
    EditorOption2[EditorOption2["fontInfo"] = 36] = "fontInfo";
    EditorOption2[EditorOption2["fontLigatures"] = 37] = "fontLigatures";
    EditorOption2[EditorOption2["fontSize"] = 38] = "fontSize";
    EditorOption2[EditorOption2["fontWeight"] = 39] = "fontWeight";
    EditorOption2[EditorOption2["formatOnPaste"] = 40] = "formatOnPaste";
    EditorOption2[EditorOption2["formatOnType"] = 41] = "formatOnType";
    EditorOption2[EditorOption2["glyphMargin"] = 42] = "glyphMargin";
    EditorOption2[EditorOption2["gotoLocation"] = 43] = "gotoLocation";
    EditorOption2[EditorOption2["hideCursorInOverviewRuler"] = 44] = "hideCursorInOverviewRuler";
    EditorOption2[EditorOption2["highlightActiveIndentGuide"] = 45] = "highlightActiveIndentGuide";
    EditorOption2[EditorOption2["hover"] = 46] = "hover";
    EditorOption2[EditorOption2["inDiffEditor"] = 47] = "inDiffEditor";
    EditorOption2[EditorOption2["letterSpacing"] = 48] = "letterSpacing";
    EditorOption2[EditorOption2["lightbulb"] = 49] = "lightbulb";
    EditorOption2[EditorOption2["lineDecorationsWidth"] = 50] = "lineDecorationsWidth";
    EditorOption2[EditorOption2["lineHeight"] = 51] = "lineHeight";
    EditorOption2[EditorOption2["lineNumbers"] = 52] = "lineNumbers";
    EditorOption2[EditorOption2["lineNumbersMinChars"] = 53] = "lineNumbersMinChars";
    EditorOption2[EditorOption2["links"] = 54] = "links";
    EditorOption2[EditorOption2["matchBrackets"] = 55] = "matchBrackets";
    EditorOption2[EditorOption2["minimap"] = 56] = "minimap";
    EditorOption2[EditorOption2["mouseStyle"] = 57] = "mouseStyle";
    EditorOption2[EditorOption2["mouseWheelScrollSensitivity"] = 58] = "mouseWheelScrollSensitivity";
    EditorOption2[EditorOption2["mouseWheelZoom"] = 59] = "mouseWheelZoom";
    EditorOption2[EditorOption2["multiCursorMergeOverlapping"] = 60] = "multiCursorMergeOverlapping";
    EditorOption2[EditorOption2["multiCursorModifier"] = 61] = "multiCursorModifier";
    EditorOption2[EditorOption2["multiCursorPaste"] = 62] = "multiCursorPaste";
    EditorOption2[EditorOption2["occurrencesHighlight"] = 63] = "occurrencesHighlight";
    EditorOption2[EditorOption2["overviewRulerBorder"] = 64] = "overviewRulerBorder";
    EditorOption2[EditorOption2["overviewRulerLanes"] = 65] = "overviewRulerLanes";
    EditorOption2[EditorOption2["padding"] = 66] = "padding";
    EditorOption2[EditorOption2["parameterHints"] = 67] = "parameterHints";
    EditorOption2[EditorOption2["peekWidgetDefaultFocus"] = 68] = "peekWidgetDefaultFocus";
    EditorOption2[EditorOption2["definitionLinkOpensInPeek"] = 69] = "definitionLinkOpensInPeek";
    EditorOption2[EditorOption2["quickSuggestions"] = 70] = "quickSuggestions";
    EditorOption2[EditorOption2["quickSuggestionsDelay"] = 71] = "quickSuggestionsDelay";
    EditorOption2[EditorOption2["readOnly"] = 72] = "readOnly";
    EditorOption2[EditorOption2["renameOnType"] = 73] = "renameOnType";
    EditorOption2[EditorOption2["renderControlCharacters"] = 74] = "renderControlCharacters";
    EditorOption2[EditorOption2["renderIndentGuides"] = 75] = "renderIndentGuides";
    EditorOption2[EditorOption2["renderFinalNewline"] = 76] = "renderFinalNewline";
    EditorOption2[EditorOption2["renderLineHighlight"] = 77] = "renderLineHighlight";
    EditorOption2[EditorOption2["renderLineHighlightOnlyWhenFocus"] = 78] = "renderLineHighlightOnlyWhenFocus";
    EditorOption2[EditorOption2["renderValidationDecorations"] = 79] = "renderValidationDecorations";
    EditorOption2[EditorOption2["renderWhitespace"] = 80] = "renderWhitespace";
    EditorOption2[EditorOption2["revealHorizontalRightPadding"] = 81] = "revealHorizontalRightPadding";
    EditorOption2[EditorOption2["roundedSelection"] = 82] = "roundedSelection";
    EditorOption2[EditorOption2["rulers"] = 83] = "rulers";
    EditorOption2[EditorOption2["scrollbar"] = 84] = "scrollbar";
    EditorOption2[EditorOption2["scrollBeyondLastColumn"] = 85] = "scrollBeyondLastColumn";
    EditorOption2[EditorOption2["scrollBeyondLastLine"] = 86] = "scrollBeyondLastLine";
    EditorOption2[EditorOption2["scrollPredominantAxis"] = 87] = "scrollPredominantAxis";
    EditorOption2[EditorOption2["selectionClipboard"] = 88] = "selectionClipboard";
    EditorOption2[EditorOption2["selectionHighlight"] = 89] = "selectionHighlight";
    EditorOption2[EditorOption2["selectOnLineNumbers"] = 90] = "selectOnLineNumbers";
    EditorOption2[EditorOption2["showFoldingControls"] = 91] = "showFoldingControls";
    EditorOption2[EditorOption2["showUnused"] = 92] = "showUnused";
    EditorOption2[EditorOption2["snippetSuggestions"] = 93] = "snippetSuggestions";
    EditorOption2[EditorOption2["smoothScrolling"] = 94] = "smoothScrolling";
    EditorOption2[EditorOption2["stopRenderingLineAfter"] = 95] = "stopRenderingLineAfter";
    EditorOption2[EditorOption2["suggest"] = 96] = "suggest";
    EditorOption2[EditorOption2["suggestFontSize"] = 97] = "suggestFontSize";
    EditorOption2[EditorOption2["suggestLineHeight"] = 98] = "suggestLineHeight";
    EditorOption2[EditorOption2["suggestOnTriggerCharacters"] = 99] = "suggestOnTriggerCharacters";
    EditorOption2[EditorOption2["suggestSelection"] = 100] = "suggestSelection";
    EditorOption2[EditorOption2["tabCompletion"] = 101] = "tabCompletion";
    EditorOption2[EditorOption2["tabIndex"] = 102] = "tabIndex";
    EditorOption2[EditorOption2["unusualLineTerminators"] = 103] = "unusualLineTerminators";
    EditorOption2[EditorOption2["useTabStops"] = 104] = "useTabStops";
    EditorOption2[EditorOption2["wordSeparators"] = 105] = "wordSeparators";
    EditorOption2[EditorOption2["wordWrap"] = 106] = "wordWrap";
    EditorOption2[EditorOption2["wordWrapBreakAfterCharacters"] = 107] = "wordWrapBreakAfterCharacters";
    EditorOption2[EditorOption2["wordWrapBreakBeforeCharacters"] = 108] = "wordWrapBreakBeforeCharacters";
    EditorOption2[EditorOption2["wordWrapColumn"] = 109] = "wordWrapColumn";
    EditorOption2[EditorOption2["wordWrapMinified"] = 110] = "wordWrapMinified";
    EditorOption2[EditorOption2["wrappingIndent"] = 111] = "wrappingIndent";
    EditorOption2[EditorOption2["wrappingStrategy"] = 112] = "wrappingStrategy";
    EditorOption2[EditorOption2["showDeprecated"] = 113] = "showDeprecated";
    EditorOption2[EditorOption2["editorClassName"] = 114] = "editorClassName";
    EditorOption2[EditorOption2["pixelRatio"] = 115] = "pixelRatio";
    EditorOption2[EditorOption2["tabFocusMode"] = 116] = "tabFocusMode";
    EditorOption2[EditorOption2["layoutInfo"] = 117] = "layoutInfo";
    EditorOption2[EditorOption2["wrappingInfo"] = 118] = "wrappingInfo";
  })(EditorOption || (EditorOption = {}));
  var EndOfLinePreference;
  (function(EndOfLinePreference2) {
    EndOfLinePreference2[EndOfLinePreference2["TextDefined"] = 0] = "TextDefined";
    EndOfLinePreference2[EndOfLinePreference2["LF"] = 1] = "LF";
    EndOfLinePreference2[EndOfLinePreference2["CRLF"] = 2] = "CRLF";
  })(EndOfLinePreference || (EndOfLinePreference = {}));
  var EndOfLineSequence;
  (function(EndOfLineSequence2) {
    EndOfLineSequence2[EndOfLineSequence2["LF"] = 0] = "LF";
    EndOfLineSequence2[EndOfLineSequence2["CRLF"] = 1] = "CRLF";
  })(EndOfLineSequence || (EndOfLineSequence = {}));
  var IndentAction;
  (function(IndentAction2) {
    IndentAction2[IndentAction2["None"] = 0] = "None";
    IndentAction2[IndentAction2["Indent"] = 1] = "Indent";
    IndentAction2[IndentAction2["IndentOutdent"] = 2] = "IndentOutdent";
    IndentAction2[IndentAction2["Outdent"] = 3] = "Outdent";
  })(IndentAction || (IndentAction = {}));
  var KeyCode;
  (function(KeyCode2) {
    KeyCode2[KeyCode2["Unknown"] = 0] = "Unknown";
    KeyCode2[KeyCode2["Backspace"] = 1] = "Backspace";
    KeyCode2[KeyCode2["Tab"] = 2] = "Tab";
    KeyCode2[KeyCode2["Enter"] = 3] = "Enter";
    KeyCode2[KeyCode2["Shift"] = 4] = "Shift";
    KeyCode2[KeyCode2["Ctrl"] = 5] = "Ctrl";
    KeyCode2[KeyCode2["Alt"] = 6] = "Alt";
    KeyCode2[KeyCode2["PauseBreak"] = 7] = "PauseBreak";
    KeyCode2[KeyCode2["CapsLock"] = 8] = "CapsLock";
    KeyCode2[KeyCode2["Escape"] = 9] = "Escape";
    KeyCode2[KeyCode2["Space"] = 10] = "Space";
    KeyCode2[KeyCode2["PageUp"] = 11] = "PageUp";
    KeyCode2[KeyCode2["PageDown"] = 12] = "PageDown";
    KeyCode2[KeyCode2["End"] = 13] = "End";
    KeyCode2[KeyCode2["Home"] = 14] = "Home";
    KeyCode2[KeyCode2["LeftArrow"] = 15] = "LeftArrow";
    KeyCode2[KeyCode2["UpArrow"] = 16] = "UpArrow";
    KeyCode2[KeyCode2["RightArrow"] = 17] = "RightArrow";
    KeyCode2[KeyCode2["DownArrow"] = 18] = "DownArrow";
    KeyCode2[KeyCode2["Insert"] = 19] = "Insert";
    KeyCode2[KeyCode2["Delete"] = 20] = "Delete";
    KeyCode2[KeyCode2["KEY_0"] = 21] = "KEY_0";
    KeyCode2[KeyCode2["KEY_1"] = 22] = "KEY_1";
    KeyCode2[KeyCode2["KEY_2"] = 23] = "KEY_2";
    KeyCode2[KeyCode2["KEY_3"] = 24] = "KEY_3";
    KeyCode2[KeyCode2["KEY_4"] = 25] = "KEY_4";
    KeyCode2[KeyCode2["KEY_5"] = 26] = "KEY_5";
    KeyCode2[KeyCode2["KEY_6"] = 27] = "KEY_6";
    KeyCode2[KeyCode2["KEY_7"] = 28] = "KEY_7";
    KeyCode2[KeyCode2["KEY_8"] = 29] = "KEY_8";
    KeyCode2[KeyCode2["KEY_9"] = 30] = "KEY_9";
    KeyCode2[KeyCode2["KEY_A"] = 31] = "KEY_A";
    KeyCode2[KeyCode2["KEY_B"] = 32] = "KEY_B";
    KeyCode2[KeyCode2["KEY_C"] = 33] = "KEY_C";
    KeyCode2[KeyCode2["KEY_D"] = 34] = "KEY_D";
    KeyCode2[KeyCode2["KEY_E"] = 35] = "KEY_E";
    KeyCode2[KeyCode2["KEY_F"] = 36] = "KEY_F";
    KeyCode2[KeyCode2["KEY_G"] = 37] = "KEY_G";
    KeyCode2[KeyCode2["KEY_H"] = 38] = "KEY_H";
    KeyCode2[KeyCode2["KEY_I"] = 39] = "KEY_I";
    KeyCode2[KeyCode2["KEY_J"] = 40] = "KEY_J";
    KeyCode2[KeyCode2["KEY_K"] = 41] = "KEY_K";
    KeyCode2[KeyCode2["KEY_L"] = 42] = "KEY_L";
    KeyCode2[KeyCode2["KEY_M"] = 43] = "KEY_M";
    KeyCode2[KeyCode2["KEY_N"] = 44] = "KEY_N";
    KeyCode2[KeyCode2["KEY_O"] = 45] = "KEY_O";
    KeyCode2[KeyCode2["KEY_P"] = 46] = "KEY_P";
    KeyCode2[KeyCode2["KEY_Q"] = 47] = "KEY_Q";
    KeyCode2[KeyCode2["KEY_R"] = 48] = "KEY_R";
    KeyCode2[KeyCode2["KEY_S"] = 49] = "KEY_S";
    KeyCode2[KeyCode2["KEY_T"] = 50] = "KEY_T";
    KeyCode2[KeyCode2["KEY_U"] = 51] = "KEY_U";
    KeyCode2[KeyCode2["KEY_V"] = 52] = "KEY_V";
    KeyCode2[KeyCode2["KEY_W"] = 53] = "KEY_W";
    KeyCode2[KeyCode2["KEY_X"] = 54] = "KEY_X";
    KeyCode2[KeyCode2["KEY_Y"] = 55] = "KEY_Y";
    KeyCode2[KeyCode2["KEY_Z"] = 56] = "KEY_Z";
    KeyCode2[KeyCode2["Meta"] = 57] = "Meta";
    KeyCode2[KeyCode2["ContextMenu"] = 58] = "ContextMenu";
    KeyCode2[KeyCode2["F1"] = 59] = "F1";
    KeyCode2[KeyCode2["F2"] = 60] = "F2";
    KeyCode2[KeyCode2["F3"] = 61] = "F3";
    KeyCode2[KeyCode2["F4"] = 62] = "F4";
    KeyCode2[KeyCode2["F5"] = 63] = "F5";
    KeyCode2[KeyCode2["F6"] = 64] = "F6";
    KeyCode2[KeyCode2["F7"] = 65] = "F7";
    KeyCode2[KeyCode2["F8"] = 66] = "F8";
    KeyCode2[KeyCode2["F9"] = 67] = "F9";
    KeyCode2[KeyCode2["F10"] = 68] = "F10";
    KeyCode2[KeyCode2["F11"] = 69] = "F11";
    KeyCode2[KeyCode2["F12"] = 70] = "F12";
    KeyCode2[KeyCode2["F13"] = 71] = "F13";
    KeyCode2[KeyCode2["F14"] = 72] = "F14";
    KeyCode2[KeyCode2["F15"] = 73] = "F15";
    KeyCode2[KeyCode2["F16"] = 74] = "F16";
    KeyCode2[KeyCode2["F17"] = 75] = "F17";
    KeyCode2[KeyCode2["F18"] = 76] = "F18";
    KeyCode2[KeyCode2["F19"] = 77] = "F19";
    KeyCode2[KeyCode2["NumLock"] = 78] = "NumLock";
    KeyCode2[KeyCode2["ScrollLock"] = 79] = "ScrollLock";
    KeyCode2[KeyCode2["US_SEMICOLON"] = 80] = "US_SEMICOLON";
    KeyCode2[KeyCode2["US_EQUAL"] = 81] = "US_EQUAL";
    KeyCode2[KeyCode2["US_COMMA"] = 82] = "US_COMMA";
    KeyCode2[KeyCode2["US_MINUS"] = 83] = "US_MINUS";
    KeyCode2[KeyCode2["US_DOT"] = 84] = "US_DOT";
    KeyCode2[KeyCode2["US_SLASH"] = 85] = "US_SLASH";
    KeyCode2[KeyCode2["US_BACKTICK"] = 86] = "US_BACKTICK";
    KeyCode2[KeyCode2["US_OPEN_SQUARE_BRACKET"] = 87] = "US_OPEN_SQUARE_BRACKET";
    KeyCode2[KeyCode2["US_BACKSLASH"] = 88] = "US_BACKSLASH";
    KeyCode2[KeyCode2["US_CLOSE_SQUARE_BRACKET"] = 89] = "US_CLOSE_SQUARE_BRACKET";
    KeyCode2[KeyCode2["US_QUOTE"] = 90] = "US_QUOTE";
    KeyCode2[KeyCode2["OEM_8"] = 91] = "OEM_8";
    KeyCode2[KeyCode2["OEM_102"] = 92] = "OEM_102";
    KeyCode2[KeyCode2["NUMPAD_0"] = 93] = "NUMPAD_0";
    KeyCode2[KeyCode2["NUMPAD_1"] = 94] = "NUMPAD_1";
    KeyCode2[KeyCode2["NUMPAD_2"] = 95] = "NUMPAD_2";
    KeyCode2[KeyCode2["NUMPAD_3"] = 96] = "NUMPAD_3";
    KeyCode2[KeyCode2["NUMPAD_4"] = 97] = "NUMPAD_4";
    KeyCode2[KeyCode2["NUMPAD_5"] = 98] = "NUMPAD_5";
    KeyCode2[KeyCode2["NUMPAD_6"] = 99] = "NUMPAD_6";
    KeyCode2[KeyCode2["NUMPAD_7"] = 100] = "NUMPAD_7";
    KeyCode2[KeyCode2["NUMPAD_8"] = 101] = "NUMPAD_8";
    KeyCode2[KeyCode2["NUMPAD_9"] = 102] = "NUMPAD_9";
    KeyCode2[KeyCode2["NUMPAD_MULTIPLY"] = 103] = "NUMPAD_MULTIPLY";
    KeyCode2[KeyCode2["NUMPAD_ADD"] = 104] = "NUMPAD_ADD";
    KeyCode2[KeyCode2["NUMPAD_SEPARATOR"] = 105] = "NUMPAD_SEPARATOR";
    KeyCode2[KeyCode2["NUMPAD_SUBTRACT"] = 106] = "NUMPAD_SUBTRACT";
    KeyCode2[KeyCode2["NUMPAD_DECIMAL"] = 107] = "NUMPAD_DECIMAL";
    KeyCode2[KeyCode2["NUMPAD_DIVIDE"] = 108] = "NUMPAD_DIVIDE";
    KeyCode2[KeyCode2["KEY_IN_COMPOSITION"] = 109] = "KEY_IN_COMPOSITION";
    KeyCode2[KeyCode2["ABNT_C1"] = 110] = "ABNT_C1";
    KeyCode2[KeyCode2["ABNT_C2"] = 111] = "ABNT_C2";
    KeyCode2[KeyCode2["MAX_VALUE"] = 112] = "MAX_VALUE";
  })(KeyCode || (KeyCode = {}));
  var MarkerSeverity;
  (function(MarkerSeverity2) {
    MarkerSeverity2[MarkerSeverity2["Hint"] = 1] = "Hint";
    MarkerSeverity2[MarkerSeverity2["Info"] = 2] = "Info";
    MarkerSeverity2[MarkerSeverity2["Warning"] = 4] = "Warning";
    MarkerSeverity2[MarkerSeverity2["Error"] = 8] = "Error";
  })(MarkerSeverity || (MarkerSeverity = {}));
  var MarkerTag;
  (function(MarkerTag2) {
    MarkerTag2[MarkerTag2["Unnecessary"] = 1] = "Unnecessary";
    MarkerTag2[MarkerTag2["Deprecated"] = 2] = "Deprecated";
  })(MarkerTag || (MarkerTag = {}));
  var MinimapPosition;
  (function(MinimapPosition2) {
    MinimapPosition2[MinimapPosition2["Inline"] = 1] = "Inline";
    MinimapPosition2[MinimapPosition2["Gutter"] = 2] = "Gutter";
  })(MinimapPosition || (MinimapPosition = {}));
  var MouseTargetType;
  (function(MouseTargetType2) {
    MouseTargetType2[MouseTargetType2["UNKNOWN"] = 0] = "UNKNOWN";
    MouseTargetType2[MouseTargetType2["TEXTAREA"] = 1] = "TEXTAREA";
    MouseTargetType2[MouseTargetType2["GUTTER_GLYPH_MARGIN"] = 2] = "GUTTER_GLYPH_MARGIN";
    MouseTargetType2[MouseTargetType2["GUTTER_LINE_NUMBERS"] = 3] = "GUTTER_LINE_NUMBERS";
    MouseTargetType2[MouseTargetType2["GUTTER_LINE_DECORATIONS"] = 4] = "GUTTER_LINE_DECORATIONS";
    MouseTargetType2[MouseTargetType2["GUTTER_VIEW_ZONE"] = 5] = "GUTTER_VIEW_ZONE";
    MouseTargetType2[MouseTargetType2["CONTENT_TEXT"] = 6] = "CONTENT_TEXT";
    MouseTargetType2[MouseTargetType2["CONTENT_EMPTY"] = 7] = "CONTENT_EMPTY";
    MouseTargetType2[MouseTargetType2["CONTENT_VIEW_ZONE"] = 8] = "CONTENT_VIEW_ZONE";
    MouseTargetType2[MouseTargetType2["CONTENT_WIDGET"] = 9] = "CONTENT_WIDGET";
    MouseTargetType2[MouseTargetType2["OVERVIEW_RULER"] = 10] = "OVERVIEW_RULER";
    MouseTargetType2[MouseTargetType2["SCROLLBAR"] = 11] = "SCROLLBAR";
    MouseTargetType2[MouseTargetType2["OVERLAY_WIDGET"] = 12] = "OVERLAY_WIDGET";
    MouseTargetType2[MouseTargetType2["OUTSIDE_EDITOR"] = 13] = "OUTSIDE_EDITOR";
  })(MouseTargetType || (MouseTargetType = {}));
  var OverlayWidgetPositionPreference;
  (function(OverlayWidgetPositionPreference2) {
    OverlayWidgetPositionPreference2[OverlayWidgetPositionPreference2["TOP_RIGHT_CORNER"] = 0] = "TOP_RIGHT_CORNER";
    OverlayWidgetPositionPreference2[OverlayWidgetPositionPreference2["BOTTOM_RIGHT_CORNER"] = 1] = "BOTTOM_RIGHT_CORNER";
    OverlayWidgetPositionPreference2[OverlayWidgetPositionPreference2["TOP_CENTER"] = 2] = "TOP_CENTER";
  })(OverlayWidgetPositionPreference || (OverlayWidgetPositionPreference = {}));
  var OverviewRulerLane;
  (function(OverviewRulerLane2) {
    OverviewRulerLane2[OverviewRulerLane2["Left"] = 1] = "Left";
    OverviewRulerLane2[OverviewRulerLane2["Center"] = 2] = "Center";
    OverviewRulerLane2[OverviewRulerLane2["Right"] = 4] = "Right";
    OverviewRulerLane2[OverviewRulerLane2["Full"] = 7] = "Full";
  })(OverviewRulerLane || (OverviewRulerLane = {}));
  var RenderLineNumbersType;
  (function(RenderLineNumbersType2) {
    RenderLineNumbersType2[RenderLineNumbersType2["Off"] = 0] = "Off";
    RenderLineNumbersType2[RenderLineNumbersType2["On"] = 1] = "On";
    RenderLineNumbersType2[RenderLineNumbersType2["Relative"] = 2] = "Relative";
    RenderLineNumbersType2[RenderLineNumbersType2["Interval"] = 3] = "Interval";
    RenderLineNumbersType2[RenderLineNumbersType2["Custom"] = 4] = "Custom";
  })(RenderLineNumbersType || (RenderLineNumbersType = {}));
  var RenderMinimap;
  (function(RenderMinimap2) {
    RenderMinimap2[RenderMinimap2["None"] = 0] = "None";
    RenderMinimap2[RenderMinimap2["Text"] = 1] = "Text";
    RenderMinimap2[RenderMinimap2["Blocks"] = 2] = "Blocks";
  })(RenderMinimap || (RenderMinimap = {}));
  var ScrollType;
  (function(ScrollType2) {
    ScrollType2[ScrollType2["Smooth"] = 0] = "Smooth";
    ScrollType2[ScrollType2["Immediate"] = 1] = "Immediate";
  })(ScrollType || (ScrollType = {}));
  var ScrollbarVisibility;
  (function(ScrollbarVisibility2) {
    ScrollbarVisibility2[ScrollbarVisibility2["Auto"] = 1] = "Auto";
    ScrollbarVisibility2[ScrollbarVisibility2["Hidden"] = 2] = "Hidden";
    ScrollbarVisibility2[ScrollbarVisibility2["Visible"] = 3] = "Visible";
  })(ScrollbarVisibility || (ScrollbarVisibility = {}));
  var SelectionDirection;
  (function(SelectionDirection2) {
    SelectionDirection2[SelectionDirection2["LTR"] = 0] = "LTR";
    SelectionDirection2[SelectionDirection2["RTL"] = 1] = "RTL";
  })(SelectionDirection || (SelectionDirection = {}));
  var SignatureHelpTriggerKind;
  (function(SignatureHelpTriggerKind2) {
    SignatureHelpTriggerKind2[SignatureHelpTriggerKind2["Invoke"] = 1] = "Invoke";
    SignatureHelpTriggerKind2[SignatureHelpTriggerKind2["TriggerCharacter"] = 2] = "TriggerCharacter";
    SignatureHelpTriggerKind2[SignatureHelpTriggerKind2["ContentChange"] = 3] = "ContentChange";
  })(SignatureHelpTriggerKind || (SignatureHelpTriggerKind = {}));
  var SymbolKind;
  (function(SymbolKind3) {
    SymbolKind3[SymbolKind3["File"] = 0] = "File";
    SymbolKind3[SymbolKind3["Module"] = 1] = "Module";
    SymbolKind3[SymbolKind3["Namespace"] = 2] = "Namespace";
    SymbolKind3[SymbolKind3["Package"] = 3] = "Package";
    SymbolKind3[SymbolKind3["Class"] = 4] = "Class";
    SymbolKind3[SymbolKind3["Method"] = 5] = "Method";
    SymbolKind3[SymbolKind3["Property"] = 6] = "Property";
    SymbolKind3[SymbolKind3["Field"] = 7] = "Field";
    SymbolKind3[SymbolKind3["Constructor"] = 8] = "Constructor";
    SymbolKind3[SymbolKind3["Enum"] = 9] = "Enum";
    SymbolKind3[SymbolKind3["Interface"] = 10] = "Interface";
    SymbolKind3[SymbolKind3["Function"] = 11] = "Function";
    SymbolKind3[SymbolKind3["Variable"] = 12] = "Variable";
    SymbolKind3[SymbolKind3["Constant"] = 13] = "Constant";
    SymbolKind3[SymbolKind3["String"] = 14] = "String";
    SymbolKind3[SymbolKind3["Number"] = 15] = "Number";
    SymbolKind3[SymbolKind3["Boolean"] = 16] = "Boolean";
    SymbolKind3[SymbolKind3["Array"] = 17] = "Array";
    SymbolKind3[SymbolKind3["Object"] = 18] = "Object";
    SymbolKind3[SymbolKind3["Key"] = 19] = "Key";
    SymbolKind3[SymbolKind3["Null"] = 20] = "Null";
    SymbolKind3[SymbolKind3["EnumMember"] = 21] = "EnumMember";
    SymbolKind3[SymbolKind3["Struct"] = 22] = "Struct";
    SymbolKind3[SymbolKind3["Event"] = 23] = "Event";
    SymbolKind3[SymbolKind3["Operator"] = 24] = "Operator";
    SymbolKind3[SymbolKind3["TypeParameter"] = 25] = "TypeParameter";
  })(SymbolKind || (SymbolKind = {}));
  var SymbolTag;
  (function(SymbolTag3) {
    SymbolTag3[SymbolTag3["Deprecated"] = 1] = "Deprecated";
  })(SymbolTag || (SymbolTag = {}));
  var TextEditorCursorBlinkingStyle;
  (function(TextEditorCursorBlinkingStyle2) {
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Hidden"] = 0] = "Hidden";
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Blink"] = 1] = "Blink";
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Smooth"] = 2] = "Smooth";
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Phase"] = 3] = "Phase";
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Expand"] = 4] = "Expand";
    TextEditorCursorBlinkingStyle2[TextEditorCursorBlinkingStyle2["Solid"] = 5] = "Solid";
  })(TextEditorCursorBlinkingStyle || (TextEditorCursorBlinkingStyle = {}));
  var TextEditorCursorStyle;
  (function(TextEditorCursorStyle2) {
    TextEditorCursorStyle2[TextEditorCursorStyle2["Line"] = 1] = "Line";
    TextEditorCursorStyle2[TextEditorCursorStyle2["Block"] = 2] = "Block";
    TextEditorCursorStyle2[TextEditorCursorStyle2["Underline"] = 3] = "Underline";
    TextEditorCursorStyle2[TextEditorCursorStyle2["LineThin"] = 4] = "LineThin";
    TextEditorCursorStyle2[TextEditorCursorStyle2["BlockOutline"] = 5] = "BlockOutline";
    TextEditorCursorStyle2[TextEditorCursorStyle2["UnderlineThin"] = 6] = "UnderlineThin";
  })(TextEditorCursorStyle || (TextEditorCursorStyle = {}));
  var TrackedRangeStickiness;
  (function(TrackedRangeStickiness2) {
    TrackedRangeStickiness2[TrackedRangeStickiness2["AlwaysGrowsWhenTypingAtEdges"] = 0] = "AlwaysGrowsWhenTypingAtEdges";
    TrackedRangeStickiness2[TrackedRangeStickiness2["NeverGrowsWhenTypingAtEdges"] = 1] = "NeverGrowsWhenTypingAtEdges";
    TrackedRangeStickiness2[TrackedRangeStickiness2["GrowsOnlyWhenTypingBefore"] = 2] = "GrowsOnlyWhenTypingBefore";
    TrackedRangeStickiness2[TrackedRangeStickiness2["GrowsOnlyWhenTypingAfter"] = 3] = "GrowsOnlyWhenTypingAfter";
  })(TrackedRangeStickiness || (TrackedRangeStickiness = {}));
  var WrappingIndent;
  (function(WrappingIndent2) {
    WrappingIndent2[WrappingIndent2["None"] = 0] = "None";
    WrappingIndent2[WrappingIndent2["Same"] = 1] = "Same";
    WrappingIndent2[WrappingIndent2["Indent"] = 2] = "Indent";
    WrappingIndent2[WrappingIndent2["DeepIndent"] = 3] = "DeepIndent";
  })(WrappingIndent || (WrappingIndent = {}));

  // node_modules/monaco-editor/esm/vs/editor/common/standalone/standaloneBase.js
  var KeyMod = class {
    static chord(firstPart, secondPart) {
      return KeyChord(firstPart, secondPart);
    }
  };
  KeyMod.CtrlCmd = 2048;
  KeyMod.Shift = 1024;
  KeyMod.Alt = 512;
  KeyMod.WinCtrl = 256;
  function createMonacoBaseAPI() {
    return {
      editor: void 0,
      languages: void 0,
      CancellationTokenSource,
      Emitter,
      KeyCode,
      KeyMod,
      Position,
      Range,
      Selection,
      SelectionDirection,
      MarkerSeverity,
      MarkerTag,
      Uri: URI,
      Token
    };
  }

  // node_modules/monaco-editor/esm/vs/editor/common/services/editorSimpleWorker.js
  var __awaiter = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var MirrorModel = class extends MirrorTextModel {
    get uri() {
      return this._uri;
    }
    get version() {
      return this._versionId;
    }
    get eol() {
      return this._eol;
    }
    getValue() {
      return this.getText();
    }
    getLinesContent() {
      return this._lines.slice(0);
    }
    getLineCount() {
      return this._lines.length;
    }
    getLineContent(lineNumber) {
      return this._lines[lineNumber - 1];
    }
    getWordAtPosition(position, wordDefinition) {
      let wordAtText = getWordAtText(position.column, ensureValidWordDefinition(wordDefinition), this._lines[position.lineNumber - 1], 0);
      if (wordAtText) {
        return new Range(position.lineNumber, wordAtText.startColumn, position.lineNumber, wordAtText.endColumn);
      }
      return null;
    }
    words(wordDefinition) {
      const lines = this._lines;
      const wordenize = this._wordenize.bind(this);
      let lineNumber = 0;
      let lineText = "";
      let wordRangesIdx = 0;
      let wordRanges = [];
      return {
        *[Symbol.iterator]() {
          while (true) {
            if (wordRangesIdx < wordRanges.length) {
              const value = lineText.substring(wordRanges[wordRangesIdx].start, wordRanges[wordRangesIdx].end);
              wordRangesIdx += 1;
              yield value;
            } else {
              if (lineNumber < lines.length) {
                lineText = lines[lineNumber];
                wordRanges = wordenize(lineText, wordDefinition);
                wordRangesIdx = 0;
                lineNumber += 1;
              } else {
                break;
              }
            }
          }
        }
      };
    }
    getLineWords(lineNumber, wordDefinition) {
      let content = this._lines[lineNumber - 1];
      let ranges = this._wordenize(content, wordDefinition);
      let words = [];
      for (const range of ranges) {
        words.push({
          word: content.substring(range.start, range.end),
          startColumn: range.start + 1,
          endColumn: range.end + 1
        });
      }
      return words;
    }
    _wordenize(content, wordDefinition) {
      const result = [];
      let match;
      wordDefinition.lastIndex = 0;
      while (match = wordDefinition.exec(content)) {
        if (match[0].length === 0) {
          break;
        }
        result.push({ start: match.index, end: match.index + match[0].length });
      }
      return result;
    }
    getValueInRange(range) {
      range = this._validateRange(range);
      if (range.startLineNumber === range.endLineNumber) {
        return this._lines[range.startLineNumber - 1].substring(range.startColumn - 1, range.endColumn - 1);
      }
      let lineEnding = this._eol;
      let startLineIndex = range.startLineNumber - 1;
      let endLineIndex = range.endLineNumber - 1;
      let resultLines = [];
      resultLines.push(this._lines[startLineIndex].substring(range.startColumn - 1));
      for (let i = startLineIndex + 1; i < endLineIndex; i++) {
        resultLines.push(this._lines[i]);
      }
      resultLines.push(this._lines[endLineIndex].substring(0, range.endColumn - 1));
      return resultLines.join(lineEnding);
    }
    offsetAt(position) {
      position = this._validatePosition(position);
      this._ensureLineStarts();
      return this._lineStarts.getAccumulatedValue(position.lineNumber - 2) + (position.column - 1);
    }
    positionAt(offset) {
      offset = Math.floor(offset);
      offset = Math.max(0, offset);
      this._ensureLineStarts();
      let out = this._lineStarts.getIndexOf(offset);
      let lineLength = this._lines[out.index].length;
      return {
        lineNumber: 1 + out.index,
        column: 1 + Math.min(out.remainder, lineLength)
      };
    }
    _validateRange(range) {
      const start = this._validatePosition({ lineNumber: range.startLineNumber, column: range.startColumn });
      const end = this._validatePosition({ lineNumber: range.endLineNumber, column: range.endColumn });
      if (start.lineNumber !== range.startLineNumber || start.column !== range.startColumn || end.lineNumber !== range.endLineNumber || end.column !== range.endColumn) {
        return {
          startLineNumber: start.lineNumber,
          startColumn: start.column,
          endLineNumber: end.lineNumber,
          endColumn: end.column
        };
      }
      return range;
    }
    _validatePosition(position) {
      if (!Position.isIPosition(position)) {
        throw new Error("bad position");
      }
      let { lineNumber, column } = position;
      let hasChanged = false;
      if (lineNumber < 1) {
        lineNumber = 1;
        column = 1;
        hasChanged = true;
      } else if (lineNumber > this._lines.length) {
        lineNumber = this._lines.length;
        column = this._lines[lineNumber - 1].length + 1;
        hasChanged = true;
      } else {
        let maxCharacter = this._lines[lineNumber - 1].length + 1;
        if (column < 1) {
          column = 1;
          hasChanged = true;
        } else if (column > maxCharacter) {
          column = maxCharacter;
          hasChanged = true;
        }
      }
      if (!hasChanged) {
        return position;
      } else {
        return { lineNumber, column };
      }
    }
  };
  var EditorSimpleWorker = class {
    constructor(host, foreignModuleFactory) {
      this._host = host;
      this._models = /* @__PURE__ */ Object.create(null);
      this._foreignModuleFactory = foreignModuleFactory;
      this._foreignModule = null;
    }
    dispose() {
      this._models = /* @__PURE__ */ Object.create(null);
    }
    _getModel(uri) {
      return this._models[uri];
    }
    _getModels() {
      let all = [];
      Object.keys(this._models).forEach((key) => all.push(this._models[key]));
      return all;
    }
    acceptNewModel(data) {
      this._models[data.url] = new MirrorModel(URI.parse(data.url), data.lines, data.EOL, data.versionId);
    }
    acceptModelChanged(strURL, e) {
      if (!this._models[strURL]) {
        return;
      }
      let model = this._models[strURL];
      model.onEvents(e);
    }
    acceptRemovedModel(strURL) {
      if (!this._models[strURL]) {
        return;
      }
      delete this._models[strURL];
    }
    computeDiff(originalUrl, modifiedUrl, ignoreTrimWhitespace, maxComputationTime) {
      return __awaiter(this, void 0, void 0, function* () {
        const original = this._getModel(originalUrl);
        const modified = this._getModel(modifiedUrl);
        if (!original || !modified) {
          return null;
        }
        const originalLines = original.getLinesContent();
        const modifiedLines = modified.getLinesContent();
        const diffComputer = new DiffComputer(originalLines, modifiedLines, {
          shouldComputeCharChanges: true,
          shouldPostProcessCharChanges: true,
          shouldIgnoreTrimWhitespace: ignoreTrimWhitespace,
          shouldMakePrettyDiff: true,
          maxComputationTime
        });
        const diffResult = diffComputer.computeDiff();
        const identical = diffResult.changes.length > 0 ? false : this._modelsAreIdentical(original, modified);
        return {
          quitEarly: diffResult.quitEarly,
          identical,
          changes: diffResult.changes
        };
      });
    }
    _modelsAreIdentical(original, modified) {
      const originalLineCount = original.getLineCount();
      const modifiedLineCount = modified.getLineCount();
      if (originalLineCount !== modifiedLineCount) {
        return false;
      }
      for (let line = 1; line <= originalLineCount; line++) {
        const originalLine = original.getLineContent(line);
        const modifiedLine = modified.getLineContent(line);
        if (originalLine !== modifiedLine) {
          return false;
        }
      }
      return true;
    }
    computeMoreMinimalEdits(modelUrl, edits) {
      return __awaiter(this, void 0, void 0, function* () {
        const model = this._getModel(modelUrl);
        if (!model) {
          return edits;
        }
        const result = [];
        let lastEol = void 0;
        edits = mergeSort(edits, (a2, b) => {
          if (a2.range && b.range) {
            return Range.compareRangesUsingStarts(a2.range, b.range);
          }
          let aRng = a2.range ? 0 : 1;
          let bRng = b.range ? 0 : 1;
          return aRng - bRng;
        });
        for (let { range, text, eol } of edits) {
          if (typeof eol === "number") {
            lastEol = eol;
          }
          if (Range.isEmpty(range) && !text) {
            continue;
          }
          const original = model.getValueInRange(range);
          text = text.replace(/\r\n|\n|\r/g, model.eol);
          if (original === text) {
            continue;
          }
          if (Math.max(text.length, original.length) > EditorSimpleWorker._diffLimit) {
            result.push({ range, text });
            continue;
          }
          const changes = stringDiff(original, text, false);
          const editOffset = model.offsetAt(Range.lift(range).getStartPosition());
          for (const change of changes) {
            const start = model.positionAt(editOffset + change.originalStart);
            const end = model.positionAt(editOffset + change.originalStart + change.originalLength);
            const newEdit = {
              text: text.substr(change.modifiedStart, change.modifiedLength),
              range: { startLineNumber: start.lineNumber, startColumn: start.column, endLineNumber: end.lineNumber, endColumn: end.column }
            };
            if (model.getValueInRange(newEdit.range) !== newEdit.text) {
              result.push(newEdit);
            }
          }
        }
        if (typeof lastEol === "number") {
          result.push({ eol: lastEol, text: "", range: { startLineNumber: 0, startColumn: 0, endLineNumber: 0, endColumn: 0 } });
        }
        return result;
      });
    }
    computeLinks(modelUrl) {
      return __awaiter(this, void 0, void 0, function* () {
        let model = this._getModel(modelUrl);
        if (!model) {
          return null;
        }
        return computeLinks(model);
      });
    }
    textualSuggest(modelUrl, position, wordDef, wordDefFlags) {
      return __awaiter(this, void 0, void 0, function* () {
        const model = this._getModel(modelUrl);
        if (!model) {
          return null;
        }
        const words = [];
        const seen = /* @__PURE__ */ new Set();
        const wordDefRegExp = new RegExp(wordDef, wordDefFlags);
        const wordAt = model.getWordAtPosition(position, wordDefRegExp);
        if (wordAt) {
          seen.add(model.getValueInRange(wordAt));
        }
        for (let word of model.words(wordDefRegExp)) {
          if (seen.has(word)) {
            continue;
          }
          seen.add(word);
          if (!isNaN(Number(word))) {
            continue;
          }
          words.push(word);
          if (seen.size > EditorSimpleWorker._suggestionsLimit) {
            break;
          }
        }
        return words;
      });
    }
    computeWordRanges(modelUrl, range, wordDef, wordDefFlags) {
      return __awaiter(this, void 0, void 0, function* () {
        let model = this._getModel(modelUrl);
        if (!model) {
          return /* @__PURE__ */ Object.create(null);
        }
        const wordDefRegExp = new RegExp(wordDef, wordDefFlags);
        const result = /* @__PURE__ */ Object.create(null);
        for (let line = range.startLineNumber; line < range.endLineNumber; line++) {
          let words = model.getLineWords(line, wordDefRegExp);
          for (const word of words) {
            if (!isNaN(Number(word.word))) {
              continue;
            }
            let array = result[word.word];
            if (!array) {
              array = [];
              result[word.word] = array;
            }
            array.push({
              startLineNumber: line,
              startColumn: word.startColumn,
              endLineNumber: line,
              endColumn: word.endColumn
            });
          }
        }
        return result;
      });
    }
    navigateValueSet(modelUrl, range, up, wordDef, wordDefFlags) {
      return __awaiter(this, void 0, void 0, function* () {
        let model = this._getModel(modelUrl);
        if (!model) {
          return null;
        }
        let wordDefRegExp = new RegExp(wordDef, wordDefFlags);
        if (range.startColumn === range.endColumn) {
          range = {
            startLineNumber: range.startLineNumber,
            startColumn: range.startColumn,
            endLineNumber: range.endLineNumber,
            endColumn: range.endColumn + 1
          };
        }
        let selectionText = model.getValueInRange(range);
        let wordRange = model.getWordAtPosition({ lineNumber: range.startLineNumber, column: range.startColumn }, wordDefRegExp);
        if (!wordRange) {
          return null;
        }
        let word = model.getValueInRange(wordRange);
        let result = BasicInplaceReplace.INSTANCE.navigateValueSet(range, selectionText, wordRange, word, up);
        return result;
      });
    }
    loadForeignModule(moduleId, createData, foreignHostMethods) {
      const proxyMethodRequest = (method, args) => {
        return this._host.fhr(method, args);
      };
      const foreignHost = createProxyObject(foreignHostMethods, proxyMethodRequest);
      let ctx = {
        host: foreignHost,
        getMirrorModels: () => {
          return this._getModels();
        }
      };
      if (this._foreignModuleFactory) {
        this._foreignModule = this._foreignModuleFactory(ctx, createData);
        return Promise.resolve(getAllMethodNames(this._foreignModule));
      }
      return Promise.reject(new Error(`Unexpected usage`));
    }
    fmr(method, args) {
      if (!this._foreignModule || typeof this._foreignModule[method] !== "function") {
        return Promise.reject(new Error("Missing requestHandler or method: " + method));
      }
      try {
        return Promise.resolve(this._foreignModule[method].apply(this._foreignModule, args));
      } catch (e) {
        return Promise.reject(e);
      }
    }
  };
  EditorSimpleWorker._diffLimit = 1e5;
  EditorSimpleWorker._suggestionsLimit = 1e4;
  if (typeof importScripts === "function") {
    globals.monaco = createMonacoBaseAPI();
  }

  // node_modules/monaco-editor/esm/vs/editor/editor.worker.js
  var initialized = false;
  function initialize(foreignModule) {
    if (initialized) {
      return;
    }
    initialized = true;
    const simpleWorker = new SimpleWorkerServer((msg) => {
      self.postMessage(msg);
    }, (host) => new EditorSimpleWorker(host, foreignModule));
    self.onmessage = (e) => {
      simpleWorker.onmessage(e.data);
    };
  }
  self.onmessage = (e) => {
    if (!initialized) {
      initialize(null);
    }
  };

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/cssScanner.js
  var TokenType;
  (function(TokenType2) {
    TokenType2[TokenType2["Ident"] = 0] = "Ident";
    TokenType2[TokenType2["AtKeyword"] = 1] = "AtKeyword";
    TokenType2[TokenType2["String"] = 2] = "String";
    TokenType2[TokenType2["BadString"] = 3] = "BadString";
    TokenType2[TokenType2["UnquotedString"] = 4] = "UnquotedString";
    TokenType2[TokenType2["Hash"] = 5] = "Hash";
    TokenType2[TokenType2["Num"] = 6] = "Num";
    TokenType2[TokenType2["Percentage"] = 7] = "Percentage";
    TokenType2[TokenType2["Dimension"] = 8] = "Dimension";
    TokenType2[TokenType2["UnicodeRange"] = 9] = "UnicodeRange";
    TokenType2[TokenType2["CDO"] = 10] = "CDO";
    TokenType2[TokenType2["CDC"] = 11] = "CDC";
    TokenType2[TokenType2["Colon"] = 12] = "Colon";
    TokenType2[TokenType2["SemiColon"] = 13] = "SemiColon";
    TokenType2[TokenType2["CurlyL"] = 14] = "CurlyL";
    TokenType2[TokenType2["CurlyR"] = 15] = "CurlyR";
    TokenType2[TokenType2["ParenthesisL"] = 16] = "ParenthesisL";
    TokenType2[TokenType2["ParenthesisR"] = 17] = "ParenthesisR";
    TokenType2[TokenType2["BracketL"] = 18] = "BracketL";
    TokenType2[TokenType2["BracketR"] = 19] = "BracketR";
    TokenType2[TokenType2["Whitespace"] = 20] = "Whitespace";
    TokenType2[TokenType2["Includes"] = 21] = "Includes";
    TokenType2[TokenType2["Dashmatch"] = 22] = "Dashmatch";
    TokenType2[TokenType2["SubstringOperator"] = 23] = "SubstringOperator";
    TokenType2[TokenType2["PrefixOperator"] = 24] = "PrefixOperator";
    TokenType2[TokenType2["SuffixOperator"] = 25] = "SuffixOperator";
    TokenType2[TokenType2["Delim"] = 26] = "Delim";
    TokenType2[TokenType2["EMS"] = 27] = "EMS";
    TokenType2[TokenType2["EXS"] = 28] = "EXS";
    TokenType2[TokenType2["Length"] = 29] = "Length";
    TokenType2[TokenType2["Angle"] = 30] = "Angle";
    TokenType2[TokenType2["Time"] = 31] = "Time";
    TokenType2[TokenType2["Freq"] = 32] = "Freq";
    TokenType2[TokenType2["Exclamation"] = 33] = "Exclamation";
    TokenType2[TokenType2["Resolution"] = 34] = "Resolution";
    TokenType2[TokenType2["Comma"] = 35] = "Comma";
    TokenType2[TokenType2["Charset"] = 36] = "Charset";
    TokenType2[TokenType2["EscapedJavaScript"] = 37] = "EscapedJavaScript";
    TokenType2[TokenType2["BadEscapedJavaScript"] = 38] = "BadEscapedJavaScript";
    TokenType2[TokenType2["Comment"] = 39] = "Comment";
    TokenType2[TokenType2["SingleLineComment"] = 40] = "SingleLineComment";
    TokenType2[TokenType2["EOF"] = 41] = "EOF";
    TokenType2[TokenType2["CustomToken"] = 42] = "CustomToken";
  })(TokenType || (TokenType = {}));
  var MultiLineStream = function() {
    function MultiLineStream2(source) {
      this.source = source;
      this.len = source.length;
      this.position = 0;
    }
    MultiLineStream2.prototype.substring = function(from, to) {
      if (to === void 0) {
        to = this.position;
      }
      return this.source.substring(from, to);
    };
    MultiLineStream2.prototype.eos = function() {
      return this.len <= this.position;
    };
    MultiLineStream2.prototype.pos = function() {
      return this.position;
    };
    MultiLineStream2.prototype.goBackTo = function(pos) {
      this.position = pos;
    };
    MultiLineStream2.prototype.goBack = function(n) {
      this.position -= n;
    };
    MultiLineStream2.prototype.advance = function(n) {
      this.position += n;
    };
    MultiLineStream2.prototype.nextChar = function() {
      return this.source.charCodeAt(this.position++) || 0;
    };
    MultiLineStream2.prototype.peekChar = function(n) {
      if (n === void 0) {
        n = 0;
      }
      return this.source.charCodeAt(this.position + n) || 0;
    };
    MultiLineStream2.prototype.lookbackChar = function(n) {
      if (n === void 0) {
        n = 0;
      }
      return this.source.charCodeAt(this.position - n) || 0;
    };
    MultiLineStream2.prototype.advanceIfChar = function(ch) {
      if (ch === this.source.charCodeAt(this.position)) {
        this.position++;
        return true;
      }
      return false;
    };
    MultiLineStream2.prototype.advanceIfChars = function(ch) {
      if (this.position + ch.length > this.source.length) {
        return false;
      }
      var i = 0;
      for (; i < ch.length; i++) {
        if (this.source.charCodeAt(this.position + i) !== ch[i]) {
          return false;
        }
      }
      this.advance(i);
      return true;
    };
    MultiLineStream2.prototype.advanceWhileChar = function(condition) {
      var posNow = this.position;
      while (this.position < this.len && condition(this.source.charCodeAt(this.position))) {
        this.position++;
      }
      return this.position - posNow;
    };
    return MultiLineStream2;
  }();
  var _a = "a".charCodeAt(0);
  var _f = "f".charCodeAt(0);
  var _z = "z".charCodeAt(0);
  var _A = "A".charCodeAt(0);
  var _F = "F".charCodeAt(0);
  var _Z = "Z".charCodeAt(0);
  var _0 = "0".charCodeAt(0);
  var _9 = "9".charCodeAt(0);
  var _TLD = "~".charCodeAt(0);
  var _HAT = "^".charCodeAt(0);
  var _EQS = "=".charCodeAt(0);
  var _PIP = "|".charCodeAt(0);
  var _MIN = "-".charCodeAt(0);
  var _USC = "_".charCodeAt(0);
  var _PRC = "%".charCodeAt(0);
  var _MUL = "*".charCodeAt(0);
  var _LPA = "(".charCodeAt(0);
  var _RPA = ")".charCodeAt(0);
  var _LAN = "<".charCodeAt(0);
  var _RAN = ">".charCodeAt(0);
  var _ATS = "@".charCodeAt(0);
  var _HSH = "#".charCodeAt(0);
  var _DLR = "$".charCodeAt(0);
  var _BSL = "\\".charCodeAt(0);
  var _FSL = "/".charCodeAt(0);
  var _NWL = "\n".charCodeAt(0);
  var _CAR = "\r".charCodeAt(0);
  var _LFD = "\f".charCodeAt(0);
  var _DQO = '"'.charCodeAt(0);
  var _SQO = "'".charCodeAt(0);
  var _WSP = " ".charCodeAt(0);
  var _TAB = "	".charCodeAt(0);
  var _SEM = ";".charCodeAt(0);
  var _COL = ":".charCodeAt(0);
  var _CUL = "{".charCodeAt(0);
  var _CUR = "}".charCodeAt(0);
  var _BRL = "[".charCodeAt(0);
  var _BRR = "]".charCodeAt(0);
  var _CMA = ",".charCodeAt(0);
  var _DOT = ".".charCodeAt(0);
  var _BNG = "!".charCodeAt(0);
  var staticTokenTable = {};
  staticTokenTable[_SEM] = TokenType.SemiColon;
  staticTokenTable[_COL] = TokenType.Colon;
  staticTokenTable[_CUL] = TokenType.CurlyL;
  staticTokenTable[_CUR] = TokenType.CurlyR;
  staticTokenTable[_BRR] = TokenType.BracketR;
  staticTokenTable[_BRL] = TokenType.BracketL;
  staticTokenTable[_LPA] = TokenType.ParenthesisL;
  staticTokenTable[_RPA] = TokenType.ParenthesisR;
  staticTokenTable[_CMA] = TokenType.Comma;
  var staticUnitTable = {};
  staticUnitTable["em"] = TokenType.EMS;
  staticUnitTable["ex"] = TokenType.EXS;
  staticUnitTable["px"] = TokenType.Length;
  staticUnitTable["cm"] = TokenType.Length;
  staticUnitTable["mm"] = TokenType.Length;
  staticUnitTable["in"] = TokenType.Length;
  staticUnitTable["pt"] = TokenType.Length;
  staticUnitTable["pc"] = TokenType.Length;
  staticUnitTable["deg"] = TokenType.Angle;
  staticUnitTable["rad"] = TokenType.Angle;
  staticUnitTable["grad"] = TokenType.Angle;
  staticUnitTable["ms"] = TokenType.Time;
  staticUnitTable["s"] = TokenType.Time;
  staticUnitTable["hz"] = TokenType.Freq;
  staticUnitTable["khz"] = TokenType.Freq;
  staticUnitTable["%"] = TokenType.Percentage;
  staticUnitTable["fr"] = TokenType.Percentage;
  staticUnitTable["dpi"] = TokenType.Resolution;
  staticUnitTable["dpcm"] = TokenType.Resolution;
  var Scanner = function() {
    function Scanner2() {
      this.stream = new MultiLineStream("");
      this.ignoreComment = true;
      this.ignoreWhitespace = true;
      this.inURL = false;
    }
    Scanner2.prototype.setSource = function(input) {
      this.stream = new MultiLineStream(input);
    };
    Scanner2.prototype.finishToken = function(offset, type, text) {
      return {
        offset,
        len: this.stream.pos() - offset,
        type,
        text: text || this.stream.substring(offset)
      };
    };
    Scanner2.prototype.substring = function(offset, len) {
      return this.stream.substring(offset, offset + len);
    };
    Scanner2.prototype.pos = function() {
      return this.stream.pos();
    };
    Scanner2.prototype.goBackTo = function(pos) {
      this.stream.goBackTo(pos);
    };
    Scanner2.prototype.scanUnquotedString = function() {
      var offset = this.stream.pos();
      var content = [];
      if (this._unquotedString(content)) {
        return this.finishToken(offset, TokenType.UnquotedString, content.join(""));
      }
      return null;
    };
    Scanner2.prototype.scan = function() {
      var triviaToken = this.trivia();
      if (triviaToken !== null) {
        return triviaToken;
      }
      var offset = this.stream.pos();
      if (this.stream.eos()) {
        return this.finishToken(offset, TokenType.EOF);
      }
      return this.scanNext(offset);
    };
    Scanner2.prototype.scanNext = function(offset) {
      if (this.stream.advanceIfChars([_LAN, _BNG, _MIN, _MIN])) {
        return this.finishToken(offset, TokenType.CDO);
      }
      if (this.stream.advanceIfChars([_MIN, _MIN, _RAN])) {
        return this.finishToken(offset, TokenType.CDC);
      }
      var content = [];
      if (this.ident(content)) {
        return this.finishToken(offset, TokenType.Ident, content.join(""));
      }
      if (this.stream.advanceIfChar(_ATS)) {
        content = ["@"];
        if (this._name(content)) {
          var keywordText = content.join("");
          if (keywordText === "@charset") {
            return this.finishToken(offset, TokenType.Charset, keywordText);
          }
          return this.finishToken(offset, TokenType.AtKeyword, keywordText);
        } else {
          return this.finishToken(offset, TokenType.Delim);
        }
      }
      if (this.stream.advanceIfChar(_HSH)) {
        content = ["#"];
        if (this._name(content)) {
          return this.finishToken(offset, TokenType.Hash, content.join(""));
        } else {
          return this.finishToken(offset, TokenType.Delim);
        }
      }
      if (this.stream.advanceIfChar(_BNG)) {
        return this.finishToken(offset, TokenType.Exclamation);
      }
      if (this._number()) {
        var pos = this.stream.pos();
        content = [this.stream.substring(offset, pos)];
        if (this.stream.advanceIfChar(_PRC)) {
          return this.finishToken(offset, TokenType.Percentage);
        } else if (this.ident(content)) {
          var dim = this.stream.substring(pos).toLowerCase();
          var tokenType_1 = staticUnitTable[dim];
          if (typeof tokenType_1 !== "undefined") {
            return this.finishToken(offset, tokenType_1, content.join(""));
          } else {
            return this.finishToken(offset, TokenType.Dimension, content.join(""));
          }
        }
        return this.finishToken(offset, TokenType.Num);
      }
      content = [];
      var tokenType = this._string(content);
      if (tokenType !== null) {
        return this.finishToken(offset, tokenType, content.join(""));
      }
      tokenType = staticTokenTable[this.stream.peekChar()];
      if (typeof tokenType !== "undefined") {
        this.stream.advance(1);
        return this.finishToken(offset, tokenType);
      }
      if (this.stream.peekChar(0) === _TLD && this.stream.peekChar(1) === _EQS) {
        this.stream.advance(2);
        return this.finishToken(offset, TokenType.Includes);
      }
      if (this.stream.peekChar(0) === _PIP && this.stream.peekChar(1) === _EQS) {
        this.stream.advance(2);
        return this.finishToken(offset, TokenType.Dashmatch);
      }
      if (this.stream.peekChar(0) === _MUL && this.stream.peekChar(1) === _EQS) {
        this.stream.advance(2);
        return this.finishToken(offset, TokenType.SubstringOperator);
      }
      if (this.stream.peekChar(0) === _HAT && this.stream.peekChar(1) === _EQS) {
        this.stream.advance(2);
        return this.finishToken(offset, TokenType.PrefixOperator);
      }
      if (this.stream.peekChar(0) === _DLR && this.stream.peekChar(1) === _EQS) {
        this.stream.advance(2);
        return this.finishToken(offset, TokenType.SuffixOperator);
      }
      this.stream.nextChar();
      return this.finishToken(offset, TokenType.Delim);
    };
    Scanner2.prototype.trivia = function() {
      while (true) {
        var offset = this.stream.pos();
        if (this._whitespace()) {
          if (!this.ignoreWhitespace) {
            return this.finishToken(offset, TokenType.Whitespace);
          }
        } else if (this.comment()) {
          if (!this.ignoreComment) {
            return this.finishToken(offset, TokenType.Comment);
          }
        } else {
          return null;
        }
      }
    };
    Scanner2.prototype.comment = function() {
      if (this.stream.advanceIfChars([_FSL, _MUL])) {
        var success_1 = false, hot_1 = false;
        this.stream.advanceWhileChar(function(ch) {
          if (hot_1 && ch === _FSL) {
            success_1 = true;
            return false;
          }
          hot_1 = ch === _MUL;
          return true;
        });
        if (success_1) {
          this.stream.advance(1);
        }
        return true;
      }
      return false;
    };
    Scanner2.prototype._number = function() {
      var npeek = 0, ch;
      if (this.stream.peekChar() === _DOT) {
        npeek = 1;
      }
      ch = this.stream.peekChar(npeek);
      if (ch >= _0 && ch <= _9) {
        this.stream.advance(npeek + 1);
        this.stream.advanceWhileChar(function(ch2) {
          return ch2 >= _0 && ch2 <= _9 || npeek === 0 && ch2 === _DOT;
        });
        return true;
      }
      return false;
    };
    Scanner2.prototype._newline = function(result) {
      var ch = this.stream.peekChar();
      switch (ch) {
        case _CAR:
        case _LFD:
        case _NWL:
          this.stream.advance(1);
          result.push(String.fromCharCode(ch));
          if (ch === _CAR && this.stream.advanceIfChar(_NWL)) {
            result.push("\n");
          }
          return true;
      }
      return false;
    };
    Scanner2.prototype._escape = function(result, includeNewLines) {
      var ch = this.stream.peekChar();
      if (ch === _BSL) {
        this.stream.advance(1);
        ch = this.stream.peekChar();
        var hexNumCount = 0;
        while (hexNumCount < 6 && (ch >= _0 && ch <= _9 || ch >= _a && ch <= _f || ch >= _A && ch <= _F)) {
          this.stream.advance(1);
          ch = this.stream.peekChar();
          hexNumCount++;
        }
        if (hexNumCount > 0) {
          try {
            var hexVal = parseInt(this.stream.substring(this.stream.pos() - hexNumCount), 16);
            if (hexVal) {
              result.push(String.fromCharCode(hexVal));
            }
          } catch (e) {
          }
          if (ch === _WSP || ch === _TAB) {
            this.stream.advance(1);
          } else {
            this._newline([]);
          }
          return true;
        }
        if (ch !== _CAR && ch !== _LFD && ch !== _NWL) {
          this.stream.advance(1);
          result.push(String.fromCharCode(ch));
          return true;
        } else if (includeNewLines) {
          return this._newline(result);
        }
      }
      return false;
    };
    Scanner2.prototype._stringChar = function(closeQuote, result) {
      var ch = this.stream.peekChar();
      if (ch !== 0 && ch !== closeQuote && ch !== _BSL && ch !== _CAR && ch !== _LFD && ch !== _NWL) {
        this.stream.advance(1);
        result.push(String.fromCharCode(ch));
        return true;
      }
      return false;
    };
    Scanner2.prototype._string = function(result) {
      if (this.stream.peekChar() === _SQO || this.stream.peekChar() === _DQO) {
        var closeQuote = this.stream.nextChar();
        result.push(String.fromCharCode(closeQuote));
        while (this._stringChar(closeQuote, result) || this._escape(result, true)) {
        }
        if (this.stream.peekChar() === closeQuote) {
          this.stream.nextChar();
          result.push(String.fromCharCode(closeQuote));
          return TokenType.String;
        } else {
          return TokenType.BadString;
        }
      }
      return null;
    };
    Scanner2.prototype._unquotedChar = function(result) {
      var ch = this.stream.peekChar();
      if (ch !== 0 && ch !== _BSL && ch !== _SQO && ch !== _DQO && ch !== _LPA && ch !== _RPA && ch !== _WSP && ch !== _TAB && ch !== _NWL && ch !== _LFD && ch !== _CAR) {
        this.stream.advance(1);
        result.push(String.fromCharCode(ch));
        return true;
      }
      return false;
    };
    Scanner2.prototype._unquotedString = function(result) {
      var hasContent = false;
      while (this._unquotedChar(result) || this._escape(result)) {
        hasContent = true;
      }
      return hasContent;
    };
    Scanner2.prototype._whitespace = function() {
      var n = this.stream.advanceWhileChar(function(ch) {
        return ch === _WSP || ch === _TAB || ch === _NWL || ch === _LFD || ch === _CAR;
      });
      return n > 0;
    };
    Scanner2.prototype._name = function(result) {
      var matched = false;
      while (this._identChar(result) || this._escape(result)) {
        matched = true;
      }
      return matched;
    };
    Scanner2.prototype.ident = function(result) {
      var pos = this.stream.pos();
      var hasMinus = this._minus(result);
      if (hasMinus && this._minus(result)) {
        if (this._identFirstChar(result) || this._escape(result)) {
          while (this._identChar(result) || this._escape(result)) {
          }
          return true;
        }
      } else if (this._identFirstChar(result) || this._escape(result)) {
        while (this._identChar(result) || this._escape(result)) {
        }
        return true;
      }
      this.stream.goBackTo(pos);
      return false;
    };
    Scanner2.prototype._identFirstChar = function(result) {
      var ch = this.stream.peekChar();
      if (ch === _USC || ch >= _a && ch <= _z || ch >= _A && ch <= _Z || ch >= 128 && ch <= 65535) {
        this.stream.advance(1);
        result.push(String.fromCharCode(ch));
        return true;
      }
      return false;
    };
    Scanner2.prototype._minus = function(result) {
      var ch = this.stream.peekChar();
      if (ch === _MIN) {
        this.stream.advance(1);
        result.push(String.fromCharCode(ch));
        return true;
      }
      return false;
    };
    Scanner2.prototype._identChar = function(result) {
      var ch = this.stream.peekChar();
      if (ch === _USC || ch === _MIN || ch >= _a && ch <= _z || ch >= _A && ch <= _Z || ch >= _0 && ch <= _9 || ch >= 128 && ch <= 65535) {
        this.stream.advance(1);
        result.push(String.fromCharCode(ch));
        return true;
      }
      return false;
    };
    return Scanner2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/utils/strings.js
  function startsWith(haystack, needle) {
    if (haystack.length < needle.length) {
      return false;
    }
    for (var i = 0; i < needle.length; i++) {
      if (haystack[i] !== needle[i]) {
        return false;
      }
    }
    return true;
  }
  function endsWith(haystack, needle) {
    var diff = haystack.length - needle.length;
    if (diff > 0) {
      return haystack.lastIndexOf(needle) === diff;
    } else if (diff === 0) {
      return haystack === needle;
    } else {
      return false;
    }
  }
  function difference(first, second, maxLenDelta) {
    if (maxLenDelta === void 0) {
      maxLenDelta = 4;
    }
    var lengthDifference = Math.abs(first.length - second.length);
    if (lengthDifference > maxLenDelta) {
      return 0;
    }
    var LCS = [];
    var zeroArray = [];
    var i, j;
    for (i = 0; i < second.length + 1; ++i) {
      zeroArray.push(0);
    }
    for (i = 0; i < first.length + 1; ++i) {
      LCS.push(zeroArray);
    }
    for (i = 1; i < first.length + 1; ++i) {
      for (j = 1; j < second.length + 1; ++j) {
        if (first[i - 1] === second[j - 1]) {
          LCS[i][j] = LCS[i - 1][j - 1] + 1;
        } else {
          LCS[i][j] = Math.max(LCS[i - 1][j], LCS[i][j - 1]);
        }
      }
    }
    return LCS[first.length][second.length] - Math.sqrt(lengthDifference);
  }
  function getLimitedString(str, ellipsis) {
    if (ellipsis === void 0) {
      ellipsis = true;
    }
    if (!str) {
      return "";
    }
    if (str.length < 140) {
      return str;
    }
    return str.slice(0, 140) + (ellipsis ? "\u2026" : "");
  }
  function trim(str, regexp) {
    var m = regexp.exec(str);
    if (m && m[0].length) {
      return str.substr(0, str.length - m[0].length);
    }
    return str;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/cssNodes.js
  var __extends = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var NodeType;
  (function(NodeType2) {
    NodeType2[NodeType2["Undefined"] = 0] = "Undefined";
    NodeType2[NodeType2["Identifier"] = 1] = "Identifier";
    NodeType2[NodeType2["Stylesheet"] = 2] = "Stylesheet";
    NodeType2[NodeType2["Ruleset"] = 3] = "Ruleset";
    NodeType2[NodeType2["Selector"] = 4] = "Selector";
    NodeType2[NodeType2["SimpleSelector"] = 5] = "SimpleSelector";
    NodeType2[NodeType2["SelectorInterpolation"] = 6] = "SelectorInterpolation";
    NodeType2[NodeType2["SelectorCombinator"] = 7] = "SelectorCombinator";
    NodeType2[NodeType2["SelectorCombinatorParent"] = 8] = "SelectorCombinatorParent";
    NodeType2[NodeType2["SelectorCombinatorSibling"] = 9] = "SelectorCombinatorSibling";
    NodeType2[NodeType2["SelectorCombinatorAllSiblings"] = 10] = "SelectorCombinatorAllSiblings";
    NodeType2[NodeType2["SelectorCombinatorShadowPiercingDescendant"] = 11] = "SelectorCombinatorShadowPiercingDescendant";
    NodeType2[NodeType2["Page"] = 12] = "Page";
    NodeType2[NodeType2["PageBoxMarginBox"] = 13] = "PageBoxMarginBox";
    NodeType2[NodeType2["ClassSelector"] = 14] = "ClassSelector";
    NodeType2[NodeType2["IdentifierSelector"] = 15] = "IdentifierSelector";
    NodeType2[NodeType2["ElementNameSelector"] = 16] = "ElementNameSelector";
    NodeType2[NodeType2["PseudoSelector"] = 17] = "PseudoSelector";
    NodeType2[NodeType2["AttributeSelector"] = 18] = "AttributeSelector";
    NodeType2[NodeType2["Declaration"] = 19] = "Declaration";
    NodeType2[NodeType2["Declarations"] = 20] = "Declarations";
    NodeType2[NodeType2["Property"] = 21] = "Property";
    NodeType2[NodeType2["Expression"] = 22] = "Expression";
    NodeType2[NodeType2["BinaryExpression"] = 23] = "BinaryExpression";
    NodeType2[NodeType2["Term"] = 24] = "Term";
    NodeType2[NodeType2["Operator"] = 25] = "Operator";
    NodeType2[NodeType2["Value"] = 26] = "Value";
    NodeType2[NodeType2["StringLiteral"] = 27] = "StringLiteral";
    NodeType2[NodeType2["URILiteral"] = 28] = "URILiteral";
    NodeType2[NodeType2["EscapedValue"] = 29] = "EscapedValue";
    NodeType2[NodeType2["Function"] = 30] = "Function";
    NodeType2[NodeType2["NumericValue"] = 31] = "NumericValue";
    NodeType2[NodeType2["HexColorValue"] = 32] = "HexColorValue";
    NodeType2[NodeType2["MixinDeclaration"] = 33] = "MixinDeclaration";
    NodeType2[NodeType2["MixinReference"] = 34] = "MixinReference";
    NodeType2[NodeType2["VariableName"] = 35] = "VariableName";
    NodeType2[NodeType2["VariableDeclaration"] = 36] = "VariableDeclaration";
    NodeType2[NodeType2["Prio"] = 37] = "Prio";
    NodeType2[NodeType2["Interpolation"] = 38] = "Interpolation";
    NodeType2[NodeType2["NestedProperties"] = 39] = "NestedProperties";
    NodeType2[NodeType2["ExtendsReference"] = 40] = "ExtendsReference";
    NodeType2[NodeType2["SelectorPlaceholder"] = 41] = "SelectorPlaceholder";
    NodeType2[NodeType2["Debug"] = 42] = "Debug";
    NodeType2[NodeType2["If"] = 43] = "If";
    NodeType2[NodeType2["Else"] = 44] = "Else";
    NodeType2[NodeType2["For"] = 45] = "For";
    NodeType2[NodeType2["Each"] = 46] = "Each";
    NodeType2[NodeType2["While"] = 47] = "While";
    NodeType2[NodeType2["MixinContentReference"] = 48] = "MixinContentReference";
    NodeType2[NodeType2["MixinContentDeclaration"] = 49] = "MixinContentDeclaration";
    NodeType2[NodeType2["Media"] = 50] = "Media";
    NodeType2[NodeType2["Keyframe"] = 51] = "Keyframe";
    NodeType2[NodeType2["FontFace"] = 52] = "FontFace";
    NodeType2[NodeType2["Import"] = 53] = "Import";
    NodeType2[NodeType2["Namespace"] = 54] = "Namespace";
    NodeType2[NodeType2["Invocation"] = 55] = "Invocation";
    NodeType2[NodeType2["FunctionDeclaration"] = 56] = "FunctionDeclaration";
    NodeType2[NodeType2["ReturnStatement"] = 57] = "ReturnStatement";
    NodeType2[NodeType2["MediaQuery"] = 58] = "MediaQuery";
    NodeType2[NodeType2["FunctionParameter"] = 59] = "FunctionParameter";
    NodeType2[NodeType2["FunctionArgument"] = 60] = "FunctionArgument";
    NodeType2[NodeType2["KeyframeSelector"] = 61] = "KeyframeSelector";
    NodeType2[NodeType2["ViewPort"] = 62] = "ViewPort";
    NodeType2[NodeType2["Document"] = 63] = "Document";
    NodeType2[NodeType2["AtApplyRule"] = 64] = "AtApplyRule";
    NodeType2[NodeType2["CustomPropertyDeclaration"] = 65] = "CustomPropertyDeclaration";
    NodeType2[NodeType2["CustomPropertySet"] = 66] = "CustomPropertySet";
    NodeType2[NodeType2["ListEntry"] = 67] = "ListEntry";
    NodeType2[NodeType2["Supports"] = 68] = "Supports";
    NodeType2[NodeType2["SupportsCondition"] = 69] = "SupportsCondition";
    NodeType2[NodeType2["NamespacePrefix"] = 70] = "NamespacePrefix";
    NodeType2[NodeType2["GridLine"] = 71] = "GridLine";
    NodeType2[NodeType2["Plugin"] = 72] = "Plugin";
    NodeType2[NodeType2["UnknownAtRule"] = 73] = "UnknownAtRule";
    NodeType2[NodeType2["Use"] = 74] = "Use";
    NodeType2[NodeType2["ModuleConfiguration"] = 75] = "ModuleConfiguration";
    NodeType2[NodeType2["Forward"] = 76] = "Forward";
    NodeType2[NodeType2["ForwardVisibility"] = 77] = "ForwardVisibility";
    NodeType2[NodeType2["Module"] = 78] = "Module";
  })(NodeType || (NodeType = {}));
  var ReferenceType;
  (function(ReferenceType2) {
    ReferenceType2[ReferenceType2["Mixin"] = 0] = "Mixin";
    ReferenceType2[ReferenceType2["Rule"] = 1] = "Rule";
    ReferenceType2[ReferenceType2["Variable"] = 2] = "Variable";
    ReferenceType2[ReferenceType2["Function"] = 3] = "Function";
    ReferenceType2[ReferenceType2["Keyframe"] = 4] = "Keyframe";
    ReferenceType2[ReferenceType2["Unknown"] = 5] = "Unknown";
    ReferenceType2[ReferenceType2["Module"] = 6] = "Module";
    ReferenceType2[ReferenceType2["Forward"] = 7] = "Forward";
    ReferenceType2[ReferenceType2["ForwardVisibility"] = 8] = "ForwardVisibility";
  })(ReferenceType || (ReferenceType = {}));
  function getNodeAtOffset(node, offset) {
    var candidate = null;
    if (!node || offset < node.offset || offset > node.end) {
      return null;
    }
    node.accept(function(node2) {
      if (node2.offset === -1 && node2.length === -1) {
        return true;
      }
      if (node2.offset <= offset && node2.end >= offset) {
        if (!candidate) {
          candidate = node2;
        } else if (node2.length <= candidate.length) {
          candidate = node2;
        }
        return true;
      }
      return false;
    });
    return candidate;
  }
  function getNodePath(node, offset) {
    var candidate = getNodeAtOffset(node, offset);
    var path = [];
    while (candidate) {
      path.unshift(candidate);
      candidate = candidate.parent;
    }
    return path;
  }
  function getParentDeclaration(node) {
    var decl = node.findParent(NodeType.Declaration);
    var value = decl && decl.getValue();
    if (value && value.encloses(node)) {
      return decl;
    }
    return null;
  }
  var Node2 = function() {
    function Node3(offset, len, nodeType) {
      if (offset === void 0) {
        offset = -1;
      }
      if (len === void 0) {
        len = -1;
      }
      this.parent = null;
      this.offset = offset;
      this.length = len;
      if (nodeType) {
        this.nodeType = nodeType;
      }
    }
    Object.defineProperty(Node3.prototype, "end", {
      get: function() {
        return this.offset + this.length;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(Node3.prototype, "type", {
      get: function() {
        return this.nodeType || NodeType.Undefined;
      },
      set: function(type) {
        this.nodeType = type;
      },
      enumerable: true,
      configurable: true
    });
    Node3.prototype.getTextProvider = function() {
      var node = this;
      while (node && !node.textProvider) {
        node = node.parent;
      }
      if (node) {
        return node.textProvider;
      }
      return function() {
        return "unknown";
      };
    };
    Node3.prototype.getText = function() {
      return this.getTextProvider()(this.offset, this.length);
    };
    Node3.prototype.matches = function(str) {
      return this.length === str.length && this.getTextProvider()(this.offset, this.length) === str;
    };
    Node3.prototype.startsWith = function(str) {
      return this.length >= str.length && this.getTextProvider()(this.offset, str.length) === str;
    };
    Node3.prototype.endsWith = function(str) {
      return this.length >= str.length && this.getTextProvider()(this.end - str.length, str.length) === str;
    };
    Node3.prototype.accept = function(visitor) {
      if (visitor(this) && this.children) {
        for (var _i = 0, _a3 = this.children; _i < _a3.length; _i++) {
          var child = _a3[_i];
          child.accept(visitor);
        }
      }
    };
    Node3.prototype.acceptVisitor = function(visitor) {
      this.accept(visitor.visitNode.bind(visitor));
    };
    Node3.prototype.adoptChild = function(node, index) {
      if (index === void 0) {
        index = -1;
      }
      if (node.parent && node.parent.children) {
        var idx = node.parent.children.indexOf(node);
        if (idx >= 0) {
          node.parent.children.splice(idx, 1);
        }
      }
      node.parent = this;
      var children = this.children;
      if (!children) {
        children = this.children = [];
      }
      if (index !== -1) {
        children.splice(index, 0, node);
      } else {
        children.push(node);
      }
      return node;
    };
    Node3.prototype.attachTo = function(parent, index) {
      if (index === void 0) {
        index = -1;
      }
      if (parent) {
        parent.adoptChild(this, index);
      }
      return this;
    };
    Node3.prototype.collectIssues = function(results) {
      if (this.issues) {
        results.push.apply(results, this.issues);
      }
    };
    Node3.prototype.addIssue = function(issue) {
      if (!this.issues) {
        this.issues = [];
      }
      this.issues.push(issue);
    };
    Node3.prototype.hasIssue = function(rule) {
      return Array.isArray(this.issues) && this.issues.some(function(i) {
        return i.getRule() === rule;
      });
    };
    Node3.prototype.isErroneous = function(recursive) {
      if (recursive === void 0) {
        recursive = false;
      }
      if (this.issues && this.issues.length > 0) {
        return true;
      }
      return recursive && Array.isArray(this.children) && this.children.some(function(c) {
        return c.isErroneous(true);
      });
    };
    Node3.prototype.setNode = function(field, node, index) {
      if (index === void 0) {
        index = -1;
      }
      if (node) {
        node.attachTo(this, index);
        this[field] = node;
        return true;
      }
      return false;
    };
    Node3.prototype.addChild = function(node) {
      if (node) {
        if (!this.children) {
          this.children = [];
        }
        node.attachTo(this);
        this.updateOffsetAndLength(node);
        return true;
      }
      return false;
    };
    Node3.prototype.updateOffsetAndLength = function(node) {
      if (node.offset < this.offset || this.offset === -1) {
        this.offset = node.offset;
      }
      var nodeEnd = node.end;
      if (nodeEnd > this.end || this.length === -1) {
        this.length = nodeEnd - this.offset;
      }
    };
    Node3.prototype.hasChildren = function() {
      return !!this.children && this.children.length > 0;
    };
    Node3.prototype.getChildren = function() {
      return this.children ? this.children.slice(0) : [];
    };
    Node3.prototype.getChild = function(index) {
      if (this.children && index < this.children.length) {
        return this.children[index];
      }
      return null;
    };
    Node3.prototype.addChildren = function(nodes) {
      for (var _i = 0, nodes_1 = nodes; _i < nodes_1.length; _i++) {
        var node = nodes_1[_i];
        this.addChild(node);
      }
    };
    Node3.prototype.findFirstChildBeforeOffset = function(offset) {
      if (this.children) {
        var current = null;
        for (var i = this.children.length - 1; i >= 0; i--) {
          current = this.children[i];
          if (current.offset <= offset) {
            return current;
          }
        }
      }
      return null;
    };
    Node3.prototype.findChildAtOffset = function(offset, goDeep) {
      var current = this.findFirstChildBeforeOffset(offset);
      if (current && current.end >= offset) {
        if (goDeep) {
          return current.findChildAtOffset(offset, true) || current;
        }
        return current;
      }
      return null;
    };
    Node3.prototype.encloses = function(candidate) {
      return this.offset <= candidate.offset && this.offset + this.length >= candidate.offset + candidate.length;
    };
    Node3.prototype.getParent = function() {
      var result = this.parent;
      while (result instanceof Nodelist) {
        result = result.parent;
      }
      return result;
    };
    Node3.prototype.findParent = function(type) {
      var result = this;
      while (result && result.type !== type) {
        result = result.parent;
      }
      return result;
    };
    Node3.prototype.findAParent = function() {
      var types = [];
      for (var _i = 0; _i < arguments.length; _i++) {
        types[_i] = arguments[_i];
      }
      var result = this;
      while (result && !types.some(function(t) {
        return result.type === t;
      })) {
        result = result.parent;
      }
      return result;
    };
    Node3.prototype.setData = function(key, value) {
      if (!this.options) {
        this.options = {};
      }
      this.options[key] = value;
    };
    Node3.prototype.getData = function(key) {
      if (!this.options || !this.options.hasOwnProperty(key)) {
        return null;
      }
      return this.options[key];
    };
    return Node3;
  }();
  var Nodelist = function(_super) {
    __extends(Nodelist2, _super);
    function Nodelist2(parent, index) {
      if (index === void 0) {
        index = -1;
      }
      var _this = _super.call(this, -1, -1) || this;
      _this.attachTo(parent, index);
      _this.offset = -1;
      _this.length = -1;
      return _this;
    }
    return Nodelist2;
  }(Node2);
  var Identifier = function(_super) {
    __extends(Identifier2, _super);
    function Identifier2(offset, length) {
      var _this = _super.call(this, offset, length) || this;
      _this.isCustomProperty = false;
      return _this;
    }
    Object.defineProperty(Identifier2.prototype, "type", {
      get: function() {
        return NodeType.Identifier;
      },
      enumerable: true,
      configurable: true
    });
    Identifier2.prototype.containsInterpolation = function() {
      return this.hasChildren();
    };
    return Identifier2;
  }(Node2);
  var Stylesheet = function(_super) {
    __extends(Stylesheet2, _super);
    function Stylesheet2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Stylesheet2.prototype, "type", {
      get: function() {
        return NodeType.Stylesheet;
      },
      enumerable: true,
      configurable: true
    });
    return Stylesheet2;
  }(Node2);
  var Declarations = function(_super) {
    __extends(Declarations2, _super);
    function Declarations2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Declarations2.prototype, "type", {
      get: function() {
        return NodeType.Declarations;
      },
      enumerable: true,
      configurable: true
    });
    return Declarations2;
  }(Node2);
  var BodyDeclaration = function(_super) {
    __extends(BodyDeclaration2, _super);
    function BodyDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    BodyDeclaration2.prototype.getDeclarations = function() {
      return this.declarations;
    };
    BodyDeclaration2.prototype.setDeclarations = function(decls) {
      return this.setNode("declarations", decls);
    };
    return BodyDeclaration2;
  }(Node2);
  var RuleSet = function(_super) {
    __extends(RuleSet2, _super);
    function RuleSet2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(RuleSet2.prototype, "type", {
      get: function() {
        return NodeType.Ruleset;
      },
      enumerable: true,
      configurable: true
    });
    RuleSet2.prototype.getSelectors = function() {
      if (!this.selectors) {
        this.selectors = new Nodelist(this);
      }
      return this.selectors;
    };
    RuleSet2.prototype.isNested = function() {
      return !!this.parent && this.parent.findParent(NodeType.Declarations) !== null;
    };
    return RuleSet2;
  }(BodyDeclaration);
  var Selector = function(_super) {
    __extends(Selector2, _super);
    function Selector2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Selector2.prototype, "type", {
      get: function() {
        return NodeType.Selector;
      },
      enumerable: true,
      configurable: true
    });
    return Selector2;
  }(Node2);
  var SimpleSelector = function(_super) {
    __extends(SimpleSelector2, _super);
    function SimpleSelector2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(SimpleSelector2.prototype, "type", {
      get: function() {
        return NodeType.SimpleSelector;
      },
      enumerable: true,
      configurable: true
    });
    return SimpleSelector2;
  }(Node2);
  var AtApplyRule = function(_super) {
    __extends(AtApplyRule2, _super);
    function AtApplyRule2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(AtApplyRule2.prototype, "type", {
      get: function() {
        return NodeType.AtApplyRule;
      },
      enumerable: true,
      configurable: true
    });
    AtApplyRule2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    AtApplyRule2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    AtApplyRule2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    return AtApplyRule2;
  }(Node2);
  var AbstractDeclaration = function(_super) {
    __extends(AbstractDeclaration2, _super);
    function AbstractDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    return AbstractDeclaration2;
  }(Node2);
  var CustomPropertyDeclaration = function(_super) {
    __extends(CustomPropertyDeclaration2, _super);
    function CustomPropertyDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(CustomPropertyDeclaration2.prototype, "type", {
      get: function() {
        return NodeType.CustomPropertyDeclaration;
      },
      enumerable: true,
      configurable: true
    });
    CustomPropertyDeclaration2.prototype.setProperty = function(node) {
      return this.setNode("property", node);
    };
    CustomPropertyDeclaration2.prototype.getProperty = function() {
      return this.property;
    };
    CustomPropertyDeclaration2.prototype.setValue = function(value) {
      return this.setNode("value", value);
    };
    CustomPropertyDeclaration2.prototype.getValue = function() {
      return this.value;
    };
    CustomPropertyDeclaration2.prototype.setPropertySet = function(value) {
      return this.setNode("propertySet", value);
    };
    CustomPropertyDeclaration2.prototype.getPropertySet = function() {
      return this.propertySet;
    };
    return CustomPropertyDeclaration2;
  }(AbstractDeclaration);
  var CustomPropertySet = function(_super) {
    __extends(CustomPropertySet2, _super);
    function CustomPropertySet2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(CustomPropertySet2.prototype, "type", {
      get: function() {
        return NodeType.CustomPropertySet;
      },
      enumerable: true,
      configurable: true
    });
    return CustomPropertySet2;
  }(BodyDeclaration);
  var Declaration = function(_super) {
    __extends(Declaration2, _super);
    function Declaration2(offset, length) {
      var _this = _super.call(this, offset, length) || this;
      _this.property = null;
      return _this;
    }
    Object.defineProperty(Declaration2.prototype, "type", {
      get: function() {
        return NodeType.Declaration;
      },
      enumerable: true,
      configurable: true
    });
    Declaration2.prototype.setProperty = function(node) {
      return this.setNode("property", node);
    };
    Declaration2.prototype.getProperty = function() {
      return this.property;
    };
    Declaration2.prototype.getFullPropertyName = function() {
      var propertyName = this.property ? this.property.getName() : "unknown";
      if (this.parent instanceof Declarations && this.parent.getParent() instanceof NestedProperties) {
        var parentDecl = this.parent.getParent().getParent();
        if (parentDecl instanceof Declaration2) {
          return parentDecl.getFullPropertyName() + propertyName;
        }
      }
      return propertyName;
    };
    Declaration2.prototype.getNonPrefixedPropertyName = function() {
      var propertyName = this.getFullPropertyName();
      if (propertyName && propertyName.charAt(0) === "-") {
        var vendorPrefixEnd = propertyName.indexOf("-", 1);
        if (vendorPrefixEnd !== -1) {
          return propertyName.substring(vendorPrefixEnd + 1);
        }
      }
      return propertyName;
    };
    Declaration2.prototype.setValue = function(value) {
      return this.setNode("value", value);
    };
    Declaration2.prototype.getValue = function() {
      return this.value;
    };
    Declaration2.prototype.setNestedProperties = function(value) {
      return this.setNode("nestedProperties", value);
    };
    Declaration2.prototype.getNestedProperties = function() {
      return this.nestedProperties;
    };
    return Declaration2;
  }(AbstractDeclaration);
  var Property = function(_super) {
    __extends(Property2, _super);
    function Property2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Property2.prototype, "type", {
      get: function() {
        return NodeType.Property;
      },
      enumerable: true,
      configurable: true
    });
    Property2.prototype.setIdentifier = function(value) {
      return this.setNode("identifier", value);
    };
    Property2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    Property2.prototype.getName = function() {
      return trim(this.getText(), /[_\+]+$/);
    };
    Property2.prototype.isCustomProperty = function() {
      return !!this.identifier && this.identifier.isCustomProperty;
    };
    return Property2;
  }(Node2);
  var Invocation = function(_super) {
    __extends(Invocation2, _super);
    function Invocation2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Invocation2.prototype, "type", {
      get: function() {
        return NodeType.Invocation;
      },
      enumerable: true,
      configurable: true
    });
    Invocation2.prototype.getArguments = function() {
      if (!this.arguments) {
        this.arguments = new Nodelist(this);
      }
      return this.arguments;
    };
    return Invocation2;
  }(Node2);
  var Function = function(_super) {
    __extends(Function2, _super);
    function Function2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Function2.prototype, "type", {
      get: function() {
        return NodeType.Function;
      },
      enumerable: true,
      configurable: true
    });
    Function2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    Function2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    Function2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    return Function2;
  }(Invocation);
  var FunctionParameter = function(_super) {
    __extends(FunctionParameter2, _super);
    function FunctionParameter2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(FunctionParameter2.prototype, "type", {
      get: function() {
        return NodeType.FunctionParameter;
      },
      enumerable: true,
      configurable: true
    });
    FunctionParameter2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    FunctionParameter2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    FunctionParameter2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    FunctionParameter2.prototype.setDefaultValue = function(node) {
      return this.setNode("defaultValue", node, 0);
    };
    FunctionParameter2.prototype.getDefaultValue = function() {
      return this.defaultValue;
    };
    return FunctionParameter2;
  }(Node2);
  var FunctionArgument = function(_super) {
    __extends(FunctionArgument2, _super);
    function FunctionArgument2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(FunctionArgument2.prototype, "type", {
      get: function() {
        return NodeType.FunctionArgument;
      },
      enumerable: true,
      configurable: true
    });
    FunctionArgument2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    FunctionArgument2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    FunctionArgument2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    FunctionArgument2.prototype.setValue = function(node) {
      return this.setNode("value", node, 0);
    };
    FunctionArgument2.prototype.getValue = function() {
      return this.value;
    };
    return FunctionArgument2;
  }(Node2);
  var IfStatement = function(_super) {
    __extends(IfStatement2, _super);
    function IfStatement2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(IfStatement2.prototype, "type", {
      get: function() {
        return NodeType.If;
      },
      enumerable: true,
      configurable: true
    });
    IfStatement2.prototype.setExpression = function(node) {
      return this.setNode("expression", node, 0);
    };
    IfStatement2.prototype.setElseClause = function(elseClause) {
      return this.setNode("elseClause", elseClause);
    };
    return IfStatement2;
  }(BodyDeclaration);
  var ForStatement = function(_super) {
    __extends(ForStatement2, _super);
    function ForStatement2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(ForStatement2.prototype, "type", {
      get: function() {
        return NodeType.For;
      },
      enumerable: true,
      configurable: true
    });
    ForStatement2.prototype.setVariable = function(node) {
      return this.setNode("variable", node, 0);
    };
    return ForStatement2;
  }(BodyDeclaration);
  var EachStatement = function(_super) {
    __extends(EachStatement2, _super);
    function EachStatement2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(EachStatement2.prototype, "type", {
      get: function() {
        return NodeType.Each;
      },
      enumerable: true,
      configurable: true
    });
    EachStatement2.prototype.getVariables = function() {
      if (!this.variables) {
        this.variables = new Nodelist(this);
      }
      return this.variables;
    };
    return EachStatement2;
  }(BodyDeclaration);
  var WhileStatement = function(_super) {
    __extends(WhileStatement2, _super);
    function WhileStatement2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(WhileStatement2.prototype, "type", {
      get: function() {
        return NodeType.While;
      },
      enumerable: true,
      configurable: true
    });
    return WhileStatement2;
  }(BodyDeclaration);
  var ElseStatement = function(_super) {
    __extends(ElseStatement2, _super);
    function ElseStatement2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(ElseStatement2.prototype, "type", {
      get: function() {
        return NodeType.Else;
      },
      enumerable: true,
      configurable: true
    });
    return ElseStatement2;
  }(BodyDeclaration);
  var FunctionDeclaration = function(_super) {
    __extends(FunctionDeclaration2, _super);
    function FunctionDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(FunctionDeclaration2.prototype, "type", {
      get: function() {
        return NodeType.FunctionDeclaration;
      },
      enumerable: true,
      configurable: true
    });
    FunctionDeclaration2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    FunctionDeclaration2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    FunctionDeclaration2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    FunctionDeclaration2.prototype.getParameters = function() {
      if (!this.parameters) {
        this.parameters = new Nodelist(this);
      }
      return this.parameters;
    };
    return FunctionDeclaration2;
  }(BodyDeclaration);
  var ViewPort = function(_super) {
    __extends(ViewPort2, _super);
    function ViewPort2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(ViewPort2.prototype, "type", {
      get: function() {
        return NodeType.ViewPort;
      },
      enumerable: true,
      configurable: true
    });
    return ViewPort2;
  }(BodyDeclaration);
  var FontFace = function(_super) {
    __extends(FontFace2, _super);
    function FontFace2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(FontFace2.prototype, "type", {
      get: function() {
        return NodeType.FontFace;
      },
      enumerable: true,
      configurable: true
    });
    return FontFace2;
  }(BodyDeclaration);
  var NestedProperties = function(_super) {
    __extends(NestedProperties2, _super);
    function NestedProperties2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(NestedProperties2.prototype, "type", {
      get: function() {
        return NodeType.NestedProperties;
      },
      enumerable: true,
      configurable: true
    });
    return NestedProperties2;
  }(BodyDeclaration);
  var Keyframe = function(_super) {
    __extends(Keyframe2, _super);
    function Keyframe2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Keyframe2.prototype, "type", {
      get: function() {
        return NodeType.Keyframe;
      },
      enumerable: true,
      configurable: true
    });
    Keyframe2.prototype.setKeyword = function(keyword) {
      return this.setNode("keyword", keyword, 0);
    };
    Keyframe2.prototype.getKeyword = function() {
      return this.keyword;
    };
    Keyframe2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    Keyframe2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    Keyframe2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    return Keyframe2;
  }(BodyDeclaration);
  var KeyframeSelector = function(_super) {
    __extends(KeyframeSelector2, _super);
    function KeyframeSelector2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(KeyframeSelector2.prototype, "type", {
      get: function() {
        return NodeType.KeyframeSelector;
      },
      enumerable: true,
      configurable: true
    });
    return KeyframeSelector2;
  }(BodyDeclaration);
  var Import = function(_super) {
    __extends(Import2, _super);
    function Import2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Import2.prototype, "type", {
      get: function() {
        return NodeType.Import;
      },
      enumerable: true,
      configurable: true
    });
    Import2.prototype.setMedialist = function(node) {
      if (node) {
        node.attachTo(this);
        return true;
      }
      return false;
    };
    return Import2;
  }(Node2);
  var Use = function(_super) {
    __extends(Use2, _super);
    function Use2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(Use2.prototype, "type", {
      get: function() {
        return NodeType.Use;
      },
      enumerable: true,
      configurable: true
    });
    Use2.prototype.getParameters = function() {
      if (!this.parameters) {
        this.parameters = new Nodelist(this);
      }
      return this.parameters;
    };
    Use2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    Use2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    return Use2;
  }(Node2);
  var ModuleConfiguration = function(_super) {
    __extends(ModuleConfiguration2, _super);
    function ModuleConfiguration2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(ModuleConfiguration2.prototype, "type", {
      get: function() {
        return NodeType.ModuleConfiguration;
      },
      enumerable: true,
      configurable: true
    });
    ModuleConfiguration2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    ModuleConfiguration2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    ModuleConfiguration2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    ModuleConfiguration2.prototype.setValue = function(node) {
      return this.setNode("value", node, 0);
    };
    ModuleConfiguration2.prototype.getValue = function() {
      return this.value;
    };
    return ModuleConfiguration2;
  }(Node2);
  var Forward = function(_super) {
    __extends(Forward2, _super);
    function Forward2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(Forward2.prototype, "type", {
      get: function() {
        return NodeType.Forward;
      },
      enumerable: true,
      configurable: true
    });
    Forward2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    Forward2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    return Forward2;
  }(Node2);
  var ForwardVisibility = function(_super) {
    __extends(ForwardVisibility2, _super);
    function ForwardVisibility2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(ForwardVisibility2.prototype, "type", {
      get: function() {
        return NodeType.ForwardVisibility;
      },
      enumerable: true,
      configurable: true
    });
    ForwardVisibility2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    ForwardVisibility2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    return ForwardVisibility2;
  }(Node2);
  var Namespace = function(_super) {
    __extends(Namespace2, _super);
    function Namespace2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Namespace2.prototype, "type", {
      get: function() {
        return NodeType.Namespace;
      },
      enumerable: true,
      configurable: true
    });
    return Namespace2;
  }(Node2);
  var Media = function(_super) {
    __extends(Media2, _super);
    function Media2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Media2.prototype, "type", {
      get: function() {
        return NodeType.Media;
      },
      enumerable: true,
      configurable: true
    });
    return Media2;
  }(BodyDeclaration);
  var Supports = function(_super) {
    __extends(Supports2, _super);
    function Supports2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Supports2.prototype, "type", {
      get: function() {
        return NodeType.Supports;
      },
      enumerable: true,
      configurable: true
    });
    return Supports2;
  }(BodyDeclaration);
  var Document = function(_super) {
    __extends(Document2, _super);
    function Document2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Document2.prototype, "type", {
      get: function() {
        return NodeType.Document;
      },
      enumerable: true,
      configurable: true
    });
    return Document2;
  }(BodyDeclaration);
  var Medialist = function(_super) {
    __extends(Medialist2, _super);
    function Medialist2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Medialist2.prototype.getMediums = function() {
      if (!this.mediums) {
        this.mediums = new Nodelist(this);
      }
      return this.mediums;
    };
    return Medialist2;
  }(Node2);
  var MediaQuery = function(_super) {
    __extends(MediaQuery2, _super);
    function MediaQuery2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(MediaQuery2.prototype, "type", {
      get: function() {
        return NodeType.MediaQuery;
      },
      enumerable: true,
      configurable: true
    });
    return MediaQuery2;
  }(Node2);
  var SupportsCondition = function(_super) {
    __extends(SupportsCondition2, _super);
    function SupportsCondition2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(SupportsCondition2.prototype, "type", {
      get: function() {
        return NodeType.SupportsCondition;
      },
      enumerable: true,
      configurable: true
    });
    return SupportsCondition2;
  }(Node2);
  var Page = function(_super) {
    __extends(Page2, _super);
    function Page2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Page2.prototype, "type", {
      get: function() {
        return NodeType.Page;
      },
      enumerable: true,
      configurable: true
    });
    return Page2;
  }(BodyDeclaration);
  var PageBoxMarginBox = function(_super) {
    __extends(PageBoxMarginBox2, _super);
    function PageBoxMarginBox2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(PageBoxMarginBox2.prototype, "type", {
      get: function() {
        return NodeType.PageBoxMarginBox;
      },
      enumerable: true,
      configurable: true
    });
    return PageBoxMarginBox2;
  }(BodyDeclaration);
  var Expression = function(_super) {
    __extends(Expression2, _super);
    function Expression2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Expression2.prototype, "type", {
      get: function() {
        return NodeType.Expression;
      },
      enumerable: true,
      configurable: true
    });
    return Expression2;
  }(Node2);
  var BinaryExpression = function(_super) {
    __extends(BinaryExpression2, _super);
    function BinaryExpression2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(BinaryExpression2.prototype, "type", {
      get: function() {
        return NodeType.BinaryExpression;
      },
      enumerable: true,
      configurable: true
    });
    BinaryExpression2.prototype.setLeft = function(left) {
      return this.setNode("left", left);
    };
    BinaryExpression2.prototype.getLeft = function() {
      return this.left;
    };
    BinaryExpression2.prototype.setRight = function(right) {
      return this.setNode("right", right);
    };
    BinaryExpression2.prototype.getRight = function() {
      return this.right;
    };
    BinaryExpression2.prototype.setOperator = function(value) {
      return this.setNode("operator", value);
    };
    BinaryExpression2.prototype.getOperator = function() {
      return this.operator;
    };
    return BinaryExpression2;
  }(Node2);
  var Term = function(_super) {
    __extends(Term2, _super);
    function Term2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Term2.prototype, "type", {
      get: function() {
        return NodeType.Term;
      },
      enumerable: true,
      configurable: true
    });
    Term2.prototype.setOperator = function(value) {
      return this.setNode("operator", value);
    };
    Term2.prototype.getOperator = function() {
      return this.operator;
    };
    Term2.prototype.setExpression = function(value) {
      return this.setNode("expression", value);
    };
    Term2.prototype.getExpression = function() {
      return this.expression;
    };
    return Term2;
  }(Node2);
  var AttributeSelector = function(_super) {
    __extends(AttributeSelector2, _super);
    function AttributeSelector2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(AttributeSelector2.prototype, "type", {
      get: function() {
        return NodeType.AttributeSelector;
      },
      enumerable: true,
      configurable: true
    });
    AttributeSelector2.prototype.setNamespacePrefix = function(value) {
      return this.setNode("namespacePrefix", value);
    };
    AttributeSelector2.prototype.getNamespacePrefix = function() {
      return this.namespacePrefix;
    };
    AttributeSelector2.prototype.setIdentifier = function(value) {
      return this.setNode("identifier", value);
    };
    AttributeSelector2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    AttributeSelector2.prototype.setOperator = function(operator) {
      return this.setNode("operator", operator);
    };
    AttributeSelector2.prototype.getOperator = function() {
      return this.operator;
    };
    AttributeSelector2.prototype.setValue = function(value) {
      return this.setNode("value", value);
    };
    AttributeSelector2.prototype.getValue = function() {
      return this.value;
    };
    return AttributeSelector2;
  }(Node2);
  var Operator = function(_super) {
    __extends(Operator2, _super);
    function Operator2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Operator2.prototype, "type", {
      get: function() {
        return NodeType.Operator;
      },
      enumerable: true,
      configurable: true
    });
    return Operator2;
  }(Node2);
  var HexColorValue = function(_super) {
    __extends(HexColorValue2, _super);
    function HexColorValue2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(HexColorValue2.prototype, "type", {
      get: function() {
        return NodeType.HexColorValue;
      },
      enumerable: true,
      configurable: true
    });
    return HexColorValue2;
  }(Node2);
  var _dot = ".".charCodeAt(0);
  var _02 = "0".charCodeAt(0);
  var _92 = "9".charCodeAt(0);
  var NumericValue = function(_super) {
    __extends(NumericValue2, _super);
    function NumericValue2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(NumericValue2.prototype, "type", {
      get: function() {
        return NodeType.NumericValue;
      },
      enumerable: true,
      configurable: true
    });
    NumericValue2.prototype.getValue = function() {
      var raw = this.getText();
      var unitIdx = 0;
      var code;
      for (var i = 0, len = raw.length; i < len; i++) {
        code = raw.charCodeAt(i);
        if (!(_02 <= code && code <= _92 || code === _dot)) {
          break;
        }
        unitIdx += 1;
      }
      return {
        value: raw.substring(0, unitIdx),
        unit: unitIdx < raw.length ? raw.substring(unitIdx) : void 0
      };
    };
    return NumericValue2;
  }(Node2);
  var VariableDeclaration = function(_super) {
    __extends(VariableDeclaration2, _super);
    function VariableDeclaration2(offset, length) {
      var _this = _super.call(this, offset, length) || this;
      _this.variable = null;
      _this.value = null;
      _this.needsSemicolon = true;
      return _this;
    }
    Object.defineProperty(VariableDeclaration2.prototype, "type", {
      get: function() {
        return NodeType.VariableDeclaration;
      },
      enumerable: true,
      configurable: true
    });
    VariableDeclaration2.prototype.setVariable = function(node) {
      if (node) {
        node.attachTo(this);
        this.variable = node;
        return true;
      }
      return false;
    };
    VariableDeclaration2.prototype.getVariable = function() {
      return this.variable;
    };
    VariableDeclaration2.prototype.getName = function() {
      return this.variable ? this.variable.getName() : "";
    };
    VariableDeclaration2.prototype.setValue = function(node) {
      if (node) {
        node.attachTo(this);
        this.value = node;
        return true;
      }
      return false;
    };
    VariableDeclaration2.prototype.getValue = function() {
      return this.value;
    };
    return VariableDeclaration2;
  }(AbstractDeclaration);
  var Interpolation = function(_super) {
    __extends(Interpolation2, _super);
    function Interpolation2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Interpolation2.prototype, "type", {
      get: function() {
        return NodeType.Interpolation;
      },
      enumerable: true,
      configurable: true
    });
    return Interpolation2;
  }(Node2);
  var Variable = function(_super) {
    __extends(Variable2, _super);
    function Variable2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(Variable2.prototype, "type", {
      get: function() {
        return NodeType.VariableName;
      },
      enumerable: true,
      configurable: true
    });
    Variable2.prototype.getName = function() {
      return this.getText();
    };
    return Variable2;
  }(Node2);
  var ExtendsReference = function(_super) {
    __extends(ExtendsReference2, _super);
    function ExtendsReference2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(ExtendsReference2.prototype, "type", {
      get: function() {
        return NodeType.ExtendsReference;
      },
      enumerable: true,
      configurable: true
    });
    ExtendsReference2.prototype.getSelectors = function() {
      if (!this.selectors) {
        this.selectors = new Nodelist(this);
      }
      return this.selectors;
    };
    return ExtendsReference2;
  }(Node2);
  var MixinContentReference = function(_super) {
    __extends(MixinContentReference2, _super);
    function MixinContentReference2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(MixinContentReference2.prototype, "type", {
      get: function() {
        return NodeType.MixinContentReference;
      },
      enumerable: true,
      configurable: true
    });
    MixinContentReference2.prototype.getArguments = function() {
      if (!this.arguments) {
        this.arguments = new Nodelist(this);
      }
      return this.arguments;
    };
    return MixinContentReference2;
  }(Node2);
  var MixinContentDeclaration = function(_super) {
    __extends(MixinContentDeclaration2, _super);
    function MixinContentDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(MixinContentDeclaration2.prototype, "type", {
      get: function() {
        return NodeType.MixinContentReference;
      },
      enumerable: true,
      configurable: true
    });
    MixinContentDeclaration2.prototype.getParameters = function() {
      if (!this.parameters) {
        this.parameters = new Nodelist(this);
      }
      return this.parameters;
    };
    return MixinContentDeclaration2;
  }(BodyDeclaration);
  var MixinReference = function(_super) {
    __extends(MixinReference2, _super);
    function MixinReference2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(MixinReference2.prototype, "type", {
      get: function() {
        return NodeType.MixinReference;
      },
      enumerable: true,
      configurable: true
    });
    MixinReference2.prototype.getNamespaces = function() {
      if (!this.namespaces) {
        this.namespaces = new Nodelist(this);
      }
      return this.namespaces;
    };
    MixinReference2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    MixinReference2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    MixinReference2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    MixinReference2.prototype.getArguments = function() {
      if (!this.arguments) {
        this.arguments = new Nodelist(this);
      }
      return this.arguments;
    };
    MixinReference2.prototype.setContent = function(node) {
      return this.setNode("content", node);
    };
    MixinReference2.prototype.getContent = function() {
      return this.content;
    };
    return MixinReference2;
  }(Node2);
  var MixinDeclaration = function(_super) {
    __extends(MixinDeclaration2, _super);
    function MixinDeclaration2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(MixinDeclaration2.prototype, "type", {
      get: function() {
        return NodeType.MixinDeclaration;
      },
      enumerable: true,
      configurable: true
    });
    MixinDeclaration2.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    MixinDeclaration2.prototype.getIdentifier = function() {
      return this.identifier;
    };
    MixinDeclaration2.prototype.getName = function() {
      return this.identifier ? this.identifier.getText() : "";
    };
    MixinDeclaration2.prototype.getParameters = function() {
      if (!this.parameters) {
        this.parameters = new Nodelist(this);
      }
      return this.parameters;
    };
    MixinDeclaration2.prototype.setGuard = function(node) {
      if (node) {
        node.attachTo(this);
        this.guard = node;
      }
      return false;
    };
    return MixinDeclaration2;
  }(BodyDeclaration);
  var UnknownAtRule = function(_super) {
    __extends(UnknownAtRule2, _super);
    function UnknownAtRule2(offset, length) {
      return _super.call(this, offset, length) || this;
    }
    Object.defineProperty(UnknownAtRule2.prototype, "type", {
      get: function() {
        return NodeType.UnknownAtRule;
      },
      enumerable: true,
      configurable: true
    });
    UnknownAtRule2.prototype.setAtRuleName = function(atRuleName) {
      this.atRuleName = atRuleName;
    };
    UnknownAtRule2.prototype.getAtRuleName = function() {
      return this.atRuleName;
    };
    return UnknownAtRule2;
  }(BodyDeclaration);
  var ListEntry = function(_super) {
    __extends(ListEntry2, _super);
    function ListEntry2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(ListEntry2.prototype, "type", {
      get: function() {
        return NodeType.ListEntry;
      },
      enumerable: true,
      configurable: true
    });
    ListEntry2.prototype.setKey = function(node) {
      return this.setNode("key", node, 0);
    };
    ListEntry2.prototype.setValue = function(node) {
      return this.setNode("value", node, 1);
    };
    return ListEntry2;
  }(Node2);
  var LessGuard = function(_super) {
    __extends(LessGuard2, _super);
    function LessGuard2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    LessGuard2.prototype.getConditions = function() {
      if (!this.conditions) {
        this.conditions = new Nodelist(this);
      }
      return this.conditions;
    };
    return LessGuard2;
  }(Node2);
  var GuardCondition = function(_super) {
    __extends(GuardCondition2, _super);
    function GuardCondition2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    GuardCondition2.prototype.setVariable = function(node) {
      return this.setNode("variable", node);
    };
    return GuardCondition2;
  }(Node2);
  var Module = function(_super) {
    __extends(Module3, _super);
    function Module3() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    Object.defineProperty(Module3.prototype, "type", {
      get: function() {
        return NodeType.Module;
      },
      enumerable: true,
      configurable: true
    });
    Module3.prototype.setIdentifier = function(node) {
      return this.setNode("identifier", node, 0);
    };
    Module3.prototype.getIdentifier = function() {
      return this.identifier;
    };
    return Module3;
  }(Node2);
  var Level;
  (function(Level2) {
    Level2[Level2["Ignore"] = 1] = "Ignore";
    Level2[Level2["Warning"] = 2] = "Warning";
    Level2[Level2["Error"] = 4] = "Error";
  })(Level || (Level = {}));
  var Marker = function() {
    function Marker2(node, rule, level, message, offset, length) {
      if (offset === void 0) {
        offset = node.offset;
      }
      if (length === void 0) {
        length = node.length;
      }
      this.node = node;
      this.rule = rule;
      this.level = level;
      this.message = message || rule.message;
      this.offset = offset;
      this.length = length;
    }
    Marker2.prototype.getRule = function() {
      return this.rule;
    };
    Marker2.prototype.getLevel = function() {
      return this.level;
    };
    Marker2.prototype.getOffset = function() {
      return this.offset;
    };
    Marker2.prototype.getLength = function() {
      return this.length;
    };
    Marker2.prototype.getNode = function() {
      return this.node;
    };
    Marker2.prototype.getMessage = function() {
      return this.message;
    };
    return Marker2;
  }();
  var ParseErrorCollector = function() {
    function ParseErrorCollector2() {
      this.entries = [];
    }
    ParseErrorCollector2.entries = function(node) {
      var visitor = new ParseErrorCollector2();
      node.acceptVisitor(visitor);
      return visitor.entries;
    };
    ParseErrorCollector2.prototype.visitNode = function(node) {
      if (node.isErroneous()) {
        node.collectIssues(this.entries);
      }
      return true;
    };
    return ParseErrorCollector2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/fillers/vscode-nls.js
  function format(message, args) {
    var result;
    if (args.length === 0) {
      result = message;
    } else {
      result = message.replace(/\{(\d+)\}/g, function(match, rest) {
        var index = rest[0];
        return typeof args[index] !== "undefined" ? args[index] : match;
      });
    }
    return result;
  }
  function localize(key, message) {
    var args = [];
    for (var _i = 2; _i < arguments.length; _i++) {
      args[_i - 2] = arguments[_i];
    }
    return format(message, args);
  }
  function loadMessageBundle(file) {
    return localize;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/cssErrors.js
  var localize2 = loadMessageBundle();
  var CSSIssueType = function() {
    function CSSIssueType2(id, message) {
      this.id = id;
      this.message = message;
    }
    return CSSIssueType2;
  }();
  var ParseError = {
    NumberExpected: new CSSIssueType("css-numberexpected", localize2("expected.number", "number expected")),
    ConditionExpected: new CSSIssueType("css-conditionexpected", localize2("expected.condt", "condition expected")),
    RuleOrSelectorExpected: new CSSIssueType("css-ruleorselectorexpected", localize2("expected.ruleorselector", "at-rule or selector expected")),
    DotExpected: new CSSIssueType("css-dotexpected", localize2("expected.dot", "dot expected")),
    ColonExpected: new CSSIssueType("css-colonexpected", localize2("expected.colon", "colon expected")),
    SemiColonExpected: new CSSIssueType("css-semicolonexpected", localize2("expected.semicolon", "semi-colon expected")),
    TermExpected: new CSSIssueType("css-termexpected", localize2("expected.term", "term expected")),
    ExpressionExpected: new CSSIssueType("css-expressionexpected", localize2("expected.expression", "expression expected")),
    OperatorExpected: new CSSIssueType("css-operatorexpected", localize2("expected.operator", "operator expected")),
    IdentifierExpected: new CSSIssueType("css-identifierexpected", localize2("expected.ident", "identifier expected")),
    PercentageExpected: new CSSIssueType("css-percentageexpected", localize2("expected.percentage", "percentage expected")),
    URIOrStringExpected: new CSSIssueType("css-uriorstringexpected", localize2("expected.uriorstring", "uri or string expected")),
    URIExpected: new CSSIssueType("css-uriexpected", localize2("expected.uri", "URI expected")),
    VariableNameExpected: new CSSIssueType("css-varnameexpected", localize2("expected.varname", "variable name expected")),
    VariableValueExpected: new CSSIssueType("css-varvalueexpected", localize2("expected.varvalue", "variable value expected")),
    PropertyValueExpected: new CSSIssueType("css-propertyvalueexpected", localize2("expected.propvalue", "property value expected")),
    LeftCurlyExpected: new CSSIssueType("css-lcurlyexpected", localize2("expected.lcurly", "{ expected")),
    RightCurlyExpected: new CSSIssueType("css-rcurlyexpected", localize2("expected.rcurly", "} expected")),
    LeftSquareBracketExpected: new CSSIssueType("css-rbracketexpected", localize2("expected.lsquare", "[ expected")),
    RightSquareBracketExpected: new CSSIssueType("css-lbracketexpected", localize2("expected.rsquare", "] expected")),
    LeftParenthesisExpected: new CSSIssueType("css-lparentexpected", localize2("expected.lparen", "( expected")),
    RightParenthesisExpected: new CSSIssueType("css-rparentexpected", localize2("expected.rparent", ") expected")),
    CommaExpected: new CSSIssueType("css-commaexpected", localize2("expected.comma", "comma expected")),
    PageDirectiveOrDeclarationExpected: new CSSIssueType("css-pagedirordeclexpected", localize2("expected.pagedirordecl", "page directive or declaraton expected")),
    UnknownAtRule: new CSSIssueType("css-unknownatrule", localize2("unknown.atrule", "at-rule unknown")),
    UnknownKeyword: new CSSIssueType("css-unknownkeyword", localize2("unknown.keyword", "unknown keyword")),
    SelectorExpected: new CSSIssueType("css-selectorexpected", localize2("expected.selector", "selector expected")),
    StringLiteralExpected: new CSSIssueType("css-stringliteralexpected", localize2("expected.stringliteral", "string literal expected")),
    WhitespaceExpected: new CSSIssueType("css-whitespaceexpected", localize2("expected.whitespace", "whitespace expected")),
    MediaQueryExpected: new CSSIssueType("css-mediaqueryexpected", localize2("expected.mediaquery", "media query expected")),
    IdentifierOrWildcardExpected: new CSSIssueType("css-idorwildcardexpected", localize2("expected.idorwildcard", "identifier or wildcard expected")),
    WildcardExpected: new CSSIssueType("css-wildcardexpected", localize2("expected.wildcard", "wildcard expected")),
    IdentifierOrVariableExpected: new CSSIssueType("css-idorvarexpected", localize2("expected.idorvar", "identifier or variable expected"))
  };

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/languageFacts/entry.js
  var browserNames = {
    E: "Edge",
    FF: "Firefox",
    S: "Safari",
    C: "Chrome",
    IE: "IE",
    O: "Opera"
  };
  function getEntryStatus(status) {
    switch (status) {
      case "experimental":
        return "\u26A0\uFE0F Property is experimental. Be cautious when using it.\uFE0F\n\n";
      case "nonstandard":
        return "\u{1F6A8}\uFE0F Property is nonstandard. Avoid using it.\n\n";
      case "obsolete":
        return "\u{1F6A8}\uFE0F\uFE0F\uFE0F Property is obsolete. Avoid using it.\n\n";
      default:
        return "";
    }
  }
  function getEntryDescription(entry, doesSupportMarkdown) {
    var result;
    if (doesSupportMarkdown) {
      result = {
        kind: "markdown",
        value: getEntryMarkdownDescription(entry)
      };
    } else {
      result = {
        kind: "plaintext",
        value: getEntryStringDescription(entry)
      };
    }
    if (result.value === "") {
      return void 0;
    }
    return result;
  }
  function textToMarkedString(text) {
    text = text.replace(/[\\`*_{}[\]()#+\-.!]/g, "\\$&");
    return text.replace(/</g, "&lt;").replace(/>/g, "&gt;");
  }
  function getEntryStringDescription(entry) {
    if (!entry.description || entry.description === "") {
      return "";
    }
    if (typeof entry.description !== "string") {
      return entry.description.value;
    }
    var result = "";
    if (entry.status) {
      result += getEntryStatus(entry.status);
    }
    result += entry.description;
    var browserLabel = getBrowserLabel(entry.browsers);
    if (browserLabel) {
      result += "\n(" + browserLabel + ")";
    }
    if ("syntax" in entry) {
      result += "\n\nSyntax: " + entry.syntax;
    }
    if (entry.references && entry.references.length > 0) {
      result += "\n\n";
      result += entry.references.map(function(r) {
        return r.name + ": " + r.url;
      }).join(" | ");
    }
    return result;
  }
  function getEntryMarkdownDescription(entry) {
    if (!entry.description || entry.description === "") {
      return "";
    }
    var result = "";
    if (entry.status) {
      result += getEntryStatus(entry.status);
    }
    var description = typeof entry.description === "string" ? entry.description : entry.description.value;
    result += textToMarkedString(description);
    var browserLabel = getBrowserLabel(entry.browsers);
    if (browserLabel) {
      result += "\n\n(" + textToMarkedString(browserLabel) + ")";
    }
    if ("syntax" in entry && entry.syntax) {
      result += "\n\nSyntax: " + textToMarkedString(entry.syntax);
    }
    if (entry.references && entry.references.length > 0) {
      result += "\n\n";
      result += entry.references.map(function(r) {
        return "[" + r.name + "](" + r.url + ")";
      }).join(" | ");
    }
    return result;
  }
  function getBrowserLabel(browsers) {
    if (browsers === void 0) {
      browsers = [];
    }
    if (browsers.length === 0) {
      return null;
    }
    return browsers.map(function(b) {
      var result = "";
      var matches2 = b.match(/([A-Z]+)(\d+)?/);
      var name = matches2[1];
      var version = matches2[2];
      if (name in browserNames) {
        result += browserNames[name];
      }
      if (version) {
        result += " " + version;
      }
      return result;
    }).join(", ");
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/languageFacts/colors.js
  var localize3 = loadMessageBundle();
  var colorFunctions = [
    { func: "rgb($red, $green, $blue)", desc: localize3("css.builtin.rgb", "Creates a Color from red, green, and blue values.") },
    { func: "rgba($red, $green, $blue, $alpha)", desc: localize3("css.builtin.rgba", "Creates a Color from red, green, blue, and alpha values.") },
    { func: "hsl($hue, $saturation, $lightness)", desc: localize3("css.builtin.hsl", "Creates a Color from hue, saturation, and lightness values.") },
    { func: "hsla($hue, $saturation, $lightness, $alpha)", desc: localize3("css.builtin.hsla", "Creates a Color from hue, saturation, lightness, and alpha values.") }
  ];
  var colors = {
    aliceblue: "#f0f8ff",
    antiquewhite: "#faebd7",
    aqua: "#00ffff",
    aquamarine: "#7fffd4",
    azure: "#f0ffff",
    beige: "#f5f5dc",
    bisque: "#ffe4c4",
    black: "#000000",
    blanchedalmond: "#ffebcd",
    blue: "#0000ff",
    blueviolet: "#8a2be2",
    brown: "#a52a2a",
    burlywood: "#deb887",
    cadetblue: "#5f9ea0",
    chartreuse: "#7fff00",
    chocolate: "#d2691e",
    coral: "#ff7f50",
    cornflowerblue: "#6495ed",
    cornsilk: "#fff8dc",
    crimson: "#dc143c",
    cyan: "#00ffff",
    darkblue: "#00008b",
    darkcyan: "#008b8b",
    darkgoldenrod: "#b8860b",
    darkgray: "#a9a9a9",
    darkgrey: "#a9a9a9",
    darkgreen: "#006400",
    darkkhaki: "#bdb76b",
    darkmagenta: "#8b008b",
    darkolivegreen: "#556b2f",
    darkorange: "#ff8c00",
    darkorchid: "#9932cc",
    darkred: "#8b0000",
    darksalmon: "#e9967a",
    darkseagreen: "#8fbc8f",
    darkslateblue: "#483d8b",
    darkslategray: "#2f4f4f",
    darkslategrey: "#2f4f4f",
    darkturquoise: "#00ced1",
    darkviolet: "#9400d3",
    deeppink: "#ff1493",
    deepskyblue: "#00bfff",
    dimgray: "#696969",
    dimgrey: "#696969",
    dodgerblue: "#1e90ff",
    firebrick: "#b22222",
    floralwhite: "#fffaf0",
    forestgreen: "#228b22",
    fuchsia: "#ff00ff",
    gainsboro: "#dcdcdc",
    ghostwhite: "#f8f8ff",
    gold: "#ffd700",
    goldenrod: "#daa520",
    gray: "#808080",
    grey: "#808080",
    green: "#008000",
    greenyellow: "#adff2f",
    honeydew: "#f0fff0",
    hotpink: "#ff69b4",
    indianred: "#cd5c5c",
    indigo: "#4b0082",
    ivory: "#fffff0",
    khaki: "#f0e68c",
    lavender: "#e6e6fa",
    lavenderblush: "#fff0f5",
    lawngreen: "#7cfc00",
    lemonchiffon: "#fffacd",
    lightblue: "#add8e6",
    lightcoral: "#f08080",
    lightcyan: "#e0ffff",
    lightgoldenrodyellow: "#fafad2",
    lightgray: "#d3d3d3",
    lightgrey: "#d3d3d3",
    lightgreen: "#90ee90",
    lightpink: "#ffb6c1",
    lightsalmon: "#ffa07a",
    lightseagreen: "#20b2aa",
    lightskyblue: "#87cefa",
    lightslategray: "#778899",
    lightslategrey: "#778899",
    lightsteelblue: "#b0c4de",
    lightyellow: "#ffffe0",
    lime: "#00ff00",
    limegreen: "#32cd32",
    linen: "#faf0e6",
    magenta: "#ff00ff",
    maroon: "#800000",
    mediumaquamarine: "#66cdaa",
    mediumblue: "#0000cd",
    mediumorchid: "#ba55d3",
    mediumpurple: "#9370d8",
    mediumseagreen: "#3cb371",
    mediumslateblue: "#7b68ee",
    mediumspringgreen: "#00fa9a",
    mediumturquoise: "#48d1cc",
    mediumvioletred: "#c71585",
    midnightblue: "#191970",
    mintcream: "#f5fffa",
    mistyrose: "#ffe4e1",
    moccasin: "#ffe4b5",
    navajowhite: "#ffdead",
    navy: "#000080",
    oldlace: "#fdf5e6",
    olive: "#808000",
    olivedrab: "#6b8e23",
    orange: "#ffa500",
    orangered: "#ff4500",
    orchid: "#da70d6",
    palegoldenrod: "#eee8aa",
    palegreen: "#98fb98",
    paleturquoise: "#afeeee",
    palevioletred: "#d87093",
    papayawhip: "#ffefd5",
    peachpuff: "#ffdab9",
    peru: "#cd853f",
    pink: "#ffc0cb",
    plum: "#dda0dd",
    powderblue: "#b0e0e6",
    purple: "#800080",
    red: "#ff0000",
    rebeccapurple: "#663399",
    rosybrown: "#bc8f8f",
    royalblue: "#4169e1",
    saddlebrown: "#8b4513",
    salmon: "#fa8072",
    sandybrown: "#f4a460",
    seagreen: "#2e8b57",
    seashell: "#fff5ee",
    sienna: "#a0522d",
    silver: "#c0c0c0",
    skyblue: "#87ceeb",
    slateblue: "#6a5acd",
    slategray: "#708090",
    slategrey: "#708090",
    snow: "#fffafa",
    springgreen: "#00ff7f",
    steelblue: "#4682b4",
    tan: "#d2b48c",
    teal: "#008080",
    thistle: "#d8bfd8",
    tomato: "#ff6347",
    turquoise: "#40e0d0",
    violet: "#ee82ee",
    wheat: "#f5deb3",
    white: "#ffffff",
    whitesmoke: "#f5f5f5",
    yellow: "#ffff00",
    yellowgreen: "#9acd32"
  };
  var colorKeywords = {
    "currentColor": "The value of the 'color' property. The computed value of the 'currentColor' keyword is the computed value of the 'color' property. If the 'currentColor' keyword is set on the 'color' property itself, it is treated as 'color:inherit' at parse time.",
    "transparent": "Fully transparent. This keyword can be considered a shorthand for rgba(0,0,0,0) which is its computed value."
  };
  function getNumericValue(node, factor) {
    var val = node.getText();
    var m = val.match(/^([-+]?[0-9]*\.?[0-9]+)(%?)$/);
    if (m) {
      if (m[2]) {
        factor = 100;
      }
      var result = parseFloat(m[1]) / factor;
      if (result >= 0 && result <= 1) {
        return result;
      }
    }
    throw new Error();
  }
  function getAngle(node) {
    var val = node.getText();
    var m = val.match(/^([-+]?[0-9]*\.?[0-9]+)(deg)?$/);
    if (m) {
      return parseFloat(val) % 360;
    }
    throw new Error();
  }
  function isColorConstructor(node) {
    var name = node.getName();
    if (!name) {
      return false;
    }
    return /^(rgb|rgba|hsl|hsla)$/gi.test(name);
  }
  var Digit0 = 48;
  var Digit9 = 57;
  var A = 65;
  var a = 97;
  var f = 102;
  function hexDigit(charCode) {
    if (charCode < Digit0) {
      return 0;
    }
    if (charCode <= Digit9) {
      return charCode - Digit0;
    }
    if (charCode < a) {
      charCode += a - A;
    }
    if (charCode >= a && charCode <= f) {
      return charCode - a + 10;
    }
    return 0;
  }
  function colorFromHex(text) {
    if (text[0] !== "#") {
      return null;
    }
    switch (text.length) {
      case 4:
        return {
          red: hexDigit(text.charCodeAt(1)) * 17 / 255,
          green: hexDigit(text.charCodeAt(2)) * 17 / 255,
          blue: hexDigit(text.charCodeAt(3)) * 17 / 255,
          alpha: 1
        };
      case 5:
        return {
          red: hexDigit(text.charCodeAt(1)) * 17 / 255,
          green: hexDigit(text.charCodeAt(2)) * 17 / 255,
          blue: hexDigit(text.charCodeAt(3)) * 17 / 255,
          alpha: hexDigit(text.charCodeAt(4)) * 17 / 255
        };
      case 7:
        return {
          red: (hexDigit(text.charCodeAt(1)) * 16 + hexDigit(text.charCodeAt(2))) / 255,
          green: (hexDigit(text.charCodeAt(3)) * 16 + hexDigit(text.charCodeAt(4))) / 255,
          blue: (hexDigit(text.charCodeAt(5)) * 16 + hexDigit(text.charCodeAt(6))) / 255,
          alpha: 1
        };
      case 9:
        return {
          red: (hexDigit(text.charCodeAt(1)) * 16 + hexDigit(text.charCodeAt(2))) / 255,
          green: (hexDigit(text.charCodeAt(3)) * 16 + hexDigit(text.charCodeAt(4))) / 255,
          blue: (hexDigit(text.charCodeAt(5)) * 16 + hexDigit(text.charCodeAt(6))) / 255,
          alpha: (hexDigit(text.charCodeAt(7)) * 16 + hexDigit(text.charCodeAt(8))) / 255
        };
    }
    return null;
  }
  function colorFromHSL(hue, sat, light, alpha) {
    if (alpha === void 0) {
      alpha = 1;
    }
    hue = hue / 60;
    if (sat === 0) {
      return { red: light, green: light, blue: light, alpha };
    } else {
      var hueToRgb = function(t12, t22, hue2) {
        while (hue2 < 0) {
          hue2 += 6;
        }
        while (hue2 >= 6) {
          hue2 -= 6;
        }
        if (hue2 < 1) {
          return (t22 - t12) * hue2 + t12;
        }
        if (hue2 < 3) {
          return t22;
        }
        if (hue2 < 4) {
          return (t22 - t12) * (4 - hue2) + t12;
        }
        return t12;
      };
      var t2 = light <= 0.5 ? light * (sat + 1) : light + sat - light * sat;
      var t1 = light * 2 - t2;
      return { red: hueToRgb(t1, t2, hue + 2), green: hueToRgb(t1, t2, hue), blue: hueToRgb(t1, t2, hue - 2), alpha };
    }
  }
  function hslFromColor(rgba) {
    var r = rgba.red;
    var g = rgba.green;
    var b = rgba.blue;
    var a2 = rgba.alpha;
    var max = Math.max(r, g, b);
    var min = Math.min(r, g, b);
    var h = 0;
    var s = 0;
    var l = (min + max) / 2;
    var chroma = max - min;
    if (chroma > 0) {
      s = Math.min(l <= 0.5 ? chroma / (2 * l) : chroma / (2 - 2 * l), 1);
      switch (max) {
        case r:
          h = (g - b) / chroma + (g < b ? 6 : 0);
          break;
        case g:
          h = (b - r) / chroma + 2;
          break;
        case b:
          h = (r - g) / chroma + 4;
          break;
      }
      h *= 60;
      h = Math.round(h);
    }
    return { h, s, l, a: a2 };
  }
  function getColorValue(node) {
    if (node.type === NodeType.HexColorValue) {
      var text = node.getText();
      return colorFromHex(text);
    } else if (node.type === NodeType.Function) {
      var functionNode = node;
      var name = functionNode.getName();
      var colorValues = functionNode.getArguments().getChildren();
      if (!name || colorValues.length < 3 || colorValues.length > 4) {
        return null;
      }
      try {
        var alpha = colorValues.length === 4 ? getNumericValue(colorValues[3], 1) : 1;
        if (name === "rgb" || name === "rgba") {
          return {
            red: getNumericValue(colorValues[0], 255),
            green: getNumericValue(colorValues[1], 255),
            blue: getNumericValue(colorValues[2], 255),
            alpha
          };
        } else if (name === "hsl" || name === "hsla") {
          var h = getAngle(colorValues[0]);
          var s = getNumericValue(colorValues[1], 100);
          var l = getNumericValue(colorValues[2], 100);
          return colorFromHSL(h, s, l, alpha);
        }
      } catch (e) {
        return null;
      }
    } else if (node.type === NodeType.Identifier) {
      if (node.parent && node.parent.type !== NodeType.Term) {
        return null;
      }
      var term = node.parent;
      if (term && term.parent && term.parent.type === NodeType.BinaryExpression) {
        var expression = term.parent;
        if (expression.parent && expression.parent.type === NodeType.ListEntry && expression.parent.key === expression) {
          return null;
        }
      }
      var candidateColor = node.getText().toLowerCase();
      if (candidateColor === "none") {
        return null;
      }
      var colorHex = colors[candidateColor];
      if (colorHex) {
        return colorFromHex(colorHex);
      }
    }
    return null;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/languageFacts/builtinData.js
  var positionKeywords = {
    "bottom": "Computes to \u2018100%\u2019 for the vertical position if one or two values are given, otherwise specifies the bottom edge as the origin for the next offset.",
    "center": "Computes to \u201850%\u2019 (\u2018left 50%\u2019) for the horizontal position if the horizontal position is not otherwise specified, or \u201850%\u2019 (\u2018top 50%\u2019) for the vertical position if it is.",
    "left": "Computes to \u20180%\u2019 for the horizontal position if one or two values are given, otherwise specifies the left edge as the origin for the next offset.",
    "right": "Computes to \u2018100%\u2019 for the horizontal position if one or two values are given, otherwise specifies the right edge as the origin for the next offset.",
    "top": "Computes to \u20180%\u2019 for the vertical position if one or two values are given, otherwise specifies the top edge as the origin for the next offset."
  };
  var repeatStyleKeywords = {
    "no-repeat": "Placed once and not repeated in this direction.",
    "repeat": "Repeated in this direction as often as needed to cover the background painting area.",
    "repeat-x": "Computes to \u2018repeat no-repeat\u2019.",
    "repeat-y": "Computes to \u2018no-repeat repeat\u2019.",
    "round": "Repeated as often as will fit within the background positioning area. If it doesn\u2019t fit a whole number of times, it is rescaled so that it does.",
    "space": "Repeated as often as will fit within the background positioning area without being clipped and then the images are spaced out to fill the area."
  };
  var lineStyleKeywords = {
    "dashed": "A series of square-ended dashes.",
    "dotted": "A series of round dots.",
    "double": "Two parallel solid lines with some space between them.",
    "groove": "Looks as if it were carved in the canvas.",
    "hidden": "Same as \u2018none\u2019, but has different behavior in the border conflict resolution rules for border-collapsed tables.",
    "inset": "Looks as if the content on the inside of the border is sunken into the canvas.",
    "none": "No border. Color and width are ignored.",
    "outset": "Looks as if the content on the inside of the border is coming out of the canvas.",
    "ridge": "Looks as if it were coming out of the canvas.",
    "solid": "A single line segment."
  };
  var lineWidthKeywords = ["medium", "thick", "thin"];
  var boxKeywords = {
    "border-box": "The background is painted within (clipped to) the border box.",
    "content-box": "The background is painted within (clipped to) the content box.",
    "padding-box": "The background is painted within (clipped to) the padding box."
  };
  var geometryBoxKeywords = {
    "margin-box": "Uses the margin box as reference box.",
    "fill-box": "Uses the object bounding box as reference box.",
    "stroke-box": "Uses the stroke bounding box as reference box.",
    "view-box": "Uses the nearest SVG viewport as reference box."
  };
  var cssWideKeywords = {
    "initial": "Represents the value specified as the property\u2019s initial value.",
    "inherit": "Represents the computed value of the property on the element\u2019s parent.",
    "unset": "Acts as either `inherit` or `initial`, depending on whether the property is inherited or not."
  };
  var imageFunctions = {
    "url()": "Reference an image file by URL",
    "image()": "Provide image fallbacks and annotations.",
    "-webkit-image-set()": "Provide multiple resolutions. Remember to use unprefixed image-set() in addition.",
    "image-set()": "Provide multiple resolutions of an image and const the UA decide which is most appropriate in a given situation.",
    "-moz-element()": "Use an element in the document as an image. Remember to use unprefixed element() in addition.",
    "element()": "Use an element in the document as an image.",
    "cross-fade()": "Indicates the two images to be combined and how far along in the transition the combination is.",
    "-webkit-gradient()": "Deprecated. Use modern linear-gradient() or radial-gradient() instead.",
    "-webkit-linear-gradient()": "Linear gradient. Remember to use unprefixed version in addition.",
    "-moz-linear-gradient()": "Linear gradient. Remember to use unprefixed version in addition.",
    "-o-linear-gradient()": "Linear gradient. Remember to use unprefixed version in addition.",
    "linear-gradient()": "A linear gradient is created by specifying a straight gradient line, and then several colors placed along that line.",
    "-webkit-repeating-linear-gradient()": "Repeating Linear gradient. Remember to use unprefixed version in addition.",
    "-moz-repeating-linear-gradient()": "Repeating Linear gradient. Remember to use unprefixed version in addition.",
    "-o-repeating-linear-gradient()": "Repeating Linear gradient. Remember to use unprefixed version in addition.",
    "repeating-linear-gradient()": "Same as linear-gradient, except the color-stops are repeated infinitely in both directions, with their positions shifted by multiples of the difference between the last specified color-stop\u2019s position and the first specified color-stop\u2019s position.",
    "-webkit-radial-gradient()": "Radial gradient. Remember to use unprefixed version in addition.",
    "-moz-radial-gradient()": "Radial gradient. Remember to use unprefixed version in addition.",
    "radial-gradient()": "Colors emerge from a single point and smoothly spread outward in a circular or elliptical shape.",
    "-webkit-repeating-radial-gradient()": "Repeating radial gradient. Remember to use unprefixed version in addition.",
    "-moz-repeating-radial-gradient()": "Repeating radial gradient. Remember to use unprefixed version in addition.",
    "repeating-radial-gradient()": "Same as radial-gradient, except the color-stops are repeated infinitely in both directions, with their positions shifted by multiples of the difference between the last specified color-stop\u2019s position and the first specified color-stop\u2019s position."
  };
  var transitionTimingFunctions = {
    "ease": "Equivalent to cubic-bezier(0.25, 0.1, 0.25, 1.0).",
    "ease-in": "Equivalent to cubic-bezier(0.42, 0, 1.0, 1.0).",
    "ease-in-out": "Equivalent to cubic-bezier(0.42, 0, 0.58, 1.0).",
    "ease-out": "Equivalent to cubic-bezier(0, 0, 0.58, 1.0).",
    "linear": "Equivalent to cubic-bezier(0.0, 0.0, 1.0, 1.0).",
    "step-end": "Equivalent to steps(1, end).",
    "step-start": "Equivalent to steps(1, start).",
    "steps()": "The first parameter specifies the number of intervals in the function. The second parameter, which is optional, is either the value \u201Cstart\u201D or \u201Cend\u201D.",
    "cubic-bezier()": "Specifies a cubic-bezier curve. The four values specify points P1 and P2  of the curve as (x1, y1, x2, y2).",
    "cubic-bezier(0.6, -0.28, 0.735, 0.045)": "Ease-in Back. Overshoots.",
    "cubic-bezier(0.68, -0.55, 0.265, 1.55)": "Ease-in-out Back. Overshoots.",
    "cubic-bezier(0.175, 0.885, 0.32, 1.275)": "Ease-out Back. Overshoots.",
    "cubic-bezier(0.6, 0.04, 0.98, 0.335)": "Ease-in Circular. Based on half circle.",
    "cubic-bezier(0.785, 0.135, 0.15, 0.86)": "Ease-in-out Circular. Based on half circle.",
    "cubic-bezier(0.075, 0.82, 0.165, 1)": "Ease-out Circular. Based on half circle.",
    "cubic-bezier(0.55, 0.055, 0.675, 0.19)": "Ease-in Cubic. Based on power of three.",
    "cubic-bezier(0.645, 0.045, 0.355, 1)": "Ease-in-out Cubic. Based on power of three.",
    "cubic-bezier(0.215, 0.610, 0.355, 1)": "Ease-out Cubic. Based on power of three.",
    "cubic-bezier(0.95, 0.05, 0.795, 0.035)": "Ease-in Exponential. Based on two to the power ten.",
    "cubic-bezier(1, 0, 0, 1)": "Ease-in-out Exponential. Based on two to the power ten.",
    "cubic-bezier(0.19, 1, 0.22, 1)": "Ease-out Exponential. Based on two to the power ten.",
    "cubic-bezier(0.47, 0, 0.745, 0.715)": "Ease-in Sine.",
    "cubic-bezier(0.445, 0.05, 0.55, 0.95)": "Ease-in-out Sine.",
    "cubic-bezier(0.39, 0.575, 0.565, 1)": "Ease-out Sine.",
    "cubic-bezier(0.55, 0.085, 0.68, 0.53)": "Ease-in Quadratic. Based on power of two.",
    "cubic-bezier(0.455, 0.03, 0.515, 0.955)": "Ease-in-out Quadratic. Based on power of two.",
    "cubic-bezier(0.25, 0.46, 0.45, 0.94)": "Ease-out Quadratic. Based on power of two.",
    "cubic-bezier(0.895, 0.03, 0.685, 0.22)": "Ease-in Quartic. Based on power of four.",
    "cubic-bezier(0.77, 0, 0.175, 1)": "Ease-in-out Quartic. Based on power of four.",
    "cubic-bezier(0.165, 0.84, 0.44, 1)": "Ease-out Quartic. Based on power of four.",
    "cubic-bezier(0.755, 0.05, 0.855, 0.06)": "Ease-in Quintic. Based on power of five.",
    "cubic-bezier(0.86, 0, 0.07, 1)": "Ease-in-out Quintic. Based on power of five.",
    "cubic-bezier(0.23, 1, 0.320, 1)": "Ease-out Quintic. Based on power of five."
  };
  var basicShapeFunctions = {
    "circle()": "Defines a circle.",
    "ellipse()": "Defines an ellipse.",
    "inset()": "Defines an inset rectangle.",
    "polygon()": "Defines a polygon."
  };
  var units = {
    "length": ["em", "rem", "ex", "px", "cm", "mm", "in", "pt", "pc", "ch", "vw", "vh", "vmin", "vmax"],
    "angle": ["deg", "rad", "grad", "turn"],
    "time": ["ms", "s"],
    "frequency": ["Hz", "kHz"],
    "resolution": ["dpi", "dpcm", "dppx"],
    "percentage": ["%", "fr"]
  };
  var html5Tags = [
    "a",
    "abbr",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "bdi",
    "bdo",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "data",
    "datalist",
    "dd",
    "del",
    "details",
    "dfn",
    "dialog",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hgroup",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "input",
    "ins",
    "kbd",
    "keygen",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "map",
    "mark",
    "menu",
    "menuitem",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "param",
    "picture",
    "pre",
    "progress",
    "q",
    "rb",
    "rp",
    "rt",
    "rtc",
    "ruby",
    "s",
    "samp",
    "script",
    "section",
    "select",
    "small",
    "source",
    "span",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "table",
    "tbody",
    "td",
    "template",
    "textarea",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "track",
    "u",
    "ul",
    "const",
    "video",
    "wbr"
  ];
  var svgElements = [
    "circle",
    "clipPath",
    "cursor",
    "defs",
    "desc",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "foreignObject",
    "g",
    "hatch",
    "hatchpath",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "mesh",
    "meshpatch",
    "meshrow",
    "metadata",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "set",
    "solidcolor",
    "stop",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "tspan",
    "use",
    "view"
  ];
  var pageBoxDirectives = [
    "@bottom-center",
    "@bottom-left",
    "@bottom-left-corner",
    "@bottom-right",
    "@bottom-right-corner",
    "@left-bottom",
    "@left-middle",
    "@left-top",
    "@right-bottom",
    "@right-middle",
    "@right-top",
    "@top-center",
    "@top-left",
    "@top-left-corner",
    "@top-right",
    "@top-right-corner"
  ];

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/utils/objects.js
  function values(obj) {
    return Object.keys(obj).map(function(key) {
      return obj[key];
    });
  }
  function isDefined(obj) {
    return typeof obj !== "undefined";
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/cssParser.js
  var __spreadArrays = function() {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++)
      s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
      for (var a2 = arguments[i], j = 0, jl = a2.length; j < jl; j++, k++)
        r[k] = a2[j];
    return r;
  };
  var Parser = function() {
    function Parser2(scnr) {
      if (scnr === void 0) {
        scnr = new Scanner();
      }
      this.keyframeRegex = /^@(\-(webkit|ms|moz|o)\-)?keyframes$/i;
      this.scanner = scnr;
      this.token = { type: TokenType.EOF, offset: -1, len: 0, text: "" };
      this.prevToken = void 0;
    }
    Parser2.prototype.peekIdent = function(text) {
      return TokenType.Ident === this.token.type && text.length === this.token.text.length && text === this.token.text.toLowerCase();
    };
    Parser2.prototype.peekKeyword = function(text) {
      return TokenType.AtKeyword === this.token.type && text.length === this.token.text.length && text === this.token.text.toLowerCase();
    };
    Parser2.prototype.peekDelim = function(text) {
      return TokenType.Delim === this.token.type && text === this.token.text;
    };
    Parser2.prototype.peek = function(type) {
      return type === this.token.type;
    };
    Parser2.prototype.peekRegExp = function(type, regEx) {
      if (type !== this.token.type) {
        return false;
      }
      return regEx.test(this.token.text);
    };
    Parser2.prototype.hasWhitespace = function() {
      return !!this.prevToken && this.prevToken.offset + this.prevToken.len !== this.token.offset;
    };
    Parser2.prototype.consumeToken = function() {
      this.prevToken = this.token;
      this.token = this.scanner.scan();
    };
    Parser2.prototype.mark = function() {
      return {
        prev: this.prevToken,
        curr: this.token,
        pos: this.scanner.pos()
      };
    };
    Parser2.prototype.restoreAtMark = function(mark) {
      this.prevToken = mark.prev;
      this.token = mark.curr;
      this.scanner.goBackTo(mark.pos);
    };
    Parser2.prototype.try = function(func) {
      var pos = this.mark();
      var node = func();
      if (!node) {
        this.restoreAtMark(pos);
        return null;
      }
      return node;
    };
    Parser2.prototype.acceptOneKeyword = function(keywords) {
      if (TokenType.AtKeyword === this.token.type) {
        for (var _i = 0, keywords_1 = keywords; _i < keywords_1.length; _i++) {
          var keyword = keywords_1[_i];
          if (keyword.length === this.token.text.length && keyword === this.token.text.toLowerCase()) {
            this.consumeToken();
            return true;
          }
        }
      }
      return false;
    };
    Parser2.prototype.accept = function(type) {
      if (type === this.token.type) {
        this.consumeToken();
        return true;
      }
      return false;
    };
    Parser2.prototype.acceptIdent = function(text) {
      if (this.peekIdent(text)) {
        this.consumeToken();
        return true;
      }
      return false;
    };
    Parser2.prototype.acceptKeyword = function(text) {
      if (this.peekKeyword(text)) {
        this.consumeToken();
        return true;
      }
      return false;
    };
    Parser2.prototype.acceptDelim = function(text) {
      if (this.peekDelim(text)) {
        this.consumeToken();
        return true;
      }
      return false;
    };
    Parser2.prototype.acceptRegexp = function(regEx) {
      if (regEx.test(this.token.text)) {
        this.consumeToken();
        return true;
      }
      return false;
    };
    Parser2.prototype._parseRegexp = function(regEx) {
      var node = this.createNode(NodeType.Identifier);
      do {
      } while (this.acceptRegexp(regEx));
      return this.finish(node);
    };
    Parser2.prototype.acceptUnquotedString = function() {
      var pos = this.scanner.pos();
      this.scanner.goBackTo(this.token.offset);
      var unquoted = this.scanner.scanUnquotedString();
      if (unquoted) {
        this.token = unquoted;
        this.consumeToken();
        return true;
      }
      this.scanner.goBackTo(pos);
      return false;
    };
    Parser2.prototype.resync = function(resyncTokens, resyncStopTokens) {
      while (true) {
        if (resyncTokens && resyncTokens.indexOf(this.token.type) !== -1) {
          this.consumeToken();
          return true;
        } else if (resyncStopTokens && resyncStopTokens.indexOf(this.token.type) !== -1) {
          return true;
        } else {
          if (this.token.type === TokenType.EOF) {
            return false;
          }
          this.token = this.scanner.scan();
        }
      }
    };
    Parser2.prototype.createNode = function(nodeType) {
      return new Node2(this.token.offset, this.token.len, nodeType);
    };
    Parser2.prototype.create = function(ctor) {
      return new ctor(this.token.offset, this.token.len);
    };
    Parser2.prototype.finish = function(node, error, resyncTokens, resyncStopTokens) {
      if (!(node instanceof Nodelist)) {
        if (error) {
          this.markError(node, error, resyncTokens, resyncStopTokens);
        }
        if (this.prevToken) {
          var prevEnd = this.prevToken.offset + this.prevToken.len;
          node.length = prevEnd > node.offset ? prevEnd - node.offset : 0;
        }
      }
      return node;
    };
    Parser2.prototype.markError = function(node, error, resyncTokens, resyncStopTokens) {
      if (this.token !== this.lastErrorToken) {
        node.addIssue(new Marker(node, error, Level.Error, void 0, this.token.offset, this.token.len));
        this.lastErrorToken = this.token;
      }
      if (resyncTokens || resyncStopTokens) {
        this.resync(resyncTokens, resyncStopTokens);
      }
    };
    Parser2.prototype.parseStylesheet = function(textDocument) {
      var versionId = textDocument.version;
      var text = textDocument.getText();
      var textProvider = function(offset, length) {
        if (textDocument.version !== versionId) {
          throw new Error("Underlying model has changed, AST is no longer valid");
        }
        return text.substr(offset, length);
      };
      return this.internalParse(text, this._parseStylesheet, textProvider);
    };
    Parser2.prototype.internalParse = function(input, parseFunc, textProvider) {
      this.scanner.setSource(input);
      this.token = this.scanner.scan();
      var node = parseFunc.bind(this)();
      if (node) {
        if (textProvider) {
          node.textProvider = textProvider;
        } else {
          node.textProvider = function(offset, length) {
            return input.substr(offset, length);
          };
        }
      }
      return node;
    };
    Parser2.prototype._parseStylesheet = function() {
      var node = this.create(Stylesheet);
      while (node.addChild(this._parseStylesheetStart())) {
      }
      var inRecovery = false;
      do {
        var hasMatch = false;
        do {
          hasMatch = false;
          var statement = this._parseStylesheetStatement();
          if (statement) {
            node.addChild(statement);
            hasMatch = true;
            inRecovery = false;
            if (!this.peek(TokenType.EOF) && this._needsSemicolonAfter(statement) && !this.accept(TokenType.SemiColon)) {
              this.markError(node, ParseError.SemiColonExpected);
            }
          }
          while (this.accept(TokenType.SemiColon) || this.accept(TokenType.CDO) || this.accept(TokenType.CDC)) {
            hasMatch = true;
            inRecovery = false;
          }
        } while (hasMatch);
        if (this.peek(TokenType.EOF)) {
          break;
        }
        if (!inRecovery) {
          if (this.peek(TokenType.AtKeyword)) {
            this.markError(node, ParseError.UnknownAtRule);
          } else {
            this.markError(node, ParseError.RuleOrSelectorExpected);
          }
          inRecovery = true;
        }
        this.consumeToken();
      } while (!this.peek(TokenType.EOF));
      return this.finish(node);
    };
    Parser2.prototype._parseStylesheetStart = function() {
      return this._parseCharset();
    };
    Parser2.prototype._parseStylesheetStatement = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseStylesheetAtStatement(isNested);
      }
      return this._parseRuleset(isNested);
    };
    Parser2.prototype._parseStylesheetAtStatement = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      return this._parseImport() || this._parseMedia(isNested) || this._parsePage() || this._parseFontFace() || this._parseKeyframe() || this._parseSupports(isNested) || this._parseViewPort() || this._parseNamespace() || this._parseDocument() || this._parseUnknownAtRule();
    };
    Parser2.prototype._tryParseRuleset = function(isNested) {
      var mark = this.mark();
      if (this._parseSelector(isNested)) {
        while (this.accept(TokenType.Comma) && this._parseSelector(isNested)) {
        }
        if (this.accept(TokenType.CurlyL)) {
          this.restoreAtMark(mark);
          return this._parseRuleset(isNested);
        }
      }
      this.restoreAtMark(mark);
      return null;
    };
    Parser2.prototype._parseRuleset = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      var node = this.create(RuleSet);
      var selectors = node.getSelectors();
      if (!selectors.addChild(this._parseSelector(isNested))) {
        return null;
      }
      while (this.accept(TokenType.Comma)) {
        if (!selectors.addChild(this._parseSelector(isNested))) {
          return this.finish(node, ParseError.SelectorExpected);
        }
      }
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._parseRuleSetDeclarationAtStatement = function() {
      return this._parseAtApply() || this._parseUnknownAtRule();
    };
    Parser2.prototype._parseRuleSetDeclaration = function() {
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseRuleSetDeclarationAtStatement();
      }
      return this._tryParseCustomPropertyDeclaration() || this._parseDeclaration();
    };
    Parser2.prototype._parseAtApply = function() {
      if (!this.peekKeyword("@apply")) {
        return null;
      }
      var node = this.create(AtApplyRule);
      this.consumeToken();
      if (!node.setIdentifier(this._parseIdent([ReferenceType.Variable]))) {
        return this.finish(node, ParseError.IdentifierExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._needsSemicolonAfter = function(node) {
      switch (node.type) {
        case NodeType.Keyframe:
        case NodeType.ViewPort:
        case NodeType.Media:
        case NodeType.Ruleset:
        case NodeType.Namespace:
        case NodeType.If:
        case NodeType.For:
        case NodeType.Each:
        case NodeType.While:
        case NodeType.MixinDeclaration:
        case NodeType.FunctionDeclaration:
        case NodeType.MixinContentDeclaration:
          return false;
        case NodeType.ExtendsReference:
        case NodeType.MixinContentReference:
        case NodeType.ReturnStatement:
        case NodeType.MediaQuery:
        case NodeType.Debug:
        case NodeType.Import:
        case NodeType.AtApplyRule:
        case NodeType.CustomPropertyDeclaration:
          return true;
        case NodeType.VariableDeclaration:
          return node.needsSemicolon;
        case NodeType.MixinReference:
          return !node.getContent();
        case NodeType.Declaration:
          return !node.getNestedProperties();
      }
      return false;
    };
    Parser2.prototype._parseDeclarations = function(parseDeclaration) {
      var node = this.create(Declarations);
      if (!this.accept(TokenType.CurlyL)) {
        return null;
      }
      var decl = parseDeclaration();
      while (node.addChild(decl)) {
        if (this.peek(TokenType.CurlyR)) {
          break;
        }
        if (this._needsSemicolonAfter(decl) && !this.accept(TokenType.SemiColon)) {
          return this.finish(node, ParseError.SemiColonExpected, [TokenType.SemiColon, TokenType.CurlyR]);
        }
        if (decl && this.prevToken && this.prevToken.type === TokenType.SemiColon) {
          decl.semicolonPosition = this.prevToken.offset;
        }
        while (this.accept(TokenType.SemiColon)) {
        }
        decl = parseDeclaration();
      }
      if (!this.accept(TokenType.CurlyR)) {
        return this.finish(node, ParseError.RightCurlyExpected, [TokenType.CurlyR, TokenType.SemiColon]);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseBody = function(node, parseDeclaration) {
      if (!node.setDeclarations(this._parseDeclarations(parseDeclaration))) {
        return this.finish(node, ParseError.LeftCurlyExpected, [TokenType.CurlyR, TokenType.SemiColon]);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseSelector = function(isNested) {
      var node = this.create(Selector);
      var hasContent = false;
      if (isNested) {
        hasContent = node.addChild(this._parseCombinator());
      }
      while (node.addChild(this._parseSimpleSelector())) {
        hasContent = true;
        node.addChild(this._parseCombinator());
      }
      return hasContent ? this.finish(node) : null;
    };
    Parser2.prototype._parseDeclaration = function(resyncStopTokens) {
      var node = this.create(Declaration);
      if (!node.setProperty(this._parseProperty())) {
        return null;
      }
      if (!this.accept(TokenType.Colon)) {
        var stopTokens = resyncStopTokens ? __spreadArrays(resyncStopTokens, [TokenType.SemiColon]) : [TokenType.SemiColon];
        return this.finish(node, ParseError.ColonExpected, [TokenType.Colon], stopTokens);
      }
      if (this.prevToken) {
        node.colonPosition = this.prevToken.offset;
      }
      if (!node.setValue(this._parseExpr())) {
        return this.finish(node, ParseError.PropertyValueExpected);
      }
      node.addChild(this._parsePrio());
      if (this.peek(TokenType.SemiColon)) {
        node.semicolonPosition = this.token.offset;
      }
      return this.finish(node);
    };
    Parser2.prototype._tryParseCustomPropertyDeclaration = function() {
      if (!this.peekRegExp(TokenType.Ident, /^--/)) {
        return null;
      }
      var node = this.create(CustomPropertyDeclaration);
      if (!node.setProperty(this._parseProperty())) {
        return null;
      }
      if (!this.accept(TokenType.Colon)) {
        return this.finish(node, ParseError.ColonExpected, [TokenType.Colon]);
      }
      if (this.prevToken) {
        node.colonPosition = this.prevToken.offset;
      }
      var mark = this.mark();
      if (this.peek(TokenType.CurlyL)) {
        var propertySet = this.create(CustomPropertySet);
        var declarations = this._parseDeclarations(this._parseRuleSetDeclaration.bind(this));
        if (propertySet.setDeclarations(declarations) && !declarations.isErroneous(true)) {
          propertySet.addChild(this._parsePrio());
          if (this.peek(TokenType.SemiColon)) {
            this.finish(propertySet);
            node.setPropertySet(propertySet);
            node.semicolonPosition = this.token.offset;
            return this.finish(node);
          }
        }
        this.restoreAtMark(mark);
      }
      var expression = this._parseExpr();
      if (expression && !expression.isErroneous(true)) {
        this._parsePrio();
        if (this.peek(TokenType.SemiColon)) {
          node.setValue(expression);
          node.semicolonPosition = this.token.offset;
          return this.finish(node);
        }
      }
      this.restoreAtMark(mark);
      node.addChild(this._parseCustomPropertyValue());
      node.addChild(this._parsePrio());
      if (isDefined(node.colonPosition) && this.token.offset === node.colonPosition + 1) {
        return this.finish(node, ParseError.PropertyValueExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseCustomPropertyValue = function() {
      var node = this.create(Node2);
      var isTopLevel = function() {
        return curlyDepth === 0 && parensDepth === 0 && bracketsDepth === 0;
      };
      var curlyDepth = 0;
      var parensDepth = 0;
      var bracketsDepth = 0;
      done:
        while (true) {
          switch (this.token.type) {
            case TokenType.SemiColon:
              if (isTopLevel()) {
                break done;
              }
              break;
            case TokenType.Exclamation:
              if (isTopLevel()) {
                break done;
              }
              break;
            case TokenType.CurlyL:
              curlyDepth++;
              break;
            case TokenType.CurlyR:
              curlyDepth--;
              if (curlyDepth < 0) {
                if (parensDepth === 0 && bracketsDepth === 0) {
                  break done;
                }
                return this.finish(node, ParseError.LeftCurlyExpected);
              }
              break;
            case TokenType.ParenthesisL:
              parensDepth++;
              break;
            case TokenType.ParenthesisR:
              parensDepth--;
              if (parensDepth < 0) {
                return this.finish(node, ParseError.LeftParenthesisExpected);
              }
              break;
            case TokenType.BracketL:
              bracketsDepth++;
              break;
            case TokenType.BracketR:
              bracketsDepth--;
              if (bracketsDepth < 0) {
                return this.finish(node, ParseError.LeftSquareBracketExpected);
              }
              break;
            case TokenType.BadString:
              break done;
            case TokenType.EOF:
              var error = ParseError.RightCurlyExpected;
              if (bracketsDepth > 0) {
                error = ParseError.RightSquareBracketExpected;
              } else if (parensDepth > 0) {
                error = ParseError.RightParenthesisExpected;
              }
              return this.finish(node, error);
          }
          this.consumeToken();
        }
      return this.finish(node);
    };
    Parser2.prototype._tryToParseDeclaration = function() {
      var mark = this.mark();
      if (this._parseProperty() && this.accept(TokenType.Colon)) {
        this.restoreAtMark(mark);
        return this._parseDeclaration();
      }
      this.restoreAtMark(mark);
      return null;
    };
    Parser2.prototype._parseProperty = function() {
      var node = this.create(Property);
      var mark = this.mark();
      if (this.acceptDelim("*") || this.acceptDelim("_")) {
        if (this.hasWhitespace()) {
          this.restoreAtMark(mark);
          return null;
        }
      }
      if (node.setIdentifier(this._parsePropertyIdentifier())) {
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._parsePropertyIdentifier = function() {
      return this._parseIdent();
    };
    Parser2.prototype._parseCharset = function() {
      if (!this.peek(TokenType.Charset)) {
        return null;
      }
      var node = this.create(Node2);
      this.consumeToken();
      if (!this.accept(TokenType.String)) {
        return this.finish(node, ParseError.IdentifierExpected);
      }
      if (!this.accept(TokenType.SemiColon)) {
        return this.finish(node, ParseError.SemiColonExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseImport = function() {
      if (!this.peekKeyword("@import")) {
        return null;
      }
      var node = this.create(Import);
      this.consumeToken();
      if (!node.addChild(this._parseURILiteral()) && !node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.URIOrStringExpected);
      }
      if (!this.peek(TokenType.SemiColon) && !this.peek(TokenType.EOF)) {
        node.setMedialist(this._parseMediaQueryList());
      }
      return this.finish(node);
    };
    Parser2.prototype._parseNamespace = function() {
      if (!this.peekKeyword("@namespace")) {
        return null;
      }
      var node = this.create(Namespace);
      this.consumeToken();
      if (!node.addChild(this._parseURILiteral())) {
        node.addChild(this._parseIdent());
        if (!node.addChild(this._parseURILiteral()) && !node.addChild(this._parseStringLiteral())) {
          return this.finish(node, ParseError.URIExpected, [TokenType.SemiColon]);
        }
      }
      if (!this.accept(TokenType.SemiColon)) {
        return this.finish(node, ParseError.SemiColonExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseFontFace = function() {
      if (!this.peekKeyword("@font-face")) {
        return null;
      }
      var node = this.create(FontFace);
      this.consumeToken();
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._parseViewPort = function() {
      if (!this.peekKeyword("@-ms-viewport") && !this.peekKeyword("@-o-viewport") && !this.peekKeyword("@viewport")) {
        return null;
      }
      var node = this.create(ViewPort);
      this.consumeToken();
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._parseKeyframe = function() {
      if (!this.peekRegExp(TokenType.AtKeyword, this.keyframeRegex)) {
        return null;
      }
      var node = this.create(Keyframe);
      var atNode = this.create(Node2);
      this.consumeToken();
      node.setKeyword(this.finish(atNode));
      if (atNode.matches("@-ms-keyframes")) {
        this.markError(atNode, ParseError.UnknownKeyword);
      }
      if (!node.setIdentifier(this._parseKeyframeIdent())) {
        return this.finish(node, ParseError.IdentifierExpected, [TokenType.CurlyR]);
      }
      return this._parseBody(node, this._parseKeyframeSelector.bind(this));
    };
    Parser2.prototype._parseKeyframeIdent = function() {
      return this._parseIdent([ReferenceType.Keyframe]);
    };
    Parser2.prototype._parseKeyframeSelector = function() {
      var node = this.create(KeyframeSelector);
      if (!node.addChild(this._parseIdent()) && !this.accept(TokenType.Percentage)) {
        return null;
      }
      while (this.accept(TokenType.Comma)) {
        if (!node.addChild(this._parseIdent()) && !this.accept(TokenType.Percentage)) {
          return this.finish(node, ParseError.PercentageExpected);
        }
      }
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._tryParseKeyframeSelector = function() {
      var node = this.create(KeyframeSelector);
      var pos = this.mark();
      if (!node.addChild(this._parseIdent()) && !this.accept(TokenType.Percentage)) {
        return null;
      }
      while (this.accept(TokenType.Comma)) {
        if (!node.addChild(this._parseIdent()) && !this.accept(TokenType.Percentage)) {
          this.restoreAtMark(pos);
          return null;
        }
      }
      if (!this.peek(TokenType.CurlyL)) {
        this.restoreAtMark(pos);
        return null;
      }
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._parseSupports = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (!this.peekKeyword("@supports")) {
        return null;
      }
      var node = this.create(Supports);
      this.consumeToken();
      node.addChild(this._parseSupportsCondition());
      return this._parseBody(node, this._parseSupportsDeclaration.bind(this, isNested));
    };
    Parser2.prototype._parseSupportsDeclaration = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (isNested) {
        return this._tryParseRuleset(true) || this._tryToParseDeclaration() || this._parseStylesheetStatement(true);
      }
      return this._parseStylesheetStatement(false);
    };
    Parser2.prototype._parseSupportsCondition = function() {
      var node = this.create(SupportsCondition);
      if (this.acceptIdent("not")) {
        node.addChild(this._parseSupportsConditionInParens());
      } else {
        node.addChild(this._parseSupportsConditionInParens());
        if (this.peekRegExp(TokenType.Ident, /^(and|or)$/i)) {
          var text = this.token.text.toLowerCase();
          while (this.acceptIdent(text)) {
            node.addChild(this._parseSupportsConditionInParens());
          }
        }
      }
      return this.finish(node);
    };
    Parser2.prototype._parseSupportsConditionInParens = function() {
      var node = this.create(SupportsCondition);
      if (this.accept(TokenType.ParenthesisL)) {
        if (this.prevToken) {
          node.lParent = this.prevToken.offset;
        }
        if (!node.addChild(this._tryToParseDeclaration())) {
          if (!this._parseSupportsCondition()) {
            return this.finish(node, ParseError.ConditionExpected);
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected, [TokenType.ParenthesisR], []);
        }
        if (this.prevToken) {
          node.rParent = this.prevToken.offset;
        }
        return this.finish(node);
      } else if (this.peek(TokenType.Ident)) {
        var pos = this.mark();
        this.consumeToken();
        if (!this.hasWhitespace() && this.accept(TokenType.ParenthesisL)) {
          var openParentCount = 1;
          while (this.token.type !== TokenType.EOF && openParentCount !== 0) {
            if (this.token.type === TokenType.ParenthesisL) {
              openParentCount++;
            } else if (this.token.type === TokenType.ParenthesisR) {
              openParentCount--;
            }
            this.consumeToken();
          }
          return this.finish(node);
        } else {
          this.restoreAtMark(pos);
        }
      }
      return this.finish(node, ParseError.LeftParenthesisExpected, [], [TokenType.ParenthesisL]);
    };
    Parser2.prototype._parseMediaDeclaration = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (isNested) {
        return this._tryParseRuleset(true) || this._tryToParseDeclaration() || this._parseStylesheetStatement(true);
      }
      return this._parseStylesheetStatement(false);
    };
    Parser2.prototype._parseMedia = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (!this.peekKeyword("@media")) {
        return null;
      }
      var node = this.create(Media);
      this.consumeToken();
      if (!node.addChild(this._parseMediaQueryList())) {
        return this.finish(node, ParseError.MediaQueryExpected);
      }
      return this._parseBody(node, this._parseMediaDeclaration.bind(this, isNested));
    };
    Parser2.prototype._parseMediaQueryList = function() {
      var node = this.create(Medialist);
      if (!node.addChild(this._parseMediaQuery([TokenType.CurlyL]))) {
        return this.finish(node, ParseError.MediaQueryExpected);
      }
      while (this.accept(TokenType.Comma)) {
        if (!node.addChild(this._parseMediaQuery([TokenType.CurlyL]))) {
          return this.finish(node, ParseError.MediaQueryExpected);
        }
      }
      return this.finish(node);
    };
    Parser2.prototype._parseMediaQuery = function(resyncStopToken) {
      var node = this.create(MediaQuery);
      var parseExpression = true;
      var hasContent = false;
      if (!this.peek(TokenType.ParenthesisL)) {
        if (this.acceptIdent("only") || this.acceptIdent("not")) {
        }
        if (!node.addChild(this._parseIdent())) {
          return null;
        }
        hasContent = true;
        parseExpression = this.acceptIdent("and");
      }
      while (parseExpression) {
        if (node.addChild(this._parseMediaContentStart())) {
          parseExpression = this.acceptIdent("and");
          continue;
        }
        if (!this.accept(TokenType.ParenthesisL)) {
          if (hasContent) {
            return this.finish(node, ParseError.LeftParenthesisExpected, [], resyncStopToken);
          }
          return null;
        }
        if (!node.addChild(this._parseMediaFeatureName())) {
          return this.finish(node, ParseError.IdentifierExpected, [], resyncStopToken);
        }
        if (this.accept(TokenType.Colon)) {
          if (!node.addChild(this._parseExpr())) {
            return this.finish(node, ParseError.TermExpected, [], resyncStopToken);
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected, [], resyncStopToken);
        }
        parseExpression = this.acceptIdent("and");
      }
      return this.finish(node);
    };
    Parser2.prototype._parseMediaContentStart = function() {
      return null;
    };
    Parser2.prototype._parseMediaFeatureName = function() {
      return this._parseIdent();
    };
    Parser2.prototype._parseMedium = function() {
      var node = this.create(Node2);
      if (node.addChild(this._parseIdent())) {
        return this.finish(node);
      } else {
        return null;
      }
    };
    Parser2.prototype._parsePageDeclaration = function() {
      return this._parsePageMarginBox() || this._parseRuleSetDeclaration();
    };
    Parser2.prototype._parsePage = function() {
      if (!this.peekKeyword("@page")) {
        return null;
      }
      var node = this.create(Page);
      this.consumeToken();
      if (node.addChild(this._parsePageSelector())) {
        while (this.accept(TokenType.Comma)) {
          if (!node.addChild(this._parsePageSelector())) {
            return this.finish(node, ParseError.IdentifierExpected);
          }
        }
      }
      return this._parseBody(node, this._parsePageDeclaration.bind(this));
    };
    Parser2.prototype._parsePageMarginBox = function() {
      if (!this.peek(TokenType.AtKeyword)) {
        return null;
      }
      var node = this.create(PageBoxMarginBox);
      if (!this.acceptOneKeyword(pageBoxDirectives)) {
        this.markError(node, ParseError.UnknownAtRule, [], [TokenType.CurlyL]);
      }
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    Parser2.prototype._parsePageSelector = function() {
      if (!this.peek(TokenType.Ident) && !this.peek(TokenType.Colon)) {
        return null;
      }
      var node = this.create(Node2);
      node.addChild(this._parseIdent());
      if (this.accept(TokenType.Colon)) {
        if (!node.addChild(this._parseIdent())) {
          return this.finish(node, ParseError.IdentifierExpected);
        }
      }
      return this.finish(node);
    };
    Parser2.prototype._parseDocument = function() {
      if (!this.peekKeyword("@-moz-document")) {
        return null;
      }
      var node = this.create(Document);
      this.consumeToken();
      this.resync([], [TokenType.CurlyL]);
      return this._parseBody(node, this._parseStylesheetStatement.bind(this));
    };
    Parser2.prototype._parseUnknownAtRule = function() {
      if (!this.peek(TokenType.AtKeyword)) {
        return null;
      }
      var node = this.create(UnknownAtRule);
      node.addChild(this._parseUnknownAtRuleName());
      var isTopLevel = function() {
        return curlyDepth === 0 && parensDepth === 0 && bracketsDepth === 0;
      };
      var curlyLCount = 0;
      var curlyDepth = 0;
      var parensDepth = 0;
      var bracketsDepth = 0;
      done:
        while (true) {
          switch (this.token.type) {
            case TokenType.SemiColon:
              if (isTopLevel()) {
                break done;
              }
              break;
            case TokenType.EOF:
              if (curlyDepth > 0) {
                return this.finish(node, ParseError.RightCurlyExpected);
              } else if (bracketsDepth > 0) {
                return this.finish(node, ParseError.RightSquareBracketExpected);
              } else if (parensDepth > 0) {
                return this.finish(node, ParseError.RightParenthesisExpected);
              } else {
                return this.finish(node);
              }
            case TokenType.CurlyL:
              curlyLCount++;
              curlyDepth++;
              break;
            case TokenType.CurlyR:
              curlyDepth--;
              if (curlyLCount > 0 && curlyDepth === 0) {
                this.consumeToken();
                if (bracketsDepth > 0) {
                  return this.finish(node, ParseError.RightSquareBracketExpected);
                } else if (parensDepth > 0) {
                  return this.finish(node, ParseError.RightParenthesisExpected);
                }
                break done;
              }
              if (curlyDepth < 0) {
                if (parensDepth === 0 && bracketsDepth === 0) {
                  break done;
                }
                return this.finish(node, ParseError.LeftCurlyExpected);
              }
              break;
            case TokenType.ParenthesisL:
              parensDepth++;
              break;
            case TokenType.ParenthesisR:
              parensDepth--;
              if (parensDepth < 0) {
                return this.finish(node, ParseError.LeftParenthesisExpected);
              }
              break;
            case TokenType.BracketL:
              bracketsDepth++;
              break;
            case TokenType.BracketR:
              bracketsDepth--;
              if (bracketsDepth < 0) {
                return this.finish(node, ParseError.LeftSquareBracketExpected);
              }
              break;
          }
          this.consumeToken();
        }
      return node;
    };
    Parser2.prototype._parseUnknownAtRuleName = function() {
      var node = this.create(Node2);
      if (this.accept(TokenType.AtKeyword)) {
        return this.finish(node);
      }
      return node;
    };
    Parser2.prototype._parseOperator = function() {
      if (this.peekDelim("/") || this.peekDelim("*") || this.peekDelim("+") || this.peekDelim("-") || this.peek(TokenType.Dashmatch) || this.peek(TokenType.Includes) || this.peek(TokenType.SubstringOperator) || this.peek(TokenType.PrefixOperator) || this.peek(TokenType.SuffixOperator) || this.peekDelim("=")) {
        var node = this.createNode(NodeType.Operator);
        this.consumeToken();
        return this.finish(node);
      } else {
        return null;
      }
    };
    Parser2.prototype._parseUnaryOperator = function() {
      if (!this.peekDelim("+") && !this.peekDelim("-")) {
        return null;
      }
      var node = this.create(Node2);
      this.consumeToken();
      return this.finish(node);
    };
    Parser2.prototype._parseCombinator = function() {
      if (this.peekDelim(">")) {
        var node = this.create(Node2);
        this.consumeToken();
        var mark = this.mark();
        if (!this.hasWhitespace() && this.acceptDelim(">")) {
          if (!this.hasWhitespace() && this.acceptDelim(">")) {
            node.type = NodeType.SelectorCombinatorShadowPiercingDescendant;
            return this.finish(node);
          }
          this.restoreAtMark(mark);
        }
        node.type = NodeType.SelectorCombinatorParent;
        return this.finish(node);
      } else if (this.peekDelim("+")) {
        var node = this.create(Node2);
        this.consumeToken();
        node.type = NodeType.SelectorCombinatorSibling;
        return this.finish(node);
      } else if (this.peekDelim("~")) {
        var node = this.create(Node2);
        this.consumeToken();
        node.type = NodeType.SelectorCombinatorAllSiblings;
        return this.finish(node);
      } else if (this.peekDelim("/")) {
        var node = this.create(Node2);
        this.consumeToken();
        var mark = this.mark();
        if (!this.hasWhitespace() && this.acceptIdent("deep") && !this.hasWhitespace() && this.acceptDelim("/")) {
          node.type = NodeType.SelectorCombinatorShadowPiercingDescendant;
          return this.finish(node);
        }
        this.restoreAtMark(mark);
      }
      return null;
    };
    Parser2.prototype._parseSimpleSelector = function() {
      var node = this.create(SimpleSelector);
      var c = 0;
      if (node.addChild(this._parseElementName())) {
        c++;
      }
      while ((c === 0 || !this.hasWhitespace()) && node.addChild(this._parseSimpleSelectorBody())) {
        c++;
      }
      return c > 0 ? this.finish(node) : null;
    };
    Parser2.prototype._parseSimpleSelectorBody = function() {
      return this._parsePseudo() || this._parseHash() || this._parseClass() || this._parseAttrib();
    };
    Parser2.prototype._parseSelectorIdent = function() {
      return this._parseIdent();
    };
    Parser2.prototype._parseHash = function() {
      if (!this.peek(TokenType.Hash) && !this.peekDelim("#")) {
        return null;
      }
      var node = this.createNode(NodeType.IdentifierSelector);
      if (this.acceptDelim("#")) {
        if (this.hasWhitespace() || !node.addChild(this._parseSelectorIdent())) {
          return this.finish(node, ParseError.IdentifierExpected);
        }
      } else {
        this.consumeToken();
      }
      return this.finish(node);
    };
    Parser2.prototype._parseClass = function() {
      if (!this.peekDelim(".")) {
        return null;
      }
      var node = this.createNode(NodeType.ClassSelector);
      this.consumeToken();
      if (this.hasWhitespace() || !node.addChild(this._parseSelectorIdent())) {
        return this.finish(node, ParseError.IdentifierExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseElementName = function() {
      var pos = this.mark();
      var node = this.createNode(NodeType.ElementNameSelector);
      node.addChild(this._parseNamespacePrefix());
      if (!node.addChild(this._parseSelectorIdent()) && !this.acceptDelim("*")) {
        this.restoreAtMark(pos);
        return null;
      }
      return this.finish(node);
    };
    Parser2.prototype._parseNamespacePrefix = function() {
      var pos = this.mark();
      var node = this.createNode(NodeType.NamespacePrefix);
      if (!node.addChild(this._parseIdent()) && !this.acceptDelim("*")) {
      }
      if (!this.acceptDelim("|")) {
        this.restoreAtMark(pos);
        return null;
      }
      return this.finish(node);
    };
    Parser2.prototype._parseAttrib = function() {
      if (!this.peek(TokenType.BracketL)) {
        return null;
      }
      var node = this.create(AttributeSelector);
      this.consumeToken();
      node.setNamespacePrefix(this._parseNamespacePrefix());
      if (!node.setIdentifier(this._parseIdent())) {
        return this.finish(node, ParseError.IdentifierExpected);
      }
      if (node.setOperator(this._parseOperator())) {
        node.setValue(this._parseBinaryExpr());
        this.acceptIdent("i");
      }
      if (!this.accept(TokenType.BracketR)) {
        return this.finish(node, ParseError.RightSquareBracketExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parsePseudo = function() {
      var _this = this;
      var node = this._tryParsePseudoIdentifier();
      if (node) {
        if (!this.hasWhitespace() && this.accept(TokenType.ParenthesisL)) {
          var tryAsSelector = function() {
            var selectors = _this.create(Node2);
            if (!selectors.addChild(_this._parseSelector(false))) {
              return null;
            }
            while (_this.accept(TokenType.Comma) && selectors.addChild(_this._parseSelector(false))) {
            }
            if (_this.peek(TokenType.ParenthesisR)) {
              return _this.finish(selectors);
            }
            return null;
          };
          node.addChild(this.try(tryAsSelector) || this._parseBinaryExpr());
          if (!this.accept(TokenType.ParenthesisR)) {
            return this.finish(node, ParseError.RightParenthesisExpected);
          }
        }
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._tryParsePseudoIdentifier = function() {
      if (!this.peek(TokenType.Colon)) {
        return null;
      }
      var pos = this.mark();
      var node = this.createNode(NodeType.PseudoSelector);
      this.consumeToken();
      if (this.hasWhitespace()) {
        this.restoreAtMark(pos);
        return null;
      }
      if (this.accept(TokenType.Colon) && this.hasWhitespace()) {
        this.markError(node, ParseError.IdentifierExpected);
      }
      if (!node.addChild(this._parseIdent())) {
        this.markError(node, ParseError.IdentifierExpected);
      }
      return node;
    };
    Parser2.prototype._tryParsePrio = function() {
      var mark = this.mark();
      var prio = this._parsePrio();
      if (prio) {
        return prio;
      }
      this.restoreAtMark(mark);
      return null;
    };
    Parser2.prototype._parsePrio = function() {
      if (!this.peek(TokenType.Exclamation)) {
        return null;
      }
      var node = this.createNode(NodeType.Prio);
      if (this.accept(TokenType.Exclamation) && this.acceptIdent("important")) {
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._parseExpr = function(stopOnComma) {
      if (stopOnComma === void 0) {
        stopOnComma = false;
      }
      var node = this.create(Expression);
      if (!node.addChild(this._parseBinaryExpr())) {
        return null;
      }
      while (true) {
        if (this.peek(TokenType.Comma)) {
          if (stopOnComma) {
            return this.finish(node);
          }
          this.consumeToken();
        }
        if (!node.addChild(this._parseBinaryExpr())) {
          break;
        }
      }
      return this.finish(node);
    };
    Parser2.prototype._parseNamedLine = function() {
      if (!this.peek(TokenType.BracketL)) {
        return null;
      }
      var node = this.createNode(NodeType.GridLine);
      this.consumeToken();
      while (node.addChild(this._parseIdent())) {
      }
      if (!this.accept(TokenType.BracketR)) {
        return this.finish(node, ParseError.RightSquareBracketExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseBinaryExpr = function(preparsedLeft, preparsedOper) {
      var node = this.create(BinaryExpression);
      if (!node.setLeft(preparsedLeft || this._parseTerm())) {
        return null;
      }
      if (!node.setOperator(preparsedOper || this._parseOperator())) {
        return this.finish(node);
      }
      if (!node.setRight(this._parseTerm())) {
        return this.finish(node, ParseError.TermExpected);
      }
      node = this.finish(node);
      var operator = this._parseOperator();
      if (operator) {
        node = this._parseBinaryExpr(node, operator);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseTerm = function() {
      var node = this.create(Term);
      node.setOperator(this._parseUnaryOperator());
      if (node.setExpression(this._parseTermExpression())) {
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._parseTermExpression = function() {
      return this._parseURILiteral() || this._parseFunction() || this._parseIdent() || this._parseStringLiteral() || this._parseNumeric() || this._parseHexColor() || this._parseOperation() || this._parseNamedLine();
    };
    Parser2.prototype._parseOperation = function() {
      if (!this.peek(TokenType.ParenthesisL)) {
        return null;
      }
      var node = this.create(Node2);
      this.consumeToken();
      node.addChild(this._parseExpr());
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseNumeric = function() {
      if (this.peek(TokenType.Num) || this.peek(TokenType.Percentage) || this.peek(TokenType.Resolution) || this.peek(TokenType.Length) || this.peek(TokenType.EMS) || this.peek(TokenType.EXS) || this.peek(TokenType.Angle) || this.peek(TokenType.Time) || this.peek(TokenType.Dimension) || this.peek(TokenType.Freq)) {
        var node = this.create(NumericValue);
        this.consumeToken();
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._parseStringLiteral = function() {
      if (!this.peek(TokenType.String) && !this.peek(TokenType.BadString)) {
        return null;
      }
      var node = this.createNode(NodeType.StringLiteral);
      this.consumeToken();
      return this.finish(node);
    };
    Parser2.prototype._parseURILiteral = function() {
      if (!this.peekRegExp(TokenType.Ident, /^url(-prefix)?$/i)) {
        return null;
      }
      var pos = this.mark();
      var node = this.createNode(NodeType.URILiteral);
      this.accept(TokenType.Ident);
      if (this.hasWhitespace() || !this.peek(TokenType.ParenthesisL)) {
        this.restoreAtMark(pos);
        return null;
      }
      this.scanner.inURL = true;
      this.consumeToken();
      node.addChild(this._parseURLArgument());
      this.scanner.inURL = false;
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseURLArgument = function() {
      var node = this.create(Node2);
      if (!this.accept(TokenType.String) && !this.accept(TokenType.BadString) && !this.acceptUnquotedString()) {
        return null;
      }
      return this.finish(node);
    };
    Parser2.prototype._parseIdent = function(referenceTypes) {
      if (!this.peek(TokenType.Ident)) {
        return null;
      }
      var node = this.create(Identifier);
      if (referenceTypes) {
        node.referenceTypes = referenceTypes;
      }
      node.isCustomProperty = this.peekRegExp(TokenType.Ident, /^--/);
      this.consumeToken();
      return this.finish(node);
    };
    Parser2.prototype._parseFunction = function() {
      var pos = this.mark();
      var node = this.create(Function);
      if (!node.setIdentifier(this._parseFunctionIdentifier())) {
        return null;
      }
      if (this.hasWhitespace() || !this.accept(TokenType.ParenthesisL)) {
        this.restoreAtMark(pos);
        return null;
      }
      if (node.getArguments().addChild(this._parseFunctionArgument())) {
        while (this.accept(TokenType.Comma)) {
          if (this.peek(TokenType.ParenthesisR)) {
            break;
          }
          if (!node.getArguments().addChild(this._parseFunctionArgument())) {
            this.markError(node, ParseError.ExpressionExpected);
          }
        }
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    Parser2.prototype._parseFunctionIdentifier = function() {
      if (!this.peek(TokenType.Ident)) {
        return null;
      }
      var node = this.create(Identifier);
      node.referenceTypes = [ReferenceType.Function];
      if (this.acceptIdent("progid")) {
        if (this.accept(TokenType.Colon)) {
          while (this.accept(TokenType.Ident) && this.acceptDelim(".")) {
          }
        }
        return this.finish(node);
      }
      this.consumeToken();
      return this.finish(node);
    };
    Parser2.prototype._parseFunctionArgument = function() {
      var node = this.create(FunctionArgument);
      if (node.setValue(this._parseExpr(true))) {
        return this.finish(node);
      }
      return null;
    };
    Parser2.prototype._parseHexColor = function() {
      if (this.peekRegExp(TokenType.Hash, /^#([A-Fa-f0-9]{3}|[A-Fa-f0-9]{4}|[A-Fa-f0-9]{6}|[A-Fa-f0-9]{8})$/g)) {
        var node = this.create(HexColorValue);
        this.consumeToken();
        return this.finish(node);
      } else {
        return null;
      }
    };
    return Parser2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/utils/arrays.js
  function findFirst(array, p) {
    var low = 0, high = array.length;
    if (high === 0) {
      return 0;
    }
    while (low < high) {
      var mid = Math.floor((low + high) / 2);
      if (p(array[mid])) {
        high = mid;
      } else {
        low = mid + 1;
      }
    }
    return low;
  }
  function includes(array, item) {
    return array.indexOf(item) !== -1;
  }
  function union() {
    var arrays = [];
    for (var _i = 0; _i < arguments.length; _i++) {
      arrays[_i] = arguments[_i];
    }
    var result = [];
    for (var _a3 = 0, arrays_1 = arrays; _a3 < arrays_1.length; _a3++) {
      var array = arrays_1[_a3];
      for (var _b = 0, array_1 = array; _b < array_1.length; _b++) {
        var item = array_1[_b];
        if (!includes(result, item)) {
          result.push(item);
        }
      }
    }
    return result;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/cssSymbolScope.js
  var __extends2 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var Scope = function() {
    function Scope2(offset, length) {
      this.offset = offset;
      this.length = length;
      this.symbols = [];
      this.parent = null;
      this.children = [];
    }
    Scope2.prototype.addChild = function(scope) {
      this.children.push(scope);
      scope.setParent(this);
    };
    Scope2.prototype.setParent = function(scope) {
      this.parent = scope;
    };
    Scope2.prototype.findScope = function(offset, length) {
      if (length === void 0) {
        length = 0;
      }
      if (this.offset <= offset && this.offset + this.length > offset + length || this.offset === offset && this.length === length) {
        return this.findInScope(offset, length);
      }
      return null;
    };
    Scope2.prototype.findInScope = function(offset, length) {
      if (length === void 0) {
        length = 0;
      }
      var end = offset + length;
      var idx = findFirst(this.children, function(s) {
        return s.offset > end;
      });
      if (idx === 0) {
        return this;
      }
      var res = this.children[idx - 1];
      if (res.offset <= offset && res.offset + res.length >= offset + length) {
        return res.findInScope(offset, length);
      }
      return this;
    };
    Scope2.prototype.addSymbol = function(symbol) {
      this.symbols.push(symbol);
    };
    Scope2.prototype.getSymbol = function(name, type) {
      for (var index = 0; index < this.symbols.length; index++) {
        var symbol = this.symbols[index];
        if (symbol.name === name && symbol.type === type) {
          return symbol;
        }
      }
      return null;
    };
    Scope2.prototype.getSymbols = function() {
      return this.symbols;
    };
    return Scope2;
  }();
  var GlobalScope = function(_super) {
    __extends2(GlobalScope2, _super);
    function GlobalScope2() {
      return _super.call(this, 0, Number.MAX_VALUE) || this;
    }
    return GlobalScope2;
  }(Scope);
  var Symbol2 = function() {
    function Symbol3(name, value, node, type) {
      this.name = name;
      this.value = value;
      this.node = node;
      this.type = type;
    }
    return Symbol3;
  }();
  var ScopeBuilder = function() {
    function ScopeBuilder2(scope) {
      this.scope = scope;
    }
    ScopeBuilder2.prototype.addSymbol = function(node, name, value, type) {
      if (node.offset !== -1) {
        var current = this.scope.findScope(node.offset, node.length);
        if (current) {
          current.addSymbol(new Symbol2(name, value, node, type));
        }
      }
    };
    ScopeBuilder2.prototype.addScope = function(node) {
      if (node.offset !== -1) {
        var current = this.scope.findScope(node.offset, node.length);
        if (current && (current.offset !== node.offset || current.length !== node.length)) {
          var newScope = new Scope(node.offset, node.length);
          current.addChild(newScope);
          return newScope;
        }
        return current;
      }
      return null;
    };
    ScopeBuilder2.prototype.addSymbolToChildScope = function(scopeNode, node, name, value, type) {
      if (scopeNode && scopeNode.offset !== -1) {
        var current = this.addScope(scopeNode);
        if (current) {
          current.addSymbol(new Symbol2(name, value, node, type));
        }
      }
    };
    ScopeBuilder2.prototype.visitNode = function(node) {
      switch (node.type) {
        case NodeType.Keyframe:
          this.addSymbol(node, node.getName(), void 0, ReferenceType.Keyframe);
          return true;
        case NodeType.CustomPropertyDeclaration:
          return this.visitCustomPropertyDeclarationNode(node);
        case NodeType.VariableDeclaration:
          return this.visitVariableDeclarationNode(node);
        case NodeType.Ruleset:
          return this.visitRuleSet(node);
        case NodeType.MixinDeclaration:
          this.addSymbol(node, node.getName(), void 0, ReferenceType.Mixin);
          return true;
        case NodeType.FunctionDeclaration:
          this.addSymbol(node, node.getName(), void 0, ReferenceType.Function);
          return true;
        case NodeType.FunctionParameter: {
          return this.visitFunctionParameterNode(node);
        }
        case NodeType.Declarations:
          this.addScope(node);
          return true;
        case NodeType.For:
          var forNode = node;
          var scopeNode = forNode.getDeclarations();
          if (scopeNode && forNode.variable) {
            this.addSymbolToChildScope(scopeNode, forNode.variable, forNode.variable.getName(), void 0, ReferenceType.Variable);
          }
          return true;
        case NodeType.Each: {
          var eachNode = node;
          var scopeNode_1 = eachNode.getDeclarations();
          if (scopeNode_1) {
            var variables = eachNode.getVariables().getChildren();
            for (var _i = 0, variables_1 = variables; _i < variables_1.length; _i++) {
              var variable = variables_1[_i];
              this.addSymbolToChildScope(scopeNode_1, variable, variable.getName(), void 0, ReferenceType.Variable);
            }
          }
          return true;
        }
      }
      return true;
    };
    ScopeBuilder2.prototype.visitRuleSet = function(node) {
      var current = this.scope.findScope(node.offset, node.length);
      if (current) {
        for (var _i = 0, _a3 = node.getSelectors().getChildren(); _i < _a3.length; _i++) {
          var child = _a3[_i];
          if (child instanceof Selector) {
            if (child.getChildren().length === 1) {
              current.addSymbol(new Symbol2(child.getChild(0).getText(), void 0, child, ReferenceType.Rule));
            }
          }
        }
      }
      return true;
    };
    ScopeBuilder2.prototype.visitVariableDeclarationNode = function(node) {
      var value = node.getValue() ? node.getValue().getText() : void 0;
      this.addSymbol(node, node.getName(), value, ReferenceType.Variable);
      return true;
    };
    ScopeBuilder2.prototype.visitFunctionParameterNode = function(node) {
      var scopeNode = node.getParent().getDeclarations();
      if (scopeNode) {
        var valueNode = node.getDefaultValue();
        var value = valueNode ? valueNode.getText() : void 0;
        this.addSymbolToChildScope(scopeNode, node, node.getName(), value, ReferenceType.Variable);
      }
      return true;
    };
    ScopeBuilder2.prototype.visitCustomPropertyDeclarationNode = function(node) {
      var value = node.getValue() ? node.getValue().getText() : "";
      this.addCSSVariable(node.getProperty(), node.getProperty().getName(), value, ReferenceType.Variable);
      return true;
    };
    ScopeBuilder2.prototype.addCSSVariable = function(node, name, value, type) {
      if (node.offset !== -1) {
        this.scope.addSymbol(new Symbol2(name, value, node, type));
      }
    };
    return ScopeBuilder2;
  }();
  var Symbols = function() {
    function Symbols2(node) {
      this.global = new GlobalScope();
      node.acceptVisitor(new ScopeBuilder(this.global));
    }
    Symbols2.prototype.findSymbolsAtOffset = function(offset, referenceType) {
      var scope = this.global.findScope(offset, 0);
      var result = [];
      var names = {};
      while (scope) {
        var symbols = scope.getSymbols();
        for (var i = 0; i < symbols.length; i++) {
          var symbol = symbols[i];
          if (symbol.type === referenceType && !names[symbol.name]) {
            result.push(symbol);
            names[symbol.name] = true;
          }
        }
        scope = scope.parent;
      }
      return result;
    };
    Symbols2.prototype.internalFindSymbol = function(node, referenceTypes) {
      var scopeNode = node;
      if (node.parent instanceof FunctionParameter && node.parent.getParent() instanceof BodyDeclaration) {
        scopeNode = node.parent.getParent().getDeclarations();
      }
      if (node.parent instanceof FunctionArgument && node.parent.getParent() instanceof Function) {
        var funcId = node.parent.getParent().getIdentifier();
        if (funcId) {
          var functionSymbol = this.internalFindSymbol(funcId, [ReferenceType.Function]);
          if (functionSymbol) {
            scopeNode = functionSymbol.node.getDeclarations();
          }
        }
      }
      if (!scopeNode) {
        return null;
      }
      var name = node.getText();
      var scope = this.global.findScope(scopeNode.offset, scopeNode.length);
      while (scope) {
        for (var index = 0; index < referenceTypes.length; index++) {
          var type = referenceTypes[index];
          var symbol = scope.getSymbol(name, type);
          if (symbol) {
            return symbol;
          }
        }
        scope = scope.parent;
      }
      return null;
    };
    Symbols2.prototype.evaluateReferenceTypes = function(node) {
      if (node instanceof Identifier) {
        var referenceTypes = node.referenceTypes;
        if (referenceTypes) {
          return referenceTypes;
        } else {
          if (node.isCustomProperty) {
            return [ReferenceType.Variable];
          }
          var decl = getParentDeclaration(node);
          if (decl) {
            var propertyName = decl.getNonPrefixedPropertyName();
            if ((propertyName === "animation" || propertyName === "animation-name") && decl.getValue() && decl.getValue().offset === node.offset) {
              return [ReferenceType.Keyframe];
            }
          }
        }
      } else if (node instanceof Variable) {
        return [ReferenceType.Variable];
      }
      var selector = node.findAParent(NodeType.Selector, NodeType.ExtendsReference);
      if (selector) {
        return [ReferenceType.Rule];
      }
      return null;
    };
    Symbols2.prototype.findSymbolFromNode = function(node) {
      if (!node) {
        return null;
      }
      while (node.type === NodeType.Interpolation) {
        node = node.getParent();
      }
      var referenceTypes = this.evaluateReferenceTypes(node);
      if (referenceTypes) {
        return this.internalFindSymbol(node, referenceTypes);
      }
      return null;
    };
    Symbols2.prototype.matchesSymbol = function(node, symbol) {
      if (!node) {
        return false;
      }
      while (node.type === NodeType.Interpolation) {
        node = node.getParent();
      }
      if (!node.matches(symbol.name)) {
        return false;
      }
      var referenceTypes = this.evaluateReferenceTypes(node);
      if (!referenceTypes || referenceTypes.indexOf(symbol.type) === -1) {
        return false;
      }
      var nodeSymbol = this.internalFindSymbol(node, referenceTypes);
      return nodeSymbol === symbol;
    };
    Symbols2.prototype.findSymbol = function(name, type, offset) {
      var scope = this.global.findScope(offset);
      while (scope) {
        var symbol = scope.getSymbol(name, type);
        if (symbol) {
          return symbol;
        }
        scope = scope.parent;
      }
      return null;
    };
    return Symbols2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/_deps/vscode-languageserver-types/main.js
  var Position2;
  (function(Position3) {
    function create(line, character) {
      return { line, character };
    }
    Position3.create = create;
    function is(value) {
      var candidate = value;
      return Is.objectLiteral(candidate) && Is.number(candidate.line) && Is.number(candidate.character);
    }
    Position3.is = is;
  })(Position2 || (Position2 = {}));
  var Range2;
  (function(Range3) {
    function create(one, two, three, four) {
      if (Is.number(one) && Is.number(two) && Is.number(three) && Is.number(four)) {
        return { start: Position2.create(one, two), end: Position2.create(three, four) };
      } else if (Position2.is(one) && Position2.is(two)) {
        return { start: one, end: two };
      } else {
        throw new Error("Range#create called with invalid arguments[" + one + ", " + two + ", " + three + ", " + four + "]");
      }
    }
    Range3.create = create;
    function is(value) {
      var candidate = value;
      return Is.objectLiteral(candidate) && Position2.is(candidate.start) && Position2.is(candidate.end);
    }
    Range3.is = is;
  })(Range2 || (Range2 = {}));
  var Location;
  (function(Location2) {
    function create(uri, range) {
      return { uri, range };
    }
    Location2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Range2.is(candidate.range) && (Is.string(candidate.uri) || Is.undefined(candidate.uri));
    }
    Location2.is = is;
  })(Location || (Location = {}));
  var LocationLink;
  (function(LocationLink2) {
    function create(targetUri, targetRange, targetSelectionRange, originSelectionRange) {
      return { targetUri, targetRange, targetSelectionRange, originSelectionRange };
    }
    LocationLink2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Range2.is(candidate.targetRange) && Is.string(candidate.targetUri) && (Range2.is(candidate.targetSelectionRange) || Is.undefined(candidate.targetSelectionRange)) && (Range2.is(candidate.originSelectionRange) || Is.undefined(candidate.originSelectionRange));
    }
    LocationLink2.is = is;
  })(LocationLink || (LocationLink = {}));
  var Color;
  (function(Color2) {
    function create(red, green, blue, alpha) {
      return {
        red,
        green,
        blue,
        alpha
      };
    }
    Color2.create = create;
    function is(value) {
      var candidate = value;
      return Is.number(candidate.red) && Is.number(candidate.green) && Is.number(candidate.blue) && Is.number(candidate.alpha);
    }
    Color2.is = is;
  })(Color || (Color = {}));
  var ColorInformation;
  (function(ColorInformation2) {
    function create(range, color) {
      return {
        range,
        color
      };
    }
    ColorInformation2.create = create;
    function is(value) {
      var candidate = value;
      return Range2.is(candidate.range) && Color.is(candidate.color);
    }
    ColorInformation2.is = is;
  })(ColorInformation || (ColorInformation = {}));
  var ColorPresentation;
  (function(ColorPresentation2) {
    function create(label, textEdit, additionalTextEdits) {
      return {
        label,
        textEdit,
        additionalTextEdits
      };
    }
    ColorPresentation2.create = create;
    function is(value) {
      var candidate = value;
      return Is.string(candidate.label) && (Is.undefined(candidate.textEdit) || TextEdit.is(candidate)) && (Is.undefined(candidate.additionalTextEdits) || Is.typedArray(candidate.additionalTextEdits, TextEdit.is));
    }
    ColorPresentation2.is = is;
  })(ColorPresentation || (ColorPresentation = {}));
  var FoldingRangeKind;
  (function(FoldingRangeKind2) {
    FoldingRangeKind2["Comment"] = "comment";
    FoldingRangeKind2["Imports"] = "imports";
    FoldingRangeKind2["Region"] = "region";
  })(FoldingRangeKind || (FoldingRangeKind = {}));
  var FoldingRange;
  (function(FoldingRange2) {
    function create(startLine, endLine, startCharacter, endCharacter, kind) {
      var result = {
        startLine,
        endLine
      };
      if (Is.defined(startCharacter)) {
        result.startCharacter = startCharacter;
      }
      if (Is.defined(endCharacter)) {
        result.endCharacter = endCharacter;
      }
      if (Is.defined(kind)) {
        result.kind = kind;
      }
      return result;
    }
    FoldingRange2.create = create;
    function is(value) {
      var candidate = value;
      return Is.number(candidate.startLine) && Is.number(candidate.startLine) && (Is.undefined(candidate.startCharacter) || Is.number(candidate.startCharacter)) && (Is.undefined(candidate.endCharacter) || Is.number(candidate.endCharacter)) && (Is.undefined(candidate.kind) || Is.string(candidate.kind));
    }
    FoldingRange2.is = is;
  })(FoldingRange || (FoldingRange = {}));
  var DiagnosticRelatedInformation;
  (function(DiagnosticRelatedInformation2) {
    function create(location, message) {
      return {
        location,
        message
      };
    }
    DiagnosticRelatedInformation2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Location.is(candidate.location) && Is.string(candidate.message);
    }
    DiagnosticRelatedInformation2.is = is;
  })(DiagnosticRelatedInformation || (DiagnosticRelatedInformation = {}));
  var DiagnosticSeverity;
  (function(DiagnosticSeverity2) {
    DiagnosticSeverity2.Error = 1;
    DiagnosticSeverity2.Warning = 2;
    DiagnosticSeverity2.Information = 3;
    DiagnosticSeverity2.Hint = 4;
  })(DiagnosticSeverity || (DiagnosticSeverity = {}));
  var DiagnosticTag;
  (function(DiagnosticTag2) {
    DiagnosticTag2.Unnecessary = 1;
    DiagnosticTag2.Deprecated = 2;
  })(DiagnosticTag || (DiagnosticTag = {}));
  var DiagnosticCode;
  (function(DiagnosticCode2) {
    function is(value) {
      var candidate = value;
      return candidate !== void 0 && candidate !== null && (Is.number(candidate.value) || Is.string(candidate.value)) && Is.string(candidate.target);
    }
    DiagnosticCode2.is = is;
  })(DiagnosticCode || (DiagnosticCode = {}));
  var Diagnostic;
  (function(Diagnostic2) {
    function create(range, message, severity, code, source, relatedInformation) {
      var result = { range, message };
      if (Is.defined(severity)) {
        result.severity = severity;
      }
      if (Is.defined(code)) {
        result.code = code;
      }
      if (Is.defined(source)) {
        result.source = source;
      }
      if (Is.defined(relatedInformation)) {
        result.relatedInformation = relatedInformation;
      }
      return result;
    }
    Diagnostic2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Range2.is(candidate.range) && Is.string(candidate.message) && (Is.number(candidate.severity) || Is.undefined(candidate.severity)) && (Is.number(candidate.code) || Is.string(candidate.code) || Is.undefined(candidate.code)) && (Is.string(candidate.source) || Is.undefined(candidate.source)) && (Is.undefined(candidate.relatedInformation) || Is.typedArray(candidate.relatedInformation, DiagnosticRelatedInformation.is));
    }
    Diagnostic2.is = is;
  })(Diagnostic || (Diagnostic = {}));
  var Command;
  (function(Command2) {
    function create(title, command) {
      var args = [];
      for (var _i = 2; _i < arguments.length; _i++) {
        args[_i - 2] = arguments[_i];
      }
      var result = { title, command };
      if (Is.defined(args) && args.length > 0) {
        result.arguments = args;
      }
      return result;
    }
    Command2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.string(candidate.title) && Is.string(candidate.command);
    }
    Command2.is = is;
  })(Command || (Command = {}));
  var TextEdit;
  (function(TextEdit2) {
    function replace(range, newText) {
      return { range, newText };
    }
    TextEdit2.replace = replace;
    function insert(position, newText) {
      return { range: { start: position, end: position }, newText };
    }
    TextEdit2.insert = insert;
    function del(range) {
      return { range, newText: "" };
    }
    TextEdit2.del = del;
    function is(value) {
      var candidate = value;
      return Is.objectLiteral(candidate) && Is.string(candidate.newText) && Range2.is(candidate.range);
    }
    TextEdit2.is = is;
  })(TextEdit || (TextEdit = {}));
  var TextDocumentEdit;
  (function(TextDocumentEdit2) {
    function create(textDocument, edits) {
      return { textDocument, edits };
    }
    TextDocumentEdit2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && VersionedTextDocumentIdentifier.is(candidate.textDocument) && Array.isArray(candidate.edits);
    }
    TextDocumentEdit2.is = is;
  })(TextDocumentEdit || (TextDocumentEdit = {}));
  var CreateFile;
  (function(CreateFile2) {
    function create(uri, options) {
      var result = {
        kind: "create",
        uri
      };
      if (options !== void 0 && (options.overwrite !== void 0 || options.ignoreIfExists !== void 0)) {
        result.options = options;
      }
      return result;
    }
    CreateFile2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && candidate.kind === "create" && Is.string(candidate.uri) && (candidate.options === void 0 || (candidate.options.overwrite === void 0 || Is.boolean(candidate.options.overwrite)) && (candidate.options.ignoreIfExists === void 0 || Is.boolean(candidate.options.ignoreIfExists)));
    }
    CreateFile2.is = is;
  })(CreateFile || (CreateFile = {}));
  var RenameFile;
  (function(RenameFile2) {
    function create(oldUri, newUri, options) {
      var result = {
        kind: "rename",
        oldUri,
        newUri
      };
      if (options !== void 0 && (options.overwrite !== void 0 || options.ignoreIfExists !== void 0)) {
        result.options = options;
      }
      return result;
    }
    RenameFile2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && candidate.kind === "rename" && Is.string(candidate.oldUri) && Is.string(candidate.newUri) && (candidate.options === void 0 || (candidate.options.overwrite === void 0 || Is.boolean(candidate.options.overwrite)) && (candidate.options.ignoreIfExists === void 0 || Is.boolean(candidate.options.ignoreIfExists)));
    }
    RenameFile2.is = is;
  })(RenameFile || (RenameFile = {}));
  var DeleteFile;
  (function(DeleteFile2) {
    function create(uri, options) {
      var result = {
        kind: "delete",
        uri
      };
      if (options !== void 0 && (options.recursive !== void 0 || options.ignoreIfNotExists !== void 0)) {
        result.options = options;
      }
      return result;
    }
    DeleteFile2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && candidate.kind === "delete" && Is.string(candidate.uri) && (candidate.options === void 0 || (candidate.options.recursive === void 0 || Is.boolean(candidate.options.recursive)) && (candidate.options.ignoreIfNotExists === void 0 || Is.boolean(candidate.options.ignoreIfNotExists)));
    }
    DeleteFile2.is = is;
  })(DeleteFile || (DeleteFile = {}));
  var WorkspaceEdit;
  (function(WorkspaceEdit2) {
    function is(value) {
      var candidate = value;
      return candidate && (candidate.changes !== void 0 || candidate.documentChanges !== void 0) && (candidate.documentChanges === void 0 || candidate.documentChanges.every(function(change) {
        if (Is.string(change.kind)) {
          return CreateFile.is(change) || RenameFile.is(change) || DeleteFile.is(change);
        } else {
          return TextDocumentEdit.is(change);
        }
      }));
    }
    WorkspaceEdit2.is = is;
  })(WorkspaceEdit || (WorkspaceEdit = {}));
  var TextEditChangeImpl = function() {
    function TextEditChangeImpl2(edits) {
      this.edits = edits;
    }
    TextEditChangeImpl2.prototype.insert = function(position, newText) {
      this.edits.push(TextEdit.insert(position, newText));
    };
    TextEditChangeImpl2.prototype.replace = function(range, newText) {
      this.edits.push(TextEdit.replace(range, newText));
    };
    TextEditChangeImpl2.prototype.delete = function(range) {
      this.edits.push(TextEdit.del(range));
    };
    TextEditChangeImpl2.prototype.add = function(edit) {
      this.edits.push(edit);
    };
    TextEditChangeImpl2.prototype.all = function() {
      return this.edits;
    };
    TextEditChangeImpl2.prototype.clear = function() {
      this.edits.splice(0, this.edits.length);
    };
    return TextEditChangeImpl2;
  }();
  var WorkspaceChange = function() {
    function WorkspaceChange2(workspaceEdit) {
      var _this = this;
      this._textEditChanges = /* @__PURE__ */ Object.create(null);
      if (workspaceEdit) {
        this._workspaceEdit = workspaceEdit;
        if (workspaceEdit.documentChanges) {
          workspaceEdit.documentChanges.forEach(function(change) {
            if (TextDocumentEdit.is(change)) {
              var textEditChange = new TextEditChangeImpl(change.edits);
              _this._textEditChanges[change.textDocument.uri] = textEditChange;
            }
          });
        } else if (workspaceEdit.changes) {
          Object.keys(workspaceEdit.changes).forEach(function(key) {
            var textEditChange = new TextEditChangeImpl(workspaceEdit.changes[key]);
            _this._textEditChanges[key] = textEditChange;
          });
        }
      }
    }
    Object.defineProperty(WorkspaceChange2.prototype, "edit", {
      get: function() {
        if (this._workspaceEdit === void 0) {
          return { documentChanges: [] };
        }
        return this._workspaceEdit;
      },
      enumerable: true,
      configurable: true
    });
    WorkspaceChange2.prototype.getTextEditChange = function(key) {
      if (VersionedTextDocumentIdentifier.is(key)) {
        if (!this._workspaceEdit) {
          this._workspaceEdit = {
            documentChanges: []
          };
        }
        if (!this._workspaceEdit.documentChanges) {
          throw new Error("Workspace edit is not configured for document changes.");
        }
        var textDocument = key;
        var result = this._textEditChanges[textDocument.uri];
        if (!result) {
          var edits = [];
          var textDocumentEdit = {
            textDocument,
            edits
          };
          this._workspaceEdit.documentChanges.push(textDocumentEdit);
          result = new TextEditChangeImpl(edits);
          this._textEditChanges[textDocument.uri] = result;
        }
        return result;
      } else {
        if (!this._workspaceEdit) {
          this._workspaceEdit = {
            changes: /* @__PURE__ */ Object.create(null)
          };
        }
        if (!this._workspaceEdit.changes) {
          throw new Error("Workspace edit is not configured for normal text edit changes.");
        }
        var result = this._textEditChanges[key];
        if (!result) {
          var edits = [];
          this._workspaceEdit.changes[key] = edits;
          result = new TextEditChangeImpl(edits);
          this._textEditChanges[key] = result;
        }
        return result;
      }
    };
    WorkspaceChange2.prototype.createFile = function(uri, options) {
      this.checkDocumentChanges();
      this._workspaceEdit.documentChanges.push(CreateFile.create(uri, options));
    };
    WorkspaceChange2.prototype.renameFile = function(oldUri, newUri, options) {
      this.checkDocumentChanges();
      this._workspaceEdit.documentChanges.push(RenameFile.create(oldUri, newUri, options));
    };
    WorkspaceChange2.prototype.deleteFile = function(uri, options) {
      this.checkDocumentChanges();
      this._workspaceEdit.documentChanges.push(DeleteFile.create(uri, options));
    };
    WorkspaceChange2.prototype.checkDocumentChanges = function() {
      if (!this._workspaceEdit || !this._workspaceEdit.documentChanges) {
        throw new Error("Workspace edit is not configured for document changes.");
      }
    };
    return WorkspaceChange2;
  }();
  var TextDocumentIdentifier;
  (function(TextDocumentIdentifier2) {
    function create(uri) {
      return { uri };
    }
    TextDocumentIdentifier2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.string(candidate.uri);
    }
    TextDocumentIdentifier2.is = is;
  })(TextDocumentIdentifier || (TextDocumentIdentifier = {}));
  var VersionedTextDocumentIdentifier;
  (function(VersionedTextDocumentIdentifier2) {
    function create(uri, version) {
      return { uri, version };
    }
    VersionedTextDocumentIdentifier2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.string(candidate.uri) && (candidate.version === null || Is.number(candidate.version));
    }
    VersionedTextDocumentIdentifier2.is = is;
  })(VersionedTextDocumentIdentifier || (VersionedTextDocumentIdentifier = {}));
  var TextDocumentItem;
  (function(TextDocumentItem2) {
    function create(uri, languageId, version, text) {
      return { uri, languageId, version, text };
    }
    TextDocumentItem2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.string(candidate.uri) && Is.string(candidate.languageId) && Is.number(candidate.version) && Is.string(candidate.text);
    }
    TextDocumentItem2.is = is;
  })(TextDocumentItem || (TextDocumentItem = {}));
  var MarkupKind;
  (function(MarkupKind2) {
    MarkupKind2.PlainText = "plaintext";
    MarkupKind2.Markdown = "markdown";
  })(MarkupKind || (MarkupKind = {}));
  (function(MarkupKind2) {
    function is(value) {
      var candidate = value;
      return candidate === MarkupKind2.PlainText || candidate === MarkupKind2.Markdown;
    }
    MarkupKind2.is = is;
  })(MarkupKind || (MarkupKind = {}));
  var MarkupContent;
  (function(MarkupContent2) {
    function is(value) {
      var candidate = value;
      return Is.objectLiteral(value) && MarkupKind.is(candidate.kind) && Is.string(candidate.value);
    }
    MarkupContent2.is = is;
  })(MarkupContent || (MarkupContent = {}));
  var CompletionItemKind2;
  (function(CompletionItemKind3) {
    CompletionItemKind3.Text = 1;
    CompletionItemKind3.Method = 2;
    CompletionItemKind3.Function = 3;
    CompletionItemKind3.Constructor = 4;
    CompletionItemKind3.Field = 5;
    CompletionItemKind3.Variable = 6;
    CompletionItemKind3.Class = 7;
    CompletionItemKind3.Interface = 8;
    CompletionItemKind3.Module = 9;
    CompletionItemKind3.Property = 10;
    CompletionItemKind3.Unit = 11;
    CompletionItemKind3.Value = 12;
    CompletionItemKind3.Enum = 13;
    CompletionItemKind3.Keyword = 14;
    CompletionItemKind3.Snippet = 15;
    CompletionItemKind3.Color = 16;
    CompletionItemKind3.File = 17;
    CompletionItemKind3.Reference = 18;
    CompletionItemKind3.Folder = 19;
    CompletionItemKind3.EnumMember = 20;
    CompletionItemKind3.Constant = 21;
    CompletionItemKind3.Struct = 22;
    CompletionItemKind3.Event = 23;
    CompletionItemKind3.Operator = 24;
    CompletionItemKind3.TypeParameter = 25;
  })(CompletionItemKind2 || (CompletionItemKind2 = {}));
  var InsertTextFormat;
  (function(InsertTextFormat2) {
    InsertTextFormat2.PlainText = 1;
    InsertTextFormat2.Snippet = 2;
  })(InsertTextFormat || (InsertTextFormat = {}));
  var CompletionItemTag2;
  (function(CompletionItemTag3) {
    CompletionItemTag3.Deprecated = 1;
  })(CompletionItemTag2 || (CompletionItemTag2 = {}));
  var InsertReplaceEdit;
  (function(InsertReplaceEdit2) {
    function create(newText, insert, replace) {
      return { newText, insert, replace };
    }
    InsertReplaceEdit2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && Is.string(candidate.newText) && Range2.is(candidate.insert) && Range2.is(candidate.replace);
    }
    InsertReplaceEdit2.is = is;
  })(InsertReplaceEdit || (InsertReplaceEdit = {}));
  var CompletionItem;
  (function(CompletionItem2) {
    function create(label) {
      return { label };
    }
    CompletionItem2.create = create;
  })(CompletionItem || (CompletionItem = {}));
  var CompletionList;
  (function(CompletionList2) {
    function create(items, isIncomplete) {
      return { items: items ? items : [], isIncomplete: !!isIncomplete };
    }
    CompletionList2.create = create;
  })(CompletionList || (CompletionList = {}));
  var MarkedString;
  (function(MarkedString2) {
    function fromPlainText(plainText) {
      return plainText.replace(/[\\`*_{}[\]()#+\-.!]/g, "\\$&");
    }
    MarkedString2.fromPlainText = fromPlainText;
    function is(value) {
      var candidate = value;
      return Is.string(candidate) || Is.objectLiteral(candidate) && Is.string(candidate.language) && Is.string(candidate.value);
    }
    MarkedString2.is = is;
  })(MarkedString || (MarkedString = {}));
  var Hover;
  (function(Hover2) {
    function is(value) {
      var candidate = value;
      return !!candidate && Is.objectLiteral(candidate) && (MarkupContent.is(candidate.contents) || MarkedString.is(candidate.contents) || Is.typedArray(candidate.contents, MarkedString.is)) && (value.range === void 0 || Range2.is(value.range));
    }
    Hover2.is = is;
  })(Hover || (Hover = {}));
  var ParameterInformation;
  (function(ParameterInformation2) {
    function create(label, documentation) {
      return documentation ? { label, documentation } : { label };
    }
    ParameterInformation2.create = create;
  })(ParameterInformation || (ParameterInformation = {}));
  var SignatureInformation;
  (function(SignatureInformation2) {
    function create(label, documentation) {
      var parameters = [];
      for (var _i = 2; _i < arguments.length; _i++) {
        parameters[_i - 2] = arguments[_i];
      }
      var result = { label };
      if (Is.defined(documentation)) {
        result.documentation = documentation;
      }
      if (Is.defined(parameters)) {
        result.parameters = parameters;
      } else {
        result.parameters = [];
      }
      return result;
    }
    SignatureInformation2.create = create;
  })(SignatureInformation || (SignatureInformation = {}));
  var DocumentHighlightKind2;
  (function(DocumentHighlightKind3) {
    DocumentHighlightKind3.Text = 1;
    DocumentHighlightKind3.Read = 2;
    DocumentHighlightKind3.Write = 3;
  })(DocumentHighlightKind2 || (DocumentHighlightKind2 = {}));
  var DocumentHighlight;
  (function(DocumentHighlight2) {
    function create(range, kind) {
      var result = { range };
      if (Is.number(kind)) {
        result.kind = kind;
      }
      return result;
    }
    DocumentHighlight2.create = create;
  })(DocumentHighlight || (DocumentHighlight = {}));
  var SymbolKind2;
  (function(SymbolKind3) {
    SymbolKind3.File = 1;
    SymbolKind3.Module = 2;
    SymbolKind3.Namespace = 3;
    SymbolKind3.Package = 4;
    SymbolKind3.Class = 5;
    SymbolKind3.Method = 6;
    SymbolKind3.Property = 7;
    SymbolKind3.Field = 8;
    SymbolKind3.Constructor = 9;
    SymbolKind3.Enum = 10;
    SymbolKind3.Interface = 11;
    SymbolKind3.Function = 12;
    SymbolKind3.Variable = 13;
    SymbolKind3.Constant = 14;
    SymbolKind3.String = 15;
    SymbolKind3.Number = 16;
    SymbolKind3.Boolean = 17;
    SymbolKind3.Array = 18;
    SymbolKind3.Object = 19;
    SymbolKind3.Key = 20;
    SymbolKind3.Null = 21;
    SymbolKind3.EnumMember = 22;
    SymbolKind3.Struct = 23;
    SymbolKind3.Event = 24;
    SymbolKind3.Operator = 25;
    SymbolKind3.TypeParameter = 26;
  })(SymbolKind2 || (SymbolKind2 = {}));
  var SymbolTag2;
  (function(SymbolTag3) {
    SymbolTag3.Deprecated = 1;
  })(SymbolTag2 || (SymbolTag2 = {}));
  var SymbolInformation;
  (function(SymbolInformation2) {
    function create(name, kind, range, uri, containerName) {
      var result = {
        name,
        kind,
        location: { uri, range }
      };
      if (containerName) {
        result.containerName = containerName;
      }
      return result;
    }
    SymbolInformation2.create = create;
  })(SymbolInformation || (SymbolInformation = {}));
  var DocumentSymbol;
  (function(DocumentSymbol2) {
    function create(name, detail, kind, range, selectionRange, children) {
      var result = {
        name,
        detail,
        kind,
        range,
        selectionRange
      };
      if (children !== void 0) {
        result.children = children;
      }
      return result;
    }
    DocumentSymbol2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && Is.string(candidate.name) && Is.number(candidate.kind) && Range2.is(candidate.range) && Range2.is(candidate.selectionRange) && (candidate.detail === void 0 || Is.string(candidate.detail)) && (candidate.deprecated === void 0 || Is.boolean(candidate.deprecated)) && (candidate.children === void 0 || Array.isArray(candidate.children)) && (candidate.tags === void 0 || Array.isArray(candidate.tags));
    }
    DocumentSymbol2.is = is;
  })(DocumentSymbol || (DocumentSymbol = {}));
  var CodeActionKind;
  (function(CodeActionKind2) {
    CodeActionKind2.Empty = "";
    CodeActionKind2.QuickFix = "quickfix";
    CodeActionKind2.Refactor = "refactor";
    CodeActionKind2.RefactorExtract = "refactor.extract";
    CodeActionKind2.RefactorInline = "refactor.inline";
    CodeActionKind2.RefactorRewrite = "refactor.rewrite";
    CodeActionKind2.Source = "source";
    CodeActionKind2.SourceOrganizeImports = "source.organizeImports";
    CodeActionKind2.SourceFixAll = "source.fixAll";
  })(CodeActionKind || (CodeActionKind = {}));
  var CodeActionContext;
  (function(CodeActionContext2) {
    function create(diagnostics, only) {
      var result = { diagnostics };
      if (only !== void 0 && only !== null) {
        result.only = only;
      }
      return result;
    }
    CodeActionContext2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.typedArray(candidate.diagnostics, Diagnostic.is) && (candidate.only === void 0 || Is.typedArray(candidate.only, Is.string));
    }
    CodeActionContext2.is = is;
  })(CodeActionContext || (CodeActionContext = {}));
  var CodeAction;
  (function(CodeAction2) {
    function create(title, commandOrEdit, kind) {
      var result = { title };
      if (Command.is(commandOrEdit)) {
        result.command = commandOrEdit;
      } else {
        result.edit = commandOrEdit;
      }
      if (kind !== void 0) {
        result.kind = kind;
      }
      return result;
    }
    CodeAction2.create = create;
    function is(value) {
      var candidate = value;
      return candidate && Is.string(candidate.title) && (candidate.diagnostics === void 0 || Is.typedArray(candidate.diagnostics, Diagnostic.is)) && (candidate.kind === void 0 || Is.string(candidate.kind)) && (candidate.edit !== void 0 || candidate.command !== void 0) && (candidate.command === void 0 || Command.is(candidate.command)) && (candidate.isPreferred === void 0 || Is.boolean(candidate.isPreferred)) && (candidate.edit === void 0 || WorkspaceEdit.is(candidate.edit));
    }
    CodeAction2.is = is;
  })(CodeAction || (CodeAction = {}));
  var CodeLens;
  (function(CodeLens2) {
    function create(range, data) {
      var result = { range };
      if (Is.defined(data)) {
        result.data = data;
      }
      return result;
    }
    CodeLens2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Range2.is(candidate.range) && (Is.undefined(candidate.command) || Command.is(candidate.command));
    }
    CodeLens2.is = is;
  })(CodeLens || (CodeLens = {}));
  var FormattingOptions;
  (function(FormattingOptions2) {
    function create(tabSize, insertSpaces) {
      return { tabSize, insertSpaces };
    }
    FormattingOptions2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.number(candidate.tabSize) && Is.boolean(candidate.insertSpaces);
    }
    FormattingOptions2.is = is;
  })(FormattingOptions || (FormattingOptions = {}));
  var DocumentLink;
  (function(DocumentLink2) {
    function create(range, target, data) {
      return { range, target, data };
    }
    DocumentLink2.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Range2.is(candidate.range) && (Is.undefined(candidate.target) || Is.string(candidate.target));
    }
    DocumentLink2.is = is;
  })(DocumentLink || (DocumentLink = {}));
  var SelectionRange;
  (function(SelectionRange2) {
    function create(range, parent) {
      return { range, parent };
    }
    SelectionRange2.create = create;
    function is(value) {
      var candidate = value;
      return candidate !== void 0 && Range2.is(candidate.range) && (candidate.parent === void 0 || SelectionRange2.is(candidate.parent));
    }
    SelectionRange2.is = is;
  })(SelectionRange || (SelectionRange = {}));
  var TextDocument;
  (function(TextDocument3) {
    function create(uri, languageId, version, content) {
      return new FullTextDocument(uri, languageId, version, content);
    }
    TextDocument3.create = create;
    function is(value) {
      var candidate = value;
      return Is.defined(candidate) && Is.string(candidate.uri) && (Is.undefined(candidate.languageId) || Is.string(candidate.languageId)) && Is.number(candidate.lineCount) && Is.func(candidate.getText) && Is.func(candidate.positionAt) && Is.func(candidate.offsetAt) ? true : false;
    }
    TextDocument3.is = is;
    function applyEdits(document, edits) {
      var text = document.getText();
      var sortedEdits = mergeSort3(edits, function(a2, b) {
        var diff = a2.range.start.line - b.range.start.line;
        if (diff === 0) {
          return a2.range.start.character - b.range.start.character;
        }
        return diff;
      });
      var lastModifiedOffset = text.length;
      for (var i = sortedEdits.length - 1; i >= 0; i--) {
        var e = sortedEdits[i];
        var startOffset = document.offsetAt(e.range.start);
        var endOffset = document.offsetAt(e.range.end);
        if (endOffset <= lastModifiedOffset) {
          text = text.substring(0, startOffset) + e.newText + text.substring(endOffset, text.length);
        } else {
          throw new Error("Overlapping edit");
        }
        lastModifiedOffset = startOffset;
      }
      return text;
    }
    TextDocument3.applyEdits = applyEdits;
    function mergeSort3(data, compare) {
      if (data.length <= 1) {
        return data;
      }
      var p = data.length / 2 | 0;
      var left = data.slice(0, p);
      var right = data.slice(p);
      mergeSort3(left, compare);
      mergeSort3(right, compare);
      var leftIdx = 0;
      var rightIdx = 0;
      var i = 0;
      while (leftIdx < left.length && rightIdx < right.length) {
        var ret = compare(left[leftIdx], right[rightIdx]);
        if (ret <= 0) {
          data[i++] = left[leftIdx++];
        } else {
          data[i++] = right[rightIdx++];
        }
      }
      while (leftIdx < left.length) {
        data[i++] = left[leftIdx++];
      }
      while (rightIdx < right.length) {
        data[i++] = right[rightIdx++];
      }
      return data;
    }
  })(TextDocument || (TextDocument = {}));
  var FullTextDocument = function() {
    function FullTextDocument3(uri, languageId, version, content) {
      this._uri = uri;
      this._languageId = languageId;
      this._version = version;
      this._content = content;
      this._lineOffsets = void 0;
    }
    Object.defineProperty(FullTextDocument3.prototype, "uri", {
      get: function() {
        return this._uri;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(FullTextDocument3.prototype, "languageId", {
      get: function() {
        return this._languageId;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(FullTextDocument3.prototype, "version", {
      get: function() {
        return this._version;
      },
      enumerable: true,
      configurable: true
    });
    FullTextDocument3.prototype.getText = function(range) {
      if (range) {
        var start = this.offsetAt(range.start);
        var end = this.offsetAt(range.end);
        return this._content.substring(start, end);
      }
      return this._content;
    };
    FullTextDocument3.prototype.update = function(event, version) {
      this._content = event.text;
      this._version = version;
      this._lineOffsets = void 0;
    };
    FullTextDocument3.prototype.getLineOffsets = function() {
      if (this._lineOffsets === void 0) {
        var lineOffsets = [];
        var text = this._content;
        var isLineStart = true;
        for (var i = 0; i < text.length; i++) {
          if (isLineStart) {
            lineOffsets.push(i);
            isLineStart = false;
          }
          var ch = text.charAt(i);
          isLineStart = ch === "\r" || ch === "\n";
          if (ch === "\r" && i + 1 < text.length && text.charAt(i + 1) === "\n") {
            i++;
          }
        }
        if (isLineStart && text.length > 0) {
          lineOffsets.push(text.length);
        }
        this._lineOffsets = lineOffsets;
      }
      return this._lineOffsets;
    };
    FullTextDocument3.prototype.positionAt = function(offset) {
      offset = Math.max(Math.min(offset, this._content.length), 0);
      var lineOffsets = this.getLineOffsets();
      var low = 0, high = lineOffsets.length;
      if (high === 0) {
        return Position2.create(0, offset);
      }
      while (low < high) {
        var mid = Math.floor((low + high) / 2);
        if (lineOffsets[mid] > offset) {
          high = mid;
        } else {
          low = mid + 1;
        }
      }
      var line = low - 1;
      return Position2.create(line, offset - lineOffsets[line]);
    };
    FullTextDocument3.prototype.offsetAt = function(position) {
      var lineOffsets = this.getLineOffsets();
      if (position.line >= lineOffsets.length) {
        return this._content.length;
      } else if (position.line < 0) {
        return 0;
      }
      var lineOffset = lineOffsets[position.line];
      var nextLineOffset = position.line + 1 < lineOffsets.length ? lineOffsets[position.line + 1] : this._content.length;
      return Math.max(Math.min(lineOffset + position.character, nextLineOffset), lineOffset);
    };
    Object.defineProperty(FullTextDocument3.prototype, "lineCount", {
      get: function() {
        return this.getLineOffsets().length;
      },
      enumerable: true,
      configurable: true
    });
    return FullTextDocument3;
  }();
  var Is;
  (function(Is2) {
    var toString = Object.prototype.toString;
    function defined(value) {
      return typeof value !== "undefined";
    }
    Is2.defined = defined;
    function undefined2(value) {
      return typeof value === "undefined";
    }
    Is2.undefined = undefined2;
    function boolean(value) {
      return value === true || value === false;
    }
    Is2.boolean = boolean;
    function string(value) {
      return toString.call(value) === "[object String]";
    }
    Is2.string = string;
    function number(value) {
      return toString.call(value) === "[object Number]";
    }
    Is2.number = number;
    function func(value) {
      return toString.call(value) === "[object Function]";
    }
    Is2.func = func;
    function objectLiteral(value) {
      return value !== null && typeof value === "object";
    }
    Is2.objectLiteral = objectLiteral;
    function typedArray(value, check) {
      return Array.isArray(value) && value.every(check);
    }
    Is2.typedArray = typedArray;
  })(Is || (Is = {}));

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-languageserver-textdocument/lib/esm/main.js
  var FullTextDocument2 = function() {
    function FullTextDocument3(uri, languageId, version, content) {
      this._uri = uri;
      this._languageId = languageId;
      this._version = version;
      this._content = content;
      this._lineOffsets = void 0;
    }
    Object.defineProperty(FullTextDocument3.prototype, "uri", {
      get: function() {
        return this._uri;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(FullTextDocument3.prototype, "languageId", {
      get: function() {
        return this._languageId;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(FullTextDocument3.prototype, "version", {
      get: function() {
        return this._version;
      },
      enumerable: true,
      configurable: true
    });
    FullTextDocument3.prototype.getText = function(range) {
      if (range) {
        var start = this.offsetAt(range.start);
        var end = this.offsetAt(range.end);
        return this._content.substring(start, end);
      }
      return this._content;
    };
    FullTextDocument3.prototype.update = function(changes, version) {
      for (var _i = 0, changes_1 = changes; _i < changes_1.length; _i++) {
        var change = changes_1[_i];
        if (FullTextDocument3.isIncremental(change)) {
          var range = getWellformedRange(change.range);
          var startOffset = this.offsetAt(range.start);
          var endOffset = this.offsetAt(range.end);
          this._content = this._content.substring(0, startOffset) + change.text + this._content.substring(endOffset, this._content.length);
          var startLine = Math.max(range.start.line, 0);
          var endLine = Math.max(range.end.line, 0);
          var lineOffsets = this._lineOffsets;
          var addedLineOffsets = computeLineOffsets(change.text, false, startOffset);
          if (endLine - startLine === addedLineOffsets.length) {
            for (var i = 0, len = addedLineOffsets.length; i < len; i++) {
              lineOffsets[i + startLine + 1] = addedLineOffsets[i];
            }
          } else {
            if (addedLineOffsets.length < 1e4) {
              lineOffsets.splice.apply(lineOffsets, [startLine + 1, endLine - startLine].concat(addedLineOffsets));
            } else {
              this._lineOffsets = lineOffsets = lineOffsets.slice(0, startLine + 1).concat(addedLineOffsets, lineOffsets.slice(endLine + 1));
            }
          }
          var diff = change.text.length - (endOffset - startOffset);
          if (diff !== 0) {
            for (var i = startLine + 1 + addedLineOffsets.length, len = lineOffsets.length; i < len; i++) {
              lineOffsets[i] = lineOffsets[i] + diff;
            }
          }
        } else if (FullTextDocument3.isFull(change)) {
          this._content = change.text;
          this._lineOffsets = void 0;
        } else {
          throw new Error("Unknown change event received");
        }
      }
      this._version = version;
    };
    FullTextDocument3.prototype.getLineOffsets = function() {
      if (this._lineOffsets === void 0) {
        this._lineOffsets = computeLineOffsets(this._content, true);
      }
      return this._lineOffsets;
    };
    FullTextDocument3.prototype.positionAt = function(offset) {
      offset = Math.max(Math.min(offset, this._content.length), 0);
      var lineOffsets = this.getLineOffsets();
      var low = 0, high = lineOffsets.length;
      if (high === 0) {
        return { line: 0, character: offset };
      }
      while (low < high) {
        var mid = Math.floor((low + high) / 2);
        if (lineOffsets[mid] > offset) {
          high = mid;
        } else {
          low = mid + 1;
        }
      }
      var line = low - 1;
      return { line, character: offset - lineOffsets[line] };
    };
    FullTextDocument3.prototype.offsetAt = function(position) {
      var lineOffsets = this.getLineOffsets();
      if (position.line >= lineOffsets.length) {
        return this._content.length;
      } else if (position.line < 0) {
        return 0;
      }
      var lineOffset = lineOffsets[position.line];
      var nextLineOffset = position.line + 1 < lineOffsets.length ? lineOffsets[position.line + 1] : this._content.length;
      return Math.max(Math.min(lineOffset + position.character, nextLineOffset), lineOffset);
    };
    Object.defineProperty(FullTextDocument3.prototype, "lineCount", {
      get: function() {
        return this.getLineOffsets().length;
      },
      enumerable: true,
      configurable: true
    });
    FullTextDocument3.isIncremental = function(event) {
      var candidate = event;
      return candidate !== void 0 && candidate !== null && typeof candidate.text === "string" && candidate.range !== void 0 && (candidate.rangeLength === void 0 || typeof candidate.rangeLength === "number");
    };
    FullTextDocument3.isFull = function(event) {
      var candidate = event;
      return candidate !== void 0 && candidate !== null && typeof candidate.text === "string" && candidate.range === void 0 && candidate.rangeLength === void 0;
    };
    return FullTextDocument3;
  }();
  var TextDocument2;
  (function(TextDocument3) {
    function create(uri, languageId, version, content) {
      return new FullTextDocument2(uri, languageId, version, content);
    }
    TextDocument3.create = create;
    function update(document, changes, version) {
      if (document instanceof FullTextDocument2) {
        document.update(changes, version);
        return document;
      } else {
        throw new Error("TextDocument.update: document must be created by TextDocument.create");
      }
    }
    TextDocument3.update = update;
    function applyEdits(document, edits) {
      var text = document.getText();
      var sortedEdits = mergeSort2(edits.map(getWellformedEdit), function(a2, b) {
        var diff = a2.range.start.line - b.range.start.line;
        if (diff === 0) {
          return a2.range.start.character - b.range.start.character;
        }
        return diff;
      });
      var lastModifiedOffset = 0;
      var spans = [];
      for (var _i = 0, sortedEdits_1 = sortedEdits; _i < sortedEdits_1.length; _i++) {
        var e = sortedEdits_1[_i];
        var startOffset = document.offsetAt(e.range.start);
        if (startOffset < lastModifiedOffset) {
          throw new Error("Overlapping edit");
        } else if (startOffset > lastModifiedOffset) {
          spans.push(text.substring(lastModifiedOffset, startOffset));
        }
        if (e.newText.length) {
          spans.push(e.newText);
        }
        lastModifiedOffset = document.offsetAt(e.range.end);
      }
      spans.push(text.substr(lastModifiedOffset));
      return spans.join("");
    }
    TextDocument3.applyEdits = applyEdits;
  })(TextDocument2 || (TextDocument2 = {}));
  function mergeSort2(data, compare) {
    if (data.length <= 1) {
      return data;
    }
    var p = data.length / 2 | 0;
    var left = data.slice(0, p);
    var right = data.slice(p);
    mergeSort2(left, compare);
    mergeSort2(right, compare);
    var leftIdx = 0;
    var rightIdx = 0;
    var i = 0;
    while (leftIdx < left.length && rightIdx < right.length) {
      var ret = compare(left[leftIdx], right[rightIdx]);
      if (ret <= 0) {
        data[i++] = left[leftIdx++];
      } else {
        data[i++] = right[rightIdx++];
      }
    }
    while (leftIdx < left.length) {
      data[i++] = left[leftIdx++];
    }
    while (rightIdx < right.length) {
      data[i++] = right[rightIdx++];
    }
    return data;
  }
  function computeLineOffsets(text, isAtLineStart, textOffset) {
    if (textOffset === void 0) {
      textOffset = 0;
    }
    var result = isAtLineStart ? [textOffset] : [];
    for (var i = 0; i < text.length; i++) {
      var ch = text.charCodeAt(i);
      if (ch === 13 || ch === 10) {
        if (ch === 13 && i + 1 < text.length && text.charCodeAt(i + 1) === 10) {
          i++;
        }
        result.push(textOffset + i + 1);
      }
    }
    return result;
  }
  function getWellformedRange(range) {
    var start = range.start;
    var end = range.end;
    if (start.line > end.line || start.line === end.line && start.character > end.character) {
      return { start: end, end: start };
    }
    return range;
  }
  function getWellformedEdit(textEdit) {
    var range = getWellformedRange(textEdit.range);
    if (range !== textEdit.range) {
      return { newText: textEdit.newText, range };
    }
    return textEdit;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/cssLanguageTypes.js
  var ClientCapabilities;
  (function(ClientCapabilities2) {
    ClientCapabilities2.LATEST = {
      textDocument: {
        completion: {
          completionItem: {
            documentationFormat: [MarkupKind.Markdown, MarkupKind.PlainText]
          }
        },
        hover: {
          contentFormat: [MarkupKind.Markdown, MarkupKind.PlainText]
        }
      }
    };
  })(ClientCapabilities || (ClientCapabilities = {}));
  var FileType;
  (function(FileType2) {
    FileType2[FileType2["Unknown"] = 0] = "Unknown";
    FileType2[FileType2["File"] = 1] = "File";
    FileType2[FileType2["Directory"] = 2] = "Directory";
    FileType2[FileType2["SymbolicLink"] = 64] = "SymbolicLink";
  })(FileType || (FileType = {}));

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-uri/index.js
  var __extends3 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var _a2;
  var isWindows2;
  if (typeof process === "object") {
    isWindows2 = process.platform === "win32";
  } else if (typeof navigator === "object") {
    userAgent = navigator.userAgent;
    isWindows2 = userAgent.indexOf("Windows") >= 0;
  }
  var userAgent;
  var _schemePattern2 = /^\w[\w\d+.-]*$/;
  var _singleSlashStart2 = /^\//;
  var _doubleSlashStart2 = /^\/\//;
  function _validateUri2(ret, _strict) {
    if (!ret.scheme && _strict) {
      throw new Error('[UriError]: Scheme is missing: {scheme: "", authority: "' + ret.authority + '", path: "' + ret.path + '", query: "' + ret.query + '", fragment: "' + ret.fragment + '"}');
    }
    if (ret.scheme && !_schemePattern2.test(ret.scheme)) {
      throw new Error("[UriError]: Scheme contains illegal characters.");
    }
    if (ret.path) {
      if (ret.authority) {
        if (!_singleSlashStart2.test(ret.path)) {
          throw new Error('[UriError]: If a URI contains an authority component, then the path component must either be empty or begin with a slash ("/") character');
        }
      } else {
        if (_doubleSlashStart2.test(ret.path)) {
          throw new Error('[UriError]: If a URI does not contain an authority component, then the path cannot begin with two slash characters ("//")');
        }
      }
    }
  }
  function _schemeFix2(scheme, _strict) {
    if (!scheme && !_strict) {
      return "file";
    }
    return scheme;
  }
  function _referenceResolution2(scheme, path) {
    switch (scheme) {
      case "https":
      case "http":
      case "file":
        if (!path) {
          path = _slash2;
        } else if (path[0] !== _slash2) {
          path = _slash2 + path;
        }
        break;
    }
    return path;
  }
  var _empty2 = "";
  var _slash2 = "/";
  var _regexp2 = /^(([^:/?#]+?):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/;
  var URI2 = function() {
    function URI3(schemeOrData, authority, path, query, fragment, _strict) {
      if (_strict === void 0) {
        _strict = false;
      }
      if (typeof schemeOrData === "object") {
        this.scheme = schemeOrData.scheme || _empty2;
        this.authority = schemeOrData.authority || _empty2;
        this.path = schemeOrData.path || _empty2;
        this.query = schemeOrData.query || _empty2;
        this.fragment = schemeOrData.fragment || _empty2;
      } else {
        this.scheme = _schemeFix2(schemeOrData, _strict);
        this.authority = authority || _empty2;
        this.path = _referenceResolution2(this.scheme, path || _empty2);
        this.query = query || _empty2;
        this.fragment = fragment || _empty2;
        _validateUri2(this, _strict);
      }
    }
    URI3.isUri = function(thing) {
      if (thing instanceof URI3) {
        return true;
      }
      if (!thing) {
        return false;
      }
      return typeof thing.authority === "string" && typeof thing.fragment === "string" && typeof thing.path === "string" && typeof thing.query === "string" && typeof thing.scheme === "string" && typeof thing.fsPath === "function" && typeof thing.with === "function" && typeof thing.toString === "function";
    };
    Object.defineProperty(URI3.prototype, "fsPath", {
      get: function() {
        return uriToFsPath2(this, false);
      },
      enumerable: true,
      configurable: true
    });
    URI3.prototype.with = function(change) {
      if (!change) {
        return this;
      }
      var scheme = change.scheme, authority = change.authority, path = change.path, query = change.query, fragment = change.fragment;
      if (scheme === void 0) {
        scheme = this.scheme;
      } else if (scheme === null) {
        scheme = _empty2;
      }
      if (authority === void 0) {
        authority = this.authority;
      } else if (authority === null) {
        authority = _empty2;
      }
      if (path === void 0) {
        path = this.path;
      } else if (path === null) {
        path = _empty2;
      }
      if (query === void 0) {
        query = this.query;
      } else if (query === null) {
        query = _empty2;
      }
      if (fragment === void 0) {
        fragment = this.fragment;
      } else if (fragment === null) {
        fragment = _empty2;
      }
      if (scheme === this.scheme && authority === this.authority && path === this.path && query === this.query && fragment === this.fragment) {
        return this;
      }
      return new _URI(scheme, authority, path, query, fragment);
    };
    URI3.parse = function(value, _strict) {
      if (_strict === void 0) {
        _strict = false;
      }
      var match = _regexp2.exec(value);
      if (!match) {
        return new _URI(_empty2, _empty2, _empty2, _empty2, _empty2);
      }
      return new _URI(match[2] || _empty2, percentDecode2(match[4] || _empty2), percentDecode2(match[5] || _empty2), percentDecode2(match[7] || _empty2), percentDecode2(match[9] || _empty2), _strict);
    };
    URI3.file = function(path) {
      var authority = _empty2;
      if (isWindows2) {
        path = path.replace(/\\/g, _slash2);
      }
      if (path[0] === _slash2 && path[1] === _slash2) {
        var idx = path.indexOf(_slash2, 2);
        if (idx === -1) {
          authority = path.substring(2);
          path = _slash2;
        } else {
          authority = path.substring(2, idx);
          path = path.substring(idx) || _slash2;
        }
      }
      return new _URI("file", authority, path, _empty2, _empty2);
    };
    URI3.from = function(components) {
      return new _URI(components.scheme, components.authority, components.path, components.query, components.fragment);
    };
    URI3.prototype.toString = function(skipEncoding) {
      if (skipEncoding === void 0) {
        skipEncoding = false;
      }
      return _asFormatted2(this, skipEncoding);
    };
    URI3.prototype.toJSON = function() {
      return this;
    };
    URI3.revive = function(data) {
      if (!data) {
        return data;
      } else if (data instanceof URI3) {
        return data;
      } else {
        var result = new _URI(data);
        result._formatted = data.external;
        result._fsPath = data._sep === _pathSepMarker2 ? data.fsPath : null;
        return result;
      }
    };
    return URI3;
  }();
  var _pathSepMarker2 = isWindows2 ? 1 : void 0;
  var _URI = function(_super) {
    __extends3(_URI2, _super);
    function _URI2() {
      var _this = _super !== null && _super.apply(this, arguments) || this;
      _this._formatted = null;
      _this._fsPath = null;
      return _this;
    }
    Object.defineProperty(_URI2.prototype, "fsPath", {
      get: function() {
        if (!this._fsPath) {
          this._fsPath = uriToFsPath2(this, false);
        }
        return this._fsPath;
      },
      enumerable: true,
      configurable: true
    });
    _URI2.prototype.toString = function(skipEncoding) {
      if (skipEncoding === void 0) {
        skipEncoding = false;
      }
      if (!skipEncoding) {
        if (!this._formatted) {
          this._formatted = _asFormatted2(this, false);
        }
        return this._formatted;
      } else {
        return _asFormatted2(this, true);
      }
    };
    _URI2.prototype.toJSON = function() {
      var res = {
        $mid: 1
      };
      if (this._fsPath) {
        res.fsPath = this._fsPath;
        res._sep = _pathSepMarker2;
      }
      if (this._formatted) {
        res.external = this._formatted;
      }
      if (this.path) {
        res.path = this.path;
      }
      if (this.scheme) {
        res.scheme = this.scheme;
      }
      if (this.authority) {
        res.authority = this.authority;
      }
      if (this.query) {
        res.query = this.query;
      }
      if (this.fragment) {
        res.fragment = this.fragment;
      }
      return res;
    };
    return _URI2;
  }(URI2);
  var encodeTable2 = (_a2 = {}, _a2[58] = "%3A", _a2[47] = "%2F", _a2[63] = "%3F", _a2[35] = "%23", _a2[91] = "%5B", _a2[93] = "%5D", _a2[64] = "%40", _a2[33] = "%21", _a2[36] = "%24", _a2[38] = "%26", _a2[39] = "%27", _a2[40] = "%28", _a2[41] = "%29", _a2[42] = "%2A", _a2[43] = "%2B", _a2[44] = "%2C", _a2[59] = "%3B", _a2[61] = "%3D", _a2[32] = "%20", _a2);
  function encodeURIComponentFast2(uriComponent, allowSlash) {
    var res = void 0;
    var nativeEncodePos = -1;
    for (var pos = 0; pos < uriComponent.length; pos++) {
      var code = uriComponent.charCodeAt(pos);
      if (code >= 97 && code <= 122 || code >= 65 && code <= 90 || code >= 48 && code <= 57 || code === 45 || code === 46 || code === 95 || code === 126 || allowSlash && code === 47) {
        if (nativeEncodePos !== -1) {
          res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
          nativeEncodePos = -1;
        }
        if (res !== void 0) {
          res += uriComponent.charAt(pos);
        }
      } else {
        if (res === void 0) {
          res = uriComponent.substr(0, pos);
        }
        var escaped = encodeTable2[code];
        if (escaped !== void 0) {
          if (nativeEncodePos !== -1) {
            res += encodeURIComponent(uriComponent.substring(nativeEncodePos, pos));
            nativeEncodePos = -1;
          }
          res += escaped;
        } else if (nativeEncodePos === -1) {
          nativeEncodePos = pos;
        }
      }
    }
    if (nativeEncodePos !== -1) {
      res += encodeURIComponent(uriComponent.substring(nativeEncodePos));
    }
    return res !== void 0 ? res : uriComponent;
  }
  function encodeURIComponentMinimal2(path) {
    var res = void 0;
    for (var pos = 0; pos < path.length; pos++) {
      var code = path.charCodeAt(pos);
      if (code === 35 || code === 63) {
        if (res === void 0) {
          res = path.substr(0, pos);
        }
        res += encodeTable2[code];
      } else {
        if (res !== void 0) {
          res += path[pos];
        }
      }
    }
    return res !== void 0 ? res : path;
  }
  function uriToFsPath2(uri, keepDriveLetterCasing) {
    var value;
    if (uri.authority && uri.path.length > 1 && uri.scheme === "file") {
      value = "//" + uri.authority + uri.path;
    } else if (uri.path.charCodeAt(0) === 47 && (uri.path.charCodeAt(1) >= 65 && uri.path.charCodeAt(1) <= 90 || uri.path.charCodeAt(1) >= 97 && uri.path.charCodeAt(1) <= 122) && uri.path.charCodeAt(2) === 58) {
      if (!keepDriveLetterCasing) {
        value = uri.path[1].toLowerCase() + uri.path.substr(2);
      } else {
        value = uri.path.substr(1);
      }
    } else {
      value = uri.path;
    }
    if (isWindows2) {
      value = value.replace(/\//g, "\\");
    }
    return value;
  }
  function _asFormatted2(uri, skipEncoding) {
    var encoder = !skipEncoding ? encodeURIComponentFast2 : encodeURIComponentMinimal2;
    var res = "";
    var scheme = uri.scheme, authority = uri.authority, path = uri.path, query = uri.query, fragment = uri.fragment;
    if (scheme) {
      res += scheme;
      res += ":";
    }
    if (authority || scheme === "file") {
      res += _slash2;
      res += _slash2;
    }
    if (authority) {
      var idx = authority.indexOf("@");
      if (idx !== -1) {
        var userinfo = authority.substr(0, idx);
        authority = authority.substr(idx + 1);
        idx = userinfo.indexOf(":");
        if (idx === -1) {
          res += encoder(userinfo, false);
        } else {
          res += encoder(userinfo.substr(0, idx), false);
          res += ":";
          res += encoder(userinfo.substr(idx + 1), false);
        }
        res += "@";
      }
      authority = authority.toLowerCase();
      idx = authority.indexOf(":");
      if (idx === -1) {
        res += encoder(authority, false);
      } else {
        res += encoder(authority.substr(0, idx), false);
        res += authority.substr(idx);
      }
    }
    if (path) {
      if (path.length >= 3 && path.charCodeAt(0) === 47 && path.charCodeAt(2) === 58) {
        var code = path.charCodeAt(1);
        if (code >= 65 && code <= 90) {
          path = "/" + String.fromCharCode(code + 32) + ":" + path.substr(3);
        }
      } else if (path.length >= 2 && path.charCodeAt(1) === 58) {
        var code = path.charCodeAt(0);
        if (code >= 65 && code <= 90) {
          path = String.fromCharCode(code + 32) + ":" + path.substr(2);
        }
      }
      res += encoder(path, true);
    }
    if (query) {
      res += "?";
      res += encoder(query, false);
    }
    if (fragment) {
      res += "#";
      res += !skipEncoding ? encodeURIComponentFast2(fragment, false) : fragment;
    }
    return res;
  }
  function decodeURIComponentGraceful2(str) {
    try {
      return decodeURIComponent(str);
    } catch (_a3) {
      if (str.length > 3) {
        return str.substr(0, 3) + decodeURIComponentGraceful2(str.substr(3));
      } else {
        return str;
      }
    }
  }
  var _rEncodedAsHex2 = /(%[0-9A-Za-z][0-9A-Za-z])+/g;
  function percentDecode2(str) {
    if (!str.match(_rEncodedAsHex2)) {
      return str;
    }
    return str.replace(_rEncodedAsHex2, function(match) {
      return decodeURIComponentGraceful2(match);
    });
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/utils/resources.js
  var Slash = "/".charCodeAt(0);
  var Dot = ".".charCodeAt(0);
  function dirname2(uri) {
    var lastIndexOfSlash = uri.lastIndexOf("/");
    return lastIndexOfSlash !== -1 ? uri.substr(0, lastIndexOfSlash) : "";
  }
  function extname2(uri) {
    for (var i = uri.length - 1; i >= 0; i--) {
      var ch = uri.charCodeAt(i);
      if (ch === Dot) {
        if (i > 0 && uri.charCodeAt(i - 1) !== Slash) {
          return uri.substr(i);
        } else {
          break;
        }
      } else if (ch === Slash) {
        break;
      }
    }
    return "";
  }
  function normalizePath(parts) {
    var newParts = [];
    for (var _i = 0, parts_1 = parts; _i < parts_1.length; _i++) {
      var part = parts_1[_i];
      if (part.length === 0 || part.length === 1 && part.charCodeAt(0) === Dot) {
      } else if (part.length === 2 && part.charCodeAt(0) === Dot && part.charCodeAt(1) === Dot) {
        newParts.pop();
      } else {
        newParts.push(part);
      }
    }
    if (parts.length > 1 && parts[parts.length - 1].length === 0) {
      newParts.push("");
    }
    var res = newParts.join("/");
    if (parts[0].length === 0) {
      res = "/" + res;
    }
    return res;
  }
  function joinPath(uriString) {
    var paths = [];
    for (var _i = 1; _i < arguments.length; _i++) {
      paths[_i - 1] = arguments[_i];
    }
    var uri = URI2.parse(uriString);
    var parts = uri.path.split("/");
    for (var _a3 = 0, paths_1 = paths; _a3 < paths_1.length; _a3++) {
      var path = paths_1[_a3];
      parts.push.apply(parts, path.split("/"));
    }
    return uri.with({ path: normalizePath(parts) }).toString();
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/pathCompletion.js
  var __awaiter2 = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var __generator = function(thisArg, body) {
    var _ = { label: 0, sent: function() {
      if (t[0] & 1)
        throw t[1];
      return t[1];
    }, trys: [], ops: [] }, f2, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() {
      return this;
    }), g;
    function verb(n) {
      return function(v) {
        return step([n, v]);
      };
    }
    function step(op) {
      if (f2)
        throw new TypeError("Generator is already executing.");
      while (_)
        try {
          if (f2 = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
            return t;
          if (y = 0, t)
            op = [op[0] & 2, t.value];
          switch (op[0]) {
            case 0:
            case 1:
              t = op;
              break;
            case 4:
              _.label++;
              return { value: op[1], done: false };
            case 5:
              _.label++;
              y = op[1];
              op = [0];
              continue;
            case 7:
              op = _.ops.pop();
              _.trys.pop();
              continue;
            default:
              if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                _ = 0;
                continue;
              }
              if (op[0] === 3 && (!t || op[1] > t[0] && op[1] < t[3])) {
                _.label = op[1];
                break;
              }
              if (op[0] === 6 && _.label < t[1]) {
                _.label = t[1];
                t = op;
                break;
              }
              if (t && _.label < t[2]) {
                _.label = t[2];
                _.ops.push(op);
                break;
              }
              if (t[2])
                _.ops.pop();
              _.trys.pop();
              continue;
          }
          op = body.call(thisArg, _);
        } catch (e) {
          op = [6, e];
          y = 0;
        } finally {
          f2 = t = 0;
        }
      if (op[0] & 5)
        throw op[1];
      return { value: op[0] ? op[1] : void 0, done: true };
    }
  };
  var PathCompletionParticipant = function() {
    function PathCompletionParticipant2(readDirectory) {
      this.readDirectory = readDirectory;
      this.literalCompletions = [];
      this.importCompletions = [];
    }
    PathCompletionParticipant2.prototype.onCssURILiteralValue = function(context) {
      this.literalCompletions.push(context);
    };
    PathCompletionParticipant2.prototype.onCssImportPath = function(context) {
      this.importCompletions.push(context);
    };
    PathCompletionParticipant2.prototype.computeCompletions = function(document, documentContext) {
      return __awaiter2(this, void 0, void 0, function() {
        var result, _i, _a3, literalCompletion, uriValue, fullValue, items, _b, items_1, item, _c, _d, importCompletion, pathValue, fullValue, suggestions, _e, suggestions_1, item;
        return __generator(this, function(_f2) {
          switch (_f2.label) {
            case 0:
              result = { items: [], isIncomplete: false };
              _i = 0, _a3 = this.literalCompletions;
              _f2.label = 1;
            case 1:
              if (!(_i < _a3.length))
                return [3, 5];
              literalCompletion = _a3[_i];
              uriValue = literalCompletion.uriValue;
              fullValue = stripQuotes(uriValue);
              if (!(fullValue === "." || fullValue === ".."))
                return [3, 2];
              result.isIncomplete = true;
              return [3, 4];
            case 2:
              return [4, this.providePathSuggestions(uriValue, literalCompletion.position, literalCompletion.range, document, documentContext)];
            case 3:
              items = _f2.sent();
              for (_b = 0, items_1 = items; _b < items_1.length; _b++) {
                item = items_1[_b];
                result.items.push(item);
              }
              _f2.label = 4;
            case 4:
              _i++;
              return [3, 1];
            case 5:
              _c = 0, _d = this.importCompletions;
              _f2.label = 6;
            case 6:
              if (!(_c < _d.length))
                return [3, 10];
              importCompletion = _d[_c];
              pathValue = importCompletion.pathValue;
              fullValue = stripQuotes(pathValue);
              if (!(fullValue === "." || fullValue === ".."))
                return [3, 7];
              result.isIncomplete = true;
              return [3, 9];
            case 7:
              return [4, this.providePathSuggestions(pathValue, importCompletion.position, importCompletion.range, document, documentContext)];
            case 8:
              suggestions = _f2.sent();
              if (document.languageId === "scss") {
                suggestions.forEach(function(s) {
                  if (startsWith(s.label, "_") && endsWith(s.label, ".scss")) {
                    if (s.textEdit) {
                      s.textEdit.newText = s.label.slice(1, -5);
                    } else {
                      s.label = s.label.slice(1, -5);
                    }
                  }
                });
              }
              for (_e = 0, suggestions_1 = suggestions; _e < suggestions_1.length; _e++) {
                item = suggestions_1[_e];
                result.items.push(item);
              }
              _f2.label = 9;
            case 9:
              _c++;
              return [3, 6];
            case 10:
              return [2, result];
          }
        });
      });
    };
    PathCompletionParticipant2.prototype.providePathSuggestions = function(pathValue, position, range, document, documentContext) {
      return __awaiter2(this, void 0, void 0, function() {
        var fullValue, isValueQuoted, valueBeforeCursor, currentDocUri, fullValueRange, replaceRange, valueBeforeLastSlash, parentDir, result, infos, _i, infos_1, _a3, name, type, e_1;
        return __generator(this, function(_b) {
          switch (_b.label) {
            case 0:
              fullValue = stripQuotes(pathValue);
              isValueQuoted = startsWith(pathValue, "'") || startsWith(pathValue, '"');
              valueBeforeCursor = isValueQuoted ? fullValue.slice(0, position.character - (range.start.character + 1)) : fullValue.slice(0, position.character - range.start.character);
              currentDocUri = document.uri;
              fullValueRange = isValueQuoted ? shiftRange(range, 1, -1) : range;
              replaceRange = pathToReplaceRange(valueBeforeCursor, fullValue, fullValueRange);
              valueBeforeLastSlash = valueBeforeCursor.substring(0, valueBeforeCursor.lastIndexOf("/") + 1);
              parentDir = documentContext.resolveReference(valueBeforeLastSlash || ".", currentDocUri);
              if (!parentDir)
                return [3, 4];
              _b.label = 1;
            case 1:
              _b.trys.push([1, 3, , 4]);
              result = [];
              return [4, this.readDirectory(parentDir)];
            case 2:
              infos = _b.sent();
              for (_i = 0, infos_1 = infos; _i < infos_1.length; _i++) {
                _a3 = infos_1[_i], name = _a3[0], type = _a3[1];
                if (name.charCodeAt(0) !== CharCode_dot && (type === FileType.Directory || joinPath(parentDir, name) !== currentDocUri)) {
                  result.push(createCompletionItem(name, type === FileType.Directory, replaceRange));
                }
              }
              return [2, result];
            case 3:
              e_1 = _b.sent();
              return [3, 4];
            case 4:
              return [2, []];
          }
        });
      });
    };
    return PathCompletionParticipant2;
  }();
  var CharCode_dot = ".".charCodeAt(0);
  function stripQuotes(fullValue) {
    if (startsWith(fullValue, "'") || startsWith(fullValue, '"')) {
      return fullValue.slice(1, -1);
    } else {
      return fullValue;
    }
  }
  function pathToReplaceRange(valueBeforeCursor, fullValue, fullValueRange) {
    var replaceRange;
    var lastIndexOfSlash = valueBeforeCursor.lastIndexOf("/");
    if (lastIndexOfSlash === -1) {
      replaceRange = fullValueRange;
    } else {
      var valueAfterLastSlash = fullValue.slice(lastIndexOfSlash + 1);
      var startPos = shiftPosition(fullValueRange.end, -valueAfterLastSlash.length);
      var whitespaceIndex = valueAfterLastSlash.indexOf(" ");
      var endPos = void 0;
      if (whitespaceIndex !== -1) {
        endPos = shiftPosition(startPos, whitespaceIndex);
      } else {
        endPos = fullValueRange.end;
      }
      replaceRange = Range2.create(startPos, endPos);
    }
    return replaceRange;
  }
  function createCompletionItem(name, isDir, replaceRange) {
    if (isDir) {
      name = name + "/";
      return {
        label: escapePath(name),
        kind: CompletionItemKind2.Folder,
        textEdit: TextEdit.replace(replaceRange, escapePath(name)),
        command: {
          title: "Suggest",
          command: "editor.action.triggerSuggest"
        }
      };
    } else {
      return {
        label: escapePath(name),
        kind: CompletionItemKind2.File,
        textEdit: TextEdit.replace(replaceRange, escapePath(name))
      };
    }
  }
  function escapePath(p) {
    return p.replace(/(\s|\(|\)|,|"|')/g, "\\$1");
  }
  function shiftPosition(pos, offset) {
    return Position2.create(pos.line, pos.character + offset);
  }
  function shiftRange(range, startOffset, endOffset) {
    var start = shiftPosition(range.start, startOffset);
    var end = shiftPosition(range.end, endOffset);
    return Range2.create(start, end);
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssCompletion.js
  var __awaiter3 = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var __generator2 = function(thisArg, body) {
    var _ = { label: 0, sent: function() {
      if (t[0] & 1)
        throw t[1];
      return t[1];
    }, trys: [], ops: [] }, f2, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() {
      return this;
    }), g;
    function verb(n) {
      return function(v) {
        return step([n, v]);
      };
    }
    function step(op) {
      if (f2)
        throw new TypeError("Generator is already executing.");
      while (_)
        try {
          if (f2 = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
            return t;
          if (y = 0, t)
            op = [op[0] & 2, t.value];
          switch (op[0]) {
            case 0:
            case 1:
              t = op;
              break;
            case 4:
              _.label++;
              return { value: op[1], done: false };
            case 5:
              _.label++;
              y = op[1];
              op = [0];
              continue;
            case 7:
              op = _.ops.pop();
              _.trys.pop();
              continue;
            default:
              if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                _ = 0;
                continue;
              }
              if (op[0] === 3 && (!t || op[1] > t[0] && op[1] < t[3])) {
                _.label = op[1];
                break;
              }
              if (op[0] === 6 && _.label < t[1]) {
                _.label = t[1];
                t = op;
                break;
              }
              if (t && _.label < t[2]) {
                _.label = t[2];
                _.ops.push(op);
                break;
              }
              if (t[2])
                _.ops.pop();
              _.trys.pop();
              continue;
          }
          op = body.call(thisArg, _);
        } catch (e) {
          op = [6, e];
          y = 0;
        } finally {
          f2 = t = 0;
        }
      if (op[0] & 5)
        throw op[1];
      return { value: op[0] ? op[1] : void 0, done: true };
    }
  };
  var localize4 = loadMessageBundle();
  var SnippetFormat = InsertTextFormat.Snippet;
  var SortTexts;
  (function(SortTexts2) {
    SortTexts2["Enums"] = " ";
    SortTexts2["Normal"] = "d";
    SortTexts2["VendorPrefixed"] = "x";
    SortTexts2["Term"] = "y";
    SortTexts2["Variable"] = "z";
  })(SortTexts || (SortTexts = {}));
  var CSSCompletion = function() {
    function CSSCompletion2(variablePrefix, lsOptions, cssDataManager) {
      if (variablePrefix === void 0) {
        variablePrefix = null;
      }
      this.variablePrefix = variablePrefix;
      this.lsOptions = lsOptions;
      this.cssDataManager = cssDataManager;
      this.completionParticipants = [];
    }
    CSSCompletion2.prototype.configure = function(settings) {
      this.settings = settings;
    };
    CSSCompletion2.prototype.getSymbolContext = function() {
      if (!this.symbolContext) {
        this.symbolContext = new Symbols(this.styleSheet);
      }
      return this.symbolContext;
    };
    CSSCompletion2.prototype.setCompletionParticipants = function(registeredCompletionParticipants) {
      this.completionParticipants = registeredCompletionParticipants || [];
    };
    CSSCompletion2.prototype.doComplete2 = function(document, position, styleSheet, documentContext) {
      return __awaiter3(this, void 0, void 0, function() {
        var participant, contributedParticipants, result, pathCompletionResult;
        return __generator2(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              if (!this.lsOptions.fileSystemProvider || !this.lsOptions.fileSystemProvider.readDirectory) {
                return [2, this.doComplete(document, position, styleSheet)];
              }
              participant = new PathCompletionParticipant(this.lsOptions.fileSystemProvider.readDirectory);
              contributedParticipants = this.completionParticipants;
              this.completionParticipants = [participant].concat(contributedParticipants);
              result = this.doComplete(document, position, styleSheet);
              _a3.label = 1;
            case 1:
              _a3.trys.push([1, , 3, 4]);
              return [4, participant.computeCompletions(document, documentContext)];
            case 2:
              pathCompletionResult = _a3.sent();
              return [2, {
                isIncomplete: result.isIncomplete || pathCompletionResult.isIncomplete,
                items: pathCompletionResult.items.concat(result.items)
              }];
            case 3:
              this.completionParticipants = contributedParticipants;
              return [7];
            case 4:
              return [2];
          }
        });
      });
    };
    CSSCompletion2.prototype.doComplete = function(document, position, styleSheet) {
      this.offset = document.offsetAt(position);
      this.position = position;
      this.currentWord = getCurrentWord(document, this.offset);
      this.defaultReplaceRange = Range2.create(Position2.create(this.position.line, this.position.character - this.currentWord.length), this.position);
      this.textDocument = document;
      this.styleSheet = styleSheet;
      try {
        var result = { isIncomplete: false, items: [] };
        this.nodePath = getNodePath(this.styleSheet, this.offset);
        for (var i = this.nodePath.length - 1; i >= 0; i--) {
          var node = this.nodePath[i];
          if (node instanceof Property) {
            this.getCompletionsForDeclarationProperty(node.getParent(), result);
          } else if (node instanceof Expression) {
            if (node.parent instanceof Interpolation) {
              this.getVariableProposals(null, result);
            } else {
              this.getCompletionsForExpression(node, result);
            }
          } else if (node instanceof SimpleSelector) {
            var parentRef = node.findAParent(NodeType.ExtendsReference, NodeType.Ruleset);
            if (parentRef) {
              if (parentRef.type === NodeType.ExtendsReference) {
                this.getCompletionsForExtendsReference(parentRef, node, result);
              } else {
                var parentRuleSet = parentRef;
                this.getCompletionsForSelector(parentRuleSet, parentRuleSet && parentRuleSet.isNested(), result);
              }
            }
          } else if (node instanceof FunctionArgument) {
            this.getCompletionsForFunctionArgument(node, node.getParent(), result);
          } else if (node instanceof Declarations) {
            this.getCompletionsForDeclarations(node, result);
          } else if (node instanceof VariableDeclaration) {
            this.getCompletionsForVariableDeclaration(node, result);
          } else if (node instanceof RuleSet) {
            this.getCompletionsForRuleSet(node, result);
          } else if (node instanceof Interpolation) {
            this.getCompletionsForInterpolation(node, result);
          } else if (node instanceof FunctionDeclaration) {
            this.getCompletionsForFunctionDeclaration(node, result);
          } else if (node instanceof MixinReference) {
            this.getCompletionsForMixinReference(node, result);
          } else if (node instanceof Function) {
            this.getCompletionsForFunctionArgument(null, node, result);
          } else if (node instanceof Supports) {
            this.getCompletionsForSupports(node, result);
          } else if (node instanceof SupportsCondition) {
            this.getCompletionsForSupportsCondition(node, result);
          } else if (node instanceof ExtendsReference) {
            this.getCompletionsForExtendsReference(node, null, result);
          } else if (node.type === NodeType.URILiteral) {
            this.getCompletionForUriLiteralValue(node, result);
          } else if (node.parent === null) {
            this.getCompletionForTopLevel(result);
          } else if (node.type === NodeType.StringLiteral && this.isImportPathParent(node.parent.type)) {
            this.getCompletionForImportPath(node, result);
          } else {
            continue;
          }
          if (result.items.length > 0 || this.offset > node.offset) {
            return this.finalize(result);
          }
        }
        this.getCompletionsForStylesheet(result);
        if (result.items.length === 0) {
          if (this.variablePrefix && this.currentWord.indexOf(this.variablePrefix) === 0) {
            this.getVariableProposals(null, result);
          }
        }
        return this.finalize(result);
      } finally {
        this.position = null;
        this.currentWord = null;
        this.textDocument = null;
        this.styleSheet = null;
        this.symbolContext = null;
        this.defaultReplaceRange = null;
        this.nodePath = null;
      }
    };
    CSSCompletion2.prototype.isImportPathParent = function(type) {
      return type === NodeType.Import;
    };
    CSSCompletion2.prototype.finalize = function(result) {
      return result;
    };
    CSSCompletion2.prototype.findInNodePath = function() {
      var types = [];
      for (var _i = 0; _i < arguments.length; _i++) {
        types[_i] = arguments[_i];
      }
      for (var i = this.nodePath.length - 1; i >= 0; i--) {
        var node = this.nodePath[i];
        if (types.indexOf(node.type) !== -1) {
          return node;
        }
      }
      return null;
    };
    CSSCompletion2.prototype.getCompletionsForDeclarationProperty = function(declaration, result) {
      return this.getPropertyProposals(declaration, result);
    };
    CSSCompletion2.prototype.getPropertyProposals = function(declaration, result) {
      var _this = this;
      var triggerPropertyValueCompletion = this.isTriggerPropertyValueCompletionEnabled;
      var completePropertyWithSemicolon = this.isCompletePropertyWithSemicolonEnabled;
      var properties = this.cssDataManager.getProperties();
      properties.forEach(function(entry) {
        var range;
        var insertText;
        var retrigger = false;
        if (declaration) {
          range = _this.getCompletionRange(declaration.getProperty());
          insertText = entry.name;
          if (!isDefined(declaration.colonPosition)) {
            insertText += ": ";
            retrigger = true;
          }
        } else {
          range = _this.getCompletionRange(null);
          insertText = entry.name + ": ";
          retrigger = true;
        }
        if (!declaration && completePropertyWithSemicolon) {
          insertText += "$0;";
        }
        if (declaration && !declaration.semicolonPosition) {
          if (completePropertyWithSemicolon && _this.offset >= _this.textDocument.offsetAt(range.end)) {
            insertText += "$0;";
          }
        }
        var item = {
          label: entry.name,
          documentation: getEntryDescription(entry, _this.doesSupportMarkdown()),
          tags: isDeprecated(entry) ? [CompletionItemTag2.Deprecated] : [],
          textEdit: TextEdit.replace(range, insertText),
          insertTextFormat: InsertTextFormat.Snippet,
          kind: CompletionItemKind2.Property
        };
        if (!entry.restrictions) {
          retrigger = false;
        }
        if (triggerPropertyValueCompletion && retrigger) {
          item.command = {
            title: "Suggest",
            command: "editor.action.triggerSuggest"
          };
        }
        var relevance = typeof entry.relevance === "number" ? Math.min(Math.max(entry.relevance, 0), 99) : 50;
        var sortTextSuffix = (255 - relevance).toString(16);
        var sortTextPrefix = startsWith(entry.name, "-") ? SortTexts.VendorPrefixed : SortTexts.Normal;
        item.sortText = sortTextPrefix + "_" + sortTextSuffix;
        result.items.push(item);
      });
      this.completionParticipants.forEach(function(participant) {
        if (participant.onCssProperty) {
          participant.onCssProperty({
            propertyName: _this.currentWord,
            range: _this.defaultReplaceRange
          });
        }
      });
      return result;
    };
    Object.defineProperty(CSSCompletion2.prototype, "isTriggerPropertyValueCompletionEnabled", {
      get: function() {
        if (!this.settings || !this.settings.completion || this.settings.completion.triggerPropertyValueCompletion === void 0) {
          return true;
        }
        return this.settings.completion.triggerPropertyValueCompletion;
      },
      enumerable: true,
      configurable: true
    });
    Object.defineProperty(CSSCompletion2.prototype, "isCompletePropertyWithSemicolonEnabled", {
      get: function() {
        if (!this.settings || !this.settings.completion || this.settings.completion.completePropertyWithSemicolon === void 0) {
          return true;
        }
        return this.settings.completion.completePropertyWithSemicolon;
      },
      enumerable: true,
      configurable: true
    });
    CSSCompletion2.prototype.getCompletionsForDeclarationValue = function(node, result) {
      var _this = this;
      var propertyName = node.getFullPropertyName();
      var entry = this.cssDataManager.getProperty(propertyName);
      var existingNode = node.getValue() || null;
      while (existingNode && existingNode.hasChildren()) {
        existingNode = existingNode.findChildAtOffset(this.offset, false);
      }
      this.completionParticipants.forEach(function(participant) {
        if (participant.onCssPropertyValue) {
          participant.onCssPropertyValue({
            propertyName,
            propertyValue: _this.currentWord,
            range: _this.getCompletionRange(existingNode)
          });
        }
      });
      if (entry) {
        if (entry.restrictions) {
          for (var _i = 0, _a3 = entry.restrictions; _i < _a3.length; _i++) {
            var restriction = _a3[_i];
            switch (restriction) {
              case "color":
                this.getColorProposals(entry, existingNode, result);
                break;
              case "position":
                this.getPositionProposals(entry, existingNode, result);
                break;
              case "repeat":
                this.getRepeatStyleProposals(entry, existingNode, result);
                break;
              case "line-style":
                this.getLineStyleProposals(entry, existingNode, result);
                break;
              case "line-width":
                this.getLineWidthProposals(entry, existingNode, result);
                break;
              case "geometry-box":
                this.getGeometryBoxProposals(entry, existingNode, result);
                break;
              case "box":
                this.getBoxProposals(entry, existingNode, result);
                break;
              case "image":
                this.getImageProposals(entry, existingNode, result);
                break;
              case "timing-function":
                this.getTimingFunctionProposals(entry, existingNode, result);
                break;
              case "shape":
                this.getBasicShapeProposals(entry, existingNode, result);
                break;
            }
          }
        }
        this.getValueEnumProposals(entry, existingNode, result);
        this.getCSSWideKeywordProposals(entry, existingNode, result);
        this.getUnitProposals(entry, existingNode, result);
      } else {
        var existingValues = collectValues(this.styleSheet, node);
        for (var _b = 0, _c = existingValues.getEntries(); _b < _c.length; _b++) {
          var existingValue = _c[_b];
          result.items.push({
            label: existingValue,
            textEdit: TextEdit.replace(this.getCompletionRange(existingNode), existingValue),
            kind: CompletionItemKind2.Value
          });
        }
      }
      this.getVariableProposals(existingNode, result);
      this.getTermProposals(entry, existingNode, result);
      return result;
    };
    CSSCompletion2.prototype.getValueEnumProposals = function(entry, existingNode, result) {
      if (entry.values) {
        for (var _i = 0, _a3 = entry.values; _i < _a3.length; _i++) {
          var value = _a3[_i];
          var insertString = value.name;
          var insertTextFormat = void 0;
          if (endsWith(insertString, ")")) {
            var from = insertString.lastIndexOf("(");
            if (from !== -1) {
              insertString = insertString.substr(0, from) + "($1)";
              insertTextFormat = SnippetFormat;
            }
          }
          var sortText = SortTexts.Enums;
          if (startsWith(value.name, "-")) {
            sortText += SortTexts.VendorPrefixed;
          }
          var item = {
            label: value.name,
            documentation: getEntryDescription(value, this.doesSupportMarkdown()),
            tags: isDeprecated(entry) ? [CompletionItemTag2.Deprecated] : [],
            textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertString),
            sortText,
            kind: CompletionItemKind2.Value,
            insertTextFormat
          };
          result.items.push(item);
        }
      }
      return result;
    };
    CSSCompletion2.prototype.getCSSWideKeywordProposals = function(entry, existingNode, result) {
      for (var keywords in cssWideKeywords) {
        result.items.push({
          label: keywords,
          documentation: cssWideKeywords[keywords],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), keywords),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForInterpolation = function(node, result) {
      if (this.offset >= node.offset + 2) {
        this.getVariableProposals(null, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getVariableProposals = function(existingNode, result) {
      var symbols = this.getSymbolContext().findSymbolsAtOffset(this.offset, ReferenceType.Variable);
      for (var _i = 0, symbols_1 = symbols; _i < symbols_1.length; _i++) {
        var symbol = symbols_1[_i];
        var insertText = startsWith(symbol.name, "--") ? "var(" + symbol.name + ")" : symbol.name;
        var completionItem = {
          label: symbol.name,
          documentation: symbol.value ? getLimitedString(symbol.value) : symbol.value,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
          kind: CompletionItemKind2.Variable,
          sortText: SortTexts.Variable
        };
        if (typeof completionItem.documentation === "string" && isColorString(completionItem.documentation)) {
          completionItem.kind = CompletionItemKind2.Color;
        }
        if (symbol.node.type === NodeType.FunctionParameter) {
          var mixinNode = symbol.node.getParent();
          if (mixinNode.type === NodeType.MixinDeclaration) {
            completionItem.detail = localize4("completion.argument", "argument from '{0}'", mixinNode.getName());
          }
        }
        result.items.push(completionItem);
      }
      return result;
    };
    CSSCompletion2.prototype.getVariableProposalsForCSSVarFunction = function(result) {
      var symbols = this.getSymbolContext().findSymbolsAtOffset(this.offset, ReferenceType.Variable);
      symbols = symbols.filter(function(symbol2) {
        return startsWith(symbol2.name, "--");
      });
      for (var _i = 0, symbols_2 = symbols; _i < symbols_2.length; _i++) {
        var symbol = symbols_2[_i];
        var completionItem = {
          label: symbol.name,
          documentation: symbol.value ? getLimitedString(symbol.value) : symbol.value,
          textEdit: TextEdit.replace(this.getCompletionRange(null), symbol.name),
          kind: CompletionItemKind2.Variable
        };
        if (typeof completionItem.documentation === "string" && isColorString(completionItem.documentation)) {
          completionItem.kind = CompletionItemKind2.Color;
        }
        result.items.push(completionItem);
      }
      return result;
    };
    CSSCompletion2.prototype.getUnitProposals = function(entry, existingNode, result) {
      var currentWord = "0";
      if (this.currentWord.length > 0) {
        var numMatch = this.currentWord.match(/^-?\d[\.\d+]*/);
        if (numMatch) {
          currentWord = numMatch[0];
          result.isIncomplete = currentWord.length === this.currentWord.length;
        }
      } else if (this.currentWord.length === 0) {
        result.isIncomplete = true;
      }
      if (existingNode && existingNode.parent && existingNode.parent.type === NodeType.Term) {
        existingNode = existingNode.getParent();
      }
      if (entry.restrictions) {
        for (var _i = 0, _a3 = entry.restrictions; _i < _a3.length; _i++) {
          var restriction = _a3[_i];
          var units2 = units[restriction];
          if (units2) {
            for (var _b = 0, units_1 = units2; _b < units_1.length; _b++) {
              var unit = units_1[_b];
              var insertText = currentWord + unit;
              result.items.push({
                label: insertText,
                textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
                kind: CompletionItemKind2.Unit
              });
            }
          }
        }
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionRange = function(existingNode) {
      if (existingNode && existingNode.offset <= this.offset && this.offset <= existingNode.end) {
        var end = existingNode.end !== -1 ? this.textDocument.positionAt(existingNode.end) : this.position;
        var start = this.textDocument.positionAt(existingNode.offset);
        if (start.line === end.line) {
          return Range2.create(start, end);
        }
      }
      return this.defaultReplaceRange;
    };
    CSSCompletion2.prototype.getColorProposals = function(entry, existingNode, result) {
      for (var color in colors) {
        result.items.push({
          label: color,
          documentation: colors[color],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), color),
          kind: CompletionItemKind2.Color
        });
      }
      for (var color in colorKeywords) {
        result.items.push({
          label: color,
          documentation: colorKeywords[color],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), color),
          kind: CompletionItemKind2.Value
        });
      }
      var colorValues = new Set2();
      this.styleSheet.acceptVisitor(new ColorValueCollector(colorValues, this.offset));
      for (var _i = 0, _a3 = colorValues.getEntries(); _i < _a3.length; _i++) {
        var color = _a3[_i];
        result.items.push({
          label: color,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), color),
          kind: CompletionItemKind2.Color
        });
      }
      var _loop_1 = function(p2) {
        var tabStop = 1;
        var replaceFunction = function(_match, p1) {
          return "${" + tabStop++ + ":" + p1 + "}";
        };
        var insertText = p2.func.replace(/\[?\$(\w+)\]?/g, replaceFunction);
        result.items.push({
          label: p2.func.substr(0, p2.func.indexOf("(")),
          detail: p2.func,
          documentation: p2.desc,
          textEdit: TextEdit.replace(this_1.getCompletionRange(existingNode), insertText),
          insertTextFormat: SnippetFormat,
          kind: CompletionItemKind2.Function
        });
      };
      var this_1 = this;
      for (var _b = 0, _c = colorFunctions; _b < _c.length; _b++) {
        var p = _c[_b];
        _loop_1(p);
      }
      return result;
    };
    CSSCompletion2.prototype.getPositionProposals = function(entry, existingNode, result) {
      for (var position in positionKeywords) {
        result.items.push({
          label: position,
          documentation: positionKeywords[position],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), position),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getRepeatStyleProposals = function(entry, existingNode, result) {
      for (var repeat in repeatStyleKeywords) {
        result.items.push({
          label: repeat,
          documentation: repeatStyleKeywords[repeat],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), repeat),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getLineStyleProposals = function(entry, existingNode, result) {
      for (var lineStyle in lineStyleKeywords) {
        result.items.push({
          label: lineStyle,
          documentation: lineStyleKeywords[lineStyle],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), lineStyle),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getLineWidthProposals = function(entry, existingNode, result) {
      for (var _i = 0, _a3 = lineWidthKeywords; _i < _a3.length; _i++) {
        var lineWidth = _a3[_i];
        result.items.push({
          label: lineWidth,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), lineWidth),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getGeometryBoxProposals = function(entry, existingNode, result) {
      for (var box in geometryBoxKeywords) {
        result.items.push({
          label: box,
          documentation: geometryBoxKeywords[box],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), box),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getBoxProposals = function(entry, existingNode, result) {
      for (var box in boxKeywords) {
        result.items.push({
          label: box,
          documentation: boxKeywords[box],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), box),
          kind: CompletionItemKind2.Value
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getImageProposals = function(entry, existingNode, result) {
      for (var image in imageFunctions) {
        var insertText = moveCursorInsideParenthesis(image);
        result.items.push({
          label: image,
          documentation: imageFunctions[image],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
          kind: CompletionItemKind2.Function,
          insertTextFormat: image !== insertText ? SnippetFormat : void 0
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getTimingFunctionProposals = function(entry, existingNode, result) {
      for (var timing in transitionTimingFunctions) {
        var insertText = moveCursorInsideParenthesis(timing);
        result.items.push({
          label: timing,
          documentation: transitionTimingFunctions[timing],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
          kind: CompletionItemKind2.Function,
          insertTextFormat: timing !== insertText ? SnippetFormat : void 0
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getBasicShapeProposals = function(entry, existingNode, result) {
      for (var shape in basicShapeFunctions) {
        var insertText = moveCursorInsideParenthesis(shape);
        result.items.push({
          label: shape,
          documentation: basicShapeFunctions[shape],
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
          kind: CompletionItemKind2.Function,
          insertTextFormat: shape !== insertText ? SnippetFormat : void 0
        });
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForStylesheet = function(result) {
      var node = this.styleSheet.findFirstChildBeforeOffset(this.offset);
      if (!node) {
        return this.getCompletionForTopLevel(result);
      }
      if (node instanceof RuleSet) {
        return this.getCompletionsForRuleSet(node, result);
      }
      if (node instanceof Supports) {
        return this.getCompletionsForSupports(node, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionForTopLevel = function(result) {
      var _this = this;
      this.cssDataManager.getAtDirectives().forEach(function(entry) {
        result.items.push({
          label: entry.name,
          textEdit: TextEdit.replace(_this.getCompletionRange(null), entry.name),
          documentation: getEntryDescription(entry, _this.doesSupportMarkdown()),
          tags: isDeprecated(entry) ? [CompletionItemTag2.Deprecated] : [],
          kind: CompletionItemKind2.Keyword
        });
      });
      this.getCompletionsForSelector(null, false, result);
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForRuleSet = function(ruleSet, result) {
      var declarations = ruleSet.getDeclarations();
      var isAfter = declarations && declarations.endsWith("}") && this.offset >= declarations.end;
      if (isAfter) {
        return this.getCompletionForTopLevel(result);
      }
      var isInSelectors = !declarations || this.offset <= declarations.offset;
      if (isInSelectors) {
        return this.getCompletionsForSelector(ruleSet, ruleSet.isNested(), result);
      }
      return this.getCompletionsForDeclarations(ruleSet.getDeclarations(), result);
    };
    CSSCompletion2.prototype.getCompletionsForSelector = function(ruleSet, isNested, result) {
      var _this = this;
      var existingNode = this.findInNodePath(NodeType.PseudoSelector, NodeType.IdentifierSelector, NodeType.ClassSelector, NodeType.ElementNameSelector);
      if (!existingNode && this.offset - this.currentWord.length > 0 && this.textDocument.getText()[this.offset - this.currentWord.length - 1] === ":") {
        this.currentWord = ":" + this.currentWord;
        this.defaultReplaceRange = Range2.create(Position2.create(this.position.line, this.position.character - this.currentWord.length), this.position);
      }
      var pseudoClasses = this.cssDataManager.getPseudoClasses();
      pseudoClasses.forEach(function(entry2) {
        var insertText = moveCursorInsideParenthesis(entry2.name);
        var item = {
          label: entry2.name,
          textEdit: TextEdit.replace(_this.getCompletionRange(existingNode), insertText),
          documentation: getEntryDescription(entry2, _this.doesSupportMarkdown()),
          tags: isDeprecated(entry2) ? [CompletionItemTag2.Deprecated] : [],
          kind: CompletionItemKind2.Function,
          insertTextFormat: entry2.name !== insertText ? SnippetFormat : void 0
        };
        if (startsWith(entry2.name, ":-")) {
          item.sortText = SortTexts.VendorPrefixed;
        }
        result.items.push(item);
      });
      var pseudoElements = this.cssDataManager.getPseudoElements();
      pseudoElements.forEach(function(entry2) {
        var insertText = moveCursorInsideParenthesis(entry2.name);
        var item = {
          label: entry2.name,
          textEdit: TextEdit.replace(_this.getCompletionRange(existingNode), insertText),
          documentation: getEntryDescription(entry2, _this.doesSupportMarkdown()),
          tags: isDeprecated(entry2) ? [CompletionItemTag2.Deprecated] : [],
          kind: CompletionItemKind2.Function,
          insertTextFormat: entry2.name !== insertText ? SnippetFormat : void 0
        };
        if (startsWith(entry2.name, "::-")) {
          item.sortText = SortTexts.VendorPrefixed;
        }
        result.items.push(item);
      });
      if (!isNested) {
        for (var _i = 0, _a3 = html5Tags; _i < _a3.length; _i++) {
          var entry = _a3[_i];
          result.items.push({
            label: entry,
            textEdit: TextEdit.replace(this.getCompletionRange(existingNode), entry),
            kind: CompletionItemKind2.Keyword
          });
        }
        for (var _b = 0, _c = svgElements; _b < _c.length; _b++) {
          var entry = _c[_b];
          result.items.push({
            label: entry,
            textEdit: TextEdit.replace(this.getCompletionRange(existingNode), entry),
            kind: CompletionItemKind2.Keyword
          });
        }
      }
      var visited = {};
      visited[this.currentWord] = true;
      var docText = this.textDocument.getText();
      this.styleSheet.accept(function(n) {
        if (n.type === NodeType.SimpleSelector && n.length > 0) {
          var selector2 = docText.substr(n.offset, n.length);
          if (selector2.charAt(0) === "." && !visited[selector2]) {
            visited[selector2] = true;
            result.items.push({
              label: selector2,
              textEdit: TextEdit.replace(_this.getCompletionRange(existingNode), selector2),
              kind: CompletionItemKind2.Keyword
            });
          }
          return false;
        }
        return true;
      });
      if (ruleSet && ruleSet.isNested()) {
        var selector = ruleSet.getSelectors().findFirstChildBeforeOffset(this.offset);
        if (selector && ruleSet.getSelectors().getChildren().indexOf(selector) === 0) {
          this.getPropertyProposals(null, result);
        }
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForDeclarations = function(declarations, result) {
      if (!declarations || this.offset === declarations.offset) {
        return result;
      }
      var node = declarations.findFirstChildBeforeOffset(this.offset);
      if (!node) {
        return this.getCompletionsForDeclarationProperty(null, result);
      }
      if (node instanceof AbstractDeclaration) {
        var declaration = node;
        if (!isDefined(declaration.colonPosition) || this.offset <= declaration.colonPosition) {
          return this.getCompletionsForDeclarationProperty(declaration, result);
        } else if (isDefined(declaration.semicolonPosition) && declaration.semicolonPosition < this.offset) {
          if (this.offset === declaration.semicolonPosition + 1) {
            return result;
          }
          return this.getCompletionsForDeclarationProperty(null, result);
        }
        if (declaration instanceof Declaration) {
          return this.getCompletionsForDeclarationValue(declaration, result);
        }
      } else if (node instanceof ExtendsReference) {
        this.getCompletionsForExtendsReference(node, null, result);
      } else if (this.currentWord && this.currentWord[0] === "@") {
        this.getCompletionsForDeclarationProperty(null, result);
      } else if (node instanceof RuleSet) {
        this.getCompletionsForDeclarationProperty(null, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForVariableDeclaration = function(declaration, result) {
      if (this.offset && isDefined(declaration.colonPosition) && this.offset > declaration.colonPosition) {
        this.getVariableProposals(declaration.getValue(), result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForExpression = function(expression, result) {
      var parent = expression.getParent();
      if (parent instanceof FunctionArgument) {
        this.getCompletionsForFunctionArgument(parent, parent.getParent(), result);
        return result;
      }
      var declaration = expression.findParent(NodeType.Declaration);
      if (!declaration) {
        this.getTermProposals(void 0, null, result);
        return result;
      }
      var node = expression.findChildAtOffset(this.offset, true);
      if (!node) {
        return this.getCompletionsForDeclarationValue(declaration, result);
      }
      if (node instanceof NumericValue || node instanceof Identifier) {
        return this.getCompletionsForDeclarationValue(declaration, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForFunctionArgument = function(arg, func, result) {
      var identifier = func.getIdentifier();
      if (identifier && identifier.matches("var")) {
        if (!func.getArguments().hasChildren() || func.getArguments().getChild(0) === arg) {
          this.getVariableProposalsForCSSVarFunction(result);
        }
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForFunctionDeclaration = function(decl, result) {
      var declarations = decl.getDeclarations();
      if (declarations && this.offset > declarations.offset && this.offset < declarations.end) {
        this.getTermProposals(void 0, null, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForMixinReference = function(ref, result) {
      var _this = this;
      var allMixins = this.getSymbolContext().findSymbolsAtOffset(this.offset, ReferenceType.Mixin);
      for (var _i = 0, allMixins_1 = allMixins; _i < allMixins_1.length; _i++) {
        var mixinSymbol = allMixins_1[_i];
        if (mixinSymbol.node instanceof MixinDeclaration) {
          result.items.push(this.makeTermProposal(mixinSymbol, mixinSymbol.node.getParameters(), null));
        }
      }
      var identifierNode = ref.getIdentifier() || null;
      this.completionParticipants.forEach(function(participant) {
        if (participant.onCssMixinReference) {
          participant.onCssMixinReference({
            mixinName: _this.currentWord,
            range: _this.getCompletionRange(identifierNode)
          });
        }
      });
      return result;
    };
    CSSCompletion2.prototype.getTermProposals = function(entry, existingNode, result) {
      var allFunctions = this.getSymbolContext().findSymbolsAtOffset(this.offset, ReferenceType.Function);
      for (var _i = 0, allFunctions_1 = allFunctions; _i < allFunctions_1.length; _i++) {
        var functionSymbol = allFunctions_1[_i];
        if (functionSymbol.node instanceof FunctionDeclaration) {
          result.items.push(this.makeTermProposal(functionSymbol, functionSymbol.node.getParameters(), existingNode));
        }
      }
      return result;
    };
    CSSCompletion2.prototype.makeTermProposal = function(symbol, parameters, existingNode) {
      var decl = symbol.node;
      var params = parameters.getChildren().map(function(c) {
        return c instanceof FunctionParameter ? c.getName() : c.getText();
      });
      var insertText = symbol.name + "(" + params.map(function(p, index) {
        return "${" + (index + 1) + ":" + p + "}";
      }).join(", ") + ")";
      return {
        label: symbol.name,
        detail: symbol.name + "(" + params.join(", ") + ")",
        textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
        insertTextFormat: SnippetFormat,
        kind: CompletionItemKind2.Function,
        sortText: SortTexts.Term
      };
    };
    CSSCompletion2.prototype.getCompletionsForSupportsCondition = function(supportsCondition, result) {
      var child = supportsCondition.findFirstChildBeforeOffset(this.offset);
      if (child) {
        if (child instanceof Declaration) {
          if (!isDefined(child.colonPosition) || this.offset <= child.colonPosition) {
            return this.getCompletionsForDeclarationProperty(child, result);
          } else {
            return this.getCompletionsForDeclarationValue(child, result);
          }
        } else if (child instanceof SupportsCondition) {
          return this.getCompletionsForSupportsCondition(child, result);
        }
      }
      if (isDefined(supportsCondition.lParent) && this.offset > supportsCondition.lParent && (!isDefined(supportsCondition.rParent) || this.offset <= supportsCondition.rParent)) {
        return this.getCompletionsForDeclarationProperty(null, result);
      }
      return result;
    };
    CSSCompletion2.prototype.getCompletionsForSupports = function(supports, result) {
      var declarations = supports.getDeclarations();
      var inInCondition = !declarations || this.offset <= declarations.offset;
      if (inInCondition) {
        var child = supports.findFirstChildBeforeOffset(this.offset);
        if (child instanceof SupportsCondition) {
          return this.getCompletionsForSupportsCondition(child, result);
        }
        return result;
      }
      return this.getCompletionForTopLevel(result);
    };
    CSSCompletion2.prototype.getCompletionsForExtendsReference = function(extendsRef, existingNode, result) {
      return result;
    };
    CSSCompletion2.prototype.getCompletionForUriLiteralValue = function(uriLiteralNode, result) {
      var uriValue;
      var position;
      var range;
      if (!uriLiteralNode.hasChildren()) {
        uriValue = "";
        position = this.position;
        var emptyURIValuePosition = this.textDocument.positionAt(uriLiteralNode.offset + "url(".length);
        range = Range2.create(emptyURIValuePosition, emptyURIValuePosition);
      } else {
        var uriValueNode = uriLiteralNode.getChild(0);
        uriValue = uriValueNode.getText();
        position = this.position;
        range = this.getCompletionRange(uriValueNode);
      }
      this.completionParticipants.forEach(function(participant) {
        if (participant.onCssURILiteralValue) {
          participant.onCssURILiteralValue({
            uriValue,
            position,
            range
          });
        }
      });
      return result;
    };
    CSSCompletion2.prototype.getCompletionForImportPath = function(importPathNode, result) {
      var _this = this;
      this.completionParticipants.forEach(function(participant) {
        if (participant.onCssImportPath) {
          participant.onCssImportPath({
            pathValue: importPathNode.getText(),
            position: _this.position,
            range: _this.getCompletionRange(importPathNode)
          });
        }
      });
      return result;
    };
    CSSCompletion2.prototype.doesSupportMarkdown = function() {
      var _a3, _b, _c;
      if (!isDefined(this.supportsMarkdown)) {
        if (!isDefined(this.lsOptions.clientCapabilities)) {
          this.supportsMarkdown = true;
          return this.supportsMarkdown;
        }
        var documentationFormat = (_c = (_b = (_a3 = this.lsOptions.clientCapabilities.textDocument) === null || _a3 === void 0 ? void 0 : _a3.completion) === null || _b === void 0 ? void 0 : _b.completionItem) === null || _c === void 0 ? void 0 : _c.documentationFormat;
        this.supportsMarkdown = Array.isArray(documentationFormat) && documentationFormat.indexOf(MarkupKind.Markdown) !== -1;
      }
      return this.supportsMarkdown;
    };
    return CSSCompletion2;
  }();
  function isDeprecated(entry) {
    if (entry.status && (entry.status === "nonstandard" || entry.status === "obsolete")) {
      return true;
    }
    return false;
  }
  var Set2 = function() {
    function Set3() {
      this.entries = {};
    }
    Set3.prototype.add = function(entry) {
      this.entries[entry] = true;
    };
    Set3.prototype.getEntries = function() {
      return Object.keys(this.entries);
    };
    return Set3;
  }();
  function moveCursorInsideParenthesis(text) {
    return text.replace(/\(\)$/, "($1)");
  }
  function collectValues(styleSheet, declaration) {
    var fullPropertyName = declaration.getFullPropertyName();
    var entries = new Set2();
    function visitValue(node) {
      if (node instanceof Identifier || node instanceof NumericValue || node instanceof HexColorValue) {
        entries.add(node.getText());
      }
      return true;
    }
    function matchesProperty(decl) {
      var propertyName = decl.getFullPropertyName();
      return fullPropertyName === propertyName;
    }
    function vistNode(node) {
      if (node instanceof Declaration && node !== declaration) {
        if (matchesProperty(node)) {
          var value = node.getValue();
          if (value) {
            value.accept(visitValue);
          }
        }
      }
      return true;
    }
    styleSheet.accept(vistNode);
    return entries;
  }
  var ColorValueCollector = function() {
    function ColorValueCollector2(entries, currentOffset) {
      this.entries = entries;
      this.currentOffset = currentOffset;
    }
    ColorValueCollector2.prototype.visitNode = function(node) {
      if (node instanceof HexColorValue || node instanceof Function && isColorConstructor(node)) {
        if (this.currentOffset < node.offset || node.end < this.currentOffset) {
          this.entries.add(node.getText());
        }
      }
      return true;
    };
    return ColorValueCollector2;
  }();
  function getCurrentWord(document, offset) {
    var i = offset - 1;
    var text = document.getText();
    while (i >= 0 && ' 	\n\r":{[()]},*>+'.indexOf(text.charAt(i)) === -1) {
      i--;
    }
    return text.substring(i + 1, offset);
  }
  function isColorString(s) {
    return s.toLowerCase() in colors || /(^#[0-9A-F]{6}$)|(^#[0-9A-F]{3}$)/i.test(s);
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/selectorPrinting.js
  var __extends4 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var localize5 = loadMessageBundle();
  var Element = function() {
    function Element3() {
      this.parent = null;
      this.children = null;
      this.attributes = null;
    }
    Element3.prototype.findAttribute = function(name) {
      if (this.attributes) {
        for (var _i = 0, _a3 = this.attributes; _i < _a3.length; _i++) {
          var attribute = _a3[_i];
          if (attribute.name === name) {
            return attribute.value;
          }
        }
      }
      return null;
    };
    Element3.prototype.addChild = function(child) {
      if (child instanceof Element3) {
        child.parent = this;
      }
      if (!this.children) {
        this.children = [];
      }
      this.children.push(child);
    };
    Element3.prototype.append = function(text) {
      if (this.attributes) {
        var last = this.attributes[this.attributes.length - 1];
        last.value = last.value + text;
      }
    };
    Element3.prototype.prepend = function(text) {
      if (this.attributes) {
        var first = this.attributes[0];
        first.value = text + first.value;
      }
    };
    Element3.prototype.findRoot = function() {
      var curr = this;
      while (curr.parent && !(curr.parent instanceof RootElement)) {
        curr = curr.parent;
      }
      return curr;
    };
    Element3.prototype.removeChild = function(child) {
      if (this.children) {
        var index = this.children.indexOf(child);
        if (index !== -1) {
          this.children.splice(index, 1);
          return true;
        }
      }
      return false;
    };
    Element3.prototype.addAttr = function(name, value) {
      if (!this.attributes) {
        this.attributes = [];
      }
      for (var _i = 0, _a3 = this.attributes; _i < _a3.length; _i++) {
        var attribute = _a3[_i];
        if (attribute.name === name) {
          attribute.value += " " + value;
          return;
        }
      }
      this.attributes.push({ name, value });
    };
    Element3.prototype.clone = function(cloneChildren) {
      if (cloneChildren === void 0) {
        cloneChildren = true;
      }
      var elem = new Element3();
      if (this.attributes) {
        elem.attributes = [];
        for (var _i = 0, _a3 = this.attributes; _i < _a3.length; _i++) {
          var attribute = _a3[_i];
          elem.addAttr(attribute.name, attribute.value);
        }
      }
      if (cloneChildren && this.children) {
        elem.children = [];
        for (var index = 0; index < this.children.length; index++) {
          elem.addChild(this.children[index].clone());
        }
      }
      return elem;
    };
    Element3.prototype.cloneWithParent = function() {
      var clone = this.clone(false);
      if (this.parent && !(this.parent instanceof RootElement)) {
        var parentClone = this.parent.cloneWithParent();
        parentClone.addChild(clone);
      }
      return clone;
    };
    return Element3;
  }();
  var RootElement = function(_super) {
    __extends4(RootElement2, _super);
    function RootElement2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    return RootElement2;
  }(Element);
  var LabelElement = function(_super) {
    __extends4(LabelElement2, _super);
    function LabelElement2(label) {
      var _this = _super.call(this) || this;
      _this.addAttr("name", label);
      return _this;
    }
    return LabelElement2;
  }(Element);
  var MarkedStringPrinter = function() {
    function MarkedStringPrinter2(quote) {
      this.quote = quote;
      this.result = [];
    }
    MarkedStringPrinter2.prototype.print = function(element) {
      this.result = [];
      if (element instanceof RootElement) {
        if (element.children) {
          this.doPrint(element.children, 0);
        }
      } else {
        this.doPrint([element], 0);
      }
      var value = this.result.join("\n");
      return [{ language: "html", value }];
    };
    MarkedStringPrinter2.prototype.doPrint = function(elements, indent) {
      for (var _i = 0, elements_1 = elements; _i < elements_1.length; _i++) {
        var element = elements_1[_i];
        this.doPrintElement(element, indent);
        if (element.children) {
          this.doPrint(element.children, indent + 1);
        }
      }
    };
    MarkedStringPrinter2.prototype.writeLine = function(level, content) {
      var indent = new Array(level + 1).join("  ");
      this.result.push(indent + content);
    };
    MarkedStringPrinter2.prototype.doPrintElement = function(element, indent) {
      var name = element.findAttribute("name");
      if (element instanceof LabelElement || name === "\u2026") {
        this.writeLine(indent, name);
        return;
      }
      var content = ["<"];
      if (name) {
        content.push(name);
      } else {
        content.push("element");
      }
      if (element.attributes) {
        for (var _i = 0, _a3 = element.attributes; _i < _a3.length; _i++) {
          var attr = _a3[_i];
          if (attr.name !== "name") {
            content.push(" ");
            content.push(attr.name);
            var value = attr.value;
            if (value) {
              content.push("=");
              content.push(quotes.ensure(value, this.quote));
            }
          }
        }
      }
      content.push(">");
      this.writeLine(indent, content.join(""));
    };
    return MarkedStringPrinter2;
  }();
  var quotes;
  (function(quotes2) {
    function ensure(value, which) {
      return which + remove(value) + which;
    }
    quotes2.ensure = ensure;
    function remove(value) {
      var match = value.match(/^['"](.*)["']$/);
      if (match) {
        return match[1];
      }
      return value;
    }
    quotes2.remove = remove;
  })(quotes || (quotes = {}));
  var Specificity = function() {
    function Specificity2() {
      this.id = 0;
      this.attr = 0;
      this.tag = 0;
    }
    return Specificity2;
  }();
  function toElement(node, parentElement) {
    var result = new Element();
    for (var _i = 0, _a3 = node.getChildren(); _i < _a3.length; _i++) {
      var child = _a3[_i];
      switch (child.type) {
        case NodeType.SelectorCombinator:
          if (parentElement) {
            var segments = child.getText().split("&");
            if (segments.length === 1) {
              result.addAttr("name", segments[0]);
              break;
            }
            result = parentElement.cloneWithParent();
            if (segments[0]) {
              var root = result.findRoot();
              root.prepend(segments[0]);
            }
            for (var i = 1; i < segments.length; i++) {
              if (i > 1) {
                var clone = parentElement.cloneWithParent();
                result.addChild(clone.findRoot());
                result = clone;
              }
              result.append(segments[i]);
            }
          }
          break;
        case NodeType.SelectorPlaceholder:
          if (child.matches("@at-root")) {
            return result;
          }
        case NodeType.ElementNameSelector:
          var text = child.getText();
          result.addAttr("name", text === "*" ? "element" : unescape(text));
          break;
        case NodeType.ClassSelector:
          result.addAttr("class", unescape(child.getText().substring(1)));
          break;
        case NodeType.IdentifierSelector:
          result.addAttr("id", unescape(child.getText().substring(1)));
          break;
        case NodeType.MixinDeclaration:
          result.addAttr("class", child.getName());
          break;
        case NodeType.PseudoSelector:
          result.addAttr(unescape(child.getText()), "");
          break;
        case NodeType.AttributeSelector:
          var selector = child;
          var identifier = selector.getIdentifier();
          if (identifier) {
            var expression = selector.getValue();
            var operator = selector.getOperator();
            var value = void 0;
            if (expression && operator) {
              switch (unescape(operator.getText())) {
                case "|=":
                  value = quotes.remove(unescape(expression.getText())) + "-\u2026";
                  break;
                case "^=":
                  value = quotes.remove(unescape(expression.getText())) + "\u2026";
                  break;
                case "$=":
                  value = "\u2026" + quotes.remove(unescape(expression.getText()));
                  break;
                case "~=":
                  value = " \u2026 " + quotes.remove(unescape(expression.getText())) + " \u2026 ";
                  break;
                case "*=":
                  value = "\u2026" + quotes.remove(unescape(expression.getText())) + "\u2026";
                  break;
                default:
                  value = quotes.remove(unescape(expression.getText()));
                  break;
              }
            }
            result.addAttr(unescape(identifier.getText()), value);
          }
          break;
      }
    }
    return result;
  }
  function unescape(content) {
    var scanner = new Scanner();
    scanner.setSource(content);
    var token = scanner.scanUnquotedString();
    if (token) {
      return token.text;
    }
    return content;
  }
  var SelectorPrinting = function() {
    function SelectorPrinting2(cssDataManager) {
      this.cssDataManager = cssDataManager;
    }
    SelectorPrinting2.prototype.selectorToMarkedString = function(node) {
      var root = selectorToElement(node);
      if (root) {
        var markedStrings = new MarkedStringPrinter('"').print(root);
        markedStrings.push(this.selectorToSpecificityMarkedString(node));
        return markedStrings;
      } else {
        return [];
      }
    };
    SelectorPrinting2.prototype.simpleSelectorToMarkedString = function(node) {
      var element = toElement(node);
      var markedStrings = new MarkedStringPrinter('"').print(element);
      markedStrings.push(this.selectorToSpecificityMarkedString(node));
      return markedStrings;
    };
    SelectorPrinting2.prototype.isPseudoElementIdentifier = function(text) {
      var match = text.match(/^::?([\w-]+)/);
      if (!match) {
        return false;
      }
      return !!this.cssDataManager.getPseudoElement("::" + match[1]);
    };
    SelectorPrinting2.prototype.selectorToSpecificityMarkedString = function(node) {
      var _this = this;
      var calculateScore = function(node2) {
        for (var _i = 0, _a3 = node2.getChildren(); _i < _a3.length; _i++) {
          var element = _a3[_i];
          switch (element.type) {
            case NodeType.IdentifierSelector:
              specificity.id++;
              break;
            case NodeType.ClassSelector:
            case NodeType.AttributeSelector:
              specificity.attr++;
              break;
            case NodeType.ElementNameSelector:
              if (element.matches("*")) {
                break;
              }
              specificity.tag++;
              break;
            case NodeType.PseudoSelector:
              var text = element.getText();
              if (_this.isPseudoElementIdentifier(text)) {
                specificity.tag++;
              } else {
                if (text.match(/^:not/i)) {
                  break;
                }
                specificity.attr++;
              }
              break;
          }
          if (element.getChildren().length > 0) {
            calculateScore(element);
          }
        }
      };
      var specificity = new Specificity();
      calculateScore(node);
      return localize5("specificity", "[Selector Specificity](https://developer.mozilla.org/en-US/docs/Web/CSS/Specificity): ({0}, {1}, {2})", specificity.id, specificity.attr, specificity.tag);
    };
    return SelectorPrinting2;
  }();
  var SelectorElementBuilder = function() {
    function SelectorElementBuilder2(element) {
      this.prev = null;
      this.element = element;
    }
    SelectorElementBuilder2.prototype.processSelector = function(selector) {
      var parentElement = null;
      if (!(this.element instanceof RootElement)) {
        if (selector.getChildren().some(function(c) {
          return c.hasChildren() && c.getChild(0).type === NodeType.SelectorCombinator;
        })) {
          var curr = this.element.findRoot();
          if (curr.parent instanceof RootElement) {
            parentElement = this.element;
            this.element = curr.parent;
            this.element.removeChild(curr);
            this.prev = null;
          }
        }
      }
      for (var _i = 0, _a3 = selector.getChildren(); _i < _a3.length; _i++) {
        var selectorChild = _a3[_i];
        if (selectorChild instanceof SimpleSelector) {
          if (this.prev instanceof SimpleSelector) {
            var labelElement = new LabelElement("\u2026");
            this.element.addChild(labelElement);
            this.element = labelElement;
          } else if (this.prev && (this.prev.matches("+") || this.prev.matches("~")) && this.element.parent) {
            this.element = this.element.parent;
          }
          if (this.prev && this.prev.matches("~")) {
            this.element.addChild(toElement(selectorChild));
            this.element.addChild(new LabelElement("\u22EE"));
          }
          var thisElement = toElement(selectorChild, parentElement);
          var root = thisElement.findRoot();
          this.element.addChild(root);
          this.element = thisElement;
        }
        if (selectorChild instanceof SimpleSelector || selectorChild.type === NodeType.SelectorCombinatorParent || selectorChild.type === NodeType.SelectorCombinatorShadowPiercingDescendant || selectorChild.type === NodeType.SelectorCombinatorSibling || selectorChild.type === NodeType.SelectorCombinatorAllSiblings) {
          this.prev = selectorChild;
        }
      }
    };
    return SelectorElementBuilder2;
  }();
  function isNewSelectorContext(node) {
    switch (node.type) {
      case NodeType.MixinDeclaration:
      case NodeType.Stylesheet:
        return true;
    }
    return false;
  }
  function selectorToElement(node) {
    if (node.matches("@at-root")) {
      return null;
    }
    var root = new RootElement();
    var parentRuleSets = [];
    var ruleSet = node.getParent();
    if (ruleSet instanceof RuleSet) {
      var parent = ruleSet.getParent();
      while (parent && !isNewSelectorContext(parent)) {
        if (parent instanceof RuleSet) {
          if (parent.getSelectors().matches("@at-root")) {
            break;
          }
          parentRuleSets.push(parent);
        }
        parent = parent.getParent();
      }
    }
    var builder = new SelectorElementBuilder(root);
    for (var i = parentRuleSets.length - 1; i >= 0; i--) {
      var selector = parentRuleSets[i].getSelectors().getChild(0);
      if (selector) {
        builder.processSelector(selector);
      }
    }
    builder.processSelector(node);
    return root;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssHover.js
  var CSSHover = function() {
    function CSSHover2(clientCapabilities, cssDataManager) {
      this.clientCapabilities = clientCapabilities;
      this.cssDataManager = cssDataManager;
      this.selectorPrinting = new SelectorPrinting(cssDataManager);
    }
    CSSHover2.prototype.doHover = function(document, position, stylesheet) {
      function getRange2(node2) {
        return Range2.create(document.positionAt(node2.offset), document.positionAt(node2.end));
      }
      var offset = document.offsetAt(position);
      var nodepath = getNodePath(stylesheet, offset);
      var hover = null;
      for (var i = 0; i < nodepath.length; i++) {
        var node = nodepath[i];
        if (node instanceof Selector) {
          hover = {
            contents: this.selectorPrinting.selectorToMarkedString(node),
            range: getRange2(node)
          };
          break;
        }
        if (node instanceof SimpleSelector) {
          if (!startsWith(node.getText(), "@")) {
            hover = {
              contents: this.selectorPrinting.simpleSelectorToMarkedString(node),
              range: getRange2(node)
            };
          }
          break;
        }
        if (node instanceof Declaration) {
          var propertyName = node.getFullPropertyName();
          var entry = this.cssDataManager.getProperty(propertyName);
          if (entry) {
            var contents = getEntryDescription(entry, this.doesSupportMarkdown());
            if (contents) {
              hover = {
                contents,
                range: getRange2(node)
              };
            } else {
              hover = null;
            }
          }
          continue;
        }
        if (node instanceof UnknownAtRule) {
          var atRuleName = node.getText();
          var entry = this.cssDataManager.getAtDirective(atRuleName);
          if (entry) {
            var contents = getEntryDescription(entry, this.doesSupportMarkdown());
            if (contents) {
              hover = {
                contents,
                range: getRange2(node)
              };
            } else {
              hover = null;
            }
          }
          continue;
        }
        if (node instanceof Node2 && node.type === NodeType.PseudoSelector) {
          var selectorName = node.getText();
          var entry = selectorName.slice(0, 2) === "::" ? this.cssDataManager.getPseudoElement(selectorName) : this.cssDataManager.getPseudoClass(selectorName);
          if (entry) {
            var contents = getEntryDescription(entry, this.doesSupportMarkdown());
            if (contents) {
              hover = {
                contents,
                range: getRange2(node)
              };
            } else {
              hover = null;
            }
          }
          continue;
        }
      }
      if (hover) {
        hover.contents = this.convertContents(hover.contents);
      }
      return hover;
    };
    CSSHover2.prototype.convertContents = function(contents) {
      if (!this.doesSupportMarkdown()) {
        if (typeof contents === "string") {
          return contents;
        } else if ("kind" in contents) {
          return {
            kind: "plaintext",
            value: contents.value
          };
        } else if (Array.isArray(contents)) {
          return contents.map(function(c) {
            return typeof c === "string" ? c : c.value;
          });
        } else {
          return contents.value;
        }
      }
      return contents;
    };
    CSSHover2.prototype.doesSupportMarkdown = function() {
      if (!isDefined(this.supportsMarkdown)) {
        if (!isDefined(this.clientCapabilities)) {
          this.supportsMarkdown = true;
          return this.supportsMarkdown;
        }
        var hover = this.clientCapabilities.textDocument && this.clientCapabilities.textDocument.hover;
        this.supportsMarkdown = hover && hover.contentFormat && Array.isArray(hover.contentFormat) && hover.contentFormat.indexOf(MarkupKind.Markdown) !== -1;
      }
      return this.supportsMarkdown;
    };
    return CSSHover2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssNavigation.js
  var __awaiter4 = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var __generator3 = function(thisArg, body) {
    var _ = { label: 0, sent: function() {
      if (t[0] & 1)
        throw t[1];
      return t[1];
    }, trys: [], ops: [] }, f2, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() {
      return this;
    }), g;
    function verb(n) {
      return function(v) {
        return step([n, v]);
      };
    }
    function step(op) {
      if (f2)
        throw new TypeError("Generator is already executing.");
      while (_)
        try {
          if (f2 = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
            return t;
          if (y = 0, t)
            op = [op[0] & 2, t.value];
          switch (op[0]) {
            case 0:
            case 1:
              t = op;
              break;
            case 4:
              _.label++;
              return { value: op[1], done: false };
            case 5:
              _.label++;
              y = op[1];
              op = [0];
              continue;
            case 7:
              op = _.ops.pop();
              _.trys.pop();
              continue;
            default:
              if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                _ = 0;
                continue;
              }
              if (op[0] === 3 && (!t || op[1] > t[0] && op[1] < t[3])) {
                _.label = op[1];
                break;
              }
              if (op[0] === 6 && _.label < t[1]) {
                _.label = t[1];
                t = op;
                break;
              }
              if (t && _.label < t[2]) {
                _.label = t[2];
                _.ops.push(op);
                break;
              }
              if (t[2])
                _.ops.pop();
              _.trys.pop();
              continue;
          }
          op = body.call(thisArg, _);
        } catch (e) {
          op = [6, e];
          y = 0;
        } finally {
          f2 = t = 0;
        }
      if (op[0] & 5)
        throw op[1];
      return { value: op[0] ? op[1] : void 0, done: true };
    }
  };
  var localize6 = loadMessageBundle();
  var CSSNavigation = function() {
    function CSSNavigation2(fileSystemProvider) {
      this.fileSystemProvider = fileSystemProvider;
    }
    CSSNavigation2.prototype.findDefinition = function(document, position, stylesheet) {
      var symbols = new Symbols(stylesheet);
      var offset = document.offsetAt(position);
      var node = getNodeAtOffset(stylesheet, offset);
      if (!node) {
        return null;
      }
      var symbol = symbols.findSymbolFromNode(node);
      if (!symbol) {
        return null;
      }
      return {
        uri: document.uri,
        range: getRange(symbol.node, document)
      };
    };
    CSSNavigation2.prototype.findReferences = function(document, position, stylesheet) {
      var highlights = this.findDocumentHighlights(document, position, stylesheet);
      return highlights.map(function(h) {
        return {
          uri: document.uri,
          range: h.range
        };
      });
    };
    CSSNavigation2.prototype.findDocumentHighlights = function(document, position, stylesheet) {
      var result = [];
      var offset = document.offsetAt(position);
      var node = getNodeAtOffset(stylesheet, offset);
      if (!node || node.type === NodeType.Stylesheet || node.type === NodeType.Declarations) {
        return result;
      }
      if (node.type === NodeType.Identifier && node.parent && node.parent.type === NodeType.ClassSelector) {
        node = node.parent;
      }
      var symbols = new Symbols(stylesheet);
      var symbol = symbols.findSymbolFromNode(node);
      var name = node.getText();
      stylesheet.accept(function(candidate) {
        if (symbol) {
          if (symbols.matchesSymbol(candidate, symbol)) {
            result.push({
              kind: getHighlightKind(candidate),
              range: getRange(candidate, document)
            });
            return false;
          }
        } else if (node && node.type === candidate.type && candidate.matches(name)) {
          result.push({
            kind: getHighlightKind(candidate),
            range: getRange(candidate, document)
          });
        }
        return true;
      });
      return result;
    };
    CSSNavigation2.prototype.isRawStringDocumentLinkNode = function(node) {
      return node.type === NodeType.Import;
    };
    CSSNavigation2.prototype.findDocumentLinks = function(document, stylesheet, documentContext) {
      var links = this.findUnresolvedLinks(document, stylesheet);
      for (var i = 0; i < links.length; i++) {
        var target = links[i].target;
        if (target && !/^\w+:\/\//g.test(target)) {
          var resolved = documentContext.resolveReference(target, document.uri);
          if (resolved) {
            links[i].target = resolved;
          }
        }
      }
      return links;
    };
    CSSNavigation2.prototype.findDocumentLinks2 = function(document, stylesheet, documentContext) {
      return __awaiter4(this, void 0, void 0, function() {
        var links, resolvedLinks, _i, links_1, link, target, resolvedTarget;
        return __generator3(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              links = this.findUnresolvedLinks(document, stylesheet);
              resolvedLinks = [];
              _i = 0, links_1 = links;
              _a3.label = 1;
            case 1:
              if (!(_i < links_1.length))
                return [3, 5];
              link = links_1[_i];
              target = link.target;
              if (!(target && !/^\w+:\/\//g.test(target)))
                return [3, 3];
              return [4, this.resolveRelativeReference(target, document.uri, documentContext)];
            case 2:
              resolvedTarget = _a3.sent();
              if (resolvedTarget !== void 0) {
                link.target = resolvedTarget;
                resolvedLinks.push(link);
              }
              return [3, 4];
            case 3:
              resolvedLinks.push(link);
              _a3.label = 4;
            case 4:
              _i++;
              return [3, 1];
            case 5:
              return [2, resolvedLinks];
          }
        });
      });
    };
    CSSNavigation2.prototype.findUnresolvedLinks = function(document, stylesheet) {
      var _this = this;
      var result = [];
      var collect = function(uriStringNode) {
        var rawUri = uriStringNode.getText();
        var range = getRange(uriStringNode, document);
        if (range.start.line === range.end.line && range.start.character === range.end.character) {
          return;
        }
        if (startsWith(rawUri, "'") || startsWith(rawUri, '"')) {
          rawUri = rawUri.slice(1, -1);
        }
        result.push({ target: rawUri, range });
      };
      stylesheet.accept(function(candidate) {
        if (candidate.type === NodeType.URILiteral) {
          var first = candidate.getChild(0);
          if (first) {
            collect(first);
          }
          return false;
        }
        if (candidate.parent && _this.isRawStringDocumentLinkNode(candidate.parent)) {
          var rawText = candidate.getText();
          if (startsWith(rawText, "'") || startsWith(rawText, '"')) {
            collect(candidate);
          }
          return false;
        }
        return true;
      });
      return result;
    };
    CSSNavigation2.prototype.findDocumentSymbols = function(document, stylesheet) {
      var result = [];
      stylesheet.accept(function(node) {
        var entry = {
          name: null,
          kind: SymbolKind2.Class,
          location: null
        };
        var locationNode = node;
        if (node instanceof Selector) {
          entry.name = node.getText();
          locationNode = node.findAParent(NodeType.Ruleset, NodeType.ExtendsReference);
          if (locationNode) {
            entry.location = Location.create(document.uri, getRange(locationNode, document));
            result.push(entry);
          }
          return false;
        } else if (node instanceof VariableDeclaration) {
          entry.name = node.getName();
          entry.kind = SymbolKind2.Variable;
        } else if (node instanceof MixinDeclaration) {
          entry.name = node.getName();
          entry.kind = SymbolKind2.Method;
        } else if (node instanceof FunctionDeclaration) {
          entry.name = node.getName();
          entry.kind = SymbolKind2.Function;
        } else if (node instanceof Keyframe) {
          entry.name = localize6("literal.keyframes", "@keyframes {0}", node.getName());
        } else if (node instanceof FontFace) {
          entry.name = localize6("literal.fontface", "@font-face");
        } else if (node instanceof Media) {
          var mediaList = node.getChild(0);
          if (mediaList instanceof Medialist) {
            entry.name = "@media " + mediaList.getText();
            entry.kind = SymbolKind2.Module;
          }
        }
        if (entry.name) {
          entry.location = Location.create(document.uri, getRange(locationNode, document));
          result.push(entry);
        }
        return true;
      });
      return result;
    };
    CSSNavigation2.prototype.findDocumentColors = function(document, stylesheet) {
      var result = [];
      stylesheet.accept(function(node) {
        var colorInfo = getColorInformation(node, document);
        if (colorInfo) {
          result.push(colorInfo);
        }
        return true;
      });
      return result;
    };
    CSSNavigation2.prototype.getColorPresentations = function(document, stylesheet, color, range) {
      var result = [];
      var red256 = Math.round(color.red * 255), green256 = Math.round(color.green * 255), blue256 = Math.round(color.blue * 255);
      var label;
      if (color.alpha === 1) {
        label = "rgb(" + red256 + ", " + green256 + ", " + blue256 + ")";
      } else {
        label = "rgba(" + red256 + ", " + green256 + ", " + blue256 + ", " + color.alpha + ")";
      }
      result.push({ label, textEdit: TextEdit.replace(range, label) });
      if (color.alpha === 1) {
        label = "#" + toTwoDigitHex(red256) + toTwoDigitHex(green256) + toTwoDigitHex(blue256);
      } else {
        label = "#" + toTwoDigitHex(red256) + toTwoDigitHex(green256) + toTwoDigitHex(blue256) + toTwoDigitHex(Math.round(color.alpha * 255));
      }
      result.push({ label, textEdit: TextEdit.replace(range, label) });
      var hsl = hslFromColor(color);
      if (hsl.a === 1) {
        label = "hsl(" + hsl.h + ", " + Math.round(hsl.s * 100) + "%, " + Math.round(hsl.l * 100) + "%)";
      } else {
        label = "hsla(" + hsl.h + ", " + Math.round(hsl.s * 100) + "%, " + Math.round(hsl.l * 100) + "%, " + hsl.a + ")";
      }
      result.push({ label, textEdit: TextEdit.replace(range, label) });
      return result;
    };
    CSSNavigation2.prototype.doRename = function(document, position, newName, stylesheet) {
      var _a3;
      var highlights = this.findDocumentHighlights(document, position, stylesheet);
      var edits = highlights.map(function(h) {
        return TextEdit.replace(h.range, newName);
      });
      return {
        changes: (_a3 = {}, _a3[document.uri] = edits, _a3)
      };
    };
    CSSNavigation2.prototype.resolveRelativeReference = function(ref, documentUri, documentContext) {
      return __awaiter4(this, void 0, void 0, function() {
        var moduleName, rootFolderUri, documentFolderUri, modulePath, pathWithinModule;
        return __generator3(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              if (!(ref[0] === "~" && ref[1] !== "/" && this.fileSystemProvider))
                return [3, 3];
              ref = ref.substring(1);
              if (!startsWith(documentUri, "file://"))
                return [3, 2];
              moduleName = getModuleNameFromPath(ref);
              rootFolderUri = documentContext.resolveReference("/", documentUri);
              documentFolderUri = dirname2(documentUri);
              return [4, this.resolvePathToModule(moduleName, documentFolderUri, rootFolderUri)];
            case 1:
              modulePath = _a3.sent();
              if (modulePath) {
                pathWithinModule = ref.substring(moduleName.length + 1);
                return [2, joinPath(modulePath, pathWithinModule)];
              }
              _a3.label = 2;
            case 2:
              return [2, documentContext.resolveReference(ref, documentUri)];
            case 3:
              return [2, documentContext.resolveReference(ref, documentUri)];
          }
        });
      });
    };
    CSSNavigation2.prototype.resolvePathToModule = function(_moduleName, documentFolderUri, rootFolderUri) {
      return __awaiter4(this, void 0, void 0, function() {
        var packPath;
        return __generator3(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              packPath = joinPath(documentFolderUri, "node_modules", _moduleName, "package.json");
              return [4, this.fileExists(packPath)];
            case 1:
              if (_a3.sent()) {
                return [2, dirname2(packPath)];
              } else if (rootFolderUri && documentFolderUri.startsWith(rootFolderUri) && documentFolderUri.length !== rootFolderUri.length) {
                return [2, this.resolvePathToModule(_moduleName, dirname2(documentFolderUri), rootFolderUri)];
              }
              return [2, void 0];
          }
        });
      });
    };
    CSSNavigation2.prototype.fileExists = function(uri) {
      return __awaiter4(this, void 0, void 0, function() {
        var stat, err_1;
        return __generator3(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              if (!this.fileSystemProvider) {
                return [2, false];
              }
              _a3.label = 1;
            case 1:
              _a3.trys.push([1, 3, , 4]);
              return [4, this.fileSystemProvider.stat(uri)];
            case 2:
              stat = _a3.sent();
              if (stat.type === FileType.Unknown && stat.size === -1) {
                return [2, false];
              }
              return [2, true];
            case 3:
              err_1 = _a3.sent();
              return [2, false];
            case 4:
              return [2];
          }
        });
      });
    };
    return CSSNavigation2;
  }();
  function getColorInformation(node, document) {
    var color = getColorValue(node);
    if (color) {
      var range = getRange(node, document);
      return { color, range };
    }
    return null;
  }
  function getRange(node, document) {
    return Range2.create(document.positionAt(node.offset), document.positionAt(node.end));
  }
  function getHighlightKind(node) {
    if (node.type === NodeType.Selector) {
      return DocumentHighlightKind2.Write;
    }
    if (node instanceof Identifier) {
      if (node.parent && node.parent instanceof Property) {
        if (node.isCustomProperty) {
          return DocumentHighlightKind2.Write;
        }
      }
    }
    if (node.parent) {
      switch (node.parent.type) {
        case NodeType.FunctionDeclaration:
        case NodeType.MixinDeclaration:
        case NodeType.Keyframe:
        case NodeType.VariableDeclaration:
        case NodeType.FunctionParameter:
          return DocumentHighlightKind2.Write;
      }
    }
    return DocumentHighlightKind2.Read;
  }
  function toTwoDigitHex(n) {
    var r = n.toString(16);
    return r.length !== 2 ? "0" + r : r;
  }
  function getModuleNameFromPath(path) {
    if (path[0] === "@") {
      return path.substring(0, path.indexOf("/", path.indexOf("/") + 1));
    }
    return path.substring(0, path.indexOf("/"));
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/lintRules.js
  var localize7 = loadMessageBundle();
  var Warning = Level.Warning;
  var Error2 = Level.Error;
  var Ignore = Level.Ignore;
  var Rule = function() {
    function Rule2(id, message, defaultValue) {
      this.id = id;
      this.message = message;
      this.defaultValue = defaultValue;
    }
    return Rule2;
  }();
  var Setting = function() {
    function Setting2(id, message, defaultValue) {
      this.id = id;
      this.message = message;
      this.defaultValue = defaultValue;
    }
    return Setting2;
  }();
  var Rules = {
    AllVendorPrefixes: new Rule("compatibleVendorPrefixes", localize7("rule.vendorprefixes.all", "When using a vendor-specific prefix make sure to also include all other vendor-specific properties"), Ignore),
    IncludeStandardPropertyWhenUsingVendorPrefix: new Rule("vendorPrefix", localize7("rule.standardvendorprefix.all", "When using a vendor-specific prefix also include the standard property"), Warning),
    DuplicateDeclarations: new Rule("duplicateProperties", localize7("rule.duplicateDeclarations", "Do not use duplicate style definitions"), Ignore),
    EmptyRuleSet: new Rule("emptyRules", localize7("rule.emptyRuleSets", "Do not use empty rulesets"), Warning),
    ImportStatemement: new Rule("importStatement", localize7("rule.importDirective", "Import statements do not load in parallel"), Ignore),
    BewareOfBoxModelSize: new Rule("boxModel", localize7("rule.bewareOfBoxModelSize", "Do not use width or height when using padding or border"), Ignore),
    UniversalSelector: new Rule("universalSelector", localize7("rule.universalSelector", "The universal selector (*) is known to be slow"), Ignore),
    ZeroWithUnit: new Rule("zeroUnits", localize7("rule.zeroWidthUnit", "No unit for zero needed"), Ignore),
    RequiredPropertiesForFontFace: new Rule("fontFaceProperties", localize7("rule.fontFaceProperties", "@font-face rule must define 'src' and 'font-family' properties"), Warning),
    HexColorLength: new Rule("hexColorLength", localize7("rule.hexColor", "Hex colors must consist of three, four, six or eight hex numbers"), Error2),
    ArgsInColorFunction: new Rule("argumentsInColorFunction", localize7("rule.colorFunction", "Invalid number of parameters"), Error2),
    UnknownProperty: new Rule("unknownProperties", localize7("rule.unknownProperty", "Unknown property."), Warning),
    UnknownAtRules: new Rule("unknownAtRules", localize7("rule.unknownAtRules", "Unknown at-rule."), Warning),
    IEStarHack: new Rule("ieHack", localize7("rule.ieHack", "IE hacks are only necessary when supporting IE7 and older"), Ignore),
    UnknownVendorSpecificProperty: new Rule("unknownVendorSpecificProperties", localize7("rule.unknownVendorSpecificProperty", "Unknown vendor specific property."), Ignore),
    PropertyIgnoredDueToDisplay: new Rule("propertyIgnoredDueToDisplay", localize7("rule.propertyIgnoredDueToDisplay", "Property is ignored due to the display."), Warning),
    AvoidImportant: new Rule("important", localize7("rule.avoidImportant", "Avoid using !important. It is an indication that the specificity of the entire CSS has gotten out of control and needs to be refactored."), Ignore),
    AvoidFloat: new Rule("float", localize7("rule.avoidFloat", "Avoid using 'float'. Floats lead to fragile CSS that is easy to break if one aspect of the layout changes."), Ignore),
    AvoidIdSelector: new Rule("idSelector", localize7("rule.avoidIdSelector", "Selectors should not contain IDs because these rules are too tightly coupled with the HTML."), Ignore)
  };
  var Settings = {
    ValidProperties: new Setting("validProperties", localize7("rule.validProperties", "A list of properties that are not validated against the `unknownProperties` rule."), [])
  };
  var LintConfigurationSettings = function() {
    function LintConfigurationSettings2(conf) {
      if (conf === void 0) {
        conf = {};
      }
      this.conf = conf;
    }
    LintConfigurationSettings2.prototype.getRule = function(rule) {
      if (this.conf.hasOwnProperty(rule.id)) {
        var level = toLevel(this.conf[rule.id]);
        if (level) {
          return level;
        }
      }
      return rule.defaultValue;
    };
    LintConfigurationSettings2.prototype.getSetting = function(setting) {
      return this.conf[setting.id];
    };
    return LintConfigurationSettings2;
  }();
  function toLevel(level) {
    switch (level) {
      case "ignore":
        return Level.Ignore;
      case "warning":
        return Level.Warning;
      case "error":
        return Level.Error;
    }
    return null;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssCodeActions.js
  var localize8 = loadMessageBundle();
  var CSSCodeActions = function() {
    function CSSCodeActions2(cssDataManager) {
      this.cssDataManager = cssDataManager;
    }
    CSSCodeActions2.prototype.doCodeActions = function(document, range, context, stylesheet) {
      return this.doCodeActions2(document, range, context, stylesheet).map(function(ca) {
        var textDocumentEdit = ca.edit && ca.edit.documentChanges && ca.edit.documentChanges[0];
        return Command.create(ca.title, "_css.applyCodeAction", document.uri, document.version, textDocumentEdit && textDocumentEdit.edits);
      });
    };
    CSSCodeActions2.prototype.doCodeActions2 = function(document, range, context, stylesheet) {
      var result = [];
      if (context.diagnostics) {
        for (var _i = 0, _a3 = context.diagnostics; _i < _a3.length; _i++) {
          var diagnostic = _a3[_i];
          this.appendFixesForMarker(document, stylesheet, diagnostic, result);
        }
      }
      return result;
    };
    CSSCodeActions2.prototype.getFixesForUnknownProperty = function(document, property, marker, result) {
      var propertyName = property.getName();
      var candidates = [];
      this.cssDataManager.getProperties().forEach(function(p) {
        var score = difference(propertyName, p.name);
        if (score >= propertyName.length / 2) {
          candidates.push({ property: p.name, score });
        }
      });
      candidates.sort(function(a2, b) {
        return b.score - a2.score || a2.property.localeCompare(b.property);
      });
      var maxActions = 3;
      for (var _i = 0, candidates_1 = candidates; _i < candidates_1.length; _i++) {
        var candidate = candidates_1[_i];
        var propertyName_1 = candidate.property;
        var title = localize8("css.codeaction.rename", "Rename to '{0}'", propertyName_1);
        var edit = TextEdit.replace(marker.range, propertyName_1);
        var documentIdentifier = VersionedTextDocumentIdentifier.create(document.uri, document.version);
        var workspaceEdit = { documentChanges: [TextDocumentEdit.create(documentIdentifier, [edit])] };
        var codeAction = CodeAction.create(title, workspaceEdit, CodeActionKind.QuickFix);
        codeAction.diagnostics = [marker];
        result.push(codeAction);
        if (--maxActions <= 0) {
          return;
        }
      }
    };
    CSSCodeActions2.prototype.appendFixesForMarker = function(document, stylesheet, marker, result) {
      if (marker.code !== Rules.UnknownProperty.id) {
        return;
      }
      var offset = document.offsetAt(marker.range.start);
      var end = document.offsetAt(marker.range.end);
      var nodepath = getNodePath(stylesheet, offset);
      for (var i = nodepath.length - 1; i >= 0; i--) {
        var node = nodepath[i];
        if (node instanceof Declaration) {
          var property = node.getProperty();
          if (property && property.offset === offset && property.end === end) {
            this.getFixesForUnknownProperty(document, property, marker, result);
            return;
          }
        }
      }
    };
    return CSSCodeActions2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/lintUtil.js
  var Element2 = function() {
    function Element3(decl) {
      this.fullPropertyName = decl.getFullPropertyName().toLowerCase();
      this.node = decl;
    }
    return Element3;
  }();
  function setSide(model, side, value, property) {
    var state = model[side];
    state.value = value;
    if (value) {
      if (!includes(state.properties, property)) {
        state.properties.push(property);
      }
    }
  }
  function setAllSides(model, value, property) {
    setSide(model, "top", value, property);
    setSide(model, "right", value, property);
    setSide(model, "bottom", value, property);
    setSide(model, "left", value, property);
  }
  function updateModelWithValue(model, side, value, property) {
    if (side === "top" || side === "right" || side === "bottom" || side === "left") {
      setSide(model, side, value, property);
    } else {
      setAllSides(model, value, property);
    }
  }
  function updateModelWithList(model, values2, property) {
    switch (values2.length) {
      case 1:
        updateModelWithValue(model, void 0, values2[0], property);
        break;
      case 2:
        updateModelWithValue(model, "top", values2[0], property);
        updateModelWithValue(model, "bottom", values2[0], property);
        updateModelWithValue(model, "right", values2[1], property);
        updateModelWithValue(model, "left", values2[1], property);
        break;
      case 3:
        updateModelWithValue(model, "top", values2[0], property);
        updateModelWithValue(model, "right", values2[1], property);
        updateModelWithValue(model, "left", values2[1], property);
        updateModelWithValue(model, "bottom", values2[2], property);
        break;
      case 4:
        updateModelWithValue(model, "top", values2[0], property);
        updateModelWithValue(model, "right", values2[1], property);
        updateModelWithValue(model, "bottom", values2[2], property);
        updateModelWithValue(model, "left", values2[3], property);
        break;
    }
  }
  function matches(value, candidates) {
    for (var _i = 0, candidates_1 = candidates; _i < candidates_1.length; _i++) {
      var candidate = candidates_1[_i];
      if (value.matches(candidate)) {
        return true;
      }
    }
    return false;
  }
  function checkLineWidth(value, allowsKeywords) {
    if (allowsKeywords === void 0) {
      allowsKeywords = true;
    }
    if (allowsKeywords && matches(value, ["initial", "unset"])) {
      return false;
    }
    return parseFloat(value.getText()) !== 0;
  }
  function checkLineWidthList(nodes, allowsKeywords) {
    if (allowsKeywords === void 0) {
      allowsKeywords = true;
    }
    return nodes.map(function(node) {
      return checkLineWidth(node, allowsKeywords);
    });
  }
  function checkLineStyle(valueNode, allowsKeywords) {
    if (allowsKeywords === void 0) {
      allowsKeywords = true;
    }
    if (matches(valueNode, ["none", "hidden"])) {
      return false;
    }
    if (allowsKeywords && matches(valueNode, ["initial", "unset"])) {
      return false;
    }
    return true;
  }
  function checkLineStyleList(nodes, allowsKeywords) {
    if (allowsKeywords === void 0) {
      allowsKeywords = true;
    }
    return nodes.map(function(node) {
      return checkLineStyle(node, allowsKeywords);
    });
  }
  function checkBorderShorthand(node) {
    var children = node.getChildren();
    if (children.length === 1) {
      var value = children[0];
      return checkLineWidth(value) && checkLineStyle(value);
    }
    for (var _i = 0, children_1 = children; _i < children_1.length; _i++) {
      var child = children_1[_i];
      var value = child;
      if (!checkLineWidth(value, false) || !checkLineStyle(value, false)) {
        return false;
      }
    }
    return true;
  }
  function calculateBoxModel(propertyTable) {
    var model = {
      top: { value: false, properties: [] },
      right: { value: false, properties: [] },
      bottom: { value: false, properties: [] },
      left: { value: false, properties: [] }
    };
    for (var _i = 0, propertyTable_1 = propertyTable; _i < propertyTable_1.length; _i++) {
      var property = propertyTable_1[_i];
      var value = property.node.value;
      if (typeof value === "undefined") {
        continue;
      }
      switch (property.fullPropertyName) {
        case "box-sizing":
          return {
            top: { value: false, properties: [] },
            right: { value: false, properties: [] },
            bottom: { value: false, properties: [] },
            left: { value: false, properties: [] }
          };
        case "width":
          model.width = property;
          break;
        case "height":
          model.height = property;
          break;
        default:
          var segments = property.fullPropertyName.split("-");
          switch (segments[0]) {
            case "border":
              switch (segments[1]) {
                case void 0:
                case "top":
                case "right":
                case "bottom":
                case "left":
                  switch (segments[2]) {
                    case void 0:
                      updateModelWithValue(model, segments[1], checkBorderShorthand(value), property);
                      break;
                    case "width":
                      updateModelWithValue(model, segments[1], checkLineWidth(value, false), property);
                      break;
                    case "style":
                      updateModelWithValue(model, segments[1], checkLineStyle(value, true), property);
                      break;
                  }
                  break;
                case "width":
                  updateModelWithList(model, checkLineWidthList(value.getChildren(), false), property);
                  break;
                case "style":
                  updateModelWithList(model, checkLineStyleList(value.getChildren(), true), property);
                  break;
              }
              break;
            case "padding":
              if (segments.length === 1) {
                updateModelWithList(model, checkLineWidthList(value.getChildren(), true), property);
              } else {
                updateModelWithValue(model, segments[1], checkLineWidth(value, true), property);
              }
              break;
          }
          break;
      }
    }
    return model;
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/lint.js
  var localize9 = loadMessageBundle();
  var NodesByRootMap = function() {
    function NodesByRootMap2() {
      this.data = {};
    }
    NodesByRootMap2.prototype.add = function(root, name, node) {
      var entry = this.data[root];
      if (!entry) {
        entry = { nodes: [], names: [] };
        this.data[root] = entry;
      }
      entry.names.push(name);
      if (node) {
        entry.nodes.push(node);
      }
    };
    return NodesByRootMap2;
  }();
  var LintVisitor = function() {
    function LintVisitor2(document, settings, cssDataManager) {
      var _this = this;
      this.cssDataManager = cssDataManager;
      this.warnings = [];
      this.settings = settings;
      this.documentText = document.getText();
      this.keyframes = new NodesByRootMap();
      this.validProperties = {};
      var properties = settings.getSetting(Settings.ValidProperties);
      if (Array.isArray(properties)) {
        properties.forEach(function(p) {
          if (typeof p === "string") {
            var name = p.trim().toLowerCase();
            if (name.length) {
              _this.validProperties[name] = true;
            }
          }
        });
      }
    }
    LintVisitor2.entries = function(node, document, settings, cssDataManager, entryFilter) {
      var visitor = new LintVisitor2(document, settings, cssDataManager);
      node.acceptVisitor(visitor);
      visitor.completeValidations();
      return visitor.getEntries(entryFilter);
    };
    LintVisitor2.prototype.isValidPropertyDeclaration = function(element) {
      var propertyName = element.fullPropertyName;
      return this.validProperties[propertyName];
    };
    LintVisitor2.prototype.fetch = function(input, s) {
      var elements = [];
      for (var _i = 0, input_1 = input; _i < input_1.length; _i++) {
        var curr = input_1[_i];
        if (curr.fullPropertyName === s) {
          elements.push(curr);
        }
      }
      return elements;
    };
    LintVisitor2.prototype.fetchWithValue = function(input, s, v) {
      var elements = [];
      for (var _i = 0, input_2 = input; _i < input_2.length; _i++) {
        var inputElement = input_2[_i];
        if (inputElement.fullPropertyName === s) {
          var expression = inputElement.node.getValue();
          if (expression && this.findValueInExpression(expression, v)) {
            elements.push(inputElement);
          }
        }
      }
      return elements;
    };
    LintVisitor2.prototype.findValueInExpression = function(expression, v) {
      var found = false;
      expression.accept(function(node) {
        if (node.type === NodeType.Identifier && node.matches(v)) {
          found = true;
        }
        return !found;
      });
      return found;
    };
    LintVisitor2.prototype.getEntries = function(filter) {
      if (filter === void 0) {
        filter = Level.Warning | Level.Error;
      }
      return this.warnings.filter(function(entry) {
        return (entry.getLevel() & filter) !== 0;
      });
    };
    LintVisitor2.prototype.addEntry = function(node, rule, details) {
      var entry = new Marker(node, rule, this.settings.getRule(rule), details);
      this.warnings.push(entry);
    };
    LintVisitor2.prototype.getMissingNames = function(expected, actual) {
      var expectedClone = expected.slice(0);
      for (var i = 0; i < actual.length; i++) {
        var k = expectedClone.indexOf(actual[i]);
        if (k !== -1) {
          expectedClone[k] = null;
        }
      }
      var result = null;
      for (var i = 0; i < expectedClone.length; i++) {
        var curr = expectedClone[i];
        if (curr) {
          if (result === null) {
            result = localize9("namelist.single", "'{0}'", curr);
          } else {
            result = localize9("namelist.concatenated", "{0}, '{1}'", result, curr);
          }
        }
      }
      return result;
    };
    LintVisitor2.prototype.visitNode = function(node) {
      switch (node.type) {
        case NodeType.UnknownAtRule:
          return this.visitUnknownAtRule(node);
        case NodeType.Keyframe:
          return this.visitKeyframe(node);
        case NodeType.FontFace:
          return this.visitFontFace(node);
        case NodeType.Ruleset:
          return this.visitRuleSet(node);
        case NodeType.SimpleSelector:
          return this.visitSimpleSelector(node);
        case NodeType.Function:
          return this.visitFunction(node);
        case NodeType.NumericValue:
          return this.visitNumericValue(node);
        case NodeType.Import:
          return this.visitImport(node);
        case NodeType.HexColorValue:
          return this.visitHexColorValue(node);
        case NodeType.Prio:
          return this.visitPrio(node);
      }
      return true;
    };
    LintVisitor2.prototype.completeValidations = function() {
      this.validateKeyframes();
    };
    LintVisitor2.prototype.visitUnknownAtRule = function(node) {
      var atRuleName = node.getChild(0);
      if (!atRuleName) {
        return false;
      }
      var atDirective = this.cssDataManager.getAtDirective(atRuleName.getText());
      if (atDirective) {
        return false;
      }
      this.addEntry(atRuleName, Rules.UnknownAtRules, "Unknown at rule " + atRuleName.getText());
      return true;
    };
    LintVisitor2.prototype.visitKeyframe = function(node) {
      var keyword = node.getKeyword();
      if (!keyword) {
        return false;
      }
      var text = keyword.getText();
      this.keyframes.add(node.getName(), text, text !== "@keyframes" ? keyword : null);
      return true;
    };
    LintVisitor2.prototype.validateKeyframes = function() {
      var expected = ["@-webkit-keyframes", "@-moz-keyframes", "@-o-keyframes"];
      for (var name in this.keyframes.data) {
        var actual = this.keyframes.data[name].names;
        var needsStandard = actual.indexOf("@keyframes") === -1;
        if (!needsStandard && actual.length === 1) {
          continue;
        }
        var missingVendorSpecific = this.getMissingNames(expected, actual);
        if (missingVendorSpecific || needsStandard) {
          for (var _i = 0, _a3 = this.keyframes.data[name].nodes; _i < _a3.length; _i++) {
            var node = _a3[_i];
            if (needsStandard) {
              var message = localize9("keyframes.standardrule.missing", "Always define standard rule '@keyframes' when defining keyframes.");
              this.addEntry(node, Rules.IncludeStandardPropertyWhenUsingVendorPrefix, message);
            }
            if (missingVendorSpecific) {
              var message = localize9("keyframes.vendorspecific.missing", "Always include all vendor specific rules: Missing: {0}", missingVendorSpecific);
              this.addEntry(node, Rules.AllVendorPrefixes, message);
            }
          }
        }
      }
      return true;
    };
    LintVisitor2.prototype.visitSimpleSelector = function(node) {
      var firstChar = this.documentText.charAt(node.offset);
      if (node.length === 1 && firstChar === "*") {
        this.addEntry(node, Rules.UniversalSelector);
      }
      if (firstChar === "#") {
        this.addEntry(node, Rules.AvoidIdSelector);
      }
      return true;
    };
    LintVisitor2.prototype.visitImport = function(node) {
      this.addEntry(node, Rules.ImportStatemement);
      return true;
    };
    LintVisitor2.prototype.visitRuleSet = function(node) {
      var declarations = node.getDeclarations();
      if (!declarations) {
        return false;
      }
      if (!declarations.hasChildren()) {
        this.addEntry(node.getSelectors(), Rules.EmptyRuleSet);
      }
      var propertyTable = [];
      for (var _i = 0, _a3 = declarations.getChildren(); _i < _a3.length; _i++) {
        var element = _a3[_i];
        if (element instanceof Declaration) {
          propertyTable.push(new Element2(element));
        }
      }
      var boxModel = calculateBoxModel(propertyTable);
      if (boxModel.width) {
        var properties = [];
        if (boxModel.right.value) {
          properties = union(properties, boxModel.right.properties);
        }
        if (boxModel.left.value) {
          properties = union(properties, boxModel.left.properties);
        }
        if (properties.length !== 0) {
          for (var _b = 0, properties_1 = properties; _b < properties_1.length; _b++) {
            var item = properties_1[_b];
            this.addEntry(item.node, Rules.BewareOfBoxModelSize);
          }
          this.addEntry(boxModel.width.node, Rules.BewareOfBoxModelSize);
        }
      }
      if (boxModel.height) {
        var properties = [];
        if (boxModel.top.value) {
          properties = union(properties, boxModel.top.properties);
        }
        if (boxModel.bottom.value) {
          properties = union(properties, boxModel.bottom.properties);
        }
        if (properties.length !== 0) {
          for (var _c = 0, properties_2 = properties; _c < properties_2.length; _c++) {
            var item = properties_2[_c];
            this.addEntry(item.node, Rules.BewareOfBoxModelSize);
          }
          this.addEntry(boxModel.height.node, Rules.BewareOfBoxModelSize);
        }
      }
      var displayElems = this.fetchWithValue(propertyTable, "display", "inline");
      if (displayElems.length > 0) {
        for (var _d = 0, _e = ["width", "height", "margin-top", "margin-bottom", "float"]; _d < _e.length; _d++) {
          var prop = _e[_d];
          var elem = this.fetch(propertyTable, prop);
          for (var index = 0; index < elem.length; index++) {
            var node_1 = elem[index].node;
            var value = node_1.getValue();
            if (prop === "float" && (!value || value.matches("none"))) {
              continue;
            }
            this.addEntry(node_1, Rules.PropertyIgnoredDueToDisplay, localize9("rule.propertyIgnoredDueToDisplayInline", "Property is ignored due to the display. With 'display: inline', the width, height, margin-top, margin-bottom, and float properties have no effect."));
          }
        }
      }
      displayElems = this.fetchWithValue(propertyTable, "display", "inline-block");
      if (displayElems.length > 0) {
        var elem = this.fetch(propertyTable, "float");
        for (var index = 0; index < elem.length; index++) {
          var node_2 = elem[index].node;
          var value = node_2.getValue();
          if (value && !value.matches("none")) {
            this.addEntry(node_2, Rules.PropertyIgnoredDueToDisplay, localize9("rule.propertyIgnoredDueToDisplayInlineBlock", "inline-block is ignored due to the float. If 'float' has a value other than 'none', the box is floated and 'display' is treated as 'block'"));
          }
        }
      }
      displayElems = this.fetchWithValue(propertyTable, "display", "block");
      if (displayElems.length > 0) {
        var elem = this.fetch(propertyTable, "vertical-align");
        for (var index = 0; index < elem.length; index++) {
          this.addEntry(elem[index].node, Rules.PropertyIgnoredDueToDisplay, localize9("rule.propertyIgnoredDueToDisplayBlock", "Property is ignored due to the display. With 'display: block', vertical-align should not be used."));
        }
      }
      var elements = this.fetch(propertyTable, "float");
      for (var index = 0; index < elements.length; index++) {
        var element = elements[index];
        if (!this.isValidPropertyDeclaration(element)) {
          this.addEntry(element.node, Rules.AvoidFloat);
        }
      }
      for (var i = 0; i < propertyTable.length; i++) {
        var element = propertyTable[i];
        if (element.fullPropertyName !== "background" && !this.validProperties[element.fullPropertyName]) {
          var value = element.node.getValue();
          if (value && this.documentText.charAt(value.offset) !== "-") {
            var elements_1 = this.fetch(propertyTable, element.fullPropertyName);
            if (elements_1.length > 1) {
              for (var k = 0; k < elements_1.length; k++) {
                var value_1 = elements_1[k].node.getValue();
                if (value_1 && this.documentText.charAt(value_1.offset) !== "-" && elements_1[k] !== element) {
                  this.addEntry(element.node, Rules.DuplicateDeclarations);
                }
              }
            }
          }
        }
      }
      var isExportBlock = node.getSelectors().matches(":export");
      if (!isExportBlock) {
        var propertiesBySuffix = new NodesByRootMap();
        var containsUnknowns = false;
        for (var _f2 = 0, propertyTable_1 = propertyTable; _f2 < propertyTable_1.length; _f2++) {
          var element = propertyTable_1[_f2];
          var decl = element.node;
          if (this.isCSSDeclaration(decl)) {
            var name = element.fullPropertyName;
            var firstChar = name.charAt(0);
            if (firstChar === "-") {
              if (name.charAt(1) !== "-") {
                if (!this.cssDataManager.isKnownProperty(name) && !this.validProperties[name]) {
                  this.addEntry(decl.getProperty(), Rules.UnknownVendorSpecificProperty);
                }
                var nonPrefixedName = decl.getNonPrefixedPropertyName();
                propertiesBySuffix.add(nonPrefixedName, name, decl.getProperty());
              }
            } else {
              var fullName = name;
              if (firstChar === "*" || firstChar === "_") {
                this.addEntry(decl.getProperty(), Rules.IEStarHack);
                name = name.substr(1);
              }
              if (!this.cssDataManager.isKnownProperty(fullName) && !this.cssDataManager.isKnownProperty(name)) {
                if (!this.validProperties[name]) {
                  this.addEntry(decl.getProperty(), Rules.UnknownProperty, localize9("property.unknownproperty.detailed", "Unknown property: '{0}'", decl.getFullPropertyName()));
                }
              }
              propertiesBySuffix.add(name, name, null);
            }
          } else {
            containsUnknowns = true;
          }
        }
        if (!containsUnknowns) {
          for (var suffix in propertiesBySuffix.data) {
            var entry = propertiesBySuffix.data[suffix];
            var actual = entry.names;
            var needsStandard = this.cssDataManager.isStandardProperty(suffix) && actual.indexOf(suffix) === -1;
            if (!needsStandard && actual.length === 1) {
              continue;
            }
            var expected = [];
            for (var i = 0, len = LintVisitor2.prefixes.length; i < len; i++) {
              var prefix = LintVisitor2.prefixes[i];
              if (this.cssDataManager.isStandardProperty(prefix + suffix)) {
                expected.push(prefix + suffix);
              }
            }
            var missingVendorSpecific = this.getMissingNames(expected, actual);
            if (missingVendorSpecific || needsStandard) {
              for (var _g = 0, _h = entry.nodes; _g < _h.length; _g++) {
                var node_3 = _h[_g];
                if (needsStandard) {
                  var message = localize9("property.standard.missing", "Also define the standard property '{0}' for compatibility", suffix);
                  this.addEntry(node_3, Rules.IncludeStandardPropertyWhenUsingVendorPrefix, message);
                }
                if (missingVendorSpecific) {
                  var message = localize9("property.vendorspecific.missing", "Always include all vendor specific properties: Missing: {0}", missingVendorSpecific);
                  this.addEntry(node_3, Rules.AllVendorPrefixes, message);
                }
              }
            }
          }
        }
      }
      return true;
    };
    LintVisitor2.prototype.visitPrio = function(node) {
      this.addEntry(node, Rules.AvoidImportant);
      return true;
    };
    LintVisitor2.prototype.visitNumericValue = function(node) {
      var funcDecl = node.findParent(NodeType.Function);
      if (funcDecl && funcDecl.getName() === "calc") {
        return true;
      }
      var decl = node.findParent(NodeType.Declaration);
      if (decl) {
        var declValue = decl.getValue();
        if (declValue) {
          var value = node.getValue();
          if (!value.unit || units.length.indexOf(value.unit.toLowerCase()) === -1) {
            return true;
          }
          if (parseFloat(value.value) === 0 && !!value.unit && !this.validProperties[decl.getFullPropertyName()]) {
            this.addEntry(node, Rules.ZeroWithUnit);
          }
        }
      }
      return true;
    };
    LintVisitor2.prototype.visitFontFace = function(node) {
      var declarations = node.getDeclarations();
      if (!declarations) {
        return false;
      }
      var definesSrc = false, definesFontFamily = false;
      var containsUnknowns = false;
      for (var _i = 0, _a3 = declarations.getChildren(); _i < _a3.length; _i++) {
        var node_4 = _a3[_i];
        if (this.isCSSDeclaration(node_4)) {
          var name = node_4.getProperty().getName().toLowerCase();
          if (name === "src") {
            definesSrc = true;
          }
          if (name === "font-family") {
            definesFontFamily = true;
          }
        } else {
          containsUnknowns = true;
        }
      }
      if (!containsUnknowns && (!definesSrc || !definesFontFamily)) {
        this.addEntry(node, Rules.RequiredPropertiesForFontFace);
      }
      return true;
    };
    LintVisitor2.prototype.isCSSDeclaration = function(node) {
      if (node instanceof Declaration) {
        if (!node.getValue()) {
          return false;
        }
        var property = node.getProperty();
        if (!property) {
          return false;
        }
        var identifier = property.getIdentifier();
        if (!identifier || identifier.containsInterpolation()) {
          return false;
        }
        return true;
      }
      return false;
    };
    LintVisitor2.prototype.visitHexColorValue = function(node) {
      var length = node.length;
      if (length !== 9 && length !== 7 && length !== 5 && length !== 4) {
        this.addEntry(node, Rules.HexColorLength);
      }
      return false;
    };
    LintVisitor2.prototype.visitFunction = function(node) {
      var fnName = node.getName().toLowerCase();
      var expectedAttrCount = -1;
      var actualAttrCount = 0;
      switch (fnName) {
        case "rgb(":
        case "hsl(":
          expectedAttrCount = 3;
          break;
        case "rgba(":
        case "hsla(":
          expectedAttrCount = 4;
          break;
      }
      if (expectedAttrCount !== -1) {
        node.getArguments().accept(function(n) {
          if (n instanceof BinaryExpression) {
            actualAttrCount += 1;
            return false;
          }
          return true;
        });
        if (actualAttrCount !== expectedAttrCount) {
          this.addEntry(node, Rules.ArgsInColorFunction);
        }
      }
      return true;
    };
    LintVisitor2.prefixes = [
      "-ms-",
      "-moz-",
      "-o-",
      "-webkit-"
    ];
    return LintVisitor2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssValidation.js
  var CSSValidation = function() {
    function CSSValidation2(cssDataManager) {
      this.cssDataManager = cssDataManager;
    }
    CSSValidation2.prototype.configure = function(settings) {
      this.settings = settings;
    };
    CSSValidation2.prototype.doValidation = function(document, stylesheet, settings) {
      if (settings === void 0) {
        settings = this.settings;
      }
      if (settings && settings.validate === false) {
        return [];
      }
      var entries = [];
      entries.push.apply(entries, ParseErrorCollector.entries(stylesheet));
      entries.push.apply(entries, LintVisitor.entries(stylesheet, document, new LintConfigurationSettings(settings && settings.lint), this.cssDataManager));
      var ruleIds = [];
      for (var r in Rules) {
        ruleIds.push(Rules[r].id);
      }
      function toDiagnostic(marker) {
        var range = Range2.create(document.positionAt(marker.getOffset()), document.positionAt(marker.getOffset() + marker.getLength()));
        var source = document.languageId;
        return {
          code: marker.getRule().id,
          source,
          message: marker.getMessage(),
          severity: marker.getLevel() === Level.Warning ? DiagnosticSeverity.Warning : DiagnosticSeverity.Error,
          range
        };
      }
      return entries.filter(function(entry) {
        return entry.getLevel() !== Level.Ignore;
      }).map(toDiagnostic);
    };
    return CSSValidation2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/scssScanner.js
  var __extends5 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var _FSL2 = "/".charCodeAt(0);
  var _NWL2 = "\n".charCodeAt(0);
  var _CAR2 = "\r".charCodeAt(0);
  var _LFD2 = "\f".charCodeAt(0);
  var _DLR2 = "$".charCodeAt(0);
  var _HSH2 = "#".charCodeAt(0);
  var _CUL2 = "{".charCodeAt(0);
  var _EQS2 = "=".charCodeAt(0);
  var _BNG2 = "!".charCodeAt(0);
  var _LAN2 = "<".charCodeAt(0);
  var _RAN2 = ">".charCodeAt(0);
  var _DOT2 = ".".charCodeAt(0);
  var _ATS2 = "@".charCodeAt(0);
  var customTokenValue = TokenType.CustomToken;
  var VariableName = customTokenValue++;
  var InterpolationFunction = customTokenValue++;
  var Default = customTokenValue++;
  var EqualsOperator = customTokenValue++;
  var NotEqualsOperator = customTokenValue++;
  var GreaterEqualsOperator = customTokenValue++;
  var SmallerEqualsOperator = customTokenValue++;
  var Ellipsis = customTokenValue++;
  var Module2 = customTokenValue++;
  var SCSSScanner = function(_super) {
    __extends5(SCSSScanner2, _super);
    function SCSSScanner2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    SCSSScanner2.prototype.scanNext = function(offset) {
      if (this.stream.advanceIfChar(_DLR2)) {
        var content = ["$"];
        if (this.ident(content)) {
          return this.finishToken(offset, VariableName, content.join(""));
        } else {
          this.stream.goBackTo(offset);
        }
      }
      if (this.stream.advanceIfChars([_HSH2, _CUL2])) {
        return this.finishToken(offset, InterpolationFunction);
      }
      if (this.stream.advanceIfChars([_EQS2, _EQS2])) {
        return this.finishToken(offset, EqualsOperator);
      }
      if (this.stream.advanceIfChars([_BNG2, _EQS2])) {
        return this.finishToken(offset, NotEqualsOperator);
      }
      if (this.stream.advanceIfChar(_LAN2)) {
        if (this.stream.advanceIfChar(_EQS2)) {
          return this.finishToken(offset, SmallerEqualsOperator);
        }
        return this.finishToken(offset, TokenType.Delim);
      }
      if (this.stream.advanceIfChar(_RAN2)) {
        if (this.stream.advanceIfChar(_EQS2)) {
          return this.finishToken(offset, GreaterEqualsOperator);
        }
        return this.finishToken(offset, TokenType.Delim);
      }
      if (this.stream.advanceIfChars([_DOT2, _DOT2, _DOT2])) {
        return this.finishToken(offset, Ellipsis);
      }
      return _super.prototype.scanNext.call(this, offset);
    };
    SCSSScanner2.prototype.comment = function() {
      if (_super.prototype.comment.call(this)) {
        return true;
      }
      if (!this.inURL && this.stream.advanceIfChars([_FSL2, _FSL2])) {
        this.stream.advanceWhileChar(function(ch) {
          switch (ch) {
            case _NWL2:
            case _CAR2:
            case _LFD2:
              return false;
            default:
              return true;
          }
        });
        return true;
      } else {
        return false;
      }
    };
    return SCSSScanner2;
  }(Scanner);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/scssErrors.js
  var localize10 = loadMessageBundle();
  var SCSSIssueType = function() {
    function SCSSIssueType2(id, message) {
      this.id = id;
      this.message = message;
    }
    return SCSSIssueType2;
  }();
  var SCSSParseError = {
    FromExpected: new SCSSIssueType("scss-fromexpected", localize10("expected.from", "'from' expected")),
    ThroughOrToExpected: new SCSSIssueType("scss-throughexpected", localize10("expected.through", "'through' or 'to' expected")),
    InExpected: new SCSSIssueType("scss-fromexpected", localize10("expected.in", "'in' expected"))
  };

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/scssParser.js
  var __extends6 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var SCSSParser = function(_super) {
    __extends6(SCSSParser2, _super);
    function SCSSParser2() {
      return _super.call(this, new SCSSScanner()) || this;
    }
    SCSSParser2.prototype._parseStylesheetStatement = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseWarnAndDebug() || this._parseControlStatement() || this._parseMixinDeclaration() || this._parseMixinContent() || this._parseMixinReference() || this._parseFunctionDeclaration() || this._parseForward() || this._parseUse() || this._parseRuleset(isNested) || _super.prototype._parseStylesheetAtStatement.call(this, isNested);
      }
      return this._parseRuleset(true) || this._parseVariableDeclaration();
    };
    SCSSParser2.prototype._parseImport = function() {
      if (!this.peekKeyword("@import")) {
        return null;
      }
      var node = this.create(Import);
      this.consumeToken();
      if (!node.addChild(this._parseURILiteral()) && !node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.URIOrStringExpected);
      }
      while (this.accept(TokenType.Comma)) {
        if (!node.addChild(this._parseURILiteral()) && !node.addChild(this._parseStringLiteral())) {
          return this.finish(node, ParseError.URIOrStringExpected);
        }
      }
      if (!this.peek(TokenType.SemiColon) && !this.peek(TokenType.EOF)) {
        node.setMedialist(this._parseMediaQueryList());
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseVariableDeclaration = function(panic) {
      if (panic === void 0) {
        panic = [];
      }
      if (!this.peek(VariableName)) {
        return null;
      }
      var node = this.create(VariableDeclaration);
      if (!node.setVariable(this._parseVariable())) {
        return null;
      }
      if (!this.accept(TokenType.Colon)) {
        return this.finish(node, ParseError.ColonExpected);
      }
      if (this.prevToken) {
        node.colonPosition = this.prevToken.offset;
      }
      if (!node.setValue(this._parseExpr())) {
        return this.finish(node, ParseError.VariableValueExpected, [], panic);
      }
      while (this.peek(TokenType.Exclamation)) {
        if (node.addChild(this._tryParsePrio())) {
        } else {
          this.consumeToken();
          if (!this.peekRegExp(TokenType.Ident, /^(default|global)$/)) {
            return this.finish(node, ParseError.UnknownKeyword);
          }
          this.consumeToken();
        }
      }
      if (this.peek(TokenType.SemiColon)) {
        node.semicolonPosition = this.token.offset;
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMediaContentStart = function() {
      return this._parseInterpolation();
    };
    SCSSParser2.prototype._parseMediaFeatureName = function() {
      return this._parseModuleMember() || this._parseFunction() || this._parseIdent() || this._parseVariable();
    };
    SCSSParser2.prototype._parseKeyframeSelector = function() {
      return this._tryParseKeyframeSelector() || this._parseControlStatement(this._parseKeyframeSelector.bind(this)) || this._parseVariableDeclaration() || this._parseMixinContent();
    };
    SCSSParser2.prototype._parseVariable = function() {
      if (!this.peek(VariableName)) {
        return null;
      }
      var node = this.create(Variable);
      this.consumeToken();
      return node;
    };
    SCSSParser2.prototype._parseModuleMember = function() {
      var pos = this.mark();
      var node = this.create(Module);
      if (!node.setIdentifier(this._parseIdent([ReferenceType.Module]))) {
        return null;
      }
      if (this.hasWhitespace() || !this.acceptDelim(".") || this.hasWhitespace()) {
        this.restoreAtMark(pos);
        return null;
      }
      if (!node.addChild(this._parseVariable() || this._parseFunction())) {
        return this.finish(node, ParseError.IdentifierOrVariableExpected);
      }
      return node;
    };
    SCSSParser2.prototype._parseIdent = function(referenceTypes) {
      var _this = this;
      if (!this.peek(TokenType.Ident) && !this.peek(InterpolationFunction) && !this.peekDelim("-")) {
        return null;
      }
      var node = this.create(Identifier);
      node.referenceTypes = referenceTypes;
      node.isCustomProperty = this.peekRegExp(TokenType.Ident, /^--/);
      var hasContent = false;
      var indentInterpolation = function() {
        var pos = _this.mark();
        if (_this.acceptDelim("-")) {
          if (!_this.hasWhitespace()) {
            _this.acceptDelim("-");
          }
          if (_this.hasWhitespace()) {
            _this.restoreAtMark(pos);
            return null;
          }
        }
        return _this._parseInterpolation();
      };
      while (this.accept(TokenType.Ident) || node.addChild(indentInterpolation()) || hasContent && this.acceptRegexp(/[\w-]/)) {
        hasContent = true;
        if (this.hasWhitespace()) {
          break;
        }
      }
      return hasContent ? this.finish(node) : null;
    };
    SCSSParser2.prototype._parseTermExpression = function() {
      return this._parseModuleMember() || this._parseVariable() || this._parseSelectorCombinator() || _super.prototype._parseTermExpression.call(this);
    };
    SCSSParser2.prototype._parseInterpolation = function() {
      if (this.peek(InterpolationFunction)) {
        var node = this.create(Interpolation);
        this.consumeToken();
        if (!node.addChild(this._parseExpr()) && !this._parseSelectorCombinator()) {
          if (this.accept(TokenType.CurlyR)) {
            return this.finish(node);
          }
          return this.finish(node, ParseError.ExpressionExpected);
        }
        if (!this.accept(TokenType.CurlyR)) {
          return this.finish(node, ParseError.RightCurlyExpected);
        }
        return this.finish(node);
      }
      return null;
    };
    SCSSParser2.prototype._parseOperator = function() {
      if (this.peek(EqualsOperator) || this.peek(NotEqualsOperator) || this.peek(GreaterEqualsOperator) || this.peek(SmallerEqualsOperator) || this.peekDelim(">") || this.peekDelim("<") || this.peekIdent("and") || this.peekIdent("or") || this.peekDelim("%")) {
        var node = this.createNode(NodeType.Operator);
        this.consumeToken();
        return this.finish(node);
      }
      return _super.prototype._parseOperator.call(this);
    };
    SCSSParser2.prototype._parseUnaryOperator = function() {
      if (this.peekIdent("not")) {
        var node = this.create(Node2);
        this.consumeToken();
        return this.finish(node);
      }
      return _super.prototype._parseUnaryOperator.call(this);
    };
    SCSSParser2.prototype._parseRuleSetDeclaration = function() {
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseKeyframe() || this._parseImport() || this._parseMedia(true) || this._parseFontFace() || this._parseWarnAndDebug() || this._parseControlStatement() || this._parseFunctionDeclaration() || this._parseExtends() || this._parseMixinReference() || this._parseMixinContent() || this._parseMixinDeclaration() || this._parseRuleset(true) || this._parseSupports(true) || _super.prototype._parseRuleSetDeclarationAtStatement.call(this);
      }
      return this._parseVariableDeclaration() || this._tryParseRuleset(true) || _super.prototype._parseRuleSetDeclaration.call(this);
    };
    SCSSParser2.prototype._parseDeclaration = function(resyncStopTokens) {
      var node = this.create(Declaration);
      if (!node.setProperty(this._parseProperty())) {
        return null;
      }
      if (!this.accept(TokenType.Colon)) {
        return this.finish(node, ParseError.ColonExpected, [TokenType.Colon], resyncStopTokens);
      }
      if (this.prevToken) {
        node.colonPosition = this.prevToken.offset;
      }
      var hasContent = false;
      if (node.setValue(this._parseExpr())) {
        hasContent = true;
        node.addChild(this._parsePrio());
      }
      if (this.peek(TokenType.CurlyL)) {
        node.setNestedProperties(this._parseNestedProperties());
      } else {
        if (!hasContent) {
          return this.finish(node, ParseError.PropertyValueExpected);
        }
      }
      if (this.peek(TokenType.SemiColon)) {
        node.semicolonPosition = this.token.offset;
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseNestedProperties = function() {
      var node = this.create(NestedProperties);
      return this._parseBody(node, this._parseDeclaration.bind(this));
    };
    SCSSParser2.prototype._parseExtends = function() {
      if (this.peekKeyword("@extend")) {
        var node = this.create(ExtendsReference);
        this.consumeToken();
        if (!node.getSelectors().addChild(this._parseSimpleSelector())) {
          return this.finish(node, ParseError.SelectorExpected);
        }
        while (this.accept(TokenType.Comma)) {
          node.getSelectors().addChild(this._parseSimpleSelector());
        }
        if (this.accept(TokenType.Exclamation)) {
          if (!this.acceptIdent("optional")) {
            return this.finish(node, ParseError.UnknownKeyword);
          }
        }
        return this.finish(node);
      }
      return null;
    };
    SCSSParser2.prototype._parseSimpleSelectorBody = function() {
      return this._parseSelectorCombinator() || this._parseSelectorPlaceholder() || _super.prototype._parseSimpleSelectorBody.call(this);
    };
    SCSSParser2.prototype._parseSelectorCombinator = function() {
      if (this.peekDelim("&")) {
        var node = this.createNode(NodeType.SelectorCombinator);
        this.consumeToken();
        while (!this.hasWhitespace() && (this.acceptDelim("-") || this.accept(TokenType.Num) || this.accept(TokenType.Dimension) || node.addChild(this._parseIdent()) || this.acceptDelim("&"))) {
        }
        return this.finish(node);
      }
      return null;
    };
    SCSSParser2.prototype._parseSelectorPlaceholder = function() {
      if (this.peekDelim("%")) {
        var node = this.createNode(NodeType.SelectorPlaceholder);
        this.consumeToken();
        this._parseIdent();
        return this.finish(node);
      } else if (this.peekKeyword("@at-root")) {
        var node = this.createNode(NodeType.SelectorPlaceholder);
        this.consumeToken();
        return this.finish(node);
      }
      return null;
    };
    SCSSParser2.prototype._parseElementName = function() {
      var pos = this.mark();
      var node = _super.prototype._parseElementName.call(this);
      if (node && !this.hasWhitespace() && this.peek(TokenType.ParenthesisL)) {
        this.restoreAtMark(pos);
        return null;
      }
      return node;
    };
    SCSSParser2.prototype._tryParsePseudoIdentifier = function() {
      return this._parseInterpolation() || _super.prototype._tryParsePseudoIdentifier.call(this);
    };
    SCSSParser2.prototype._parseWarnAndDebug = function() {
      if (!this.peekKeyword("@debug") && !this.peekKeyword("@warn") && !this.peekKeyword("@error")) {
        return null;
      }
      var node = this.createNode(NodeType.Debug);
      this.consumeToken();
      node.addChild(this._parseExpr());
      return this.finish(node);
    };
    SCSSParser2.prototype._parseControlStatement = function(parseStatement) {
      if (parseStatement === void 0) {
        parseStatement = this._parseRuleSetDeclaration.bind(this);
      }
      if (!this.peek(TokenType.AtKeyword)) {
        return null;
      }
      return this._parseIfStatement(parseStatement) || this._parseForStatement(parseStatement) || this._parseEachStatement(parseStatement) || this._parseWhileStatement(parseStatement);
    };
    SCSSParser2.prototype._parseIfStatement = function(parseStatement) {
      if (!this.peekKeyword("@if")) {
        return null;
      }
      return this._internalParseIfStatement(parseStatement);
    };
    SCSSParser2.prototype._internalParseIfStatement = function(parseStatement) {
      var node = this.create(IfStatement);
      this.consumeToken();
      if (!node.setExpression(this._parseExpr(true))) {
        return this.finish(node, ParseError.ExpressionExpected);
      }
      this._parseBody(node, parseStatement);
      if (this.acceptKeyword("@else")) {
        if (this.peekIdent("if")) {
          node.setElseClause(this._internalParseIfStatement(parseStatement));
        } else if (this.peek(TokenType.CurlyL)) {
          var elseNode = this.create(ElseStatement);
          this._parseBody(elseNode, parseStatement);
          node.setElseClause(elseNode);
        }
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseForStatement = function(parseStatement) {
      if (!this.peekKeyword("@for")) {
        return null;
      }
      var node = this.create(ForStatement);
      this.consumeToken();
      if (!node.setVariable(this._parseVariable())) {
        return this.finish(node, ParseError.VariableNameExpected, [TokenType.CurlyR]);
      }
      if (!this.acceptIdent("from")) {
        return this.finish(node, SCSSParseError.FromExpected, [TokenType.CurlyR]);
      }
      if (!node.addChild(this._parseBinaryExpr())) {
        return this.finish(node, ParseError.ExpressionExpected, [TokenType.CurlyR]);
      }
      if (!this.acceptIdent("to") && !this.acceptIdent("through")) {
        return this.finish(node, SCSSParseError.ThroughOrToExpected, [TokenType.CurlyR]);
      }
      if (!node.addChild(this._parseBinaryExpr())) {
        return this.finish(node, ParseError.ExpressionExpected, [TokenType.CurlyR]);
      }
      return this._parseBody(node, parseStatement);
    };
    SCSSParser2.prototype._parseEachStatement = function(parseStatement) {
      if (!this.peekKeyword("@each")) {
        return null;
      }
      var node = this.create(EachStatement);
      this.consumeToken();
      var variables = node.getVariables();
      if (!variables.addChild(this._parseVariable())) {
        return this.finish(node, ParseError.VariableNameExpected, [TokenType.CurlyR]);
      }
      while (this.accept(TokenType.Comma)) {
        if (!variables.addChild(this._parseVariable())) {
          return this.finish(node, ParseError.VariableNameExpected, [TokenType.CurlyR]);
        }
      }
      this.finish(variables);
      if (!this.acceptIdent("in")) {
        return this.finish(node, SCSSParseError.InExpected, [TokenType.CurlyR]);
      }
      if (!node.addChild(this._parseExpr())) {
        return this.finish(node, ParseError.ExpressionExpected, [TokenType.CurlyR]);
      }
      return this._parseBody(node, parseStatement);
    };
    SCSSParser2.prototype._parseWhileStatement = function(parseStatement) {
      if (!this.peekKeyword("@while")) {
        return null;
      }
      var node = this.create(WhileStatement);
      this.consumeToken();
      if (!node.addChild(this._parseBinaryExpr())) {
        return this.finish(node, ParseError.ExpressionExpected, [TokenType.CurlyR]);
      }
      return this._parseBody(node, parseStatement);
    };
    SCSSParser2.prototype._parseFunctionBodyDeclaration = function() {
      return this._parseVariableDeclaration() || this._parseReturnStatement() || this._parseWarnAndDebug() || this._parseControlStatement(this._parseFunctionBodyDeclaration.bind(this));
    };
    SCSSParser2.prototype._parseFunctionDeclaration = function() {
      if (!this.peekKeyword("@function")) {
        return null;
      }
      var node = this.create(FunctionDeclaration);
      this.consumeToken();
      if (!node.setIdentifier(this._parseIdent([ReferenceType.Function]))) {
        return this.finish(node, ParseError.IdentifierExpected, [TokenType.CurlyR]);
      }
      if (!this.accept(TokenType.ParenthesisL)) {
        return this.finish(node, ParseError.LeftParenthesisExpected, [TokenType.CurlyR]);
      }
      if (node.getParameters().addChild(this._parseParameterDeclaration())) {
        while (this.accept(TokenType.Comma)) {
          if (this.peek(TokenType.ParenthesisR)) {
            break;
          }
          if (!node.getParameters().addChild(this._parseParameterDeclaration())) {
            return this.finish(node, ParseError.VariableNameExpected);
          }
        }
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected, [TokenType.CurlyR]);
      }
      return this._parseBody(node, this._parseFunctionBodyDeclaration.bind(this));
    };
    SCSSParser2.prototype._parseReturnStatement = function() {
      if (!this.peekKeyword("@return")) {
        return null;
      }
      var node = this.createNode(NodeType.ReturnStatement);
      this.consumeToken();
      if (!node.addChild(this._parseExpr())) {
        return this.finish(node, ParseError.ExpressionExpected);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMixinDeclaration = function() {
      if (!this.peekKeyword("@mixin")) {
        return null;
      }
      var node = this.create(MixinDeclaration);
      this.consumeToken();
      if (!node.setIdentifier(this._parseIdent([ReferenceType.Mixin]))) {
        return this.finish(node, ParseError.IdentifierExpected, [TokenType.CurlyR]);
      }
      if (this.accept(TokenType.ParenthesisL)) {
        if (node.getParameters().addChild(this._parseParameterDeclaration())) {
          while (this.accept(TokenType.Comma)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getParameters().addChild(this._parseParameterDeclaration())) {
              return this.finish(node, ParseError.VariableNameExpected);
            }
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected, [TokenType.CurlyR]);
        }
      }
      return this._parseBody(node, this._parseRuleSetDeclaration.bind(this));
    };
    SCSSParser2.prototype._parseParameterDeclaration = function() {
      var node = this.create(FunctionParameter);
      if (!node.setIdentifier(this._parseVariable())) {
        return null;
      }
      if (this.accept(Ellipsis)) {
      }
      if (this.accept(TokenType.Colon)) {
        if (!node.setDefaultValue(this._parseExpr(true))) {
          return this.finish(node, ParseError.VariableValueExpected, [], [TokenType.Comma, TokenType.ParenthesisR]);
        }
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMixinContent = function() {
      if (!this.peekKeyword("@content")) {
        return null;
      }
      var node = this.create(MixinContentReference);
      this.consumeToken();
      if (this.accept(TokenType.ParenthesisL)) {
        if (node.getArguments().addChild(this._parseFunctionArgument())) {
          while (this.accept(TokenType.Comma)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getArguments().addChild(this._parseFunctionArgument())) {
              return this.finish(node, ParseError.ExpressionExpected);
            }
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected);
        }
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMixinReference = function() {
      if (!this.peekKeyword("@include")) {
        return null;
      }
      var node = this.create(MixinReference);
      this.consumeToken();
      var firstIdent = this._parseIdent([ReferenceType.Mixin]);
      if (!node.setIdentifier(firstIdent)) {
        return this.finish(node, ParseError.IdentifierExpected, [TokenType.CurlyR]);
      }
      if (!this.hasWhitespace() && this.acceptDelim(".") && !this.hasWhitespace()) {
        var secondIdent = this._parseIdent([ReferenceType.Mixin]);
        if (!secondIdent) {
          return this.finish(node, ParseError.IdentifierExpected, [TokenType.CurlyR]);
        }
        var moduleToken = this.create(Module);
        firstIdent.referenceTypes = [ReferenceType.Module];
        moduleToken.setIdentifier(firstIdent);
        node.setIdentifier(secondIdent);
        node.addChild(moduleToken);
      }
      if (this.accept(TokenType.ParenthesisL)) {
        if (node.getArguments().addChild(this._parseFunctionArgument())) {
          while (this.accept(TokenType.Comma)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getArguments().addChild(this._parseFunctionArgument())) {
              return this.finish(node, ParseError.ExpressionExpected);
            }
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected);
        }
      }
      if (this.peekIdent("using") || this.peek(TokenType.CurlyL)) {
        node.setContent(this._parseMixinContentDeclaration());
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMixinContentDeclaration = function() {
      var node = this.create(MixinContentDeclaration);
      if (this.acceptIdent("using")) {
        if (!this.accept(TokenType.ParenthesisL)) {
          return this.finish(node, ParseError.LeftParenthesisExpected, [TokenType.CurlyL]);
        }
        if (node.getParameters().addChild(this._parseParameterDeclaration())) {
          while (this.accept(TokenType.Comma)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getParameters().addChild(this._parseParameterDeclaration())) {
              return this.finish(node, ParseError.VariableNameExpected);
            }
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected, [TokenType.CurlyL]);
        }
      }
      if (this.peek(TokenType.CurlyL)) {
        this._parseBody(node, this._parseMixinReferenceBodyStatement.bind(this));
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseMixinReferenceBodyStatement = function() {
      return this._tryParseKeyframeSelector() || this._parseRuleSetDeclaration();
    };
    SCSSParser2.prototype._parseFunctionArgument = function() {
      var node = this.create(FunctionArgument);
      var pos = this.mark();
      var argument = this._parseVariable();
      if (argument) {
        if (!this.accept(TokenType.Colon)) {
          if (this.accept(Ellipsis)) {
            node.setValue(argument);
            return this.finish(node);
          } else {
            this.restoreAtMark(pos);
          }
        } else {
          node.setIdentifier(argument);
        }
      }
      if (node.setValue(this._parseExpr(true))) {
        this.accept(Ellipsis);
        node.addChild(this._parsePrio());
        return this.finish(node);
      } else if (node.setValue(this._tryParsePrio())) {
        return this.finish(node);
      }
      return null;
    };
    SCSSParser2.prototype._parseURLArgument = function() {
      var pos = this.mark();
      var node = _super.prototype._parseURLArgument.call(this);
      if (!node || !this.peek(TokenType.ParenthesisR)) {
        this.restoreAtMark(pos);
        var node_1 = this.create(Node2);
        node_1.addChild(this._parseBinaryExpr());
        return this.finish(node_1);
      }
      return node;
    };
    SCSSParser2.prototype._parseOperation = function() {
      if (!this.peek(TokenType.ParenthesisL)) {
        return null;
      }
      var node = this.create(Node2);
      this.consumeToken();
      while (node.addChild(this._parseListElement())) {
        this.accept(TokenType.Comma);
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseListElement = function() {
      var node = this.create(ListEntry);
      var child = this._parseBinaryExpr();
      if (!child) {
        return null;
      }
      if (this.accept(TokenType.Colon)) {
        node.setKey(child);
        if (!node.setValue(this._parseBinaryExpr())) {
          return this.finish(node, ParseError.ExpressionExpected);
        }
      } else {
        node.setValue(child);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseUse = function() {
      if (!this.peekKeyword("@use")) {
        return null;
      }
      var node = this.create(Use);
      this.consumeToken();
      if (!node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.StringLiteralExpected);
      }
      if (!this.peek(TokenType.SemiColon) && !this.peek(TokenType.EOF)) {
        if (!this.peekRegExp(TokenType.Ident, /as|with/)) {
          return this.finish(node, ParseError.UnknownKeyword);
        }
        if (this.acceptIdent("as") && (!node.setIdentifier(this._parseIdent([ReferenceType.Module])) && !this.acceptDelim("*"))) {
          return this.finish(node, ParseError.IdentifierOrWildcardExpected);
        }
        if (this.acceptIdent("with")) {
          if (!this.accept(TokenType.ParenthesisL)) {
            return this.finish(node, ParseError.LeftParenthesisExpected, [TokenType.ParenthesisR]);
          }
          if (!node.getParameters().addChild(this._parseModuleConfigDeclaration())) {
            return this.finish(node, ParseError.VariableNameExpected);
          }
          while (this.accept(TokenType.Comma)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getParameters().addChild(this._parseModuleConfigDeclaration())) {
              return this.finish(node, ParseError.VariableNameExpected);
            }
          }
          if (!this.accept(TokenType.ParenthesisR)) {
            return this.finish(node, ParseError.RightParenthesisExpected);
          }
        }
      }
      if (!this.accept(TokenType.SemiColon) && !this.accept(TokenType.EOF)) {
        return this.finish(node, ParseError.SemiColonExpected);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseModuleConfigDeclaration = function() {
      var node = this.create(ModuleConfiguration);
      if (!node.setIdentifier(this._parseVariable())) {
        return null;
      }
      if (!this.accept(TokenType.Colon) || !node.setValue(this._parseExpr(true))) {
        return this.finish(node, ParseError.VariableValueExpected, [], [TokenType.Comma, TokenType.ParenthesisR]);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseForward = function() {
      if (!this.peekKeyword("@forward")) {
        return null;
      }
      var node = this.create(Forward);
      this.consumeToken();
      if (!node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.StringLiteralExpected);
      }
      if (!this.peek(TokenType.SemiColon) && !this.peek(TokenType.EOF)) {
        if (!this.peekRegExp(TokenType.Ident, /as|hide|show/)) {
          return this.finish(node, ParseError.UnknownKeyword);
        }
        if (this.acceptIdent("as")) {
          var identifier = this._parseIdent([ReferenceType.Forward]);
          if (!node.setIdentifier(identifier)) {
            return this.finish(node, ParseError.IdentifierExpected);
          }
          if (this.hasWhitespace() || !this.acceptDelim("*")) {
            return this.finish(node, ParseError.WildcardExpected);
          }
        }
        if (this.peekIdent("hide") || this.peekIdent("show")) {
          if (!node.addChild(this._parseForwardVisibility())) {
            return this.finish(node, ParseError.IdentifierOrVariableExpected);
          }
        }
      }
      if (!this.accept(TokenType.SemiColon) && !this.accept(TokenType.EOF)) {
        return this.finish(node, ParseError.SemiColonExpected);
      }
      return this.finish(node);
    };
    SCSSParser2.prototype._parseForwardVisibility = function() {
      var node = this.create(ForwardVisibility);
      node.setIdentifier(this._parseIdent());
      while (node.addChild(this._parseVariable() || this._parseIdent())) {
      }
      return node.getChildren().length > 1 ? node : null;
    };
    SCSSParser2.prototype._parseSupportsCondition = function() {
      return this._parseInterpolation() || _super.prototype._parseSupportsCondition.call(this);
    };
    return SCSSParser2;
  }(Parser);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/scssCompletion.js
  var __extends7 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var localize11 = loadMessageBundle();
  var SCSSCompletion = function(_super) {
    __extends7(SCSSCompletion2, _super);
    function SCSSCompletion2(lsServiceOptions, cssDataManager) {
      var _this = _super.call(this, "$", lsServiceOptions, cssDataManager) || this;
      addReferencesToDocumentation(SCSSCompletion2.scssModuleLoaders);
      addReferencesToDocumentation(SCSSCompletion2.scssModuleBuiltIns);
      return _this;
    }
    SCSSCompletion2.prototype.isImportPathParent = function(type) {
      return type === NodeType.Forward || type === NodeType.Use || _super.prototype.isImportPathParent.call(this, type);
    };
    SCSSCompletion2.prototype.getCompletionForImportPath = function(importPathNode, result) {
      var parentType = importPathNode.getParent().type;
      if (parentType === NodeType.Forward || parentType === NodeType.Use) {
        for (var _i = 0, _a3 = SCSSCompletion2.scssModuleBuiltIns; _i < _a3.length; _i++) {
          var p = _a3[_i];
          var item = {
            label: p.label,
            documentation: p.documentation,
            textEdit: TextEdit.replace(this.getCompletionRange(importPathNode), "'" + p.label + "'"),
            kind: CompletionItemKind2.Module
          };
          result.items.push(item);
        }
      }
      return _super.prototype.getCompletionForImportPath.call(this, importPathNode, result);
    };
    SCSSCompletion2.prototype.createReplaceFunction = function() {
      var tabStopCounter = 1;
      return function(_match, p1) {
        return "\\" + p1 + ": ${" + tabStopCounter++ + ":" + (SCSSCompletion2.variableDefaults[p1] || "") + "}";
      };
    };
    SCSSCompletion2.prototype.createFunctionProposals = function(proposals, existingNode, sortToEnd, result) {
      for (var _i = 0, proposals_1 = proposals; _i < proposals_1.length; _i++) {
        var p = proposals_1[_i];
        var insertText = p.func.replace(/\[?(\$\w+)\]?/g, this.createReplaceFunction());
        var label = p.func.substr(0, p.func.indexOf("("));
        var item = {
          label,
          detail: p.func,
          documentation: p.desc,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), insertText),
          insertTextFormat: InsertTextFormat.Snippet,
          kind: CompletionItemKind2.Function
        };
        if (sortToEnd) {
          item.sortText = "z";
        }
        result.items.push(item);
      }
      return result;
    };
    SCSSCompletion2.prototype.getCompletionsForSelector = function(ruleSet, isNested, result) {
      this.createFunctionProposals(SCSSCompletion2.selectorFuncs, null, true, result);
      return _super.prototype.getCompletionsForSelector.call(this, ruleSet, isNested, result);
    };
    SCSSCompletion2.prototype.getTermProposals = function(entry, existingNode, result) {
      var functions = SCSSCompletion2.builtInFuncs;
      if (entry) {
        functions = functions.filter(function(f2) {
          return !f2.type || !entry.restrictions || entry.restrictions.indexOf(f2.type) !== -1;
        });
      }
      this.createFunctionProposals(functions, existingNode, true, result);
      return _super.prototype.getTermProposals.call(this, entry, existingNode, result);
    };
    SCSSCompletion2.prototype.getColorProposals = function(entry, existingNode, result) {
      this.createFunctionProposals(SCSSCompletion2.colorProposals, existingNode, false, result);
      return _super.prototype.getColorProposals.call(this, entry, existingNode, result);
    };
    SCSSCompletion2.prototype.getCompletionsForDeclarationProperty = function(declaration, result) {
      this.getCompletionForAtDirectives(result);
      this.getCompletionsForSelector(null, true, result);
      return _super.prototype.getCompletionsForDeclarationProperty.call(this, declaration, result);
    };
    SCSSCompletion2.prototype.getCompletionsForExtendsReference = function(_extendsRef, existingNode, result) {
      var symbols = this.getSymbolContext().findSymbolsAtOffset(this.offset, ReferenceType.Rule);
      for (var _i = 0, symbols_1 = symbols; _i < symbols_1.length; _i++) {
        var symbol = symbols_1[_i];
        var suggest = {
          label: symbol.name,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), symbol.name),
          kind: CompletionItemKind2.Function
        };
        result.items.push(suggest);
      }
      return result;
    };
    SCSSCompletion2.prototype.getCompletionForAtDirectives = function(result) {
      var _a3;
      (_a3 = result.items).push.apply(_a3, SCSSCompletion2.scssAtDirectives);
      return result;
    };
    SCSSCompletion2.prototype.getCompletionForTopLevel = function(result) {
      this.getCompletionForAtDirectives(result);
      this.getCompletionForModuleLoaders(result);
      _super.prototype.getCompletionForTopLevel.call(this, result);
      return result;
    };
    SCSSCompletion2.prototype.getCompletionForModuleLoaders = function(result) {
      var _a3;
      (_a3 = result.items).push.apply(_a3, SCSSCompletion2.scssModuleLoaders);
      return result;
    };
    SCSSCompletion2.variableDefaults = {
      "$red": "1",
      "$green": "2",
      "$blue": "3",
      "$alpha": "1.0",
      "$color": "#000000",
      "$weight": "0.5",
      "$hue": "0",
      "$saturation": "0%",
      "$lightness": "0%",
      "$degrees": "0",
      "$amount": "0",
      "$string": '""',
      "$substring": '"s"',
      "$number": "0",
      "$limit": "1"
    };
    SCSSCompletion2.colorProposals = [
      { func: "red($color)", desc: localize11("scss.builtin.red", "Gets the red component of a color.") },
      { func: "green($color)", desc: localize11("scss.builtin.green", "Gets the green component of a color.") },
      { func: "blue($color)", desc: localize11("scss.builtin.blue", "Gets the blue component of a color.") },
      { func: "mix($color, $color, [$weight])", desc: localize11("scss.builtin.mix", "Mixes two colors together.") },
      { func: "hue($color)", desc: localize11("scss.builtin.hue", "Gets the hue component of a color.") },
      { func: "saturation($color)", desc: localize11("scss.builtin.saturation", "Gets the saturation component of a color.") },
      { func: "lightness($color)", desc: localize11("scss.builtin.lightness", "Gets the lightness component of a color.") },
      { func: "adjust-hue($color, $degrees)", desc: localize11("scss.builtin.adjust-hue", "Changes the hue of a color.") },
      { func: "lighten($color, $amount)", desc: localize11("scss.builtin.lighten", "Makes a color lighter.") },
      { func: "darken($color, $amount)", desc: localize11("scss.builtin.darken", "Makes a color darker.") },
      { func: "saturate($color, $amount)", desc: localize11("scss.builtin.saturate", "Makes a color more saturated.") },
      { func: "desaturate($color, $amount)", desc: localize11("scss.builtin.desaturate", "Makes a color less saturated.") },
      { func: "grayscale($color)", desc: localize11("scss.builtin.grayscale", "Converts a color to grayscale.") },
      { func: "complement($color)", desc: localize11("scss.builtin.complement", "Returns the complement of a color.") },
      { func: "invert($color)", desc: localize11("scss.builtin.invert", "Returns the inverse of a color.") },
      { func: "alpha($color)", desc: localize11("scss.builtin.alpha", "Gets the opacity component of a color.") },
      { func: "opacity($color)", desc: "Gets the alpha component (opacity) of a color." },
      { func: "rgba($color, $alpha)", desc: localize11("scss.builtin.rgba", "Changes the alpha component for a color.") },
      { func: "opacify($color, $amount)", desc: localize11("scss.builtin.opacify", "Makes a color more opaque.") },
      { func: "fade-in($color, $amount)", desc: localize11("scss.builtin.fade-in", "Makes a color more opaque.") },
      { func: "transparentize($color, $amount)", desc: localize11("scss.builtin.transparentize", "Makes a color more transparent.") },
      { func: "fade-out($color, $amount)", desc: localize11("scss.builtin.fade-out", "Makes a color more transparent.") },
      { func: "adjust-color($color, [$red], [$green], [$blue], [$hue], [$saturation], [$lightness], [$alpha])", desc: localize11("scss.builtin.adjust-color", "Increases or decreases one or more components of a color.") },
      { func: "scale-color($color, [$red], [$green], [$blue], [$saturation], [$lightness], [$alpha])", desc: localize11("scss.builtin.scale-color", "Fluidly scales one or more properties of a color.") },
      { func: "change-color($color, [$red], [$green], [$blue], [$hue], [$saturation], [$lightness], [$alpha])", desc: localize11("scss.builtin.change-color", "Changes one or more properties of a color.") },
      { func: "ie-hex-str($color)", desc: localize11("scss.builtin.ie-hex-str", "Converts a color into the format understood by IE filters.") }
    ];
    SCSSCompletion2.selectorFuncs = [
      { func: "selector-nest($selectors\u2026)", desc: localize11("scss.builtin.selector-nest", "Nests selector beneath one another like they would be nested in the stylesheet.") },
      { func: "selector-append($selectors\u2026)", desc: localize11("scss.builtin.selector-append", "Appends selectors to one another without spaces in between.") },
      { func: "selector-extend($selector, $extendee, $extender)", desc: localize11("scss.builtin.selector-extend", "Extends $extendee with $extender within $selector.") },
      { func: "selector-replace($selector, $original, $replacement)", desc: localize11("scss.builtin.selector-replace", "Replaces $original with $replacement within $selector.") },
      { func: "selector-unify($selector1, $selector2)", desc: localize11("scss.builtin.selector-unify", "Unifies two selectors to produce a selector that matches elements matched by both.") },
      { func: "is-superselector($super, $sub)", desc: localize11("scss.builtin.is-superselector", "Returns whether $super matches all the elements $sub does, and possibly more.") },
      { func: "simple-selectors($selector)", desc: localize11("scss.builtin.simple-selectors", "Returns the simple selectors that comprise a compound selector.") },
      { func: "selector-parse($selector)", desc: localize11("scss.builtin.selector-parse", "Parses a selector into the format returned by &.") }
    ];
    SCSSCompletion2.builtInFuncs = [
      { func: "unquote($string)", desc: localize11("scss.builtin.unquote", "Removes quotes from a string.") },
      { func: "quote($string)", desc: localize11("scss.builtin.quote", "Adds quotes to a string.") },
      { func: "str-length($string)", desc: localize11("scss.builtin.str-length", "Returns the number of characters in a string.") },
      { func: "str-insert($string, $insert, $index)", desc: localize11("scss.builtin.str-insert", "Inserts $insert into $string at $index.") },
      { func: "str-index($string, $substring)", desc: localize11("scss.builtin.str-index", "Returns the index of the first occurance of $substring in $string.") },
      { func: "str-slice($string, $start-at, [$end-at])", desc: localize11("scss.builtin.str-slice", "Extracts a substring from $string.") },
      { func: "to-upper-case($string)", desc: localize11("scss.builtin.to-upper-case", "Converts a string to upper case.") },
      { func: "to-lower-case($string)", desc: localize11("scss.builtin.to-lower-case", "Converts a string to lower case.") },
      { func: "percentage($number)", desc: localize11("scss.builtin.percentage", "Converts a unitless number to a percentage."), type: "percentage" },
      { func: "round($number)", desc: localize11("scss.builtin.round", "Rounds a number to the nearest whole number.") },
      { func: "ceil($number)", desc: localize11("scss.builtin.ceil", "Rounds a number up to the next whole number.") },
      { func: "floor($number)", desc: localize11("scss.builtin.floor", "Rounds a number down to the previous whole number.") },
      { func: "abs($number)", desc: localize11("scss.builtin.abs", "Returns the absolute value of a number.") },
      { func: "min($numbers)", desc: localize11("scss.builtin.min", "Finds the minimum of several numbers.") },
      { func: "max($numbers)", desc: localize11("scss.builtin.max", "Finds the maximum of several numbers.") },
      { func: "random([$limit])", desc: localize11("scss.builtin.random", "Returns a random number.") },
      { func: "length($list)", desc: localize11("scss.builtin.length", "Returns the length of a list.") },
      { func: "nth($list, $n)", desc: localize11("scss.builtin.nth", "Returns a specific item in a list.") },
      { func: "set-nth($list, $n, $value)", desc: localize11("scss.builtin.set-nth", "Replaces the nth item in a list.") },
      { func: "join($list1, $list2, [$separator])", desc: localize11("scss.builtin.join", "Joins together two lists into one.") },
      { func: "append($list1, $val, [$separator])", desc: localize11("scss.builtin.append", "Appends a single value onto the end of a list.") },
      { func: "zip($lists)", desc: localize11("scss.builtin.zip", "Combines several lists into a single multidimensional list.") },
      { func: "index($list, $value)", desc: localize11("scss.builtin.index", "Returns the position of a value within a list.") },
      { func: "list-separator(#list)", desc: localize11("scss.builtin.list-separator", "Returns the separator of a list.") },
      { func: "map-get($map, $key)", desc: localize11("scss.builtin.map-get", "Returns the value in a map associated with a given key.") },
      { func: "map-merge($map1, $map2)", desc: localize11("scss.builtin.map-merge", "Merges two maps together into a new map.") },
      { func: "map-remove($map, $keys)", desc: localize11("scss.builtin.map-remove", "Returns a new map with keys removed.") },
      { func: "map-keys($map)", desc: localize11("scss.builtin.map-keys", "Returns a list of all keys in a map.") },
      { func: "map-values($map)", desc: localize11("scss.builtin.map-values", "Returns a list of all values in a map.") },
      { func: "map-has-key($map, $key)", desc: localize11("scss.builtin.map-has-key", "Returns whether a map has a value associated with a given key.") },
      { func: "keywords($args)", desc: localize11("scss.builtin.keywords", "Returns the keywords passed to a function that takes variable arguments.") },
      { func: "feature-exists($feature)", desc: localize11("scss.builtin.feature-exists", "Returns whether a feature exists in the current Sass runtime.") },
      { func: "variable-exists($name)", desc: localize11("scss.builtin.variable-exists", "Returns whether a variable with the given name exists in the current scope.") },
      { func: "global-variable-exists($name)", desc: localize11("scss.builtin.global-variable-exists", "Returns whether a variable with the given name exists in the global scope.") },
      { func: "function-exists($name)", desc: localize11("scss.builtin.function-exists", "Returns whether a function with the given name exists.") },
      { func: "mixin-exists($name)", desc: localize11("scss.builtin.mixin-exists", "Returns whether a mixin with the given name exists.") },
      { func: "inspect($value)", desc: localize11("scss.builtin.inspect", "Returns the string representation of a value as it would be represented in Sass.") },
      { func: "type-of($value)", desc: localize11("scss.builtin.type-of", "Returns the type of a value.") },
      { func: "unit($number)", desc: localize11("scss.builtin.unit", "Returns the unit(s) associated with a number.") },
      { func: "unitless($number)", desc: localize11("scss.builtin.unitless", "Returns whether a number has units.") },
      { func: "comparable($number1, $number2)", desc: localize11("scss.builtin.comparable", "Returns whether two numbers can be added, subtracted, or compared.") },
      { func: "call($name, $args\u2026)", desc: localize11("scss.builtin.call", "Dynamically calls a Sass function.") }
    ];
    SCSSCompletion2.scssAtDirectives = [
      {
        label: "@extend",
        documentation: localize11("scss.builtin.@extend", "Inherits the styles of another selector."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@at-root",
        documentation: localize11("scss.builtin.@at-root", "Causes one or more rules to be emitted at the root of the document."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@debug",
        documentation: localize11("scss.builtin.@debug", "Prints the value of an expression to the standard error output stream. Useful for debugging complicated Sass files."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@warn",
        documentation: localize11("scss.builtin.@warn", "Prints the value of an expression to the standard error output stream. Useful for libraries that need to warn users of deprecations or recovering from minor mixin usage mistakes. Warnings can be turned off with the `--quiet` command-line option or the `:quiet` Sass option."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@error",
        documentation: localize11("scss.builtin.@error", "Throws the value of an expression as a fatal error with stack trace. Useful for validating arguments to mixins and functions."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@if",
        documentation: localize11("scss.builtin.@if", "Includes the body if the expression does not evaluate to `false` or `null`."),
        insertText: "@if ${1:expr} {\n	$0\n}",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@for",
        documentation: localize11("scss.builtin.@for", "For loop that repeatedly outputs a set of styles for each `$var` in the `from/through` or `from/to` clause."),
        insertText: "@for \\$${1:var} from ${2:start} ${3|to,through|} ${4:end} {\n	$0\n}",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@each",
        documentation: localize11("scss.builtin.@each", "Each loop that sets `$var` to each item in the list or map, then outputs the styles it contains using that value of `$var`."),
        insertText: "@each \\$${1:var} in ${2:list} {\n	$0\n}",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@while",
        documentation: localize11("scss.builtin.@while", "While loop that takes an expression and repeatedly outputs the nested styles until the statement evaluates to `false`."),
        insertText: "@while ${1:condition} {\n	$0\n}",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@mixin",
        documentation: localize11("scss.builtin.@mixin", "Defines styles that can be re-used throughout the stylesheet with `@include`."),
        insertText: "@mixin ${1:name} {\n	$0\n}",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@include",
        documentation: localize11("scss.builtin.@include", "Includes the styles defined by another mixin into the current rule."),
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@function",
        documentation: localize11("scss.builtin.@function", "Defines complex operations that can be re-used throughout stylesheets."),
        kind: CompletionItemKind2.Keyword
      }
    ];
    SCSSCompletion2.scssModuleLoaders = [
      {
        label: "@use",
        documentation: localize11("scss.builtin.@use", "Loads mixins, functions, and variables from other Sass stylesheets as 'modules', and combines CSS from multiple stylesheets together."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/at-rules/use" }],
        insertText: "@use '$0';",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      },
      {
        label: "@forward",
        documentation: localize11("scss.builtin.@forward", "Loads a Sass stylesheet and makes its mixins, functions, and variables available when this stylesheet is loaded with the @use rule."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/at-rules/forward" }],
        insertText: "@forward '$0';",
        insertTextFormat: InsertTextFormat.Snippet,
        kind: CompletionItemKind2.Keyword
      }
    ];
    SCSSCompletion2.scssModuleBuiltIns = [
      {
        label: "sass:math",
        documentation: localize11("scss.builtin.sass:math", "Provides functions that operate on numbers."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/math" }]
      },
      {
        label: "sass:string",
        documentation: localize11("scss.builtin.sass:string", "Makes it easy to combine, search, or split apart strings."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/string" }]
      },
      {
        label: "sass:color",
        documentation: localize11("scss.builtin.sass:color", "Generates new colors based on existing ones, making it easy to build color themes."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/color" }]
      },
      {
        label: "sass:list",
        documentation: localize11("scss.builtin.sass:list", "Lets you access and modify values in lists."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/list" }]
      },
      {
        label: "sass:map",
        documentation: localize11("scss.builtin.sass:map", "Makes it possible to look up the value associated with a key in a map, and much more."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/map" }]
      },
      {
        label: "sass:selector",
        documentation: localize11("scss.builtin.sass:selector", "Provides access to Sass\u2019s powerful selector engine."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/selector" }]
      },
      {
        label: "sass:meta",
        documentation: localize11("scss.builtin.sass:meta", "Exposes the details of Sass\u2019s inner workings."),
        references: [{ name: "Sass documentation", url: "https://sass-lang.com/documentation/modules/meta" }]
      }
    ];
    return SCSSCompletion2;
  }(CSSCompletion);
  function addReferencesToDocumentation(items) {
    items.forEach(function(i) {
      if (i.documentation && i.references && i.references.length > 0) {
        var markdownDoc = typeof i.documentation === "string" ? { kind: "markdown", value: i.documentation } : { kind: "markdown", value: i.documentation.value };
        markdownDoc.value += "\n\n";
        markdownDoc.value += i.references.map(function(r) {
          return "[" + r.name + "](" + r.url + ")";
        }).join(" | ");
        i.documentation = markdownDoc;
      }
    });
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/lessScanner.js
  var __extends8 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var _FSL3 = "/".charCodeAt(0);
  var _NWL3 = "\n".charCodeAt(0);
  var _CAR3 = "\r".charCodeAt(0);
  var _LFD3 = "\f".charCodeAt(0);
  var _TIC = "`".charCodeAt(0);
  var _DOT3 = ".".charCodeAt(0);
  var customTokenValue2 = TokenType.CustomToken;
  var Ellipsis2 = customTokenValue2++;
  var LESSScanner = function(_super) {
    __extends8(LESSScanner2, _super);
    function LESSScanner2() {
      return _super !== null && _super.apply(this, arguments) || this;
    }
    LESSScanner2.prototype.scanNext = function(offset) {
      var tokenType = this.escapedJavaScript();
      if (tokenType !== null) {
        return this.finishToken(offset, tokenType);
      }
      if (this.stream.advanceIfChars([_DOT3, _DOT3, _DOT3])) {
        return this.finishToken(offset, Ellipsis2);
      }
      return _super.prototype.scanNext.call(this, offset);
    };
    LESSScanner2.prototype.comment = function() {
      if (_super.prototype.comment.call(this)) {
        return true;
      }
      if (!this.inURL && this.stream.advanceIfChars([_FSL3, _FSL3])) {
        this.stream.advanceWhileChar(function(ch) {
          switch (ch) {
            case _NWL3:
            case _CAR3:
            case _LFD3:
              return false;
            default:
              return true;
          }
        });
        return true;
      } else {
        return false;
      }
    };
    LESSScanner2.prototype.escapedJavaScript = function() {
      var ch = this.stream.peekChar();
      if (ch === _TIC) {
        this.stream.advance(1);
        this.stream.advanceWhileChar(function(ch2) {
          return ch2 !== _TIC;
        });
        return this.stream.advanceIfChar(_TIC) ? TokenType.EscapedJavaScript : TokenType.BadEscapedJavaScript;
      }
      return null;
    };
    return LESSScanner2;
  }(Scanner);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/parser/lessParser.js
  var __extends9 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var LESSParser = function(_super) {
    __extends9(LESSParser2, _super);
    function LESSParser2() {
      return _super.call(this, new LESSScanner()) || this;
    }
    LESSParser2.prototype._parseStylesheetStatement = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseVariableDeclaration() || this._parsePlugin() || _super.prototype._parseStylesheetAtStatement.call(this, isNested);
      }
      return this._tryParseMixinDeclaration() || this._tryParseMixinReference() || this._parseFunction() || this._parseRuleset(true);
    };
    LESSParser2.prototype._parseImport = function() {
      if (!this.peekKeyword("@import") && !this.peekKeyword("@import-once")) {
        return null;
      }
      var node = this.create(Import);
      this.consumeToken();
      if (this.accept(TokenType.ParenthesisL)) {
        if (!this.accept(TokenType.Ident)) {
          return this.finish(node, ParseError.IdentifierExpected, [TokenType.SemiColon]);
        }
        do {
          if (!this.accept(TokenType.Comma)) {
            break;
          }
        } while (this.accept(TokenType.Ident));
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected, [TokenType.SemiColon]);
        }
      }
      if (!node.addChild(this._parseURILiteral()) && !node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.URIOrStringExpected, [TokenType.SemiColon]);
      }
      if (!this.peek(TokenType.SemiColon) && !this.peek(TokenType.EOF)) {
        node.setMedialist(this._parseMediaQueryList());
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parsePlugin = function() {
      if (!this.peekKeyword("@plugin")) {
        return null;
      }
      var node = this.createNode(NodeType.Plugin);
      this.consumeToken();
      if (!node.addChild(this._parseStringLiteral())) {
        return this.finish(node, ParseError.StringLiteralExpected);
      }
      if (!this.accept(TokenType.SemiColon)) {
        return this.finish(node, ParseError.SemiColonExpected);
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseMediaQuery = function(resyncStopToken) {
      var node = _super.prototype._parseMediaQuery.call(this, resyncStopToken);
      if (!node) {
        var node_1 = this.create(MediaQuery);
        if (node_1.addChild(this._parseVariable())) {
          return this.finish(node_1);
        }
        return null;
      }
      return node;
    };
    LESSParser2.prototype._parseMediaDeclaration = function(isNested) {
      if (isNested === void 0) {
        isNested = false;
      }
      return this._tryParseRuleset(isNested) || this._tryToParseDeclaration() || this._tryParseMixinDeclaration() || this._tryParseMixinReference() || this._parseDetachedRuleSetMixin() || this._parseStylesheetStatement(isNested);
    };
    LESSParser2.prototype._parseMediaFeatureName = function() {
      return this._parseIdent() || this._parseVariable();
    };
    LESSParser2.prototype._parseVariableDeclaration = function(panic) {
      if (panic === void 0) {
        panic = [];
      }
      var node = this.create(VariableDeclaration);
      var mark = this.mark();
      if (!node.setVariable(this._parseVariable(true))) {
        return null;
      }
      if (this.accept(TokenType.Colon)) {
        if (this.prevToken) {
          node.colonPosition = this.prevToken.offset;
        }
        if (node.setValue(this._parseDetachedRuleSet())) {
          node.needsSemicolon = false;
        } else if (!node.setValue(this._parseExpr())) {
          return this.finish(node, ParseError.VariableValueExpected, [], panic);
        }
        node.addChild(this._parsePrio());
      } else {
        this.restoreAtMark(mark);
        return null;
      }
      if (this.peek(TokenType.SemiColon)) {
        node.semicolonPosition = this.token.offset;
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseDetachedRuleSet = function() {
      var mark = this.mark();
      if (this.peekDelim("#") || this.peekDelim(".")) {
        this.consumeToken();
        if (!this.hasWhitespace() && this.accept(TokenType.ParenthesisL)) {
          var node = this.create(MixinDeclaration);
          if (node.getParameters().addChild(this._parseMixinParameter())) {
            while (this.accept(TokenType.Comma) || this.accept(TokenType.SemiColon)) {
              if (this.peek(TokenType.ParenthesisR)) {
                break;
              }
              if (!node.getParameters().addChild(this._parseMixinParameter())) {
                this.markError(node, ParseError.IdentifierExpected, [], [TokenType.ParenthesisR]);
              }
            }
          }
          if (!this.accept(TokenType.ParenthesisR)) {
            this.restoreAtMark(mark);
            return null;
          }
        } else {
          this.restoreAtMark(mark);
          return null;
        }
      }
      if (!this.peek(TokenType.CurlyL)) {
        return null;
      }
      var content = this.create(BodyDeclaration);
      this._parseBody(content, this._parseDetachedRuleSetBody.bind(this));
      return this.finish(content);
    };
    LESSParser2.prototype._parseDetachedRuleSetBody = function() {
      return this._tryParseKeyframeSelector() || this._parseRuleSetDeclaration();
    };
    LESSParser2.prototype._addLookupChildren = function(node) {
      if (!node.addChild(this._parseLookupValue())) {
        return false;
      }
      var expectsValue = false;
      while (true) {
        if (this.peek(TokenType.BracketL)) {
          expectsValue = true;
        }
        if (!node.addChild(this._parseLookupValue())) {
          break;
        }
        expectsValue = false;
      }
      return !expectsValue;
    };
    LESSParser2.prototype._parseLookupValue = function() {
      var node = this.create(Node2);
      var mark = this.mark();
      if (!this.accept(TokenType.BracketL)) {
        this.restoreAtMark(mark);
        return null;
      }
      if ((node.addChild(this._parseVariable(false, true)) || node.addChild(this._parsePropertyIdentifier())) && this.accept(TokenType.BracketR) || this.accept(TokenType.BracketR)) {
        return node;
      }
      this.restoreAtMark(mark);
      return null;
    };
    LESSParser2.prototype._parseVariable = function(declaration, insideLookup) {
      if (declaration === void 0) {
        declaration = false;
      }
      if (insideLookup === void 0) {
        insideLookup = false;
      }
      var isPropertyReference = !declaration && this.peekDelim("$");
      if (!this.peekDelim("@") && !isPropertyReference && !this.peek(TokenType.AtKeyword)) {
        return null;
      }
      var node = this.create(Variable);
      var mark = this.mark();
      while (this.acceptDelim("@") || !declaration && this.acceptDelim("$")) {
        if (this.hasWhitespace()) {
          this.restoreAtMark(mark);
          return null;
        }
      }
      if (!this.accept(TokenType.AtKeyword) && !this.accept(TokenType.Ident)) {
        this.restoreAtMark(mark);
        return null;
      }
      if (!insideLookup && this.peek(TokenType.BracketL)) {
        if (!this._addLookupChildren(node)) {
          this.restoreAtMark(mark);
          return null;
        }
      }
      return node;
    };
    LESSParser2.prototype._parseTermExpression = function() {
      return this._parseVariable() || this._parseEscaped() || _super.prototype._parseTermExpression.call(this) || this._tryParseMixinReference(false);
    };
    LESSParser2.prototype._parseEscaped = function() {
      if (this.peek(TokenType.EscapedJavaScript) || this.peek(TokenType.BadEscapedJavaScript)) {
        var node = this.createNode(NodeType.EscapedValue);
        this.consumeToken();
        return this.finish(node);
      }
      if (this.peekDelim("~")) {
        var node = this.createNode(NodeType.EscapedValue);
        this.consumeToken();
        if (this.accept(TokenType.String) || this.accept(TokenType.EscapedJavaScript)) {
          return this.finish(node);
        } else {
          return this.finish(node, ParseError.TermExpected);
        }
      }
      return null;
    };
    LESSParser2.prototype._parseOperator = function() {
      var node = this._parseGuardOperator();
      if (node) {
        return node;
      } else {
        return _super.prototype._parseOperator.call(this);
      }
    };
    LESSParser2.prototype._parseGuardOperator = function() {
      if (this.peekDelim(">")) {
        var node = this.createNode(NodeType.Operator);
        this.consumeToken();
        this.acceptDelim("=");
        return node;
      } else if (this.peekDelim("=")) {
        var node = this.createNode(NodeType.Operator);
        this.consumeToken();
        this.acceptDelim("<");
        return node;
      } else if (this.peekDelim("<")) {
        var node = this.createNode(NodeType.Operator);
        this.consumeToken();
        this.acceptDelim("=");
        return node;
      }
      return null;
    };
    LESSParser2.prototype._parseRuleSetDeclaration = function() {
      if (this.peek(TokenType.AtKeyword)) {
        return this._parseKeyframe() || this._parseMedia(true) || this._parseImport() || this._parseSupports(true) || this._parseDetachedRuleSetMixin() || this._parseVariableDeclaration() || _super.prototype._parseRuleSetDeclarationAtStatement.call(this);
      }
      return this._tryParseMixinDeclaration() || this._tryParseRuleset(true) || this._tryParseMixinReference() || this._parseFunction() || this._parseExtend() || _super.prototype._parseRuleSetDeclaration.call(this);
    };
    LESSParser2.prototype._parseKeyframeIdent = function() {
      return this._parseIdent([ReferenceType.Keyframe]) || this._parseVariable();
    };
    LESSParser2.prototype._parseKeyframeSelector = function() {
      return this._parseDetachedRuleSetMixin() || _super.prototype._parseKeyframeSelector.call(this);
    };
    LESSParser2.prototype._parseSimpleSelectorBody = function() {
      return this._parseSelectorCombinator() || _super.prototype._parseSimpleSelectorBody.call(this);
    };
    LESSParser2.prototype._parseSelector = function(isNested) {
      var node = this.create(Selector);
      var hasContent = false;
      if (isNested) {
        hasContent = node.addChild(this._parseCombinator());
      }
      while (node.addChild(this._parseSimpleSelector())) {
        hasContent = true;
        var mark = this.mark();
        if (node.addChild(this._parseGuard()) && this.peek(TokenType.CurlyL)) {
          break;
        }
        this.restoreAtMark(mark);
        node.addChild(this._parseCombinator());
      }
      return hasContent ? this.finish(node) : null;
    };
    LESSParser2.prototype._parseSelectorCombinator = function() {
      if (this.peekDelim("&")) {
        var node = this.createNode(NodeType.SelectorCombinator);
        this.consumeToken();
        while (!this.hasWhitespace() && (this.acceptDelim("-") || this.accept(TokenType.Num) || this.accept(TokenType.Dimension) || node.addChild(this._parseIdent()) || this.acceptDelim("&"))) {
        }
        return this.finish(node);
      }
      return null;
    };
    LESSParser2.prototype._parseSelectorIdent = function() {
      if (!this.peekInterpolatedIdent()) {
        return null;
      }
      var node = this.createNode(NodeType.SelectorInterpolation);
      var hasContent = this._acceptInterpolatedIdent(node);
      return hasContent ? this.finish(node) : null;
    };
    LESSParser2.prototype._parsePropertyIdentifier = function(inLookup) {
      if (inLookup === void 0) {
        inLookup = false;
      }
      var propertyRegex = /^[\w-]+/;
      if (!this.peekInterpolatedIdent() && !this.peekRegExp(this.token.type, propertyRegex)) {
        return null;
      }
      var mark = this.mark();
      var node = this.create(Identifier);
      node.isCustomProperty = this.acceptDelim("-") && this.acceptDelim("-");
      var childAdded = false;
      if (!inLookup) {
        if (node.isCustomProperty) {
          childAdded = this._acceptInterpolatedIdent(node);
        } else {
          childAdded = this._acceptInterpolatedIdent(node, propertyRegex);
        }
      } else {
        if (node.isCustomProperty) {
          childAdded = node.addChild(this._parseIdent());
        } else {
          childAdded = node.addChild(this._parseRegexp(propertyRegex));
        }
      }
      if (!childAdded) {
        this.restoreAtMark(mark);
        return null;
      }
      if (!inLookup && !this.hasWhitespace()) {
        this.acceptDelim("+");
        if (!this.hasWhitespace()) {
          this.acceptIdent("_");
        }
      }
      return this.finish(node);
    };
    LESSParser2.prototype.peekInterpolatedIdent = function() {
      return this.peek(TokenType.Ident) || this.peekDelim("@") || this.peekDelim("$") || this.peekDelim("-");
    };
    LESSParser2.prototype._acceptInterpolatedIdent = function(node, identRegex) {
      var _this = this;
      var hasContent = false;
      var indentInterpolation = function() {
        var pos = _this.mark();
        if (_this.acceptDelim("-")) {
          if (!_this.hasWhitespace()) {
            _this.acceptDelim("-");
          }
          if (_this.hasWhitespace()) {
            _this.restoreAtMark(pos);
            return null;
          }
        }
        return _this._parseInterpolation();
      };
      var accept = identRegex ? function() {
        return _this.acceptRegexp(identRegex);
      } : function() {
        return _this.accept(TokenType.Ident);
      };
      while (accept() || node.addChild(this._parseInterpolation() || this.try(indentInterpolation))) {
        hasContent = true;
        if (this.hasWhitespace()) {
          break;
        }
      }
      return hasContent;
    };
    LESSParser2.prototype._parseInterpolation = function() {
      var mark = this.mark();
      if (this.peekDelim("@") || this.peekDelim("$")) {
        var node = this.createNode(NodeType.Interpolation);
        this.consumeToken();
        if (this.hasWhitespace() || !this.accept(TokenType.CurlyL)) {
          this.restoreAtMark(mark);
          return null;
        }
        if (!node.addChild(this._parseIdent())) {
          return this.finish(node, ParseError.IdentifierExpected);
        }
        if (!this.accept(TokenType.CurlyR)) {
          return this.finish(node, ParseError.RightCurlyExpected);
        }
        return this.finish(node);
      }
      return null;
    };
    LESSParser2.prototype._tryParseMixinDeclaration = function() {
      var mark = this.mark();
      var node = this.create(MixinDeclaration);
      if (!node.setIdentifier(this._parseMixinDeclarationIdentifier()) || !this.accept(TokenType.ParenthesisL)) {
        this.restoreAtMark(mark);
        return null;
      }
      if (node.getParameters().addChild(this._parseMixinParameter())) {
        while (this.accept(TokenType.Comma) || this.accept(TokenType.SemiColon)) {
          if (this.peek(TokenType.ParenthesisR)) {
            break;
          }
          if (!node.getParameters().addChild(this._parseMixinParameter())) {
            this.markError(node, ParseError.IdentifierExpected, [], [TokenType.ParenthesisR]);
          }
        }
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        this.restoreAtMark(mark);
        return null;
      }
      node.setGuard(this._parseGuard());
      if (!this.peek(TokenType.CurlyL)) {
        this.restoreAtMark(mark);
        return null;
      }
      return this._parseBody(node, this._parseMixInBodyDeclaration.bind(this));
    };
    LESSParser2.prototype._parseMixInBodyDeclaration = function() {
      return this._parseFontFace() || this._parseRuleSetDeclaration();
    };
    LESSParser2.prototype._parseMixinDeclarationIdentifier = function() {
      var identifier;
      if (this.peekDelim("#") || this.peekDelim(".")) {
        identifier = this.create(Identifier);
        this.consumeToken();
        if (this.hasWhitespace() || !identifier.addChild(this._parseIdent())) {
          return null;
        }
      } else if (this.peek(TokenType.Hash)) {
        identifier = this.create(Identifier);
        this.consumeToken();
      } else {
        return null;
      }
      identifier.referenceTypes = [ReferenceType.Mixin];
      return this.finish(identifier);
    };
    LESSParser2.prototype._parsePseudo = function() {
      if (!this.peek(TokenType.Colon)) {
        return null;
      }
      var mark = this.mark();
      var node = this.create(ExtendsReference);
      this.consumeToken();
      if (this.acceptIdent("extend")) {
        return this._completeExtends(node);
      }
      this.restoreAtMark(mark);
      return _super.prototype._parsePseudo.call(this);
    };
    LESSParser2.prototype._parseExtend = function() {
      if (!this.peekDelim("&")) {
        return null;
      }
      var mark = this.mark();
      var node = this.create(ExtendsReference);
      this.consumeToken();
      if (this.hasWhitespace() || !this.accept(TokenType.Colon) || !this.acceptIdent("extend")) {
        this.restoreAtMark(mark);
        return null;
      }
      return this._completeExtends(node);
    };
    LESSParser2.prototype._completeExtends = function(node) {
      if (!this.accept(TokenType.ParenthesisL)) {
        return this.finish(node, ParseError.LeftParenthesisExpected);
      }
      var selectors = node.getSelectors();
      if (!selectors.addChild(this._parseSelector(true))) {
        return this.finish(node, ParseError.SelectorExpected);
      }
      while (this.accept(TokenType.Comma)) {
        if (!selectors.addChild(this._parseSelector(true))) {
          return this.finish(node, ParseError.SelectorExpected);
        }
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseDetachedRuleSetMixin = function() {
      if (!this.peek(TokenType.AtKeyword)) {
        return null;
      }
      var mark = this.mark();
      var node = this.create(MixinReference);
      if (node.addChild(this._parseVariable(true)) && (this.hasWhitespace() || !this.accept(TokenType.ParenthesisL))) {
        this.restoreAtMark(mark);
        return null;
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    LESSParser2.prototype._tryParseMixinReference = function(atRoot) {
      if (atRoot === void 0) {
        atRoot = true;
      }
      var mark = this.mark();
      var node = this.create(MixinReference);
      var identifier = this._parseMixinDeclarationIdentifier();
      while (identifier) {
        this.acceptDelim(">");
        var nextId = this._parseMixinDeclarationIdentifier();
        if (nextId) {
          node.getNamespaces().addChild(identifier);
          identifier = nextId;
        } else {
          break;
        }
      }
      if (!node.setIdentifier(identifier)) {
        this.restoreAtMark(mark);
        return null;
      }
      var hasArguments = false;
      if (this.accept(TokenType.ParenthesisL)) {
        hasArguments = true;
        if (node.getArguments().addChild(this._parseMixinArgument())) {
          while (this.accept(TokenType.Comma) || this.accept(TokenType.SemiColon)) {
            if (this.peek(TokenType.ParenthesisR)) {
              break;
            }
            if (!node.getArguments().addChild(this._parseMixinArgument())) {
              return this.finish(node, ParseError.ExpressionExpected);
            }
          }
        }
        if (!this.accept(TokenType.ParenthesisR)) {
          return this.finish(node, ParseError.RightParenthesisExpected);
        }
        identifier.referenceTypes = [ReferenceType.Mixin];
      } else {
        identifier.referenceTypes = [ReferenceType.Mixin, ReferenceType.Rule];
      }
      if (this.peek(TokenType.BracketL)) {
        if (!atRoot) {
          this._addLookupChildren(node);
        }
      } else {
        node.addChild(this._parsePrio());
      }
      if (!hasArguments && !this.peek(TokenType.SemiColon) && !this.peek(TokenType.CurlyR) && !this.peek(TokenType.EOF)) {
        this.restoreAtMark(mark);
        return null;
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseMixinArgument = function() {
      var node = this.create(FunctionArgument);
      var pos = this.mark();
      var argument = this._parseVariable();
      if (argument) {
        if (!this.accept(TokenType.Colon)) {
          this.restoreAtMark(pos);
        } else {
          node.setIdentifier(argument);
        }
      }
      if (node.setValue(this._parseDetachedRuleSet() || this._parseExpr(true))) {
        return this.finish(node);
      }
      this.restoreAtMark(pos);
      return null;
    };
    LESSParser2.prototype._parseMixinParameter = function() {
      var node = this.create(FunctionParameter);
      if (this.peekKeyword("@rest")) {
        var restNode = this.create(Node2);
        this.consumeToken();
        if (!this.accept(Ellipsis2)) {
          return this.finish(node, ParseError.DotExpected, [], [TokenType.Comma, TokenType.ParenthesisR]);
        }
        node.setIdentifier(this.finish(restNode));
        return this.finish(node);
      }
      if (this.peek(Ellipsis2)) {
        var varargsNode = this.create(Node2);
        this.consumeToken();
        node.setIdentifier(this.finish(varargsNode));
        return this.finish(node);
      }
      var hasContent = false;
      if (node.setIdentifier(this._parseVariable())) {
        this.accept(TokenType.Colon);
        hasContent = true;
      }
      if (!node.setDefaultValue(this._parseDetachedRuleSet() || this._parseExpr(true)) && !hasContent) {
        return null;
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseGuard = function() {
      if (!this.peekIdent("when")) {
        return null;
      }
      var node = this.create(LessGuard);
      this.consumeToken();
      node.isNegated = this.acceptIdent("not");
      if (!node.getConditions().addChild(this._parseGuardCondition())) {
        return this.finish(node, ParseError.ConditionExpected);
      }
      while (this.acceptIdent("and") || this.accept(TokenType.Comma)) {
        if (!node.getConditions().addChild(this._parseGuardCondition())) {
          return this.finish(node, ParseError.ConditionExpected);
        }
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseGuardCondition = function() {
      if (!this.peek(TokenType.ParenthesisL)) {
        return null;
      }
      var node = this.create(GuardCondition);
      this.consumeToken();
      if (!node.addChild(this._parseExpr())) {
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseFunction = function() {
      var pos = this.mark();
      var node = this.create(Function);
      if (!node.setIdentifier(this._parseFunctionIdentifier())) {
        return null;
      }
      if (this.hasWhitespace() || !this.accept(TokenType.ParenthesisL)) {
        this.restoreAtMark(pos);
        return null;
      }
      if (node.getArguments().addChild(this._parseMixinArgument())) {
        while (this.accept(TokenType.Comma) || this.accept(TokenType.SemiColon)) {
          if (this.peek(TokenType.ParenthesisR)) {
            break;
          }
          if (!node.getArguments().addChild(this._parseMixinArgument())) {
            return this.finish(node, ParseError.ExpressionExpected);
          }
        }
      }
      if (!this.accept(TokenType.ParenthesisR)) {
        return this.finish(node, ParseError.RightParenthesisExpected);
      }
      return this.finish(node);
    };
    LESSParser2.prototype._parseFunctionIdentifier = function() {
      if (this.peekDelim("%")) {
        var node = this.create(Identifier);
        node.referenceTypes = [ReferenceType.Function];
        this.consumeToken();
        return this.finish(node);
      }
      return _super.prototype._parseFunctionIdentifier.call(this);
    };
    LESSParser2.prototype._parseURLArgument = function() {
      var pos = this.mark();
      var node = _super.prototype._parseURLArgument.call(this);
      if (!node || !this.peek(TokenType.ParenthesisR)) {
        this.restoreAtMark(pos);
        var node_2 = this.create(Node2);
        node_2.addChild(this._parseBinaryExpr());
        return this.finish(node_2);
      }
      return node;
    };
    return LESSParser2;
  }(Parser);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/lessCompletion.js
  var __extends10 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var localize12 = loadMessageBundle();
  var LESSCompletion = function(_super) {
    __extends10(LESSCompletion2, _super);
    function LESSCompletion2(lsOptions, cssDataManager) {
      return _super.call(this, "@", lsOptions, cssDataManager) || this;
    }
    LESSCompletion2.prototype.createFunctionProposals = function(proposals, existingNode, sortToEnd, result) {
      for (var _i = 0, proposals_1 = proposals; _i < proposals_1.length; _i++) {
        var p = proposals_1[_i];
        var item = {
          label: p.name,
          detail: p.example,
          documentation: p.description,
          textEdit: TextEdit.replace(this.getCompletionRange(existingNode), p.name + "($0)"),
          insertTextFormat: InsertTextFormat.Snippet,
          kind: CompletionItemKind2.Function
        };
        if (sortToEnd) {
          item.sortText = "z";
        }
        result.items.push(item);
      }
      return result;
    };
    LESSCompletion2.prototype.getTermProposals = function(entry, existingNode, result) {
      var functions = LESSCompletion2.builtInProposals;
      if (entry) {
        functions = functions.filter(function(f2) {
          return !f2.type || !entry.restrictions || entry.restrictions.indexOf(f2.type) !== -1;
        });
      }
      this.createFunctionProposals(functions, existingNode, true, result);
      return _super.prototype.getTermProposals.call(this, entry, existingNode, result);
    };
    LESSCompletion2.prototype.getColorProposals = function(entry, existingNode, result) {
      this.createFunctionProposals(LESSCompletion2.colorProposals, existingNode, false, result);
      return _super.prototype.getColorProposals.call(this, entry, existingNode, result);
    };
    LESSCompletion2.prototype.getCompletionsForDeclarationProperty = function(declaration, result) {
      this.getCompletionsForSelector(null, true, result);
      return _super.prototype.getCompletionsForDeclarationProperty.call(this, declaration, result);
    };
    LESSCompletion2.builtInProposals = [
      {
        "name": "if",
        "example": "if(condition, trueValue [, falseValue]);",
        "description": localize12("less.builtin.if", "returns one of two values depending on a condition.")
      },
      {
        "name": "boolean",
        "example": "boolean(condition);",
        "description": localize12("less.builtin.boolean", '"store" a boolean test for later evaluation in a guard or if().')
      },
      {
        "name": "length",
        "example": "length(@list);",
        "description": localize12("less.builtin.length", "returns the number of elements in a value list")
      },
      {
        "name": "extract",
        "example": "extract(@list, index);",
        "description": localize12("less.builtin.extract", "returns a value at the specified position in the list")
      },
      {
        "name": "range",
        "example": "range([start, ] end [, step]);",
        "description": localize12("less.builtin.range", "generate a list spanning a range of values")
      },
      {
        "name": "each",
        "example": "each(@list, ruleset);",
        "description": localize12("less.builtin.each", "bind the evaluation of a ruleset to each member of a list.")
      },
      {
        "name": "escape",
        "example": "escape(@string);",
        "description": localize12("less.builtin.escape", "URL encodes a string")
      },
      {
        "name": "e",
        "example": "e(@string);",
        "description": localize12("less.builtin.e", "escape string content")
      },
      {
        "name": "replace",
        "example": "replace(@string, @pattern, @replacement[, @flags]);",
        "description": localize12("less.builtin.replace", "string replace")
      },
      {
        "name": "unit",
        "example": "unit(@dimension, [@unit: '']);",
        "description": localize12("less.builtin.unit", "remove or change the unit of a dimension")
      },
      {
        "name": "color",
        "example": "color(@string);",
        "description": localize12("less.builtin.color", "parses a string to a color"),
        "type": "color"
      },
      {
        "name": "convert",
        "example": "convert(@value, unit);",
        "description": localize12("less.builtin.convert", "converts numbers from one type into another")
      },
      {
        "name": "data-uri",
        "example": "data-uri([mimetype,] url);",
        "description": localize12("less.builtin.data-uri", "inlines a resource and falls back to `url()`"),
        "type": "url"
      },
      {
        "name": "abs",
        "description": localize12("less.builtin.abs", "absolute value of a number"),
        "example": "abs(number);"
      },
      {
        "name": "acos",
        "description": localize12("less.builtin.acos", "arccosine - inverse of cosine function"),
        "example": "acos(number);"
      },
      {
        "name": "asin",
        "description": localize12("less.builtin.asin", "arcsine - inverse of sine function"),
        "example": "asin(number);"
      },
      {
        "name": "ceil",
        "example": "ceil(@number);",
        "description": localize12("less.builtin.ceil", "rounds up to an integer")
      },
      {
        "name": "cos",
        "description": localize12("less.builtin.cos", "cosine function"),
        "example": "cos(number);"
      },
      {
        "name": "floor",
        "description": localize12("less.builtin.floor", "rounds down to an integer"),
        "example": "floor(@number);"
      },
      {
        "name": "percentage",
        "description": localize12("less.builtin.percentage", "converts to a %, e.g. 0.5 > 50%"),
        "example": "percentage(@number);",
        "type": "percentage"
      },
      {
        "name": "round",
        "description": localize12("less.builtin.round", "rounds a number to a number of places"),
        "example": "round(number, [places: 0]);"
      },
      {
        "name": "sqrt",
        "description": localize12("less.builtin.sqrt", "calculates square root of a number"),
        "example": "sqrt(number);"
      },
      {
        "name": "sin",
        "description": localize12("less.builtin.sin", "sine function"),
        "example": "sin(number);"
      },
      {
        "name": "tan",
        "description": localize12("less.builtin.tan", "tangent function"),
        "example": "tan(number);"
      },
      {
        "name": "atan",
        "description": localize12("less.builtin.atan", "arctangent - inverse of tangent function"),
        "example": "atan(number);"
      },
      {
        "name": "pi",
        "description": localize12("less.builtin.pi", "returns pi"),
        "example": "pi();"
      },
      {
        "name": "pow",
        "description": localize12("less.builtin.pow", "first argument raised to the power of the second argument"),
        "example": "pow(@base, @exponent);"
      },
      {
        "name": "mod",
        "description": localize12("less.builtin.mod", "first argument modulus second argument"),
        "example": "mod(number, number);"
      },
      {
        "name": "min",
        "description": localize12("less.builtin.min", "returns the lowest of one or more values"),
        "example": "min(@x, @y);"
      },
      {
        "name": "max",
        "description": localize12("less.builtin.max", "returns the lowest of one or more values"),
        "example": "max(@x, @y);"
      }
    ];
    LESSCompletion2.colorProposals = [
      {
        "name": "argb",
        "example": "argb(@color);",
        "description": localize12("less.builtin.argb", "creates a #AARRGGBB")
      },
      {
        "name": "hsl",
        "example": "hsl(@hue, @saturation, @lightness);",
        "description": localize12("less.builtin.hsl", "creates a color")
      },
      {
        "name": "hsla",
        "example": "hsla(@hue, @saturation, @lightness, @alpha);",
        "description": localize12("less.builtin.hsla", "creates a color")
      },
      {
        "name": "hsv",
        "example": "hsv(@hue, @saturation, @value);",
        "description": localize12("less.builtin.hsv", "creates a color")
      },
      {
        "name": "hsva",
        "example": "hsva(@hue, @saturation, @value, @alpha);",
        "description": localize12("less.builtin.hsva", "creates a color")
      },
      {
        "name": "hue",
        "example": "hue(@color);",
        "description": localize12("less.builtin.hue", "returns the `hue` channel of `@color` in the HSL space")
      },
      {
        "name": "saturation",
        "example": "saturation(@color);",
        "description": localize12("less.builtin.saturation", "returns the `saturation` channel of `@color` in the HSL space")
      },
      {
        "name": "lightness",
        "example": "lightness(@color);",
        "description": localize12("less.builtin.lightness", "returns the `lightness` channel of `@color` in the HSL space")
      },
      {
        "name": "hsvhue",
        "example": "hsvhue(@color);",
        "description": localize12("less.builtin.hsvhue", "returns the `hue` channel of `@color` in the HSV space")
      },
      {
        "name": "hsvsaturation",
        "example": "hsvsaturation(@color);",
        "description": localize12("less.builtin.hsvsaturation", "returns the `saturation` channel of `@color` in the HSV space")
      },
      {
        "name": "hsvvalue",
        "example": "hsvvalue(@color);",
        "description": localize12("less.builtin.hsvvalue", "returns the `value` channel of `@color` in the HSV space")
      },
      {
        "name": "red",
        "example": "red(@color);",
        "description": localize12("less.builtin.red", "returns the `red` channel of `@color`")
      },
      {
        "name": "green",
        "example": "green(@color);",
        "description": localize12("less.builtin.green", "returns the `green` channel of `@color`")
      },
      {
        "name": "blue",
        "example": "blue(@color);",
        "description": localize12("less.builtin.blue", "returns the `blue` channel of `@color`")
      },
      {
        "name": "alpha",
        "example": "alpha(@color);",
        "description": localize12("less.builtin.alpha", "returns the `alpha` channel of `@color`")
      },
      {
        "name": "luma",
        "example": "luma(@color);",
        "description": localize12("less.builtin.luma", "returns the `luma` value (perceptual brightness) of `@color`")
      },
      {
        "name": "saturate",
        "example": "saturate(@color, 10%);",
        "description": localize12("less.builtin.saturate", "return `@color` 10% points more saturated")
      },
      {
        "name": "desaturate",
        "example": "desaturate(@color, 10%);",
        "description": localize12("less.builtin.desaturate", "return `@color` 10% points less saturated")
      },
      {
        "name": "lighten",
        "example": "lighten(@color, 10%);",
        "description": localize12("less.builtin.lighten", "return `@color` 10% points lighter")
      },
      {
        "name": "darken",
        "example": "darken(@color, 10%);",
        "description": localize12("less.builtin.darken", "return `@color` 10% points darker")
      },
      {
        "name": "fadein",
        "example": "fadein(@color, 10%);",
        "description": localize12("less.builtin.fadein", "return `@color` 10% points less transparent")
      },
      {
        "name": "fadeout",
        "example": "fadeout(@color, 10%);",
        "description": localize12("less.builtin.fadeout", "return `@color` 10% points more transparent")
      },
      {
        "name": "fade",
        "example": "fade(@color, 50%);",
        "description": localize12("less.builtin.fade", "return `@color` with 50% transparency")
      },
      {
        "name": "spin",
        "example": "spin(@color, 10);",
        "description": localize12("less.builtin.spin", "return `@color` with a 10 degree larger in hue")
      },
      {
        "name": "mix",
        "example": "mix(@color1, @color2, [@weight: 50%]);",
        "description": localize12("less.builtin.mix", "return a mix of `@color1` and `@color2`")
      },
      {
        "name": "greyscale",
        "example": "greyscale(@color);",
        "description": localize12("less.builtin.greyscale", "returns a grey, 100% desaturated color")
      },
      {
        "name": "contrast",
        "example": "contrast(@color1, [@darkcolor: black], [@lightcolor: white], [@threshold: 43%]);",
        "description": localize12("less.builtin.contrast", "return `@darkcolor` if `@color1 is> 43% luma` otherwise return `@lightcolor`, see notes")
      },
      {
        "name": "multiply",
        "example": "multiply(@color1, @color2);"
      },
      {
        "name": "screen",
        "example": "screen(@color1, @color2);"
      },
      {
        "name": "overlay",
        "example": "overlay(@color1, @color2);"
      },
      {
        "name": "softlight",
        "example": "softlight(@color1, @color2);"
      },
      {
        "name": "hardlight",
        "example": "hardlight(@color1, @color2);"
      },
      {
        "name": "difference",
        "example": "difference(@color1, @color2);"
      },
      {
        "name": "exclusion",
        "example": "exclusion(@color1, @color2);"
      },
      {
        "name": "average",
        "example": "average(@color1, @color2);"
      },
      {
        "name": "negation",
        "example": "negation(@color1, @color2);"
      }
    ];
    return LESSCompletion2;
  }(CSSCompletion);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssFolding.js
  function getFoldingRanges(document, context) {
    var ranges = computeFoldingRanges(document);
    return limitFoldingRanges(ranges, context);
  }
  function computeFoldingRanges(document) {
    function getStartLine(t) {
      return document.positionAt(t.offset).line;
    }
    function getEndLine(t) {
      return document.positionAt(t.offset + t.len).line;
    }
    function getScanner() {
      switch (document.languageId) {
        case "scss":
          return new SCSSScanner();
        case "less":
          return new LESSScanner();
        default:
          return new Scanner();
      }
    }
    function tokenToRange(t, kind) {
      var startLine = getStartLine(t);
      var endLine = getEndLine(t);
      if (startLine !== endLine) {
        return {
          startLine,
          endLine,
          kind
        };
      } else {
        return null;
      }
    }
    var ranges = [];
    var delimiterStack = [];
    var scanner = getScanner();
    scanner.ignoreComment = false;
    scanner.setSource(document.getText());
    var token = scanner.scan();
    var prevToken = null;
    var _loop_1 = function() {
      switch (token.type) {
        case TokenType.CurlyL:
        case InterpolationFunction: {
          delimiterStack.push({ line: getStartLine(token), type: "brace", isStart: true });
          break;
        }
        case TokenType.CurlyR: {
          if (delimiterStack.length !== 0) {
            var prevDelimiter = popPrevStartDelimiterOfType(delimiterStack, "brace");
            if (!prevDelimiter) {
              break;
            }
            var endLine = getEndLine(token);
            if (prevDelimiter.type === "brace") {
              if (prevToken && getEndLine(prevToken) !== endLine) {
                endLine--;
              }
              if (prevDelimiter.line !== endLine) {
                ranges.push({
                  startLine: prevDelimiter.line,
                  endLine,
                  kind: void 0
                });
              }
            }
          }
          break;
        }
        case TokenType.Comment: {
          var commentRegionMarkerToDelimiter_1 = function(marker) {
            if (marker === "#region") {
              return { line: getStartLine(token), type: "comment", isStart: true };
            } else {
              return { line: getEndLine(token), type: "comment", isStart: false };
            }
          };
          var getCurrDelimiter = function(token2) {
            var matches2 = token2.text.match(/^\s*\/\*\s*(#region|#endregion)\b\s*(.*?)\s*\*\//);
            if (matches2) {
              return commentRegionMarkerToDelimiter_1(matches2[1]);
            } else if (document.languageId === "scss" || document.languageId === "less") {
              var matches_1 = token2.text.match(/^\s*\/\/\s*(#region|#endregion)\b\s*(.*?)\s*/);
              if (matches_1) {
                return commentRegionMarkerToDelimiter_1(matches_1[1]);
              }
            }
            return null;
          };
          var currDelimiter = getCurrDelimiter(token);
          if (currDelimiter) {
            if (currDelimiter.isStart) {
              delimiterStack.push(currDelimiter);
            } else {
              var prevDelimiter = popPrevStartDelimiterOfType(delimiterStack, "comment");
              if (!prevDelimiter) {
                break;
              }
              if (prevDelimiter.type === "comment") {
                if (prevDelimiter.line !== currDelimiter.line) {
                  ranges.push({
                    startLine: prevDelimiter.line,
                    endLine: currDelimiter.line,
                    kind: "region"
                  });
                }
              }
            }
          } else {
            var range = tokenToRange(token, "comment");
            if (range) {
              ranges.push(range);
            }
          }
          break;
        }
      }
      prevToken = token;
      token = scanner.scan();
    };
    while (token.type !== TokenType.EOF) {
      _loop_1();
    }
    return ranges;
  }
  function popPrevStartDelimiterOfType(stack, type) {
    if (stack.length === 0) {
      return null;
    }
    for (var i = stack.length - 1; i >= 0; i--) {
      if (stack[i].type === type && stack[i].isStart) {
        return stack.splice(i, 1)[0];
      }
    }
    return null;
  }
  function limitFoldingRanges(ranges, context) {
    var maxRanges = context && context.rangeLimit || Number.MAX_VALUE;
    var sortedRanges = ranges.sort(function(r1, r2) {
      var diff = r1.startLine - r2.startLine;
      if (diff === 0) {
        diff = r1.endLine - r2.endLine;
      }
      return diff;
    });
    var validRanges = [];
    var prevEndLine = -1;
    sortedRanges.forEach(function(r) {
      if (!(r.startLine < prevEndLine && prevEndLine < r.endLine)) {
        validRanges.push(r);
        prevEndLine = r.endLine;
      }
    });
    if (validRanges.length < maxRanges) {
      return validRanges;
    } else {
      return validRanges.slice(0, maxRanges);
    }
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/data/webCustomData.js
  var cssData = {
    "version": 1.1,
    "properties": [
      {
        "name": "additive-symbols",
        "browsers": [
          "FF33"
        ],
        "syntax": "[ <integer> && <symbol> ]#",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies the symbols used by the marker-construction algorithm specified by the system descriptor. Needs to be specified if the counter system is 'additive'.",
        "restrictions": [
          "integer",
          "string",
          "image",
          "identifier"
        ]
      },
      {
        "name": "align-content",
        "values": [
          {
            "name": "center",
            "description": "Lines are packed toward the center of the flex container."
          },
          {
            "name": "flex-end",
            "description": "Lines are packed toward the end of the flex container."
          },
          {
            "name": "flex-start",
            "description": "Lines are packed toward the start of the flex container."
          },
          {
            "name": "space-around",
            "description": "Lines are evenly distributed in the flex container, with half-size spaces on either end."
          },
          {
            "name": "space-between",
            "description": "Lines are evenly distributed in the flex container."
          },
          {
            "name": "stretch",
            "description": "Lines stretch to take up the remaining space."
          }
        ],
        "syntax": "normal | <baseline-position> | <content-distribution> | <overflow-position>? <content-position>",
        "relevance": 59,
        "description": "Aligns a flex container\u2019s lines within the flex container when there is extra space in the cross-axis, similar to how 'justify-content' aligns individual items within the main-axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "align-items",
        "values": [
          {
            "name": "baseline",
            "description": "If the flex item\u2019s inline axis is the same as the cross axis, this value is identical to 'flex-start'. Otherwise, it participates in baseline alignment."
          },
          {
            "name": "center",
            "description": "The flex item\u2019s margin box is centered in the cross axis within the line."
          },
          {
            "name": "flex-end",
            "description": "The cross-end margin edge of the flex item is placed flush with the cross-end edge of the line."
          },
          {
            "name": "flex-start",
            "description": "The cross-start margin edge of the flex item is placed flush with the cross-start edge of the line."
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flex item computes to auto, and neither of the cross-axis margins are auto, the flex item is stretched."
          }
        ],
        "syntax": "normal | stretch | <baseline-position> | [ <overflow-position>? <self-position> ]",
        "relevance": 81,
        "description": "Aligns flex items along the cross axis of the current line of the flex container.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "justify-items",
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "normal"
          },
          {
            "name": "end"
          },
          {
            "name": "start"
          },
          {
            "name": "flex-end",
            "description": '"Flex items are packed toward the end of the line."'
          },
          {
            "name": "flex-start",
            "description": '"Flex items are packed toward the start of the line."'
          },
          {
            "name": "self-end",
            "description": "The item is packed flush to the edge of the alignment container of the end side of the item, in the appropriate axis."
          },
          {
            "name": "self-start",
            "description": "The item is packed flush to the edge of the alignment container of the start side of the item, in the appropriate axis.."
          },
          {
            "name": "center",
            "description": "The items are packed flush to each other toward the center of the of the alignment container."
          },
          {
            "name": "left"
          },
          {
            "name": "right"
          },
          {
            "name": "baseline"
          },
          {
            "name": "first baseline"
          },
          {
            "name": "last baseline"
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flex item computes to auto, and neither of the cross-axis margins are auto, the flex item is stretched."
          },
          {
            "name": "save"
          },
          {
            "name": "unsave"
          },
          {
            "name": "legacy"
          }
        ],
        "syntax": "normal | stretch | <baseline-position> | <overflow-position>? [ <self-position> | left | right ] | legacy | legacy && [ left | right | center ]",
        "relevance": 50,
        "description": "Defines the default justify-self for all items of the box, giving them the default way of justifying each box along the appropriate axis",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "justify-self",
        "browsers": [
          "E16",
          "FF45",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "normal"
          },
          {
            "name": "end"
          },
          {
            "name": "start"
          },
          {
            "name": "flex-end",
            "description": '"Flex items are packed toward the end of the line."'
          },
          {
            "name": "flex-start",
            "description": '"Flex items are packed toward the start of the line."'
          },
          {
            "name": "self-end",
            "description": "The item is packed flush to the edge of the alignment container of the end side of the item, in the appropriate axis."
          },
          {
            "name": "self-start",
            "description": "The item is packed flush to the edge of the alignment container of the start side of the item, in the appropriate axis.."
          },
          {
            "name": "center",
            "description": "The items are packed flush to each other toward the center of the of the alignment container."
          },
          {
            "name": "left"
          },
          {
            "name": "right"
          },
          {
            "name": "baseline"
          },
          {
            "name": "first baseline"
          },
          {
            "name": "last baseline"
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flex item computes to auto, and neither of the cross-axis margins are auto, the flex item is stretched."
          },
          {
            "name": "save"
          },
          {
            "name": "unsave"
          }
        ],
        "syntax": "auto | normal | stretch | <baseline-position> | <overflow-position>? [ <self-position> | left | right ]",
        "relevance": 52,
        "description": "Defines the way of justifying a box inside its container along the appropriate axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "align-self",
        "values": [
          {
            "name": "auto",
            "description": "Computes to the value of 'align-items' on the element\u2019s parent, or 'stretch' if the element has no parent. On absolutely positioned elements, it computes to itself."
          },
          {
            "name": "baseline",
            "description": "If the flex item\u2019s inline axis is the same as the cross axis, this value is identical to 'flex-start'. Otherwise, it participates in baseline alignment."
          },
          {
            "name": "center",
            "description": "The flex item\u2019s margin box is centered in the cross axis within the line."
          },
          {
            "name": "flex-end",
            "description": "The cross-end margin edge of the flex item is placed flush with the cross-end edge of the line."
          },
          {
            "name": "flex-start",
            "description": "The cross-start margin edge of the flex item is placed flush with the cross-start edge of the line."
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flex item computes to auto, and neither of the cross-axis margins are auto, the flex item is stretched."
          }
        ],
        "syntax": "auto | normal | stretch | <baseline-position> | <overflow-position>? <self-position>",
        "relevance": 69,
        "description": "Allows the default alignment along the cross axis to be overridden for individual flex items.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "all",
        "browsers": [
          "E79",
          "FF27",
          "S9.1",
          "C37",
          "O24"
        ],
        "values": [],
        "syntax": "initial | inherit | unset | revert",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/all"
          }
        ],
        "description": "Shorthand that resets all properties except 'direction' and 'unicode-bidi'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "alt",
        "browsers": [
          "S9"
        ],
        "values": [],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/alt"
          }
        ],
        "description": "Provides alternative text for assistive technology to replace the generated content of a ::before or ::after element.",
        "restrictions": [
          "string",
          "enum"
        ]
      },
      {
        "name": "animation",
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          },
          {
            "name": "none",
            "description": "No animation is performed"
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "syntax": "<single-animation>#",
        "relevance": 79,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation"
          }
        ],
        "description": "Shorthand property combines six of the animation properties into a single property.",
        "restrictions": [
          "time",
          "timing-function",
          "enum",
          "identifier",
          "number"
        ]
      },
      {
        "name": "animation-delay",
        "syntax": "<time>#",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-delay"
          }
        ],
        "description": "Defines when the animation will start.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "animation-direction",
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "syntax": "<single-animation-direction>#",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-direction"
          }
        ],
        "description": "Defines whether or not the animation should play in reverse on alternate cycles.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "animation-duration",
        "syntax": "<time>#",
        "relevance": 64,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-duration"
          }
        ],
        "description": "Defines the length of time that an animation takes to complete one cycle.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "animation-fill-mode",
        "values": [
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "none",
            "description": "There is no change to the property value between the time the animation is applied and the time the animation begins playing or after the animation completes."
          }
        ],
        "syntax": "<single-animation-fill-mode>#",
        "relevance": 61,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-fill-mode"
          }
        ],
        "description": "Defines what values are applied by the animation outside the time it is executing.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "animation-iteration-count",
        "values": [
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          }
        ],
        "syntax": "<single-animation-iteration-count>#",
        "relevance": 59,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-iteration-count"
          }
        ],
        "description": "Defines the number of times an animation cycle is played. The default value is one, meaning the animation will play from beginning to end once.",
        "restrictions": [
          "number",
          "enum"
        ]
      },
      {
        "name": "animation-name",
        "values": [
          {
            "name": "none",
            "description": "No animation is performed"
          }
        ],
        "syntax": "[ none | <keyframes-name> ]#",
        "relevance": 64,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-name"
          }
        ],
        "description": "Defines a list of animations that apply. Each name is used to select the keyframe at-rule that provides the property values for the animation.",
        "restrictions": [
          "identifier",
          "enum"
        ]
      },
      {
        "name": "animation-play-state",
        "values": [
          {
            "name": "paused",
            "description": "A running animation will be paused."
          },
          {
            "name": "running",
            "description": "Resume playback of a paused animation."
          }
        ],
        "syntax": "<single-animation-play-state>#",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-play-state"
          }
        ],
        "description": "Defines whether the animation is running or paused.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "animation-timing-function",
        "syntax": "<timing-function>#",
        "relevance": 68,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/animation-timing-function"
          }
        ],
        "description": "Describes how the animation will progress over one cycle of its duration.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "backface-visibility",
        "values": [
          {
            "name": "hidden",
            "description": "Back side is hidden."
          },
          {
            "name": "visible",
            "description": "Back side is visible."
          }
        ],
        "syntax": "visible | hidden",
        "relevance": 59,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/backface-visibility"
          }
        ],
        "description": "Determines whether or not the 'back' side of a transformed element is visible when facing the viewer. With an identity transform, the front side of an element faces the viewer.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "background",
        "values": [
          {
            "name": "fixed",
            "description": "The background is fixed with regard to the viewport. In paged media where there is no viewport, a 'fixed' background is fixed with respect to the page box and therefore replicated on every page."
          },
          {
            "name": "local",
            "description": "The background is fixed with regard to the element's contents: if the element has a scrolling mechanism, the background scrolls with the element's contents."
          },
          {
            "name": "none",
            "description": "A value of 'none' counts as an image layer but draws nothing."
          },
          {
            "name": "scroll",
            "description": "The background is fixed with regard to the element itself and does not scroll with its contents. (It is effectively attached to the element's border.)"
          }
        ],
        "syntax": "[ <bg-layer> , ]* <final-bg-layer>",
        "relevance": 93,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background"
          }
        ],
        "description": "Shorthand property for setting most background properties at the same place in the style sheet.",
        "restrictions": [
          "enum",
          "image",
          "color",
          "position",
          "length",
          "repeat",
          "percentage",
          "box"
        ]
      },
      {
        "name": "background-attachment",
        "values": [
          {
            "name": "fixed",
            "description": "The background is fixed with regard to the viewport. In paged media where there is no viewport, a 'fixed' background is fixed with respect to the page box and therefore replicated on every page."
          },
          {
            "name": "local",
            "description": "The background is fixed with regard to the element\u2019s contents: if the element has a scrolling mechanism, the background scrolls with the element\u2019s contents."
          },
          {
            "name": "scroll",
            "description": "The background is fixed with regard to the element itself and does not scroll with its contents. (It is effectively attached to the element\u2019s border.)"
          }
        ],
        "syntax": "<attachment>#",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-attachment"
          }
        ],
        "description": "Specifies whether the background images are fixed with regard to the viewport ('fixed') or scroll along with the element ('scroll') or its contents ('local').",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "background-blend-mode",
        "browsers": [
          "E79",
          "FF30",
          "S8",
          "C35",
          "O22"
        ],
        "values": [
          {
            "name": "normal",
            "description": "Default attribute which specifies no blending"
          },
          {
            "name": "multiply",
            "description": "The source color is multiplied by the destination color and replaces the destination."
          },
          {
            "name": "screen",
            "description": "Multiplies the complements of the backdrop and source color values, then complements the result."
          },
          {
            "name": "overlay",
            "description": "Multiplies or screens the colors, depending on the backdrop color value."
          },
          {
            "name": "darken",
            "description": "Selects the darker of the backdrop and source colors."
          },
          {
            "name": "lighten",
            "description": "Selects the lighter of the backdrop and source colors."
          },
          {
            "name": "color-dodge",
            "description": "Brightens the backdrop color to reflect the source color."
          },
          {
            "name": "color-burn",
            "description": "Darkens the backdrop color to reflect the source color."
          },
          {
            "name": "hard-light",
            "description": "Multiplies or screens the colors, depending on the source color value."
          },
          {
            "name": "soft-light",
            "description": "Darkens or lightens the colors, depending on the source color value."
          },
          {
            "name": "difference",
            "description": "Subtracts the darker of the two constituent colors from the lighter color.."
          },
          {
            "name": "exclusion",
            "description": "Produces an effect similar to that of the Difference mode but lower in contrast."
          },
          {
            "name": "hue",
            "browsers": [
              "E79",
              "FF30",
              "S8",
              "C35",
              "O22"
            ],
            "description": "Creates a color with the hue of the source color and the saturation and luminosity of the backdrop color."
          },
          {
            "name": "saturation",
            "browsers": [
              "E79",
              "FF30",
              "S8",
              "C35",
              "O22"
            ],
            "description": "Creates a color with the saturation of the source color and the hue and luminosity of the backdrop color."
          },
          {
            "name": "color",
            "browsers": [
              "E79",
              "FF30",
              "S8",
              "C35",
              "O22"
            ],
            "description": "Creates a color with the hue and saturation of the source color and the luminosity of the backdrop color."
          },
          {
            "name": "luminosity",
            "browsers": [
              "E79",
              "FF30",
              "S8",
              "C35",
              "O22"
            ],
            "description": "Creates a color with the luminosity of the source color and the hue and saturation of the backdrop color."
          }
        ],
        "syntax": "<blend-mode>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-blend-mode"
          }
        ],
        "description": "Defines the blending mode of each background layer.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "background-clip",
        "syntax": "<box>#",
        "relevance": 67,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-clip"
          }
        ],
        "description": "Determines the background painting area.",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "background-color",
        "syntax": "<color>",
        "relevance": 94,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-color"
          }
        ],
        "description": "Sets the background color of an element.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "background-image",
        "values": [
          {
            "name": "none",
            "description": "Counts as an image layer but draws nothing."
          }
        ],
        "syntax": "<bg-image>#",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-image"
          }
        ],
        "description": "Sets the background image(s) of an element.",
        "restrictions": [
          "image",
          "enum"
        ]
      },
      {
        "name": "background-origin",
        "syntax": "<box>#",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-origin"
          }
        ],
        "description": "For elements rendered as a single box, specifies the background positioning area. For elements rendered as multiple boxes (e.g., inline boxes on several lines, boxes on several pages) specifies which boxes 'box-decoration-break' operates on to determine the background positioning area(s).",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "background-position",
        "syntax": "<bg-position>#",
        "relevance": 87,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-position"
          }
        ],
        "description": "Specifies the initial position of the background image(s) (after any resizing) within their corresponding background positioning area.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "background-position-x",
        "values": [
          {
            "name": "center",
            "description": "Equivalent to '50%' ('left 50%') for the horizontal position if the horizontal position is not otherwise specified, or '50%' ('top 50%') for the vertical position if it is."
          },
          {
            "name": "left",
            "description": "Equivalent to '0%' for the horizontal position if one or two values are given, otherwise specifies the left edge as the origin for the next offset."
          },
          {
            "name": "right",
            "description": "Equivalent to '100%' for the horizontal position if one or two values are given, otherwise specifies the right edge as the origin for the next offset."
          }
        ],
        "status": "experimental",
        "syntax": "[ center | [ [ left | right | x-start | x-end ]? <length-percentage>? ]! ]#",
        "relevance": 54,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-position-x"
          }
        ],
        "description": "If background images have been specified, this property specifies their initial position (after any resizing) within their corresponding background positioning area.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "background-position-y",
        "values": [
          {
            "name": "bottom",
            "description": "Equivalent to '100%' for the vertical position if one or two values are given, otherwise specifies the bottom edge as the origin for the next offset."
          },
          {
            "name": "center",
            "description": "Equivalent to '50%' ('left 50%') for the horizontal position if the horizontal position is not otherwise specified, or '50%' ('top 50%') for the vertical position if it is."
          },
          {
            "name": "top",
            "description": "Equivalent to '0%' for the vertical position if one or two values are given, otherwise specifies the top edge as the origin for the next offset."
          }
        ],
        "status": "experimental",
        "syntax": "[ center | [ [ top | bottom | y-start | y-end ]? <length-percentage>? ]! ]#",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-position-y"
          }
        ],
        "description": "If background images have been specified, this property specifies their initial position (after any resizing) within their corresponding background positioning area.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "background-repeat",
        "values": [],
        "syntax": "<repeat-style>#",
        "relevance": 85,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-repeat"
          }
        ],
        "description": "Specifies how background images are tiled after they have been sized and positioned.",
        "restrictions": [
          "repeat"
        ]
      },
      {
        "name": "background-size",
        "values": [
          {
            "name": "auto",
            "description": "Resolved by using the image\u2019s intrinsic ratio and the size of the other dimension, or failing that, using the image\u2019s intrinsic size, or failing that, treating it as 100%."
          },
          {
            "name": "contain",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the largest size such that both its width and its height can fit inside the background positioning area."
          },
          {
            "name": "cover",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the smallest size such that both its width and its height can completely cover the background positioning area."
          }
        ],
        "syntax": "<bg-size>#",
        "relevance": 85,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/background-size"
          }
        ],
        "description": "Specifies the size of the background images.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "behavior",
        "browsers": [
          "IE6"
        ],
        "relevance": 50,
        "description": "IE only. Used to extend behaviors of the browser.",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "block-size",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Depends on the values of other properties."
          }
        ],
        "syntax": "<'width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/block-size"
          }
        ],
        "description": "Logical 'width'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border",
        "syntax": "<line-width> || <line-style> || <color>",
        "relevance": 95,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border"
          }
        ],
        "description": "Shorthand property for setting border width, style, and color.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-block-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-end"
          }
        ],
        "description": "Logical 'border-bottom'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-block-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-start"
          }
        ],
        "description": "Logical 'border-top'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-block-end-color",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-end-color"
          }
        ],
        "description": "Logical 'border-bottom-color'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-block-start-color",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-start-color"
          }
        ],
        "description": "Logical 'border-top-color'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-block-end-style",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-end-style"
          }
        ],
        "description": "Logical 'border-bottom-style'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-block-start-style",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-start-style"
          }
        ],
        "description": "Logical 'border-top-style'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-block-end-width",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-end-width"
          }
        ],
        "description": "Logical 'border-bottom-width'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-block-start-width",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-start-width"
          }
        ],
        "description": "Logical 'border-top-width'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-bottom",
        "syntax": "<line-width> || <line-style> || <color>",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom"
          }
        ],
        "description": "Shorthand property for setting border width, style and color.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-bottom-color",
        "syntax": "<'border-top-color'>",
        "relevance": 71,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom-color"
          }
        ],
        "description": "Sets the color of the bottom border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-bottom-left-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom-left-radius"
          }
        ],
        "description": "Defines the radii of the bottom left outer border edge.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border-bottom-right-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 73,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom-right-radius"
          }
        ],
        "description": "Defines the radii of the bottom right outer border edge.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border-bottom-style",
        "syntax": "<line-style>",
        "relevance": 57,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom-style"
          }
        ],
        "description": "Sets the style of the bottom border.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-bottom-width",
        "syntax": "<line-width>",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-bottom-width"
          }
        ],
        "description": "Sets the thickness of the bottom border.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-collapse",
        "values": [
          {
            "name": "collapse",
            "description": "Selects the collapsing borders model."
          },
          {
            "name": "separate",
            "description": "Selects the separated borders border model."
          }
        ],
        "syntax": "collapse | separate",
        "relevance": 75,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-collapse"
          }
        ],
        "description": "Selects a table's border model.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "border-color",
        "values": [],
        "syntax": "<color>{1,4}",
        "relevance": 86,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-color"
          }
        ],
        "description": "The color of the border around all four edges of an element.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-image",
        "values": [
          {
            "name": "auto",
            "description": "If 'auto' is specified then the border image width is the intrinsic width or height (whichever is applicable) of the corresponding image slice. If the image does not have the required intrinsic dimension then the corresponding border-width is used instead."
          },
          {
            "name": "fill",
            "description": "Causes the middle part of the border-image to be preserved."
          },
          {
            "name": "none",
            "description": "Use the border styles."
          },
          {
            "name": "repeat",
            "description": "The image is tiled (repeated) to fill the area."
          },
          {
            "name": "round",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the image is rescaled so that it does."
          },
          {
            "name": "space",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the extra space is distributed around the tiles."
          },
          {
            "name": "stretch",
            "description": "The image is stretched to fill the area."
          },
          {
            "name": "url()"
          }
        ],
        "syntax": "<'border-image-source'> || <'border-image-slice'> [ / <'border-image-width'> | / <'border-image-width'>? / <'border-image-outset'> ]? || <'border-image-repeat'>",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image"
          }
        ],
        "description": "Shorthand property for setting 'border-image-source', 'border-image-slice', 'border-image-width', 'border-image-outset' and 'border-image-repeat'. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "percentage",
          "number",
          "url",
          "enum"
        ]
      },
      {
        "name": "border-image-outset",
        "syntax": "[ <length> | <number> ]{1,4}",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image-outset"
          }
        ],
        "description": "The values specify the amount by which the border image area extends beyond the border box on the top, right, bottom, and left sides respectively. If the fourth value is absent, it is the same as the second. If the third one is also absent, it is the same as the first. If the second one is also absent, it is the same as the first. Numbers represent multiples of the corresponding border-width.",
        "restrictions": [
          "length",
          "number"
        ]
      },
      {
        "name": "border-image-repeat",
        "values": [
          {
            "name": "repeat",
            "description": "The image is tiled (repeated) to fill the area."
          },
          {
            "name": "round",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the image is rescaled so that it does."
          },
          {
            "name": "space",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the extra space is distributed around the tiles."
          },
          {
            "name": "stretch",
            "description": "The image is stretched to fill the area."
          }
        ],
        "syntax": "[ stretch | repeat | round | space ]{1,2}",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image-repeat"
          }
        ],
        "description": "Specifies how the images for the sides and the middle part of the border image are scaled and tiled. If the second keyword is absent, it is assumed to be the same as the first.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "border-image-slice",
        "values": [
          {
            "name": "fill",
            "description": "Causes the middle part of the border-image to be preserved."
          }
        ],
        "syntax": "<number-percentage>{1,4} && fill?",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image-slice"
          }
        ],
        "description": "Specifies inward offsets from the top, right, bottom, and left edges of the image, dividing it into nine regions: four corners, four edges and a middle.",
        "restrictions": [
          "number",
          "percentage"
        ]
      },
      {
        "name": "border-image-source",
        "values": [
          {
            "name": "none",
            "description": "Use the border styles."
          }
        ],
        "syntax": "none | <image>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image-source"
          }
        ],
        "description": "Specifies an image to use instead of the border styles given by the 'border-style' properties and as an additional background layer for the element. If the value is 'none' or if the image cannot be displayed, the border styles will be used.",
        "restrictions": [
          "image"
        ]
      },
      {
        "name": "border-image-width",
        "values": [
          {
            "name": "auto",
            "description": "The border image width is the intrinsic width or height (whichever is applicable) of the corresponding image slice. If the image does not have the required intrinsic dimension then the corresponding border-width is used instead."
          }
        ],
        "syntax": "[ <length-percentage> | <number> | auto ]{1,4}",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-image-width"
          }
        ],
        "description": "The four values of 'border-image-width' specify offsets that are used to divide the border image area into nine parts. They represent inward distances from the top, right, bottom, and left sides of the area, respectively.",
        "restrictions": [
          "length",
          "percentage",
          "number"
        ]
      },
      {
        "name": "border-inline-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-end"
          }
        ],
        "description": "Logical 'border-right'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-inline-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-start"
          }
        ],
        "description": "Logical 'border-left'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-inline-end-color",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-end-color"
          }
        ],
        "description": "Logical 'border-right-color'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-inline-start-color",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-color'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-start-color"
          }
        ],
        "description": "Logical 'border-left-color'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-inline-end-style",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-end-style"
          }
        ],
        "description": "Logical 'border-right-style'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-inline-start-style",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-start-style"
          }
        ],
        "description": "Logical 'border-left-style'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-inline-end-width",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-end-width"
          }
        ],
        "description": "Logical 'border-right-width'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-inline-start-width",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-start-width"
          }
        ],
        "description": "Logical 'border-left-width'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-left",
        "syntax": "<line-width> || <line-style> || <color>",
        "relevance": 82,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-left"
          }
        ],
        "description": "Shorthand property for setting border width, style and color",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-left-color",
        "syntax": "<color>",
        "relevance": 65,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-left-color"
          }
        ],
        "description": "Sets the color of the left border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-left-style",
        "syntax": "<line-style>",
        "relevance": 54,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-left-style"
          }
        ],
        "description": "Sets the style of the left border.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-left-width",
        "syntax": "<line-width>",
        "relevance": 58,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-left-width"
          }
        ],
        "description": "Sets the thickness of the left border.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-radius",
        "syntax": "<length-percentage>{1,4} [ / <length-percentage>{1,4} ]?",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-radius"
          }
        ],
        "description": "Defines the radii of the outer border edge.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border-right",
        "syntax": "<line-width> || <line-style> || <color>",
        "relevance": 81,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-right"
          }
        ],
        "description": "Shorthand property for setting border width, style and color",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-right-color",
        "syntax": "<color>",
        "relevance": 64,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-right-color"
          }
        ],
        "description": "Sets the color of the right border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-right-style",
        "syntax": "<line-style>",
        "relevance": 54,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-right-style"
          }
        ],
        "description": "Sets the style of the right border.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-right-width",
        "syntax": "<line-width>",
        "relevance": 60,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-right-width"
          }
        ],
        "description": "Sets the thickness of the right border.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-spacing",
        "syntax": "<length> <length>?",
        "relevance": 68,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-spacing"
          }
        ],
        "description": "The lengths specify the distance that separates adjoining cell borders. If one length is specified, it gives both the horizontal and vertical spacing. If two are specified, the first gives the horizontal spacing and the second the vertical spacing. Lengths may not be negative.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "border-style",
        "values": [],
        "syntax": "<line-style>{1,4}",
        "relevance": 79,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-style"
          }
        ],
        "description": "The style of the border around edges of an element.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-top",
        "syntax": "<line-width> || <line-style> || <color>",
        "relevance": 87,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top"
          }
        ],
        "description": "Shorthand property for setting border width, style and color",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "border-top-color",
        "syntax": "<color>",
        "relevance": 71,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top-color"
          }
        ],
        "description": "Sets the color of the top border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "border-top-left-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top-left-radius"
          }
        ],
        "description": "Defines the radii of the top left outer border edge.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border-top-right-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 72,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top-right-radius"
          }
        ],
        "description": "Defines the radii of the top right outer border edge.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "border-top-style",
        "syntax": "<line-style>",
        "relevance": 57,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top-style"
          }
        ],
        "description": "Sets the style of the top border.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "border-top-width",
        "syntax": "<line-width>",
        "relevance": 61,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-top-width"
          }
        ],
        "description": "Sets the thickness of the top border.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "border-width",
        "values": [],
        "syntax": "<line-width>{1,4}",
        "relevance": 81,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-width"
          }
        ],
        "description": "Shorthand that sets the four 'border-*-width' properties. If it has four values, they set top, right, bottom and left in that order. If left is missing, it is the same as right; if bottom is missing, it is the same as top; if right is missing, it is the same as top.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "bottom",
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 89,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/bottom"
          }
        ],
        "description": "Specifies how far an absolutely positioned box's bottom margin edge is offset above the bottom edge of the box's 'containing block'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "box-decoration-break",
        "browsers": [
          "E79",
          "FF32",
          "S6.1",
          "C22",
          "O15"
        ],
        "values": [
          {
            "name": "clone",
            "description": "Each box is independently wrapped with the border and padding."
          },
          {
            "name": "slice",
            "description": "The effect is as though the element were rendered with no breaks present, and then sliced by the breaks afterward."
          }
        ],
        "syntax": "slice | clone",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-decoration-break"
          }
        ],
        "description": "Specifies whether individual boxes are treated as broken pieces of one continuous box, or whether each box is individually wrapped with the border and padding.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "box-shadow",
        "values": [
          {
            "name": "inset",
            "description": "Changes the drop shadow from an outer shadow (one that shadows the box onto the canvas, as if it were lifted above the canvas) to an inner shadow (one that shadows the canvas onto the box, as if the box were cut out of the canvas and shifted behind it)."
          },
          {
            "name": "none",
            "description": "No shadow."
          }
        ],
        "syntax": "none | <shadow>#",
        "relevance": 89,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-shadow"
          }
        ],
        "description": "Attaches one or more drop-shadows to the box. The property is a comma-separated list of shadows, each specified by 2-4 length values, an optional color, and an optional 'inset' keyword. Omitted lengths are 0; omitted colors are a user agent chosen color.",
        "restrictions": [
          "length",
          "color",
          "enum"
        ]
      },
      {
        "name": "box-sizing",
        "values": [
          {
            "name": "border-box",
            "description": "The specified width and height (and respective min/max properties) on this element determine the border box of the element."
          },
          {
            "name": "content-box",
            "description": "Behavior of width and height as specified by CSS2.1. The specified width and height (and respective min/max properties) apply to the width and height respectively of the content box of the element."
          }
        ],
        "syntax": "content-box | border-box",
        "relevance": 92,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-sizing"
          }
        ],
        "description": "Specifies the behavior of the 'width' and 'height' properties.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "break-after",
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the principal box."
          },
          {
            "name": "avoid",
            "description": "Avoid a break before/after the principal box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the principal box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the principal box."
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the principal box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the principal box."
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "syntax": "auto | avoid | always | all | avoid-page | page | left | right | recto | verso | avoid-column | column | avoid-region | region",
        "relevance": 50,
        "description": "Describes the page/column/region break behavior after the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "break-before",
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the principal box."
          },
          {
            "name": "avoid",
            "description": "Avoid a break before/after the principal box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the principal box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the principal box."
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the principal box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the principal box."
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "syntax": "auto | avoid | always | all | avoid-page | page | left | right | recto | verso | avoid-column | column | avoid-region | region",
        "relevance": 50,
        "description": "Describes the page/column/region break behavior before the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "break-inside",
        "values": [
          {
            "name": "auto",
            "description": "Impose no additional breaking constraints within the box."
          },
          {
            "name": "avoid",
            "description": "Avoid breaks within the box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break within the box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break within the box."
          }
        ],
        "syntax": "auto | avoid | avoid-page | avoid-column | avoid-region",
        "relevance": 50,
        "description": "Describes the page/column/region break behavior inside the principal box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "caption-side",
        "values": [
          {
            "name": "bottom",
            "description": "Positions the caption box below the table box."
          },
          {
            "name": "top",
            "description": "Positions the caption box above the table box."
          }
        ],
        "syntax": "top | bottom | block-start | block-end | inline-start | inline-end",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/caption-side"
          }
        ],
        "description": "Specifies the position of the caption box with respect to the table box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "caret-color",
        "browsers": [
          "E79",
          "FF53",
          "S11.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent selects an appropriate color for the caret. This is generally currentcolor, but the user agent may choose a different color to ensure good visibility and contrast with the surrounding content, taking into account the value of currentcolor, the background, shadows, and other factors."
          }
        ],
        "syntax": "auto | <color>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/caret-color"
          }
        ],
        "description": "Controls the color of the text insertion indicator.",
        "restrictions": [
          "color",
          "enum"
        ]
      },
      {
        "name": "clear",
        "values": [
          {
            "name": "both",
            "description": "The clearance of the generated box is set to the amount necessary to place the top border edge below the bottom outer edge of any right-floating and left-floating boxes that resulted from elements earlier in the source document."
          },
          {
            "name": "left",
            "description": "The clearance of the generated box is set to the amount necessary to place the top border edge below the bottom outer edge of any left-floating boxes that resulted from elements earlier in the source document."
          },
          {
            "name": "none",
            "description": "No constraint on the box's position with respect to floats."
          },
          {
            "name": "right",
            "description": "The clearance of the generated box is set to the amount necessary to place the top border edge below the bottom outer edge of any right-floating boxes that resulted from elements earlier in the source document."
          }
        ],
        "syntax": "none | left | right | both | inline-start | inline-end",
        "relevance": 84,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/clear"
          }
        ],
        "description": "Indicates which sides of an element's box(es) may not be adjacent to an earlier floating box. The 'clear' property does not consider floats inside the element itself or in other block formatting contexts.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "clip",
        "values": [
          {
            "name": "auto",
            "description": "The element does not clip."
          },
          {
            "name": "rect()",
            "description": "Specifies offsets from the edges of the border box."
          }
        ],
        "syntax": "<shape> | auto",
        "relevance": 73,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/clip"
          }
        ],
        "description": "Deprecated. Use the 'clip-path' property when support allows. Defines the visible portion of an element\u2019s box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "clip-path",
        "values": [
          {
            "name": "none",
            "description": "No clipping path gets created."
          },
          {
            "name": "url()",
            "description": "References a <clipPath> element to create a clipping path."
          }
        ],
        "syntax": "<clip-source> | [ <basic-shape> || <geometry-box> ] | none",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/clip-path"
          }
        ],
        "description": "Specifies a clipping path where everything inside the path is visible and everything outside is clipped out.",
        "restrictions": [
          "url",
          "shape",
          "geometry-box",
          "enum"
        ]
      },
      {
        "name": "clip-rule",
        "browsers": [
          "E",
          "C5",
          "FF3",
          "IE10",
          "O9",
          "S6"
        ],
        "values": [
          {
            "name": "evenodd",
            "description": "Determines the \u2018insideness\u2019 of a point on the canvas by drawing a ray from that point to infinity in any direction and counting the number of path segments from the given shape that the ray crosses."
          },
          {
            "name": "nonzero",
            "description": "Determines the \u2018insideness\u2019 of a point on the canvas by drawing a ray from that point to infinity in any direction and then examining the places where a segment of the shape crosses the ray."
          }
        ],
        "relevance": 50,
        "description": "Indicates the algorithm which is to be used to determine what parts of the canvas are included inside the shape.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "color",
        "syntax": "<color>",
        "relevance": 94,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/color"
          }
        ],
        "description": "Sets the color of an element's text",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "color-interpolation-filters",
        "browsers": [
          "E",
          "C5",
          "FF3",
          "IE10",
          "O9",
          "S6"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Color operations are not required to occur in a particular color space."
          },
          {
            "name": "linearRGB",
            "description": "Color operations should occur in the linearized RGB color space."
          },
          {
            "name": "sRGB",
            "description": "Color operations should occur in the sRGB color space."
          }
        ],
        "relevance": 50,
        "description": "Specifies the color space for imaging operations performed via filter effects.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "column-count",
        "values": [
          {
            "name": "auto",
            "description": "Determines the number of columns by the 'column-width' property and the element width."
          }
        ],
        "syntax": "<integer> | auto",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-count"
          }
        ],
        "description": "Describes the optimal number of columns into which the content of the element will be flowed.",
        "restrictions": [
          "integer",
          "enum"
        ]
      },
      {
        "name": "column-fill",
        "values": [
          {
            "name": "auto",
            "description": "Fills columns sequentially."
          },
          {
            "name": "balance",
            "description": "Balance content equally between columns, if possible."
          }
        ],
        "syntax": "auto | balance | balance-all",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-fill"
          }
        ],
        "description": "In continuous media, this property will only be consulted if the length of columns has been constrained. Otherwise, columns will automatically be balanced.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "column-gap",
        "values": [
          {
            "name": "normal",
            "description": "User agent specific and typically equivalent to 1em."
          }
        ],
        "syntax": "normal | <length-percentage>",
        "relevance": 52,
        "description": "Sets the gap between columns. If there is a column rule between columns, it will appear in the middle of the gap.",
        "restrictions": [
          "length",
          "enum"
        ]
      },
      {
        "name": "column-rule",
        "syntax": "<'column-rule-width'> || <'column-rule-style'> || <'column-rule-color'>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-rule"
          }
        ],
        "description": "Shorthand for setting 'column-rule-width', 'column-rule-style', and 'column-rule-color' at the same place in the style sheet. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "column-rule-color",
        "syntax": "<color>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-rule-color"
          }
        ],
        "description": "Sets the color of the column rule",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "column-rule-style",
        "syntax": "<'border-style'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-rule-style"
          }
        ],
        "description": "Sets the style of the rule between columns of an element.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "column-rule-width",
        "syntax": "<'border-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-rule-width"
          }
        ],
        "description": "Sets the width of the rule between columns. Negative values are not allowed.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "columns",
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "syntax": "<'column-width'> || <'column-count'>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/columns"
          }
        ],
        "description": "A shorthand property which sets both 'column-width' and 'column-count'.",
        "restrictions": [
          "length",
          "integer",
          "enum"
        ]
      },
      {
        "name": "column-span",
        "values": [
          {
            "name": "all",
            "description": "The element spans across all columns. Content in the normal flow that appears before the element is automatically balanced across all columns before the element appear."
          },
          {
            "name": "none",
            "description": "The element does not span multiple columns."
          }
        ],
        "syntax": "none | all",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-span"
          }
        ],
        "description": "Describes the page/column break behavior after the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "column-width",
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "syntax": "<length> | auto",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/column-width"
          }
        ],
        "description": "Describes the width of columns in multicol elements.",
        "restrictions": [
          "length",
          "enum"
        ]
      },
      {
        "name": "contain",
        "browsers": [
          "E79",
          "FF69",
          "C52",
          "O40"
        ],
        "values": [
          {
            "name": "none",
            "description": "Indicates that the property has no effect."
          },
          {
            "name": "strict",
            "description": "Turns on all forms of containment for the element."
          },
          {
            "name": "content",
            "description": "All containment rules except size are applied to the element."
          },
          {
            "name": "size",
            "description": "For properties that can have effects on more than just an element and its descendants, those effects don't escape the containing element."
          },
          {
            "name": "layout",
            "description": "Turns on layout containment for the element."
          },
          {
            "name": "style",
            "description": "Turns on style containment for the element."
          },
          {
            "name": "paint",
            "description": "Turns on paint containment for the element."
          }
        ],
        "syntax": "none | strict | content | [ size || layout || style || paint ]",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/contain"
          }
        ],
        "description": "Indicates that an element and its contents are, as much as possible, independent of the rest of the document tree.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "content",
        "values": [
          {
            "name": "attr()",
            "description": "The attr(n) function returns as a string the value of attribute n for the subject of the selector."
          },
          {
            "name": "counter(name)",
            "description": "Counters are denoted by identifiers (see the 'counter-increment' and 'counter-reset' properties)."
          },
          {
            "name": "icon",
            "description": "The (pseudo-)element is replaced in its entirety by the resource referenced by its 'icon' property, and treated as a replaced element."
          },
          {
            "name": "none",
            "description": "On elements, this inhibits the children of the element from being rendered as children of this element, as if the element was empty. On pseudo-elements it causes the pseudo-element to have no content."
          },
          {
            "name": "normal",
            "description": "See http://www.w3.org/TR/css3-content/#content for computation rules."
          },
          {
            "name": "url()"
          }
        ],
        "syntax": "normal | none | [ <content-replacement> | <content-list> ] [/ <string> ]?",
        "relevance": 89,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/content"
          }
        ],
        "description": "Determines which page-based occurrence of a given element is applied to a counter or string value.",
        "restrictions": [
          "string",
          "url"
        ]
      },
      {
        "name": "counter-increment",
        "values": [
          {
            "name": "none",
            "description": "This element does not alter the value of any counters."
          }
        ],
        "syntax": "[ <custom-ident> <integer>? ]+ | none",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/counter-increment"
          }
        ],
        "description": "Manipulate the value of existing counters.",
        "restrictions": [
          "identifier",
          "integer"
        ]
      },
      {
        "name": "counter-reset",
        "values": [
          {
            "name": "none",
            "description": "The counter is not modified."
          }
        ],
        "syntax": "[ <custom-ident> <integer>? ]+ | none",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/counter-reset"
          }
        ],
        "description": "Property accepts one or more names of counters (identifiers), each one optionally followed by an integer. The integer gives the value that the counter is set to on each occurrence of the element.",
        "restrictions": [
          "identifier",
          "integer"
        ]
      },
      {
        "name": "cursor",
        "values": [
          {
            "name": "alias",
            "description": "Indicates an alias of/shortcut to something is to be created. Often rendered as an arrow with a small curved arrow next to it."
          },
          {
            "name": "all-scroll",
            "description": "Indicates that the something can be scrolled in any direction. Often rendered as arrows pointing up, down, left, and right with a dot in the middle."
          },
          {
            "name": "auto",
            "description": "The UA determines the cursor to display based on the current context."
          },
          {
            "name": "cell",
            "description": "Indicates that a cell or set of cells may be selected. Often rendered as a thick plus-sign with a dot in the middle."
          },
          {
            "name": "col-resize",
            "description": "Indicates that the item/column can be resized horizontally. Often rendered as arrows pointing left and right with a vertical bar separating them."
          },
          {
            "name": "context-menu",
            "description": "A context menu is available for the object under the cursor. Often rendered as an arrow with a small menu-like graphic next to it."
          },
          {
            "name": "copy",
            "description": "Indicates something is to be copied. Often rendered as an arrow with a small plus sign next to it."
          },
          {
            "name": "crosshair",
            "description": "A simple crosshair (e.g., short line segments resembling a '+' sign). Often used to indicate a two dimensional bitmap selection mode."
          },
          {
            "name": "default",
            "description": "The platform-dependent default cursor. Often rendered as an arrow."
          },
          {
            "name": "e-resize",
            "description": "Indicates that east edge is to be moved."
          },
          {
            "name": "ew-resize",
            "description": "Indicates a bidirectional east-west resize cursor."
          },
          {
            "name": "grab",
            "description": "Indicates that something can be grabbed."
          },
          {
            "name": "grabbing",
            "description": "Indicates that something is being grabbed."
          },
          {
            "name": "help",
            "description": "Help is available for the object under the cursor. Often rendered as a question mark or a balloon."
          },
          {
            "name": "move",
            "description": "Indicates something is to be moved."
          },
          {
            "name": "-moz-grab",
            "description": "Indicates that something can be grabbed."
          },
          {
            "name": "-moz-grabbing",
            "description": "Indicates that something is being grabbed."
          },
          {
            "name": "-moz-zoom-in",
            "description": "Indicates that something can be zoomed (magnified) in."
          },
          {
            "name": "-moz-zoom-out",
            "description": "Indicates that something can be zoomed (magnified) out."
          },
          {
            "name": "ne-resize",
            "description": "Indicates that movement starts from north-east corner."
          },
          {
            "name": "nesw-resize",
            "description": "Indicates a bidirectional north-east/south-west cursor."
          },
          {
            "name": "no-drop",
            "description": "Indicates that the dragged item cannot be dropped at the current cursor location. Often rendered as a hand or pointer with a small circle with a line through it."
          },
          {
            "name": "none",
            "description": "No cursor is rendered for the element."
          },
          {
            "name": "not-allowed",
            "description": "Indicates that the requested action will not be carried out. Often rendered as a circle with a line through it."
          },
          {
            "name": "n-resize",
            "description": "Indicates that north edge is to be moved."
          },
          {
            "name": "ns-resize",
            "description": "Indicates a bidirectional north-south cursor."
          },
          {
            "name": "nw-resize",
            "description": "Indicates that movement starts from north-west corner."
          },
          {
            "name": "nwse-resize",
            "description": "Indicates a bidirectional north-west/south-east cursor."
          },
          {
            "name": "pointer",
            "description": "The cursor is a pointer that indicates a link."
          },
          {
            "name": "progress",
            "description": "A progress indicator. The program is performing some processing, but is different from 'wait' in that the user may still interact with the program. Often rendered as a spinning beach ball, or an arrow with a watch or hourglass."
          },
          {
            "name": "row-resize",
            "description": "Indicates that the item/row can be resized vertically. Often rendered as arrows pointing up and down with a horizontal bar separating them."
          },
          {
            "name": "se-resize",
            "description": "Indicates that movement starts from south-east corner."
          },
          {
            "name": "s-resize",
            "description": "Indicates that south edge is to be moved."
          },
          {
            "name": "sw-resize",
            "description": "Indicates that movement starts from south-west corner."
          },
          {
            "name": "text",
            "description": "Indicates text that may be selected. Often rendered as a vertical I-beam."
          },
          {
            "name": "vertical-text",
            "description": "Indicates vertical-text that may be selected. Often rendered as a horizontal I-beam."
          },
          {
            "name": "wait",
            "description": "Indicates that the program is busy and the user should wait. Often rendered as a watch or hourglass."
          },
          {
            "name": "-webkit-grab",
            "description": "Indicates that something can be grabbed."
          },
          {
            "name": "-webkit-grabbing",
            "description": "Indicates that something is being grabbed."
          },
          {
            "name": "-webkit-zoom-in",
            "description": "Indicates that something can be zoomed (magnified) in."
          },
          {
            "name": "-webkit-zoom-out",
            "description": "Indicates that something can be zoomed (magnified) out."
          },
          {
            "name": "w-resize",
            "description": "Indicates that west edge is to be moved."
          },
          {
            "name": "zoom-in",
            "description": "Indicates that something can be zoomed (magnified) in."
          },
          {
            "name": "zoom-out",
            "description": "Indicates that something can be zoomed (magnified) out."
          }
        ],
        "syntax": "[ [ <url> [ <x> <y> ]? , ]* [ auto | default | none | context-menu | help | pointer | progress | wait | cell | crosshair | text | vertical-text | alias | copy | move | no-drop | not-allowed | e-resize | n-resize | ne-resize | nw-resize | s-resize | se-resize | sw-resize | w-resize | ew-resize | ns-resize | nesw-resize | nwse-resize | col-resize | row-resize | all-scroll | zoom-in | zoom-out | grab | grabbing ] ]",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/cursor"
          }
        ],
        "description": "Allows control over cursor appearance in an element",
        "restrictions": [
          "url",
          "number",
          "enum"
        ]
      },
      {
        "name": "direction",
        "values": [
          {
            "name": "ltr",
            "description": "Left-to-right direction."
          },
          {
            "name": "rtl",
            "description": "Right-to-left direction."
          }
        ],
        "syntax": "ltr | rtl",
        "relevance": 68,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/direction"
          }
        ],
        "description": "Specifies the inline base direction or directionality of any bidi paragraph, embedding, isolate, or override established by the box. Note: for HTML content use the 'dir' attribute and 'bdo' element rather than this property.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "display",
        "values": [
          {
            "name": "block",
            "description": "The element generates a block-level box"
          },
          {
            "name": "contents",
            "description": "The element itself does not generate any boxes, but its children and pseudo-elements still generate boxes as normal."
          },
          {
            "name": "flex",
            "description": "The element generates a principal flex container box and establishes a flex formatting context."
          },
          {
            "name": "flexbox",
            "description": "The element lays out its contents using flow layout (block-and-inline layout). Standardized as 'flex'."
          },
          {
            "name": "flow-root",
            "description": "The element generates a block container box, and lays out its contents using flow layout."
          },
          {
            "name": "grid",
            "description": "The element generates a principal grid container box, and establishes a grid formatting context."
          },
          {
            "name": "inline",
            "description": "The element generates an inline-level box."
          },
          {
            "name": "inline-block",
            "description": "A block box, which itself is flowed as a single inline box, similar to a replaced element. The inside of an inline-block is formatted as a block box, and the box itself is formatted as an inline box."
          },
          {
            "name": "inline-flex",
            "description": "Inline-level flex container."
          },
          {
            "name": "inline-flexbox",
            "description": "Inline-level flex container. Standardized as 'inline-flex'"
          },
          {
            "name": "inline-table",
            "description": "Inline-level table wrapper box containing table box."
          },
          {
            "name": "list-item",
            "description": "One or more block boxes and one marker box."
          },
          {
            "name": "-moz-box",
            "description": "The element lays out its contents using flow layout (block-and-inline layout). Standardized as 'flex'."
          },
          {
            "name": "-moz-deck"
          },
          {
            "name": "-moz-grid"
          },
          {
            "name": "-moz-grid-group"
          },
          {
            "name": "-moz-grid-line"
          },
          {
            "name": "-moz-groupbox"
          },
          {
            "name": "-moz-inline-box",
            "description": "Inline-level flex container. Standardized as 'inline-flex'"
          },
          {
            "name": "-moz-inline-grid"
          },
          {
            "name": "-moz-inline-stack"
          },
          {
            "name": "-moz-marker"
          },
          {
            "name": "-moz-popup"
          },
          {
            "name": "-moz-stack"
          },
          {
            "name": "-ms-flexbox",
            "description": "The element lays out its contents using flow layout (block-and-inline layout). Standardized as 'flex'."
          },
          {
            "name": "-ms-grid",
            "description": "The element generates a principal grid container box, and establishes a grid formatting context."
          },
          {
            "name": "-ms-inline-flexbox",
            "description": "Inline-level flex container. Standardized as 'inline-flex'"
          },
          {
            "name": "-ms-inline-grid",
            "description": "Inline-level grid container."
          },
          {
            "name": "none",
            "description": "The element and its descendants generates no boxes."
          },
          {
            "name": "ruby",
            "description": "The element generates a principal ruby container box, and establishes a ruby formatting context."
          },
          {
            "name": "ruby-base"
          },
          {
            "name": "ruby-base-container"
          },
          {
            "name": "ruby-text"
          },
          {
            "name": "ruby-text-container"
          },
          {
            "name": "run-in",
            "description": "The element generates a run-in box. Run-in elements act like inlines or blocks, depending on the surrounding elements."
          },
          {
            "name": "table",
            "description": "The element generates a principal table wrapper box containing an additionally-generated table box, and establishes a table formatting context."
          },
          {
            "name": "table-caption"
          },
          {
            "name": "table-cell"
          },
          {
            "name": "table-column"
          },
          {
            "name": "table-column-group"
          },
          {
            "name": "table-footer-group"
          },
          {
            "name": "table-header-group"
          },
          {
            "name": "table-row"
          },
          {
            "name": "table-row-group"
          },
          {
            "name": "-webkit-box",
            "description": "The element lays out its contents using flow layout (block-and-inline layout). Standardized as 'flex'."
          },
          {
            "name": "-webkit-flex",
            "description": "The element lays out its contents using flow layout (block-and-inline layout)."
          },
          {
            "name": "-webkit-inline-box",
            "description": "Inline-level flex container. Standardized as 'inline-flex'"
          },
          {
            "name": "-webkit-inline-flex",
            "description": "Inline-level flex container."
          }
        ],
        "syntax": "[ <display-outside> || <display-inside> ] | <display-listitem> | <display-internal> | <display-box> | <display-legacy>",
        "relevance": 96,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/display"
          }
        ],
        "description": "In combination with 'float' and 'position', determines the type of box or boxes that are generated for an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "empty-cells",
        "values": [
          {
            "name": "hide",
            "description": "No borders or backgrounds are drawn around/behind empty cells."
          },
          {
            "name": "-moz-show-background"
          },
          {
            "name": "show",
            "description": "Borders and backgrounds are drawn around/behind empty cells (like normal cells)."
          }
        ],
        "syntax": "show | hide",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/empty-cells"
          }
        ],
        "description": "In the separated borders model, this property controls the rendering of borders and backgrounds around cells that have no visible content.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "enable-background",
        "values": [
          {
            "name": "accumulate",
            "description": "If the ancestor container element has a property of new, then all graphics elements within the current container are rendered both on the parent's background image and onto the target."
          },
          {
            "name": "new",
            "description": "Create a new background image canvas. All children of the current container element can access the background, and they will be rendered onto both the parent's background image canvas in addition to the target device."
          }
        ],
        "relevance": 50,
        "description": "Deprecated. Use 'isolation' property instead when support allows. Specifies how the accumulation of the background image is managed.",
        "restrictions": [
          "integer",
          "length",
          "percentage",
          "enum"
        ]
      },
      {
        "name": "fallback",
        "browsers": [
          "FF33"
        ],
        "syntax": "<counter-style-name>",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies a fallback counter style to be used when the current counter style can\u2019t create a representation for a given counter value.",
        "restrictions": [
          "identifier"
        ]
      },
      {
        "name": "fill",
        "values": [
          {
            "name": "url()",
            "description": "A URL reference to a paint server element, which is an element that defines a paint server: \u2018hatch\u2019, \u2018linearGradient\u2019, \u2018mesh\u2019, \u2018pattern\u2019, \u2018radialGradient\u2019 and \u2018solidcolor\u2019."
          },
          {
            "name": "none",
            "description": "No paint is applied in this layer."
          }
        ],
        "relevance": 74,
        "description": "Paints the interior of the given graphical element.",
        "restrictions": [
          "color",
          "enum",
          "url"
        ]
      },
      {
        "name": "fill-opacity",
        "relevance": 52,
        "description": "Specifies the opacity of the painting operation used to paint the interior the current object.",
        "restrictions": [
          "number(0-1)"
        ]
      },
      {
        "name": "fill-rule",
        "values": [
          {
            "name": "evenodd",
            "description": "Determines the \u2018insideness\u2019 of a point on the canvas by drawing a ray from that point to infinity in any direction and counting the number of path segments from the given shape that the ray crosses."
          },
          {
            "name": "nonzero",
            "description": "Determines the \u2018insideness\u2019 of a point on the canvas by drawing a ray from that point to infinity in any direction and then examining the places where a segment of the shape crosses the ray."
          }
        ],
        "relevance": 50,
        "description": "Indicates the algorithm (or winding rule) which is to be used to determine what parts of the canvas are included inside the shape.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "filter",
        "browsers": [
          "E12",
          "FF35",
          "S9.1",
          "C53",
          "O40"
        ],
        "values": [
          {
            "name": "none",
            "description": "No filter effects are applied."
          },
          {
            "name": "blur()",
            "description": "Applies a Gaussian blur to the input image."
          },
          {
            "name": "brightness()",
            "description": "Applies a linear multiplier to input image, making it appear more or less bright."
          },
          {
            "name": "contrast()",
            "description": "Adjusts the contrast of the input."
          },
          {
            "name": "drop-shadow()",
            "description": "Applies a drop shadow effect to the input image."
          },
          {
            "name": "grayscale()",
            "description": "Converts the input image to grayscale."
          },
          {
            "name": "hue-rotate()",
            "description": "Applies a hue rotation on the input image. "
          },
          {
            "name": "invert()",
            "description": "Inverts the samples in the input image."
          },
          {
            "name": "opacity()",
            "description": "Applies transparency to the samples in the input image."
          },
          {
            "name": "saturate()",
            "description": "Saturates the input image."
          },
          {
            "name": "sepia()",
            "description": "Converts the input image to sepia."
          },
          {
            "name": "url()",
            "browsers": [
              "E12",
              "FF35",
              "S9.1",
              "C53",
              "O40"
            ],
            "description": "A filter reference to a <filter> element."
          }
        ],
        "syntax": "none | <filter-function-list>",
        "relevance": 64,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/filter"
          }
        ],
        "description": "Processes an element\u2019s rendering before it is displayed in the document, by applying one or more filter effects.",
        "restrictions": [
          "enum",
          "url"
        ]
      },
      {
        "name": "flex",
        "values": [
          {
            "name": "auto",
            "description": "Retrieves the value of the main size property as the used 'flex-basis'."
          },
          {
            "name": "content",
            "description": "Indicates automatic sizing, based on the flex item\u2019s content."
          },
          {
            "name": "none",
            "description": "Expands to '0 0 auto'."
          }
        ],
        "syntax": "none | [ <'flex-grow'> <'flex-shrink'>? || <'flex-basis'> ]",
        "relevance": 77,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex"
          }
        ],
        "description": "Specifies the components of a flexible length: the flex grow factor and flex shrink factor, and the flex basis.",
        "restrictions": [
          "length",
          "number",
          "percentage"
        ]
      },
      {
        "name": "flex-basis",
        "values": [
          {
            "name": "auto",
            "description": "Retrieves the value of the main size property as the used 'flex-basis'."
          },
          {
            "name": "content",
            "description": "Indicates automatic sizing, based on the flex item\u2019s content."
          }
        ],
        "syntax": "content | <'width'>",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-basis"
          }
        ],
        "description": "Sets the flex basis.",
        "restrictions": [
          "length",
          "number",
          "percentage"
        ]
      },
      {
        "name": "flex-direction",
        "values": [
          {
            "name": "column",
            "description": "The flex container\u2019s main axis has the same orientation as the block axis of the current writing mode."
          },
          {
            "name": "column-reverse",
            "description": "Same as 'column', except the main-start and main-end directions are swapped."
          },
          {
            "name": "row",
            "description": "The flex container\u2019s main axis has the same orientation as the inline axis of the current writing mode."
          },
          {
            "name": "row-reverse",
            "description": "Same as 'row', except the main-start and main-end directions are swapped."
          }
        ],
        "syntax": "row | row-reverse | column | column-reverse",
        "relevance": 78,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-direction"
          }
        ],
        "description": "Specifies how flex items are placed in the flex container, by setting the direction of the flex container\u2019s main axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "flex-flow",
        "values": [
          {
            "name": "column",
            "description": "The flex container\u2019s main axis has the same orientation as the block axis of the current writing mode."
          },
          {
            "name": "column-reverse",
            "description": "Same as 'column', except the main-start and main-end directions are swapped."
          },
          {
            "name": "nowrap",
            "description": "The flex container is single-line."
          },
          {
            "name": "row",
            "description": "The flex container\u2019s main axis has the same orientation as the inline axis of the current writing mode."
          },
          {
            "name": "row-reverse",
            "description": "Same as 'row', except the main-start and main-end directions are swapped."
          },
          {
            "name": "wrap",
            "description": "The flexbox is multi-line."
          },
          {
            "name": "wrap-reverse",
            "description": "Same as 'wrap', except the cross-start and cross-end directions are swapped."
          }
        ],
        "syntax": "<'flex-direction'> || <'flex-wrap'>",
        "relevance": 58,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-flow"
          }
        ],
        "description": "Specifies how flexbox items are placed in the flexbox.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "flex-grow",
        "syntax": "<number>",
        "relevance": 71,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-grow"
          }
        ],
        "description": "Sets the flex grow factor. Negative numbers are invalid.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "flex-shrink",
        "syntax": "<number>",
        "relevance": 69,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-shrink"
          }
        ],
        "description": "Sets the flex shrink factor. Negative numbers are invalid.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "flex-wrap",
        "values": [
          {
            "name": "nowrap",
            "description": "The flex container is single-line."
          },
          {
            "name": "wrap",
            "description": "The flexbox is multi-line."
          },
          {
            "name": "wrap-reverse",
            "description": "Same as 'wrap', except the cross-start and cross-end directions are swapped."
          }
        ],
        "syntax": "nowrap | wrap | wrap-reverse",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/flex-wrap"
          }
        ],
        "description": "Controls whether the flex container is single-line or multi-line, and the direction of the cross-axis, which determines the direction new lines are stacked in.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "float",
        "values": [
          {
            "name": "inline-end",
            "description": "A keyword indicating that the element must float on the end side of its containing block. That is the right side with ltr scripts, and the left side with rtl scripts."
          },
          {
            "name": "inline-start",
            "description": "A keyword indicating that the element must float on the start side of its containing block. That is the left side with ltr scripts, and the right side with rtl scripts."
          },
          {
            "name": "left",
            "description": "The element generates a block box that is floated to the left. Content flows on the right side of the box, starting at the top (subject to the 'clear' property)."
          },
          {
            "name": "none",
            "description": "The box is not floated."
          },
          {
            "name": "right",
            "description": "Similar to 'left', except the box is floated to the right, and content flows on the left side of the box, starting at the top."
          }
        ],
        "syntax": "left | right | none | inline-start | inline-end",
        "relevance": 92,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/float"
          }
        ],
        "description": "Specifies how a box should be floated. It may be set for any element, but only applies to elements that generate boxes that are not absolutely positioned.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "flood-color",
        "browsers": [
          "E",
          "C5",
          "FF3",
          "IE10",
          "O9",
          "S6"
        ],
        "relevance": 50,
        "description": "Indicates what color to use to flood the current filter primitive subregion.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "flood-opacity",
        "browsers": [
          "E",
          "C5",
          "FF3",
          "IE10",
          "O9",
          "S6"
        ],
        "relevance": 50,
        "description": "Indicates what opacity to use to flood the current filter primitive subregion.",
        "restrictions": [
          "number(0-1)",
          "percentage"
        ]
      },
      {
        "name": "font",
        "values": [
          {
            "name": "100",
            "description": "Thin"
          },
          {
            "name": "200",
            "description": "Extra Light (Ultra Light)"
          },
          {
            "name": "300",
            "description": "Light"
          },
          {
            "name": "400",
            "description": "Normal"
          },
          {
            "name": "500",
            "description": "Medium"
          },
          {
            "name": "600",
            "description": "Semi Bold (Demi Bold)"
          },
          {
            "name": "700",
            "description": "Bold"
          },
          {
            "name": "800",
            "description": "Extra Bold (Ultra Bold)"
          },
          {
            "name": "900",
            "description": "Black (Heavy)"
          },
          {
            "name": "bold",
            "description": "Same as 700"
          },
          {
            "name": "bolder",
            "description": "Specifies the weight of the face bolder than the inherited value."
          },
          {
            "name": "caption",
            "description": "The font used for captioned controls (e.g., buttons, drop-downs, etc.)."
          },
          {
            "name": "icon",
            "description": "The font used to label icons."
          },
          {
            "name": "italic",
            "description": "Selects a font that is labeled 'italic', or, if that is not available, one labeled 'oblique'."
          },
          {
            "name": "large"
          },
          {
            "name": "larger"
          },
          {
            "name": "lighter",
            "description": "Specifies the weight of the face lighter than the inherited value."
          },
          {
            "name": "medium"
          },
          {
            "name": "menu",
            "description": "The font used in menus (e.g., dropdown menus and menu lists)."
          },
          {
            "name": "message-box",
            "description": "The font used in dialog boxes."
          },
          {
            "name": "normal",
            "description": "Specifies a face that is not labeled as a small-caps font."
          },
          {
            "name": "oblique",
            "description": "Selects a font that is labeled 'oblique'."
          },
          {
            "name": "small"
          },
          {
            "name": "small-caps",
            "description": "Specifies a font that is labeled as a small-caps font. If a genuine small-caps font is not available, user agents should simulate a small-caps font."
          },
          {
            "name": "small-caption",
            "description": "The font used for labeling small controls."
          },
          {
            "name": "smaller"
          },
          {
            "name": "status-bar",
            "description": "The font used in window status bars."
          },
          {
            "name": "x-large"
          },
          {
            "name": "x-small"
          },
          {
            "name": "xx-large"
          },
          {
            "name": "xx-small"
          }
        ],
        "syntax": "[ [ <'font-style'> || <font-variant-css21> || <'font-weight'> || <'font-stretch'> ]? <'font-size'> [ / <'line-height'> ]? <'font-family'> ] | caption | icon | menu | message-box | small-caption | status-bar",
        "relevance": 82,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font"
          }
        ],
        "description": "Shorthand property for setting 'font-style', 'font-variant', 'font-weight', 'font-size', 'line-height', and 'font-family', at the same place in the style sheet. The syntax of this property is based on a traditional typographical shorthand notation to set multiple properties related to fonts.",
        "restrictions": [
          "font"
        ]
      },
      {
        "name": "font-family",
        "values": [
          {
            "name": "-apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, Cantarell, 'Open Sans', 'Helvetica Neue', sans-serif"
          },
          {
            "name": "Arial, Helvetica, sans-serif"
          },
          {
            "name": "Cambria, Cochin, Georgia, Times, 'Times New Roman', serif"
          },
          {
            "name": "'Courier New', Courier, monospace"
          },
          {
            "name": "cursive"
          },
          {
            "name": "fantasy"
          },
          {
            "name": "'Franklin Gothic Medium', 'Arial Narrow', Arial, sans-serif"
          },
          {
            "name": "Georgia, 'Times New Roman', Times, serif"
          },
          {
            "name": "'Gill Sans', 'Gill Sans MT', Calibri, 'Trebuchet MS', sans-serif"
          },
          {
            "name": "Impact, Haettenschweiler, 'Arial Narrow Bold', sans-serif"
          },
          {
            "name": "'Lucida Sans', 'Lucida Sans Regular', 'Lucida Grande', 'Lucida Sans Unicode', Geneva, Verdana, sans-serif"
          },
          {
            "name": "monospace"
          },
          {
            "name": "sans-serif"
          },
          {
            "name": "'Segoe UI', Tahoma, Geneva, Verdana, sans-serif"
          },
          {
            "name": "serif"
          },
          {
            "name": "'Times New Roman', Times, serif"
          },
          {
            "name": "'Trebuchet MS', 'Lucida Sans Unicode', 'Lucida Grande', 'Lucida Sans', Arial, sans-serif"
          },
          {
            "name": "Verdana, Geneva, Tahoma, sans-serif"
          }
        ],
        "syntax": "<family-name>",
        "relevance": 92,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-family"
          }
        ],
        "description": "Specifies a prioritized list of font family names or generic family names. A user agent iterates through the list of family names until it matches an available font that contains a glyph for the character to be rendered.",
        "restrictions": [
          "font"
        ]
      },
      {
        "name": "font-feature-settings",
        "values": [
          {
            "name": '"aalt"',
            "description": "Access All Alternates."
          },
          {
            "name": '"abvf"',
            "description": "Above-base Forms. Required in Khmer script."
          },
          {
            "name": '"abvm"',
            "description": "Above-base Mark Positioning. Required in Indic scripts."
          },
          {
            "name": '"abvs"',
            "description": "Above-base Substitutions. Required in Indic scripts."
          },
          {
            "name": '"afrc"',
            "description": "Alternative Fractions."
          },
          {
            "name": '"akhn"',
            "description": "Akhand. Required in most Indic scripts."
          },
          {
            "name": '"blwf"',
            "description": "Below-base Form. Required in a number of Indic scripts."
          },
          {
            "name": '"blwm"',
            "description": "Below-base Mark Positioning. Required in Indic scripts."
          },
          {
            "name": '"blws"',
            "description": "Below-base Substitutions. Required in Indic scripts."
          },
          {
            "name": '"calt"',
            "description": "Contextual Alternates."
          },
          {
            "name": '"case"',
            "description": "Case-Sensitive Forms. Applies only to European scripts; particularly prominent in Spanish-language setting."
          },
          {
            "name": '"ccmp"',
            "description": "Glyph Composition/Decomposition."
          },
          {
            "name": '"cfar"',
            "description": "Conjunct Form After Ro. Required in Khmer scripts."
          },
          {
            "name": '"cjct"',
            "description": "Conjunct Forms. Required in Indic scripts that show similarity to Devanagari."
          },
          {
            "name": '"clig"',
            "description": "Contextual Ligatures."
          },
          {
            "name": '"cpct"',
            "description": "Centered CJK Punctuation. Used primarily in Chinese fonts."
          },
          {
            "name": '"cpsp"',
            "description": "Capital Spacing. Should not be used in connecting scripts (e.g. most Arabic)."
          },
          {
            "name": '"cswh"',
            "description": "Contextual Swash."
          },
          {
            "name": '"curs"',
            "description": "Cursive Positioning. Can be used in any cursive script."
          },
          {
            "name": '"c2pc"',
            "description": "Petite Capitals From Capitals. Applies only to bicameral scripts."
          },
          {
            "name": '"c2sc"',
            "description": "Small Capitals From Capitals. Applies only to bicameral scripts."
          },
          {
            "name": '"dist"',
            "description": "Distances. Required in Indic scripts."
          },
          {
            "name": '"dlig"',
            "description": "Discretionary ligatures."
          },
          {
            "name": '"dnom"',
            "description": "Denominators."
          },
          {
            "name": '"dtls"',
            "description": "Dotless Forms. Applied to math formula layout."
          },
          {
            "name": '"expt"',
            "description": "Expert Forms. Applies only to Japanese."
          },
          {
            "name": '"falt"',
            "description": "Final Glyph on Line Alternates. Can be used in any cursive script."
          },
          {
            "name": '"fin2"',
            "description": "Terminal Form #2. Used only with the Syriac script."
          },
          {
            "name": '"fin3"',
            "description": "Terminal Form #3. Used only with the Syriac script."
          },
          {
            "name": '"fina"',
            "description": "Terminal Forms. Can be used in any alphabetic script."
          },
          {
            "name": '"flac"',
            "description": "Flattened ascent forms. Applied to math formula layout."
          },
          {
            "name": '"frac"',
            "description": "Fractions."
          },
          {
            "name": '"fwid"',
            "description": "Full Widths. Applies to any script which can use monospaced forms."
          },
          {
            "name": '"half"',
            "description": "Half Forms. Required in Indic scripts that show similarity to Devanagari."
          },
          {
            "name": '"haln"',
            "description": "Halant Forms. Required in Indic scripts."
          },
          {
            "name": '"halt"',
            "description": "Alternate Half Widths. Used only in CJKV fonts."
          },
          {
            "name": '"hist"',
            "description": "Historical Forms."
          },
          {
            "name": '"hkna"',
            "description": "Horizontal Kana Alternates. Applies only to fonts that support kana (hiragana and katakana)."
          },
          {
            "name": '"hlig"',
            "description": "Historical Ligatures."
          },
          {
            "name": '"hngl"',
            "description": "Hangul. Korean only."
          },
          {
            "name": '"hojo"',
            "description": "Hojo Kanji Forms (JIS X 0212-1990 Kanji Forms). Used only with Kanji script."
          },
          {
            "name": '"hwid"',
            "description": "Half Widths. Generally used only in CJKV fonts."
          },
          {
            "name": '"init"',
            "description": "Initial Forms. Can be used in any alphabetic script."
          },
          {
            "name": '"isol"',
            "description": "Isolated Forms. Can be used in any cursive script."
          },
          {
            "name": '"ital"',
            "description": "Italics. Applies mostly to Latin; note that many non-Latin fonts contain Latin as well."
          },
          {
            "name": '"jalt"',
            "description": "Justification Alternates. Can be used in any cursive script."
          },
          {
            "name": '"jp78"',
            "description": "JIS78 Forms. Applies only to Japanese."
          },
          {
            "name": '"jp83"',
            "description": "JIS83 Forms. Applies only to Japanese."
          },
          {
            "name": '"jp90"',
            "description": "JIS90 Forms. Applies only to Japanese."
          },
          {
            "name": '"jp04"',
            "description": "JIS2004 Forms. Applies only to Japanese."
          },
          {
            "name": '"kern"',
            "description": "Kerning."
          },
          {
            "name": '"lfbd"',
            "description": "Left Bounds."
          },
          {
            "name": '"liga"',
            "description": "Standard Ligatures."
          },
          {
            "name": '"ljmo"',
            "description": "Leading Jamo Forms. Required for Hangul script when Ancient Hangul writing system is supported."
          },
          {
            "name": '"lnum"',
            "description": "Lining Figures."
          },
          {
            "name": '"locl"',
            "description": "Localized Forms."
          },
          {
            "name": '"ltra"',
            "description": "Left-to-right glyph alternates."
          },
          {
            "name": '"ltrm"',
            "description": "Left-to-right mirrored forms."
          },
          {
            "name": '"mark"',
            "description": "Mark Positioning."
          },
          {
            "name": '"med2"',
            "description": "Medial Form #2. Used only with the Syriac script."
          },
          {
            "name": '"medi"',
            "description": "Medial Forms."
          },
          {
            "name": '"mgrk"',
            "description": "Mathematical Greek."
          },
          {
            "name": '"mkmk"',
            "description": "Mark to Mark Positioning."
          },
          {
            "name": '"nalt"',
            "description": "Alternate Annotation Forms."
          },
          {
            "name": '"nlck"',
            "description": "NLC Kanji Forms. Used only with Kanji script."
          },
          {
            "name": '"nukt"',
            "description": "Nukta Forms. Required in Indic scripts.."
          },
          {
            "name": '"numr"',
            "description": "Numerators."
          },
          {
            "name": '"onum"',
            "description": "Oldstyle Figures."
          },
          {
            "name": '"opbd"',
            "description": "Optical Bounds."
          },
          {
            "name": '"ordn"',
            "description": "Ordinals. Applies mostly to Latin script."
          },
          {
            "name": '"ornm"',
            "description": "Ornaments."
          },
          {
            "name": '"palt"',
            "description": "Proportional Alternate Widths. Used mostly in CJKV fonts."
          },
          {
            "name": '"pcap"',
            "description": "Petite Capitals."
          },
          {
            "name": '"pkna"',
            "description": "Proportional Kana. Generally used only in Japanese fonts."
          },
          {
            "name": '"pnum"',
            "description": "Proportional Figures."
          },
          {
            "name": '"pref"',
            "description": "Pre-base Forms. Required in Khmer and Myanmar (Burmese) scripts and southern Indic scripts that may display a pre-base form of Ra."
          },
          {
            "name": '"pres"',
            "description": "Pre-base Substitutions. Required in Indic scripts."
          },
          {
            "name": '"pstf"',
            "description": "Post-base Forms. Required in scripts of south and southeast Asia that have post-base forms for consonants eg: Gurmukhi, Malayalam, Khmer."
          },
          {
            "name": '"psts"',
            "description": "Post-base Substitutions."
          },
          {
            "name": '"pwid"',
            "description": "Proportional Widths."
          },
          {
            "name": '"qwid"',
            "description": "Quarter Widths. Generally used only in CJKV fonts."
          },
          {
            "name": '"rand"',
            "description": "Randomize."
          },
          {
            "name": '"rclt"',
            "description": "Required Contextual Alternates. May apply to any script, but is especially important for many styles of Arabic."
          },
          {
            "name": '"rlig"',
            "description": "Required Ligatures. Applies to Arabic and Syriac. May apply to some other scripts."
          },
          {
            "name": '"rkrf"',
            "description": "Rakar Forms. Required in Devanagari and Gujarati scripts."
          },
          {
            "name": '"rphf"',
            "description": "Reph Form. Required in Indic scripts. E.g. Devanagari, Kannada."
          },
          {
            "name": '"rtbd"',
            "description": "Right Bounds."
          },
          {
            "name": '"rtla"',
            "description": "Right-to-left alternates."
          },
          {
            "name": '"rtlm"',
            "description": "Right-to-left mirrored forms."
          },
          {
            "name": '"ruby"',
            "description": "Ruby Notation Forms. Applies only to Japanese."
          },
          {
            "name": '"salt"',
            "description": "Stylistic Alternates."
          },
          {
            "name": '"sinf"',
            "description": "Scientific Inferiors."
          },
          {
            "name": '"size"',
            "description": "Optical size."
          },
          {
            "name": '"smcp"',
            "description": "Small Capitals. Applies only to bicameral scripts."
          },
          {
            "name": '"smpl"',
            "description": "Simplified Forms. Applies only to Chinese and Japanese."
          },
          {
            "name": '"ssty"',
            "description": "Math script style alternates."
          },
          {
            "name": '"stch"',
            "description": "Stretching Glyph Decomposition."
          },
          {
            "name": '"subs"',
            "description": "Subscript."
          },
          {
            "name": '"sups"',
            "description": "Superscript."
          },
          {
            "name": '"swsh"',
            "description": "Swash. Does not apply to ideographic scripts."
          },
          {
            "name": '"titl"',
            "description": "Titling."
          },
          {
            "name": '"tjmo"',
            "description": "Trailing Jamo Forms. Required for Hangul script when Ancient Hangul writing system is supported."
          },
          {
            "name": '"tnam"',
            "description": "Traditional Name Forms. Applies only to Japanese."
          },
          {
            "name": '"tnum"',
            "description": "Tabular Figures."
          },
          {
            "name": '"trad"',
            "description": "Traditional Forms. Applies only to Chinese and Japanese."
          },
          {
            "name": '"twid"',
            "description": "Third Widths. Generally used only in CJKV fonts."
          },
          {
            "name": '"unic"',
            "description": "Unicase."
          },
          {
            "name": '"valt"',
            "description": "Alternate Vertical Metrics. Applies only to scripts with vertical writing modes."
          },
          {
            "name": '"vatu"',
            "description": "Vattu Variants. Used for Indic scripts. E.g. Devanagari."
          },
          {
            "name": '"vert"',
            "description": "Vertical Alternates. Applies only to scripts with vertical writing modes."
          },
          {
            "name": '"vhal"',
            "description": "Alternate Vertical Half Metrics. Used only in CJKV fonts."
          },
          {
            "name": '"vjmo"',
            "description": "Vowel Jamo Forms. Required for Hangul script when Ancient Hangul writing system is supported."
          },
          {
            "name": '"vkna"',
            "description": "Vertical Kana Alternates. Applies only to fonts that support kana (hiragana and katakana)."
          },
          {
            "name": '"vkrn"',
            "description": "Vertical Kerning."
          },
          {
            "name": '"vpal"',
            "description": "Proportional Alternate Vertical Metrics. Used mostly in CJKV fonts."
          },
          {
            "name": '"vrt2"',
            "description": "Vertical Alternates and Rotation. Applies only to scripts with vertical writing modes."
          },
          {
            "name": '"zero"',
            "description": "Slashed Zero."
          },
          {
            "name": "normal",
            "description": "No change in glyph substitution or positioning occurs."
          },
          {
            "name": "off",
            "description": "Disable feature."
          },
          {
            "name": "on",
            "description": "Enable feature."
          }
        ],
        "syntax": "normal | <feature-tag-value>#",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-feature-settings"
          }
        ],
        "description": "Provides low-level control over OpenType font features. It is intended as a way of providing access to font features that are not widely used but are needed for a particular use case.",
        "restrictions": [
          "string",
          "integer"
        ]
      },
      {
        "name": "font-kerning",
        "browsers": [
          "E79",
          "FF32",
          "S9",
          "C33",
          "O20"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Specifies that kerning is applied at the discretion of the user agent."
          },
          {
            "name": "none",
            "description": "Specifies that kerning is not applied."
          },
          {
            "name": "normal",
            "description": "Specifies that kerning is applied."
          }
        ],
        "syntax": "auto | normal | none",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-kerning"
          }
        ],
        "description": "Kerning is the contextual adjustment of inter-glyph spacing. This property controls metric kerning, kerning that utilizes adjustment data contained in the font.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-language-override",
        "browsers": [
          "FF34"
        ],
        "values": [
          {
            "name": "normal",
            "description": "Implies that when rendering with OpenType fonts the language of the document is used to infer the OpenType language system, used to select language specific features when rendering."
          }
        ],
        "syntax": "normal | <string>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-language-override"
          }
        ],
        "description": "The value of 'normal' implies that when rendering with OpenType fonts the language of the document is used to infer the OpenType language system, used to select language specific features when rendering.",
        "restrictions": [
          "string"
        ]
      },
      {
        "name": "font-size",
        "values": [
          {
            "name": "large"
          },
          {
            "name": "larger"
          },
          {
            "name": "medium"
          },
          {
            "name": "small"
          },
          {
            "name": "smaller"
          },
          {
            "name": "x-large"
          },
          {
            "name": "x-small"
          },
          {
            "name": "xx-large"
          },
          {
            "name": "xx-small"
          }
        ],
        "syntax": "<absolute-size> | <relative-size> | <length-percentage>",
        "relevance": 94,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-size"
          }
        ],
        "description": "Indicates the desired height of glyphs from the font. For scalable fonts, the font-size is a scale factor applied to the EM unit of the font. (Note that certain glyphs may bleed outside their EM box.) For non-scalable fonts, the font-size is converted into absolute units and matched against the declared font-size of the font, using the same absolute coordinate space for both of the matched values.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "font-size-adjust",
        "browsers": [
          "E79",
          "FF40",
          "C43",
          "O30"
        ],
        "values": [
          {
            "name": "none",
            "description": "Do not preserve the font\u2019s x-height."
          }
        ],
        "syntax": "none | <number>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-size-adjust"
          }
        ],
        "description": "Preserves the readability of text when font fallback occurs by adjusting the font-size so that the x-height is the same regardless of the font used.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "font-stretch",
        "values": [
          {
            "name": "condensed"
          },
          {
            "name": "expanded"
          },
          {
            "name": "extra-condensed"
          },
          {
            "name": "extra-expanded"
          },
          {
            "name": "narrower",
            "description": "Indicates a narrower value relative to the width of the parent element."
          },
          {
            "name": "normal"
          },
          {
            "name": "semi-condensed"
          },
          {
            "name": "semi-expanded"
          },
          {
            "name": "ultra-condensed"
          },
          {
            "name": "ultra-expanded"
          },
          {
            "name": "wider",
            "description": "Indicates a wider value relative to the width of the parent element."
          }
        ],
        "syntax": "<font-stretch-absolute>{1,2}",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-stretch"
          }
        ],
        "description": "Selects a normal, condensed, or expanded face from a font family.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-style",
        "values": [
          {
            "name": "italic",
            "description": "Selects a font that is labeled as an 'italic' face, or an 'oblique' face if one is not"
          },
          {
            "name": "normal",
            "description": "Selects a face that is classified as 'normal'."
          },
          {
            "name": "oblique",
            "description": "Selects a font that is labeled as an 'oblique' face, or an 'italic' face if one is not."
          }
        ],
        "syntax": "normal | italic | oblique <angle>{0,2}",
        "relevance": 83,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-style"
          }
        ],
        "description": "Allows italic or oblique faces to be selected. Italic forms are generally cursive in nature while oblique faces are typically sloped versions of the regular face.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-synthesis",
        "browsers": [
          "FF34",
          "S9"
        ],
        "values": [
          {
            "name": "none",
            "description": "Disallow all synthetic faces."
          },
          {
            "name": "style",
            "description": "Allow synthetic italic faces."
          },
          {
            "name": "weight",
            "description": "Allow synthetic bold faces."
          }
        ],
        "syntax": "none | [ weight || style ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-synthesis"
          }
        ],
        "description": "Controls whether user agents are allowed to synthesize bold or oblique font faces when a font family lacks bold or italic faces.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant",
        "values": [
          {
            "name": "normal",
            "description": "Specifies a face that is not labeled as a small-caps font."
          },
          {
            "name": "small-caps",
            "description": "Specifies a font that is labeled as a small-caps font. If a genuine small-caps font is not available, user agents should simulate a small-caps font."
          }
        ],
        "syntax": "normal | none | [ <common-lig-values> || <discretionary-lig-values> || <historical-lig-values> || <contextual-alt-values> || stylistic(<feature-value-name>) || historical-forms || styleset(<feature-value-name>#) || character-variant(<feature-value-name>#) || swash(<feature-value-name>) || ornaments(<feature-value-name>) || annotation(<feature-value-name>) || [ small-caps | all-small-caps | petite-caps | all-petite-caps | unicase | titling-caps ] || <numeric-figure-values> || <numeric-spacing-values> || <numeric-fraction-values> || ordinal || slashed-zero || <east-asian-variant-values> || <east-asian-width-values> || ruby ]",
        "relevance": 63,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant"
          }
        ],
        "description": "Specifies variant representations of the font",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-alternates",
        "browsers": [
          "FF34"
        ],
        "values": [
          {
            "name": "annotation()",
            "description": "Enables display of alternate annotation forms."
          },
          {
            "name": "character-variant()",
            "description": "Enables display of specific character variants."
          },
          {
            "name": "historical-forms",
            "description": "Enables display of historical forms."
          },
          {
            "name": "normal",
            "description": "None of the features are enabled."
          },
          {
            "name": "ornaments()",
            "description": "Enables replacement of default glyphs with ornaments, if provided in the font."
          },
          {
            "name": "styleset()",
            "description": "Enables display with stylistic sets."
          },
          {
            "name": "stylistic()",
            "description": "Enables display of stylistic alternates."
          },
          {
            "name": "swash()",
            "description": "Enables display of swash glyphs."
          }
        ],
        "syntax": "normal | [ stylistic( <feature-value-name> ) || historical-forms || styleset( <feature-value-name># ) || character-variant( <feature-value-name># ) || swash( <feature-value-name> ) || ornaments( <feature-value-name> ) || annotation( <feature-value-name> ) ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-alternates"
          }
        ],
        "description": "For any given character, fonts can provide a variety of alternate glyphs in addition to the default glyph for that character. This property provides control over the selection of these alternate glyphs.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-caps",
        "browsers": [
          "E79",
          "FF34",
          "C52",
          "O39"
        ],
        "values": [
          {
            "name": "all-petite-caps",
            "description": "Enables display of petite capitals for both upper and lowercase letters."
          },
          {
            "name": "all-small-caps",
            "description": "Enables display of small capitals for both upper and lowercase letters."
          },
          {
            "name": "normal",
            "description": "None of the features are enabled."
          },
          {
            "name": "petite-caps",
            "description": "Enables display of petite capitals."
          },
          {
            "name": "small-caps",
            "description": "Enables display of small capitals. Small-caps glyphs typically use the form of uppercase letters but are reduced to the size of lowercase letters."
          },
          {
            "name": "titling-caps",
            "description": "Enables display of titling capitals."
          },
          {
            "name": "unicase",
            "description": "Enables display of mixture of small capitals for uppercase letters with normal lowercase letters."
          }
        ],
        "syntax": "normal | small-caps | all-small-caps | petite-caps | all-petite-caps | unicase | titling-caps",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-caps"
          }
        ],
        "description": "Specifies control over capitalized forms.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-east-asian",
        "browsers": [
          "E79",
          "FF34",
          "C63",
          "O50"
        ],
        "values": [
          {
            "name": "full-width",
            "description": "Enables rendering of full-width variants."
          },
          {
            "name": "jis04",
            "description": "Enables rendering of JIS04 forms."
          },
          {
            "name": "jis78",
            "description": "Enables rendering of JIS78 forms."
          },
          {
            "name": "jis83",
            "description": "Enables rendering of JIS83 forms."
          },
          {
            "name": "jis90",
            "description": "Enables rendering of JIS90 forms."
          },
          {
            "name": "normal",
            "description": "None of the features are enabled."
          },
          {
            "name": "proportional-width",
            "description": "Enables rendering of proportionally-spaced variants."
          },
          {
            "name": "ruby",
            "description": "Enables display of ruby variant glyphs."
          },
          {
            "name": "simplified",
            "description": "Enables rendering of simplified forms."
          },
          {
            "name": "traditional",
            "description": "Enables rendering of traditional forms."
          }
        ],
        "syntax": "normal | [ <east-asian-variant-values> || <east-asian-width-values> || ruby ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-east-asian"
          }
        ],
        "description": "Allows control of glyph substitute and positioning in East Asian text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-ligatures",
        "browsers": [
          "E79",
          "FF34",
          "S9.1",
          "C34",
          "O21"
        ],
        "values": [
          {
            "name": "additional-ligatures",
            "description": "Enables display of additional ligatures."
          },
          {
            "name": "common-ligatures",
            "description": "Enables display of common ligatures."
          },
          {
            "name": "contextual",
            "browsers": [
              "E79",
              "FF34",
              "S9.1",
              "C34",
              "O21"
            ],
            "description": "Enables display of contextual alternates."
          },
          {
            "name": "discretionary-ligatures",
            "description": "Enables display of discretionary ligatures."
          },
          {
            "name": "historical-ligatures",
            "description": "Enables display of historical ligatures."
          },
          {
            "name": "no-additional-ligatures",
            "description": "Disables display of additional ligatures."
          },
          {
            "name": "no-common-ligatures",
            "description": "Disables display of common ligatures."
          },
          {
            "name": "no-contextual",
            "browsers": [
              "E79",
              "FF34",
              "S9.1",
              "C34",
              "O21"
            ],
            "description": "Disables display of contextual alternates."
          },
          {
            "name": "no-discretionary-ligatures",
            "description": "Disables display of discretionary ligatures."
          },
          {
            "name": "no-historical-ligatures",
            "description": "Disables display of historical ligatures."
          },
          {
            "name": "none",
            "browsers": [
              "E79",
              "FF34",
              "S9.1",
              "C34",
              "O21"
            ],
            "description": "Disables all ligatures."
          },
          {
            "name": "normal",
            "description": "Implies that the defaults set by the font are used."
          }
        ],
        "syntax": "normal | none | [ <common-lig-values> || <discretionary-lig-values> || <historical-lig-values> || <contextual-alt-values> ]",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-ligatures"
          }
        ],
        "description": "Specifies control over which ligatures are enabled or disabled. A value of \u2018normal\u2019 implies that the defaults set by the font are used.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-numeric",
        "browsers": [
          "E79",
          "FF34",
          "S9.1",
          "C52",
          "O39"
        ],
        "values": [
          {
            "name": "diagonal-fractions",
            "description": "Enables display of lining diagonal fractions."
          },
          {
            "name": "lining-nums",
            "description": "Enables display of lining numerals."
          },
          {
            "name": "normal",
            "description": "None of the features are enabled."
          },
          {
            "name": "oldstyle-nums",
            "description": "Enables display of old-style numerals."
          },
          {
            "name": "ordinal",
            "description": "Enables display of letter forms used with ordinal numbers."
          },
          {
            "name": "proportional-nums",
            "description": "Enables display of proportional numerals."
          },
          {
            "name": "slashed-zero",
            "description": "Enables display of slashed zeros."
          },
          {
            "name": "stacked-fractions",
            "description": "Enables display of lining stacked fractions."
          },
          {
            "name": "tabular-nums",
            "description": "Enables display of tabular numerals."
          }
        ],
        "syntax": "normal | [ <numeric-figure-values> || <numeric-spacing-values> || <numeric-fraction-values> || ordinal || slashed-zero ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-numeric"
          }
        ],
        "description": "Specifies control over numerical forms.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-variant-position",
        "browsers": [
          "FF34"
        ],
        "values": [
          {
            "name": "normal",
            "description": "None of the features are enabled."
          },
          {
            "name": "sub",
            "description": "Enables display of subscript variants (OpenType feature: subs)."
          },
          {
            "name": "super",
            "description": "Enables display of superscript variants (OpenType feature: sups)."
          }
        ],
        "syntax": "normal | sub | super",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variant-position"
          }
        ],
        "description": "Specifies the vertical position",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "font-weight",
        "values": [
          {
            "name": "100",
            "description": "Thin"
          },
          {
            "name": "200",
            "description": "Extra Light (Ultra Light)"
          },
          {
            "name": "300",
            "description": "Light"
          },
          {
            "name": "400",
            "description": "Normal"
          },
          {
            "name": "500",
            "description": "Medium"
          },
          {
            "name": "600",
            "description": "Semi Bold (Demi Bold)"
          },
          {
            "name": "700",
            "description": "Bold"
          },
          {
            "name": "800",
            "description": "Extra Bold (Ultra Bold)"
          },
          {
            "name": "900",
            "description": "Black (Heavy)"
          },
          {
            "name": "bold",
            "description": "Same as 700"
          },
          {
            "name": "bolder",
            "description": "Specifies the weight of the face bolder than the inherited value."
          },
          {
            "name": "lighter",
            "description": "Specifies the weight of the face lighter than the inherited value."
          },
          {
            "name": "normal",
            "description": "Same as 400"
          }
        ],
        "syntax": "<font-weight-absolute>{1,2}",
        "relevance": 93,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-weight"
          }
        ],
        "description": "Specifies weight of glyphs in the font, their degree of blackness or stroke thickness.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "glyph-orientation-horizontal",
        "relevance": 50,
        "description": "Controls glyph orientation when the inline-progression-direction is horizontal.",
        "restrictions": [
          "angle",
          "number"
        ]
      },
      {
        "name": "glyph-orientation-vertical",
        "values": [
          {
            "name": "auto",
            "description": "Sets the orientation based on the fullwidth or non-fullwidth characters and the most common orientation."
          }
        ],
        "relevance": 50,
        "description": "Controls glyph orientation when the inline-progression-direction is vertical.",
        "restrictions": [
          "angle",
          "number",
          "enum"
        ]
      },
      {
        "name": "grid-area",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line> [ / <grid-line> ]{0,3}",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-area"
          }
        ],
        "description": "Determine a grid item\u2019s size and location within the grid by contributing a line, a span, or nothing (automatic) to its grid placement. Shorthand for 'grid-row-start', 'grid-column-start', 'grid-row-end', and 'grid-column-end'.",
        "restrictions": [
          "identifier",
          "integer"
        ]
      },
      {
        "name": "grid",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "syntax": "<'grid-template'> | <'grid-template-rows'> / [ auto-flow && dense? ] <'grid-auto-columns'>? | [ auto-flow && dense? ] <'grid-auto-rows'>? / <'grid-template-columns'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid"
          }
        ],
        "description": "The grid CSS property is a shorthand property that sets all of the explicit grid properties ('grid-template-rows', 'grid-template-columns', and 'grid-template-areas'), and all the implicit grid properties ('grid-auto-rows', 'grid-auto-columns', and 'grid-auto-flow'), in a single declaration.",
        "restrictions": [
          "identifier",
          "length",
          "percentage",
          "string",
          "enum"
        ]
      },
      {
        "name": "grid-auto-columns",
        "values": [
          {
            "name": "min-content",
            "description": "Represents the largest min-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "max-content",
            "description": "Represents the largest max-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "auto",
            "description": "As a maximum, identical to 'max-content'. As a minimum, represents the largest minimum size (as specified by min-width/min-height) of the grid items occupying the grid track."
          },
          {
            "name": "minmax()",
            "description": "Defines a size range greater than or equal to min and less than or equal to max."
          }
        ],
        "syntax": "<track-size>+",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-auto-columns"
          }
        ],
        "description": "Specifies the size of implicitly created columns.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "grid-auto-flow",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "row",
            "description": "The auto-placement algorithm places items by filling each row in turn, adding new rows as necessary."
          },
          {
            "name": "column",
            "description": "The auto-placement algorithm places items by filling each column in turn, adding new columns as necessary."
          },
          {
            "name": "dense",
            "description": "If specified, the auto-placement algorithm uses a \u201Cdense\u201D packing algorithm, which attempts to fill in holes earlier in the grid if smaller items come up later."
          }
        ],
        "syntax": "[ row | column ] || dense",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-auto-flow"
          }
        ],
        "description": "Controls how the auto-placement algorithm works, specifying exactly how auto-placed items get flowed into the grid.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "grid-auto-rows",
        "values": [
          {
            "name": "min-content",
            "description": "Represents the largest min-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "max-content",
            "description": "Represents the largest max-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "auto",
            "description": "As a maximum, identical to 'max-content'. As a minimum, represents the largest minimum size (as specified by min-width/min-height) of the grid items occupying the grid track."
          },
          {
            "name": "minmax()",
            "description": "Defines a size range greater than or equal to min and less than or equal to max."
          }
        ],
        "syntax": "<track-size>+",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-auto-rows"
          }
        ],
        "description": "Specifies the size of implicitly created rows.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "grid-column",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line> [ / <grid-line> ]?",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-column"
          }
        ],
        "description": "Shorthand for 'grid-column-start' and 'grid-column-end'.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-column-end",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-column-end"
          }
        ],
        "description": "Determine a grid item\u2019s size and location within the grid by contributing a line, a span, or nothing (automatic) to its grid placement.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-column-gap",
        "browsers": [
          "FF52",
          "C57",
          "S10.1",
          "O44"
        ],
        "status": "obsolete",
        "syntax": "<length-percentage>",
        "relevance": 1,
        "description": "Specifies the gutters between grid columns. Replaced by 'column-gap' property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "grid-column-start",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-column-start"
          }
        ],
        "description": "Determine a grid item\u2019s size and location within the grid by contributing a line, a span, or nothing (automatic) to its grid placement.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-gap",
        "browsers": [
          "FF52",
          "C57",
          "S10.1",
          "O44"
        ],
        "status": "obsolete",
        "syntax": "<'grid-row-gap'> <'grid-column-gap'>?",
        "relevance": 1,
        "description": "Shorthand that specifies the gutters between grid columns and grid rows in one declaration. Replaced by 'gap' property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "grid-row",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line> [ / <grid-line> ]?",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-row"
          }
        ],
        "description": "Shorthand for 'grid-row-start' and 'grid-row-end'.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-row-end",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-row-end"
          }
        ],
        "description": "Determine a grid item\u2019s size and location within the grid by contributing a line, a span, or nothing (automatic) to its grid placement.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-row-gap",
        "browsers": [
          "FF52",
          "C57",
          "S10.1",
          "O44"
        ],
        "status": "obsolete",
        "syntax": "<length-percentage>",
        "relevance": 1,
        "description": "Specifies the gutters between grid rows. Replaced by 'row-gap' property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "grid-row-start",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The property contributes nothing to the grid item\u2019s placement, indicating auto-placement, an automatic span, or a default span of one."
          },
          {
            "name": "span",
            "description": "Contributes a grid span to the grid item\u2019s placement such that the corresponding edge of the grid item\u2019s grid area is N lines from its opposite edge."
          }
        ],
        "syntax": "<grid-line>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-row-start"
          }
        ],
        "description": "Determine a grid item\u2019s size and location within the grid by contributing a line, a span, or nothing (automatic) to its grid placement.",
        "restrictions": [
          "identifier",
          "integer",
          "enum"
        ]
      },
      {
        "name": "grid-template",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "Sets all three properties to their initial values."
          },
          {
            "name": "min-content",
            "description": "Represents the largest min-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "max-content",
            "description": "Represents the largest max-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "auto",
            "description": "As a maximum, identical to 'max-content'. As a minimum, represents the largest minimum size (as specified by min-width/min-height) of the grid items occupying the grid track."
          },
          {
            "name": "subgrid",
            "description": "Sets 'grid-template-rows' and 'grid-template-columns' to 'subgrid', and 'grid-template-areas' to its initial value."
          },
          {
            "name": "minmax()",
            "description": "Defines a size range greater than or equal to min and less than or equal to max."
          },
          {
            "name": "repeat()",
            "description": "Represents a repeated fragment of the track list, allowing a large number of columns or rows that exhibit a recurring pattern to be written in a more compact form."
          }
        ],
        "syntax": "none | [ <'grid-template-rows'> / <'grid-template-columns'> ] | [ <line-names>? <string> <track-size>? <line-names>? ]+ [ / <explicit-track-list> ]?",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-template"
          }
        ],
        "description": "Shorthand for setting grid-template-columns, grid-template-rows, and grid-template-areas in a single declaration.",
        "restrictions": [
          "identifier",
          "length",
          "percentage",
          "string",
          "enum"
        ]
      },
      {
        "name": "grid-template-areas",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "The grid container doesn\u2019t define any named grid areas."
          }
        ],
        "syntax": "none | <string>+",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-template-areas"
          }
        ],
        "description": "Specifies named grid areas, which are not associated with any particular grid item, but can be referenced from the grid-placement properties.",
        "restrictions": [
          "string"
        ]
      },
      {
        "name": "grid-template-columns",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "There is no explicit grid; any rows/columns will be implicitly generated."
          },
          {
            "name": "min-content",
            "description": "Represents the largest min-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "max-content",
            "description": "Represents the largest max-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "auto",
            "description": "As a maximum, identical to 'max-content'. As a minimum, represents the largest minimum size (as specified by min-width/min-height) of the grid items occupying the grid track."
          },
          {
            "name": "subgrid",
            "description": "Indicates that the grid will align to its parent grid in that axis."
          },
          {
            "name": "minmax()",
            "description": "Defines a size range greater than or equal to min and less than or equal to max."
          },
          {
            "name": "repeat()",
            "description": "Represents a repeated fragment of the track list, allowing a large number of columns or rows that exhibit a recurring pattern to be written in a more compact form."
          }
        ],
        "syntax": "none | <track-list> | <auto-track-list> | subgrid <line-name-list>?",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-template-columns"
          }
        ],
        "description": "specifies, as a space-separated track list, the line names and track sizing functions of the grid.",
        "restrictions": [
          "identifier",
          "length",
          "percentage",
          "enum"
        ]
      },
      {
        "name": "grid-template-rows",
        "browsers": [
          "E16",
          "FF52",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "There is no explicit grid; any rows/columns will be implicitly generated."
          },
          {
            "name": "min-content",
            "description": "Represents the largest min-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "max-content",
            "description": "Represents the largest max-content contribution of the grid items occupying the grid track."
          },
          {
            "name": "auto",
            "description": "As a maximum, identical to 'max-content'. As a minimum, represents the largest minimum size (as specified by min-width/min-height) of the grid items occupying the grid track."
          },
          {
            "name": "subgrid",
            "description": "Indicates that the grid will align to its parent grid in that axis."
          },
          {
            "name": "minmax()",
            "description": "Defines a size range greater than or equal to min and less than or equal to max."
          },
          {
            "name": "repeat()",
            "description": "Represents a repeated fragment of the track list, allowing a large number of columns or rows that exhibit a recurring pattern to be written in a more compact form."
          }
        ],
        "syntax": "none | <track-list> | <auto-track-list> | subgrid <line-name-list>?",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/grid-template-rows"
          }
        ],
        "description": "specifies, as a space-separated track list, the line names and track sizing functions of the grid.",
        "restrictions": [
          "identifier",
          "length",
          "percentage",
          "string",
          "enum"
        ]
      },
      {
        "name": "height",
        "values": [
          {
            "name": "auto",
            "description": "The height depends on the values of other properties."
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>{1,2}",
        "relevance": 96,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/height"
          }
        ],
        "description": "Specifies the height of the content area, padding area or border area (depending on 'box-sizing') of certain boxes.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "hyphens",
        "values": [
          {
            "name": "auto",
            "description": "Conditional hyphenation characters inside a word, if present, take priority over automatic resources when determining hyphenation points within the word."
          },
          {
            "name": "manual",
            "description": "Words are only broken at line breaks where there are characters inside the word that suggest line break opportunities"
          },
          {
            "name": "none",
            "description": "Words are not broken at line breaks, even if characters inside the word suggest line break points."
          }
        ],
        "syntax": "none | manual | auto",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/hyphens"
          }
        ],
        "description": "Controls whether hyphenation is allowed to create more break opportunities within a line of text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "image-orientation",
        "browsers": [
          "E81",
          "FF26",
          "S13.1",
          "C81",
          "O67"
        ],
        "values": [
          {
            "name": "flip",
            "description": "After rotating by the precededing angle, the image is flipped horizontally. Defaults to 0deg if the angle is ommitted."
          },
          {
            "name": "from-image",
            "description": "If the image has an orientation specified in its metadata, such as EXIF, this value computes to the angle that the metadata specifies is necessary to correctly orient the image."
          }
        ],
        "syntax": "from-image | <angle> | [ <angle>? flip ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/image-orientation"
          }
        ],
        "description": "Specifies an orthogonal rotation to be applied to an image before it is laid out.",
        "restrictions": [
          "angle"
        ]
      },
      {
        "name": "image-rendering",
        "browsers": [
          "E79",
          "FF3.6",
          "S6",
          "C13",
          "O15"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The image should be scaled with an algorithm that maximizes the appearance of the image."
          },
          {
            "name": "crisp-edges",
            "description": "The image must be scaled with an algorithm that preserves contrast and edges in the image, and which does not smooth colors or introduce blur to the image in the process."
          },
          {
            "name": "-moz-crisp-edges",
            "browsers": [
              "E79",
              "FF3.6",
              "S6",
              "C13",
              "O15"
            ]
          },
          {
            "name": "optimizeQuality",
            "description": "Deprecated."
          },
          {
            "name": "optimizeSpeed",
            "description": "Deprecated."
          },
          {
            "name": "pixelated",
            "description": "When scaling the image up, the 'nearest neighbor' or similar algorithm must be used, so that the image appears to be simply composed of very large pixels."
          }
        ],
        "syntax": "auto | crisp-edges | pixelated",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/image-rendering"
          }
        ],
        "description": "Provides a hint to the user-agent about what aspects of an image are most important to preserve when the image is scaled, to aid the user-agent in the choice of an appropriate scaling algorithm.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "ime-mode",
        "browsers": [
          "E12",
          "FF3",
          "IE5"
        ],
        "values": [
          {
            "name": "active",
            "description": "The input method editor is initially active; text entry is performed using it unless the user specifically dismisses it."
          },
          {
            "name": "auto",
            "description": "No change is made to the current input method editor state. This is the default."
          },
          {
            "name": "disabled",
            "description": "The input method editor is disabled and may not be activated by the user."
          },
          {
            "name": "inactive",
            "description": "The input method editor is initially inactive, but the user may activate it if they wish."
          },
          {
            "name": "normal",
            "description": "The IME state should be normal; this value can be used in a user style sheet to override the page setting."
          }
        ],
        "status": "obsolete",
        "syntax": "auto | normal | active | inactive | disabled",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/ime-mode"
          }
        ],
        "description": "Controls the state of the input method editor for text fields.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "inline-size",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Depends on the values of other properties."
          }
        ],
        "syntax": "<'width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inline-size"
          }
        ],
        "description": "Logical 'height'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "isolation",
        "browsers": [
          "E79",
          "FF36",
          "S8",
          "C41",
          "O30"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Elements are not isolated unless an operation is applied that causes the creation of a stacking context."
          },
          {
            "name": "isolate",
            "description": "In CSS will turn the element into a stacking context."
          }
        ],
        "syntax": "auto | isolate",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/isolation"
          }
        ],
        "description": "In CSS setting to 'isolate' will turn the element into a stacking context. In SVG, it defines whether an element is isolated or not.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "justify-content",
        "values": [
          {
            "name": "center",
            "description": "Flex items are packed toward the center of the line."
          },
          {
            "name": "start",
            "description": "The items are packed flush to each other toward the start edge of the alignment container in the main axis."
          },
          {
            "name": "end",
            "description": "The items are packed flush to each other toward the end edge of the alignment container in the main axis."
          },
          {
            "name": "left",
            "description": "The items are packed flush to each other toward the left edge of the alignment container in the main axis."
          },
          {
            "name": "right",
            "description": "The items are packed flush to each other toward the right edge of the alignment container in the main axis."
          },
          {
            "name": "safe",
            "description": "If the size of the item overflows the alignment container, the item is instead aligned as if the alignment mode were start."
          },
          {
            "name": "unsafe",
            "description": "Regardless of the relative sizes of the item and alignment container, the given alignment value is honored."
          },
          {
            "name": "stretch",
            "description": "If the combined size of the alignment subjects is less than the size of the alignment container, any auto-sized alignment subjects have their size increased equally (not proportionally), while still respecting the constraints imposed by max-height/max-width (or equivalent functionality), so that the combined size exactly fills the alignment container."
          },
          {
            "name": "space-evenly",
            "description": "The items are evenly distributed within the alignment container along the main axis."
          },
          {
            "name": "flex-end",
            "description": "Flex items are packed toward the end of the line."
          },
          {
            "name": "flex-start",
            "description": "Flex items are packed toward the start of the line."
          },
          {
            "name": "space-around",
            "description": "Flex items are evenly distributed in the line, with half-size spaces on either end."
          },
          {
            "name": "space-between",
            "description": "Flex items are evenly distributed in the line."
          },
          {
            "name": "baseline",
            "description": "Specifies participation in first-baseline alignment."
          },
          {
            "name": "first baseline",
            "description": "Specifies participation in first-baseline alignment."
          },
          {
            "name": "last baseline",
            "description": "Specifies participation in last-baseline alignment."
          }
        ],
        "syntax": "normal | <content-distribution> | <overflow-position>? [ <content-position> | left | right ]",
        "relevance": 82,
        "description": "Aligns flex items along the main axis of the current line of the flex container.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "kerning",
        "values": [
          {
            "name": "auto",
            "description": "Indicates that the user agent should adjust inter-glyph spacing based on kerning tables that are included in the font that will be used."
          }
        ],
        "relevance": 50,
        "description": "Indicates whether the user agent should adjust inter-glyph spacing based on kerning tables that are included in the relevant font or instead disable auto-kerning and set inter-character spacing to a specific length.",
        "restrictions": [
          "length",
          "enum"
        ]
      },
      {
        "name": "left",
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 95,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/left"
          }
        ],
        "description": "Specifies how far an absolutely positioned box's left margin edge is offset to the right of the left edge of the box's 'containing block'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "letter-spacing",
        "values": [
          {
            "name": "normal",
            "description": "The spacing is the normal spacing for the current font. It is typically zero-length."
          }
        ],
        "syntax": "normal | <length>",
        "relevance": 79,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/letter-spacing"
          }
        ],
        "description": "Specifies the minimum, maximum, and optimal spacing between grapheme clusters.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "lighting-color",
        "browsers": [
          "E",
          "C5",
          "FF3",
          "IE10",
          "O9",
          "S6"
        ],
        "relevance": 50,
        "description": "Defines the color of the light source for filter primitives 'feDiffuseLighting' and 'feSpecularLighting'.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "line-break",
        "values": [
          {
            "name": "auto",
            "description": "The UA determines the set of line-breaking restrictions to use for CJK scripts, and it may vary the restrictions based on the length of the line; e.g., use a less restrictive set of line-break rules for short lines."
          },
          {
            "name": "loose",
            "description": "Breaks text using the least restrictive set of line-breaking rules. Typically used for short lines, such as in newspapers."
          },
          {
            "name": "normal",
            "description": "Breaks text using the most common set of line-breaking rules."
          },
          {
            "name": "strict",
            "description": "Breaks CJK scripts using a more restrictive set of line-breaking rules than 'normal'."
          }
        ],
        "syntax": "auto | loose | normal | strict | anywhere",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/line-break"
          }
        ],
        "description": "Specifies what set of line breaking restrictions are in effect within the element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "line-height",
        "values": [
          {
            "name": "normal",
            "description": "Tells user agents to set the computed value to a 'reasonable' value based on the font size of the element."
          }
        ],
        "syntax": "normal | <number> | <length> | <percentage>",
        "relevance": 92,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/line-height"
          }
        ],
        "description": "Determines the block-progression dimension of the text content area of an inline box.",
        "restrictions": [
          "number",
          "length",
          "percentage"
        ]
      },
      {
        "name": "list-style",
        "values": [
          {
            "name": "armenian"
          },
          {
            "name": "circle",
            "description": "A hollow circle."
          },
          {
            "name": "decimal"
          },
          {
            "name": "decimal-leading-zero"
          },
          {
            "name": "disc",
            "description": "A filled circle."
          },
          {
            "name": "georgian"
          },
          {
            "name": "inside",
            "description": "The marker box is outside the principal block box, as described in the section on the ::marker pseudo-element below."
          },
          {
            "name": "lower-alpha"
          },
          {
            "name": "lower-greek"
          },
          {
            "name": "lower-latin"
          },
          {
            "name": "lower-roman"
          },
          {
            "name": "none"
          },
          {
            "name": "outside",
            "description": "The ::marker pseudo-element is an inline element placed immediately before all ::before pseudo-elements in the principal block box, after which the element's content flows."
          },
          {
            "name": "square",
            "description": "A filled square."
          },
          {
            "name": "symbols()",
            "description": "Allows a counter style to be defined inline."
          },
          {
            "name": "upper-alpha"
          },
          {
            "name": "upper-latin"
          },
          {
            "name": "upper-roman"
          },
          {
            "name": "url()"
          }
        ],
        "syntax": "<'list-style-type'> || <'list-style-position'> || <'list-style-image'>",
        "relevance": 84,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/list-style"
          }
        ],
        "description": "Shorthand for setting 'list-style-type', 'list-style-position' and 'list-style-image'",
        "restrictions": [
          "image",
          "enum",
          "url"
        ]
      },
      {
        "name": "list-style-image",
        "values": [
          {
            "name": "none",
            "description": "The default contents of the of the list item\u2019s marker are given by 'list-style-type' instead."
          }
        ],
        "syntax": "<url> | none",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/list-style-image"
          }
        ],
        "description": "Sets the image that will be used as the list item marker. When the image is available, it will replace the marker set with the 'list-style-type' marker.",
        "restrictions": [
          "image"
        ]
      },
      {
        "name": "list-style-position",
        "values": [
          {
            "name": "inside",
            "description": "The marker box is outside the principal block box, as described in the section on the ::marker pseudo-element below."
          },
          {
            "name": "outside",
            "description": "The ::marker pseudo-element is an inline element placed immediately before all ::before pseudo-elements in the principal block box, after which the element's content flows."
          }
        ],
        "syntax": "inside | outside",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/list-style-position"
          }
        ],
        "description": "Specifies the position of the '::marker' pseudo-element's box in the list item.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "list-style-type",
        "values": [
          {
            "name": "armenian",
            "description": "Traditional uppercase Armenian numbering."
          },
          {
            "name": "circle",
            "description": "A hollow circle."
          },
          {
            "name": "decimal",
            "description": "Western decimal numbers."
          },
          {
            "name": "decimal-leading-zero",
            "description": "Decimal numbers padded by initial zeros."
          },
          {
            "name": "disc",
            "description": "A filled circle."
          },
          {
            "name": "georgian",
            "description": "Traditional Georgian numbering."
          },
          {
            "name": "lower-alpha",
            "description": "Lowercase ASCII letters."
          },
          {
            "name": "lower-greek",
            "description": "Lowercase classical Greek."
          },
          {
            "name": "lower-latin",
            "description": "Lowercase ASCII letters."
          },
          {
            "name": "lower-roman",
            "description": "Lowercase ASCII Roman numerals."
          },
          {
            "name": "none",
            "description": "No marker"
          },
          {
            "name": "square",
            "description": "A filled square."
          },
          {
            "name": "symbols()",
            "description": "Allows a counter style to be defined inline."
          },
          {
            "name": "upper-alpha",
            "description": "Uppercase ASCII letters."
          },
          {
            "name": "upper-latin",
            "description": "Uppercase ASCII letters."
          },
          {
            "name": "upper-roman",
            "description": "Uppercase ASCII Roman numerals."
          }
        ],
        "syntax": "<counter-style> | <string> | none",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/list-style-type"
          }
        ],
        "description": "Used to construct the default contents of a list item\u2019s marker",
        "restrictions": [
          "enum",
          "string"
        ]
      },
      {
        "name": "margin",
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "[ <length> | <percentage> | auto ]{1,4}",
        "relevance": 95,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin"
          }
        ],
        "description": "Shorthand property to set values the thickness of the margin area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. Negative values for margin properties are allowed, but there may be implementation-specific limits.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-block-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<'margin-left'>",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-block-end"
          }
        ],
        "description": "Logical 'margin-bottom'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-block-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<'margin-left'>",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-block-start"
          }
        ],
        "description": "Logical 'margin-top'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-bottom",
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-bottom"
          }
        ],
        "description": "Shorthand property to set values the thickness of the margin area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. Negative values for margin properties are allowed, but there may be implementation-specific limits..",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-inline-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<'margin-left'>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-inline-end"
          }
        ],
        "description": "Logical 'margin-right'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-inline-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<'margin-left'>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-inline-start"
          }
        ],
        "description": "Logical 'margin-left'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-left",
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-left"
          }
        ],
        "description": "Shorthand property to set values the thickness of the margin area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. Negative values for margin properties are allowed, but there may be implementation-specific limits..",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-right",
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 90,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-right"
          }
        ],
        "description": "Shorthand property to set values the thickness of the margin area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. Negative values for margin properties are allowed, but there may be implementation-specific limits..",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "margin-top",
        "values": [
          {
            "name": "auto"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 95,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-top"
          }
        ],
        "description": "Shorthand property to set values the thickness of the margin area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. Negative values for margin properties are allowed, but there may be implementation-specific limits..",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "marker",
        "values": [
          {
            "name": "none",
            "description": "Indicates that no marker symbol will be drawn at the given vertex or vertices."
          },
          {
            "name": "url()",
            "description": "Indicates that the <marker> element referenced will be used."
          }
        ],
        "relevance": 50,
        "description": "Specifies the marker symbol that shall be used for all points on the sets the value for all vertices on the given \u2018path\u2019 element or basic shape.",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "marker-end",
        "values": [
          {
            "name": "none",
            "description": "Indicates that no marker symbol will be drawn at the given vertex or vertices."
          },
          {
            "name": "url()",
            "description": "Indicates that the <marker> element referenced will be used."
          }
        ],
        "relevance": 50,
        "description": "Specifies the marker that will be drawn at the last vertices of the given markable element.",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "marker-mid",
        "values": [
          {
            "name": "none",
            "description": "Indicates that no marker symbol will be drawn at the given vertex or vertices."
          },
          {
            "name": "url()",
            "description": "Indicates that the <marker> element referenced will be used."
          }
        ],
        "relevance": 50,
        "description": "Specifies the marker that will be drawn at all vertices except the first and last.",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "marker-start",
        "values": [
          {
            "name": "none",
            "description": "Indicates that no marker symbol will be drawn at the given vertex or vertices."
          },
          {
            "name": "url()",
            "description": "Indicates that the <marker> element referenced will be used."
          }
        ],
        "relevance": 50,
        "description": "Specifies the marker that will be drawn at the first vertices of the given markable element.",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "mask-image",
        "browsers": [
          "E16",
          "FF53",
          "S4",
          "C1",
          "O15"
        ],
        "values": [
          {
            "name": "none",
            "description": "Counts as a transparent black image layer."
          },
          {
            "name": "url()",
            "description": "Reference to a <mask element or to a CSS image."
          }
        ],
        "syntax": "<mask-reference>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-image"
          }
        ],
        "description": "Sets the mask layer image of an element.",
        "restrictions": [
          "url",
          "image",
          "enum"
        ]
      },
      {
        "name": "mask-mode",
        "browsers": [
          "FF53"
        ],
        "values": [
          {
            "name": "alpha",
            "description": "Alpha values of the mask layer image should be used as the mask values."
          },
          {
            "name": "auto",
            "description": "Use alpha values if 'mask-image' is an image, luminance if a <mask> element or a CSS image."
          },
          {
            "name": "luminance",
            "description": "Luminance values of the mask layer image should be used as the mask values."
          }
        ],
        "syntax": "<masking-mode>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-mode"
          }
        ],
        "description": "Indicates whether the mask layer image is treated as luminance mask or alpha mask.",
        "restrictions": [
          "url",
          "image",
          "enum"
        ]
      },
      {
        "name": "mask-origin",
        "browsers": [
          "E79",
          "FF53",
          "S4",
          "C1",
          "O15"
        ],
        "syntax": "<geometry-box>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-origin"
          }
        ],
        "description": "Specifies the mask positioning area.",
        "restrictions": [
          "geometry-box",
          "enum"
        ]
      },
      {
        "name": "mask-position",
        "browsers": [
          "E18",
          "FF53",
          "S3.2",
          "C1",
          "O15"
        ],
        "syntax": "<position>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-position"
          }
        ],
        "description": "Specifies how mask layer images are positioned.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "mask-repeat",
        "browsers": [
          "E18",
          "FF53",
          "S3.2",
          "C1",
          "O15"
        ],
        "syntax": "<repeat-style>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-repeat"
          }
        ],
        "description": "Specifies how mask layer images are tiled after they have been sized and positioned.",
        "restrictions": [
          "repeat"
        ]
      },
      {
        "name": "mask-size",
        "browsers": [
          "E18",
          "FF53",
          "S4",
          "C4",
          "O15"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Resolved by using the image\u2019s intrinsic ratio and the size of the other dimension, or failing that, using the image\u2019s intrinsic size, or failing that, treating it as 100%."
          },
          {
            "name": "contain",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the largest size such that both its width and its height can fit inside the background positioning area."
          },
          {
            "name": "cover",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the smallest size such that both its width and its height can completely cover the background positioning area."
          }
        ],
        "syntax": "<bg-size>#",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-size"
          }
        ],
        "description": "Specifies the size of the mask layer images.",
        "restrictions": [
          "length",
          "percentage",
          "enum"
        ]
      },
      {
        "name": "mask-type",
        "browsers": [
          "E79",
          "FF35",
          "S6.1",
          "C24",
          "O15"
        ],
        "values": [
          {
            "name": "alpha",
            "description": "Indicates that the alpha values of the mask should be used."
          },
          {
            "name": "luminance",
            "description": "Indicates that the luminance values of the mask should be used."
          }
        ],
        "syntax": "luminance | alpha",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-type"
          }
        ],
        "description": "Defines whether the content of the <mask> element is treated as as luminance mask or alpha mask.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "max-block-size",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "No limit on the width of the box."
          }
        ],
        "syntax": "<'max-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/max-block-size"
          }
        ],
        "description": "Logical 'max-width'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "max-height",
        "values": [
          {
            "name": "none",
            "description": "No limit on the height of the box."
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>",
        "relevance": 84,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/max-height"
          }
        ],
        "description": "Allows authors to constrain content height to a certain range.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "max-inline-size",
        "browsers": [
          "E79",
          "FF41",
          "S10.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "none",
            "description": "No limit on the height of the box."
          }
        ],
        "syntax": "<'max-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/max-inline-size"
          }
        ],
        "description": "Logical 'max-height'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "max-width",
        "values": [
          {
            "name": "none",
            "description": "No limit on the width of the box."
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>",
        "relevance": 89,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/max-width"
          }
        ],
        "description": "Allows authors to constrain content width to a certain range.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "min-block-size",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C57",
          "O44"
        ],
        "syntax": "<'min-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/min-block-size"
          }
        ],
        "description": "Logical 'min-width'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "min-height",
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/min-height"
          }
        ],
        "description": "Allows authors to constrain content height to a certain range.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "min-inline-size",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C57",
          "O44"
        ],
        "syntax": "<'min-width'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/min-inline-size"
          }
        ],
        "description": "Logical 'min-height'. Mapping depends on the element\u2019s 'writing-mode'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "min-width",
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>",
        "relevance": 87,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/min-width"
          }
        ],
        "description": "Allows authors to constrain content width to a certain range.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "mix-blend-mode",
        "browsers": [
          "E79",
          "FF32",
          "S8",
          "C41",
          "O28"
        ],
        "values": [
          {
            "name": "normal",
            "description": "Default attribute which specifies no blending"
          },
          {
            "name": "multiply",
            "description": "The source color is multiplied by the destination color and replaces the destination."
          },
          {
            "name": "screen",
            "description": "Multiplies the complements of the backdrop and source color values, then complements the result."
          },
          {
            "name": "overlay",
            "description": "Multiplies or screens the colors, depending on the backdrop color value."
          },
          {
            "name": "darken",
            "description": "Selects the darker of the backdrop and source colors."
          },
          {
            "name": "lighten",
            "description": "Selects the lighter of the backdrop and source colors."
          },
          {
            "name": "color-dodge",
            "description": "Brightens the backdrop color to reflect the source color."
          },
          {
            "name": "color-burn",
            "description": "Darkens the backdrop color to reflect the source color."
          },
          {
            "name": "hard-light",
            "description": "Multiplies or screens the colors, depending on the source color value."
          },
          {
            "name": "soft-light",
            "description": "Darkens or lightens the colors, depending on the source color value."
          },
          {
            "name": "difference",
            "description": "Subtracts the darker of the two constituent colors from the lighter color.."
          },
          {
            "name": "exclusion",
            "description": "Produces an effect similar to that of the Difference mode but lower in contrast."
          },
          {
            "name": "hue",
            "browsers": [
              "E79",
              "FF32",
              "S8",
              "C41",
              "O28"
            ],
            "description": "Creates a color with the hue of the source color and the saturation and luminosity of the backdrop color."
          },
          {
            "name": "saturation",
            "browsers": [
              "E79",
              "FF32",
              "S8",
              "C41",
              "O28"
            ],
            "description": "Creates a color with the saturation of the source color and the hue and luminosity of the backdrop color."
          },
          {
            "name": "color",
            "browsers": [
              "E79",
              "FF32",
              "S8",
              "C41",
              "O28"
            ],
            "description": "Creates a color with the hue and saturation of the source color and the luminosity of the backdrop color."
          },
          {
            "name": "luminosity",
            "browsers": [
              "E79",
              "FF32",
              "S8",
              "C41",
              "O28"
            ],
            "description": "Creates a color with the luminosity of the source color and the hue and saturation of the backdrop color."
          }
        ],
        "syntax": "<blend-mode>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mix-blend-mode"
          }
        ],
        "description": "Defines the formula that must be used to mix the colors with the backdrop.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "motion",
        "browsers": [
          "C46",
          "O33"
        ],
        "values": [
          {
            "name": "none",
            "description": "No motion path gets created."
          },
          {
            "name": "path()",
            "description": "Defines an SVG path as a string, with optional 'fill-rule' as the first argument."
          },
          {
            "name": "auto",
            "description": "Indicates that the object is rotated by the angle of the direction of the motion path."
          },
          {
            "name": "reverse",
            "description": "Indicates that the object is rotated by the angle of the direction of the motion path plus 180 degrees."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property for setting 'motion-path', 'motion-offset' and 'motion-rotation'.",
        "restrictions": [
          "url",
          "length",
          "percentage",
          "angle",
          "shape",
          "geometry-box",
          "enum"
        ]
      },
      {
        "name": "motion-offset",
        "browsers": [
          "C46",
          "O33"
        ],
        "relevance": 50,
        "description": "A distance that describes the position along the specified motion path.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "motion-path",
        "browsers": [
          "C46",
          "O33"
        ],
        "values": [
          {
            "name": "none",
            "description": "No motion path gets created."
          },
          {
            "name": "path()",
            "description": "Defines an SVG path as a string, with optional 'fill-rule' as the first argument."
          }
        ],
        "relevance": 50,
        "description": "Specifies the motion path the element gets positioned at.",
        "restrictions": [
          "url",
          "shape",
          "geometry-box",
          "enum"
        ]
      },
      {
        "name": "motion-rotation",
        "browsers": [
          "C46",
          "O33"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Indicates that the object is rotated by the angle of the direction of the motion path."
          },
          {
            "name": "reverse",
            "description": "Indicates that the object is rotated by the angle of the direction of the motion path plus 180 degrees."
          }
        ],
        "relevance": 50,
        "description": "Defines the direction of the element while positioning along the motion path.",
        "restrictions": [
          "angle"
        ]
      },
      {
        "name": "-moz-animation",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          },
          {
            "name": "none",
            "description": "No animation is performed"
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines six of the animation properties into a single property.",
        "restrictions": [
          "time",
          "enum",
          "timing-function",
          "identifier",
          "number"
        ]
      },
      {
        "name": "-moz-animation-delay",
        "browsers": [
          "FF9"
        ],
        "relevance": 50,
        "description": "Defines when the animation will start.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-moz-animation-direction",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Defines whether or not the animation should play in reverse on alternate cycles.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-animation-duration",
        "browsers": [
          "FF9"
        ],
        "relevance": 50,
        "description": "Defines the length of time that an animation takes to complete one cycle.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-moz-animation-iteration-count",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          }
        ],
        "relevance": 50,
        "description": "Defines the number of times an animation cycle is played. The default value is one, meaning the animation will play from beginning to end once.",
        "restrictions": [
          "number",
          "enum"
        ]
      },
      {
        "name": "-moz-animation-name",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "none",
            "description": "No animation is performed"
          }
        ],
        "relevance": 50,
        "description": "Defines a list of animations that apply. Each name is used to select the keyframe at-rule that provides the property values for the animation.",
        "restrictions": [
          "identifier",
          "enum"
        ]
      },
      {
        "name": "-moz-animation-play-state",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "paused",
            "description": "A running animation will be paused."
          },
          {
            "name": "running",
            "description": "Resume playback of a paused animation."
          }
        ],
        "relevance": 50,
        "description": "Defines whether the animation is running or paused.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-animation-timing-function",
        "browsers": [
          "FF9"
        ],
        "relevance": 50,
        "description": "Describes how the animation will progress over one cycle of its duration. See the 'transition-timing-function'.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "-moz-appearance",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "button"
          },
          {
            "name": "button-arrow-down"
          },
          {
            "name": "button-arrow-next"
          },
          {
            "name": "button-arrow-previous"
          },
          {
            "name": "button-arrow-up"
          },
          {
            "name": "button-bevel"
          },
          {
            "name": "checkbox"
          },
          {
            "name": "checkbox-container"
          },
          {
            "name": "checkbox-label"
          },
          {
            "name": "dialog"
          },
          {
            "name": "groupbox"
          },
          {
            "name": "listbox"
          },
          {
            "name": "menuarrow"
          },
          {
            "name": "menuimage"
          },
          {
            "name": "menuitem"
          },
          {
            "name": "menuitemtext"
          },
          {
            "name": "menulist"
          },
          {
            "name": "menulist-button"
          },
          {
            "name": "menulist-text"
          },
          {
            "name": "menulist-textfield"
          },
          {
            "name": "menupopup"
          },
          {
            "name": "menuradio"
          },
          {
            "name": "menuseparator"
          },
          {
            "name": "-moz-mac-unified-toolbar"
          },
          {
            "name": "-moz-win-borderless-glass"
          },
          {
            "name": "-moz-win-browsertabbar-toolbox"
          },
          {
            "name": "-moz-win-communications-toolbox"
          },
          {
            "name": "-moz-win-glass"
          },
          {
            "name": "-moz-win-media-toolbox"
          },
          {
            "name": "none"
          },
          {
            "name": "progressbar"
          },
          {
            "name": "progresschunk"
          },
          {
            "name": "radio"
          },
          {
            "name": "radio-container"
          },
          {
            "name": "radio-label"
          },
          {
            "name": "radiomenuitem"
          },
          {
            "name": "resizer"
          },
          {
            "name": "resizerpanel"
          },
          {
            "name": "scrollbarbutton-down"
          },
          {
            "name": "scrollbarbutton-left"
          },
          {
            "name": "scrollbarbutton-right"
          },
          {
            "name": "scrollbarbutton-up"
          },
          {
            "name": "scrollbar-small"
          },
          {
            "name": "scrollbartrack-horizontal"
          },
          {
            "name": "scrollbartrack-vertical"
          },
          {
            "name": "separator"
          },
          {
            "name": "spinner"
          },
          {
            "name": "spinner-downbutton"
          },
          {
            "name": "spinner-textfield"
          },
          {
            "name": "spinner-upbutton"
          },
          {
            "name": "statusbar"
          },
          {
            "name": "statusbarpanel"
          },
          {
            "name": "tab"
          },
          {
            "name": "tabpanels"
          },
          {
            "name": "tab-scroll-arrow-back"
          },
          {
            "name": "tab-scroll-arrow-forward"
          },
          {
            "name": "textfield"
          },
          {
            "name": "textfield-multiline"
          },
          {
            "name": "toolbar"
          },
          {
            "name": "toolbox"
          },
          {
            "name": "tooltip"
          },
          {
            "name": "treeheadercell"
          },
          {
            "name": "treeheadersortarrow"
          },
          {
            "name": "treeitem"
          },
          {
            "name": "treetwistyopen"
          },
          {
            "name": "treeview"
          },
          {
            "name": "treewisty"
          },
          {
            "name": "window"
          }
        ],
        "status": "nonstandard",
        "syntax": "none | button | button-arrow-down | button-arrow-next | button-arrow-previous | button-arrow-up | button-bevel | button-focus | caret | checkbox | checkbox-container | checkbox-label | checkmenuitem | dualbutton | groupbox | listbox | listitem | menuarrow | menubar | menucheckbox | menuimage | menuitem | menuitemtext | menulist | menulist-button | menulist-text | menulist-textfield | menupopup | menuradio | menuseparator | meterbar | meterchunk | progressbar | progressbar-vertical | progresschunk | progresschunk-vertical | radio | radio-container | radio-label | radiomenuitem | range | range-thumb | resizer | resizerpanel | scale-horizontal | scalethumbend | scalethumb-horizontal | scalethumbstart | scalethumbtick | scalethumb-vertical | scale-vertical | scrollbarbutton-down | scrollbarbutton-left | scrollbarbutton-right | scrollbarbutton-up | scrollbarthumb-horizontal | scrollbarthumb-vertical | scrollbartrack-horizontal | scrollbartrack-vertical | searchfield | separator | sheet | spinner | spinner-downbutton | spinner-textfield | spinner-upbutton | splitter | statusbar | statusbarpanel | tab | tabpanel | tabpanels | tab-scroll-arrow-back | tab-scroll-arrow-forward | textfield | textfield-multiline | toolbar | toolbarbutton | toolbarbutton-dropdown | toolbargripper | toolbox | tooltip | treeheader | treeheadercell | treeheadersortarrow | treeitem | treeline | treetwisty | treetwistyopen | treeview | -moz-mac-unified-toolbar | -moz-win-borderless-glass | -moz-win-browsertabbar-toolbox | -moz-win-communicationstext | -moz-win-communications-toolbox | -moz-win-exclude-glass | -moz-win-glass | -moz-win-mediatext | -moz-win-media-toolbox | -moz-window-button-box | -moz-window-button-box-maximized | -moz-window-button-close | -moz-window-button-maximize | -moz-window-button-minimize | -moz-window-button-restore | -moz-window-frame-bottom | -moz-window-frame-left | -moz-window-frame-right | -moz-window-titlebar | -moz-window-titlebar-maximized",
        "relevance": 0,
        "description": "Used in Gecko (Firefox) to display an element using a platform-native styling based on the operating system's theme.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-backface-visibility",
        "browsers": [
          "FF10"
        ],
        "values": [
          {
            "name": "hidden"
          },
          {
            "name": "visible"
          }
        ],
        "relevance": 50,
        "description": "Determines whether or not the 'back' side of a transformed element is visible when facing the viewer. With an identity transform, the front side of an element faces the viewer.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-background-clip",
        "browsers": [
          "FF1-3.6"
        ],
        "values": [
          {
            "name": "padding"
          }
        ],
        "relevance": 50,
        "description": "Determines the background painting area.",
        "restrictions": [
          "box",
          "enum"
        ]
      },
      {
        "name": "-moz-background-inline-policy",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "bounding-box"
          },
          {
            "name": "continuous"
          },
          {
            "name": "each-box"
          }
        ],
        "relevance": 50,
        "description": "In Gecko-based applications like Firefox, the -moz-background-inline-policy CSS property specifies how the background image of an inline element is determined when the content of the inline element wraps onto multiple lines. The choice of position has significant effects on repetition.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-background-origin",
        "browsers": [
          "FF1"
        ],
        "relevance": 50,
        "description": "For elements rendered as a single box, specifies the background positioning area. For elements rendered as multiple boxes (e.g., inline boxes on several lines, boxes on several pages) specifies which boxes 'box-decoration-break' operates on to determine the background positioning area(s).",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "-moz-border-bottom-colors",
        "browsers": [
          "FF1"
        ],
        "status": "nonstandard",
        "syntax": "<color>+ | none",
        "relevance": 0,
        "description": "Sets a list of colors for the bottom border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-border-image",
        "browsers": [
          "FF3.6"
        ],
        "values": [
          {
            "name": "auto",
            "description": "If 'auto' is specified then the border image width is the intrinsic width or height (whichever is applicable) of the corresponding image slice. If the image does not have the required intrinsic dimension then the corresponding border-width is used instead."
          },
          {
            "name": "fill",
            "description": "Causes the middle part of the border-image to be preserved."
          },
          {
            "name": "none"
          },
          {
            "name": "repeat",
            "description": "The image is tiled (repeated) to fill the area."
          },
          {
            "name": "round",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the image is rescaled so that it does."
          },
          {
            "name": "space",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the extra space is distributed around the tiles."
          },
          {
            "name": "stretch",
            "description": "The image is stretched to fill the area."
          },
          {
            "name": "url()"
          }
        ],
        "relevance": 50,
        "description": "Shorthand property for setting 'border-image-source', 'border-image-slice', 'border-image-width', 'border-image-outset' and 'border-image-repeat'. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "percentage",
          "number",
          "url",
          "enum"
        ]
      },
      {
        "name": "-moz-border-left-colors",
        "browsers": [
          "FF1"
        ],
        "status": "nonstandard",
        "syntax": "<color>+ | none",
        "relevance": 0,
        "description": "Sets a list of colors for the bottom border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-border-right-colors",
        "browsers": [
          "FF1"
        ],
        "status": "nonstandard",
        "syntax": "<color>+ | none",
        "relevance": 0,
        "description": "Sets a list of colors for the bottom border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-border-top-colors",
        "browsers": [
          "FF1"
        ],
        "status": "nonstandard",
        "syntax": "<color>+ | none",
        "relevance": 0,
        "description": "Ske Firefox, -moz-border-bottom-colors sets a list of colors for the bottom border.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-box-align",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "baseline",
            "description": "If this box orientation is inline-axis or horizontal, all children are placed with their baselines aligned, and extra space placed before or after as necessary. For block flows, the baseline of the first non-empty line box located within the element is used. For tables, the baseline of the first cell is used."
          },
          {
            "name": "center",
            "description": "Any extra space is divided evenly, with half placed above the child and the other half placed after the child."
          },
          {
            "name": "end",
            "description": "For normal direction boxes, the bottom edge of each child is placed along the bottom of the box. Extra space is placed above the element. For reverse direction boxes, the top edge of each child is placed along the top of the box. Extra space is placed below the element."
          },
          {
            "name": "start",
            "description": "For normal direction boxes, the top edge of each child is placed along the top of the box. Extra space is placed below the element. For reverse direction boxes, the bottom edge of each child is placed along the bottom of the box. Extra space is placed above the element."
          },
          {
            "name": "stretch",
            "description": "The height of each child is adjusted to that of the containing block."
          }
        ],
        "relevance": 50,
        "description": "Specifies how a XUL box aligns its contents across (perpendicular to) the direction of its layout. The effect of this is only visible if there is extra space in the box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-box-direction",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "normal",
            "description": "A box with a computed value of horizontal for box-orient displays its children from left to right. A box with a computed value of vertical displays its children from top to bottom."
          },
          {
            "name": "reverse",
            "description": "A box with a computed value of horizontal for box-orient displays its children from right to left. A box with a computed value of vertical displays its children from bottom to top."
          }
        ],
        "relevance": 50,
        "description": "Specifies whether a box lays out its contents normally (from the top or left edge), or in reverse (from the bottom or right edge).",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-box-flex",
        "browsers": [
          "FF1"
        ],
        "relevance": 50,
        "description": "Specifies how a box grows to fill the box that contains it, in the direction of the containing box's layout.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "-moz-box-flexgroup",
        "browsers": [
          "FF1"
        ],
        "relevance": 50,
        "description": "Flexible elements can be assigned to flex groups using the 'box-flex-group' property.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-moz-box-ordinal-group",
        "browsers": [
          "FF1"
        ],
        "relevance": 50,
        "description": "Indicates the ordinal group the element belongs to. Elements with a lower ordinal group are displayed before those with a higher ordinal group.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-moz-box-orient",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "block-axis",
            "description": "Elements are oriented along the box's axis."
          },
          {
            "name": "horizontal",
            "description": "The box displays its children from left to right in a horizontal line."
          },
          {
            "name": "inline-axis",
            "description": "Elements are oriented vertically."
          },
          {
            "name": "vertical",
            "description": "The box displays its children from stacked from top to bottom vertically."
          }
        ],
        "relevance": 50,
        "description": "In Mozilla applications, -moz-box-orient specifies whether a box lays out its contents horizontally or vertically.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-box-pack",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "center",
            "description": "The extra space is divided evenly, with half placed before the first child and the other half placed after the last child."
          },
          {
            "name": "end",
            "description": "For normal direction boxes, the right edge of the last child is placed at the right side, with all extra space placed before the first child. For reverse direction boxes, the left edge of the first child is placed at the left side, with all extra space placed after the last child."
          },
          {
            "name": "justify",
            "description": "The space is divided evenly in-between each child, with none of the extra space placed before the first child or after the last child. If there is only one child, treat the pack value as if it were start."
          },
          {
            "name": "start",
            "description": "For normal direction boxes, the left edge of the first child is placed at the left side, with all extra space placed after the last child. For reverse direction boxes, the right edge of the last child is placed at the right side, with all extra space placed before the first child."
          }
        ],
        "relevance": 50,
        "description": "Specifies how a box packs its contents in the direction of its layout. The effect of this is only visible if there is extra space in the box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-box-sizing",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "border-box",
            "description": "The specified width and height (and respective min/max properties) on this element determine the border box of the element."
          },
          {
            "name": "content-box",
            "description": "Behavior of width and height as specified by CSS2.1. The specified width and height (and respective min/max properties) apply to the width and height respectively of the content box of the element."
          },
          {
            "name": "padding-box",
            "description": "The specified width and height (and respective min/max properties) on this element determine the padding box of the element."
          }
        ],
        "relevance": 50,
        "description": "Box Model addition in CSS3.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-column-count",
        "browsers": [
          "FF3.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Determines the number of columns by the 'column-width' property and the element width."
          }
        ],
        "relevance": 50,
        "description": "Describes the optimal number of columns into which the content of the element will be flowed.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-moz-column-gap",
        "browsers": [
          "FF3.5"
        ],
        "values": [
          {
            "name": "normal",
            "description": "User agent specific and typically equivalent to 1em."
          }
        ],
        "relevance": 50,
        "description": "Sets the gap between columns. If there is a column rule between columns, it will appear in the middle of the gap.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-moz-column-rule",
        "browsers": [
          "FF3.5"
        ],
        "relevance": 50,
        "description": "Shorthand for setting 'column-rule-width', 'column-rule-style', and 'column-rule-color' at the same place in the style sheet. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "-moz-column-rule-color",
        "browsers": [
          "FF3.5"
        ],
        "relevance": 50,
        "description": "Sets the color of the column rule",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-column-rule-style",
        "browsers": [
          "FF3.5"
        ],
        "relevance": 50,
        "description": "Sets the style of the rule between columns of an element.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "-moz-column-rule-width",
        "browsers": [
          "FF3.5"
        ],
        "relevance": 50,
        "description": "Sets the width of the rule between columns. Negative values are not allowed.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "-moz-columns",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "relevance": 50,
        "description": "A shorthand property which sets both 'column-width' and 'column-count'.",
        "restrictions": [
          "length",
          "integer"
        ]
      },
      {
        "name": "-moz-column-width",
        "browsers": [
          "FF3.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "relevance": 50,
        "description": "This property describes the width of columns in multicol elements.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-moz-font-feature-settings",
        "browsers": [
          "FF4"
        ],
        "values": [
          {
            "name": '"c2cs"'
          },
          {
            "name": '"dlig"'
          },
          {
            "name": '"kern"'
          },
          {
            "name": '"liga"'
          },
          {
            "name": '"lnum"'
          },
          {
            "name": '"onum"'
          },
          {
            "name": '"smcp"'
          },
          {
            "name": '"swsh"'
          },
          {
            "name": '"tnum"'
          },
          {
            "name": "normal",
            "description": "No change in glyph substitution or positioning occurs."
          },
          {
            "name": "off",
            "browsers": [
              "FF4"
            ]
          },
          {
            "name": "on",
            "browsers": [
              "FF4"
            ]
          }
        ],
        "relevance": 50,
        "description": "Provides low-level control over OpenType font features. It is intended as a way of providing access to font features that are not widely used but are needed for a particular use case.",
        "restrictions": [
          "string",
          "integer"
        ]
      },
      {
        "name": "-moz-hyphens",
        "browsers": [
          "FF9"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Conditional hyphenation characters inside a word, if present, take priority over automatic resources when determining hyphenation points within the word."
          },
          {
            "name": "manual",
            "description": "Words are only broken at line breaks where there are characters inside the word that suggest line break opportunities"
          },
          {
            "name": "none",
            "description": "Words are not broken at line breaks, even if characters inside the word suggest line break points."
          }
        ],
        "relevance": 50,
        "description": "Controls whether hyphenation is allowed to create more break opportunities within a line of text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-perspective",
        "browsers": [
          "FF10"
        ],
        "values": [
          {
            "name": "none",
            "description": "No perspective transform is applied."
          }
        ],
        "relevance": 50,
        "description": "Applies the same transform as the perspective(<number>) transform function, except that it applies only to the positioned or transformed children of the element, not to the transform on the element itself.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-moz-perspective-origin",
        "browsers": [
          "FF10"
        ],
        "relevance": 50,
        "description": "Establishes the origin for the perspective property. It effectively sets the X and Y position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "-moz-text-align-last",
        "browsers": [
          "FF12"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "center",
            "description": "The inline contents are centered within the line box."
          },
          {
            "name": "justify",
            "description": "The text is justified according to the method specified by the 'text-justify' property."
          },
          {
            "name": "left",
            "description": "The inline contents are aligned to the left edge of the line box. In vertical text, 'left' aligns to the edge of the line box that would be the start edge for left-to-right text."
          },
          {
            "name": "right",
            "description": "The inline contents are aligned to the right edge of the line box. In vertical text, 'right' aligns to the edge of the line box that would be the end edge for left-to-right text."
          }
        ],
        "relevance": 50,
        "description": "Describes how the last line of a block or a line right before a forced line break is aligned when 'text-align' is set to 'justify'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-text-decoration-color",
        "browsers": [
          "FF6"
        ],
        "relevance": 50,
        "description": "Specifies the color of text decoration (underlines overlines, and line-throughs) set on the element with text-decoration-line.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-moz-text-decoration-line",
        "browsers": [
          "FF6"
        ],
        "values": [
          {
            "name": "line-through",
            "description": "Each line of text has a line through the middle."
          },
          {
            "name": "none",
            "description": "Neither produces nor inhibits text decoration."
          },
          {
            "name": "overline",
            "description": "Each line of text has a line above it."
          },
          {
            "name": "underline",
            "description": "Each line of text is underlined."
          }
        ],
        "relevance": 50,
        "description": "Specifies what line decorations, if any, are added to the element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-text-decoration-style",
        "browsers": [
          "FF6"
        ],
        "values": [
          {
            "name": "dashed",
            "description": "Produces a dashed line style."
          },
          {
            "name": "dotted",
            "description": "Produces a dotted line."
          },
          {
            "name": "double",
            "description": "Produces a double line."
          },
          {
            "name": "none",
            "description": "Produces no line."
          },
          {
            "name": "solid",
            "description": "Produces a solid line."
          },
          {
            "name": "wavy",
            "description": "Produces a wavy line."
          }
        ],
        "relevance": 50,
        "description": "Specifies the line style for underline, line-through and overline text decoration.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-text-size-adjust",
        "browsers": [
          "FF"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Renderers must use the default size adjustment when displaying on a small device."
          },
          {
            "name": "none",
            "description": "Renderers must not do size adjustment when displaying on a small device."
          }
        ],
        "relevance": 50,
        "description": "Specifies a size adjustment for displaying text content in mobile browsers.",
        "restrictions": [
          "enum",
          "percentage"
        ]
      },
      {
        "name": "-moz-transform",
        "browsers": [
          "FF3.5"
        ],
        "values": [
          {
            "name": "matrix()",
            "description": "Specifies a 2D transformation in the form of a transformation matrix of six values. matrix(a,b,c,d,e,f) is equivalent to applying the transformation matrix [a b c d e f]"
          },
          {
            "name": "matrix3d()",
            "description": "Specifies a 3D transformation as a 4x4 homogeneous matrix of 16 values in column-major order."
          },
          {
            "name": "none"
          },
          {
            "name": "perspective",
            "description": "Specifies a perspective projection matrix."
          },
          {
            "name": "rotate()",
            "description": "Specifies a 2D rotation by the angle specified in the parameter about the origin of the element, as defined by the transform-origin property."
          },
          {
            "name": "rotate3d()",
            "description": "Specifies a clockwise 3D rotation by the angle specified in last parameter about the [x,y,z] direction vector described by the first 3 parameters."
          },
          {
            "name": "rotateX('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the X axis."
          },
          {
            "name": "rotateY('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Y axis."
          },
          {
            "name": "rotateZ('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Z axis."
          },
          {
            "name": "scale()",
            "description": "Specifies a 2D scale operation by the [sx,sy] scaling vector described by the 2 parameters. If the second parameter is not provided, it is takes a value equal to the first."
          },
          {
            "name": "scale3d()",
            "description": "Specifies a 3D scale operation by the [sx,sy,sz] scaling vector described by the 3 parameters."
          },
          {
            "name": "scaleX()",
            "description": "Specifies a scale operation using the [sx,1] scaling vector, where sx is given as the parameter."
          },
          {
            "name": "scaleY()",
            "description": "Specifies a scale operation using the [sy,1] scaling vector, where sy is given as the parameter."
          },
          {
            "name": "scaleZ()",
            "description": "Specifies a scale operation using the [1,1,sz] scaling vector, where sz is given as the parameter."
          },
          {
            "name": "skew()",
            "description": "Specifies a skew transformation along the X and Y axes. The first angle parameter specifies the skew on the X axis. The second angle parameter specifies the skew on the Y axis. If the second parameter is not given then a value of 0 is used for the Y angle (ie: no skew on the Y axis)."
          },
          {
            "name": "skewX()",
            "description": "Specifies a skew transformation along the X axis by the given angle."
          },
          {
            "name": "skewY()",
            "description": "Specifies a skew transformation along the Y axis by the given angle."
          },
          {
            "name": "translate()",
            "description": "Specifies a 2D translation by the vector [tx, ty], where tx is the first translation-value parameter and ty is the optional second translation-value parameter."
          },
          {
            "name": "translate3d()",
            "description": "Specifies a 3D translation by the vector [tx,ty,tz], with tx, ty and tz being the first, second and third translation-value parameters respectively."
          },
          {
            "name": "translateX()",
            "description": "Specifies a translation by the given amount in the X direction."
          },
          {
            "name": "translateY()",
            "description": "Specifies a translation by the given amount in the Y direction."
          },
          {
            "name": "translateZ()",
            "description": "Specifies a translation by the given amount in the Z direction. Note that percentage values are not allowed in the translateZ translation-value, and if present are evaluated as 0."
          }
        ],
        "relevance": 50,
        "description": "A two-dimensional transformation is applied to an element through the 'transform' property. This property contains a list of transform functions similar to those allowed by SVG.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-moz-transform-origin",
        "browsers": [
          "FF3.5"
        ],
        "relevance": 50,
        "description": "Establishes the origin of transformation for an element.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-moz-transition",
        "browsers": [
          "FF4"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines four of the transition properties into a single property.",
        "restrictions": [
          "time",
          "property",
          "timing-function",
          "enum"
        ]
      },
      {
        "name": "-moz-transition-delay",
        "browsers": [
          "FF4"
        ],
        "relevance": 50,
        "description": "Defines when the transition will start. It allows a transition to begin execution some period of time from when it is applied.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-moz-transition-duration",
        "browsers": [
          "FF4"
        ],
        "relevance": 50,
        "description": "Specifies how long the transition from the old value to the new value should take.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-moz-transition-property",
        "browsers": [
          "FF4"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Specifies the name of the CSS property to which the transition is applied.",
        "restrictions": [
          "property"
        ]
      },
      {
        "name": "-moz-transition-timing-function",
        "browsers": [
          "FF4"
        ],
        "relevance": 50,
        "description": "Describes how the intermediate values used during a transition will be calculated.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "-moz-user-focus",
        "browsers": [
          "FF1"
        ],
        "values": [
          {
            "name": "ignore"
          },
          {
            "name": "normal"
          }
        ],
        "status": "nonstandard",
        "syntax": "ignore | normal | select-after | select-before | select-menu | select-same | select-all | none",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-user-focus"
          }
        ],
        "description": "Used to indicate whether the element can have focus."
      },
      {
        "name": "-moz-user-select",
        "browsers": [
          "FF1.5"
        ],
        "values": [
          {
            "name": "all"
          },
          {
            "name": "element"
          },
          {
            "name": "elements"
          },
          {
            "name": "-moz-all"
          },
          {
            "name": "-moz-none"
          },
          {
            "name": "none"
          },
          {
            "name": "text"
          },
          {
            "name": "toggle"
          }
        ],
        "relevance": 50,
        "description": "Controls the appearance of selection.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-accelerator",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "false",
            "description": "The element does not contain an accelerator key sequence."
          },
          {
            "name": "true",
            "description": "The element contains an accelerator key sequence."
          }
        ],
        "status": "nonstandard",
        "syntax": "false | true",
        "relevance": 0,
        "description": "IE only. Has the ability to turn off its system underlines for accelerator keys until the ALT key is pressed",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-behavior",
        "browsers": [
          "IE8"
        ],
        "relevance": 50,
        "description": "IE only. Used to extend behaviors of the browser",
        "restrictions": [
          "url"
        ]
      },
      {
        "name": "-ms-block-progression",
        "browsers": [
          "IE8"
        ],
        "values": [
          {
            "name": "bt",
            "description": "Bottom-to-top block flow. Layout is horizontal."
          },
          {
            "name": "lr",
            "description": "Left-to-right direction. The flow orientation is vertical."
          },
          {
            "name": "rl",
            "description": "Right-to-left direction. The flow orientation is vertical."
          },
          {
            "name": "tb",
            "description": "Top-to-bottom direction. The flow orientation is horizontal."
          }
        ],
        "status": "nonstandard",
        "syntax": "tb | rl | bt | lr",
        "relevance": 0,
        "description": "Sets the block-progression value and the flow orientation",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-content-zoom-chaining",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "chained",
            "description": "The nearest zoomable parent element begins zooming when the user hits a zoom limit during a manipulation. No bounce effect is shown."
          },
          {
            "name": "none",
            "description": "A bounce effect is shown when the user hits a zoom limit during a manipulation."
          }
        ],
        "status": "nonstandard",
        "syntax": "none | chained",
        "relevance": 0,
        "description": "Specifies the zoom behavior that occurs when a user hits the zoom limit during a manipulation."
      },
      {
        "name": "-ms-content-zooming",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "The element is not zoomable."
          },
          {
            "name": "zoom",
            "description": "The element is zoomable."
          }
        ],
        "status": "nonstandard",
        "syntax": "none | zoom",
        "relevance": 0,
        "description": "Specifies whether zooming is enabled.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-content-zoom-limit",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<'-ms-content-zoom-limit-min'> <'-ms-content-zoom-limit-max'>",
        "relevance": 0,
        "description": "Shorthand property for the -ms-content-zoom-limit-min and -ms-content-zoom-limit-max properties.",
        "restrictions": [
          "percentage"
        ]
      },
      {
        "name": "-ms-content-zoom-limit-max",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<percentage>",
        "relevance": 0,
        "description": "Specifies the maximum zoom factor.",
        "restrictions": [
          "percentage"
        ]
      },
      {
        "name": "-ms-content-zoom-limit-min",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<percentage>",
        "relevance": 0,
        "description": "Specifies the minimum zoom factor.",
        "restrictions": [
          "percentage"
        ]
      },
      {
        "name": "-ms-content-zoom-snap",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "mandatory",
            "description": "Indicates that the motion of the content after the contact is picked up is always adjusted so that it lands on a snap-point."
          },
          {
            "name": "none",
            "description": "Indicates that zooming is unaffected by any defined snap-points."
          },
          {
            "name": "proximity",
            "description": 'Indicates that the motion of the content after the contact is picked up may be adjusted if the content would normally stop "close enough" to a snap-point.'
          },
          {
            "name": "snapInterval(100%, 100%)",
            "description": "Specifies where the snap-points will be placed."
          },
          {
            "name": "snapList()",
            "description": "Specifies the position of individual snap-points as a comma-separated list of zoom factors."
          }
        ],
        "status": "nonstandard",
        "syntax": "<'-ms-content-zoom-snap-type'> || <'-ms-content-zoom-snap-points'>",
        "relevance": 0,
        "description": "Shorthand property for the -ms-content-zoom-snap-type and -ms-content-zoom-snap-points properties."
      },
      {
        "name": "-ms-content-zoom-snap-points",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "snapInterval(100%, 100%)",
            "description": "Specifies where the snap-points will be placed."
          },
          {
            "name": "snapList()",
            "description": "Specifies the position of individual snap-points as a comma-separated list of zoom factors."
          }
        ],
        "status": "nonstandard",
        "syntax": "snapInterval( <percentage>, <percentage> ) | snapList( <percentage># )",
        "relevance": 0,
        "description": "Defines where zoom snap-points are located."
      },
      {
        "name": "-ms-content-zoom-snap-type",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "mandatory",
            "description": "Indicates that the motion of the content after the contact is picked up is always adjusted so that it lands on a snap-point."
          },
          {
            "name": "none",
            "description": "Indicates that zooming is unaffected by any defined snap-points."
          },
          {
            "name": "proximity",
            "description": 'Indicates that the motion of the content after the contact is picked up may be adjusted if the content would normally stop "close enough" to a snap-point.'
          }
        ],
        "status": "nonstandard",
        "syntax": "none | proximity | mandatory",
        "relevance": 0,
        "description": "Specifies how zooming is affected by defined snap-points.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-filter",
        "browsers": [
          "IE8-9"
        ],
        "status": "nonstandard",
        "syntax": "<string>",
        "relevance": 0,
        "description": "IE only. Used to produce visual effects.",
        "restrictions": [
          "string"
        ]
      },
      {
        "name": "-ms-flex",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Retrieves the value of the main size property as the used 'flex-basis'."
          },
          {
            "name": "none",
            "description": "Expands to '0 0 auto'."
          }
        ],
        "relevance": 50,
        "description": "specifies the parameters of a flexible length: the positive and negative flexibility, and the preferred size.",
        "restrictions": [
          "length",
          "number",
          "percentage"
        ]
      },
      {
        "name": "-ms-flex-align",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "baseline",
            "description": "If the flex item\u2019s inline axis is the same as the cross axis, this value is identical to 'flex-start'. Otherwise, it participates in baseline alignment."
          },
          {
            "name": "center",
            "description": "The flex item\u2019s margin box is centered in the cross axis within the line."
          },
          {
            "name": "end",
            "description": "The cross-end margin edge of the flex item is placed flush with the cross-end edge of the line."
          },
          {
            "name": "start",
            "description": "The cross-start margin edge of the flexbox item is placed flush with the cross-start edge of the line."
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flexbox item is anything other than 'auto', this value is identical to 'start'."
          }
        ],
        "relevance": 50,
        "description": "Aligns flex items along the cross axis of the current line of the flex container.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-direction",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "column",
            "description": "The flex container\u2019s main axis has the same orientation as the block axis of the current writing mode."
          },
          {
            "name": "column-reverse",
            "description": "Same as 'column', except the main-start and main-end directions are swapped."
          },
          {
            "name": "row",
            "description": "The flex container\u2019s main axis has the same orientation as the inline axis of the current writing mode."
          },
          {
            "name": "row-reverse",
            "description": "Same as 'row', except the main-start and main-end directions are swapped."
          }
        ],
        "relevance": 50,
        "description": "Specifies how flex items are placed in the flex container, by setting the direction of the flex container\u2019s main axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-flow",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "column",
            "description": "The flex container\u2019s main axis has the same orientation as the block axis of the current writing mode."
          },
          {
            "name": "column-reverse",
            "description": "Same as 'column', except the main-start and main-end directions are swapped."
          },
          {
            "name": "nowrap",
            "description": "The flex container is single-line."
          },
          {
            "name": "row",
            "description": "The flex container\u2019s main axis has the same orientation as the inline axis of the current writing mode."
          },
          {
            "name": "wrap",
            "description": "The flexbox is multi-line."
          },
          {
            "name": "wrap-reverse",
            "description": "Same as 'wrap', except the cross-start and cross-end directions are swapped."
          }
        ],
        "relevance": 50,
        "description": "Specifies how flexbox items are placed in the flexbox.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-item-align",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Computes to the value of 'align-items' on the element\u2019s parent, or 'stretch' if the element has no parent. On absolutely positioned elements, it computes to itself."
          },
          {
            "name": "baseline",
            "description": "If the flex item\u2019s inline axis is the same as the cross axis, this value is identical to 'flex-start'. Otherwise, it participates in baseline alignment."
          },
          {
            "name": "center",
            "description": "The flex item\u2019s margin box is centered in the cross axis within the line."
          },
          {
            "name": "end",
            "description": "The cross-end margin edge of the flex item is placed flush with the cross-end edge of the line."
          },
          {
            "name": "start",
            "description": "The cross-start margin edge of the flex item is placed flush with the cross-start edge of the line."
          },
          {
            "name": "stretch",
            "description": "If the cross size property of the flex item computes to auto, and neither of the cross-axis margins are auto, the flex item is stretched."
          }
        ],
        "relevance": 50,
        "description": "Allows the default alignment along the cross axis to be overridden for individual flex items.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-line-pack",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "center",
            "description": "Lines are packed toward the center of the flex container."
          },
          {
            "name": "distribute",
            "description": "Lines are evenly distributed in the flex container, with half-size spaces on either end."
          },
          {
            "name": "end",
            "description": "Lines are packed toward the end of the flex container."
          },
          {
            "name": "justify",
            "description": "Lines are evenly distributed in the flex container."
          },
          {
            "name": "start",
            "description": "Lines are packed toward the start of the flex container."
          },
          {
            "name": "stretch",
            "description": "Lines stretch to take up the remaining space."
          }
        ],
        "relevance": 50,
        "description": "Aligns a flex container\u2019s lines within the flex container when there is extra space in the cross-axis, similar to how 'justify-content' aligns individual items within the main-axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-order",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "Controls the order in which children of a flex container appear within the flex container, by assigning them to ordinal groups.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-flex-pack",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "center",
            "description": "Flex items are packed toward the center of the line."
          },
          {
            "name": "distribute",
            "description": "Flex items are evenly distributed in the line, with half-size spaces on either end."
          },
          {
            "name": "end",
            "description": "Flex items are packed toward the end of the line."
          },
          {
            "name": "justify",
            "description": "Flex items are evenly distributed in the line."
          },
          {
            "name": "start",
            "description": "Flex items are packed toward the start of the line."
          }
        ],
        "relevance": 50,
        "description": "Aligns flex items along the main axis of the current line of the flex container.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flex-wrap",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "nowrap",
            "description": "The flex container is single-line."
          },
          {
            "name": "wrap",
            "description": "The flexbox is multi-line."
          },
          {
            "name": "wrap-reverse",
            "description": "Same as 'wrap', except the cross-start and cross-end directions are swapped."
          }
        ],
        "relevance": 50,
        "description": "Controls whether the flex container is single-line or multi-line, and the direction of the cross-axis, which determines the direction new lines are stacked in.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-flow-from",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "The block container is not a CSS Region."
          }
        ],
        "status": "nonstandard",
        "syntax": "[ none | <custom-ident> ]#",
        "relevance": 0,
        "description": "Makes a block container a region and associates it with a named flow.",
        "restrictions": [
          "identifier"
        ]
      },
      {
        "name": "-ms-flow-into",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "The element is not moved to a named flow and normal CSS processing takes place."
          }
        ],
        "status": "nonstandard",
        "syntax": "[ none | <custom-ident> ]#",
        "relevance": 0,
        "description": "Places an element or its contents into a named flow.",
        "restrictions": [
          "identifier"
        ]
      },
      {
        "name": "-ms-grid-column",
        "browsers": [
          "E12",
          "IE10"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "end"
          },
          {
            "name": "start"
          }
        ],
        "relevance": 50,
        "description": "Used to place grid items and explicitly defined grid cells in the Grid.",
        "restrictions": [
          "integer",
          "string",
          "enum"
        ]
      },
      {
        "name": "-ms-grid-column-align",
        "browsers": [
          "E12",
          "IE10"
        ],
        "values": [
          {
            "name": "center",
            "description": "Places the center of the Grid Item's margin box at the center of the Grid Item's column."
          },
          {
            "name": "end",
            "description": "Aligns the end edge of the Grid Item's margin box to the end edge of the Grid Item's column."
          },
          {
            "name": "start",
            "description": "Aligns the starting edge of the Grid Item's margin box to the starting edge of the Grid Item's column."
          },
          {
            "name": "stretch",
            "description": "Ensures that the Grid Item's margin box is equal to the size of the Grid Item's column."
          }
        ],
        "relevance": 50,
        "description": "Aligns the columns in a grid.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-grid-columns",
        "browsers": [
          "E12",
          "IE10"
        ],
        "relevance": 50,
        "description": "Lays out the columns of the grid."
      },
      {
        "name": "-ms-grid-column-span",
        "browsers": [
          "E12",
          "IE10"
        ],
        "relevance": 50,
        "description": "Specifies the number of columns to span.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-grid-layer",
        "browsers": [
          "E",
          "IE10"
        ],
        "relevance": 50,
        "description": "Grid-layer is similar in concept to z-index, but avoids overloading the meaning of the z-index property, which is applicable only to positioned elements.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-grid-row",
        "browsers": [
          "E12",
          "IE10"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "end"
          },
          {
            "name": "start"
          }
        ],
        "relevance": 50,
        "description": "grid-row is used to place grid items and explicitly defined grid cells in the Grid.",
        "restrictions": [
          "integer",
          "string",
          "enum"
        ]
      },
      {
        "name": "-ms-grid-row-align",
        "browsers": [
          "E12",
          "IE10"
        ],
        "values": [
          {
            "name": "center",
            "description": "Places the center of the Grid Item's margin box at the center of the Grid Item's row."
          },
          {
            "name": "end",
            "description": "Aligns the end edge of the Grid Item's margin box to the end edge of the Grid Item's row."
          },
          {
            "name": "start",
            "description": "Aligns the starting edge of the Grid Item's margin box to the starting edge of the Grid Item's row."
          },
          {
            "name": "stretch",
            "description": "Ensures that the Grid Item's margin box is equal to the size of the Grid Item's row."
          }
        ],
        "relevance": 50,
        "description": "Aligns the rows in a grid.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-grid-rows",
        "browsers": [
          "E12",
          "IE10"
        ],
        "relevance": 50,
        "description": "Lays out the columns of the grid."
      },
      {
        "name": "-ms-grid-row-span",
        "browsers": [
          "E12",
          "IE10"
        ],
        "relevance": 50,
        "description": "Specifies the number of rows to span.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-high-contrast-adjust",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Properties will be adjusted as applicable."
          },
          {
            "name": "none",
            "description": "No adjustments will be applied."
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | none",
        "relevance": 0,
        "description": "Specifies if properties should be adjusted in high contrast mode.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-hyphenate-limit-chars",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent chooses a value that adapts to the current layout."
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | <integer>{1,3}",
        "relevance": 0,
        "description": "Specifies the minimum number of characters in a hyphenated word.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-hyphenate-limit-lines",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "no-limit",
            "description": "There is no limit."
          }
        ],
        "status": "nonstandard",
        "syntax": "no-limit | <integer>",
        "relevance": 0,
        "description": "Indicates the maximum number of successive hyphenated lines in an element.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-ms-hyphenate-limit-zone",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<percentage> | <length>",
        "relevance": 0,
        "description": "Specifies the maximum amount of unfilled space (before justification) that may be left in the line box before hyphenation is triggered to pull part of a word from the next line back up into the current line.",
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "-ms-hyphens",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Conditional hyphenation characters inside a word, if present, take priority over automatic resources when determining hyphenation points within the word."
          },
          {
            "name": "manual",
            "description": "Words are only broken at line breaks where there are characters inside the word that suggest line break opportunities"
          },
          {
            "name": "none",
            "description": "Words are not broken at line breaks, even if characters inside the word suggest line break points."
          }
        ],
        "relevance": 50,
        "description": "Controls whether hyphenation is allowed to create more break opportunities within a line of text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-ime-mode",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "active",
            "description": "The input method editor is initially active; text entry is performed using it unless the user specifically dismisses it."
          },
          {
            "name": "auto",
            "description": "No change is made to the current input method editor state. This is the default."
          },
          {
            "name": "disabled",
            "description": "The input method editor is disabled and may not be activated by the user."
          },
          {
            "name": "inactive",
            "description": "The input method editor is initially inactive, but the user may activate it if they wish."
          },
          {
            "name": "normal",
            "description": "The IME state should be normal; this value can be used in a user style sheet to override the page setting."
          }
        ],
        "relevance": 50,
        "description": "Controls the state of the input method editor for text fields.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-interpolation-mode",
        "browsers": [
          "IE7"
        ],
        "values": [
          {
            "name": "bicubic"
          },
          {
            "name": "nearest-neighbor"
          }
        ],
        "relevance": 50,
        "description": "Gets or sets the interpolation (resampling) method used to stretch images.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-layout-grid",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "char",
            "description": "Any of the range of character values available to the -ms-layout-grid-char property."
          },
          {
            "name": "line",
            "description": "Any of the range of line values available to the -ms-layout-grid-line property."
          },
          {
            "name": "mode",
            "description": "Any of the range of mode values available to the -ms-layout-grid-mode property."
          },
          {
            "name": "type",
            "description": "Any of the range of type values available to the -ms-layout-grid-type property."
          }
        ],
        "relevance": 50,
        "description": "Sets or retrieves the composite document grid properties that specify the layout of text characters."
      },
      {
        "name": "-ms-layout-grid-char",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Largest character in the font of the element is used to set the character grid."
          },
          {
            "name": "none",
            "description": "Default. No character grid is set."
          }
        ],
        "relevance": 50,
        "description": "Sets or retrieves the size of the character grid used for rendering the text content of an element.",
        "restrictions": [
          "enum",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-layout-grid-line",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Largest character in the font of the element is used to set the character grid."
          },
          {
            "name": "none",
            "description": "Default. No grid line is set."
          }
        ],
        "relevance": 50,
        "description": "Sets or retrieves the gridline value used for rendering the text content of an element.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-layout-grid-mode",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "both",
            "description": "Default. Both the char and line grid modes are enabled. This setting is necessary to fully enable the layout grid on an element."
          },
          {
            "name": "char",
            "description": "Only a character grid is used. This is recommended for use with block-level elements, such as a blockquote, where the line grid is intended to be disabled."
          },
          {
            "name": "line",
            "description": "Only a line grid is used. This is recommended for use with inline elements, such as a span, to disable the horizontal grid on runs of text that act as a single entity in the grid layout."
          },
          {
            "name": "none",
            "description": "No grid is used."
          }
        ],
        "relevance": 50,
        "description": "Gets or sets whether the text layout grid uses two dimensions.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-layout-grid-type",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "fixed",
            "description": "Grid used for monospaced layout. All noncursive characters are treated as equal; every character is centered within a single grid space by default."
          },
          {
            "name": "loose",
            "description": "Default. Grid used for Japanese and Korean characters."
          },
          {
            "name": "strict",
            "description": "Grid used for Chinese, as well as Japanese (Genko) and Korean characters. Only the ideographs, kanas, and wide characters are snapped to the grid."
          }
        ],
        "relevance": 50,
        "description": "Sets or retrieves the type of grid used for rendering the text content of an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-line-break",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The UA determines the set of line-breaking restrictions to use for CJK scripts, and it may vary the restrictions based on the length of the line; e.g., use a less restrictive set of line-break rules for short lines."
          },
          {
            "name": "keep-all",
            "description": "Sequences of CJK characters can no longer break on implied break points. This option should only be used where the presence of word separator characters still creates line-breaking opportunities, as in Korean."
          },
          {
            "name": "newspaper",
            "description": "Breaks CJK scripts using the least restrictive set of line-breaking rules. Typically used for short lines, such as in newspapers."
          },
          {
            "name": "normal",
            "description": "Breaks CJK scripts using a normal set of line-breaking rules."
          },
          {
            "name": "strict",
            "description": "Breaks CJK scripts using a more restrictive set of line-breaking rules than 'normal'."
          }
        ],
        "relevance": 50,
        "description": "Specifies what set of line breaking restrictions are in effect within the element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-overflow-style",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "No preference, UA should use the first scrolling method in the list that it supports."
          },
          {
            "name": "-ms-autohiding-scrollbar",
            "description": "Indicates the element displays auto-hiding scrollbars during mouse interactions and panning indicators during touch and keyboard interactions."
          },
          {
            "name": "none",
            "description": "Indicates the element does not display scrollbars or panning indicators, even when its content overflows."
          },
          {
            "name": "scrollbar",
            "description": 'Scrollbars are typically narrow strips inserted on one or two edges of an element and which often have arrows to click on and a "thumb" to drag up and down (or left and right) to move the contents of the element.'
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | none | scrollbar | -ms-autohiding-scrollbar",
        "relevance": 0,
        "description": "Specify whether content is clipped when it overflows the element's content area.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-perspective",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "No perspective transform is applied."
          }
        ],
        "relevance": 50,
        "description": "Applies the same transform as the perspective(<number>) transform function, except that it applies only to the positioned or transformed children of the element, not to the transform on the element itself.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-perspective-origin",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "Establishes the origin for the perspective property. It effectively sets the X and Y position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "-ms-perspective-origin-x",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "Establishes the origin for the perspective property. It effectively sets the X  position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "-ms-perspective-origin-y",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "Establishes the origin for the perspective property. It effectively sets the Y position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "-ms-progress-appearance",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "bar"
          },
          {
            "name": "ring"
          }
        ],
        "relevance": 50,
        "description": "Gets or sets a value that specifies whether a progress control displays as a bar or a ring.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scrollbar-3dlight-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the top and left edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-arrow-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the arrow elements of a scroll arrow.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-base-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the main elements of a scroll bar, which include the scroll box, track, and scroll arrows.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-darkshadow-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the gutter of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-face-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-highlight-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the top and left edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-shadow-color",
        "browsers": [
          "IE8"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "description": "Determines the color of the bottom and right edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scrollbar-track-color",
        "browsers": [
          "IE5"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-ms-scrollbar-track-color"
          }
        ],
        "description": "Determines the color of the track element of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-ms-scroll-chaining",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "chained"
          },
          {
            "name": "none"
          }
        ],
        "status": "nonstandard",
        "syntax": "chained | none",
        "relevance": 0,
        "description": "Gets or sets a value that indicates the scrolling behavior that occurs when a user hits the content boundary during a manipulation.",
        "restrictions": [
          "enum",
          "length"
        ]
      },
      {
        "name": "-ms-scroll-limit",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "status": "nonstandard",
        "syntax": "<'-ms-scroll-limit-x-min'> <'-ms-scroll-limit-y-min'> <'-ms-scroll-limit-x-max'> <'-ms-scroll-limit-y-max'>",
        "relevance": 0,
        "description": "Gets or sets a shorthand value that sets values for the -ms-scroll-limit-x-min, -ms-scroll-limit-y-min, -ms-scroll-limit-x-max, and -ms-scroll-limit-y-max properties.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-scroll-limit-x-max",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | <length>",
        "relevance": 0,
        "description": "Gets or sets a value that specifies the maximum value for the scrollLeft property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-scroll-limit-x-min",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<length>",
        "relevance": 0,
        "description": "Gets or sets a value that specifies the minimum value for the scrollLeft property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-scroll-limit-y-max",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | <length>",
        "relevance": 0,
        "description": "Gets or sets a value that specifies the maximum value for the scrollTop property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-scroll-limit-y-min",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<length>",
        "relevance": 0,
        "description": "Gets or sets a value that specifies the minimum value for the scrollTop property.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-ms-scroll-rails",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none"
          },
          {
            "name": "railed"
          }
        ],
        "status": "nonstandard",
        "syntax": "none | railed",
        "relevance": 0,
        "description": "Gets or sets a value that indicates whether or not small motions perpendicular to the primary axis of motion will result in either changes to both the scrollTop and scrollLeft properties or a change to the primary axis (for instance, either the scrollTop or scrollLeft properties will change, but not both).",
        "restrictions": [
          "enum",
          "length"
        ]
      },
      {
        "name": "-ms-scroll-snap-points-x",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "snapInterval(100%, 100%)"
          },
          {
            "name": "snapList()"
          }
        ],
        "status": "nonstandard",
        "syntax": "snapInterval( <length-percentage>, <length-percentage> ) | snapList( <length-percentage># )",
        "relevance": 0,
        "description": "Gets or sets a value that defines where snap-points will be located along the x-axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scroll-snap-points-y",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "snapInterval(100%, 100%)"
          },
          {
            "name": "snapList()"
          }
        ],
        "status": "nonstandard",
        "syntax": "snapInterval( <length-percentage>, <length-percentage> ) | snapList( <length-percentage># )",
        "relevance": 0,
        "description": "Gets or sets a value that defines where snap-points will be located along the y-axis.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scroll-snap-type",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "The visual viewport of this scroll container must ignore snap points, if any, when scrolled."
          },
          {
            "name": "mandatory",
            "description": "The visual viewport of this scroll container is guaranteed to rest on a snap point when there are no active scrolling operations."
          },
          {
            "name": "proximity",
            "description": "The visual viewport of this scroll container may come to rest on a snap point at the termination of a scroll at the discretion of the UA given the parameters of the scroll."
          }
        ],
        "status": "nonstandard",
        "syntax": "none | proximity | mandatory",
        "relevance": 0,
        "description": "Gets or sets a value that defines what type of snap-point should be used for the current element. There are two type of snap-points, with the primary difference being whether or not the user is guaranteed to always stop on a snap-point.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scroll-snap-x",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "mandatory"
          },
          {
            "name": "none"
          },
          {
            "name": "proximity"
          },
          {
            "name": "snapInterval(100%, 100%)"
          },
          {
            "name": "snapList()"
          }
        ],
        "status": "nonstandard",
        "syntax": "<'-ms-scroll-snap-type'> <'-ms-scroll-snap-points-x'>",
        "relevance": 0,
        "description": "Gets or sets a shorthand value that sets values for the -ms-scroll-snap-type and -ms-scroll-snap-points-x properties.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scroll-snap-y",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "mandatory"
          },
          {
            "name": "none"
          },
          {
            "name": "proximity"
          },
          {
            "name": "snapInterval(100%, 100%)"
          },
          {
            "name": "snapList()"
          }
        ],
        "status": "nonstandard",
        "syntax": "<'-ms-scroll-snap-type'> <'-ms-scroll-snap-points-y'>",
        "relevance": 0,
        "description": "Gets or sets a shorthand value that sets values for the -ms-scroll-snap-type and -ms-scroll-snap-points-y properties.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-scroll-translation",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none"
          },
          {
            "name": "vertical-to-horizontal"
          }
        ],
        "status": "nonstandard",
        "syntax": "none | vertical-to-horizontal",
        "relevance": 0,
        "description": "Gets or sets a value that specifies whether vertical-to-horizontal scroll wheel translation occurs on the specified element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-text-align-last",
        "browsers": [
          "E",
          "IE8"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "center",
            "description": "The inline contents are centered within the line box."
          },
          {
            "name": "justify",
            "description": "The text is justified according to the method specified by the 'text-justify' property."
          },
          {
            "name": "left",
            "description": "The inline contents are aligned to the left edge of the line box. In vertical text, 'left' aligns to the edge of the line box that would be the start edge for left-to-right text."
          },
          {
            "name": "right",
            "description": "The inline contents are aligned to the right edge of the line box. In vertical text, 'right' aligns to the edge of the line box that would be the end edge for left-to-right text."
          }
        ],
        "relevance": 50,
        "description": "Describes how the last line of a block or a line right before a forced line break is aligned when 'text-align' is set to 'justify'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-text-autospace",
        "browsers": [
          "E",
          "IE8"
        ],
        "values": [
          {
            "name": "ideograph-alpha",
            "description": "Creates 1/4em extra spacing between runs of ideographic letters and non-ideographic letters, such as Latin-based, Cyrillic, Greek, Arabic or Hebrew."
          },
          {
            "name": "ideograph-numeric",
            "description": "Creates 1/4em extra spacing between runs of ideographic letters and numeric glyphs."
          },
          {
            "name": "ideograph-parenthesis",
            "description": "Creates extra spacing between normal (non wide) parenthesis and ideographs."
          },
          {
            "name": "ideograph-space",
            "description": "Extends the width of the space character while surrounded by ideographs."
          },
          {
            "name": "none",
            "description": "No extra space is created."
          },
          {
            "name": "punctuation",
            "description": "Creates extra non-breaking spacing around punctuation as required by language-specific typographic conventions."
          }
        ],
        "status": "nonstandard",
        "syntax": "none | ideograph-alpha | ideograph-numeric | ideograph-parenthesis | ideograph-space",
        "relevance": 0,
        "description": "Determines whether or not a full-width punctuation mark character should be trimmed if it appears at the beginning of a line, so that its 'ink' lines up with the first glyph in the line above and below.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-text-combine-horizontal",
        "browsers": [
          "E",
          "IE11"
        ],
        "values": [
          {
            "name": "all",
            "description": "Attempt to typeset horizontally all consecutive characters within the box such that they take up the space of a single character within the vertical line box."
          },
          {
            "name": "digits",
            "description": "Attempt to typeset horizontally each maximal sequence of consecutive ASCII digits (U+0030\u2013U+0039) that has as many or fewer characters than the specified integer such that it takes up the space of a single character within the vertical line box."
          },
          {
            "name": "none",
            "description": "No special processing."
          }
        ],
        "relevance": 50,
        "description": "This property specifies the combination of multiple characters into the space of a single character.",
        "restrictions": [
          "enum",
          "integer"
        ]
      },
      {
        "name": "-ms-text-justify",
        "browsers": [
          "E",
          "IE8"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The UA determines the justification algorithm to follow, based on a balance between performance and adequate presentation quality."
          },
          {
            "name": "distribute",
            "description": "Justification primarily changes spacing both at word separators and at grapheme cluster boundaries in all scripts except those in the connected and cursive groups. This value is sometimes used in e.g. Japanese, often with the 'text-align-last' property."
          },
          {
            "name": "inter-cluster",
            "description": "Justification primarily changes spacing at word separators and at grapheme cluster boundaries in clustered scripts. This value is typically used for Southeast Asian scripts such as Thai."
          },
          {
            "name": "inter-ideograph",
            "description": "Justification primarily changes spacing at word separators and at inter-graphemic boundaries in scripts that use no word spaces. This value is typically used for CJK languages."
          },
          {
            "name": "inter-word",
            "description": "Justification primarily changes spacing at word separators. This value is typically used for languages that separate words using spaces, like English or (sometimes) Korean."
          },
          {
            "name": "kashida",
            "description": "Justification primarily stretches Arabic and related scripts through the use of kashida or other calligraphic elongation."
          }
        ],
        "relevance": 50,
        "description": "Selects the justification algorithm used when 'text-align' is set to 'justify'. The property applies to block containers, but the UA may (but is not required to) also support it on inline elements.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-text-kashida-space",
        "browsers": [
          "E",
          "IE10"
        ],
        "relevance": 50,
        "description": "Sets or retrieves the ratio of kashida expansion to white space expansion when justifying lines of text in the object.",
        "restrictions": [
          "percentage"
        ]
      },
      {
        "name": "-ms-text-overflow",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "clip",
            "description": "Clip inline content that overflows. Characters may be only partially rendered."
          },
          {
            "name": "ellipsis",
            "description": "Render an ellipsis character (U+2026) to represent clipped inline content."
          }
        ],
        "relevance": 50,
        "description": "Text can overflow for example when it is prevented from wrapping",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-text-size-adjust",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Renderers must use the default size adjustment when displaying on a small device."
          },
          {
            "name": "none",
            "description": "Renderers must not do size adjustment when displaying on a small device."
          }
        ],
        "relevance": 50,
        "description": "Specifies a size adjustment for displaying text content in mobile browsers.",
        "restrictions": [
          "enum",
          "percentage"
        ]
      },
      {
        "name": "-ms-text-underline-position",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "alphabetic",
            "description": "The underline is aligned with the alphabetic baseline. In this case the underline is likely to cross some descenders."
          },
          {
            "name": "auto",
            "description": "The user agent may use any algorithm to determine the underline's position. In horizontal line layout, the underline should be aligned as for alphabetic. In vertical line layout, if the language is set to Japanese or Korean, the underline should be aligned as for over."
          },
          {
            "name": "over",
            "description": "The underline is aligned with the 'top' (right in vertical writing) edge of the element's em-box. In this mode, an overline also switches sides."
          },
          {
            "name": "under",
            "description": "The underline is aligned with the 'bottom' (left in vertical writing) edge of the element's em-box. In this case the underline usually does not cross the descenders. This is sometimes called 'accounting' underline."
          }
        ],
        "relevance": 50,
        "description": "Sets the position of an underline specified on the same element: it does not affect underlines specified by ancestor elements.This property is typically used in vertical writing contexts such as in Japanese documents where it often desired to have the underline appear 'over' (to the right of) the affected run of text",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-touch-action",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The element is a passive element, with several exceptions."
          },
          {
            "name": "double-tap-zoom",
            "description": "The element will zoom on double-tap."
          },
          {
            "name": "manipulation",
            "description": "The element is a manipulation-causing element."
          },
          {
            "name": "none",
            "description": "The element is a manipulation-blocking element."
          },
          {
            "name": "pan-x",
            "description": "The element permits touch-driven panning on the horizontal axis. The touch pan is performed on the nearest ancestor with horizontally scrollable content."
          },
          {
            "name": "pan-y",
            "description": "The element permits touch-driven panning on the vertical axis. The touch pan is performed on the nearest ancestor with vertically scrollable content."
          },
          {
            "name": "pinch-zoom",
            "description": "The element permits pinch-zooming. The pinch-zoom is performed on the nearest ancestor with zoomable content."
          }
        ],
        "relevance": 50,
        "description": "Gets or sets a value that indicates whether and how a given region can be manipulated by the user.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-touch-select",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "grippers",
            "description": "Grippers are always on."
          },
          {
            "name": "none",
            "description": "Grippers are always off."
          }
        ],
        "status": "nonstandard",
        "syntax": "grippers | none",
        "relevance": 0,
        "description": "Gets or sets a value that toggles the 'gripper' visual elements that enable touch text selection.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-transform",
        "browsers": [
          "IE9-9"
        ],
        "values": [
          {
            "name": "matrix()",
            "description": "Specifies a 2D transformation in the form of a transformation matrix of six values. matrix(a,b,c,d,e,f) is equivalent to applying the transformation matrix [a b c d e f]"
          },
          {
            "name": "matrix3d()",
            "description": "Specifies a 3D transformation as a 4x4 homogeneous matrix of 16 values in column-major order."
          },
          {
            "name": "none"
          },
          {
            "name": "rotate()",
            "description": "Specifies a 2D rotation by the angle specified in the parameter about the origin of the element, as defined by the transform-origin property."
          },
          {
            "name": "rotate3d()",
            "description": "Specifies a clockwise 3D rotation by the angle specified in last parameter about the [x,y,z] direction vector described by the first 3 parameters."
          },
          {
            "name": "rotateX('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the X axis."
          },
          {
            "name": "rotateY('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Y axis."
          },
          {
            "name": "rotateZ('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Z axis."
          },
          {
            "name": "scale()",
            "description": "Specifies a 2D scale operation by the [sx,sy] scaling vector described by the 2 parameters. If the second parameter is not provided, it is takes a value equal to the first."
          },
          {
            "name": "scale3d()",
            "description": "Specifies a 3D scale operation by the [sx,sy,sz] scaling vector described by the 3 parameters."
          },
          {
            "name": "scaleX()",
            "description": "Specifies a scale operation using the [sx,1] scaling vector, where sx is given as the parameter."
          },
          {
            "name": "scaleY()",
            "description": "Specifies a scale operation using the [sy,1] scaling vector, where sy is given as the parameter."
          },
          {
            "name": "scaleZ()",
            "description": "Specifies a scale operation using the [1,1,sz] scaling vector, where sz is given as the parameter."
          },
          {
            "name": "skew()",
            "description": "Specifies a skew transformation along the X and Y axes. The first angle parameter specifies the skew on the X axis. The second angle parameter specifies the skew on the Y axis. If the second parameter is not given then a value of 0 is used for the Y angle (ie: no skew on the Y axis)."
          },
          {
            "name": "skewX()",
            "description": "Specifies a skew transformation along the X axis by the given angle."
          },
          {
            "name": "skewY()",
            "description": "Specifies a skew transformation along the Y axis by the given angle."
          },
          {
            "name": "translate()",
            "description": "Specifies a 2D translation by the vector [tx, ty], where tx is the first translation-value parameter and ty is the optional second translation-value parameter."
          },
          {
            "name": "translate3d()",
            "description": "Specifies a 3D translation by the vector [tx,ty,tz], with tx, ty and tz being the first, second and third translation-value parameters respectively."
          },
          {
            "name": "translateX()",
            "description": "Specifies a translation by the given amount in the X direction."
          },
          {
            "name": "translateY()",
            "description": "Specifies a translation by the given amount in the Y direction."
          },
          {
            "name": "translateZ()",
            "description": "Specifies a translation by the given amount in the Z direction. Note that percentage values are not allowed in the translateZ translation-value, and if present are evaluated as 0."
          }
        ],
        "relevance": 50,
        "description": "A two-dimensional transformation is applied to an element through the 'transform' property. This property contains a list of transform functions similar to those allowed by SVG.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-transform-origin",
        "browsers": [
          "IE9-9"
        ],
        "relevance": 50,
        "description": "Establishes the origin of transformation for an element.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-transform-origin-x",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "The x coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-transform-origin-y",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "The y coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-transform-origin-z",
        "browsers": [
          "IE10"
        ],
        "relevance": 50,
        "description": "The z coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-user-select",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "element"
          },
          {
            "name": "none"
          },
          {
            "name": "text"
          }
        ],
        "status": "nonstandard",
        "syntax": "none | element | text",
        "relevance": 0,
        "description": "Controls the appearance of selection.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-word-break",
        "browsers": [
          "IE8"
        ],
        "values": [
          {
            "name": "break-all",
            "description": "Lines may break between any two grapheme clusters for non-CJK scripts."
          },
          {
            "name": "keep-all",
            "description": "Block characters can no longer create implied break points."
          },
          {
            "name": "normal",
            "description": "Breaks non-CJK scripts according to their own rules."
          }
        ],
        "relevance": 50,
        "description": "Specifies line break opportunities for non-CJK scripts.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-word-wrap",
        "browsers": [
          "IE8"
        ],
        "values": [
          {
            "name": "break-word",
            "description": "An unbreakable 'word' may be broken at an arbitrary point if there are no otherwise-acceptable break points in the line."
          },
          {
            "name": "normal",
            "description": "Lines may break only at allowed break points."
          }
        ],
        "relevance": 50,
        "description": "Specifies whether the UA may break within a word to prevent overflow when an otherwise-unbreakable string is too long to fit.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-wrap-flow",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "auto",
            "description": "For floats an exclusion is created, for all other elements an exclusion is not created."
          },
          {
            "name": "both",
            "description": "Inline flow content can flow on all sides of the exclusion."
          },
          {
            "name": "clear",
            "description": "Inline flow content can only wrap on top and bottom of the exclusion and must leave the areas to the start and end edges of the exclusion box empty."
          },
          {
            "name": "end",
            "description": "Inline flow content can wrap on the end side of the exclusion area but must leave the area to the start edge of the exclusion area empty."
          },
          {
            "name": "maximum",
            "description": "Inline flow content can wrap on the side of the exclusion with the largest available space for the given line, and must leave the other side of the exclusion empty."
          },
          {
            "name": "minimum",
            "description": "Inline flow content can flow around the edge of the exclusion with the smallest available space within the flow content\u2019s containing block, and must leave the other edge of the exclusion empty."
          },
          {
            "name": "start",
            "description": "Inline flow content can wrap on the start edge of the exclusion area but must leave the area to end edge of the exclusion area empty."
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | both | start | end | maximum | clear",
        "relevance": 0,
        "description": "An element becomes an exclusion when its 'wrap-flow' property has a computed value other than 'auto'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-wrap-margin",
        "browsers": [
          "E",
          "IE10"
        ],
        "status": "nonstandard",
        "syntax": "<length>",
        "relevance": 0,
        "description": "Gets or sets a value that is used to offset the inner wrap shape from other shapes.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-ms-wrap-through",
        "browsers": [
          "E",
          "IE10"
        ],
        "values": [
          {
            "name": "none",
            "description": "The exclusion element does not inherit its parent node's wrapping context. Its descendants are only subject to exclusion shapes defined inside the element."
          },
          {
            "name": "wrap",
            "description": "The exclusion element inherits its parent node's wrapping context. Its descendant inline content wraps around exclusions defined outside the element."
          }
        ],
        "status": "nonstandard",
        "syntax": "wrap | none",
        "relevance": 0,
        "description": "Specifies if an element inherits its parent wrapping context. In other words if it is subject to the exclusions defined outside the element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-writing-mode",
        "browsers": [
          "IE8"
        ],
        "values": [
          {
            "name": "bt-lr"
          },
          {
            "name": "bt-rl"
          },
          {
            "name": "lr-bt"
          },
          {
            "name": "lr-tb"
          },
          {
            "name": "rl-bt"
          },
          {
            "name": "rl-tb"
          },
          {
            "name": "tb-lr"
          },
          {
            "name": "tb-rl"
          }
        ],
        "relevance": 50,
        "description": "Shorthand property for both 'direction' and 'block-progression'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-ms-zoom",
        "browsers": [
          "IE8"
        ],
        "values": [
          {
            "name": "normal"
          }
        ],
        "relevance": 50,
        "description": "Sets or retrieves the magnification scale of the object.",
        "restrictions": [
          "enum",
          "integer",
          "number",
          "percentage"
        ]
      },
      {
        "name": "-ms-zoom-animation",
        "browsers": [
          "IE10"
        ],
        "values": [
          {
            "name": "default"
          },
          {
            "name": "none"
          }
        ],
        "relevance": 50,
        "description": "Gets or sets a value that indicates whether an animation is used when zooming.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "nav-down",
        "browsers": [
          "O9.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent automatically determines which element to navigate the focus to in response to directional navigational input."
          },
          {
            "name": "current",
            "description": "Indicates that the user agent should target the frame that the element is in."
          },
          {
            "name": "root",
            "description": "Indicates that the user agent should target the full window."
          }
        ],
        "relevance": 50,
        "description": "Provides an way to control directional focus navigation.",
        "restrictions": [
          "enum",
          "identifier",
          "string"
        ]
      },
      {
        "name": "nav-index",
        "browsers": [
          "O9.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The element's sequential navigation order is assigned automatically by the user agent."
          }
        ],
        "relevance": 50,
        "description": "Provides an input-method-neutral way of specifying the sequential navigation order (also known as 'tabbing order').",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "nav-left",
        "browsers": [
          "O9.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent automatically determines which element to navigate the focus to in response to directional navigational input."
          },
          {
            "name": "current",
            "description": "Indicates that the user agent should target the frame that the element is in."
          },
          {
            "name": "root",
            "description": "Indicates that the user agent should target the full window."
          }
        ],
        "relevance": 50,
        "description": "Provides an way to control directional focus navigation.",
        "restrictions": [
          "enum",
          "identifier",
          "string"
        ]
      },
      {
        "name": "nav-right",
        "browsers": [
          "O9.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent automatically determines which element to navigate the focus to in response to directional navigational input."
          },
          {
            "name": "current",
            "description": "Indicates that the user agent should target the frame that the element is in."
          },
          {
            "name": "root",
            "description": "Indicates that the user agent should target the full window."
          }
        ],
        "relevance": 50,
        "description": "Provides an way to control directional focus navigation.",
        "restrictions": [
          "enum",
          "identifier",
          "string"
        ]
      },
      {
        "name": "nav-up",
        "browsers": [
          "O9.5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The user agent automatically determines which element to navigate the focus to in response to directional navigational input."
          },
          {
            "name": "current",
            "description": "Indicates that the user agent should target the frame that the element is in."
          },
          {
            "name": "root",
            "description": "Indicates that the user agent should target the full window."
          }
        ],
        "relevance": 50,
        "description": "Provides an way to control directional focus navigation.",
        "restrictions": [
          "enum",
          "identifier",
          "string"
        ]
      },
      {
        "name": "negative",
        "browsers": [
          "FF33"
        ],
        "syntax": "<symbol> <symbol>?",
        "relevance": 50,
        "description": "@counter-style descriptor. Defines how to alter the representation when the counter value is negative.",
        "restrictions": [
          "image",
          "identifier",
          "string"
        ]
      },
      {
        "name": "-o-animation",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          },
          {
            "name": "none",
            "description": "No animation is performed"
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines six of the animation properties into a single property.",
        "restrictions": [
          "time",
          "enum",
          "timing-function",
          "identifier",
          "number"
        ]
      },
      {
        "name": "-o-animation-delay",
        "browsers": [
          "O12"
        ],
        "relevance": 50,
        "description": "Defines when the animation will start.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-o-animation-direction",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Defines whether or not the animation should play in reverse on alternate cycles.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-animation-duration",
        "browsers": [
          "O12"
        ],
        "relevance": 50,
        "description": "Defines the length of time that an animation takes to complete one cycle.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-o-animation-fill-mode",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "none",
            "description": "There is no change to the property value between the time the animation is applied and the time the animation begins playing or after the animation completes."
          }
        ],
        "relevance": 50,
        "description": "Defines what values are applied by the animation outside the time it is executing.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-animation-iteration-count",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          }
        ],
        "relevance": 50,
        "description": "Defines the number of times an animation cycle is played. The default value is one, meaning the animation will play from beginning to end once.",
        "restrictions": [
          "number",
          "enum"
        ]
      },
      {
        "name": "-o-animation-name",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "none",
            "description": "No animation is performed"
          }
        ],
        "relevance": 50,
        "description": "Defines a list of animations that apply. Each name is used to select the keyframe at-rule that provides the property values for the animation.",
        "restrictions": [
          "identifier",
          "enum"
        ]
      },
      {
        "name": "-o-animation-play-state",
        "browsers": [
          "O12"
        ],
        "values": [
          {
            "name": "paused",
            "description": "A running animation will be paused."
          },
          {
            "name": "running",
            "description": "Resume playback of a paused animation."
          }
        ],
        "relevance": 50,
        "description": "Defines whether the animation is running or paused.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-animation-timing-function",
        "browsers": [
          "O12"
        ],
        "relevance": 50,
        "description": "Describes how the animation will progress over one cycle of its duration. See the 'transition-timing-function'.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "object-fit",
        "browsers": [
          "E16",
          "FF36",
          "S10",
          "C31",
          "O19"
        ],
        "values": [
          {
            "name": "contain",
            "description": "The replaced content is sized to maintain its aspect ratio while fitting within the element\u2019s content box: its concrete object size is resolved as a contain constraint against the element's used width and height."
          },
          {
            "name": "cover",
            "description": "The replaced content is sized to maintain its aspect ratio while filling the element's entire content box: its concrete object size is resolved as a cover constraint against the element\u2019s used width and height."
          },
          {
            "name": "fill",
            "description": "The replaced content is sized to fill the element\u2019s content box: the object's concrete object size is the element's used width and height."
          },
          {
            "name": "none",
            "description": "The replaced content is not resized to fit inside the element's content box"
          },
          {
            "name": "scale-down",
            "description": "Size the content as if \u2018none\u2019 or \u2018contain\u2019 were specified, whichever would result in a smaller concrete object size."
          }
        ],
        "syntax": "fill | contain | cover | none | scale-down",
        "relevance": 61,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/object-fit"
          }
        ],
        "description": "Specifies how the contents of a replaced element should be scaled relative to the box established by its used height and width.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "object-position",
        "browsers": [
          "E16",
          "FF36",
          "S10",
          "C31",
          "O19"
        ],
        "syntax": "<position>",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/object-position"
          }
        ],
        "description": "Determines the alignment of the replaced element inside its box.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-o-border-image",
        "browsers": [
          "O11.6"
        ],
        "values": [
          {
            "name": "auto",
            "description": "If 'auto' is specified then the border image width is the intrinsic width or height (whichever is applicable) of the corresponding image slice. If the image does not have the required intrinsic dimension then the corresponding border-width is used instead."
          },
          {
            "name": "fill",
            "description": "Causes the middle part of the border-image to be preserved."
          },
          {
            "name": "none"
          },
          {
            "name": "repeat",
            "description": "The image is tiled (repeated) to fill the area."
          },
          {
            "name": "round",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the image is rescaled so that it does."
          },
          {
            "name": "space",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the extra space is distributed around the tiles."
          },
          {
            "name": "stretch",
            "description": "The image is stretched to fill the area."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property for setting 'border-image-source', 'border-image-slice', 'border-image-width', 'border-image-outset' and 'border-image-repeat'. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "percentage",
          "number",
          "image",
          "enum"
        ]
      },
      {
        "name": "-o-object-fit",
        "browsers": [
          "O10.6"
        ],
        "values": [
          {
            "name": "contain",
            "description": "The replaced content is sized to maintain its aspect ratio while fitting within the element\u2019s content box: its concrete object size is resolved as a contain constraint against the element's used width and height."
          },
          {
            "name": "cover",
            "description": "The replaced content is sized to maintain its aspect ratio while filling the element's entire content box: its concrete object size is resolved as a cover constraint against the element\u2019s used width and height."
          },
          {
            "name": "fill",
            "description": "The replaced content is sized to fill the element\u2019s content box: the object's concrete object size is the element's used width and height."
          },
          {
            "name": "none",
            "description": "The replaced content is not resized to fit inside the element's content box"
          },
          {
            "name": "scale-down",
            "description": "Size the content as if \u2018none\u2019 or \u2018contain\u2019 were specified, whichever would result in a smaller concrete object size."
          }
        ],
        "relevance": 50,
        "description": "Specifies how the contents of a replaced element should be scaled relative to the box established by its used height and width.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-object-position",
        "browsers": [
          "O10.6"
        ],
        "relevance": 50,
        "description": "Determines the alignment of the replaced element inside its box.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "opacity",
        "syntax": "<alpha-value>",
        "relevance": 93,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/opacity"
          }
        ],
        "description": "Opacity of an element's text, where 1 is opaque and 0 is entirely transparent.",
        "restrictions": [
          "number(0-1)"
        ]
      },
      {
        "name": "order",
        "syntax": "<integer>",
        "relevance": 61,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/order"
          }
        ],
        "description": "Controls the order in which children of a flex container appear within the flex container, by assigning them to ordinal groups.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "orphans",
        "browsers": [
          "E12",
          "S1.3",
          "C25",
          "IE8",
          "O9.2"
        ],
        "syntax": "<integer>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/orphans"
          }
        ],
        "description": "Specifies the minimum number of line boxes in a block container that must be left in a fragment before a fragmentation break.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-o-table-baseline",
        "browsers": [
          "O9.6"
        ],
        "relevance": 50,
        "description": "Determines which row of a inline-table should be used as baseline of inline-table.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-o-tab-size",
        "browsers": [
          "O10.6"
        ],
        "relevance": 50,
        "description": "This property determines the width of the tab character (U+0009), in space characters (U+0020), when rendered.",
        "restrictions": [
          "integer",
          "length"
        ]
      },
      {
        "name": "-o-text-overflow",
        "browsers": [
          "O10"
        ],
        "values": [
          {
            "name": "clip",
            "description": "Clip inline content that overflows. Characters may be only partially rendered."
          },
          {
            "name": "ellipsis",
            "description": "Render an ellipsis character (U+2026) to represent clipped inline content."
          }
        ],
        "relevance": 50,
        "description": "Text can overflow for example when it is prevented from wrapping",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-transform",
        "browsers": [
          "O10.5"
        ],
        "values": [
          {
            "name": "matrix()",
            "description": "Specifies a 2D transformation in the form of a transformation matrix of six values. matrix(a,b,c,d,e,f) is equivalent to applying the transformation matrix [a b c d e f]"
          },
          {
            "name": "matrix3d()",
            "description": "Specifies a 3D transformation as a 4x4 homogeneous matrix of 16 values in column-major order."
          },
          {
            "name": "none"
          },
          {
            "name": "rotate()",
            "description": "Specifies a 2D rotation by the angle specified in the parameter about the origin of the element, as defined by the transform-origin property."
          },
          {
            "name": "rotate3d()",
            "description": "Specifies a clockwise 3D rotation by the angle specified in last parameter about the [x,y,z] direction vector described by the first 3 parameters."
          },
          {
            "name": "rotateX('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the X axis."
          },
          {
            "name": "rotateY('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Y axis."
          },
          {
            "name": "rotateZ('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Z axis."
          },
          {
            "name": "scale()",
            "description": "Specifies a 2D scale operation by the [sx,sy] scaling vector described by the 2 parameters. If the second parameter is not provided, it is takes a value equal to the first."
          },
          {
            "name": "scale3d()",
            "description": "Specifies a 3D scale operation by the [sx,sy,sz] scaling vector described by the 3 parameters."
          },
          {
            "name": "scaleX()",
            "description": "Specifies a scale operation using the [sx,1] scaling vector, where sx is given as the parameter."
          },
          {
            "name": "scaleY()",
            "description": "Specifies a scale operation using the [sy,1] scaling vector, where sy is given as the parameter."
          },
          {
            "name": "scaleZ()",
            "description": "Specifies a scale operation using the [1,1,sz] scaling vector, where sz is given as the parameter."
          },
          {
            "name": "skew()",
            "description": "Specifies a skew transformation along the X and Y axes. The first angle parameter specifies the skew on the X axis. The second angle parameter specifies the skew on the Y axis. If the second parameter is not given then a value of 0 is used for the Y angle (ie: no skew on the Y axis)."
          },
          {
            "name": "skewX()",
            "description": "Specifies a skew transformation along the X axis by the given angle."
          },
          {
            "name": "skewY()",
            "description": "Specifies a skew transformation along the Y axis by the given angle."
          },
          {
            "name": "translate()",
            "description": "Specifies a 2D translation by the vector [tx, ty], where tx is the first translation-value parameter and ty is the optional second translation-value parameter."
          },
          {
            "name": "translate3d()",
            "description": "Specifies a 3D translation by the vector [tx,ty,tz], with tx, ty and tz being the first, second and third translation-value parameters respectively."
          },
          {
            "name": "translateX()",
            "description": "Specifies a translation by the given amount in the X direction."
          },
          {
            "name": "translateY()",
            "description": "Specifies a translation by the given amount in the Y direction."
          },
          {
            "name": "translateZ()",
            "description": "Specifies a translation by the given amount in the Z direction. Note that percentage values are not allowed in the translateZ translation-value, and if present are evaluated as 0."
          }
        ],
        "relevance": 50,
        "description": "A two-dimensional transformation is applied to an element through the 'transform' property. This property contains a list of transform functions similar to those allowed by SVG.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-o-transform-origin",
        "browsers": [
          "O10.5"
        ],
        "relevance": 50,
        "description": "Establishes the origin of transformation for an element.",
        "restrictions": [
          "positon",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-o-transition",
        "browsers": [
          "O11.5"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines four of the transition properties into a single property.",
        "restrictions": [
          "time",
          "property",
          "timing-function",
          "enum"
        ]
      },
      {
        "name": "-o-transition-delay",
        "browsers": [
          "O11.5"
        ],
        "relevance": 50,
        "description": "Defines when the transition will start. It allows a transition to begin execution some period of time from when it is applied.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-o-transition-duration",
        "browsers": [
          "O11.5"
        ],
        "relevance": 50,
        "description": "Specifies how long the transition from the old value to the new value should take.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-o-transition-property",
        "browsers": [
          "O11.5"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Specifies the name of the CSS property to which the transition is applied.",
        "restrictions": [
          "property"
        ]
      },
      {
        "name": "-o-transition-timing-function",
        "browsers": [
          "O11.5"
        ],
        "relevance": 50,
        "description": "Describes how the intermediate values used during a transition will be calculated.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "offset-block-end",
        "browsers": [
          "FF41"
        ],
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well."
          }
        ],
        "relevance": 50,
        "description": "Logical 'bottom'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "offset-block-start",
        "browsers": [
          "FF41"
        ],
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well."
          }
        ],
        "relevance": 50,
        "description": "Logical 'top'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "offset-inline-end",
        "browsers": [
          "FF41"
        ],
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well."
          }
        ],
        "relevance": 50,
        "description": "Logical 'right'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "offset-inline-start",
        "browsers": [
          "FF41"
        ],
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well."
          }
        ],
        "relevance": 50,
        "description": "Logical 'left'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "outline",
        "values": [
          {
            "name": "auto",
            "description": "Permits the user agent to render a custom outline style, typically the default platform style."
          },
          {
            "name": "invert",
            "description": "Performs a color inversion on the pixels on the screen."
          }
        ],
        "syntax": "[ <'outline-color'> || <'outline-style'> || <'outline-width'> ]",
        "relevance": 87,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/outline"
          }
        ],
        "description": "Shorthand property for 'outline-style', 'outline-width', and 'outline-color'.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color",
          "enum"
        ]
      },
      {
        "name": "outline-color",
        "values": [
          {
            "name": "invert",
            "description": "Performs a color inversion on the pixels on the screen."
          }
        ],
        "syntax": "<color> | invert",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/outline-color"
          }
        ],
        "description": "The color of the outline.",
        "restrictions": [
          "enum",
          "color"
        ]
      },
      {
        "name": "outline-offset",
        "browsers": [
          "E15",
          "FF1.5",
          "S1.2",
          "C1",
          "O9.5"
        ],
        "syntax": "<length>",
        "relevance": 59,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/outline-offset"
          }
        ],
        "description": "Offset the outline and draw it beyond the border edge.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "outline-style",
        "values": [
          {
            "name": "auto",
            "description": "Permits the user agent to render a custom outline style, typically the default platform style."
          }
        ],
        "syntax": "auto | <'border-style'>",
        "relevance": 60,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/outline-style"
          }
        ],
        "description": "Style of the outline.",
        "restrictions": [
          "line-style",
          "enum"
        ]
      },
      {
        "name": "outline-width",
        "syntax": "<line-width>",
        "relevance": 60,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/outline-width"
          }
        ],
        "description": "Width of the outline.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "overflow",
        "values": [
          {
            "name": "auto",
            "description": "The behavior of the 'auto' value is UA-dependent, but should cause a scrolling mechanism to be provided for overflowing boxes."
          },
          {
            "name": "hidden",
            "description": "Content is clipped and no scrolling mechanism should be provided to view the content outside the clipping region."
          },
          {
            "name": "-moz-hidden-unscrollable",
            "description": "Same as the standardized 'clip', except doesn\u2019t establish a block formatting context."
          },
          {
            "name": "scroll",
            "description": "Content is clipped and if the user agent uses a scrolling mechanism that is visible on the screen (such as a scroll bar or a panner), that mechanism should be displayed for a box whether or not any of its content is clipped."
          },
          {
            "name": "visible",
            "description": "Content is not clipped, i.e., it may be rendered outside the content box."
          }
        ],
        "syntax": "[ visible | hidden | clip | scroll | auto ]{1,2}",
        "relevance": 92,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow"
          }
        ],
        "description": "Shorthand for setting 'overflow-x' and 'overflow-y'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "overflow-wrap",
        "values": [
          {
            "name": "break-word",
            "description": "An otherwise unbreakable sequence of characters may be broken at an arbitrary point if there are no otherwise-acceptable break points in the line."
          },
          {
            "name": "normal",
            "description": "Lines may break only at allowed break points."
          }
        ],
        "syntax": "normal | break-word | anywhere",
        "relevance": 63,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-wrap"
          }
        ],
        "description": "Specifies whether the UA may break within a word to prevent overflow when an otherwise-unbreakable string is too long to fit within the line box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "overflow-x",
        "values": [
          {
            "name": "auto",
            "description": "The behavior of the 'auto' value is UA-dependent, but should cause a scrolling mechanism to be provided for overflowing boxes."
          },
          {
            "name": "hidden",
            "description": "Content is clipped and no scrolling mechanism should be provided to view the content outside the clipping region."
          },
          {
            "name": "scroll",
            "description": "Content is clipped and if the user agent uses a scrolling mechanism that is visible on the screen (such as a scroll bar or a panner), that mechanism should be displayed for a box whether or not any of its content is clipped."
          },
          {
            "name": "visible",
            "description": "Content is not clipped, i.e., it may be rendered outside the content box."
          }
        ],
        "syntax": "visible | hidden | clip | scroll | auto",
        "relevance": 79,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-x"
          }
        ],
        "description": "Specifies the handling of overflow in the horizontal direction.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "overflow-y",
        "values": [
          {
            "name": "auto",
            "description": "The behavior of the 'auto' value is UA-dependent, but should cause a scrolling mechanism to be provided for overflowing boxes."
          },
          {
            "name": "hidden",
            "description": "Content is clipped and no scrolling mechanism should be provided to view the content outside the clipping region."
          },
          {
            "name": "scroll",
            "description": "Content is clipped and if the user agent uses a scrolling mechanism that is visible on the screen (such as a scroll bar or a panner), that mechanism should be displayed for a box whether or not any of its content is clipped."
          },
          {
            "name": "visible",
            "description": "Content is not clipped, i.e., it may be rendered outside the content box."
          }
        ],
        "syntax": "visible | hidden | clip | scroll | auto",
        "relevance": 81,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-y"
          }
        ],
        "description": "Specifies the handling of overflow in the vertical direction.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "pad",
        "browsers": [
          "FF33"
        ],
        "syntax": "<integer> && <symbol>",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies a \u201Cfixed-width\u201D counter style, where representations shorter than the pad value are padded with a particular <symbol>",
        "restrictions": [
          "integer",
          "image",
          "string",
          "identifier"
        ]
      },
      {
        "name": "padding",
        "values": [],
        "syntax": "[ <length> | <percentage> ]{1,4}",
        "relevance": 96,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding"
          }
        ],
        "description": "Shorthand property to set values the thickness of the padding area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. The value may not be negative.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-bottom",
        "syntax": "<length> | <percentage>",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-bottom"
          }
        ],
        "description": "Shorthand property to set values the thickness of the padding area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. The value may not be negative.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-block-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'padding-left'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-block-end"
          }
        ],
        "description": "Logical 'padding-bottom'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-block-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'padding-left'>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-block-start"
          }
        ],
        "description": "Logical 'padding-top'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-inline-end",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'padding-left'>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-inline-end"
          }
        ],
        "description": "Logical 'padding-right'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-inline-start",
        "browsers": [
          "E79",
          "FF41",
          "S12.1",
          "C69",
          "O56"
        ],
        "syntax": "<'padding-left'>",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-inline-start"
          }
        ],
        "description": "Logical 'padding-left'. Mapping depends on the parent element\u2019s 'writing-mode', 'direction', and 'text-orientation'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-left",
        "syntax": "<length> | <percentage>",
        "relevance": 90,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-left"
          }
        ],
        "description": "Shorthand property to set values the thickness of the padding area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. The value may not be negative.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-right",
        "syntax": "<length> | <percentage>",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-right"
          }
        ],
        "description": "Shorthand property to set values the thickness of the padding area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. The value may not be negative.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "padding-top",
        "syntax": "<length> | <percentage>",
        "relevance": 90,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-top"
          }
        ],
        "description": "Shorthand property to set values the thickness of the padding area. If left is omitted, it is the same as right. If bottom is omitted it is the same as top, if right is omitted it is the same as top. The value may not be negative.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "page-break-after",
        "values": [
          {
            "name": "always",
            "description": "Always force a page break after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page break after generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page break after the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "right",
            "description": "Force one or two page breaks after the generated box so that the next page is formatted as a right page."
          }
        ],
        "syntax": "auto | always | avoid | left | right | recto | verso",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/page-break-after"
          }
        ],
        "description": "Defines rules for page breaks after an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "page-break-before",
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page break before the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page break before the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before the generated box so that the next page is formatted as a right page."
          }
        ],
        "syntax": "auto | always | avoid | left | right | recto | verso",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/page-break-before"
          }
        ],
        "description": "Defines rules for page breaks before an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "page-break-inside",
        "values": [
          {
            "name": "auto",
            "description": "Neither force nor forbid a page break inside the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page break inside the generated box."
          }
        ],
        "syntax": "auto | avoid",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/page-break-inside"
          }
        ],
        "description": "Defines rules for page breaks inside an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "paint-order",
        "browsers": [
          "E17",
          "FF60",
          "S8",
          "C35",
          "O22"
        ],
        "values": [
          {
            "name": "fill"
          },
          {
            "name": "markers"
          },
          {
            "name": "normal",
            "description": "The element is painted with the standard order of painting operations: the 'fill' is painted first, then its 'stroke' and finally its markers."
          },
          {
            "name": "stroke"
          }
        ],
        "syntax": "normal | [ fill || stroke || markers ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/paint-order"
          }
        ],
        "description": "Controls the order that the three paint operations that shapes and text are rendered with: their fill, their stroke and any markers they might have.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "perspective",
        "values": [
          {
            "name": "none",
            "description": "No perspective transform is applied."
          }
        ],
        "syntax": "none | <length>",
        "relevance": 55,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/perspective"
          }
        ],
        "description": "Applies the same transform as the perspective(<number>) transform function, except that it applies only to the positioned or transformed children of the element, not to the transform on the element itself.",
        "restrictions": [
          "length",
          "enum"
        ]
      },
      {
        "name": "perspective-origin",
        "syntax": "<position>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/perspective-origin"
          }
        ],
        "description": "Establishes the origin for the perspective property. It effectively sets the X and Y position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "pointer-events",
        "values": [
          {
            "name": "all",
            "description": "The given element can be the target element for pointer events whenever the pointer is over either the interior or the perimeter of the element."
          },
          {
            "name": "fill",
            "description": "The given element can be the target element for pointer events whenever the pointer is over the interior of the element."
          },
          {
            "name": "none",
            "description": "The given element does not receive pointer events."
          },
          {
            "name": "painted",
            "description": 'The given element can be the target element for pointer events when the pointer is over a "painted" area. '
          },
          {
            "name": "stroke",
            "description": "The given element can be the target element for pointer events whenever the pointer is over the perimeter of the element."
          },
          {
            "name": "visible",
            "description": "The given element can be the target element for pointer events when the \u2018visibility\u2019 property is set to visible and the pointer is over either the interior or the perimete of the element."
          },
          {
            "name": "visibleFill",
            "description": "The given element can be the target element for pointer events when the \u2018visibility\u2019 property is set to visible and when the pointer is over the interior of the element."
          },
          {
            "name": "visiblePainted",
            "description": "The given element can be the target element for pointer events when the \u2018visibility\u2019 property is set to visible and when the pointer is over a \u2018painted\u2019 area."
          },
          {
            "name": "visibleStroke",
            "description": "The given element can be the target element for pointer events when the \u2018visibility\u2019 property is set to visible and when the pointer is over the perimeter of the element."
          }
        ],
        "syntax": "auto | none | visiblePainted | visibleFill | visibleStroke | visible | painted | fill | stroke | all | inherit",
        "relevance": 80,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/pointer-events"
          }
        ],
        "description": "Specifies under what circumstances a given element can be the target element for a pointer event.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "position",
        "values": [
          {
            "name": "absolute",
            "description": "The box's position (and possibly size) is specified with the 'top', 'right', 'bottom', and 'left' properties. These properties specify offsets with respect to the box's 'containing block'."
          },
          {
            "name": "fixed",
            "description": "The box's position is calculated according to the 'absolute' model, but in addition, the box is fixed with respect to some reference. As with the 'absolute' model, the box's margins do not collapse with any other margins."
          },
          {
            "name": "-ms-page",
            "description": "The box's position is calculated according to the 'absolute' model."
          },
          {
            "name": "relative",
            "description": "The box's position is calculated according to the normal flow (this is called the position in normal flow). Then the box is offset relative to its normal position."
          },
          {
            "name": "static",
            "description": "The box is a normal box, laid out according to the normal flow. The 'top', 'right', 'bottom', and 'left' properties do not apply."
          },
          {
            "name": "sticky",
            "description": "The box's position is calculated according to the normal flow. Then the box is offset relative to its flow root and containing block and in all cases, including table elements, does not affect the position of any following boxes."
          },
          {
            "name": "-webkit-sticky",
            "description": "The box's position is calculated according to the normal flow. Then the box is offset relative to its flow root and containing block and in all cases, including table elements, does not affect the position of any following boxes."
          }
        ],
        "syntax": "static | relative | absolute | sticky | fixed",
        "relevance": 96,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/position"
          }
        ],
        "description": "The position CSS property sets how an element is positioned in a document. The top, right, bottom, and left properties determine the final location of positioned elements.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "prefix",
        "browsers": [
          "FF33"
        ],
        "syntax": "<symbol>",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies a <symbol> that is prepended to the marker representation.",
        "restrictions": [
          "image",
          "string",
          "identifier"
        ]
      },
      {
        "name": "quotes",
        "values": [
          {
            "name": "none",
            "description": "The 'open-quote' and 'close-quote' values of the 'content' property produce no quotations marks, as if they were 'no-open-quote' and 'no-close-quote' respectively."
          }
        ],
        "syntax": "none | auto | [ <string> <string> ]+",
        "relevance": 53,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/quotes"
          }
        ],
        "description": "Specifies quotation marks for any number of embedded quotations.",
        "restrictions": [
          "string"
        ]
      },
      {
        "name": "range",
        "browsers": [
          "FF33"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The range depends on the counter system."
          },
          {
            "name": "infinite",
            "description": "If used as the first value in a range, it represents negative infinity; if used as the second value, it represents positive infinity."
          }
        ],
        "syntax": "[ [ <integer> | infinite ]{2} ]# | auto",
        "relevance": 50,
        "description": "@counter-style descriptor. Defines the ranges over which the counter style is defined.",
        "restrictions": [
          "integer",
          "enum"
        ]
      },
      {
        "name": "resize",
        "browsers": [
          "E79",
          "FF4",
          "S3",
          "C1",
          "O12.1"
        ],
        "values": [
          {
            "name": "both",
            "description": "The UA presents a bidirectional resizing mechanism to allow the user to adjust both the height and the width of the element."
          },
          {
            "name": "horizontal",
            "description": "The UA presents a unidirectional horizontal resizing mechanism to allow the user to adjust only the width of the element."
          },
          {
            "name": "none",
            "description": "The UA does not present a resizing mechanism on the element, and the user is given no direct manipulation mechanism to resize the element."
          },
          {
            "name": "vertical",
            "description": "The UA presents a unidirectional vertical resizing mechanism to allow the user to adjust only the height of the element."
          }
        ],
        "syntax": "none | both | horizontal | vertical | block | inline",
        "relevance": 60,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/resize"
          }
        ],
        "description": "Specifies whether or not an element is resizable by the user, and if so, along which axis/axes.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "right",
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 90,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/right"
          }
        ],
        "description": "Specifies how far an absolutely positioned box's right margin edge is offset to the left of the right edge of the box's 'containing block'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "ruby-align",
        "browsers": [
          "FF38"
        ],
        "values": [
          {
            "name": "auto",
            "browsers": [
              "FF38"
            ],
            "description": "The user agent determines how the ruby contents are aligned. This is the initial value."
          },
          {
            "name": "center",
            "description": "The ruby content is centered within its box."
          },
          {
            "name": "distribute-letter",
            "browsers": [
              "FF38"
            ],
            "description": "If the width of the ruby text is smaller than that of the base, then the ruby text contents are evenly distributed across the width of the base, with the first and last ruby text glyphs lining up with the corresponding first and last base glyphs. If the width of the ruby text is at least the width of the base, then the letters of the base are evenly distributed across the width of the ruby text."
          },
          {
            "name": "distribute-space",
            "browsers": [
              "FF38"
            ],
            "description": "If the width of the ruby text is smaller than that of the base, then the ruby text contents are evenly distributed across the width of the base, with a certain amount of white space preceding the first and following the last character in the ruby text. That amount of white space is normally equal to half the amount of inter-character space of the ruby text."
          },
          {
            "name": "left",
            "description": "The ruby text content is aligned with the start edge of the base."
          },
          {
            "name": "line-edge",
            "browsers": [
              "FF38"
            ],
            "description": "If the ruby text is not adjacent to a line edge, it is aligned as in 'auto'. If it is adjacent to a line edge, then it is still aligned as in auto, but the side of the ruby text that touches the end of the line is lined up with the corresponding edge of the base."
          },
          {
            "name": "right",
            "browsers": [
              "FF38"
            ],
            "description": "The ruby text content is aligned with the end edge of the base."
          },
          {
            "name": "start",
            "browsers": [
              "FF38"
            ],
            "description": "The ruby text content is aligned with the start edge of the base."
          },
          {
            "name": "space-between",
            "browsers": [
              "FF38"
            ],
            "description": "The ruby content expands as defined for normal text justification (as defined by 'text-justify'),"
          },
          {
            "name": "space-around",
            "browsers": [
              "FF38"
            ],
            "description": "As for 'space-between' except that there exists an extra justification opportunities whose space is distributed half before and half after the ruby content."
          }
        ],
        "status": "experimental",
        "syntax": "start | center | space-between | space-around",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/ruby-align"
          }
        ],
        "description": "Specifies how text is distributed within the various ruby boxes when their contents do not exactly fill their respective boxes.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "ruby-overhang",
        "browsers": [
          "FF10",
          "IE5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The ruby text can overhang text adjacent to the base on either side. This is the initial value."
          },
          {
            "name": "end",
            "description": "The ruby text can overhang the text that follows it."
          },
          {
            "name": "none",
            "description": "The ruby text cannot overhang any text adjacent to its base, only its own base."
          },
          {
            "name": "start",
            "description": "The ruby text can overhang the text that precedes it."
          }
        ],
        "relevance": 50,
        "description": "Determines whether, and on which side, ruby text is allowed to partially overhang any adjacent text in addition to its own base, when the ruby text is wider than the ruby base.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "ruby-position",
        "browsers": [
          "E12",
          "FF38"
        ],
        "values": [
          {
            "name": "after",
            "description": "The ruby text appears after the base. This is a relatively rare setting used in ideographic East Asian writing systems, most easily found in educational text."
          },
          {
            "name": "before",
            "description": "The ruby text appears before the base. This is the most common setting used in ideographic East Asian writing systems."
          },
          {
            "name": "inline"
          },
          {
            "name": "right",
            "description": "The ruby text appears on the right of the base. Unlike 'before' and 'after', this value is not relative to the text flow direction."
          }
        ],
        "status": "experimental",
        "syntax": "over | under | inter-character",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/ruby-position"
          }
        ],
        "description": "Used by the parent of elements with display: ruby-text to control the position of the ruby text with respect to its base.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "ruby-span",
        "browsers": [
          "FF10"
        ],
        "values": [
          {
            "name": "attr(x)",
            "description": "The value of attribute 'x' is a string value. The string value is evaluated as a <number> to determine the number of ruby base elements to be spanned by the annotation element."
          },
          {
            "name": "none",
            "description": "No spanning. The computed value is '1'."
          }
        ],
        "relevance": 50,
        "description": "Determines whether, and on which side, ruby text is allowed to partially overhang any adjacent text in addition to its own base, when the ruby text is wider than the ruby base.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "scrollbar-3dlight-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-3dlight-color"
          }
        ],
        "description": "Determines the color of the top and left edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-arrow-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-arrow-color"
          }
        ],
        "description": "Determines the color of the arrow elements of a scroll arrow.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-base-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-base-color"
          }
        ],
        "description": "Determines the color of the main elements of a scroll bar, which include the scroll box, track, and scroll arrows.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-darkshadow-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-darkshadow-color"
          }
        ],
        "description": "Determines the color of the gutter of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-face-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-face-color"
          }
        ],
        "description": "Determines the color of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-highlight-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-highlight-color"
          }
        ],
        "description": "Determines the color of the top and left edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-shadow-color",
        "browsers": [
          "IE5"
        ],
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-shadow-color"
          }
        ],
        "description": "Determines the color of the bottom and right edges of the scroll box and scroll arrows of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scrollbar-track-color",
        "browsers": [
          "IE6"
        ],
        "relevance": 50,
        "description": "Determines the color of the track element of a scroll bar.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "scroll-behavior",
        "browsers": [
          "E79",
          "FF36",
          "C61",
          "O48"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Scrolls in an instant fashion."
          },
          {
            "name": "smooth",
            "description": "Scrolls in a smooth fashion using a user-agent-defined timing function and time period."
          }
        ],
        "syntax": "auto | smooth",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-behavior"
          }
        ],
        "description": "Specifies the scrolling behavior for a scrolling box, when scrolling happens due to navigation or CSSOM scrolling APIs.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "scroll-snap-coordinate",
        "browsers": [
          "FF39"
        ],
        "values": [
          {
            "name": "none",
            "description": "Specifies that this element does not contribute a snap point."
          }
        ],
        "status": "obsolete",
        "syntax": "none | <position>#",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-coordinate"
          }
        ],
        "description": "Defines the x and y coordinate within the element which will align with the nearest ancestor scroll container\u2019s snap-destination for the respective axis.",
        "restrictions": [
          "position",
          "length",
          "percentage",
          "enum"
        ]
      },
      {
        "name": "scroll-snap-destination",
        "browsers": [
          "FF39"
        ],
        "status": "obsolete",
        "syntax": "<position>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-destination"
          }
        ],
        "description": "Define the x and y coordinate within the scroll container\u2019s visual viewport which element snap points will align with.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "scroll-snap-points-x",
        "browsers": [
          "FF39",
          "S9"
        ],
        "values": [
          {
            "name": "none",
            "description": "No snap points are defined by this scroll container."
          },
          {
            "name": "repeat()",
            "description": "Defines an interval at which snap points are defined, starting from the container\u2019s relevant start edge."
          }
        ],
        "status": "obsolete",
        "syntax": "none | repeat( <length-percentage> )",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-points-x"
          }
        ],
        "description": "Defines the positioning of snap points along the x axis of the scroll container it is applied to.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "scroll-snap-points-y",
        "browsers": [
          "FF39",
          "S9"
        ],
        "values": [
          {
            "name": "none",
            "description": "No snap points are defined by this scroll container."
          },
          {
            "name": "repeat()",
            "description": "Defines an interval at which snap points are defined, starting from the container\u2019s relevant start edge."
          }
        ],
        "status": "obsolete",
        "syntax": "none | repeat( <length-percentage> )",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-points-y"
          }
        ],
        "description": "Defines the positioning of snap points along the y axis of the scroll container it is applied to.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "scroll-snap-type",
        "values": [
          {
            "name": "none",
            "description": "The visual viewport of this scroll container must ignore snap points, if any, when scrolled."
          },
          {
            "name": "mandatory",
            "description": "The visual viewport of this scroll container is guaranteed to rest on a snap point when there are no active scrolling operations."
          },
          {
            "name": "proximity",
            "description": "The visual viewport of this scroll container may come to rest on a snap point at the termination of a scroll at the discretion of the UA given the parameters of the scroll."
          }
        ],
        "syntax": "none | [ x | y | block | inline | both ] [ mandatory | proximity ]?",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-type"
          }
        ],
        "description": "Defines how strictly snap points are enforced on the scroll container.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "shape-image-threshold",
        "browsers": [
          "E79",
          "FF62",
          "S10.1",
          "C37",
          "O24"
        ],
        "syntax": "<alpha-value>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/shape-image-threshold"
          }
        ],
        "description": "Defines the alpha channel threshold used to extract the shape using an image. A value of 0.5 means that the shape will enclose all the pixels that are more than 50% opaque.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "shape-margin",
        "browsers": [
          "E79",
          "FF62",
          "S10.1",
          "C37",
          "O24"
        ],
        "syntax": "<length-percentage>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/shape-margin"
          }
        ],
        "description": "Adds a margin to a 'shape-outside'. This defines a new shape that is the smallest contour that includes all the points that are the 'shape-margin' distance outward in the perpendicular direction from a point on the underlying shape.",
        "restrictions": [
          "url",
          "length",
          "percentage"
        ]
      },
      {
        "name": "shape-outside",
        "browsers": [
          "E79",
          "FF62",
          "S10.1",
          "C37",
          "O24"
        ],
        "values": [
          {
            "name": "margin-box",
            "description": "The background is painted within (clipped to) the margin box."
          },
          {
            "name": "none",
            "description": "The float area is unaffected."
          }
        ],
        "syntax": "none | <shape-box> || <basic-shape> | <image>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/shape-outside"
          }
        ],
        "description": "Specifies an orthogonal rotation to be applied to an image before it is laid out.",
        "restrictions": [
          "image",
          "box",
          "shape",
          "enum"
        ]
      },
      {
        "name": "shape-rendering",
        "values": [
          {
            "name": "auto",
            "description": "Suppresses aural rendering."
          },
          {
            "name": "crispEdges",
            "description": "Emphasize the contrast between clean edges of artwork over rendering speed and geometric precision."
          },
          {
            "name": "geometricPrecision",
            "description": "Emphasize geometric precision over speed and crisp edges."
          },
          {
            "name": "optimizeSpeed",
            "description": "Emphasize rendering speed over geometric precision and crisp edges."
          }
        ],
        "relevance": 50,
        "description": "Provides hints about what tradeoffs to make as it renders vector graphics elements such as <path> elements and basic shapes such as circles and rectangles.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "size",
        "browsers": [
          "C",
          "O8"
        ],
        "syntax": "<length>{1,2} | auto | [ <page-size> || [ portrait | landscape ] ]",
        "relevance": 52,
        "description": "The size CSS at-rule descriptor, used with the @page at-rule, defines the size and orientation of the box which is used to represent a page. Most of the time, this size corresponds to the target size of the printed page if applicable.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "src",
        "values": [
          {
            "name": "url()",
            "description": "Reference font by URL"
          },
          {
            "name": "format()",
            "description": "Optional hint describing the format of the font resource."
          },
          {
            "name": "local()",
            "description": "Format-specific string that identifies a locally available copy of a given font."
          }
        ],
        "syntax": "[ <url> [ format( <string># ) ]? | local( <family-name> ) ]#",
        "relevance": 67,
        "description": "@font-face descriptor. Specifies the resource containing font data. It is required, whether the font is downloadable or locally installed.",
        "restrictions": [
          "enum",
          "url",
          "identifier"
        ]
      },
      {
        "name": "stop-color",
        "relevance": 51,
        "description": "Indicates what color to use at that gradient stop.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "stop-opacity",
        "relevance": 50,
        "description": "Defines the opacity of a given gradient stop.",
        "restrictions": [
          "number(0-1)"
        ]
      },
      {
        "name": "stroke",
        "values": [
          {
            "name": "url()",
            "description": "A URL reference to a paint server element, which is an element that defines a paint server: \u2018hatch\u2019, \u2018linearGradient\u2019, \u2018mesh\u2019, \u2018pattern\u2019, \u2018radialGradient\u2019 and \u2018solidcolor\u2019."
          },
          {
            "name": "none",
            "description": "No paint is applied in this layer."
          }
        ],
        "relevance": 63,
        "description": "Paints along the outline of the given graphical element.",
        "restrictions": [
          "color",
          "enum",
          "url"
        ]
      },
      {
        "name": "stroke-dasharray",
        "values": [
          {
            "name": "none",
            "description": "Indicates that no dashing is used."
          }
        ],
        "relevance": 58,
        "description": "Controls the pattern of dashes and gaps used to stroke paths.",
        "restrictions": [
          "length",
          "percentage",
          "number",
          "enum"
        ]
      },
      {
        "name": "stroke-dashoffset",
        "relevance": 58,
        "description": "Specifies the distance into the dash pattern to start the dash.",
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "stroke-linecap",
        "values": [
          {
            "name": "butt",
            "description": "Indicates that the stroke for each subpath does not extend beyond its two endpoints."
          },
          {
            "name": "round",
            "description": "Indicates that at each end of each subpath, the shape representing the stroke will be extended by a half circle with a radius equal to the stroke width."
          },
          {
            "name": "square",
            "description": "Indicates that at the end of each subpath, the shape representing the stroke will be extended by a rectangle with the same width as the stroke width and whose length is half of the stroke width."
          }
        ],
        "relevance": 53,
        "description": "Specifies the shape to be used at the end of open subpaths when they are stroked.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "stroke-linejoin",
        "values": [
          {
            "name": "bevel",
            "description": "Indicates that a bevelled corner is to be used to join path segments."
          },
          {
            "name": "miter",
            "description": "Indicates that a sharp corner is to be used to join path segments."
          },
          {
            "name": "round",
            "description": "Indicates that a round corner is to be used to join path segments."
          }
        ],
        "relevance": 50,
        "description": "Specifies the shape to be used at the corners of paths or basic shapes when they are stroked.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "stroke-miterlimit",
        "relevance": 50,
        "description": "When two line segments meet at a sharp angle and miter joins have been specified for 'stroke-linejoin', it is possible for the miter to extend far beyond the thickness of the line stroking the path.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "stroke-opacity",
        "relevance": 51,
        "description": "Specifies the opacity of the painting operation used to stroke the current object.",
        "restrictions": [
          "number(0-1)"
        ]
      },
      {
        "name": "stroke-width",
        "relevance": 60,
        "description": "Specifies the width of the stroke on the current object.",
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "suffix",
        "browsers": [
          "FF33"
        ],
        "syntax": "<symbol>",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies a <symbol> that is appended to the marker representation.",
        "restrictions": [
          "image",
          "string",
          "identifier"
        ]
      },
      {
        "name": "system",
        "browsers": [
          "FF33"
        ],
        "values": [
          {
            "name": "additive",
            "description": "Represents \u201Csign-value\u201D numbering systems, which, rather than using reusing digits in different positions to change their value, define additional digits with much larger values, so that the value of the number can be obtained by adding all the digits together."
          },
          {
            "name": "alphabetic",
            "description": 'Interprets the list of counter symbols as digits to an alphabetic numbering system, similar to the default lower-alpha counter style, which wraps from "a", "b", "c", to "aa", "ab", "ac".'
          },
          {
            "name": "cyclic",
            "description": "Cycles repeatedly through its provided symbols, looping back to the beginning when it reaches the end of the list."
          },
          {
            "name": "extends",
            "description": "Use the algorithm of another counter style, but alter other aspects."
          },
          {
            "name": "fixed",
            "description": "Runs through its list of counter symbols once, then falls back."
          },
          {
            "name": "numeric",
            "description": `interprets the list of counter symbols as digits to a "place-value" numbering system, similar to the default 'decimal' counter style.`
          },
          {
            "name": "symbolic",
            "description": "Cycles repeatedly through its provided symbols, doubling, tripling, etc. the symbols on each successive pass through the list."
          }
        ],
        "syntax": "cyclic | numeric | alphabetic | symbolic | additive | [ fixed <integer>? ] | [ extends <counter-style-name> ]",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies which algorithm will be used to construct the counter\u2019s representation based on the counter value.",
        "restrictions": [
          "enum",
          "integer"
        ]
      },
      {
        "name": "symbols",
        "browsers": [
          "FF33"
        ],
        "syntax": "<symbol>+",
        "relevance": 50,
        "description": "@counter-style descriptor. Specifies the symbols used by the marker-construction algorithm specified by the system descriptor.",
        "restrictions": [
          "image",
          "string",
          "identifier"
        ]
      },
      {
        "name": "table-layout",
        "values": [
          {
            "name": "auto",
            "description": "Use any automatic table layout algorithm."
          },
          {
            "name": "fixed",
            "description": "Use the fixed table layout algorithm."
          }
        ],
        "syntax": "auto | fixed",
        "relevance": 61,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/table-layout"
          }
        ],
        "description": "Controls the algorithm used to lay out the table cells, rows, and columns.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "tab-size",
        "browsers": [
          "E79",
          "FF4",
          "S6.1",
          "C21",
          "O15"
        ],
        "syntax": "<integer> | <length>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/tab-size"
          }
        ],
        "description": "Determines the width of the tab character (U+0009), in space characters (U+0020), when rendered.",
        "restrictions": [
          "integer",
          "length"
        ]
      },
      {
        "name": "text-align",
        "values": [
          {
            "name": "center",
            "description": "The inline contents are centered within the line box."
          },
          {
            "name": "end",
            "description": "The inline contents are aligned to the end edge of the line box."
          },
          {
            "name": "justify",
            "description": "The text is justified according to the method specified by the 'text-justify' property."
          },
          {
            "name": "left",
            "description": "The inline contents are aligned to the left edge of the line box. In vertical text, 'left' aligns to the edge of the line box that would be the start edge for left-to-right text."
          },
          {
            "name": "right",
            "description": "The inline contents are aligned to the right edge of the line box. In vertical text, 'right' aligns to the edge of the line box that would be the end edge for left-to-right text."
          },
          {
            "name": "start",
            "description": "The inline contents are aligned to the start edge of the line box."
          }
        ],
        "syntax": "start | end | left | right | center | justify | match-parent",
        "relevance": 93,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-align"
          }
        ],
        "description": "Describes how inline contents of a block are horizontally aligned if the contents do not completely fill the line box.",
        "restrictions": [
          "string"
        ]
      },
      {
        "name": "text-align-last",
        "browsers": [
          "E12",
          "FF49",
          "C47",
          "IE5.5",
          "O34"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Content on the affected line is aligned per 'text-align' unless 'text-align' is set to 'justify', in which case it is 'start-aligned'."
          },
          {
            "name": "center",
            "description": "The inline contents are centered within the line box."
          },
          {
            "name": "justify",
            "description": "The text is justified according to the method specified by the 'text-justify' property."
          },
          {
            "name": "left",
            "description": "The inline contents are aligned to the left edge of the line box. In vertical text, 'left' aligns to the edge of the line box that would be the start edge for left-to-right text."
          },
          {
            "name": "right",
            "description": "The inline contents are aligned to the right edge of the line box. In vertical text, 'right' aligns to the edge of the line box that would be the end edge for left-to-right text."
          }
        ],
        "syntax": "auto | start | end | left | right | center | justify",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-align-last"
          }
        ],
        "description": "Describes how the last line of a block or a line right before a forced line break is aligned when 'text-align' is set to 'justify'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-anchor",
        "values": [
          {
            "name": "end",
            "description": "The rendered characters are aligned such that the end of the resulting rendered text is at the initial current text position."
          },
          {
            "name": "middle",
            "description": "The rendered characters are aligned such that the geometric middle of the resulting rendered text is at the initial current text position."
          },
          {
            "name": "start",
            "description": "The rendered characters are aligned such that the start of the resulting rendered text is at the initial current text position."
          }
        ],
        "relevance": 50,
        "description": "Used to align (start-, middle- or end-alignment) a string of text relative to a given point.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-decoration",
        "values": [
          {
            "name": "dashed",
            "description": "Produces a dashed line style."
          },
          {
            "name": "dotted",
            "description": "Produces a dotted line."
          },
          {
            "name": "double",
            "description": "Produces a double line."
          },
          {
            "name": "line-through",
            "description": "Each line of text has a line through the middle."
          },
          {
            "name": "none",
            "description": "Produces no line."
          },
          {
            "name": "overline",
            "description": "Each line of text has a line above it."
          },
          {
            "name": "solid",
            "description": "Produces a solid line."
          },
          {
            "name": "underline",
            "description": "Each line of text is underlined."
          },
          {
            "name": "wavy",
            "description": "Produces a wavy line."
          }
        ],
        "syntax": "<'text-decoration-line'> || <'text-decoration-style'> || <'text-decoration-color'> || <'text-decoration-thickness'>",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration"
          }
        ],
        "description": "Decorations applied to font used for an element's text.",
        "restrictions": [
          "enum",
          "color"
        ]
      },
      {
        "name": "text-decoration-color",
        "browsers": [
          "E79",
          "FF36",
          "S12.1",
          "C57",
          "O44"
        ],
        "syntax": "<color>",
        "relevance": 52,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-color"
          }
        ],
        "description": "Specifies the color of text decoration (underlines overlines, and line-throughs) set on the element with text-decoration-line.",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "text-decoration-line",
        "browsers": [
          "E79",
          "FF36",
          "S12.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "line-through",
            "description": "Each line of text has a line through the middle."
          },
          {
            "name": "none",
            "description": "Neither produces nor inhibits text decoration."
          },
          {
            "name": "overline",
            "description": "Each line of text has a line above it."
          },
          {
            "name": "underline",
            "description": "Each line of text is underlined."
          }
        ],
        "syntax": "none | [ underline || overline || line-through || blink ] | spelling-error | grammar-error",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-line"
          }
        ],
        "description": "Specifies what line decorations, if any, are added to the element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-decoration-style",
        "browsers": [
          "E79",
          "FF36",
          "S12.1",
          "C57",
          "O44"
        ],
        "values": [
          {
            "name": "dashed",
            "description": "Produces a dashed line style."
          },
          {
            "name": "dotted",
            "description": "Produces a dotted line."
          },
          {
            "name": "double",
            "description": "Produces a double line."
          },
          {
            "name": "none",
            "description": "Produces no line."
          },
          {
            "name": "solid",
            "description": "Produces a solid line."
          },
          {
            "name": "wavy",
            "description": "Produces a wavy line."
          }
        ],
        "syntax": "solid | double | dotted | dashed | wavy",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-style"
          }
        ],
        "description": "Specifies the line style for underline, line-through and overline text decoration.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-indent",
        "values": [],
        "syntax": "<length-percentage> && hanging? && each-line?",
        "relevance": 68,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-indent"
          }
        ],
        "description": "Specifies the indentation applied to lines of inline content in a block. The indentation only affects the first line of inline content in the block unless the 'hanging' keyword is specified, in which case it affects all lines except the first.",
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "text-justify",
        "browsers": [
          "E12",
          "FF55",
          "C32",
          "IE11",
          "O19"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The UA determines the justification algorithm to follow, based on a balance between performance and adequate presentation quality."
          },
          {
            "name": "distribute",
            "description": "Justification primarily changes spacing both at word separators and at grapheme cluster boundaries in all scripts except those in the connected and cursive groups. This value is sometimes used in e.g. Japanese, often with the 'text-align-last' property."
          },
          {
            "name": "distribute-all-lines"
          },
          {
            "name": "inter-cluster",
            "description": "Justification primarily changes spacing at word separators and at grapheme cluster boundaries in clustered scripts. This value is typically used for Southeast Asian scripts such as Thai."
          },
          {
            "name": "inter-ideograph",
            "description": "Justification primarily changes spacing at word separators and at inter-graphemic boundaries in scripts that use no word spaces. This value is typically used for CJK languages."
          },
          {
            "name": "inter-word",
            "description": "Justification primarily changes spacing at word separators. This value is typically used for languages that separate words using spaces, like English or (sometimes) Korean."
          },
          {
            "name": "kashida",
            "description": "Justification primarily stretches Arabic and related scripts through the use of kashida or other calligraphic elongation."
          },
          {
            "name": "newspaper"
          }
        ],
        "syntax": "auto | inter-character | inter-word | none",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-justify"
          }
        ],
        "description": "Selects the justification algorithm used when 'text-align' is set to 'justify'. The property applies to block containers, but the UA may (but is not required to) also support it on inline elements.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-orientation",
        "browsers": [
          "E79",
          "FF41",
          "S5.1",
          "C48",
          "O15"
        ],
        "values": [
          {
            "name": "sideways",
            "browsers": [
              "E79",
              "FF41",
              "S5.1",
              "C48",
              "O15"
            ],
            "description": "This value is equivalent to 'sideways-right' in 'vertical-rl' writing mode and equivalent to 'sideways-left' in 'vertical-lr' writing mode."
          },
          {
            "name": "sideways-right",
            "browsers": [
              "E79",
              "FF41",
              "S5.1",
              "C48",
              "O15"
            ],
            "description": "In vertical writing modes, this causes text to be set as if in a horizontal layout, but rotated 90\xB0 clockwise."
          },
          {
            "name": "upright",
            "description": "In vertical writing modes, characters from horizontal-only scripts are rendered upright, i.e. in their standard horizontal orientation."
          }
        ],
        "syntax": "mixed | upright | sideways",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-orientation"
          }
        ],
        "description": "Specifies the orientation of text within a line.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-overflow",
        "values": [
          {
            "name": "clip",
            "description": "Clip inline content that overflows. Characters may be only partially rendered."
          },
          {
            "name": "ellipsis",
            "description": "Render an ellipsis character (U+2026) to represent clipped inline content."
          }
        ],
        "syntax": "[ clip | ellipsis | <string> ]{1,2}",
        "relevance": 81,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-overflow"
          }
        ],
        "description": "Text can overflow for example when it is prevented from wrapping.",
        "restrictions": [
          "enum",
          "string"
        ]
      },
      {
        "name": "text-rendering",
        "browsers": [
          "E79",
          "FF1",
          "S5",
          "C4",
          "O15"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "geometricPrecision",
            "description": "Indicates that the user agent shall emphasize geometric precision over legibility and rendering speed."
          },
          {
            "name": "optimizeLegibility",
            "description": "Indicates that the user agent shall emphasize legibility over rendering speed and geometric precision."
          },
          {
            "name": "optimizeSpeed",
            "description": "Indicates that the user agent shall emphasize rendering speed over legibility and geometric precision."
          }
        ],
        "syntax": "auto | optimizeSpeed | optimizeLegibility | geometricPrecision",
        "relevance": 68,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-rendering"
          }
        ],
        "description": "The creator of SVG content might want to provide a hint to the implementation about what tradeoffs to make as it renders text. The \u2018text-rendering\u2019 property provides these hints.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-shadow",
        "values": [
          {
            "name": "none",
            "description": "No shadow."
          }
        ],
        "syntax": "none | <shadow-t>#",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-shadow"
          }
        ],
        "description": "Enables shadow effects to be applied to the text of the element.",
        "restrictions": [
          "length",
          "color"
        ]
      },
      {
        "name": "text-transform",
        "values": [
          {
            "name": "capitalize",
            "description": "Puts the first typographic letter unit of each word in titlecase."
          },
          {
            "name": "lowercase",
            "description": "Puts all letters in lowercase."
          },
          {
            "name": "none",
            "description": "No effects."
          },
          {
            "name": "uppercase",
            "description": "Puts all letters in uppercase."
          }
        ],
        "syntax": "none | capitalize | uppercase | lowercase | full-width | full-size-kana",
        "relevance": 84,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-transform"
          }
        ],
        "description": "Controls capitalization effects of an element\u2019s text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "text-underline-position",
        "values": [
          {
            "name": "above"
          },
          {
            "name": "auto",
            "description": "The user agent may use any algorithm to determine the underline\u2019s position. In horizontal line layout, the underline should be aligned as for alphabetic. In vertical line layout, if the language is set to Japanese or Korean, the underline should be aligned as for over."
          },
          {
            "name": "below",
            "description": "The underline is aligned with the under edge of the element\u2019s content box."
          }
        ],
        "syntax": "auto | from-font | [ under || [ left | right ] ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-underline-position"
          }
        ],
        "description": "Sets the position of an underline specified on the same element: it does not affect underlines specified by ancestor elements. This property is typically used in vertical writing contexts such as in Japanese documents where it often desired to have the underline appear 'over' (to the right of) the affected run of text",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "top",
        "values": [
          {
            "name": "auto",
            "description": "For non-replaced elements, the effect of this value depends on which of related properties have the value 'auto' as well"
          }
        ],
        "syntax": "<length> | <percentage> | auto",
        "relevance": 95,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/top"
          }
        ],
        "description": "Specifies how far an absolutely positioned box's top margin edge is offset below the top edge of the box's 'containing block'.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "touch-action",
        "values": [
          {
            "name": "auto",
            "description": "The user agent may determine any permitted touch behaviors for touches that begin on the element."
          },
          {
            "name": "cross-slide-x"
          },
          {
            "name": "cross-slide-y"
          },
          {
            "name": "double-tap-zoom"
          },
          {
            "name": "manipulation",
            "description": "The user agent may consider touches that begin on the element only for the purposes of scrolling and continuous zooming."
          },
          {
            "name": "none",
            "description": "Touches that begin on the element must not trigger default touch behaviors."
          },
          {
            "name": "pan-x",
            "description": "The user agent may consider touches that begin on the element only for the purposes of horizontally scrolling the element\u2019s nearest ancestor with horizontally scrollable content."
          },
          {
            "name": "pan-y",
            "description": "The user agent may consider touches that begin on the element only for the purposes of vertically scrolling the element\u2019s nearest ancestor with vertically scrollable content."
          },
          {
            "name": "pinch-zoom"
          }
        ],
        "syntax": "auto | none | [ [ pan-x | pan-left | pan-right ] || [ pan-y | pan-up | pan-down ] || pinch-zoom ] | manipulation",
        "relevance": 65,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/touch-action"
          }
        ],
        "description": "Determines whether touch input may trigger default behavior supplied by user agent.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "transform",
        "values": [
          {
            "name": "matrix()",
            "description": "Specifies a 2D transformation in the form of a transformation matrix of six values. matrix(a,b,c,d,e,f) is equivalent to applying the transformation matrix [a b c d e f]"
          },
          {
            "name": "matrix3d()",
            "description": "Specifies a 3D transformation as a 4x4 homogeneous matrix of 16 values in column-major order."
          },
          {
            "name": "none"
          },
          {
            "name": "perspective()",
            "description": "Specifies a perspective projection matrix."
          },
          {
            "name": "rotate()",
            "description": "Specifies a 2D rotation by the angle specified in the parameter about the origin of the element, as defined by the transform-origin property."
          },
          {
            "name": "rotate3d()",
            "description": "Specifies a clockwise 3D rotation by the angle specified in last parameter about the [x,y,z] direction vector described by the first 3 parameters."
          },
          {
            "name": "rotateX('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the X axis."
          },
          {
            "name": "rotateY('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Y axis."
          },
          {
            "name": "rotateZ('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Z axis."
          },
          {
            "name": "scale()",
            "description": "Specifies a 2D scale operation by the [sx,sy] scaling vector described by the 2 parameters. If the second parameter is not provided, it is takes a value equal to the first."
          },
          {
            "name": "scale3d()",
            "description": "Specifies a 3D scale operation by the [sx,sy,sz] scaling vector described by the 3 parameters."
          },
          {
            "name": "scaleX()",
            "description": "Specifies a scale operation using the [sx,1] scaling vector, where sx is given as the parameter."
          },
          {
            "name": "scaleY()",
            "description": "Specifies a scale operation using the [sy,1] scaling vector, where sy is given as the parameter."
          },
          {
            "name": "scaleZ()",
            "description": "Specifies a scale operation using the [1,1,sz] scaling vector, where sz is given as the parameter."
          },
          {
            "name": "skew()",
            "description": "Specifies a skew transformation along the X and Y axes. The first angle parameter specifies the skew on the X axis. The second angle parameter specifies the skew on the Y axis. If the second parameter is not given then a value of 0 is used for the Y angle (ie: no skew on the Y axis)."
          },
          {
            "name": "skewX()",
            "description": "Specifies a skew transformation along the X axis by the given angle."
          },
          {
            "name": "skewY()",
            "description": "Specifies a skew transformation along the Y axis by the given angle."
          },
          {
            "name": "translate()",
            "description": "Specifies a 2D translation by the vector [tx, ty], where tx is the first translation-value parameter and ty is the optional second translation-value parameter."
          },
          {
            "name": "translate3d()",
            "description": "Specifies a 3D translation by the vector [tx,ty,tz], with tx, ty and tz being the first, second and third translation-value parameters respectively."
          },
          {
            "name": "translateX()",
            "description": "Specifies a translation by the given amount in the X direction."
          },
          {
            "name": "translateY()",
            "description": "Specifies a translation by the given amount in the Y direction."
          },
          {
            "name": "translateZ()",
            "description": "Specifies a translation by the given amount in the Z direction. Note that percentage values are not allowed in the translateZ translation-value, and if present are evaluated as 0."
          }
        ],
        "syntax": "none | <transform-list>",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transform"
          }
        ],
        "description": "A two-dimensional transformation is applied to an element through the 'transform' property. This property contains a list of transform functions similar to those allowed by SVG.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "transform-origin",
        "syntax": "[ <length-percentage> | left | center | right | top | bottom ] | [ [ <length-percentage> | left | center | right ] && [ <length-percentage> | top | center | bottom ] ] <length>?",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transform-origin"
          }
        ],
        "description": "Establishes the origin of transformation for an element.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "transform-style",
        "browsers": [
          "E12",
          "FF16",
          "S9",
          "C36",
          "O23"
        ],
        "values": [
          {
            "name": "flat",
            "description": "All children of this element are rendered flattened into the 2D plane of the element."
          },
          {
            "name": "preserve-3d",
            "browsers": [
              "E12",
              "FF16",
              "S9",
              "C36",
              "O23"
            ],
            "description": "Flattening is not performed, so children maintain their position in 3D space."
          }
        ],
        "syntax": "flat | preserve-3d",
        "relevance": 54,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transform-style"
          }
        ],
        "description": "Defines how nested elements are rendered in 3D space.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "transition",
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "syntax": "<single-transition>#",
        "relevance": 87,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transition"
          }
        ],
        "description": "Shorthand property combines four of the transition properties into a single property.",
        "restrictions": [
          "time",
          "property",
          "timing-function",
          "enum"
        ]
      },
      {
        "name": "transition-delay",
        "syntax": "<time>#",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transition-delay"
          }
        ],
        "description": "Defines when the transition will start. It allows a transition to begin execution some period of time from when it is applied.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "transition-duration",
        "syntax": "<time>#",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transition-duration"
          }
        ],
        "description": "Specifies how long the transition from the old value to the new value should take.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "transition-property",
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "syntax": "none | <single-transition-property>#",
        "relevance": 64,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transition-property"
          }
        ],
        "description": "Specifies the name of the CSS property to which the transition is applied.",
        "restrictions": [
          "property"
        ]
      },
      {
        "name": "transition-timing-function",
        "syntax": "<timing-function>#",
        "relevance": 60,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transition-timing-function"
          }
        ],
        "description": "Describes how the intermediate values used during a transition will be calculated.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "unicode-bidi",
        "values": [
          {
            "name": "bidi-override",
            "description": "Inside the element, reordering is strictly in sequence according to the 'direction' property; the implicit part of the bidirectional algorithm is ignored."
          },
          {
            "name": "embed",
            "description": "If the element is inline-level, this value opens an additional level of embedding with respect to the bidirectional algorithm. The direction of this embedding level is given by the 'direction' property."
          },
          {
            "name": "isolate",
            "description": "The contents of the element are considered to be inside a separate, independent paragraph."
          },
          {
            "name": "isolate-override",
            "description": "This combines the isolation behavior of 'isolate' with the directional override behavior of 'bidi-override'"
          },
          {
            "name": "normal",
            "description": "The element does not open an additional level of embedding with respect to the bidirectional algorithm. For inline-level elements, implicit reordering works across element boundaries."
          },
          {
            "name": "plaintext",
            "description": "For the purposes of the Unicode bidirectional algorithm, the base directionality of each bidi paragraph for which the element forms the containing block is determined not by the element's computed 'direction'."
          }
        ],
        "syntax": "normal | embed | isolate | bidi-override | isolate-override | plaintext",
        "relevance": 57,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/unicode-bidi"
          }
        ],
        "description": "The level of embedding with respect to the bidirectional algorithm.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "unicode-range",
        "values": [
          {
            "name": "U+26",
            "description": "Ampersand."
          },
          {
            "name": "U+20-24F, U+2B0-2FF, U+370-4FF, U+1E00-1EFF, U+2000-20CF, U+2100-23FF, U+2500-26FF, U+E000-F8FF, U+FB00\u2013FB4F",
            "description": "WGL4 character set (Pan-European)."
          },
          {
            "name": "U+20-17F, U+2B0-2FF, U+2000-206F, U+20A0-20CF, U+2100-21FF, U+2600-26FF",
            "description": "The Multilingual European Subset No. 1. Latin. Covers ~44 languages."
          },
          {
            "name": "U+20-2FF, U+370-4FF, U+1E00-20CF, U+2100-23FF, U+2500-26FF, U+FB00-FB4F, U+FFF0-FFFD",
            "description": "The Multilingual European Subset No. 2. Latin, Greek, and Cyrillic. Covers ~128 language."
          },
          {
            "name": "U+20-4FF, U+530-58F, U+10D0-10FF, U+1E00-23FF, U+2440-245F, U+2500-26FF, U+FB00-FB4F, U+FE20-FE2F, U+FFF0-FFFD",
            "description": "The Multilingual European Subset No. 3. Covers all characters belonging to European scripts."
          },
          {
            "name": "U+00-7F",
            "description": "Basic Latin (ASCII)."
          },
          {
            "name": "U+80-FF",
            "description": "Latin-1 Supplement. Accented characters for Western European languages, common punctuation characters, multiplication and division signs."
          },
          {
            "name": "U+100-17F",
            "description": "Latin Extended-A. Accented characters for for Czech, Dutch, Polish, and Turkish."
          },
          {
            "name": "U+180-24F",
            "description": "Latin Extended-B. Croatian, Slovenian, Romanian, Non-European and historic latin, Khoisan, Pinyin, Livonian, Sinology."
          },
          {
            "name": "U+1E00-1EFF",
            "description": "Latin Extended Additional. Vietnamese, German captial sharp s, Medievalist, Latin general use."
          },
          {
            "name": "U+250-2AF",
            "description": "International Phonetic Alphabet Extensions."
          },
          {
            "name": "U+370-3FF",
            "description": "Greek and Coptic."
          },
          {
            "name": "U+1F00-1FFF",
            "description": "Greek Extended. Accented characters for polytonic Greek."
          },
          {
            "name": "U+400-4FF",
            "description": "Cyrillic."
          },
          {
            "name": "U+500-52F",
            "description": "Cyrillic Supplement. Extra letters for Komi, Khanty, Chukchi, Mordvin, Kurdish, Aleut, Chuvash, Abkhaz, Azerbaijani, and Orok."
          },
          {
            "name": "U+00-52F, U+1E00-1FFF, U+2200\u201322FF",
            "description": "Latin, Greek, Cyrillic, some punctuation and symbols."
          },
          {
            "name": "U+530\u201358F",
            "description": "Armenian."
          },
          {
            "name": "U+590\u20135FF",
            "description": "Hebrew."
          },
          {
            "name": "U+600\u20136FF",
            "description": "Arabic."
          },
          {
            "name": "U+750\u201377F",
            "description": "Arabic Supplement. Additional letters for African languages, Khowar, Torwali, Burushaski, and early Persian."
          },
          {
            "name": "U+8A0\u20138FF",
            "description": "Arabic Extended-A. Additional letters for African languages, European and Central Asian languages, Rohingya, Tamazight, Arwi, and Koranic annotation signs."
          },
          {
            "name": "U+700\u201374F",
            "description": "Syriac."
          },
          {
            "name": "U+900\u201397F",
            "description": "Devanagari."
          },
          {
            "name": "U+980\u20139FF",
            "description": "Bengali."
          },
          {
            "name": "U+A00\u2013A7F",
            "description": "Gurmukhi."
          },
          {
            "name": "U+A80\u2013AFF",
            "description": "Gujarati."
          },
          {
            "name": "U+B00\u2013B7F",
            "description": "Oriya."
          },
          {
            "name": "U+B80\u2013BFF",
            "description": "Tamil."
          },
          {
            "name": "U+C00\u2013C7F",
            "description": "Telugu."
          },
          {
            "name": "U+C80\u2013CFF",
            "description": "Kannada."
          },
          {
            "name": "U+D00\u2013D7F",
            "description": "Malayalam."
          },
          {
            "name": "U+D80\u2013DFF",
            "description": "Sinhala."
          },
          {
            "name": "U+118A0\u2013118FF",
            "description": "Warang Citi."
          },
          {
            "name": "U+E00\u2013E7F",
            "description": "Thai."
          },
          {
            "name": "U+1A20\u20131AAF",
            "description": "Tai Tham."
          },
          {
            "name": "U+AA80\u2013AADF",
            "description": "Tai Viet."
          },
          {
            "name": "U+E80\u2013EFF",
            "description": "Lao."
          },
          {
            "name": "U+F00\u2013FFF",
            "description": "Tibetan."
          },
          {
            "name": "U+1000\u2013109F",
            "description": "Myanmar (Burmese)."
          },
          {
            "name": "U+10A0\u201310FF",
            "description": "Georgian."
          },
          {
            "name": "U+1200\u2013137F",
            "description": "Ethiopic."
          },
          {
            "name": "U+1380\u2013139F",
            "description": "Ethiopic Supplement. Extra Syllables for Sebatbeit, and Tonal marks"
          },
          {
            "name": "U+2D80\u20132DDF",
            "description": "Ethiopic Extended. Extra Syllables for Me'en, Blin, and Sebatbeit."
          },
          {
            "name": "U+AB00\u2013AB2F",
            "description": "Ethiopic Extended-A. Extra characters for Gamo-Gofa-Dawro, Basketo, and Gumuz."
          },
          {
            "name": "U+1780\u201317FF",
            "description": "Khmer."
          },
          {
            "name": "U+1800\u201318AF",
            "description": "Mongolian."
          },
          {
            "name": "U+1B80\u20131BBF",
            "description": "Sundanese."
          },
          {
            "name": "U+1CC0\u20131CCF",
            "description": "Sundanese Supplement. Punctuation."
          },
          {
            "name": "U+4E00\u20139FD5",
            "description": "CJK (Chinese, Japanese, Korean) Unified Ideographs. Most common ideographs for modern Chinese and Japanese."
          },
          {
            "name": "U+3400\u20134DB5",
            "description": "CJK Unified Ideographs Extension A. Rare ideographs."
          },
          {
            "name": "U+2F00\u20132FDF",
            "description": "Kangxi Radicals."
          },
          {
            "name": "U+2E80\u20132EFF",
            "description": "CJK Radicals Supplement. Alternative forms of Kangxi Radicals."
          },
          {
            "name": "U+1100\u201311FF",
            "description": "Hangul Jamo."
          },
          {
            "name": "U+AC00\u2013D7AF",
            "description": "Hangul Syllables."
          },
          {
            "name": "U+3040\u2013309F",
            "description": "Hiragana."
          },
          {
            "name": "U+30A0\u201330FF",
            "description": "Katakana."
          },
          {
            "name": "U+A5, U+4E00-9FFF, U+30??, U+FF00-FF9F",
            "description": "Japanese Kanji, Hiragana and Katakana characters plus Yen/Yuan symbol."
          },
          {
            "name": "U+A4D0\u2013A4FF",
            "description": "Lisu."
          },
          {
            "name": "U+A000\u2013A48F",
            "description": "Yi Syllables."
          },
          {
            "name": "U+A490\u2013A4CF",
            "description": "Yi Radicals."
          },
          {
            "name": "U+2000-206F",
            "description": "General Punctuation."
          },
          {
            "name": "U+3000\u2013303F",
            "description": "CJK Symbols and Punctuation."
          },
          {
            "name": "U+2070\u2013209F",
            "description": "Superscripts and Subscripts."
          },
          {
            "name": "U+20A0\u201320CF",
            "description": "Currency Symbols."
          },
          {
            "name": "U+2100\u2013214F",
            "description": "Letterlike Symbols."
          },
          {
            "name": "U+2150\u2013218F",
            "description": "Number Forms."
          },
          {
            "name": "U+2190\u201321FF",
            "description": "Arrows."
          },
          {
            "name": "U+2200\u201322FF",
            "description": "Mathematical Operators."
          },
          {
            "name": "U+2300\u201323FF",
            "description": "Miscellaneous Technical."
          },
          {
            "name": "U+E000-F8FF",
            "description": "Private Use Area."
          },
          {
            "name": "U+FB00\u2013FB4F",
            "description": "Alphabetic Presentation Forms. Ligatures for latin, Armenian, and Hebrew."
          },
          {
            "name": "U+FB50\u2013FDFF",
            "description": "Arabic Presentation Forms-A. Contextual forms / ligatures for Persian, Urdu, Sindhi, Central Asian languages, etc, Arabic pedagogical symbols, word ligatures."
          },
          {
            "name": "U+1F600\u20131F64F",
            "description": "Emoji: Emoticons."
          },
          {
            "name": "U+2600\u201326FF",
            "description": "Emoji: Miscellaneous Symbols."
          },
          {
            "name": "U+1F300\u20131F5FF",
            "description": "Emoji: Miscellaneous Symbols and Pictographs."
          },
          {
            "name": "U+1F900\u20131F9FF",
            "description": "Emoji: Supplemental Symbols and Pictographs."
          },
          {
            "name": "U+1F680\u20131F6FF",
            "description": "Emoji: Transport and Map Symbols."
          }
        ],
        "syntax": "<unicode-range>#",
        "relevance": 58,
        "description": "@font-face descriptor. Defines the set of Unicode codepoints that may be supported by the font face for which it is declared.",
        "restrictions": [
          "unicode-range"
        ]
      },
      {
        "name": "user-select",
        "values": [
          {
            "name": "all",
            "description": "The content of the element must be selected atomically"
          },
          {
            "name": "auto"
          },
          {
            "name": "contain",
            "description": "UAs must not allow a selection which is started in this element to be extended outside of this element."
          },
          {
            "name": "none",
            "description": "The UA must not allow selections to be started in this element."
          },
          {
            "name": "text",
            "description": "The element imposes no constraint on the selection."
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | text | none | contain | all",
        "relevance": 24,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/user-select"
          }
        ],
        "description": "Controls the appearance of selection.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "vertical-align",
        "values": [
          {
            "name": "auto",
            "description": "Align the dominant baseline of the parent box with the equivalent, or heuristically reconstructed, baseline of the element inline box."
          },
          {
            "name": "baseline",
            "description": "Align the 'alphabetic' baseline of the element with the 'alphabetic' baseline of the parent element."
          },
          {
            "name": "bottom",
            "description": "Align the after edge of the extended inline box with the after-edge of the line box."
          },
          {
            "name": "middle",
            "description": "Align the 'middle' baseline of the inline element with the middle baseline of the parent."
          },
          {
            "name": "sub",
            "description": "Lower the baseline of the box to the proper position for subscripts of the parent's box. (This value has no effect on the font size of the element's text.)"
          },
          {
            "name": "super",
            "description": "Raise the baseline of the box to the proper position for superscripts of the parent's box. (This value has no effect on the font size of the element's text.)"
          },
          {
            "name": "text-bottom",
            "description": "Align the bottom of the box with the after-edge of the parent element's font."
          },
          {
            "name": "text-top",
            "description": "Align the top of the box with the before-edge of the parent element's font."
          },
          {
            "name": "top",
            "description": "Align the before edge of the extended inline box with the before-edge of the line box."
          },
          {
            "name": "-webkit-baseline-middle"
          }
        ],
        "syntax": "baseline | sub | super | text-top | text-bottom | middle | top | bottom | <percentage> | <length>",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/vertical-align"
          }
        ],
        "description": "Affects the vertical positioning of the inline boxes generated by an inline-level element inside a line box.",
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "visibility",
        "values": [
          {
            "name": "collapse",
            "description": "Table-specific. If used on elements other than rows, row groups, columns, or column groups, 'collapse' has the same meaning as 'hidden'."
          },
          {
            "name": "hidden",
            "description": "The generated box is invisible (fully transparent, nothing is drawn), but still affects layout."
          },
          {
            "name": "visible",
            "description": "The generated box is visible."
          }
        ],
        "syntax": "visible | hidden | collapse",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/visibility"
          }
        ],
        "description": "Specifies whether the boxes generated by an element are rendered. Invisible boxes still affect layout (set the \u2018display\u2019 property to \u2018none\u2019 to suppress box generation altogether).",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-animation",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          },
          {
            "name": "none",
            "description": "No animation is performed"
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines six of the animation properties into a single property.",
        "restrictions": [
          "time",
          "enum",
          "timing-function",
          "identifier",
          "number"
        ]
      },
      {
        "name": "-webkit-animation-delay",
        "browsers": [
          "C",
          "S5"
        ],
        "relevance": 50,
        "description": "Defines when the animation will start.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-webkit-animation-direction",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "alternate",
            "description": "The animation cycle iterations that are odd counts are played in the normal direction, and the animation cycle iterations that are even counts are played in a reverse direction."
          },
          {
            "name": "alternate-reverse",
            "description": "The animation cycle iterations that are odd counts are played in the reverse direction, and the animation cycle iterations that are even counts are played in a normal direction."
          },
          {
            "name": "normal",
            "description": "Normal playback."
          },
          {
            "name": "reverse",
            "description": "All iterations of the animation are played in the reverse direction from the way they were specified."
          }
        ],
        "relevance": 50,
        "description": "Defines whether or not the animation should play in reverse on alternate cycles.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-animation-duration",
        "browsers": [
          "C",
          "S5"
        ],
        "relevance": 50,
        "description": "Defines the length of time that an animation takes to complete one cycle.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-webkit-animation-fill-mode",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "backwards",
            "description": "The beginning property value (as defined in the first @keyframes at-rule) is applied before the animation is displayed, during the period defined by 'animation-delay'."
          },
          {
            "name": "both",
            "description": "Both forwards and backwards fill modes are applied."
          },
          {
            "name": "forwards",
            "description": "The final property value (as defined in the last @keyframes at-rule) is maintained after the animation completes."
          },
          {
            "name": "none",
            "description": "There is no change to the property value between the time the animation is applied and the time the animation begins playing or after the animation completes."
          }
        ],
        "relevance": 50,
        "description": "Defines what values are applied by the animation outside the time it is executing.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-animation-iteration-count",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "infinite",
            "description": "Causes the animation to repeat forever."
          }
        ],
        "relevance": 50,
        "description": "Defines the number of times an animation cycle is played. The default value is one, meaning the animation will play from beginning to end once.",
        "restrictions": [
          "number",
          "enum"
        ]
      },
      {
        "name": "-webkit-animation-name",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "none",
            "description": "No animation is performed"
          }
        ],
        "relevance": 50,
        "description": "Defines a list of animations that apply. Each name is used to select the keyframe at-rule that provides the property values for the animation.",
        "restrictions": [
          "identifier",
          "enum"
        ]
      },
      {
        "name": "-webkit-animation-play-state",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "paused",
            "description": "A running animation will be paused."
          },
          {
            "name": "running",
            "description": "Resume playback of a paused animation."
          }
        ],
        "relevance": 50,
        "description": "Defines whether the animation is running or paused.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-animation-timing-function",
        "browsers": [
          "C",
          "S5"
        ],
        "relevance": 50,
        "description": "Describes how the animation will progress over one cycle of its duration. See the 'transition-timing-function'.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "-webkit-appearance",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "button"
          },
          {
            "name": "button-bevel"
          },
          {
            "name": "caps-lock-indicator"
          },
          {
            "name": "caret"
          },
          {
            "name": "checkbox"
          },
          {
            "name": "default-button"
          },
          {
            "name": "listbox"
          },
          {
            "name": "listitem"
          },
          {
            "name": "media-fullscreen-button"
          },
          {
            "name": "media-mute-button"
          },
          {
            "name": "media-play-button"
          },
          {
            "name": "media-seek-back-button"
          },
          {
            "name": "media-seek-forward-button"
          },
          {
            "name": "media-slider"
          },
          {
            "name": "media-sliderthumb"
          },
          {
            "name": "menulist"
          },
          {
            "name": "menulist-button"
          },
          {
            "name": "menulist-text"
          },
          {
            "name": "menulist-textfield"
          },
          {
            "name": "none"
          },
          {
            "name": "push-button"
          },
          {
            "name": "radio"
          },
          {
            "name": "scrollbarbutton-down"
          },
          {
            "name": "scrollbarbutton-left"
          },
          {
            "name": "scrollbarbutton-right"
          },
          {
            "name": "scrollbarbutton-up"
          },
          {
            "name": "scrollbargripper-horizontal"
          },
          {
            "name": "scrollbargripper-vertical"
          },
          {
            "name": "scrollbarthumb-horizontal"
          },
          {
            "name": "scrollbarthumb-vertical"
          },
          {
            "name": "scrollbartrack-horizontal"
          },
          {
            "name": "scrollbartrack-vertical"
          },
          {
            "name": "searchfield"
          },
          {
            "name": "searchfield-cancel-button"
          },
          {
            "name": "searchfield-decoration"
          },
          {
            "name": "searchfield-results-button"
          },
          {
            "name": "searchfield-results-decoration"
          },
          {
            "name": "slider-horizontal"
          },
          {
            "name": "sliderthumb-horizontal"
          },
          {
            "name": "sliderthumb-vertical"
          },
          {
            "name": "slider-vertical"
          },
          {
            "name": "square-button"
          },
          {
            "name": "textarea"
          },
          {
            "name": "textfield"
          }
        ],
        "status": "nonstandard",
        "syntax": "none | button | button-bevel | caret | checkbox | default-button | inner-spin-button | listbox | listitem | media-controls-background | media-controls-fullscreen-background | media-current-time-display | media-enter-fullscreen-button | media-exit-fullscreen-button | media-fullscreen-button | media-mute-button | media-overlay-play-button | media-play-button | media-seek-back-button | media-seek-forward-button | media-slider | media-sliderthumb | media-time-remaining-display | media-toggle-closed-captions-button | media-volume-slider | media-volume-slider-container | media-volume-sliderthumb | menulist | menulist-button | menulist-text | menulist-textfield | meter | progress-bar | progress-bar-value | push-button | radio | searchfield | searchfield-cancel-button | searchfield-decoration | searchfield-results-button | searchfield-results-decoration | slider-horizontal | slider-vertical | sliderthumb-horizontal | sliderthumb-vertical | square-button | textarea | textfield",
        "relevance": 0,
        "description": "Changes the appearance of buttons and other controls to resemble native controls.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-backdrop-filter",
        "browsers": [
          "S9"
        ],
        "values": [
          {
            "name": "none",
            "description": "No filter effects are applied."
          },
          {
            "name": "blur()",
            "description": "Applies a Gaussian blur to the input image."
          },
          {
            "name": "brightness()",
            "description": "Applies a linear multiplier to input image, making it appear more or less bright."
          },
          {
            "name": "contrast()",
            "description": "Adjusts the contrast of the input."
          },
          {
            "name": "drop-shadow()",
            "description": "Applies a drop shadow effect to the input image."
          },
          {
            "name": "grayscale()",
            "description": "Converts the input image to grayscale."
          },
          {
            "name": "hue-rotate()",
            "description": "Applies a hue rotation on the input image. "
          },
          {
            "name": "invert()",
            "description": "Inverts the samples in the input image."
          },
          {
            "name": "opacity()",
            "description": "Applies transparency to the samples in the input image."
          },
          {
            "name": "saturate()",
            "description": "Saturates the input image."
          },
          {
            "name": "sepia()",
            "description": "Converts the input image to sepia."
          },
          {
            "name": "url()",
            "description": "A filter reference to a <filter> element."
          }
        ],
        "relevance": 50,
        "description": "Applies a filter effect where the first filter in the list takes the element's background image as the input image.",
        "restrictions": [
          "enum",
          "url"
        ]
      },
      {
        "name": "-webkit-backface-visibility",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "hidden"
          },
          {
            "name": "visible"
          }
        ],
        "relevance": 50,
        "description": "Determines whether or not the 'back' side of a transformed element is visible when facing the viewer. With an identity transform, the front side of an element faces the viewer.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-background-clip",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Determines the background painting area.",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "-webkit-background-composite",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "border"
          },
          {
            "name": "padding"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-background-origin",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "For elements rendered as a single box, specifies the background positioning area. For elements rendered as multiple boxes (e.g., inline boxes on several lines, boxes on several pages) specifies which boxes 'box-decoration-break' operates on to determine the background positioning area(s).",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "-webkit-border-image",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "auto",
            "description": "If 'auto' is specified then the border image width is the intrinsic width or height (whichever is applicable) of the corresponding image slice. If the image does not have the required intrinsic dimension then the corresponding border-width is used instead."
          },
          {
            "name": "fill",
            "description": "Causes the middle part of the border-image to be preserved."
          },
          {
            "name": "none"
          },
          {
            "name": "repeat",
            "description": "The image is tiled (repeated) to fill the area."
          },
          {
            "name": "round",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the image is rescaled so that it does."
          },
          {
            "name": "space",
            "description": "The image is tiled (repeated) to fill the area. If it does not fill the area with a whole number of tiles, the extra space is distributed around the tiles."
          },
          {
            "name": "stretch",
            "description": "The image is stretched to fill the area."
          },
          {
            "name": "url()"
          }
        ],
        "relevance": 50,
        "description": "Shorthand property for setting 'border-image-source', 'border-image-slice', 'border-image-width', 'border-image-outset' and 'border-image-repeat'. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "percentage",
          "number",
          "url",
          "enum"
        ]
      },
      {
        "name": "-webkit-box-align",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "baseline",
            "description": "If this box orientation is inline-axis or horizontal, all children are placed with their baselines aligned, and extra space placed before or after as necessary. For block flows, the baseline of the first non-empty line box located within the element is used. For tables, the baseline of the first cell is used."
          },
          {
            "name": "center",
            "description": "Any extra space is divided evenly, with half placed above the child and the other half placed after the child."
          },
          {
            "name": "end",
            "description": "For normal direction boxes, the bottom edge of each child is placed along the bottom of the box. Extra space is placed above the element. For reverse direction boxes, the top edge of each child is placed along the top of the box. Extra space is placed below the element."
          },
          {
            "name": "start",
            "description": "For normal direction boxes, the top edge of each child is placed along the top of the box. Extra space is placed below the element. For reverse direction boxes, the bottom edge of each child is placed along the bottom of the box. Extra space is placed above the element."
          },
          {
            "name": "stretch",
            "description": "The height of each child is adjusted to that of the containing block."
          }
        ],
        "relevance": 50,
        "description": "Specifies the alignment of nested elements within an outer flexible box element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-box-direction",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "normal",
            "description": "A box with a computed value of horizontal for box-orient displays its children from left to right. A box with a computed value of vertical displays its children from top to bottom."
          },
          {
            "name": "reverse",
            "description": "A box with a computed value of horizontal for box-orient displays its children from right to left. A box with a computed value of vertical displays its children from bottom to top."
          }
        ],
        "relevance": 50,
        "description": "In webkit applications, -webkit-box-direction specifies whether a box lays out its contents normally (from the top or left edge), or in reverse (from the bottom or right edge).",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-box-flex",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Specifies an element's flexibility.",
        "restrictions": [
          "number"
        ]
      },
      {
        "name": "-webkit-box-flex-group",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Flexible elements can be assigned to flex groups using the 'box-flex-group' property.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-webkit-box-ordinal-group",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Indicates the ordinal group the element belongs to. Elements with a lower ordinal group are displayed before those with a higher ordinal group.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-webkit-box-orient",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "block-axis",
            "description": "Elements are oriented along the box's axis."
          },
          {
            "name": "horizontal",
            "description": "The box displays its children from left to right in a horizontal line."
          },
          {
            "name": "inline-axis",
            "description": "Elements are oriented vertically."
          },
          {
            "name": "vertical",
            "description": "The box displays its children from stacked from top to bottom vertically."
          }
        ],
        "relevance": 50,
        "description": "In webkit applications, -webkit-box-orient specifies whether a box lays out its contents horizontally or vertically.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-box-pack",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "center",
            "description": "The extra space is divided evenly, with half placed before the first child and the other half placed after the last child."
          },
          {
            "name": "end",
            "description": "For normal direction boxes, the right edge of the last child is placed at the right side, with all extra space placed before the first child. For reverse direction boxes, the left edge of the first child is placed at the left side, with all extra space placed after the last child."
          },
          {
            "name": "justify",
            "description": "The space is divided evenly in-between each child, with none of the extra space placed before the first child or after the last child. If there is only one child, treat the pack value as if it were start."
          },
          {
            "name": "start",
            "description": "For normal direction boxes, the left edge of the first child is placed at the left side, with all extra space placed after the last child. For reverse direction boxes, the right edge of the last child is placed at the right side, with all extra space placed before the first child."
          }
        ],
        "relevance": 50,
        "description": "Specifies alignment of child elements within the current element in the direction of orientation.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-box-reflect",
        "browsers": [
          "E79",
          "S4",
          "C4",
          "O15"
        ],
        "values": [
          {
            "name": "above",
            "description": "The reflection appears above the border box."
          },
          {
            "name": "below",
            "description": "The reflection appears below the border box."
          },
          {
            "name": "left",
            "description": "The reflection appears to the left of the border box."
          },
          {
            "name": "right",
            "description": "The reflection appears to the right of the border box."
          }
        ],
        "status": "nonstandard",
        "syntax": "[ above | below | right | left ]? <length>? <image>?",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-box-reflect"
          }
        ],
        "description": "Defines a reflection of a border box."
      },
      {
        "name": "-webkit-box-sizing",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "border-box",
            "description": "The specified width and height (and respective min/max properties) on this element determine the border box of the element."
          },
          {
            "name": "content-box",
            "description": "Behavior of width and height as specified by CSS2.1. The specified width and height (and respective min/max properties) apply to the width and height respectively of the content box of the element."
          }
        ],
        "relevance": 50,
        "description": "Box Model addition in CSS3.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-break-after",
        "browsers": [
          "S7"
        ],
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break before/after the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the generated box."
          },
          {
            "name": "avoid-region"
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "region"
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior before the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-break-before",
        "browsers": [
          "S7"
        ],
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break before/after the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the generated box."
          },
          {
            "name": "avoid-region"
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "region"
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior before the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-break-inside",
        "browsers": [
          "S7"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break inside the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break inside the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break inside the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break inside the generated box."
          },
          {
            "name": "avoid-region"
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior inside the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-column-break-after",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break before/after the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the generated box."
          },
          {
            "name": "avoid-region"
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "region"
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior before the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-column-break-before",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "always",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break before/after the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break before/after the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break before/after the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break before/after the generated box."
          },
          {
            "name": "avoid-region"
          },
          {
            "name": "column",
            "description": "Always force a column break before/after the generated box."
          },
          {
            "name": "left",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a left page."
          },
          {
            "name": "page",
            "description": "Always force a page break before/after the generated box."
          },
          {
            "name": "region"
          },
          {
            "name": "right",
            "description": "Force one or two page breaks before/after the generated box so that the next page is formatted as a right page."
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior before the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-column-break-inside",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Neither force nor forbid a page/column break inside the generated box."
          },
          {
            "name": "avoid",
            "description": "Avoid a page/column break inside the generated box."
          },
          {
            "name": "avoid-column",
            "description": "Avoid a column break inside the generated box."
          },
          {
            "name": "avoid-page",
            "description": "Avoid a page break inside the generated box."
          },
          {
            "name": "avoid-region"
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior inside the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-column-count",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Determines the number of columns by the 'column-width' property and the element width."
          }
        ],
        "relevance": 50,
        "description": "Describes the optimal number of columns into which the content of the element will be flowed.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "-webkit-column-gap",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "normal",
            "description": "User agent specific and typically equivalent to 1em."
          }
        ],
        "relevance": 50,
        "description": "Sets the gap between columns. If there is a column rule between columns, it will appear in the middle of the gap.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-webkit-column-rule",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "This property is a shorthand for setting 'column-rule-width', 'column-rule-style', and 'column-rule-color' at the same place in the style sheet. Omitted values are set to their initial values.",
        "restrictions": [
          "length",
          "line-width",
          "line-style",
          "color"
        ]
      },
      {
        "name": "-webkit-column-rule-color",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Sets the color of the column rule",
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-webkit-column-rule-style",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Sets the style of the rule between columns of an element.",
        "restrictions": [
          "line-style"
        ]
      },
      {
        "name": "-webkit-column-rule-width",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "description": "Sets the width of the rule between columns. Negative values are not allowed.",
        "restrictions": [
          "length",
          "line-width"
        ]
      },
      {
        "name": "-webkit-columns",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "relevance": 50,
        "description": "A shorthand property which sets both 'column-width' and 'column-count'.",
        "restrictions": [
          "length",
          "integer"
        ]
      },
      {
        "name": "-webkit-column-span",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "all",
            "description": "The element spans across all columns. Content in the normal flow that appears before the element is automatically balanced across all columns before the element appear."
          },
          {
            "name": "none",
            "description": "The element does not span multiple columns."
          }
        ],
        "relevance": 50,
        "description": "Describes the page/column break behavior after the generated box.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-column-width",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          }
        ],
        "relevance": 50,
        "description": "This property describes the width of columns in multicol elements.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-webkit-filter",
        "browsers": [
          "C18",
          "O15",
          "S6"
        ],
        "values": [
          {
            "name": "none",
            "description": "No filter effects are applied."
          },
          {
            "name": "blur()",
            "description": "Applies a Gaussian blur to the input image."
          },
          {
            "name": "brightness()",
            "description": "Applies a linear multiplier to input image, making it appear more or less bright."
          },
          {
            "name": "contrast()",
            "description": "Adjusts the contrast of the input."
          },
          {
            "name": "drop-shadow()",
            "description": "Applies a drop shadow effect to the input image."
          },
          {
            "name": "grayscale()",
            "description": "Converts the input image to grayscale."
          },
          {
            "name": "hue-rotate()",
            "description": "Applies a hue rotation on the input image. "
          },
          {
            "name": "invert()",
            "description": "Inverts the samples in the input image."
          },
          {
            "name": "opacity()",
            "description": "Applies transparency to the samples in the input image."
          },
          {
            "name": "saturate()",
            "description": "Saturates the input image."
          },
          {
            "name": "sepia()",
            "description": "Converts the input image to sepia."
          },
          {
            "name": "url()",
            "description": "A filter reference to a <filter> element."
          }
        ],
        "relevance": 50,
        "description": "Processes an element\u2019s rendering before it is displayed in the document, by applying one or more filter effects.",
        "restrictions": [
          "enum",
          "url"
        ]
      },
      {
        "name": "-webkit-flow-from",
        "browsers": [
          "S6.1"
        ],
        "values": [
          {
            "name": "none",
            "description": "The block container is not a CSS Region."
          }
        ],
        "relevance": 50,
        "description": "Makes a block container a region and associates it with a named flow.",
        "restrictions": [
          "identifier"
        ]
      },
      {
        "name": "-webkit-flow-into",
        "browsers": [
          "S6.1"
        ],
        "values": [
          {
            "name": "none",
            "description": "The element is not moved to a named flow and normal CSS processing takes place."
          }
        ],
        "relevance": 50,
        "description": "Places an element or its contents into a named flow.",
        "restrictions": [
          "identifier"
        ]
      },
      {
        "name": "-webkit-font-feature-settings",
        "browsers": [
          "C16"
        ],
        "values": [
          {
            "name": '"c2cs"'
          },
          {
            "name": '"dlig"'
          },
          {
            "name": '"kern"'
          },
          {
            "name": '"liga"'
          },
          {
            "name": '"lnum"'
          },
          {
            "name": '"onum"'
          },
          {
            "name": '"smcp"'
          },
          {
            "name": '"swsh"'
          },
          {
            "name": '"tnum"'
          },
          {
            "name": "normal",
            "description": "No change in glyph substitution or positioning occurs."
          },
          {
            "name": "off"
          },
          {
            "name": "on"
          }
        ],
        "relevance": 50,
        "description": "This property provides low-level control over OpenType font features. It is intended as a way of providing access to font features that are not widely used but are needed for a particular use case.",
        "restrictions": [
          "string",
          "integer"
        ]
      },
      {
        "name": "-webkit-hyphens",
        "browsers": [
          "S5.1"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Conditional hyphenation characters inside a word, if present, take priority over automatic resources when determining hyphenation points within the word."
          },
          {
            "name": "manual",
            "description": "Words are only broken at line breaks where there are characters inside the word that suggest line break opportunities"
          },
          {
            "name": "none",
            "description": "Words are not broken at line breaks, even if characters inside the word suggest line break points."
          }
        ],
        "relevance": 50,
        "description": "Controls whether hyphenation is allowed to create more break opportunities within a line of text.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-line-break",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "after-white-space"
          },
          {
            "name": "normal"
          }
        ],
        "relevance": 50,
        "description": "Specifies line-breaking rules for CJK (Chinese, Japanese, and Korean) text."
      },
      {
        "name": "-webkit-margin-bottom-collapse",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "collapse"
          },
          {
            "name": "discard"
          },
          {
            "name": "separate"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-margin-collapse",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "collapse"
          },
          {
            "name": "discard"
          },
          {
            "name": "separate"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-margin-start",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "-webkit-margin-top-collapse",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "collapse"
          },
          {
            "name": "discard"
          },
          {
            "name": "separate"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-mask-clip",
        "browsers": [
          "C",
          "O15",
          "S4"
        ],
        "status": "nonstandard",
        "syntax": "[ <box> | border | padding | content | text ]#",
        "relevance": 0,
        "description": "Determines the mask painting area, which determines the area that is affected by the mask.",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "-webkit-mask-image",
        "browsers": [
          "C",
          "O15",
          "S4"
        ],
        "values": [
          {
            "name": "none",
            "description": "Counts as a transparent black image layer."
          },
          {
            "name": "url()",
            "description": "Reference to a <mask element or to a CSS image."
          }
        ],
        "status": "nonstandard",
        "syntax": "<mask-reference>#",
        "relevance": 0,
        "description": "Sets the mask layer image of an element.",
        "restrictions": [
          "url",
          "image",
          "enum"
        ]
      },
      {
        "name": "-webkit-mask-origin",
        "browsers": [
          "C",
          "O15",
          "S4"
        ],
        "status": "nonstandard",
        "syntax": "[ <box> | border | padding | content ]#",
        "relevance": 0,
        "description": "Specifies the mask positioning area.",
        "restrictions": [
          "box"
        ]
      },
      {
        "name": "-webkit-mask-repeat",
        "browsers": [
          "C",
          "O15",
          "S4"
        ],
        "status": "nonstandard",
        "syntax": "<repeat-style>#",
        "relevance": 0,
        "description": "Specifies how mask layer images are tiled after they have been sized and positioned.",
        "restrictions": [
          "repeat"
        ]
      },
      {
        "name": "-webkit-mask-size",
        "browsers": [
          "C",
          "O15",
          "S4"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Resolved by using the image\u2019s intrinsic ratio and the size of the other dimension, or failing that, using the image\u2019s intrinsic size, or failing that, treating it as 100%."
          },
          {
            "name": "contain",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the largest size such that both its width and its height can fit inside the background positioning area."
          },
          {
            "name": "cover",
            "description": "Scale the image, while preserving its intrinsic aspect ratio (if any), to the smallest size such that both its width and its height can completely cover the background positioning area."
          }
        ],
        "status": "nonstandard",
        "syntax": "<bg-size>#",
        "relevance": 0,
        "description": "Specifies the size of the mask layer images.",
        "restrictions": [
          "length",
          "percentage",
          "enum"
        ]
      },
      {
        "name": "-webkit-nbsp-mode",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "normal"
          },
          {
            "name": "space"
          }
        ],
        "relevance": 50,
        "description": "Defines the behavior of nonbreaking spaces within text."
      },
      {
        "name": "-webkit-overflow-scrolling",
        "browsers": [
          "C",
          "S5"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "touch"
          }
        ],
        "status": "nonstandard",
        "syntax": "auto | touch",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-overflow-scrolling"
          }
        ],
        "description": "Specifies whether to use native-style scrolling in an overflow:scroll element."
      },
      {
        "name": "-webkit-padding-start",
        "browsers": [
          "C",
          "S3"
        ],
        "relevance": 50,
        "restrictions": [
          "percentage",
          "length"
        ]
      },
      {
        "name": "-webkit-perspective",
        "browsers": [
          "C",
          "S4"
        ],
        "values": [
          {
            "name": "none",
            "description": "No perspective transform is applied."
          }
        ],
        "relevance": 50,
        "description": "Applies the same transform as the perspective(<number>) transform function, except that it applies only to the positioned or transformed children of the element, not to the transform on the element itself.",
        "restrictions": [
          "length"
        ]
      },
      {
        "name": "-webkit-perspective-origin",
        "browsers": [
          "C",
          "S4"
        ],
        "relevance": 50,
        "description": "Establishes the origin for the perspective property. It effectively sets the X and Y position at which the viewer appears to be looking at the children of the element.",
        "restrictions": [
          "position",
          "percentage",
          "length"
        ]
      },
      {
        "name": "-webkit-region-fragment",
        "browsers": [
          "S7"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Content flows as it would in a regular content box."
          },
          {
            "name": "break",
            "description": "If the content fits within the CSS Region, then this property has no effect."
          }
        ],
        "relevance": 50,
        "description": "The 'region-fragment' property controls the behavior of the last region associated with a named flow.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-tap-highlight-color",
        "browsers": [
          "E12",
          "C16",
          "O\u226415"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-tap-highlight-color"
          }
        ],
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-webkit-text-fill-color",
        "browsers": [
          "E12",
          "FF49",
          "S3",
          "C1",
          "O15"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-text-fill-color"
          }
        ],
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-webkit-text-size-adjust",
        "browsers": [
          "E",
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Renderers must use the default size adjustment when displaying on a small device."
          },
          {
            "name": "none",
            "description": "Renderers must not do size adjustment when displaying on a small device."
          }
        ],
        "relevance": 50,
        "description": "Specifies a size adjustment for displaying text content in mobile browsers.",
        "restrictions": [
          "percentage"
        ]
      },
      {
        "name": "-webkit-text-stroke",
        "browsers": [
          "E15",
          "FF49",
          "S3",
          "C4",
          "O15"
        ],
        "status": "nonstandard",
        "syntax": "<length> || <color>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-text-stroke"
          }
        ],
        "restrictions": [
          "length",
          "line-width",
          "color",
          "percentage"
        ]
      },
      {
        "name": "-webkit-text-stroke-color",
        "browsers": [
          "E15",
          "FF49",
          "S3",
          "C1",
          "O15"
        ],
        "status": "nonstandard",
        "syntax": "<color>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-text-stroke-color"
          }
        ],
        "restrictions": [
          "color"
        ]
      },
      {
        "name": "-webkit-text-stroke-width",
        "browsers": [
          "E15",
          "FF49",
          "S3",
          "C1",
          "O15"
        ],
        "status": "nonstandard",
        "syntax": "<length>",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-text-stroke-width"
          }
        ],
        "restrictions": [
          "length",
          "line-width",
          "percentage"
        ]
      },
      {
        "name": "-webkit-touch-callout",
        "browsers": [
          "S3"
        ],
        "values": [
          {
            "name": "none"
          }
        ],
        "status": "nonstandard",
        "syntax": "default | none",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-touch-callout"
          }
        ],
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-transform",
        "browsers": [
          "C",
          "O12",
          "S3.1"
        ],
        "values": [
          {
            "name": "matrix()",
            "description": "Specifies a 2D transformation in the form of a transformation matrix of six values. matrix(a,b,c,d,e,f) is equivalent to applying the transformation matrix [a b c d e f]"
          },
          {
            "name": "matrix3d()",
            "description": "Specifies a 3D transformation as a 4x4 homogeneous matrix of 16 values in column-major order."
          },
          {
            "name": "none"
          },
          {
            "name": "perspective()",
            "description": "Specifies a perspective projection matrix."
          },
          {
            "name": "rotate()",
            "description": "Specifies a 2D rotation by the angle specified in the parameter about the origin of the element, as defined by the transform-origin property."
          },
          {
            "name": "rotate3d()",
            "description": "Specifies a clockwise 3D rotation by the angle specified in last parameter about the [x,y,z] direction vector described by the first 3 parameters."
          },
          {
            "name": "rotateX('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the X axis."
          },
          {
            "name": "rotateY('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Y axis."
          },
          {
            "name": "rotateZ('angle')",
            "description": "Specifies a clockwise rotation by the given angle about the Z axis."
          },
          {
            "name": "scale()",
            "description": "Specifies a 2D scale operation by the [sx,sy] scaling vector described by the 2 parameters. If the second parameter is not provided, it is takes a value equal to the first."
          },
          {
            "name": "scale3d()",
            "description": "Specifies a 3D scale operation by the [sx,sy,sz] scaling vector described by the 3 parameters."
          },
          {
            "name": "scaleX()",
            "description": "Specifies a scale operation using the [sx,1] scaling vector, where sx is given as the parameter."
          },
          {
            "name": "scaleY()",
            "description": "Specifies a scale operation using the [sy,1] scaling vector, where sy is given as the parameter."
          },
          {
            "name": "scaleZ()",
            "description": "Specifies a scale operation using the [1,1,sz] scaling vector, where sz is given as the parameter."
          },
          {
            "name": "skew()",
            "description": "Specifies a skew transformation along the X and Y axes. The first angle parameter specifies the skew on the X axis. The second angle parameter specifies the skew on the Y axis. If the second parameter is not given then a value of 0 is used for the Y angle (ie: no skew on the Y axis)."
          },
          {
            "name": "skewX()",
            "description": "Specifies a skew transformation along the X axis by the given angle."
          },
          {
            "name": "skewY()",
            "description": "Specifies a skew transformation along the Y axis by the given angle."
          },
          {
            "name": "translate()",
            "description": "Specifies a 2D translation by the vector [tx, ty], where tx is the first translation-value parameter and ty is the optional second translation-value parameter."
          },
          {
            "name": "translate3d()",
            "description": "Specifies a 3D translation by the vector [tx,ty,tz], with tx, ty and tz being the first, second and third translation-value parameters respectively."
          },
          {
            "name": "translateX()",
            "description": "Specifies a translation by the given amount in the X direction."
          },
          {
            "name": "translateY()",
            "description": "Specifies a translation by the given amount in the Y direction."
          },
          {
            "name": "translateZ()",
            "description": "Specifies a translation by the given amount in the Z direction. Note that percentage values are not allowed in the translateZ translation-value, and if present are evaluated as 0."
          }
        ],
        "relevance": 50,
        "description": "A two-dimensional transformation is applied to an element through the 'transform' property. This property contains a list of transform functions similar to those allowed by SVG.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-transform-origin",
        "browsers": [
          "C",
          "O15",
          "S3.1"
        ],
        "relevance": 50,
        "description": "Establishes the origin of transformation for an element.",
        "restrictions": [
          "position",
          "length",
          "percentage"
        ]
      },
      {
        "name": "-webkit-transform-origin-x",
        "browsers": [
          "C",
          "S3.1"
        ],
        "relevance": 50,
        "description": "The x coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-webkit-transform-origin-y",
        "browsers": [
          "C",
          "S3.1"
        ],
        "relevance": 50,
        "description": "The y coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-webkit-transform-origin-z",
        "browsers": [
          "C",
          "S4"
        ],
        "relevance": 50,
        "description": "The z coordinate of the origin for transforms applied to an element with respect to its border box.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "-webkit-transform-style",
        "browsers": [
          "C",
          "S4"
        ],
        "values": [
          {
            "name": "flat",
            "description": "All children of this element are rendered flattened into the 2D plane of the element."
          }
        ],
        "relevance": 50,
        "description": "Defines how nested elements are rendered in 3D space.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-transition",
        "browsers": [
          "C",
          "O12",
          "S5"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Shorthand property combines four of the transition properties into a single property.",
        "restrictions": [
          "time",
          "property",
          "timing-function",
          "enum"
        ]
      },
      {
        "name": "-webkit-transition-delay",
        "browsers": [
          "C",
          "O12",
          "S5"
        ],
        "relevance": 50,
        "description": "Defines when the transition will start. It allows a transition to begin execution some period of time from when it is applied.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-webkit-transition-duration",
        "browsers": [
          "C",
          "O12",
          "S5"
        ],
        "relevance": 50,
        "description": "Specifies how long the transition from the old value to the new value should take.",
        "restrictions": [
          "time"
        ]
      },
      {
        "name": "-webkit-transition-property",
        "browsers": [
          "C",
          "O12",
          "S5"
        ],
        "values": [
          {
            "name": "all",
            "description": "Every property that is able to undergo a transition will do so."
          },
          {
            "name": "none",
            "description": "No property will transition."
          }
        ],
        "relevance": 50,
        "description": "Specifies the name of the CSS property to which the transition is applied.",
        "restrictions": [
          "property"
        ]
      },
      {
        "name": "-webkit-transition-timing-function",
        "browsers": [
          "C",
          "O12",
          "S5"
        ],
        "relevance": 50,
        "description": "Describes how the intermediate values used during a transition will be calculated.",
        "restrictions": [
          "timing-function"
        ]
      },
      {
        "name": "-webkit-user-drag",
        "browsers": [
          "S3"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "element"
          },
          {
            "name": "none"
          }
        ],
        "relevance": 50,
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-user-modify",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "read-only"
          },
          {
            "name": "read-write"
          },
          {
            "name": "read-write-plaintext-only"
          }
        ],
        "status": "nonstandard",
        "syntax": "read-only | read-write | read-write-plaintext-only",
        "relevance": 0,
        "description": "Determines whether a user can edit the content of an element.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "-webkit-user-select",
        "browsers": [
          "C",
          "S3"
        ],
        "values": [
          {
            "name": "auto"
          },
          {
            "name": "none"
          },
          {
            "name": "text"
          }
        ],
        "relevance": 50,
        "description": "Controls the appearance of selection.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "white-space",
        "values": [
          {
            "name": "normal",
            "description": "Sets 'white-space-collapsing' to 'collapse' and 'text-wrap' to 'normal'."
          },
          {
            "name": "nowrap",
            "description": "Sets 'white-space-collapsing' to 'collapse' and 'text-wrap' to 'none'."
          },
          {
            "name": "pre",
            "description": "Sets 'white-space-collapsing' to 'preserve' and 'text-wrap' to 'none'."
          },
          {
            "name": "pre-line",
            "description": "Sets 'white-space-collapsing' to 'preserve-breaks' and 'text-wrap' to 'normal'."
          },
          {
            "name": "pre-wrap",
            "description": "Sets 'white-space-collapsing' to 'preserve' and 'text-wrap' to 'normal'."
          }
        ],
        "syntax": "normal | pre | nowrap | pre-wrap | pre-line | break-spaces",
        "relevance": 88,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/white-space"
          }
        ],
        "description": "Shorthand property for the 'white-space-collapsing' and 'text-wrap' properties.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "widows",
        "browsers": [
          "E12",
          "S1.3",
          "C25",
          "IE8",
          "O9.2"
        ],
        "syntax": "<integer>",
        "relevance": 51,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/widows"
          }
        ],
        "description": "Specifies the minimum number of line boxes of a block container that must be left in a fragment after a break.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "width",
        "values": [
          {
            "name": "auto",
            "description": "The width depends on the values of other properties."
          },
          {
            "name": "fit-content",
            "description": "Use the fit-content inline size or fit-content block size, as appropriate to the writing mode."
          },
          {
            "name": "max-content",
            "description": "Use the max-content inline size or max-content block size, as appropriate to the writing mode."
          },
          {
            "name": "min-content",
            "description": "Use the min-content inline size or min-content block size, as appropriate to the writing mode."
          }
        ],
        "syntax": "<viewport-length>{1,2}",
        "relevance": 96,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/width"
          }
        ],
        "description": "Specifies the width of the content area, padding area or border area (depending on 'box-sizing') of certain boxes.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "will-change",
        "browsers": [
          "E79",
          "FF36",
          "S9.1",
          "C36",
          "O24"
        ],
        "values": [
          {
            "name": "auto",
            "description": "Expresses no particular intent."
          },
          {
            "name": "contents",
            "description": "Indicates that the author expects to animate or change something about the element\u2019s contents in the near future."
          },
          {
            "name": "scroll-position",
            "description": "Indicates that the author expects to animate or change the scroll position of the element in the near future."
          }
        ],
        "syntax": "auto | <animateable-feature>#",
        "relevance": 62,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/will-change"
          }
        ],
        "description": "Provides a rendering hint to the user agent, stating what kinds of changes the author expects to perform on the element.",
        "restrictions": [
          "enum",
          "identifier"
        ]
      },
      {
        "name": "word-break",
        "values": [
          {
            "name": "break-all",
            "description": "Lines may break between any two grapheme clusters for non-CJK scripts."
          },
          {
            "name": "keep-all",
            "description": "Block characters can no longer create implied break points."
          },
          {
            "name": "normal",
            "description": "Breaks non-CJK scripts according to their own rules."
          }
        ],
        "syntax": "normal | break-all | keep-all | break-word",
        "relevance": 72,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/word-break"
          }
        ],
        "description": "Specifies line break opportunities for non-CJK scripts.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "word-spacing",
        "values": [
          {
            "name": "normal",
            "description": "No additional spacing is applied. Computes to zero."
          }
        ],
        "syntax": "normal | <length-percentage>",
        "relevance": 57,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/word-spacing"
          }
        ],
        "description": "Specifies additional spacing between \u201Cwords\u201D.",
        "restrictions": [
          "length",
          "percentage"
        ]
      },
      {
        "name": "word-wrap",
        "values": [
          {
            "name": "break-word",
            "description": "An otherwise unbreakable sequence of characters may be broken at an arbitrary point if there are no otherwise-acceptable break points in the line."
          },
          {
            "name": "normal",
            "description": "Lines may break only at allowed break points."
          }
        ],
        "syntax": "normal | break-word",
        "relevance": 77,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-wrap"
          }
        ],
        "description": "Specifies whether the UA may break within a word to prevent overflow when an otherwise-unbreakable string is too long to fit.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "writing-mode",
        "values": [
          {
            "name": "horizontal-tb",
            "description": "Top-to-bottom block flow direction. The writing mode is horizontal."
          },
          {
            "name": "sideways-lr",
            "description": "Left-to-right block flow direction. The writing mode is vertical, while the typographic mode is horizontal."
          },
          {
            "name": "sideways-rl",
            "description": "Right-to-left block flow direction. The writing mode is vertical, while the typographic mode is horizontal."
          },
          {
            "name": "vertical-lr",
            "description": "Left-to-right block flow direction. The writing mode is vertical."
          },
          {
            "name": "vertical-rl",
            "description": "Right-to-left block flow direction. The writing mode is vertical."
          }
        ],
        "syntax": "horizontal-tb | vertical-rl | vertical-lr | sideways-rl | sideways-lr",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/writing-mode"
          }
        ],
        "description": "This is a shorthand property for both 'direction' and 'block-progression'.",
        "restrictions": [
          "enum"
        ]
      },
      {
        "name": "z-index",
        "values": [
          {
            "name": "auto",
            "description": "The stack level of the generated box in the current stacking context is 0. The box does not establish a new stacking context unless it is the root element."
          }
        ],
        "syntax": "auto | <integer>",
        "relevance": 91,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/z-index"
          }
        ],
        "description": "For a positioned box, the 'z-index' property specifies the stack level of the box in the current stacking context and whether the box establishes a local stacking context.",
        "restrictions": [
          "integer"
        ]
      },
      {
        "name": "zoom",
        "browsers": [
          "E12",
          "S3.1",
          "C1",
          "IE5.5",
          "O15"
        ],
        "values": [
          {
            "name": "normal"
          }
        ],
        "syntax": "auto | <number> | <percentage>",
        "relevance": 74,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/zoom"
          }
        ],
        "description": "Non-standard. Specifies the magnification scale of the object. See 'transform: scale()' for a standards-based alternative.",
        "restrictions": [
          "enum",
          "integer",
          "number",
          "percentage"
        ]
      },
      {
        "name": "-ms-ime-align",
        "status": "nonstandard",
        "syntax": "auto | after",
        "relevance": 0,
        "description": "Aligns the Input Method Editor (IME) candidate window box relative to the element on which the IME composition is active."
      },
      {
        "name": "-moz-binding",
        "status": "nonstandard",
        "syntax": "<url> | none",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-binding"
          }
        ],
        "description": "The -moz-binding CSS property is used by Mozilla-based applications to attach an XBL binding to a DOM element."
      },
      {
        "name": "-moz-context-properties",
        "status": "nonstandard",
        "syntax": "none | [ fill | fill-opacity | stroke | stroke-opacity ]#",
        "relevance": 0,
        "browsers": [
          "FF55"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-context-properties"
          }
        ],
        "description": "If you reference an SVG image in a webpage (such as with the <img> element or as a background image), the SVG image can coordinate with the embedding element (its context) to have the image adopt property values set on the embedding element. To do this the embedding element needs to list the properties that are to be made available to the image by listing them as values of the -moz-context-properties property, and the image needs to opt in to using those properties by using values such as the context-fill value.\n\nThis feature is available since Firefox 55, but is only currently supported with SVG images loaded via chrome:// or resource:// URLs. To experiment with the feature in SVG on the Web it is necessary to set the svg.context-properties.content.enabled pref to true."
      },
      {
        "name": "-moz-float-edge",
        "status": "nonstandard",
        "syntax": "border-box | content-box | margin-box | padding-box",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-float-edge"
          }
        ],
        "description": "The non-standard -moz-float-edge CSS property specifies whether the height and width properties of the element include the margin, border, or padding thickness."
      },
      {
        "name": "-moz-force-broken-image-icon",
        "status": "nonstandard",
        "syntax": "<integer>",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-force-broken-image-icon"
          }
        ],
        "description": "The -moz-force-broken-image-icon extended CSS property can be used to force the broken image icon to be shown even when a broken image has an alt attribute."
      },
      {
        "name": "-moz-image-region",
        "status": "nonstandard",
        "syntax": "<shape> | auto",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-image-region"
          }
        ],
        "description": "For certain XUL elements and pseudo-elements that use an image from the list-style-image property, this property specifies a region of the image that is used in place of the whole image. This allows elements to use different pieces of the same image to improve performance."
      },
      {
        "name": "-moz-orient",
        "status": "nonstandard",
        "syntax": "inline | block | horizontal | vertical",
        "relevance": 0,
        "browsers": [
          "FF6"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-orient"
          }
        ],
        "description": "The -moz-orient CSS property specifies the orientation of the element to which it's applied."
      },
      {
        "name": "-moz-outline-radius",
        "status": "nonstandard",
        "syntax": "<outline-radius>{1,4} [ / <outline-radius>{1,4} ]?",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-outline-radius"
          }
        ],
        "description": "In Mozilla applications like Firefox, the -moz-outline-radius CSS property can be used to give an element's outline rounded corners."
      },
      {
        "name": "-moz-outline-radius-bottomleft",
        "status": "nonstandard",
        "syntax": "<outline-radius>",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-outline-radius-bottomleft"
          }
        ],
        "description": "In Mozilla applications, the -moz-outline-radius-bottomleft CSS property can be used to round the bottom-left corner of an element's outline."
      },
      {
        "name": "-moz-outline-radius-bottomright",
        "status": "nonstandard",
        "syntax": "<outline-radius>",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-outline-radius-bottomright"
          }
        ],
        "description": "In Mozilla applications, the -moz-outline-radius-bottomright CSS property can be used to round the bottom-right corner of an element's outline."
      },
      {
        "name": "-moz-outline-radius-topleft",
        "status": "nonstandard",
        "syntax": "<outline-radius>",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-outline-radius-topleft"
          }
        ],
        "description": "In Mozilla applications, the -moz-outline-radius-topleft CSS property can be used to round the top-left corner of an element's outline."
      },
      {
        "name": "-moz-outline-radius-topright",
        "status": "nonstandard",
        "syntax": "<outline-radius>",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-outline-radius-topright"
          }
        ],
        "description": "In Mozilla applications, the -moz-outline-radius-topright CSS property can be used to round the top-right corner of an element's outline."
      },
      {
        "name": "-moz-stack-sizing",
        "status": "nonstandard",
        "syntax": "ignore | stretch-to-fit",
        "relevance": 0,
        "description": "-moz-stack-sizing is an extended CSS property. Normally, a stack will change its size so that all of its child elements are completely visible. For example, moving a child of the stack far to the right will widen the stack so the child remains visible."
      },
      {
        "name": "-moz-text-blink",
        "status": "nonstandard",
        "syntax": "none | blink",
        "relevance": 0,
        "description": "The -moz-text-blink non-standard Mozilla CSS extension specifies the blink mode."
      },
      {
        "name": "-moz-user-input",
        "status": "nonstandard",
        "syntax": "auto | none | enabled | disabled",
        "relevance": 0,
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-moz-user-input"
          }
        ],
        "description": "In Mozilla applications, -moz-user-input determines if an element will accept user input."
      },
      {
        "name": "-moz-user-modify",
        "status": "nonstandard",
        "syntax": "read-only | read-write | write-only",
        "relevance": 0,
        "description": "The -moz-user-modify property has no effect. It was originally planned to determine whether or not the content of an element can be edited by a user."
      },
      {
        "name": "-moz-window-dragging",
        "status": "nonstandard",
        "syntax": "drag | no-drag",
        "relevance": 0,
        "description": "The -moz-window-dragging CSS property specifies whether a window is draggable or not. It only works in Chrome code, and only on Mac OS X."
      },
      {
        "name": "-moz-window-shadow",
        "status": "nonstandard",
        "syntax": "default | menu | tooltip | sheet | none",
        "relevance": 0,
        "description": "The -moz-window-shadow CSS property specifies whether a window will have a shadow. It only works on Mac OS X."
      },
      {
        "name": "-webkit-border-before",
        "status": "nonstandard",
        "syntax": "<'border-width'> || <'border-style'> || <'color'>",
        "relevance": 0,
        "browsers": [
          "E79",
          "S5.1",
          "C8",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-border-before"
          }
        ],
        "description": "The -webkit-border-before CSS property is a shorthand property for setting the individual logical block start border property values in a single place in the style sheet."
      },
      {
        "name": "-webkit-border-before-color",
        "status": "nonstandard",
        "syntax": "<'color'>",
        "relevance": 0,
        "description": "The -webkit-border-before-color CSS property sets the color of the individual logical block start border in a single place in the style sheet."
      },
      {
        "name": "-webkit-border-before-style",
        "status": "nonstandard",
        "syntax": "<'border-style'>",
        "relevance": 0,
        "description": "The -webkit-border-before-style CSS property sets the style of the individual logical block start border in a single place in the style sheet."
      },
      {
        "name": "-webkit-border-before-width",
        "status": "nonstandard",
        "syntax": "<'border-width'>",
        "relevance": 0,
        "description": "The -webkit-border-before-width CSS property sets the width of the individual logical block start border in a single place in the style sheet."
      },
      {
        "name": "-webkit-line-clamp",
        "syntax": "none | <integer>",
        "relevance": 50,
        "browsers": [
          "E17",
          "FF68",
          "S5",
          "C6",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-line-clamp"
          }
        ],
        "description": "The -webkit-line-clamp CSS property allows limiting of the contents of a block container to the specified number of lines."
      },
      {
        "name": "-webkit-mask",
        "status": "nonstandard",
        "syntax": "[ <mask-reference> || <position> [ / <bg-size> ]? || <repeat-style> || [ <box> | border | padding | content | text ] || [ <box> | border | padding | content ] ]#",
        "relevance": 0,
        "description": "The mask CSS property alters the visibility of an element by either partially or fully hiding it. This is accomplished by either masking or clipping the image at specific points."
      },
      {
        "name": "-webkit-mask-attachment",
        "status": "nonstandard",
        "syntax": "<attachment>#",
        "relevance": 0,
        "browsers": [
          "S4",
          "C1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-attachment"
          }
        ],
        "description": "If a -webkit-mask-image is specified, -webkit-mask-attachment determines whether the mask image's position is fixed within the viewport, or scrolls along with its containing block."
      },
      {
        "name": "-webkit-mask-composite",
        "status": "nonstandard",
        "syntax": "<composite-style>#",
        "relevance": 0,
        "browsers": [
          "E18",
          "FF53",
          "S3.2",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-composite"
          }
        ],
        "description": "The -webkit-mask-composite property specifies the manner in which multiple mask images applied to the same element are composited with one another. Mask images are composited in the opposite order that they are declared with the -webkit-mask-image property."
      },
      {
        "name": "-webkit-mask-position",
        "status": "nonstandard",
        "syntax": "<position>#",
        "relevance": 0,
        "description": "The mask-position CSS property sets the initial position, relative to the mask position layer defined by mask-origin, for each defined mask image."
      },
      {
        "name": "-webkit-mask-position-x",
        "status": "nonstandard",
        "syntax": "[ <length-percentage> | left | center | right ]#",
        "relevance": 0,
        "browsers": [
          "E18",
          "FF49",
          "S3.2",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-position-x"
          }
        ],
        "description": "The -webkit-mask-position-x CSS property sets the initial horizontal position of a mask image."
      },
      {
        "name": "-webkit-mask-position-y",
        "status": "nonstandard",
        "syntax": "[ <length-percentage> | top | center | bottom ]#",
        "relevance": 0,
        "browsers": [
          "E18",
          "FF49",
          "S3.2",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-position-y"
          }
        ],
        "description": "The -webkit-mask-position-y CSS property sets the initial vertical position of a mask image."
      },
      {
        "name": "-webkit-mask-repeat-x",
        "status": "nonstandard",
        "syntax": "repeat | no-repeat | space | round",
        "relevance": 0,
        "browsers": [
          "E18",
          "S5",
          "C3",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-repeat-x"
          }
        ],
        "description": "The -webkit-mask-repeat-x property specifies whether and how a mask image is repeated (tiled) horizontally."
      },
      {
        "name": "-webkit-mask-repeat-y",
        "status": "nonstandard",
        "syntax": "repeat | no-repeat | space | round",
        "relevance": 0,
        "browsers": [
          "E18",
          "S5",
          "C3",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/-webkit-mask-repeat-y"
          }
        ],
        "description": "The -webkit-mask-repeat-y property specifies whether and how a mask image is repeated (tiled) vertically."
      },
      {
        "name": "appearance",
        "status": "experimental",
        "syntax": "none | auto | button | textfield | menulist-button | <compat-auto>",
        "relevance": 60,
        "browsers": [
          "E84",
          "FF1",
          "S3",
          "C84",
          "O70"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/appearance"
          }
        ],
        "description": "Changes the appearance of buttons and other controls to resemble native controls."
      },
      {
        "name": "aspect-ratio",
        "status": "experimental",
        "syntax": "auto | <ratio>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF71",
          "C79"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/aspect-ratio"
          }
        ],
        "description": "The aspect-ratio   CSS property sets a preferred aspect ratio for the box, which will be used in the calculation of auto sizes and some other layout functions."
      },
      {
        "name": "azimuth",
        "status": "obsolete",
        "syntax": "<angle> | [ [ left-side | far-left | left | center-left | center | center-right | right | far-right | right-side ] || behind ] | leftwards | rightwards",
        "relevance": 0,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/azimuth"
          }
        ],
        "description": "In combination with elevation, the azimuth CSS property enables different audio sources to be positioned spatially for aural presentation. This is important in that it provides a natural way to tell several voices apart, as each can be positioned to originate at a different location on the sound stage. Stereo output produce a lateral sound stage, while binaural headphones and multi-speaker setups allow for a fully three-dimensional stage."
      },
      {
        "name": "backdrop-filter",
        "syntax": "none | <filter-function-list>",
        "relevance": 51,
        "browsers": [
          "E17",
          "FF70",
          "S9",
          "C76",
          "O34"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/backdrop-filter"
          }
        ],
        "description": "The backdrop-filter CSS property lets you apply graphical effects such as blurring or color shifting to the area behind an element. Because it applies to everything behind the element, to see the effect you must make the element or its background at least partially transparent."
      },
      {
        "name": "border-block",
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block"
          }
        ],
        "description": "The border-block CSS property is a shorthand property for setting the individual logical block border property values in a single place in the style sheet."
      },
      {
        "name": "border-block-color",
        "syntax": "<'border-top-color'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-color"
          }
        ],
        "description": "The border-block-color CSS property defines the color of the logical block borders of an element, which maps to a physical border color depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-color and border-bottom-color, or border-right-color and border-left-color property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-block-style",
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-style"
          }
        ],
        "description": "The border-block-style CSS property defines the style of the logical block borders of an element, which maps to a physical border style depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-style and border-bottom-style, or border-left-style and border-right-style properties depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-block-width",
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-block-width"
          }
        ],
        "description": "The border-block-width CSS property defines the width of the logical block borders of an element, which maps to a physical border width depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-width and border-bottom-width, or border-left-width, and border-right-width property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-end-end-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 50,
        "browsers": [
          "FF66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-end-end-radius"
          }
        ],
        "description": "The border-end-end-radius CSS property defines a logical border radius on an element, which maps to a physical border radius that depends on on the element's writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-end-start-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 50,
        "browsers": [
          "FF66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-end-start-radius"
          }
        ],
        "description": "The border-end-start-radius CSS property defines a logical border radius on an element, which maps to a physical border radius depending on the element's writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-inline",
        "syntax": "<'border-top-width'> || <'border-top-style'> || <'color'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline"
          }
        ],
        "description": "The border-inline CSS property is a shorthand property for setting the individual logical inline border property values in a single place in the style sheet."
      },
      {
        "name": "border-inline-color",
        "syntax": "<'border-top-color'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-color"
          }
        ],
        "description": "The border-inline-color CSS property defines the color of the logical inline borders of an element, which maps to a physical border color depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-color and border-bottom-color, or border-right-color and border-left-color property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-inline-style",
        "syntax": "<'border-top-style'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-style"
          }
        ],
        "description": "The border-inline-style CSS property defines the style of the logical inline borders of an element, which maps to a physical border style depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-style and border-bottom-style, or border-left-style and border-right-style properties depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-inline-width",
        "syntax": "<'border-top-width'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-inline-width"
          }
        ],
        "description": "The border-inline-width CSS property defines the width of the logical inline borders of an element, which maps to a physical border width depending on the element's writing mode, directionality, and text orientation. It corresponds to the border-top-width and border-bottom-width, or border-left-width, and border-right-width property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-start-end-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 50,
        "browsers": [
          "FF66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-start-end-radius"
          }
        ],
        "description": "The border-start-end-radius CSS property defines a logical border radius on an element, which maps to a physical border radius depending on the element's writing-mode, direction, and text-orientation."
      },
      {
        "name": "border-start-start-radius",
        "syntax": "<length-percentage>{1,2}",
        "relevance": 50,
        "browsers": [
          "FF66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/border-start-start-radius"
          }
        ],
        "description": "The border-start-start-radius CSS property defines a logical border radius on an element, which maps to a physical border radius that depends on the element's writing-mode, direction, and text-orientation."
      },
      {
        "name": "box-align",
        "status": "nonstandard",
        "syntax": "start | center | end | baseline | stretch",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-align"
          }
        ],
        "description": "The box-align CSS property specifies how an element aligns its contents across its layout in a perpendicular direction. The effect of the property is only visible if there is extra space in the box."
      },
      {
        "name": "box-direction",
        "status": "nonstandard",
        "syntax": "normal | reverse | inherit",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-direction"
          }
        ],
        "description": "The box-direction CSS property specifies whether a box lays out its contents normally (from the top or left edge), or in reverse (from the bottom or right edge)."
      },
      {
        "name": "box-flex",
        "status": "nonstandard",
        "syntax": "<number>",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-flex"
          }
        ],
        "description": "The -moz-box-flex and -webkit-box-flex CSS properties specify how a -moz-box or -webkit-box grows to fill the box that contains it, in the direction of the containing box's layout."
      },
      {
        "name": "box-flex-group",
        "status": "nonstandard",
        "syntax": "<integer>",
        "relevance": 0,
        "browsers": [
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-flex-group"
          }
        ],
        "description": "The box-flex-group CSS property assigns the flexbox's child elements to a flex group."
      },
      {
        "name": "box-lines",
        "status": "nonstandard",
        "syntax": "single | multiple",
        "relevance": 0,
        "browsers": [
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-lines"
          }
        ],
        "description": "The box-lines CSS property determines whether the box may have a single or multiple lines (rows for horizontally oriented boxes, columns for vertically oriented boxes)."
      },
      {
        "name": "box-ordinal-group",
        "status": "nonstandard",
        "syntax": "<integer>",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-ordinal-group"
          }
        ],
        "description": "The box-ordinal-group CSS property assigns the flexbox's child elements to an ordinal group."
      },
      {
        "name": "box-orient",
        "status": "nonstandard",
        "syntax": "horizontal | vertical | inline-axis | block-axis | inherit",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-orient"
          }
        ],
        "description": "The box-orient CSS property specifies whether an element lays out its contents horizontally or vertically."
      },
      {
        "name": "box-pack",
        "status": "nonstandard",
        "syntax": "start | center | end | justify",
        "relevance": 0,
        "browsers": [
          "E12",
          "FF1",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/box-pack"
          }
        ],
        "description": "The -moz-box-pack and -webkit-box-pack CSS properties specify how a -moz-box or -webkit-box packs its contents in the direction of its layout. The effect of this is only visible if there is extra space in the box."
      },
      {
        "name": "color-adjust",
        "syntax": "economy | exact",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF48",
          "S6",
          "C49",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/color-adjust"
          }
        ],
        "description": "The color-adjust property is a non-standard CSS extension that can be used to force printing of background colors and images in browsers based on the WebKit engine."
      },
      {
        "name": "counter-set",
        "syntax": "[ <custom-ident> <integer>? ]+ | none",
        "relevance": 50,
        "browsers": [
          "FF68"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/counter-set"
          }
        ],
        "description": "The counter-set CSS property sets a CSS counter to a given value. It manipulates the value of existing counters, and will only create new counters if there isn't already a counter of the given name on the element."
      },
      {
        "name": "font-optical-sizing",
        "syntax": "auto | none",
        "relevance": 50,
        "browsers": [
          "E17",
          "FF62",
          "S11",
          "C79",
          "O66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-optical-sizing"
          }
        ],
        "description": "The font-optical-sizing CSS property allows developers to control whether browsers render text with slightly differing visual representations to optimize viewing at different sizes, or not. This only works for fonts that have an optical size variation axis."
      },
      {
        "name": "font-variation-settings",
        "syntax": "normal | [ <string> <number> ]#",
        "relevance": 50,
        "browsers": [
          "E17",
          "FF62",
          "S11",
          "C62",
          "O49"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-variation-settings"
          }
        ],
        "description": "The font-variation-settings CSS property provides low-level control over OpenType or TrueType font variations, by specifying the four letter axis names of the features you want to vary, along with their variation values."
      },
      {
        "name": "font-smooth",
        "status": "nonstandard",
        "syntax": "auto | never | always | <absolute-size> | <length>",
        "relevance": 0,
        "browsers": [
          "E79",
          "FF25",
          "S4",
          "C5",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/font-smooth"
          }
        ],
        "description": ""
      },
      {
        "name": "gap",
        "syntax": "<'row-gap'> <'column-gap'>?",
        "relevance": 50,
        "browsers": [
          "E84",
          "FF63",
          "S10.1",
          "C84",
          "O70"
        ],
        "description": "The gap CSS property is a shorthand property for row-gap and column-gap specifying the gutters between grid rows and columns."
      },
      {
        "name": "hanging-punctuation",
        "syntax": "none | [ first || [ force-end | allow-end ] || last ]",
        "relevance": 50,
        "browsers": [
          "S10"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/hanging-punctuation"
          }
        ],
        "description": "The hanging-punctuation CSS property specifies whether a punctuation mark should hang at the start or end of a line of text. Hanging punctuation may be placed outside the line box."
      },
      {
        "name": "image-resolution",
        "status": "experimental",
        "syntax": "[ from-image || <resolution> ] && snap?",
        "relevance": 50,
        "description": "The image-resolution property specifies the intrinsic resolution of all raster images used in or on the element. It affects both content images (e.g. replaced elements and generated content) and decorative images (such as background-image). The intrinsic resolution of an image is used to determine the image\u2019s intrinsic dimensions."
      },
      {
        "name": "initial-letter",
        "status": "experimental",
        "syntax": "normal | [ <number> <integer>? ]",
        "relevance": 50,
        "browsers": [
          "S9"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/initial-letter"
          }
        ],
        "description": "The initial-letter CSS property specifies styling for dropped, raised, and sunken initial letters."
      },
      {
        "name": "initial-letter-align",
        "status": "experimental",
        "syntax": "[ auto | alphabetic | hanging | ideographic ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/initial-letter-align"
          }
        ],
        "description": "The initial-letter-align CSS property specifies the alignment of initial letters within a paragraph."
      },
      {
        "name": "inset",
        "syntax": "<'top'>{1,4}",
        "relevance": 50,
        "browsers": [
          "FF66"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset"
          }
        ],
        "description": "The inset CSS property defines the logical block and inline start and end offsets of an element, which map to physical offsets depending on the element's writing mode, directionality, and text orientation. It corresponds to the top and bottom, or right and left properties depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-block",
        "syntax": "<'top'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-block"
          }
        ],
        "description": "The inset-block CSS property defines the logical block start and end offsets of an element, which maps to physical offsets depending on the element's writing mode, directionality, and text orientation. It corresponds to the top and bottom, or right and left properties depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-block-end",
        "syntax": "<'top'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-block-end"
          }
        ],
        "description": "The inset-block-end CSS property defines the logical block end offset of an element, which maps to a physical offset depending on the element's writing mode, directionality, and text orientation. It corresponds to the top, right, bottom, or left property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-block-start",
        "syntax": "<'top'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-block-start"
          }
        ],
        "description": "The inset-block-start CSS property defines the logical block start offset of an element, which maps to a physical offset depending on the element's writing mode, directionality, and text orientation. It corresponds to the top, right, bottom, or left property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-inline",
        "syntax": "<'top'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-inline"
          }
        ],
        "description": "The inset-inline CSS property defines the logical block start and end offsets of an element, which maps to physical offsets depending on the element's writing mode, directionality, and text orientation. It corresponds to the top and bottom, or right and left properties depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-inline-end",
        "syntax": "<'top'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-inline-end"
          }
        ],
        "description": "The inset-inline-end CSS property defines the logical inline end inset of an element, which maps to a physical inset depending on the element's writing mode, directionality, and text orientation. It corresponds to the top, right, bottom, or left property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "inset-inline-start",
        "syntax": "<'top'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF63",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/inset-inline-start"
          }
        ],
        "description": "The inset-inline-start CSS property defines the logical inline start inset of an element, which maps to a physical offset depending on the element's writing mode, directionality, and text orientation. It corresponds to the top, right, bottom, or left property depending on the values defined for writing-mode, direction, and text-orientation."
      },
      {
        "name": "line-clamp",
        "status": "experimental",
        "syntax": "none | <integer>",
        "relevance": 50,
        "description": "The line-clamp property allows limiting the contents of a block container to the specified number of lines; remaining content is fragmented away and neither rendered nor measured. Optionally, it also allows inserting content into the last line box to indicate the continuity of truncated/interrupted content."
      },
      {
        "name": "line-height-step",
        "status": "experimental",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "C60",
          "O47"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/line-height-step"
          }
        ],
        "description": "The line-height-step CSS property defines the step units for line box heights. When the step unit is positive, line box heights are rounded up to the closest multiple of the unit. Negative values are invalid."
      },
      {
        "name": "margin-block",
        "syntax": "<'margin-left'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-block"
          }
        ],
        "description": "The margin-block CSS property defines the logical block start and end margins of an element, which maps to physical margins depending on the element's writing mode, directionality, and text orientation."
      },
      {
        "name": "margin-inline",
        "syntax": "<'margin-left'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-inline"
          }
        ],
        "description": "The margin-inline CSS property defines the logical inline start and end margins of an element, which maps to physical margins depending on the element's writing mode, directionality, and text orientation."
      },
      {
        "name": "margin-trim",
        "status": "experimental",
        "syntax": "none | in-flow | all",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/margin-trim"
          }
        ],
        "description": "The margin-trim property allows the container to trim the margins of its children where they adjoin the container\u2019s edges."
      },
      {
        "name": "mask",
        "syntax": "<mask-layer>#",
        "relevance": 50,
        "browsers": [
          "E12",
          "FF2",
          "S3.2",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask"
          }
        ],
        "description": "The mask CSS property alters the visibility of an element by either partially or fully hiding it. This is accomplished by either masking or clipping the image at specific points."
      },
      {
        "name": "mask-border",
        "syntax": "<'mask-border-source'> || <'mask-border-slice'> [ / <'mask-border-width'>? [ / <'mask-border-outset'> ]? ]? || <'mask-border-repeat'> || <'mask-border-mode'>",
        "relevance": 50,
        "description": "The mask-border CSS property lets you create a mask along the edge of an element's border.\n\nThis property is a shorthand for mask-border-source, mask-border-slice, mask-border-width, mask-border-outset, mask-border-repeat, and mask-border-mode. As with all shorthand properties, any omitted sub-values will be set to their initial value."
      },
      {
        "name": "mask-border-mode",
        "syntax": "luminance | alpha",
        "relevance": 50,
        "description": "The mask-border-mode CSS property specifies the blending mode used in a mask border."
      },
      {
        "name": "mask-border-outset",
        "syntax": "[ <length> | <number> ]{1,4}",
        "relevance": 50,
        "description": "The mask-border-outset CSS property specifies the distance by which an element's mask border is set out from its border box."
      },
      {
        "name": "mask-border-repeat",
        "syntax": "[ stretch | repeat | round | space ]{1,2}",
        "relevance": 50,
        "description": "The mask-border-repeat CSS property defines how the edge regions of a source image are adjusted to fit the dimensions of an element's mask border."
      },
      {
        "name": "mask-border-slice",
        "syntax": "<number-percentage>{1,4} fill?",
        "relevance": 50,
        "description": "The mask-border-slice CSS property divides the image specified by mask-border-source into regions. These regions are used to form the components of an element's mask border."
      },
      {
        "name": "mask-border-source",
        "syntax": "none | <image>",
        "relevance": 50,
        "description": "The mask-border-source CSS property specifies the source image used to create an element's mask border.\n\nThe mask-border-slice property is used to divide the source image into regions, which are then dynamically applied to the final mask border."
      },
      {
        "name": "mask-border-width",
        "syntax": "[ <length-percentage> | <number> | auto ]{1,4}",
        "relevance": 50,
        "description": "The mask-border-width CSS property specifies the width of an element's mask border."
      },
      {
        "name": "mask-clip",
        "syntax": "[ <geometry-box> | no-clip ]#",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF53",
          "S4",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-clip"
          }
        ],
        "description": "The mask-clip CSS property determines the area, which is affected by a mask. The painted content of an element must be restricted to this area."
      },
      {
        "name": "mask-composite",
        "syntax": "<compositing-operator>#",
        "relevance": 50,
        "browsers": [
          "E18",
          "FF53"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/mask-composite"
          }
        ],
        "description": "The mask-composite CSS property represents a compositing operation used on the current mask layer with the mask layers below it."
      },
      {
        "name": "max-lines",
        "status": "experimental",
        "syntax": "none | <integer>",
        "relevance": 50,
        "description": "The max-liens property forces a break after a set number of lines"
      },
      {
        "name": "offset",
        "syntax": "[ <'offset-position'>? [ <'offset-path'> [ <'offset-distance'> || <'offset-rotate'> ]? ]? ]! [ / <'offset-anchor'> ]?",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF72",
          "C55",
          "O42"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset"
          }
        ],
        "description": "The offset CSS property is a shorthand property for animating an element along a defined path."
      },
      {
        "name": "offset-anchor",
        "syntax": "auto | <position>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF72",
          "C79"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset-anchor"
          }
        ],
        "description": "Defines an anchor point of the box positioned along the path. The anchor point specifies the point of the box which is to be considered as the point that is moved along the path."
      },
      {
        "name": "offset-distance",
        "syntax": "<length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF72",
          "C55",
          "O42"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset-distance"
          }
        ],
        "description": "The offset-distance CSS property specifies a position along an offset-path."
      },
      {
        "name": "offset-path",
        "syntax": "none | ray( [ <angle> && <size>? && contain? ] ) | <path()> | <url> | [ <basic-shape> || <geometry-box> ]",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF72",
          "C55",
          "O45"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset-path"
          }
        ],
        "description": 'The offset-path CSS property specifies the offset path where the element gets positioned. The exact element\u2019s position on the offset path is determined by the offset-distance property. An offset path is either a specified path with one or multiple sub-paths or the geometry of a not-styled basic shape. Each shape or path must define an initial position for the computed value of "0" for offset-distance and an initial direction which specifies the rotation of the object to the initial position.\n\nIn this specification, a direction (or rotation) of 0 degrees is equivalent to the direction of the positive x-axis in the object\u2019s local coordinate system. In other words, a rotation of 0 degree points to the right side of the UA if the object and its ancestors have no transformation applied.'
      },
      {
        "name": "offset-position",
        "status": "experimental",
        "syntax": "auto | <position>",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset-position"
          }
        ],
        "description": "Specifies the initial position of the offset path. If position is specified with static, offset-position would be ignored."
      },
      {
        "name": "offset-rotate",
        "syntax": "[ auto | reverse ] || <angle>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF72",
          "C56",
          "O43"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/offset-rotate"
          }
        ],
        "description": "The offset-rotate CSS property defines the direction of the element while positioning along the offset path."
      },
      {
        "name": "overflow-anchor",
        "syntax": "auto | none",
        "relevance": 51,
        "browsers": [
          "E79",
          "FF66",
          "C56",
          "O43"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-anchor"
          }
        ],
        "description": "The overflow-anchor CSS property provides a way to opt out browser scroll anchoring behavior which adjusts scroll position to minimize content shifts."
      },
      {
        "name": "overflow-block",
        "syntax": "visible | hidden | clip | scroll | auto",
        "relevance": 50,
        "browsers": [
          "FF69"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-block"
          }
        ],
        "description": "The overflow-block CSS media feature can be used to test how the output device handles content that overflows the initial containing block along the block axis."
      },
      {
        "name": "overflow-clip-box",
        "status": "nonstandard",
        "syntax": "padding-box | content-box",
        "relevance": 0,
        "browsers": [
          "FF29"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Mozilla/Gecko/Chrome/CSS/overflow-clip-box"
          }
        ],
        "description": "The overflow-clip-box CSS property specifies relative to which box the clipping happens when there is an overflow. It is short hand for the overflow-clip-box-inline and overflow-clip-box-block properties."
      },
      {
        "name": "overflow-inline",
        "syntax": "visible | hidden | clip | scroll | auto",
        "relevance": 50,
        "browsers": [
          "FF69"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overflow-inline"
          }
        ],
        "description": "The overflow-inline CSS media feature can be used to test how the output device handles content that overflows the initial containing block along the inline axis."
      },
      {
        "name": "overscroll-behavior",
        "syntax": "[ contain | none | auto ]{1,2}",
        "relevance": 50,
        "browsers": [
          "E18",
          "FF59",
          "C63",
          "O50"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overscroll-behavior"
          }
        ],
        "description": "The overscroll-behavior CSS property is shorthand for the overscroll-behavior-x and overscroll-behavior-y properties, which allow you to control the browser's scroll overflow behavior \u2014 what happens when the boundary of a scrolling area is reached."
      },
      {
        "name": "overscroll-behavior-block",
        "syntax": "contain | none | auto",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF73",
          "C77",
          "O64"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overscroll-behavior-block"
          }
        ],
        "description": "The overscroll-behavior-block CSS property sets the browser's behavior when the block direction boundary of a scrolling area is reached."
      },
      {
        "name": "overscroll-behavior-inline",
        "syntax": "contain | none | auto",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF73",
          "C77",
          "O64"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overscroll-behavior-inline"
          }
        ],
        "description": "The overscroll-behavior-inline CSS property sets the browser's behavior when the inline direction boundary of a scrolling area is reached."
      },
      {
        "name": "overscroll-behavior-x",
        "syntax": "contain | none | auto",
        "relevance": 50,
        "browsers": [
          "E18",
          "FF59",
          "C63",
          "O50"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overscroll-behavior-x"
          }
        ],
        "description": "The overscroll-behavior-x CSS property is allows you to control the browser's scroll overflow behavior \u2014 what happens when the boundary of a scrolling area is reached \u2014 in the x axis direction."
      },
      {
        "name": "overscroll-behavior-y",
        "syntax": "contain | none | auto",
        "relevance": 50,
        "browsers": [
          "E18",
          "FF59",
          "C63",
          "O50"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/overscroll-behavior-y"
          }
        ],
        "description": "The overscroll-behavior-y CSS property is allows you to control the browser's scroll overflow behavior \u2014 what happens when the boundary of a scrolling area is reached \u2014 in the y axis direction."
      },
      {
        "name": "padding-block",
        "syntax": "<'padding-left'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-block"
          }
        ],
        "description": "The padding-block CSS property defines the logical block start and end padding of an element, which maps to physical padding properties depending on the element's writing mode, directionality, and text orientation."
      },
      {
        "name": "padding-inline",
        "syntax": "<'padding-left'>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF66",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/padding-inline"
          }
        ],
        "description": "The padding-inline CSS property defines the logical inline start and end padding of an element, which maps to physical padding properties depending on the element's writing mode, directionality, and text orientation."
      },
      {
        "name": "place-content",
        "syntax": "<'align-content'> <'justify-content'>?",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF53",
          "S9",
          "C59",
          "O46"
        ],
        "description": "The place-content CSS shorthand property sets both the align-content and justify-content properties."
      },
      {
        "name": "place-items",
        "syntax": "<'align-items'> <'justify-items'>?",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF45",
          "S11",
          "C59",
          "O46"
        ],
        "description": "The CSS place-items shorthand property sets both the align-items and justify-items properties. The first value is the align-items property value, the second the justify-items one. If the second value is not present, the first value is also used for it."
      },
      {
        "name": "place-self",
        "syntax": "<'align-self'> <'justify-self'>?",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF45",
          "C59",
          "O46"
        ],
        "description": "The place-self CSS property is a shorthand property sets both the align-self and justify-self properties. The first value is the align-self property value, the second the justify-self one. If the second value is not present, the first value is also used for it."
      },
      {
        "name": "rotate",
        "syntax": "none | <angle> | [ x | y | z | <number>{3} ] && <angle>",
        "relevance": 50,
        "browsers": [
          "FF72"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/rotate"
          }
        ],
        "description": "The rotate CSS property allows you to specify rotation transforms individually and independently of the transform property. This maps better to typical user interface usage, and saves having to remember the exact order of transform functions to specify in the transform value."
      },
      {
        "name": "row-gap",
        "syntax": "normal | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E84",
          "FF63",
          "S10.1",
          "C84",
          "O70"
        ],
        "description": "The row-gap CSS property specifies the gutter between grid rows."
      },
      {
        "name": "ruby-merge",
        "status": "experimental",
        "syntax": "separate | collapse | auto",
        "relevance": 50,
        "description": "This property controls how ruby annotation boxes should be rendered when there are more than one in a ruby container box: whether each pair should be kept separate, the annotations should be collapsed and rendered as a group, or the separation should be determined based on the space available."
      },
      {
        "name": "scale",
        "syntax": "none | <number>{1,3}",
        "relevance": 50,
        "browsers": [
          "FF72"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scale"
          }
        ],
        "description": "The scale CSS property allows you to specify scale transforms individually and independently of the transform property. This maps better to typical user interface usage, and saves having to remember the exact order of transform functions to specify in the transform value."
      },
      {
        "name": "scrollbar-color",
        "syntax": "auto | dark | light | <color>{2}",
        "relevance": 50,
        "browsers": [
          "FF64"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-color"
          }
        ],
        "description": "The scrollbar-color CSS property sets the color of the scrollbar track and thumb."
      },
      {
        "name": "scrollbar-width",
        "syntax": "auto | thin | none",
        "relevance": 50,
        "browsers": [
          "FF64"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scrollbar-width"
          }
        ],
        "description": "The scrollbar-width property allows the author to set the maximum thickness of an element\u2019s scrollbars when they are shown. "
      },
      {
        "name": "scroll-margin",
        "syntax": "<length>{1,4}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin"
          }
        ],
        "description": "The scroll-margin property is a shorthand property which sets all of the scroll-margin longhands, assigning values much like the margin property does for the margin-* longhands."
      },
      {
        "name": "scroll-margin-block",
        "syntax": "<length>{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-block"
          }
        ],
        "description": "The scroll-margin-block property is a shorthand property which sets the scroll-margin longhands in the block dimension."
      },
      {
        "name": "scroll-margin-block-start",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-block-start"
          }
        ],
        "description": "The scroll-margin-block-start property defines the margin of the scroll snap area at the start of the block dimension that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-block-end",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-block-end"
          }
        ],
        "description": "The scroll-margin-block-end property defines the margin of the scroll snap area at the end of the block dimension that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-bottom",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-bottom"
          }
        ],
        "description": "The scroll-margin-bottom property defines the bottom margin of the scroll snap area that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-inline",
        "syntax": "<length>{1,2}",
        "relevance": 50,
        "browsers": [
          "FF68"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-inline"
          }
        ],
        "description": "The scroll-margin-inline property is a shorthand property which sets the scroll-margin longhands in the inline dimension."
      },
      {
        "name": "scroll-margin-inline-start",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-inline-start"
          }
        ],
        "description": "The scroll-margin-inline-start property defines the margin of the scroll snap area at the start of the inline dimension that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-inline-end",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-inline-end"
          }
        ],
        "description": "The scroll-margin-inline-end property defines the margin of the scroll snap area at the end of the inline dimension that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-left",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-left"
          }
        ],
        "description": "The scroll-margin-left property defines the left margin of the scroll snap area that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-right",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-right"
          }
        ],
        "description": "The scroll-margin-right property defines the right margin of the scroll snap area that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-margin-top",
        "syntax": "<length>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-margin-top"
          }
        ],
        "description": "The scroll-margin-top property defines the top margin of the scroll snap area that is used for snapping this box to the snapport. The scroll snap area is determined by taking the transformed border box, finding its rectangular bounding box (axis-aligned in the scroll container\u2019s coordinate space), then adding the specified outsets."
      },
      {
        "name": "scroll-padding",
        "syntax": "[ auto | <length-percentage> ]{1,4}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding"
          }
        ],
        "description": "The scroll-padding property is a shorthand property which sets all of the scroll-padding longhands, assigning values much like the padding property does for the padding-* longhands."
      },
      {
        "name": "scroll-padding-block",
        "syntax": "[ auto | <length-percentage> ]{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-block"
          }
        ],
        "description": "The scroll-padding-block property is a shorthand property which sets the scroll-padding longhands for the block dimension."
      },
      {
        "name": "scroll-padding-block-start",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-block-start"
          }
        ],
        "description": "The scroll-padding-block-start property defines offsets for the start edge in the block dimension of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-block-end",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-block-end"
          }
        ],
        "description": "The scroll-padding-block-end property defines offsets for the end edge in the block dimension of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-bottom",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-bottom"
          }
        ],
        "description": "The scroll-padding-bottom property defines offsets for the bottom of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-inline",
        "syntax": "[ auto | <length-percentage> ]{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-inline"
          }
        ],
        "description": "The scroll-padding-inline property is a shorthand property which sets the scroll-padding longhands for the inline dimension."
      },
      {
        "name": "scroll-padding-inline-start",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-inline-start"
          }
        ],
        "description": "The scroll-padding-inline-start property defines offsets for the start edge in the inline dimension of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-inline-end",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-inline-end"
          }
        ],
        "description": "The scroll-padding-inline-end property defines offsets for the end edge in the inline dimension of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-left",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-left"
          }
        ],
        "description": "The scroll-padding-left property defines offsets for the left of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-right",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-right"
          }
        ],
        "description": "The scroll-padding-right property defines offsets for the right of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-padding-top",
        "syntax": "auto | <length-percentage>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-padding-top"
          }
        ],
        "description": "The scroll-padding-top property defines offsets for the top of the optimal viewing region of the scrollport: the region used as the target region for placing things in view of the user. This allows the author to exclude regions of the scrollport that are obscured by other content (such as fixed-positioned toolbars or sidebars) or simply to put more breathing room between a targeted element and the edges of the scrollport."
      },
      {
        "name": "scroll-snap-align",
        "syntax": "[ none | start | end | center ]{1,2}",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF68",
          "S11",
          "C69",
          "O56"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-align"
          }
        ],
        "description": "The scroll-snap-align property specifies the box\u2019s snap position as an alignment of its snap area (as the alignment subject) within its snap container\u2019s snapport (as the alignment container). The two values specify the snapping alignment in the block axis and inline axis, respectively. If only one value is specified, the second value defaults to the same value."
      },
      {
        "name": "scroll-snap-stop",
        "syntax": "normal | always",
        "relevance": 50,
        "browsers": [
          "E79",
          "C75",
          "O62"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-stop"
          }
        ],
        "description": 'The scroll-snap-stop CSS property defines whether the scroll container is allowed to "pass over" possible snap positions.'
      },
      {
        "name": "scroll-snap-type-x",
        "status": "obsolete",
        "syntax": "none | mandatory | proximity",
        "relevance": 0,
        "browsers": [
          "FF39",
          "S9"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-type-x"
          }
        ],
        "description": "The scroll-snap-type-x CSS property defines how strictly snap points are enforced on the horizontal axis of the scroll container in case there is one.\n\nSpecifying any precise animations or physics used to enforce those snap points is not covered by this property but instead left up to the user agent."
      },
      {
        "name": "scroll-snap-type-y",
        "status": "obsolete",
        "syntax": "none | mandatory | proximity",
        "relevance": 0,
        "browsers": [
          "FF39"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/scroll-snap-type-y"
          }
        ],
        "description": "The scroll-snap-type-y CSS property defines how strictly snap points are enforced on the vertical axis of the scroll container in case there is one.\n\nSpecifying any precise animations or physics used to enforce those snap points is not covered by this property but instead left up to the user agent."
      },
      {
        "name": "text-combine-upright",
        "syntax": "none | all | [ digits <integer>? ]",
        "relevance": 50,
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-combine-upright"
          }
        ],
        "description": "The text-combine-upright CSS property specifies the combination of multiple characters into the space of a single character. If the combined text is wider than 1em, the user agent must fit the contents within 1em. The resulting composition is treated as a single upright glyph for layout and decoration. This property only has an effect in vertical writing modes.\n\nThis is used to produce an effect that is known as tate-ch\u016B-yoko (\u7E26\u4E2D\u6A2A) in Japanese, or as \u76F4\u66F8\u6A6B\u5411 in Chinese."
      },
      {
        "name": "text-decoration-skip",
        "status": "experimental",
        "syntax": "none | [ objects || [ spaces | [ leading-spaces || trailing-spaces ] ] || edges || box-decoration ]",
        "relevance": 52,
        "browsers": [
          "S12.1",
          "C57",
          "O44"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-skip"
          }
        ],
        "description": "The text-decoration-skip CSS property specifies what parts of the element\u2019s content any text decoration affecting the element must skip over. It controls all text decoration lines drawn by the element and also any text decoration lines drawn by its ancestors."
      },
      {
        "name": "text-decoration-skip-ink",
        "syntax": "auto | all | none",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF70",
          "C64",
          "O50"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-skip-ink"
          }
        ],
        "description": "The text-decoration-skip-ink CSS property specifies how overlines and underlines are drawn when they pass over glyph ascenders and descenders."
      },
      {
        "name": "text-decoration-thickness",
        "syntax": "auto | from-font | <length> | <percentage> ",
        "relevance": 50,
        "browsers": [
          "FF70",
          "S12.1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-decoration-thickness"
          }
        ],
        "description": "The text-decoration-thickness CSS property sets the thickness, or width, of the decoration line that is used on text in an element, such as a line-through, underline, or overline."
      },
      {
        "name": "text-emphasis",
        "syntax": "<'text-emphasis-style'> || <'text-emphasis-color'>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF46",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-emphasis"
          }
        ],
        "description": "The text-emphasis CSS property is a shorthand property for setting text-emphasis-style and text-emphasis-color in one declaration. This property will apply the specified emphasis mark to each character of the element's text, except separator characters, like spaces,  and control characters."
      },
      {
        "name": "text-emphasis-color",
        "syntax": "<color>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF46",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-emphasis-color"
          }
        ],
        "description": "The text-emphasis-color CSS property defines the color used to draw emphasis marks on text being rendered in the HTML document. This value can also be set and reset using the text-emphasis shorthand."
      },
      {
        "name": "text-emphasis-position",
        "syntax": "[ over | under ] && [ right | left ]",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF46",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-emphasis-position"
          }
        ],
        "description": "The text-emphasis-position CSS property describes where emphasis marks are drawn at. The effect of emphasis marks on the line height is the same as for ruby text: if there isn't enough place, the line height is increased."
      },
      {
        "name": "text-emphasis-style",
        "syntax": "none | [ [ filled | open ] || [ dot | circle | double-circle | triangle | sesame ] ] | <string>",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF46",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-emphasis-style"
          }
        ],
        "description": "The text-emphasis-style CSS property defines the type of emphasis used. It can also be set, and reset, using the text-emphasis shorthand."
      },
      {
        "name": "text-size-adjust",
        "status": "experimental",
        "syntax": "none | auto | <percentage>",
        "relevance": 56,
        "browsers": [
          "E79",
          "C54",
          "O41"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-size-adjust"
          }
        ],
        "description": "The text-size-adjust CSS property controls the text inflation algorithm used on some smartphones and tablets. Other browsers will ignore this property."
      },
      {
        "name": "text-underline-offset",
        "syntax": "auto | <length> | <percentage> ",
        "relevance": 50,
        "browsers": [
          "FF70",
          "S12.1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/text-underline-offset"
          }
        ],
        "description": "The text-underline-offset CSS property sets the offset distance of an underline text decoration line (applied using text-decoration) from its original position."
      },
      {
        "name": "transform-box",
        "syntax": "content-box | border-box | fill-box | stroke-box | view-box",
        "relevance": 50,
        "browsers": [
          "E79",
          "FF55",
          "S11",
          "C64",
          "O51"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/transform-box"
          }
        ],
        "description": "The transform-box CSS property defines the layout box to which the transform and transform-origin properties relate."
      },
      {
        "name": "translate",
        "syntax": "none | <length-percentage> [ <length-percentage> <length>? ]?",
        "relevance": 50,
        "browsers": [
          "FF72"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/translate"
          }
        ],
        "description": "The translate CSS property allows you to specify translation transforms individually and independently of the transform property. This maps better to typical user interface usage, and saves having to remember the exact order of transform functions to specify in the transform value."
      },
      {
        "name": "speak-as",
        "syntax": "auto | bullets | numbers | words | spell-out | <counter-style-name>",
        "relevance": 50,
        "description": "The speak-as descriptor specifies how a counter symbol constructed with a given @counter-style will be represented in the spoken form. For example, an author can specify a counter symbol to be either spoken as its numerical value or just represented with an audio cue."
      },
      {
        "name": "font-display",
        "status": "experimental",
        "syntax": "[ auto | block | swap | fallback | optional ]",
        "relevance": 54,
        "description": "The font-display descriptor determines how a font face is displayed based on whether and when it is downloaded and ready to use."
      },
      {
        "name": "bleed",
        "syntax": "auto | <length>",
        "relevance": 50,
        "description": "The bleed CSS at-rule descriptor, used with the @page at-rule, specifies the extent of the page bleed area outside the page box. This property only has effect if crop marks are enabled using the marks property."
      },
      {
        "name": "marks",
        "syntax": "none | [ crop || cross ]",
        "relevance": 50,
        "description": "The marks CSS at-rule descriptor, used with the @page at-rule, adds crop and/or cross marks to the presentation of the document. Crop marks indicate where the page should be cut. Cross marks are used to align sheets."
      },
      {
        "name": "max-zoom",
        "syntax": "auto | <number> | <percentage>",
        "relevance": 50,
        "description": "The max-zoom CSS descriptor sets the maximum zoom factor of a document defined by the @viewport at-rule. The browser will not zoom in any further than this, whether automatically or at the user's request.\n\nA zoom factor of 1.0 or 100% corresponds to no zooming. Larger values are zoomed in. Smaller values are zoomed out."
      },
      {
        "name": "min-zoom",
        "syntax": "auto | <number> | <percentage>",
        "relevance": 50,
        "description": "The min-zoom CSS descriptor sets the minimum zoom factor of a document defined by the @viewport at-rule. The browser will not zoom out any further than this, whether automatically or at the user's request.\n\nA zoom factor of 1.0 or 100% corresponds to no zooming. Larger values are zoomed in. Smaller values are zoomed out."
      },
      {
        "name": "orientation",
        "syntax": "auto | portrait | landscape",
        "relevance": 50,
        "description": "The orientation CSS @media media feature can be used to apply styles based on the orientation of the viewport (or the page box, for paged media)."
      },
      {
        "name": "user-zoom",
        "syntax": "zoom | fixed",
        "relevance": 50,
        "description": "The user-zoom CSS descriptor controls whether or not the user can change the zoom factor of a document defined by @viewport."
      },
      {
        "name": "viewport-fit",
        "syntax": "auto | contain | cover",
        "relevance": 50,
        "description": "The border-block-style CSS property defines the style of the logical block borders of an element, which maps to a physical border style depending on the element's writing mode, directionality, and text orientation."
      }
    ],
    "atDirectives": [
      {
        "name": "@charset",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@charset"
          }
        ],
        "description": "Defines character set of the document."
      },
      {
        "name": "@counter-style",
        "browsers": [
          "FF33"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@counter-style"
          }
        ],
        "description": "Defines a custom counter style."
      },
      {
        "name": "@font-face",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@font-face"
          }
        ],
        "description": "Allows for linking to fonts that are automatically activated when needed. This permits authors to work around the limitation of 'web-safe' fonts, allowing for consistent rendering independent of the fonts available in a given user's environment."
      },
      {
        "name": "@font-feature-values",
        "browsers": [
          "FF34",
          "S9.1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@font-feature-values"
          }
        ],
        "description": "Defines named values for the indices used to select alternate glyphs for a given font family."
      },
      {
        "name": "@import",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@import"
          }
        ],
        "description": "Includes content of another file."
      },
      {
        "name": "@keyframes",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@keyframes"
          }
        ],
        "description": "Defines set of animation key frames."
      },
      {
        "name": "@media",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@media"
          }
        ],
        "description": "Defines a stylesheet for a particular media type."
      },
      {
        "name": "@-moz-document",
        "browsers": [
          "FF1.8"
        ],
        "description": "Gecko-specific at-rule that restricts the style rules contained within it based on the URL of the document."
      },
      {
        "name": "@-moz-keyframes",
        "browsers": [
          "FF5"
        ],
        "description": "Defines set of animation key frames."
      },
      {
        "name": "@-ms-viewport",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Specifies the size, zoom factor, and orientation of the viewport."
      },
      {
        "name": "@namespace",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@namespace"
          }
        ],
        "description": "Declares a prefix and associates it with a namespace name."
      },
      {
        "name": "@-o-keyframes",
        "browsers": [
          "O12"
        ],
        "description": "Defines set of animation key frames."
      },
      {
        "name": "@-o-viewport",
        "browsers": [
          "O11"
        ],
        "description": "Specifies the size, zoom factor, and orientation of the viewport."
      },
      {
        "name": "@page",
        "browsers": [
          "E12",
          "FF19",
          "C2",
          "IE8",
          "O6"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@page"
          }
        ],
        "description": "Directive defines various page parameters."
      },
      {
        "name": "@supports",
        "browsers": [
          "E12",
          "FF22",
          "S9",
          "C28",
          "O12.1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/@supports"
          }
        ],
        "description": "A conditional group rule whose condition tests whether the user agent supports CSS property:value pairs."
      },
      {
        "name": "@-webkit-keyframes",
        "browsers": [
          "C",
          "S4"
        ],
        "description": "Defines set of animation key frames."
      }
    ],
    "pseudoClasses": [
      {
        "name": ":active",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:active"
          }
        ],
        "description": "Applies while an element is being activated by the user. For example, between the times the user presses the mouse button and releases it."
      },
      {
        "name": ":any-link",
        "browsers": [
          "E79",
          "FF50",
          "S9",
          "C65",
          "O52"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:any-link"
          }
        ],
        "description": "Represents an element that acts as the source anchor of a hyperlink. Applies to both visited and unvisited links."
      },
      {
        "name": ":checked",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:checked"
          }
        ],
        "description": "Radio and checkbox elements can be toggled by the user. Some menu items are 'checked' when the user selects them. When such elements are toggled 'on' the :checked pseudo-class applies."
      },
      {
        "name": ":corner-present",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Indicates whether or not a scrollbar corner is present."
      },
      {
        "name": ":decrement",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Indicates whether or not the button or track piece will decrement the view\u2019s position when used."
      },
      {
        "name": ":default",
        "browsers": [
          "E79",
          "FF4",
          "S5",
          "C10",
          "O10"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:default"
          }
        ],
        "description": "Applies to the one or more UI elements that are the default among a set of similar elements. Typically applies to context menu items, buttons, and select lists/menus."
      },
      {
        "name": ":disabled",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:disabled"
          }
        ],
        "description": "Represents user interface elements that are in a disabled state; such elements have a corresponding enabled state."
      },
      {
        "name": ":double-button",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Applies when both buttons are displayed together at the same end of the scrollbar."
      },
      {
        "name": ":empty",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:empty"
          }
        ],
        "description": "Represents an element that has no children at all."
      },
      {
        "name": ":enabled",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:enabled"
          }
        ],
        "description": "Represents user interface elements that are in an enabled state; such elements have a corresponding disabled state."
      },
      {
        "name": ":end",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Indicates whether the object is placed after the thumb."
      },
      {
        "name": ":first",
        "browsers": [
          "E12",
          "S6",
          "C18",
          "IE8",
          "O9.2"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:first"
          }
        ],
        "description": "When printing double-sided documents, the page boxes on left and right pages may be different. This can be expressed through CSS pseudo-classes defined in the  page context."
      },
      {
        "name": ":first-child",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:first-child"
          }
        ],
        "description": "Same as :nth-child(1). Represents an element that is the first child of some other element."
      },
      {
        "name": ":first-of-type",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:first-of-type"
          }
        ],
        "description": "Same as :nth-of-type(1). Represents an element that is the first sibling of its type in the list of children of its parent element."
      },
      {
        "name": ":focus",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:focus"
          }
        ],
        "description": "Applies while an element has the focus (accepts keyboard or mouse events, or other forms of input)."
      },
      {
        "name": ":fullscreen",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:fullscreen"
          }
        ],
        "description": "Matches any element that has its fullscreen flag set."
      },
      {
        "name": ":future",
        "browsers": [
          "C",
          "O16",
          "S6"
        ],
        "description": "Represents any element that is defined to occur entirely after a :current element."
      },
      {
        "name": ":horizontal",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to any scrollbar pieces that have a horizontal orientation."
      },
      {
        "name": ":host",
        "browsers": [
          "E79",
          "FF63",
          "S10",
          "C54",
          "O41"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:host"
          }
        ],
        "description": "When evaluated in the context of a shadow tree, matches the shadow tree\u2019s host element."
      },
      {
        "name": ":host()",
        "browsers": [
          "C35",
          "O22"
        ],
        "description": "When evaluated in the context of a shadow tree, it matches the shadow tree\u2019s host element if the host element, in its normal context, matches the selector argument."
      },
      {
        "name": ":host-context()",
        "browsers": [
          "C35",
          "O22"
        ],
        "description": "Tests whether there is an ancestor, outside the shadow tree, which matches a particular selector."
      },
      {
        "name": ":hover",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:hover"
          }
        ],
        "description": "Applies while the user designates an element with a pointing device, but does not necessarily activate it. For example, a visual user agent could apply this pseudo-class when the cursor (mouse pointer) hovers over a box generated by the element."
      },
      {
        "name": ":increment",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Indicates whether or not the button or track piece will increment the view\u2019s position when used."
      },
      {
        "name": ":indeterminate",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:indeterminate"
          }
        ],
        "description": "Applies to UI elements whose value is in an indeterminate state."
      },
      {
        "name": ":in-range",
        "browsers": [
          "E13",
          "FF29",
          "S5.1",
          "C10",
          "O11"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:in-range"
          }
        ],
        "description": "Used in conjunction with the min and max attributes, whether on a range input, a number field, or any other types that accept those attributes."
      },
      {
        "name": ":invalid",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:invalid"
          }
        ],
        "description": "An element is :valid or :invalid when it is, respectively, valid or invalid with respect to data validity semantics defined by a different specification."
      },
      {
        "name": ":lang()",
        "browsers": [
          "E",
          "C",
          "FF1",
          "IE8",
          "O8",
          "S3"
        ],
        "description": "Represents an element that is in language specified."
      },
      {
        "name": ":last-child",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:last-child"
          }
        ],
        "description": "Same as :nth-last-child(1). Represents an element that is the last child of some other element."
      },
      {
        "name": ":last-of-type",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:last-of-type"
          }
        ],
        "description": "Same as :nth-last-of-type(1). Represents an element that is the last sibling of its type in the list of children of its parent element."
      },
      {
        "name": ":left",
        "browsers": [
          "E12",
          "S5.1",
          "C6",
          "IE8",
          "O9.2"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:left"
          }
        ],
        "description": "When printing double-sided documents, the page boxes on left and right pages may be different. This can be expressed through CSS pseudo-classes defined in the  page context."
      },
      {
        "name": ":link",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:link"
          }
        ],
        "description": "Applies to links that have not yet been visited."
      },
      {
        "name": ":matches()",
        "browsers": [
          "S9"
        ],
        "description": "Takes a selector list as its argument. It represents an element that is represented by its argument."
      },
      {
        "name": ":-moz-any()",
        "browsers": [
          "FF4"
        ],
        "description": "Represents an element that is represented by the selector list passed as its argument. Standardized as :matches()."
      },
      {
        "name": ":-moz-any-link",
        "browsers": [
          "FF1"
        ],
        "description": "Represents an element that acts as the source anchor of a hyperlink. Applies to both visited and unvisited links."
      },
      {
        "name": ":-moz-broken",
        "browsers": [
          "FF3"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-broken"
          }
        ],
        "description": "Non-standard. Matches elements representing broken images."
      },
      {
        "name": ":-moz-drag-over",
        "browsers": [
          "FF1"
        ],
        "description": "Non-standard. Matches elements when a drag-over event applies to it."
      },
      {
        "name": ":-moz-first-node",
        "browsers": [
          "FF1"
        ],
        "description": "Non-standard. Represents an element that is the first child node of some other element."
      },
      {
        "name": ":-moz-focusring",
        "browsers": [
          "FF4"
        ],
        "description": "Non-standard. Matches an element that has focus and focus ring drawing is enabled in the browser."
      },
      {
        "name": ":-moz-full-screen",
        "browsers": [
          "FF9"
        ],
        "description": "Matches any element that has its fullscreen flag set. Standardized as :fullscreen."
      },
      {
        "name": ":-moz-last-node",
        "browsers": [
          "FF1"
        ],
        "description": "Non-standard. Represents an element that is the last child node of some other element."
      },
      {
        "name": ":-moz-loading",
        "browsers": [
          "FF3"
        ],
        "description": "Non-standard. Matches elements, such as images, that haven\u2019t started loading yet."
      },
      {
        "name": ":-moz-only-whitespace",
        "browsers": [
          "FF1"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-only-whitespace"
          }
        ],
        "description": "The same as :empty, except that it additionally matches elements that only contain code points affected by whitespace processing. Standardized as :blank."
      },
      {
        "name": ":-moz-placeholder",
        "browsers": [
          "FF4"
        ],
        "description": "Deprecated. Represents placeholder text in an input field. Use ::-moz-placeholder for Firefox 19+."
      },
      {
        "name": ":-moz-submit-invalid",
        "browsers": [
          "FF4"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-submit-invalid"
          }
        ],
        "description": "Non-standard. Represents any submit button when the contents of the associated form are not valid."
      },
      {
        "name": ":-moz-suppressed",
        "browsers": [
          "FF3"
        ],
        "description": "Non-standard. Matches elements representing images that have been blocked from loading."
      },
      {
        "name": ":-moz-ui-invalid",
        "browsers": [
          "FF4"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-ui-invalid"
          }
        ],
        "description": "Non-standard. Represents any validated form element whose value isn't valid "
      },
      {
        "name": ":-moz-ui-valid",
        "browsers": [
          "FF4"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-ui-valid"
          }
        ],
        "description": "Non-standard. Represents any validated form element whose value is valid "
      },
      {
        "name": ":-moz-user-disabled",
        "browsers": [
          "FF3"
        ],
        "description": "Non-standard. Matches elements representing images that have been disabled due to the user\u2019s preferences."
      },
      {
        "name": ":-moz-window-inactive",
        "browsers": [
          "FF4"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:-moz-window-inactive"
          }
        ],
        "description": "Non-standard. Matches elements in an inactive window."
      },
      {
        "name": ":-ms-fullscreen",
        "browsers": [
          "IE11"
        ],
        "description": "Matches any element that has its fullscreen flag set."
      },
      {
        "name": ":-ms-input-placeholder",
        "browsers": [
          "IE10"
        ],
        "description": "Represents placeholder text in an input field. Note: for Edge use the pseudo-element ::-ms-input-placeholder. Standardized as ::placeholder."
      },
      {
        "name": ":-ms-keyboard-active",
        "browsers": [
          "IE10"
        ],
        "description": "Windows Store apps only. Applies one or more styles to an element when it has focus and the user presses the space bar."
      },
      {
        "name": ":-ms-lang()",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents an element that is in the language specified. Accepts a comma separated list of language tokens."
      },
      {
        "name": ":no-button",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to track pieces. Applies when there is no button at that end of the track."
      },
      {
        "name": ":not()",
        "browsers": [
          "E",
          "C",
          "FF1",
          "IE9",
          "O9.5",
          "S2"
        ],
        "description": "The negation pseudo-class, :not(X), is a functional notation taking a simple selector (excluding the negation pseudo-class itself) as an argument. It represents an element that is not represented by its argument."
      },
      {
        "name": ":nth-child()",
        "browsers": [
          "E",
          "C",
          "FF3.5",
          "IE9",
          "O9.5",
          "S3.1"
        ],
        "description": "Represents an element that has an+b-1 siblings before it in the document tree, for any positive integer or zero value of n, and has a parent element."
      },
      {
        "name": ":nth-last-child()",
        "browsers": [
          "E",
          "C",
          "FF3.5",
          "IE9",
          "O9.5",
          "S3.1"
        ],
        "description": "Represents an element that has an+b-1 siblings after it in the document tree, for any positive integer or zero value of n, and has a parent element."
      },
      {
        "name": ":nth-last-of-type()",
        "browsers": [
          "E",
          "C",
          "FF3.5",
          "IE9",
          "O9.5",
          "S3.1"
        ],
        "description": "Represents an element that has an+b-1 siblings with the same expanded element name after it in the document tree, for any zero or positive integer value of n, and has a parent element."
      },
      {
        "name": ":nth-of-type()",
        "browsers": [
          "E",
          "C",
          "FF3.5",
          "IE9",
          "O9.5",
          "S3.1"
        ],
        "description": "Represents an element that has an+b-1 siblings with the same expanded element name before it in the document tree, for any zero or positive integer value of n, and has a parent element."
      },
      {
        "name": ":only-child",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:only-child"
          }
        ],
        "description": "Represents an element that has a parent element and whose parent element has no other element children. Same as :first-child:last-child or :nth-child(1):nth-last-child(1), but with a lower specificity."
      },
      {
        "name": ":only-of-type",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:only-of-type"
          }
        ],
        "description": "Matches every element that is the only child of its type, of its parent. Same as :first-of-type:last-of-type or :nth-of-type(1):nth-last-of-type(1), but with a lower specificity."
      },
      {
        "name": ":optional",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:optional"
          }
        ],
        "description": "A form element is :required or :optional if a value for it is, respectively, required or optional before the form it belongs to is submitted. Elements that are not form elements are neither required nor optional."
      },
      {
        "name": ":out-of-range",
        "browsers": [
          "E13",
          "FF29",
          "S5.1",
          "C10",
          "O11"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:out-of-range"
          }
        ],
        "description": "Used in conjunction with the min and max attributes, whether on a range input, a number field, or any other types that accept those attributes."
      },
      {
        "name": ":past",
        "browsers": [
          "C",
          "O16",
          "S6"
        ],
        "description": "Represents any element that is defined to occur entirely prior to a :current element."
      },
      {
        "name": ":read-only",
        "browsers": [
          "E13",
          "FF78",
          "S4",
          "C1",
          "O9"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:read-only"
          }
        ],
        "description": "An element whose contents are not user-alterable is :read-only. However, elements whose contents are user-alterable (such as text input fields) are considered to be in a :read-write state. In typical documents, most elements are :read-only."
      },
      {
        "name": ":read-write",
        "browsers": [
          "E13",
          "FF78",
          "S4",
          "C1",
          "O9"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:read-write"
          }
        ],
        "description": "An element whose contents are not user-alterable is :read-only. However, elements whose contents are user-alterable (such as text input fields) are considered to be in a :read-write state. In typical documents, most elements are :read-only."
      },
      {
        "name": ":required",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:required"
          }
        ],
        "description": "A form element is :required or :optional if a value for it is, respectively, required or optional before the form it belongs to is submitted. Elements that are not form elements are neither required nor optional."
      },
      {
        "name": ":right",
        "browsers": [
          "E12",
          "S5.1",
          "C6",
          "IE8",
          "O9.2"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:right"
          }
        ],
        "description": "When printing double-sided documents, the page boxes on left and right pages may be different. This can be expressed through CSS pseudo-classes defined in the  page context."
      },
      {
        "name": ":root",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:root"
          }
        ],
        "description": "Represents an element that is the root of the document. In HTML 4, this is always the HTML element."
      },
      {
        "name": ":scope",
        "browsers": [
          "E79",
          "FF32",
          "S7",
          "C27",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:scope"
          }
        ],
        "description": "Represents any element that is in the contextual reference element set."
      },
      {
        "name": ":single-button",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Applies when both buttons are displayed separately at either end of the scrollbar."
      },
      {
        "name": ":start",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to buttons and track pieces. Indicates whether the object is placed before the thumb."
      },
      {
        "name": ":target",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:target"
          }
        ],
        "description": "Some URIs refer to a location within a resource. This kind of URI ends with a 'number sign' (#) followed by an anchor identifier (called the fragment identifier)."
      },
      {
        "name": ":valid",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:valid"
          }
        ],
        "description": "An element is :valid or :invalid when it is, respectively, valid or invalid with respect to data validity semantics defined by a different specification."
      },
      {
        "name": ":vertical",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Non-standard. Applies to any scrollbar pieces that have a vertical orientation."
      },
      {
        "name": ":visited",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:visited"
          }
        ],
        "description": "Applies once the link has been visited by the user."
      },
      {
        "name": ":-webkit-any()",
        "browsers": [
          "C",
          "S5"
        ],
        "description": "Represents an element that is represented by the selector list passed as its argument. Standardized as :matches()."
      },
      {
        "name": ":-webkit-full-screen",
        "browsers": [
          "C",
          "S6"
        ],
        "description": "Matches any element that has its fullscreen flag set. Standardized as :fullscreen."
      },
      {
        "name": ":window-inactive",
        "browsers": [
          "C",
          "S3"
        ],
        "description": "Non-standard. Applies to all scrollbar pieces. Indicates whether or not the window containing the scrollbar is currently active."
      },
      {
        "name": ":blank",
        "status": "experimental",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:blank"
          }
        ],
        "description": "The :blank CSS pseudo-class selects empty user input elements (eg. <input> or <textarea>)."
      },
      {
        "name": ":defined",
        "status": "experimental",
        "browsers": [
          "E79",
          "FF63",
          "S10",
          "C54",
          "O41"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:defined"
          }
        ],
        "description": "The :defined CSS pseudo-class represents any element that has been defined. This includes any standard element built in to the browser, and custom elements that have been successfully defined (i.e. with the CustomElementRegistry.define() method)."
      },
      {
        "name": ":dir",
        "browsers": [
          "FF49"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:dir"
          }
        ],
        "description": "The :dir() CSS pseudo-class matches elements based on the directionality of the text contained in them."
      },
      {
        "name": ":focus-visible",
        "status": "experimental",
        "browsers": [
          "E79",
          "FF4",
          "C67",
          "O54"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:focus-visible"
          }
        ],
        "description": "The :focus-visible pseudo-class applies while an element matches the :focus pseudo-class and the UA determines via heuristics that the focus should be made evident on the element."
      },
      {
        "name": ":focus-within",
        "status": "experimental",
        "browsers": [
          "E79",
          "FF52",
          "S10.1",
          "C60",
          "O47"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:focus-within"
          }
        ],
        "description": "The :focus-within pseudo-class applies to any element for which the :focus pseudo class applies as well as to an element whose descendant in the flat tree (including non-element nodes, such as text nodes) matches the conditions for matching :focus."
      },
      {
        "name": ":has",
        "status": "experimental",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:has"
          }
        ],
        "description": ":The :has() CSS pseudo-class represents an element if any of the selectors passed as parameters (relative to the :scope of the given element), match at least one element."
      },
      {
        "name": ":is",
        "status": "experimental",
        "browsers": [
          "E79",
          "FF78",
          "S9",
          "C68",
          "O55"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:is"
          }
        ],
        "description": "The :is() CSS pseudo-class function takes a selector list as its argument, and selects any element that can be selected by one of the selectors in that list. This is useful for writing large selectors in a more compact form."
      },
      {
        "name": ":placeholder-shown",
        "status": "experimental",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:placeholder-shown"
          }
        ],
        "description": "The :placeholder-shown CSS pseudo-class represents any <input> or <textarea> element that is currently displaying placeholder text."
      },
      {
        "name": ":where",
        "status": "experimental",
        "browsers": [
          "FF78",
          "C72"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/:where"
          }
        ],
        "description": "The :where() CSS pseudo-class function takes a selector list as its argument, and selects any element that can be selected by one of the selectors in that list."
      }
    ],
    "pseudoElements": [
      {
        "name": "::after",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::after"
          }
        ],
        "description": "Represents a styleable child pseudo-element immediately after the originating element\u2019s actual content."
      },
      {
        "name": "::backdrop",
        "browsers": [
          "E12",
          "FF47",
          "C37",
          "IE11",
          "O24"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::backdrop"
          }
        ],
        "description": "Used to create a backdrop that hides the underlying document for an element in a top layer (such as an element that is displayed fullscreen)."
      },
      {
        "name": "::before",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::before"
          }
        ],
        "description": "Represents a styleable child pseudo-element immediately before the originating element\u2019s actual content."
      },
      {
        "name": "::content",
        "browsers": [
          "C35",
          "O22"
        ],
        "description": "Deprecated. Matches the distribution list itself, on elements that have one. Use ::slotted for forward compatibility."
      },
      {
        "name": "::cue",
        "browsers": [
          "E79",
          "FF55",
          "S6.1",
          "C26",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::cue"
          }
        ]
      },
      {
        "name": "::cue()",
        "browsers": [
          "C",
          "O16",
          "S6"
        ]
      },
      {
        "name": "::cue-region",
        "browsers": [
          "C",
          "O16",
          "S6"
        ]
      },
      {
        "name": "::cue-region()",
        "browsers": [
          "C",
          "O16",
          "S6"
        ]
      },
      {
        "name": "::first-letter",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::first-letter"
          }
        ],
        "description": "Represents the first letter of an element, if it is not preceded by any other content (such as images or inline tables) on its line."
      },
      {
        "name": "::first-line",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::first-line"
          }
        ],
        "description": "Describes the contents of the first formatted line of its originating element."
      },
      {
        "name": "::-moz-focus-inner",
        "browsers": [
          "FF4"
        ]
      },
      {
        "name": "::-moz-focus-outer",
        "browsers": [
          "FF4"
        ]
      },
      {
        "name": "::-moz-list-bullet",
        "browsers": [
          "FF1"
        ],
        "description": "Used to style the bullet of a list element. Similar to the standardized ::marker."
      },
      {
        "name": "::-moz-list-number",
        "browsers": [
          "FF1"
        ],
        "description": "Used to style the numbers of a list element. Similar to the standardized ::marker."
      },
      {
        "name": "::-moz-placeholder",
        "browsers": [
          "FF19"
        ],
        "description": "Represents placeholder text in an input field"
      },
      {
        "name": "::-moz-progress-bar",
        "browsers": [
          "FF9"
        ],
        "description": "Represents the bar portion of a progress bar."
      },
      {
        "name": "::-moz-selection",
        "browsers": [
          "FF1"
        ],
        "description": "Represents the portion of a document that has been highlighted by the user."
      },
      {
        "name": "::-ms-backdrop",
        "browsers": [
          "IE11"
        ],
        "description": "Used to create a backdrop that hides the underlying document for an element in a top layer (such as an element that is displayed fullscreen)."
      },
      {
        "name": "::-ms-browse",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the browse button of an input type=file control."
      },
      {
        "name": "::-ms-check",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the check of a checkbox or radio button input control."
      },
      {
        "name": "::-ms-clear",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the clear button of a text input control"
      },
      {
        "name": "::-ms-expand",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the drop-down button of a select control."
      },
      {
        "name": "::-ms-fill",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the bar portion of a progress bar."
      },
      {
        "name": "::-ms-fill-lower",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the portion of the slider track from its smallest value up to the value currently selected by the thumb. In a left-to-right layout, this is the portion of the slider track to the left of the thumb."
      },
      {
        "name": "::-ms-fill-upper",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the portion of the slider track from the value currently selected by the thumb up to the slider's largest value. In a left-to-right layout, this is the portion of the slider track to the right of the thumb."
      },
      {
        "name": "::-ms-reveal",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the password reveal button of an input type=password control."
      },
      {
        "name": "::-ms-thumb",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the portion of range input control (also known as a slider control) that the user drags."
      },
      {
        "name": "::-ms-ticks-after",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the tick marks of a slider that begin just after the thumb and continue up to the slider's largest value. In a left-to-right layout, these are the ticks to the right of the thumb."
      },
      {
        "name": "::-ms-ticks-before",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the tick marks of a slider that represent its smallest values up to the value currently selected by the thumb. In a left-to-right layout, these are the ticks to the left of the thumb."
      },
      {
        "name": "::-ms-tooltip",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the tooltip of a slider (input type=range)."
      },
      {
        "name": "::-ms-track",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the track of a slider."
      },
      {
        "name": "::-ms-value",
        "browsers": [
          "E",
          "IE10"
        ],
        "description": "Represents the content of a text or password input control, or a select control."
      },
      {
        "name": "::selection",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::selection"
          }
        ],
        "description": "Represents the portion of a document that has been highlighted by the user."
      },
      {
        "name": "::shadow",
        "browsers": [
          "C35",
          "O22"
        ],
        "description": "Matches the shadow root if an element has a shadow tree."
      },
      {
        "name": "::-webkit-file-upload-button",
        "browsers": [
          "E79",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-file-upload-button"
          }
        ]
      },
      {
        "name": "::-webkit-inner-spin-button",
        "browsers": [
          "E79",
          "S5",
          "C6",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-inner-spin-button"
          }
        ]
      },
      {
        "name": "::-webkit-input-placeholder",
        "browsers": [
          "C",
          "S4"
        ]
      },
      {
        "name": "::-webkit-keygen-select",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-meter-bar",
        "browsers": [
          "E79",
          "S5.1",
          "C12",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-meter-bar"
          }
        ]
      },
      {
        "name": "::-webkit-meter-even-less-good-value",
        "browsers": [
          "E79",
          "S5.1",
          "C12",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-meter-even-less-good-value"
          }
        ]
      },
      {
        "name": "::-webkit-meter-optimum-value",
        "browsers": [
          "E79",
          "S5.1",
          "C12",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-meter-optimum-value"
          }
        ]
      },
      {
        "name": "::-webkit-meter-suboptimum-value",
        "browsers": [
          "E79",
          "S5.1",
          "C12",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-meter-suboptimum-value"
          }
        ]
      },
      {
        "name": "::-webkit-outer-spin-button",
        "browsers": [
          "S5",
          "C6"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-outer-spin-button"
          }
        ]
      },
      {
        "name": "::-webkit-progress-bar",
        "browsers": [
          "E79",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-progress-bar"
          }
        ]
      },
      {
        "name": "::-webkit-progress-inner-element",
        "browsers": [
          "E79",
          "S6.1",
          "C23",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-progress-inner-element"
          }
        ]
      },
      {
        "name": "::-webkit-progress-value",
        "browsers": [
          "E79",
          "S6.1",
          "C25",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-progress-value"
          }
        ]
      },
      {
        "name": "::-webkit-resizer",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar-button",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar-corner",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar-thumb",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar-track",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-scrollbar-track-piece",
        "browsers": [
          "E79",
          "S4",
          "C2",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-scrollbar"
          }
        ]
      },
      {
        "name": "::-webkit-search-cancel-button",
        "browsers": [
          "E79",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-search-cancel-button"
          }
        ]
      },
      {
        "name": "::-webkit-search-decoration",
        "browsers": [
          "C",
          "S4"
        ]
      },
      {
        "name": "::-webkit-search-results-button",
        "browsers": [
          "E79",
          "S3",
          "C1",
          "O15"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-webkit-search-results-button"
          }
        ]
      },
      {
        "name": "::-webkit-search-results-decoration",
        "browsers": [
          "C",
          "S4"
        ]
      },
      {
        "name": "::-webkit-slider-runnable-track",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-slider-thumb",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-textfield-decoration-container",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble-arrow",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble-arrow-clipper",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble-heading",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble-message",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-webkit-validation-bubble-text-block",
        "browsers": [
          "C",
          "O",
          "S6"
        ]
      },
      {
        "name": "::-moz-range-progress",
        "status": "nonstandard",
        "browsers": [
          "FF22"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-moz-range-progress"
          }
        ],
        "description": 'The ::-moz-range-progress CSS pseudo-element is a Mozilla extension that represents the lower portion of the track (i.e., groove) in which the indicator slides in an <input> of type="range". This portion corresponds to values lower than the value currently selected by the thumb (i.e., virtual knob).'
      },
      {
        "name": "::-moz-range-thumb",
        "status": "nonstandard",
        "browsers": [
          "FF21"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-moz-range-thumb"
          }
        ],
        "description": `The ::-moz-range-thumb CSS pseudo-element is a Mozilla extension that represents the thumb (i.e., virtual knob) of an <input> of type="range". The user can move the thumb along the input's track to alter its numerical value.`
      },
      {
        "name": "::-moz-range-track",
        "status": "nonstandard",
        "browsers": [
          "FF21"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::-moz-range-track"
          }
        ],
        "description": 'The ::-moz-range-track CSS pseudo-element is a Mozilla extension that represents the track (i.e., groove) in which the indicator slides in an <input> of type="range".'
      },
      {
        "name": "::-webkit-progress-inner-value",
        "status": "nonstandard",
        "description": "The ::-webkit-progress-value CSS pseudo-element represents the filled-in portion of the bar of a <progress> element. It is a child of the ::-webkit-progress-bar pseudo-element.\n\nIn order to let ::-webkit-progress-value take effect, -webkit-appearance needs to be set to none on the <progress> element."
      },
      {
        "name": "::grammar-error",
        "status": "experimental",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::grammar-error"
          }
        ],
        "description": "The ::grammar-error CSS pseudo-element represents a text segment which the user agent has flagged as grammatically incorrect."
      },
      {
        "name": "::marker",
        "browsers": [
          "E80",
          "FF68",
          "S11.1",
          "C80"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::marker"
          }
        ],
        "description": "The ::marker CSS pseudo-element selects the marker box of a list item, which typically contains a bullet or number. It works on any element or pseudo-element set to display: list-item, such as the <li> and <summary> elements."
      },
      {
        "name": "::part",
        "status": "experimental",
        "browsers": [
          "E79",
          "FF72",
          "S13.1",
          "C73",
          "O60"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::part"
          }
        ],
        "description": "The ::part CSS pseudo-element represents any element within a shadow tree that has a matching part attribute."
      },
      {
        "name": "::placeholder",
        "browsers": [
          "E12",
          "FF51",
          "S10.1",
          "C57",
          "O44"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::placeholder"
          }
        ],
        "description": "The ::placeholder CSS pseudo-element represents the placeholder text of a form element."
      },
      {
        "name": "::slotted",
        "browsers": [
          "E79",
          "FF63",
          "S10",
          "C50",
          "O37"
        ],
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::slotted"
          }
        ],
        "description": "The :slotted() CSS pseudo-element represents any element that has been placed into a slot inside an HTML template."
      },
      {
        "name": "::spelling-error",
        "status": "experimental",
        "references": [
          {
            "name": "MDN Reference",
            "url": "https://developer.mozilla.org/docs/Web/CSS/::spelling-error"
          }
        ],
        "description": "The ::spelling-error CSS pseudo-element represents a text segment which the user agent has flagged as incorrectly spelled."
      }
    ]
  };

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/languageFacts/dataProvider.js
  var CSSDataProvider = function() {
    function CSSDataProvider2(data) {
      this._properties = [];
      this._atDirectives = [];
      this._pseudoClasses = [];
      this._pseudoElements = [];
      this.addData(data);
    }
    CSSDataProvider2.prototype.provideProperties = function() {
      return this._properties;
    };
    CSSDataProvider2.prototype.provideAtDirectives = function() {
      return this._atDirectives;
    };
    CSSDataProvider2.prototype.providePseudoClasses = function() {
      return this._pseudoClasses;
    };
    CSSDataProvider2.prototype.providePseudoElements = function() {
      return this._pseudoElements;
    };
    CSSDataProvider2.prototype.addData = function(data) {
      if (Array.isArray(data.properties)) {
        for (var _i = 0, _a3 = data.properties; _i < _a3.length; _i++) {
          var prop = _a3[_i];
          if (isPropertyData(prop)) {
            this._properties.push(prop);
          }
        }
      }
      if (Array.isArray(data.atDirectives)) {
        for (var _b = 0, _c = data.atDirectives; _b < _c.length; _b++) {
          var prop = _c[_b];
          if (isAtDirective(prop)) {
            this._atDirectives.push(prop);
          }
        }
      }
      if (Array.isArray(data.pseudoClasses)) {
        for (var _d = 0, _e = data.pseudoClasses; _d < _e.length; _d++) {
          var prop = _e[_d];
          if (isPseudoClassData(prop)) {
            this._pseudoClasses.push(prop);
          }
        }
      }
      if (Array.isArray(data.pseudoElements)) {
        for (var _f2 = 0, _g = data.pseudoElements; _f2 < _g.length; _f2++) {
          var prop = _g[_f2];
          if (isPseudoElementData(prop)) {
            this._pseudoElements.push(prop);
          }
        }
      }
    };
    return CSSDataProvider2;
  }();
  function isPropertyData(d) {
    return typeof d.name === "string";
  }
  function isAtDirective(d) {
    return typeof d.name === "string";
  }
  function isPseudoClassData(d) {
    return typeof d.name === "string";
  }
  function isPseudoElementData(d) {
    return typeof d.name === "string";
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/languageFacts/dataManager.js
  var CSSDataManager = function() {
    function CSSDataManager2(options) {
      this.dataProviders = [];
      this._propertySet = {};
      this._atDirectiveSet = {};
      this._pseudoClassSet = {};
      this._pseudoElementSet = {};
      this._properties = [];
      this._atDirectives = [];
      this._pseudoClasses = [];
      this._pseudoElements = [];
      this.setDataProviders((options === null || options === void 0 ? void 0 : options.useDefaultDataProvider) !== false, (options === null || options === void 0 ? void 0 : options.customDataProviders) || []);
    }
    CSSDataManager2.prototype.setDataProviders = function(builtIn, providers) {
      var _a3;
      this.dataProviders = [];
      if (builtIn) {
        this.dataProviders.push(new CSSDataProvider(cssData));
      }
      (_a3 = this.dataProviders).push.apply(_a3, providers);
      this.collectData();
    };
    CSSDataManager2.prototype.collectData = function() {
      var _this = this;
      this._propertySet = {};
      this._atDirectiveSet = {};
      this._pseudoClassSet = {};
      this._pseudoElementSet = {};
      this.dataProviders.forEach(function(provider) {
        provider.provideProperties().forEach(function(p) {
          if (!_this._propertySet[p.name]) {
            _this._propertySet[p.name] = p;
          }
        });
        provider.provideAtDirectives().forEach(function(p) {
          if (!_this._atDirectiveSet[p.name]) {
            _this._atDirectiveSet[p.name] = p;
          }
        });
        provider.providePseudoClasses().forEach(function(p) {
          if (!_this._pseudoClassSet[p.name]) {
            _this._pseudoClassSet[p.name] = p;
          }
        });
        provider.providePseudoElements().forEach(function(p) {
          if (!_this._pseudoElementSet[p.name]) {
            _this._pseudoElementSet[p.name] = p;
          }
        });
      });
      this._properties = values(this._propertySet);
      this._atDirectives = values(this._atDirectiveSet);
      this._pseudoClasses = values(this._pseudoClassSet);
      this._pseudoElements = values(this._pseudoElementSet);
    };
    CSSDataManager2.prototype.getProperty = function(name) {
      return this._propertySet[name];
    };
    CSSDataManager2.prototype.getAtDirective = function(name) {
      return this._atDirectiveSet[name];
    };
    CSSDataManager2.prototype.getPseudoClass = function(name) {
      return this._pseudoClassSet[name];
    };
    CSSDataManager2.prototype.getPseudoElement = function(name) {
      return this._pseudoElementSet[name];
    };
    CSSDataManager2.prototype.getProperties = function() {
      return this._properties;
    };
    CSSDataManager2.prototype.getAtDirectives = function() {
      return this._atDirectives;
    };
    CSSDataManager2.prototype.getPseudoClasses = function() {
      return this._pseudoClasses;
    };
    CSSDataManager2.prototype.getPseudoElements = function() {
      return this._pseudoElements;
    };
    CSSDataManager2.prototype.isKnownProperty = function(name) {
      return name.toLowerCase() in this._propertySet;
    };
    CSSDataManager2.prototype.isStandardProperty = function(name) {
      return this.isKnownProperty(name) && (!this._propertySet[name.toLowerCase()].status || this._propertySet[name.toLowerCase()].status === "standard");
    };
    return CSSDataManager2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/cssSelectionRange.js
  function getSelectionRanges(document, positions, stylesheet) {
    function getSelectionRange(position) {
      var applicableRanges = getApplicableRanges(position);
      var current = void 0;
      for (var index = applicableRanges.length - 1; index >= 0; index--) {
        current = SelectionRange.create(Range2.create(document.positionAt(applicableRanges[index][0]), document.positionAt(applicableRanges[index][1])), current);
      }
      if (!current) {
        current = SelectionRange.create(Range2.create(position, position));
      }
      return current;
    }
    return positions.map(getSelectionRange);
    function getApplicableRanges(position) {
      var offset = document.offsetAt(position);
      var currNode = stylesheet.findChildAtOffset(offset, true);
      if (!currNode) {
        return [];
      }
      var result = [];
      while (currNode) {
        if (currNode.parent && currNode.offset === currNode.parent.offset && currNode.end === currNode.parent.end) {
          currNode = currNode.parent;
          continue;
        }
        if (currNode.type === NodeType.Declarations) {
          if (offset > currNode.offset && offset < currNode.end) {
            result.push([currNode.offset + 1, currNode.end - 1]);
          }
        }
        result.push([currNode.offset, currNode.end]);
        currNode = currNode.parent;
      }
      return result;
    }
  }

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/services/scssNavigation.js
  var __extends11 = function() {
    var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf || { __proto__: [] } instanceof Array && function(d2, b2) {
        d2.__proto__ = b2;
      } || function(d2, b2) {
        for (var p in b2)
          if (b2.hasOwnProperty(p))
            d2[p] = b2[p];
      };
      return extendStatics(d, b);
    };
    return function(d, b) {
      extendStatics(d, b);
      function __() {
        this.constructor = d;
      }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
  }();
  var __awaiter5 = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var __generator4 = function(thisArg, body) {
    var _ = { label: 0, sent: function() {
      if (t[0] & 1)
        throw t[1];
      return t[1];
    }, trys: [], ops: [] }, f2, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() {
      return this;
    }), g;
    function verb(n) {
      return function(v) {
        return step([n, v]);
      };
    }
    function step(op) {
      if (f2)
        throw new TypeError("Generator is already executing.");
      while (_)
        try {
          if (f2 = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
            return t;
          if (y = 0, t)
            op = [op[0] & 2, t.value];
          switch (op[0]) {
            case 0:
            case 1:
              t = op;
              break;
            case 4:
              _.label++;
              return { value: op[1], done: false };
            case 5:
              _.label++;
              y = op[1];
              op = [0];
              continue;
            case 7:
              op = _.ops.pop();
              _.trys.pop();
              continue;
            default:
              if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                _ = 0;
                continue;
              }
              if (op[0] === 3 && (!t || op[1] > t[0] && op[1] < t[3])) {
                _.label = op[1];
                break;
              }
              if (op[0] === 6 && _.label < t[1]) {
                _.label = t[1];
                t = op;
                break;
              }
              if (t && _.label < t[2]) {
                _.label = t[2];
                _.ops.push(op);
                break;
              }
              if (t[2])
                _.ops.pop();
              _.trys.pop();
              continue;
          }
          op = body.call(thisArg, _);
        } catch (e) {
          op = [6, e];
          y = 0;
        } finally {
          f2 = t = 0;
        }
      if (op[0] & 5)
        throw op[1];
      return { value: op[0] ? op[1] : void 0, done: true };
    }
  };
  var SCSSNavigation = function(_super) {
    __extends11(SCSSNavigation2, _super);
    function SCSSNavigation2(fileSystemProvider) {
      return _super.call(this, fileSystemProvider) || this;
    }
    SCSSNavigation2.prototype.isRawStringDocumentLinkNode = function(node) {
      return _super.prototype.isRawStringDocumentLinkNode.call(this, node) || node.type === NodeType.Use || node.type === NodeType.Forward;
    };
    SCSSNavigation2.prototype.resolveRelativeReference = function(ref, documentUri, documentContext) {
      return __awaiter5(this, void 0, void 0, function() {
        function toPathVariations(uri) {
          if (uri.path === "") {
            return void 0;
          }
          if (uri.path.endsWith(".scss") || uri.path.endsWith(".css")) {
            return void 0;
          }
          if (uri.path.endsWith("/")) {
            return [
              uri.with({ path: uri.path + "index.scss" }).toString(),
              uri.with({ path: uri.path + "_index.scss" }).toString()
            ];
          }
          var pathFragments = uri.path.split("/");
          var basename2 = pathFragments[pathFragments.length - 1];
          var pathWithoutBasename = uri.path.slice(0, -basename2.length);
          if (basename2.startsWith("_")) {
            if (uri.path.endsWith(".scss")) {
              return void 0;
            } else {
              return [uri.with({ path: uri.path + ".scss" }).toString()];
            }
          }
          var normalizedBasename = basename2 + ".scss";
          var documentUriWithBasename = function(newBasename) {
            return uri.with({ path: pathWithoutBasename + newBasename }).toString();
          };
          var normalizedPath = documentUriWithBasename(normalizedBasename);
          var underScorePath = documentUriWithBasename("_" + normalizedBasename);
          var indexPath = documentUriWithBasename(normalizedBasename.slice(0, -5) + "/index.scss");
          var indexUnderscoreUri = documentUriWithBasename(normalizedBasename.slice(0, -5) + "/_index.scss");
          var cssPath = documentUriWithBasename(normalizedBasename.slice(0, -5) + ".css");
          return [normalizedPath, underScorePath, indexPath, indexUnderscoreUri, cssPath];
        }
        var target, parsedUri, pathVariations, j, e_1;
        return __generator4(this, function(_a3) {
          switch (_a3.label) {
            case 0:
              if (startsWith(ref, "sass:")) {
                return [2, void 0];
              }
              return [4, _super.prototype.resolveRelativeReference.call(this, ref, documentUri, documentContext)];
            case 1:
              target = _a3.sent();
              if (!(this.fileSystemProvider && target && extname2(target).length === 0))
                return [3, 8];
              _a3.label = 2;
            case 2:
              _a3.trys.push([2, 7, , 8]);
              parsedUri = URI2.parse(target);
              pathVariations = toPathVariations(parsedUri);
              if (!pathVariations)
                return [3, 6];
              j = 0;
              _a3.label = 3;
            case 3:
              if (!(j < pathVariations.length))
                return [3, 6];
              return [4, this.fileExists(pathVariations[j])];
            case 4:
              if (_a3.sent()) {
                return [2, pathVariations[j]];
              }
              _a3.label = 5;
            case 5:
              j++;
              return [3, 3];
            case 6:
              return [2, void 0];
            case 7:
              e_1 = _a3.sent();
              return [3, 8];
            case 8:
              return [2, target];
          }
        });
      });
    };
    return SCSSNavigation2;
  }(CSSNavigation);

  // node_modules/monaco-editor/esm/vs/language/css/_deps/vscode-css-languageservice/cssLanguageService.js
  function createFacade(parser, completion, hover, navigation, codeActions, validation, cssDataManager) {
    return {
      configure: function(settings) {
        validation.configure(settings);
        completion.configure(settings);
      },
      setDataProviders: cssDataManager.setDataProviders.bind(cssDataManager),
      doValidation: validation.doValidation.bind(validation),
      parseStylesheet: parser.parseStylesheet.bind(parser),
      doComplete: completion.doComplete.bind(completion),
      doComplete2: completion.doComplete2.bind(completion),
      setCompletionParticipants: completion.setCompletionParticipants.bind(completion),
      doHover: hover.doHover.bind(hover),
      findDefinition: navigation.findDefinition.bind(navigation),
      findReferences: navigation.findReferences.bind(navigation),
      findDocumentHighlights: navigation.findDocumentHighlights.bind(navigation),
      findDocumentLinks: navigation.findDocumentLinks.bind(navigation),
      findDocumentLinks2: navigation.findDocumentLinks2.bind(navigation),
      findDocumentSymbols: navigation.findDocumentSymbols.bind(navigation),
      doCodeActions: codeActions.doCodeActions.bind(codeActions),
      doCodeActions2: codeActions.doCodeActions2.bind(codeActions),
      findColorSymbols: function(d, s) {
        return navigation.findDocumentColors(d, s).map(function(s2) {
          return s2.range;
        });
      },
      findDocumentColors: navigation.findDocumentColors.bind(navigation),
      getColorPresentations: navigation.getColorPresentations.bind(navigation),
      doRename: navigation.doRename.bind(navigation),
      getFoldingRanges,
      getSelectionRanges
    };
  }
  var defaultLanguageServiceOptions = {};
  function getCSSLanguageService(options) {
    if (options === void 0) {
      options = defaultLanguageServiceOptions;
    }
    var cssDataManager = new CSSDataManager(options);
    return createFacade(new Parser(), new CSSCompletion(null, options, cssDataManager), new CSSHover(options && options.clientCapabilities, cssDataManager), new CSSNavigation(options && options.fileSystemProvider), new CSSCodeActions(cssDataManager), new CSSValidation(cssDataManager), cssDataManager);
  }
  function getSCSSLanguageService(options) {
    if (options === void 0) {
      options = defaultLanguageServiceOptions;
    }
    var cssDataManager = new CSSDataManager(options);
    return createFacade(new SCSSParser(), new SCSSCompletion(options, cssDataManager), new CSSHover(options && options.clientCapabilities, cssDataManager), new SCSSNavigation(options && options.fileSystemProvider), new CSSCodeActions(cssDataManager), new CSSValidation(cssDataManager), cssDataManager);
  }
  function getLESSLanguageService(options) {
    if (options === void 0) {
      options = defaultLanguageServiceOptions;
    }
    var cssDataManager = new CSSDataManager(options);
    return createFacade(new LESSParser(), new LESSCompletion(options, cssDataManager), new CSSHover(options && options.clientCapabilities, cssDataManager), new CSSNavigation(options && options.fileSystemProvider), new CSSCodeActions(cssDataManager), new CSSValidation(cssDataManager), cssDataManager);
  }

  // node_modules/monaco-editor/esm/vs/language/css/cssWorker.js
  var __awaiter6 = function(thisArg, _arguments, P, generator) {
    function adopt(value) {
      return value instanceof P ? value : new P(function(resolve2) {
        resolve2(value);
      });
    }
    return new (P || (P = Promise))(function(resolve2, reject) {
      function fulfilled(value) {
        try {
          step(generator.next(value));
        } catch (e) {
          reject(e);
        }
      }
      function rejected(value) {
        try {
          step(generator["throw"](value));
        } catch (e) {
          reject(e);
        }
      }
      function step(result) {
        result.done ? resolve2(result.value) : adopt(result.value).then(fulfilled, rejected);
      }
      step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
  };
  var __generator5 = function(thisArg, body) {
    var _ = { label: 0, sent: function() {
      if (t[0] & 1)
        throw t[1];
      return t[1];
    }, trys: [], ops: [] }, f2, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() {
      return this;
    }), g;
    function verb(n) {
      return function(v) {
        return step([n, v]);
      };
    }
    function step(op) {
      if (f2)
        throw new TypeError("Generator is already executing.");
      while (_)
        try {
          if (f2 = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
            return t;
          if (y = 0, t)
            op = [op[0] & 2, t.value];
          switch (op[0]) {
            case 0:
            case 1:
              t = op;
              break;
            case 4:
              _.label++;
              return { value: op[1], done: false };
            case 5:
              _.label++;
              y = op[1];
              op = [0];
              continue;
            case 7:
              op = _.ops.pop();
              _.trys.pop();
              continue;
            default:
              if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                _ = 0;
                continue;
              }
              if (op[0] === 3 && (!t || op[1] > t[0] && op[1] < t[3])) {
                _.label = op[1];
                break;
              }
              if (op[0] === 6 && _.label < t[1]) {
                _.label = t[1];
                t = op;
                break;
              }
              if (t && _.label < t[2]) {
                _.label = t[2];
                _.ops.push(op);
                break;
              }
              if (t[2])
                _.ops.pop();
              _.trys.pop();
              continue;
          }
          op = body.call(thisArg, _);
        } catch (e) {
          op = [6, e];
          y = 0;
        } finally {
          f2 = t = 0;
        }
      if (op[0] & 5)
        throw op[1];
      return { value: op[0] ? op[1] : void 0, done: true };
    }
  };
  var CSSWorker = function() {
    function CSSWorker2(ctx, createData) {
      this._ctx = ctx;
      this._languageSettings = createData.languageSettings;
      this._languageId = createData.languageId;
      switch (this._languageId) {
        case "css":
          this._languageService = getCSSLanguageService();
          break;
        case "less":
          this._languageService = getLESSLanguageService();
          break;
        case "scss":
          this._languageService = getSCSSLanguageService();
          break;
        default:
          throw new Error("Invalid language id: " + this._languageId);
      }
      this._languageService.configure(this._languageSettings);
    }
    CSSWorker2.prototype.doValidation = function(uri) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, diagnostics;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          if (document) {
            stylesheet = this._languageService.parseStylesheet(document);
            diagnostics = this._languageService.doValidation(document, stylesheet);
            return [2, Promise.resolve(diagnostics)];
          }
          return [2, Promise.resolve([])];
        });
      });
    };
    CSSWorker2.prototype.doComplete = function(uri, position) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, completions;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          completions = this._languageService.doComplete(document, position, stylesheet);
          return [2, Promise.resolve(completions)];
        });
      });
    };
    CSSWorker2.prototype.doHover = function(uri, position) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, hover;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          hover = this._languageService.doHover(document, position, stylesheet);
          return [2, Promise.resolve(hover)];
        });
      });
    };
    CSSWorker2.prototype.findDefinition = function(uri, position) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, definition;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          definition = this._languageService.findDefinition(document, position, stylesheet);
          return [2, Promise.resolve(definition)];
        });
      });
    };
    CSSWorker2.prototype.findReferences = function(uri, position) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, references;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          references = this._languageService.findReferences(document, position, stylesheet);
          return [2, Promise.resolve(references)];
        });
      });
    };
    CSSWorker2.prototype.findDocumentHighlights = function(uri, position) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, highlights;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          highlights = this._languageService.findDocumentHighlights(document, position, stylesheet);
          return [2, Promise.resolve(highlights)];
        });
      });
    };
    CSSWorker2.prototype.findDocumentSymbols = function(uri) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, symbols;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          symbols = this._languageService.findDocumentSymbols(document, stylesheet);
          return [2, Promise.resolve(symbols)];
        });
      });
    };
    CSSWorker2.prototype.doCodeActions = function(uri, range, context) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, actions;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          actions = this._languageService.doCodeActions(document, range, context, stylesheet);
          return [2, Promise.resolve(actions)];
        });
      });
    };
    CSSWorker2.prototype.findDocumentColors = function(uri) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, colorSymbols;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          colorSymbols = this._languageService.findDocumentColors(document, stylesheet);
          return [2, Promise.resolve(colorSymbols)];
        });
      });
    };
    CSSWorker2.prototype.getColorPresentations = function(uri, color, range) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, colorPresentations;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          colorPresentations = this._languageService.getColorPresentations(document, stylesheet, color, range);
          return [2, Promise.resolve(colorPresentations)];
        });
      });
    };
    CSSWorker2.prototype.getFoldingRanges = function(uri, context) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, ranges;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          ranges = this._languageService.getFoldingRanges(document, context);
          return [2, Promise.resolve(ranges)];
        });
      });
    };
    CSSWorker2.prototype.getSelectionRanges = function(uri, positions) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, ranges;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          ranges = this._languageService.getSelectionRanges(document, positions, stylesheet);
          return [2, Promise.resolve(ranges)];
        });
      });
    };
    CSSWorker2.prototype.doRename = function(uri, position, newName) {
      return __awaiter6(this, void 0, void 0, function() {
        var document, stylesheet, renames;
        return __generator5(this, function(_a3) {
          document = this._getTextDocument(uri);
          stylesheet = this._languageService.parseStylesheet(document);
          renames = this._languageService.doRename(document, position, newName, stylesheet);
          return [2, Promise.resolve(renames)];
        });
      });
    };
    CSSWorker2.prototype._getTextDocument = function(uri) {
      var models = this._ctx.getMirrorModels();
      for (var _i = 0, models_1 = models; _i < models_1.length; _i++) {
        var model = models_1[_i];
        if (model.uri.toString() === uri) {
          return TextDocument2.create(uri, this._languageId, model.version, model.getValue());
        }
      }
      return null;
    };
    return CSSWorker2;
  }();

  // node_modules/monaco-editor/esm/vs/language/css/css.worker.js
  self.onmessage = function() {
    initialize(function(ctx, createData) {
      return new CSSWorker(ctx, createData);
    });
  };
})();
